// DiffractionImage.cpp
// authors: G.Winter, F.Remacle
// maintained by F.Remacle
// 16th December 2003
// 
// A replacement for the old DiffractionImage module - this time developed
// more carefully so that it should be much easier to extend and more
// portable.
// 
// 
#include "DiffractionImage.h"

// core C things - required for using open() and read() below
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

// including the jpeg and jasper stuff in the same file causes arguments -
// so liberate one from the file - jpeg.

#ifdef INCLUDE_JASPER
extern "C" {
#include "jasper/jasper.h"
}
#endif

using namespace std;
namespace DI 
	{
    // first start on the constructors 
    DiffractionImage::DiffractionImage(void)
		{
		constructor();
		}
    
    DiffractionImage::DiffractionImage(char* filename)
		{
		constructor(filename);
		}
    
    DiffractionImage::~DiffractionImage(void)
		{
		destructor();
		}
    
    void DiffractionImage::constructor(void)
		{
		// initalise all of the data in this object
	#ifdef INCLUDE_PYTHON
		(void) new (& headerMap) map<string, string>;
		(void) new (& headerKeys) vector<string>;
	#endif
		memset(originalHeader, 0, 4096);
		originalHeaderLength = 0;
		// assume that we are setting everything to false || 0
		bigEndian = false;
		width = 0;
		height = 0;
		// these values could be set to NaN or something to 
		// raise an exception if they are used
		beamX = 0;
		beamY = 0;
		distance = 0;
		wavelength = 0;
		twoTheta = 0;
		oscStart = 0;
		oscEnd = 0;
		pixelX = 0;
		pixelY = 0;
		exposureTime = 0.0;
		warning = false;
		// need to call the constructor on message then
		(void) new (&message) string;
		(void) new (&format) string;
		(void) new (&filename) string;
		(void) new (&serialNo) string;
		(void) new (&date) string;
		rapidAxis=new float[3];
		slowAxis=new float[3];
		pixelOrigin=new float[3];
		message = "";
		oscAxis="";
		format = "";
		manufact= "";
		filename = "";
		date = "";
		serialNo = "unknown";
		image = NULL;
		realImage=true;
		}
    
    void DiffractionImage::constructor(char* fileName)
		{
		// simply initialise and then read an image - this is more
		// a convenience function than anything else
		constructor();
		load(fileName);
		}
    
    void DiffractionImage::destructor(void)
		{
		// if there is space allocated for an image, delete it
		if (image != NULL)
			{
			delete [] image;
			image = NULL;
			}
		// the advertised behaviour is to reset so
		constructor();
		}
    
    // next the data access methods - yawn!
    // define a couple of C preprocessor macros to make this easier
    
#define GETTER_METHOD(type, arg, capitalArg) \
type DiffractionImage::get ## capitalArg(void) { return arg; }
    
#define SETTER_METHOD(type, arg, capitalArg) \
void DiffractionImage::set ## capitalArg(type new ## capitalArg) \
{ arg = new ## capitalArg; }
    
    //getting methods
    GETTER_METHOD(int, height, Height);
    GETTER_METHOD(int, width, Width);
    GETTER_METHOD(float, beamX, BeamX);
    GETTER_METHOD(float, beamY, BeamY);
    GETTER_METHOD(float, wavelength, Wavelength);
    GETTER_METHOD(float, distance, Distance);
    GETTER_METHOD(float, twoTheta, TwoTheta);
    GETTER_METHOD(float, oscStart, OscStart);
    GETTER_METHOD(float, oscEnd, OscEnd);
    GETTER_METHOD(float, pixelX, PixelX);
    GETTER_METHOD(float, pixelY, PixelY);
    GETTER_METHOD(float, exposureTime, ExposureTime);
	GETTER_METHOD(bool, warning, Warning);
	GETTER_METHOD(float*, rapidAxis, R);
    GETTER_METHOD(float*, slowAxis, S);
    GETTER_METHOD(float*, pixelOrigin, P);
	GETTER_METHOD(const char*, message.c_str(), Message);
	GETTER_METHOD(const char*, oscAxis.c_str(), OscAxis);
	GETTER_METHOD(const char*, format.c_str(), Format);
	GETTER_METHOD(const char*, manufact.c_str(), Manufacturer);
	GETTER_METHOD(const char*, filename.c_str(), Filename);
	GETTER_METHOD(const char*, serialNo.c_str(), SerialNo);
	GETTER_METHOD(const char*, date.c_str(), Date);
	//settings methods
    SETTER_METHOD(float, beamX, BeamX);
    SETTER_METHOD(float, beamY, BeamY);
    SETTER_METHOD(float, wavelength, Wavelength);
    SETTER_METHOD(float, distance, Distance);
    SETTER_METHOD(float, twoTheta, TwoTheta);
    SETTER_METHOD(float, oscStart, OscStart);
    SETTER_METHOD(float, oscEnd, OscEnd);
    SETTER_METHOD(float, pixelX, PixelX);
    SETTER_METHOD(float, pixelY, PixelY);
    SETTER_METHOD(float, exposureTime, ExposureTime);
	SETTER_METHOD(string, format, Format);
    SETTER_METHOD(string, filename, Filename);
	SETTER_METHOD(char*, oscAxis, OscAxis);
	SETTER_METHOD(int, height, Height);
    SETTER_METHOD(int, width, Width);
    
    // for accessing the image I think I will have to write the code by
    // hand - Oh what a shame

    bool DiffractionImage::isOriginalImg(void)
		{
		if (image == NULL)
			{
			throw (DiffractionImageException("image is NULL"));
			}
		return realImage;
		}
	
	unsigned int* DiffractionImage::getImage(void)
		{
		if (image == NULL)
			{
			throw (DiffractionImageException("image is NULL"));
			}
		return image;
		}
		
	void DiffractionImage::setImage(unsigned int* copy)
		{
		if (image != NULL)
			{
			free(image);
			image=new unsigned int[width*height];
			}
		for(int i=0;i<width*height;i++)
			image[i]=copy[i];
		realImage=false;
		}
		
	int DiffractionImage::getPixelAt(int x,int y)
		{
		if( (x<width && y<height) && (x>=0 && y>=0) )
			return image[y*width+x];
		else
			{
			throw(DiffractionImageException("getPixelAt: Wrong coordinates"));
			return -1;
			}
		}

    unsigned int & DiffractionImage::operator[] (int offset)
		{
		// check before we go
		if (image == NULL)
			{
			throw(DiffractionImageException("NULL image"));
			}
		if ((offset < 0) ||(offset >= width * height))
			{
			throw(DiffractionImageException("out of range"));
			}
		// if we get to here then
		return image[offset];
		}

    DiffractionImage & DiffractionImage::operator= (DiffractionImage & source)
		{
		// first call the destructor to reinitialise the structure
		destructor();
		// now copy the elements from the source image
		width = source.getWidth();
		height = source.getHeight();
		beamX = source.getBeamX();
		beamY = source.getBeamY();
		distance = source.getDistance();
		wavelength = source.getWavelength();
		twoTheta = source.getTwoTheta();	
		oscStart = source.getOscStart();
		oscEnd = source.getOscEnd();
		pixelX = source.getPixelX();
		pixelY = source.getPixelY();
		oscAxis=source.getOscAxis();
		realImage=source.isOriginalImg();
		format=source.getFormat();
		filename=source.getFilename();
		manufact=source.getManufacturer();
		try	{
			unsigned int * otherImage = source.getImage();
			// allocate some space
			allocateImage();
			// copy
			for (int i = 0; i < width * height; i++)
				image[i] = otherImage[i];
			} catch (DiffractionImageException & e) {} // do nothing - the image is NULL;
		return *this;
		}


    void DiffractionImage::load(char* fileName)
		{
		filename=fileName;
		format = guessFormat(filename);
		try {
			load();
			}catch(DiffractionImageException & e) {throw(e);}
		// if we get this far with an unknown image type it is a mar
		if (format.compare("unknown") == 0)
			format = "mar";
		}

    void DiffractionImage::load()
		{
		if (format.compare("adsc") == 0)
			{
			loadADSC();
			if(format.compare("saturn") == 0)
				loadSaturn();
			else
				{
				if(format.compare("raxisd") == 0)
					loadRAXISDtrek();
				}
			}
		else
			{
			if (format.compare("raxis") == 0)
				loadRaxis();
			else
				{
				if (format.compare("cbf") == 0) 
					{
					loadCBF();
					}
				else
					{
					if (format.compare("bruker") == 0) 
						{
						loadBruker();
						}
					else
						{
						if (format.compare("oxford") == 0) 
							{
							loadOxford();
							}
						else
							{
							try {
								loadMAR();
								}catch(DiffractionImageException & e) {throw(DiffractionImageException("unknown image format"));}
							}
						}
					}
				}
			}
		}	

    void DiffractionImage::loadHeader(char* fileName)
    {
	filename=fileName;
	format = guessFormat(filename);
	try {
	    loadHeader();
		} catch(DiffractionImageException & e) {throw(e);}
	// if we get this far with an unknown image type it is a mar
	if (format.compare("unknown") == 0)
		{
	    format = "mar";
		}
    }

    void DiffractionImage::loadHeader()
		{
		if (format.compare("adsc") == 0)
			{
			loadADSCHeader();
			if(format.compare("saturn") == 0)
				loadSaturnHeader();
			else
				{
				if(format.compare("raxisd") == 0)
					loadRAXISDtrekHeader();
				}
			}
		else 
			{
			if (format.compare("raxis") == 0)
				loadRaxisHeader();
			else 
				{
				if (format.compare("cbf") == 0)
					{
					loadCBFHeader();
					}
				else
					{
					if (format.compare("bruker") == 0)
						{
						loadBrukerHeader();
						}
					else
						{
						if (format.compare("oxford") == 0)
							{
							loadOxfordHeader();
							}
						else
							{
							try {
								loadMARHeader();
								} catch(DiffractionImageException & e){throw(DiffractionImageException("unknown image format"));}
							}
						}
					}
				}
			}
		}

    void DiffractionImage::loadCompressed(string filename)
		{
		string headerName = filename;
		headerName+=".hdr";
		string imageName = filename;
		imageName+=".jp2";
		loadCompressed(headerName, imageName);
		}

    void DiffractionImage::loadCompressed(string headerName,string imageName)
		{
        char* dummy;
        strcpy(dummy,headerName.c_str());
		loadHeader(dummy);
		loadCompressedImage(imageName);
		}

    void DiffractionImage::loadCompressedImage(string imageName)
		{
#ifndef INCLUDE_JASPER
		throw(DiffractionImageException("Jpeg2000 support not included"));
#else
		// first check header
		int newWidth, newHeight;
		
		if ((width  <= 0) || (height <= 0))
			throw(DiffractionImageException("bad size"));
		// the image will be allocated when it is needed - and not before!
		jas_stream_t * in = jas_stream_fopen(imageName.c_str(), "rb");
		if (in == NULL)
		throw(DiffractionImageException("error opening compressed file"));   // there was an error opening the file
		jas_image_t * jp2 = jas_image_decode(in, jas_image_strtofmt("jp2"), 0);
		if (jp2 == NULL)
			{
			jas_stream_close(in);
			throw(DiffractionImageException("error reading jp2 image"));
			}
		jas_stream_close(in);

		// now get the image out - this should be relatively easy
		newWidth = jas_image_cmptwidth(jp2, 0);
		newHeight = jas_image_cmptheight(jp2, 0);
		if ((newHeight != height) || (newWidth != width))
			{
			// then this is the wrong image!
			jas_image_destroy(jp2);
			throw(DiffractionImageException("wrong size in image"));
			}
		// if we get to here then all's well
		jas_matrix_t * data = jas_matrix_create(width, height);
		if (data == NULL)
			{
			jas_image_destroy(jp2);
			throw(DiffractionImageException("error allocating image component"));
			}
		jas_image_readcmpt(jp2, 0, 0, 0, width, height, data);
		// next copy this information out of the matrix and into the image
		// allocate or reallocate the image if necessary
		if (image != NULL)
			{
			delete [] image;
			image = NULL;
			}
		// again two layers of memory error trapping - since we don't want any unhandled exceptions.
		try {
			image = new unsigned short[width * height];
			} catch (bad_alloc & e) 
					{
					jas_image_destroy(jp2);    
					jas_matrix_destroy(data);
					throw(DiffractionImageException("error allocating image"));
					}
		if (image == NULL)
			{
			jas_image_destroy(jp2);
			jas_matrix_destroy(data);
			throw(DiffractionImageException("error allocating image"));
			}

		for (int i = 0; i < height; i++)
			{
			for (int j = 0; j < width; j++)
				image[i * width + j] = jas_matrix_get(data, i, j);
			}
		// free the data
		jas_image_destroy(jp2);
		jas_matrix_destroy(data);
#endif
    }

    void DiffractionImage::regenerateHeader(void)
		{
		// do nothning
		}
	
	// Save method based on the image format to call corresponding saving method
/*	void DiffractionImage::save()
		{
		if (format == "adsc")
			saveADSC();
		else 
			{
			if (format == "raxis"){} //saveRaxis(); //!! currently not working well
			else 
				{
				if(format == "mar")
					saveMAR();
				else 
					{
					if (format == "cbf"){}//saveCBF();
					}
				}
			}
		}*/
};
