// Profile.cpp
// maintained by F.Remacle
// 4th Septembre 2007
// 
// The implementation of the "PeakList" class, for performing peak searches
// on images.
// 
// 
// 
// 
// $Id: Profile.cpp,v 1.9 2007/10/31 12:06:32 fr45 Exp $

#include "Profile.h"

// this is to allow some very peculiar things to be done in the 
// Python allocator
#include <new>
#include <list>
// this is only required if not GCC
#ifdef NOT_USING_GCC
#include <math>
#else
#include <math.h>
#ifdef _MSC_VER
#include <stdlib.h>
#define drand48 rand
#endif
#endif

namespace DI
	{
	// the default constructor
	Profile::Profile(void)
		{
		constructor();
		}

	Profile::Profile(DiffractionImage * diffractionImage, int fromX, int fromY, int toX, int toY)
		{
		constructor(diffractionImage,fromX,fromY,toX,toY);
		}

	Profile::~Profile(void)
		{
		destructor();
		}

	// next a little real code - with the actual constructors and 
	// destructors. this will be pretty simple as most of the work
	// will be handled by the STL.

	// however, the constructors are not called as a part of malloc,
	// which is what python allocation uses. this means that I can't
	// combine Python with STL. bummer. even trying to explicitly call
	// the constructor fails... -> change _list to a ptr to the list?

	void Profile::constructor(void)
		{
#ifdef INCLUDE_PYTHON
		// then we need to manually call the new operator
		// to construct the STL list
		(void) new (&_list) list<ProfPeak>;
#endif
		_length = 0;
		intensities=NULL;
		Xs=NULL;
		Ys=NULL;
		}

	void Profile::constructor(DiffractionImage * diffractionImage, int fromX, int fromY, int toX, int toY)
		{
		constructor();
		loadProfile(diffractionImage,fromX,fromY,toX,toY);
		}

	void Profile::destructor(void)
		{
		// call the destructor manually - is this kosher? hope so
#ifdef INCLUDE_PYTHON
		_list.~list<ProfPeak>();
#endif
		}

	int Profile::getNPeaks(void)
		{
#ifdef TEST_CODE
		// then we will check that the structure of the list is in
		// tact - but _list.size() can be an O(n) operation so it is
		// not really desirable
		if (_list.size() != _length)
			throw (DiffractionImageException("list structure is broken"));
#endif
		return _length;
		}
		
	int Profile::getLength(void)
		{
		return npixels;
		}
  
  	unsigned int* Profile::getIntensities(void)
		{
		return intensities;
		}
		
	unsigned int Profile::getMax(void)
		{
		return intMax;
		}

	void Profile::loadProfile(DiffractionImage * diffractionImage, int fromX, int fromY, int toX, int toY)
		{
		int steps,width,height,lastgrad,grad,lastindex,npix;
		float vecX,vecY,norm,beamX,beamY,pixX,pixY,dist,waveleng,theta;
		float* autocorrelate;
		
		vecX=toX-fromX;
		vecY=toY-fromY;
		norm=sqrt(vecX*vecX+vecY*vecY);
		steps=(int) norm;
		vecX/=norm;
		vecY/=norm;
		lastindex=-1;
		npix=0;
		intMax=0;
		width=diffractionImage->getWidth();
		//first loop is to evaluate how many pixels there will be on the line
		for(int i=0;i<steps;i++)
			{
			int ivecY,ivecX,index;
			if((i*vecX)-floor(i*vecX)>0.5)
				ivecX=(int) ceil(i*vecX);
			else
				ivecX=(int) floor(i*vecX);
			if((i*vecY)-floor(i*vecY)>0.5)
				ivecY=(int) ceil(i*vecY);
			else
				ivecY=(int) floor(i*vecY);
			index=((fromY+ivecY)*width)+(fromX+ivecX);
			if(index!=lastindex)
				{
				lastindex=index;
				npix++;
				}
			}
		intensities=new unsigned int[npix];
		Xs=new int[npix];
		Ys=new int[npix];
		npixels=npix;
		npix=0;
		//second loop is to actually load these pixels in the profile
		for(int i=0;i<steps;i++)
			{
			int ivecY,ivecX,index;
			if((i*vecX)-floor(i*vecX)>0.5)
				ivecX=(int) ceil(i*vecX);
			else
				ivecX=(int) floor(i*vecX);
			if((i*vecY)-floor(i*vecY)>0.5)
				ivecY=(int) ceil(i*vecY);
			else
				ivecY=(int) floor(i*vecY);
			index=((fromY+ivecY)*width)+(fromX+ivecX);
			if(index!=lastindex)
				{
				Xs[npix]=fromX+ivecX;
				Ys[npix]=fromY+ivecY;
				intensities[npix]=diffractionImage->getPixelAt(fromX+ivecX,fromY+ivecY);
				if(intMax<intensities[npix])
					intMax=intensities[npix];
				lastindex=index;
				npix++;
				}
			}
		/*autocorrelate=new float[npix];
		//third loop is to autocorrelate the signal
		for(int i=0;i<npix;i++)
			{
			float item=0;
			for(int j=0;j<npix;j++)
				item+=intensities[j]*cos((M_PI/(npix*1.0f))*(j+0.5f)*i);
			autocorrelate[i]=item*item;
			printf("%.5f %d\n",autocorrelate[i],intensities[i]);
			if(intMax<autocorrelate[i])
				intMax=autocorrelate[i];
			}*/
			
		//Third loop is to look for peaks and calculate distance between peaks
		//second loop is to actually load these pixels in the profile
		lastgrad=intensities[1]-intensities[0];
		pixX=diffractionImage->getPixelX();
		pixY=diffractionImage->getPixelY();
		beamX=diffractionImage->getBeamX();
		beamY=diffractionImage->getBeamY();
		dist=diffractionImage->getDistance();
		theta=diffractionImage->getTwoTheta()*(M_PI/180.0f);
		waveleng=diffractionImage->getWavelength();
			
		for(int i=2;i<npix;i++)
			{
			grad=intensities[i]-intensities[i-1];
			if(lastgrad>0 && grad<=0)
				{
				if(intensities[i-1]>intMax/8)
					{
					if(_length >0 && (i-1)-_list.back().indInt < 5)
						{
						if(intensities[i-1] > _list.back().intensity)
							{
							_list.pop_back();
							_length--;
							}
						}
					else
						{
						ProfPeak peak;
						peak.posX = Xs[i-1];
						peak.posY = Ys[i-1];
						peak.indInt = i-1;
						if(_length==0)
							peak.distFromPrevious = 0;
						else
							{
							float normV1, normV2, norm2,Xcoord,Ycoord,Zcoord;
							float* V1;
							float* V2;
							double alpha;
							ProfPeak p=get(_length-1);
							V1=new float[3];
							V1[0]=p.posX*pixX-beamX;
							V1[1]=((p.posY*pixY-beamY)*cos(theta))+dist*sin(theta);
							V1[2]=((p.posY*pixY-beamY)*sin(theta))-dist*cos(theta);
							normV1=sqrt(V1[0]*V1[0]+V1[1]*V1[1]+V1[2]*V1[2]);
							V2=new float[3];
							V2[0]=Xs[i-1]*pixX-beamX;
							V2[1]=((Ys[i-1]*pixY-beamY)*cos(theta))+dist*sin(theta);
							V2[2]=((Ys[i-1]*pixY-beamY)*sin(theta))-dist*cos(theta);
							normV2=sqrt(V2[0]*V2[0]+V2[1]*V2[1]+V2[2]*V2[2]);
							V1[0]/=normV1;
							V1[1]/=normV1;
							V1[2]/=normV1;
							V1[2]+=1;
							V2[0]/=normV2;
							V2[1]/=normV2;
							V2[2]/=normV2;
							V2[2]+=1;
							Xcoord=V1[0]-V2[0];
							Ycoord=V1[1]-V2[1];
							Zcoord=V1[2]-V2[2];
							norm2=Xcoord*Xcoord+Ycoord*Ycoord+Zcoord*Zcoord;
							peak.distFromPrevious=waveleng/sqrt(norm2);
							}
						peak.intensity = intensities[i-1];
						_list.push_back(peak);
						_length ++;
						}
					}
				}
			lastgrad=grad;
			}
		}

	ProfPeak & Profile::get(int offset)
		{
		int n;
		list<ProfPeak>::iterator i;
		if (offset > _length)
			throw (DiffractionImageException("Profile: offset greater than length"));
		n = 0;
		for (i = _list.begin(); i != _list.end(); i++)
			{
			if (n == offset)
				return (*i);
			n ++;
			}
		return (* (_list.end())); // this is just to keep the compiler happy
		}

	ProfPeak* Profile::getPeaks(void)
		{
		ProfPeak* result;
		list<ProfPeak>::iterator i;
		int j=0;
		for (i = _list.begin(); i != _list.end(); i++)
			{
			result[j]=(*i);
			j++;
			}
		return result;
		}

	};
