// 
// Copyright  2001-2005 Rigaku/MSC, Inc.
//                       9009 New Trails Drive  
//                       The Woodlands, TX, USA  77381  
// 
// The contents are unpublished proprietary source  
// code of Rigaku/MSC, Inc.  
// 
// All rights reserved   
// 
// ReadRAXISImage.cpp    Initial author: T.L.Hendrixson           Aug 2001
//
//
// Description:
//
// This program reads in an R-AXIS image written using the Rigaku-style
// format and writes the values of a few of the pixels to standard output.
// It additionally writes out some of the information contained in the
// header, such as axis positions, exposure time, pixel information.
//
// This program is primarily intended for illustrative purposes, and so
// there have been no attempts made to optimize either the reads or the
// byte swapping used to get the data in the correct endianness for the
// computer being used.
//
// Although the routines used for reading the image suggest that an RAXIS
// class should be developed, that is left as an exercise for the reader.
//
// This program assumes that the image was written using either SGI Control
// software or the instrument server provided with CrystalClear.  In these
// programs, the data is written in the file so that it is big-endian.
// 
// 
// Calling syntax:
//
// ReadRAXISImage.exe filename
//
// where filename is the name of the image file to be read.
//

/****************************************************************************
 *                              Include Files                               *
 ****************************************************************************/

#include "RaxisImage.h"

/****************************************************************************
 *                                Constants                                 *
 ****************************************************************************/

const int MinimumRigakuHeaderLength = 1400;

std::string ERROR_OPENING_FILE   = "ERROR Can not open file: ";
std::string ERROR_READING_HEADER = "ERROR Can not read Rigaku-style header.";
std::string ERROR_READING_IMAGE  = "ERROR Can not read image data.";

/****************************************************************************
 ****************************************************************************/
std::string ProgramUsage(void)
{
   std::string s = "ReadRAXISImage.exe image_filename\n\n";
   s += "image_filename : Filename of image to be read.\n";

   return s;
}

/****************************************************************************
 ****************************************************************************/
bool ReadHeader(std::ifstream &InputFile, 
                RigakuHeader &Header)
{
   // Read in the minimal amount of header information.

   InputFile.read(reinterpret_cast<char *>(&Header),MinimumRigakuHeaderLength);

   if(InputFile.bad() || InputFile.fail())
      return false;

   // Check to see if endianness of the header matches endianness of this
   // architecture.  If it doesn't, then we have to swap the non-character
   // elements of the header.

   unsigned short one = 1;
   char *p = reinterpret_cast<char *>(&one);
   if(1 == static_cast<int>(*p))
      Swap(Header);

   // Read the additional padding in the header

   long size = Header.rlng-MinimumRigakuHeaderLength;
   p = new char [size];
   InputFile.read(p,size);
   delete [] p;

   return !(InputFile.bad() || InputFile.fail());
}

/****************************************************************************
 ****************************************************************************/
bool ReadImageBuffer(std::ifstream &InputFile,unsigned short * const pImageBuffer)
{
   // Figure out how much of the image is contained in the file.  Note that
   // if the image file contains only a partial image, we will NOT construct
   // an entire image.  The header element ipst should be used to determine
   // the line in the entire image corresponding to a line in the image file.
   //
   // Note that rlng is the size of a record in bytes, and we are assuming
   // 2 bytes/pixel.
   RigakuHeader Header;
   ReadHeader(InputFile, Header);
   long size = Header.rlng*Header.rnum/2;

   // Read in the data.

   InputFile.read(reinterpret_cast<char *>(pImageBuffer),size*2);

   if(InputFile.bad() || InputFile.fail())
      return false;

   // Do we need to swap the data?

   unsigned short one = 1;
   char *p = reinterpret_cast<char *>(&one);

   if(1 == static_cast<int>(*p)){

      // Swap each pixel value in the image buffer

      unsigned short *p = const_cast<unsigned short *>(pImageBuffer);
      unsigned short *pEnd = p+size;
      for(NULL; p < pEnd; p++)
         *p = Swap(*p);

   }

   return true;

}

/****************************************************************************
 ****************************************************************************/
unsigned short Swap(const unsigned short Value)
{
   unsigned short swap = Value;
   char *p = reinterpret_cast<char *>(&swap);

   char tmp = p[0];
   p[0] = p[1];
   p[1] = tmp;

   return swap;
}

/****************************************************************************
 ****************************************************************************/
long Swap(const long Value)
{
   long swap = Value;
   char *p = reinterpret_cast<char *>(&swap);

   char tmp = p[0];
   p[0] = p[3];
   p[3] = tmp;

   tmp = p[1];
   p[1] = p[2];
   p[2] = tmp;

   return swap;
}

/****************************************************************************
 ****************************************************************************/
float Swap(const float Value)
{
   float swap = Value;
   char *p = reinterpret_cast<char *>(&swap);

   char tmp = p[0];
   p[0] = p[3];
   p[3] = tmp;

   tmp = p[1];
   p[1] = p[2];
   p[2] = tmp;

   return swap;
}

/****************************************************************************
 ****************************************************************************/
void Swap(RigakuHeader &Header)
{

// Swap the bytes in the long elements of the header.

   Header.cyld = Swap(Header.cyld);
   Header.oscn = Swap(Header.oscn);
   Header.xpxl = Swap(Header.xpxl);
   Header.zpxl = Swap(Header.zpxl);
   Header.rlng = Swap(Header.rlng);
   Header.rnum = Swap(Header.rnum);
   Header.ipst = Swap(Header.ipst);
   Header.ipnm = Swap(Header.ipnm);
   Header.dr_x = Swap(Header.dr_x);
   Header.dr_z = Swap(Header.dr_z);
   Header.drxz = Swap(Header.drxz);	

// Swap the bytes in the float elements of the header.

   Header.alng = Swap(Header.alng);
   Header.blng = Swap(Header.blng);
   Header.clng = Swap(Header.clng);
   Header.alfa = Swap(Header.alfa);
   Header.beta = Swap(Header.beta);
   Header.gamm = Swap(Header.gamm);
   Header.mosc = Swap(Header.mosc);
   Header.wlng = Swap(Header.wlng);
   Header.m2ta = Swap(Header.m2ta);
   Header.camr = Swap(Header.camr);
   Header.vltg = Swap(Header.vltg);
   Header.crnt = Swap(Header.crnt);
   Header.weis = Swap(Header.weis);
   Header.phi0 = Swap(Header.phi0);
   Header.phis = Swap(Header.phis);
   Header.phie = Swap(Header.phie);
   Header.fext = Swap(Header.fext);
   Header.drtx = Swap(Header.drtx);
   Header.drtz = Swap(Header.drtz);
   Header.omga = Swap(Header.omga);
   Header.fkai = Swap(Header.fkai);
   Header.thta = Swap(Header.thta);
   Header.mu   = Swap(Header.mu);
   Header.xsiz = Swap(Header.xsiz);
   Header.zsiz = Swap(Header.zsiz);
   Header.rato = Swap(Header.rato);
   Header.ft_1 = Swap(Header.ft_1);
   Header.ft_2 = Swap(Header.ft_2);

// New variables for RAPID image headers

   Header.shft = Swap(Header.shft);
   Header.ineo = Swap(Header.ineo);
   Header.majc = Swap(Header.majc);
   Header.naxs = Swap(Header.naxs);
   for(int i = 0; i < 5; i++){
      for(int j = 0; j < 3; j++)
         Header.gvec[i][j] = Swap(Header.gvec[i][j]);
      Header.gst[i] = Swap(Header.gst[i]);
      Header.gend[i] = Swap(Header.gend[i]);
      Header.goff[i] = Swap(Header.goff[i]);
   }
   Header.saxs = Swap(Header.saxs);

   return;
}

/****************************************************************************
 ****************************************************************************/
long UnPackPixelValue(const RigakuHeader &Header,
                      const unsigned short PixelValue)
{
   // Pixels are packed so that if the most significant bit is set, then
   // the value has been bit-shifted by the amount specified in the rato
   // header element.
   //
   // Note that this form of packing is actually quite inspired as the
   // amount of data lost due to packing is much less than the error in
   // the measurement due to normal counting statistics (assuming that
   // they are Poissonian).

   if(0x7fff >= PixelValue)
      return PixelValue;

   // We only need to use the lower 15 bits of the value to uncompress it.

   long Value = 0x7fff&PixelValue;

   // Shift the value by the ratio in the header.

   Value *= Header.rato;

   return Value;
}

/****************************************************************************
 ****************************************************************************/
void OutputSamplePixelValues(const RigakuHeader &Header,
                             unsigned short * const pImageBuffer)
{
   // Output a few pixel values.  Note that the first line in the image
   // buffer may not correspond to the first line in a "complete" image.

   std::cout << std::endl << "Sample pixel information: " << std::endl;

   std::cout << "   First pixel (0," << Header.ipst << ") in image = " << 
                UnPackPixelValue(Header,*pImageBuffer) << std::endl;

   long index = (Header.rnum/2-1)*Header.xpxl+(Header.xpxl/2-1);
   std::cout << "   Middle pixel (" << Header.xpxl/2 << ',' << 
                 Header.rnum/2+Header.ipst << ") in image = " << 
                 UnPackPixelValue(Header,pImageBuffer[index]) << std::endl;

   index = Header.rnum*Header.xpxl-1;
   std::cout << "   Last pixel (" << Header.xpxl << ',' <<
                 Header.rnum+Header.ipst << ") in image = " << 
                 UnPackPixelValue(Header,pImageBuffer[index]) << std::endl;

   return;
}

/****************************************************************************
 ****************************************************************************/
void OutputHeaderInfo(const RigakuHeader &Header)
{
   std::cout << "Header information: " << std::endl;

   std::cout << "   Wavelength: " << Header.wlng << std::endl;
   std::cout << "   Exposure time: " << Header.fext << std::endl;
   std::cout << "   Direct beam position: (" << Header.drtx << ',' <<
      Header.drtz << ')' << std::endl;
   std::cout << "   Pixel information" << std::endl <<
      "      size: (" << Header.xsiz << ',' << Header.zsiz << ')' << 
      std::endl <<
      "      number: (" << Header.xpxl << ',' << Header.zpxl << ')' << 
      std::endl;
   std::cout << "   Line information" << std::endl <<
      "      total: " << Header.rnum << std::endl <<
      "      starting: " << Header.ipst << std::endl;
   std::cout << "   IP shape: " << ((1 == Header.cyld) ? "cylindrical" : 
      "flat") << std::endl;

   axis Axes[5];

   std::string s = Header.name;
   bool IsRapid = std::string::npos != s.find("R-AXIS-CS");
   std::string ScanAxis = "Phi";

   if(0 != Header.majc){
      std::istringstream is(Header.gnom);
      for(int i = 0; i < Header.naxs; i++){
         is >> Axes[i].name;
         Axes[i].start = Header.gst[i];
         Axes[i].end = Header.gend[i];
         Axes[i].offset = Header.goff[i];
      }
      ScanAxis = Axes[Header.saxs].name;
   }
   else{
      Axes[0].name = "Omega";
      Axes[2].name = "Phi";
      if(!IsRapid){
         Axes[0].start = Header.omga;
         Axes[0].end = Header.omga;
         Axes[0].offset = 0.0;
         Axes[2].start = Header.phis;
         Axes[2].end = Header.phie;
         Axes[2].offset = Header.phi0;
         ScanAxis = Axes[2].name;
      }
      else{
         Axes[2].start = Header.omga;
         Axes[2].end = Header.omga;
         Axes[2].offset = 0.0;
         Axes[0].start = Header.phis;
         Axes[0].end = Header.phie;
         Axes[0].offset = Header.phi0;
         ScanAxis = Axes[0].name;
      }
      Axes[1].name = "Chi";
      Axes[1].start = Header.fkai;
      Axes[1].end = Header.fkai;
      Axes[1].offset = 0.0;
   }

   std::cout << "   Instrument axes (using " << 
                ((0 != Header.majc) ? "extended" : "conventional") <<
                " section of header):" << std::endl;
   std::cout << "      Scan axis: " << ScanAxis << std::endl;
   for(int i = 0; i < 5 && "" != Axes[i].name; i++)
      std::cout << "      " << Axes[i] << std::endl;

   axis theta,dist;
   theta.name = "2Theta";
   theta.start = Header.thta;
   theta.end = Header.thta;
   theta.offset = 0.0;
   dist.name = "Distance";
   dist.start = Header.camr;
   dist.end = Header.camr;
   dist.offset = 0.0;
   std::cout << "      " << theta << std::endl;
   std::cout << "      " << dist << std::endl;

   return;
}

/****************************************************************************
 ****************************************************************************/
std::ostream &operator<<(std::ostream &a,
                         axis &b)
{
   a << b.name << ": start = " << b.start << ", end = " << b.end << 
        ", offset = " << b.offset;
   return a;
}
