# $Id: condition.awk,v 1.5 2008/06/09 17:38:13 mdw Exp $

#####################################################################
#
# Copyright (c) 1995 - 2004, EMBL, Peter Keller
# 
# CCIF: a library to access and output data conforming to the
# CIF (Crystallographic Information File) and mmCIF (Macromolecular
# CIF) specifications, and other related specifications.
# 
# This library is free software: you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public License
# version 3, modified in accordance with the provisions of the 
# license to address the requirements of UK law.
#
# You should have received a copy of the modified GNU Lesser General 
# Public License along with this library.  If not, copies may be 
# downloaded from http://www.ccp4.ac.uk/ccp4license.php
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
# 
#####################################################################

# Awk program to generate CCIF's condition handling module.
# *** Needs nawk! ***
# See also comments in condition.list.

# During reading in of condition.list, we build arrays of facilities
# and messages. Each has two arrays - one containing the numbers, with
# the name as subscript, the other containing the name, with the number
# as subscript. (The arrays for messages both have the facility name as
# the first subscript.)
#
# This allows easy translation between names and numbers, grouping
# of condition codes by facility, and correct layout of static
# structures in the C code output.
# 


BEGIN { fac_no = 0 ; 
        FS = "[\t]+" ; 
        SUBSEP = "_" ;

#    Translate severity codes to their numerical values:
#               0      1     2     3       4
        split("warn:success:error:inform:severe", sev_tmp, ":");

        for ( sev in sev_tmp ) {
           sev_code[sev - 1] = sev_tmp[sev];
           sev_list[sev_tmp[sev]] = sev - 1;
           delete sev_tmp[sev];
        }

# VMS-style sizes and positions of message and facility
# parts of condition code.
      fac_no_s = 12;
      fac_no_v = 16;
      msg_no_s = 13;
      msg_no_v = 3;

# Names of output files
      condition_h = "condition.h"
      condition_list = "condition_list.h"

}

{   
#   Ignore lines which don't have 4 fields.
    if ( NF != 4 ) next;

#   If this is the first time we have seen this facility, assign it
#   a new number
    if ( !( $1 in fac_list ) ) {
       fac_list[$1] = fac_no;       # Facility names are subscripts; numbers are contents
       fac_names[fac_no++] = $1;    # Facility numbers are subscripts; names are contents
       msg_no[$1] = 0;
    }

    if ( !( $3 in sev_list ) ) {
       printf ("***Unknown severity code \"%s\" at line %s of %s\n", $3, FNR, FILENAME);
       exit;
    }

    msg_list[$1, $2] = msg_no[$1];    # Second subscript is message name; contents is message number
    msg_names[$1, msg_no[$1]++] = $2; # Second subscript is message number; contents is message name
    msg_sev[$1, $2] = sev_list[$3];
    msg_fmt[$1, $2] = $4;
}

END { # Everything in place - now can generate C source for condition handling
      # module and header file.

##########################
#  generate  condition.h:
##########################
      print ("/* condition.h --- CCIF condition handler header. */") > condition_h ;
      print ("/* Generated automatically by \"condition.awk\" */") > condition_h;
      print ("\n#ifndef CONDITION_H\n#define CONDITION_H\n") > condition_h;

      print ("\n/* This file must typedef uint32 as a 32-bit unsigned integer value */") > condition_h;
      print ("#include \"uint32.h\"") > condition_h;
      
      print ("\n/* This file must define the macros __import__, __export__ and __noshare__ */") > condition_h;
      print ("#include \"import_export.h\"") > condition_h;

      print ("\n/* Length of message buffer for callback functions */") > condition_h;
      print ("#define CALLBACK_LEN 512\n")                              > condition_h;
      print ("\n/* Breakdown of condition value (not using control bits, at the moment */") > condition_h ;
      print ("\n/* Severity bits */") > condition_h;
      print ("#define CCIF_M_SEVERITY 0x07  /* Severity mask */") > condition_h ;
      for ( sev in sev_list )
         printf ("#define CCIF_K_%s %d\n", toupper(sev), sev_list[sev]) > condition_h ;

      print ("\n/* Message bits */") > condition_h
      print ("#define CCIF_V_MSG_NO " msg_no_v "   /* Position of message number */") > condition_h
      print ("#define CCIF_S_MSG_NO " msg_no_s "  /* Size of message number */") > condition_h
      print ("#define CCIF_M_MSG_NO ( 0x1fff << CCIF_V_MSG_NO ) /* Message number mask */") > condition_h;

      print ("\n/* Facility bits */") > condition_h;
      print ("#define CCIF_V_FAC_NO " fac_no_v "  /* Position of facility number */") > condition_h;
      print ("#define CCIF_S_FAC_NO " fac_no_s "  /* Size of facility number */") > condition_h;
      print ("#define CCIF_M_FAC_NO ( 0x0fff << CCIF_V_FAC_NO ) /* Facility number mask */" ) > condition_h;


     print ("\n/* Condition code symbols: */") > condition_h;
# Outer loop - loops over facilities
     for ( i = 0; i < fac_no; i++ ) {
        this_fac = fac_names[i];
        print ("\n/* Facility " this_fac " */") > condition_h;
 
# Inner loop - loops over messages within this facility
        for ( j = 0; j < msg_no[this_fac]; j++ ) {
           this_msg = msg_names[this_fac, j];
           printf ("__import__ const uint32 %s_%s;\n", this_fac, this_msg) > condition_h;
        }

     }

     print ("\n\n/* Prototype for condition handler */") > condition_h;
     print ("extern void ccif_signal(const int code, const char * const pre,") > condition_h;
     print ("            void (*callback)(), const void * ptr, int idx, ...); ") > condition_h;
     print ("extern void ccif_condition_init(void);") > condition_h;
     print ("int ccif_warnmsg_switch ( const char * varname, const int on );") > condition_h;

     print ("\n#endif /* CONDITION_H */\n") > condition_h;

#     for ( cond in msg_list ) { print cond, msg_list[cond], msg_sev[cond], msg_fmt[cond]; }


#######################################
# generate condition_list.h 
# (Static data for condition handler)
#######################################

     print ( "/* Condition data for CCIF. Generated automatically by condition.awk */" ) > condition_list;


# This bit is the same as in condition_h - we don't want to #include both condition_h and
# and condition_list in the condition handler function definition.

      print ("\n/* This file must typedef uint32 as a 32-bit unsigned integer value */") > condition_list;
      print ("#include \"uint32.h\"") > condition_list;
      
      print ("\n/* This file must define the macros __import__, __export__ and __noshare__ */") > condition_list;
      print ("#include \"import_export.h\"") > condition_list;

      print ("\n/* Length of message buffer for callback functions */") > condition_list;
      print ("#define CALLBACK_LEN 512\n")                              > condition_list;
      print ("\n/* Breakdown of condition value (not using control bits, at the moment */") > condition_list ;

      print ("\n/* Array of environment variables or control strings, to turn of messages " ) > condition_list ;
      print ("    by severity */" ) > condition_list ;
      pre = "";
      printf ("static char *message_control_vars[] = {") > condition_list;
      for ( i = 0; i in sev_code; i++ ) {
	printf ("%s\n   \"CCIF_%s\"", pre, toupper(sev_code[i]) ) > condition_list;
	pre = ",";
      }
      print ( "\n};" ) > condition_list;

      print ("\n/* Severity bits */") > condition_list;
      print ("#define CCIF_M_SEVERITY 0x07  /* Severity mask */") > condition_list ;
      for ( sev in sev_list )
         printf ("#define CCIF_K_%s %d\n", toupper(sev), sev_list[sev]) > condition_list ;

      print ("\n/* Message bits */") > condition_list
      print ("#define CCIF_V_MSG_NO " msg_no_v "   /* Position of message number */") > condition_list
      print ("#define CCIF_S_MSG_NO " msg_no_s "  /* Size of message number */") > condition_list
      print ("#define CCIF_M_MSG_NO ( 0x1fff << CCIF_V_MSG_NO ) /* Message number mask */") > condition_list;

      print ("\n/* Facility bits */") > condition_list;
      print ("#define CCIF_V_FAC_NO " fac_no_v "  /* Position of facility number */") > condition_list;
      print ("#define CCIF_S_FAC_NO " fac_no_s "  /* Size of facility number */") > condition_list;
      print ("#define CCIF_M_FAC_NO ( 0x0fff << CCIF_V_FAC_NO ) /* Facility number mask */" ) > condition_list;



     print ( "\nstatic char callback_msg[CALLBACK_LEN];\n\n")                > condition_list;


     print ("\n/* Condition code value definitions: */") > condition_list;
# Outer loop - loops over facilities
     for ( i = 0; i < fac_no; i++ ) {
        this_fac = fac_names[i];
        print ("\n/* Facility " this_fac " */") > condition_list;
 
# Inner loop - loops over messages within this facility
        for ( j = 0; j < msg_no[this_fac]; j++ ) {
           this_msg = msg_names[this_fac, j];
           msg_value = i * 2^fac_no_v + j * 2^msg_no_v + msg_sev[this_fac, this_msg];

#           printf ("__export__ const uint32 %s_%s=\t0x%08x;\n", this_fac, this_msg, msg_value) > condition_list;

#   Use octal rather than hex - i.e. the rightmost digit is the severity. 
#   For pretty output, keep facility names and message names less than 10 and 15
#   characters respectively (or change the widths in this printf statement)
           printf ("__export__ const uint32 %10s_%-15s = 0%011o;\n", this_fac, this_msg, msg_value) > condition_list;

        }

     }


# Loop over facilities, writing out all lists of messages

     print ("\n/* Messages, grouped by facility: */") > condition_list;
# Outer loop - loops over facilities
     for ( i = 0; i < fac_no; i++ ) {
        this_fac = fac_names[i];
        pre = sprintf ("\nstatic MESSAGE  %s_msgs[%d] = {\n   ", this_fac, msg_no[this_fac]);
 
# Inner loop - loops over messages within this facility
        for ( j = 0; j < msg_no[this_fac]; j++ ) {
           this_msg = msg_names[this_fac, j];
	   printf ("%s{\n      \"%s\",\n      %s,\n     -1\n   }", pre, this_msg, msg_fmt[this_fac, this_msg] ) \
                                 > condition_list;
           pre = ",\n   ";
        }
        printf("\n};\n") > condition_list;

     }

# Now, loop over facilities again, setting up pointers to the message lists.

   print ("\n/* Facilities, pointing to the message lists */") > condition_list;
   pre = sprintf ("static FACILITY CCIF_facs[%d] = {\n   ", fac_no);
   for ( i = 0; i < fac_no; i++ ) {
      this_fac = fac_names[i];
      printf ("%s{\"%s\", %s_msgs, %d}", pre, this_fac, this_fac, msg_no[this_fac]) \
                                 > condition_list;
      pre = ",\n   ";
   }
   printf("\n};\n") > condition_list;




}


