/* $Id: dict_analysis.c,v 1.7 2008/06/09 17:38:14 mdw Exp $ */

/********************************************************************
* Copyright (c) 1995 - 2004, EMBL, Peter Keller
* 
* CCIF: a library to access and output data conforming to the
* CIF (Crystallographic Information File) and mmCIF (Macromolecular
* CIF) specifications, and other related specifications.
* 
* This library is free software: you can redistribute it and/or
* modify it under the terms of the GNU Lesser General Public License
* version 3, modified in accordance with the provisions of the 
* license to address the requirements of UK law.
*
* You should have received a copy of the modified GNU Lesser General 
* Public License along with this library.  If not, copies may be 
* downloaded from http://www.ccp4.ac.uk/ccp4license.php
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU Lesser General Public License for more details.
* 
********************************************************************/

/*  This code has been written to allow various kinds of analyses of the
    mmCIF dictionary. At the moment, it builds a graph containing the
    relationships between categories, and can output dot source, and SQL,
    with the CREATE TABLE instructions ordered so that no table is created
    before another table on which it depends.

   To build with this code, run configure with the --with-analysis option

   At the moment, the programs which generate output from this code
   are cifdicdot, cifdicsql and cifdickeys.

   */

#include <assert.h>


static char *dot_header="digraph cif_mm {\n\
   fontsize=7;\n\
   ranksep=.15;\n\
   nodesep=0.08;\n\
   rankdir=LR;\n\
   concentrate=false;\n\
/*   page=\"8.26,11.69\"; */ /* A4 */ \n\
   rotate=90;            /* Need this for A3 */ \n\
   page=\"11.69,16.52\"; /* A3 */ \n\
   size=\"15.5,10.5\"; \n\
   ratio=fill; \n\
/*   page=\"8.5,11\";  */  /* US letter */ \n\
   margin=0.20; \n\
   node [shape=plaintext]; \n";

static char *pointer_key="   one_child [shape=plaintext,label=\"one\"];\n\
   one_parent [shape=plaintext,label=\"one\"];\n\
   many_child [shape=plaintext,label=\"many\"];\n\
   many_parent [shape=plaintext,label=\"many\"];\n\
   one_child -> one_parent [style=bold,color=red];\n\
   one_child -> many_parent [style=filled, color=black];\n\
   many_child -> one_parent [style=dashed, color=red];\n\
   many_child -> many_parent [style=dashed, color=black];\n\
   subgraph cluster_key {\n\
      label = \"Key\";\n\
      fontsize=14;\n\
      color=black;\n\
      one_child; one_parent; many_child; many_parent;\n\
   }\n";


/* Macro for adding a symbol table item "item" to an array (Sym * (pre ## list)[]), in which each item
   appears only once. The number of members of the array is maintained in (pre ## n ## list).
   *** NEEDS ANSI PREPROCESSOR !!! ***
   */

#define ADD_TO_UNIQUE_ITEM_LIST(count, item, pre, list) for ( count = 0; count < pre ## n ## list; count++ ) \
                                                           if ( pre ## list[count] == item ) break; \
							   if ( count ==  pre ## n ## list ) { \
                                                              count = ++pre ## n ## list; \
							      pre ## list = (Sym **) realloc(pre ## list, \
											     count * sizeof (Sym *)); \
							      assert (pre ## list ); \
							      pre ## list[count-1] = item; \
                                                         }


#define MAXTABLES 200

#if 0
#define ALL_LINKED_ITEMS
static char *ports[] = {":c", ":c", ":t"};
#else
static char *ports[] = {"", ":c", ":t"};
#endif

/* These structures rely on static storage class initialising to NULL */
static struct _edge_matrix {

  int edge,     /* set to 1 on first pass if edge exists */
    nkey_child, nkey_parent, /* These are the number of key items in the parent
				and child categories which are key items. Also 
				set on first pass */
    nother_child, nother_parent;

  Sym **Parent;
  int nparents;

  int child_port, parent_port;

  struct _link {
    Sym *child;
    int parent;  /* index into Parent array, above */
  } *Link;
  int nlinks;

} Edge_matrix[MAXTABLES][MAXTABLES]; /* Order of subscripts is
                                        [child][parent] */

static struct _table_lookup {
  Sym *p;
  int nkey;
  char charid[4];    /* Unique two-character prefix (plus '$') for table */
  Sym **items_in_port;
  int nitems_in_port;
  int nchild_cats,   /* double-purpose use: used to mark a category to be excluded
		       from the graph representation */
    nparent_cats;

  Sym **links_child_end, **links_parent_end;
  int nlinks_child_end, nlinks_parent_end;

} Table_lookup[MAXTABLES];



static int ntables, ntables_excl;


/* Builds a representation of the graph, from the symbol table form
   of mmCIF.
   parent_list is an array of category names to exclude from (when exclude != 0)
   or include in (when exclude == 0) the representation.

   This function generates no output.
*/

void build_graph (const char * parent_list[], const int exclude){


  int i,j, row=0, column=0, c_lookup, p_lookup,
    child, parent, nchild, nparent, key_to_key, nedges=0;

  const char **ch;
  static char charid[4] = "AA$";
  Sym *p, *q, *r;
  struct _edge_matrix *edge;



  /* Set up table lookup array */
  for (i = 0, ntables = 0; i < size; i++) {
    for (p = table[i]; p; p = p->next) {
      
      if ( p->category.rec == p ) {
	
	p->id = ntables;
	Table_lookup[ntables].p = p;
	
	for ( q = p->key.rec; q != p ; q = q->key.rec, Table_lookup[ntables].nkey++ );
	
	if ( exclude ) { 
	  Table_lookup[ntables].nchild_cats = 0;
	}
	else
	  Table_lookup[ntables].nchild_cats = -1;
	
	strcpy(Table_lookup[ntables].charid, charid);
	ntables++;
	if ( charid[1] == 'Z' ) {
	  charid[0]++;
	  charid[1] = 'A';
	}
	else {
	  charid[1]++;
	}
	assert(ntables <= MAXTABLES);
      }
    }
  }


  /* Mark all parent categories to exclude from diagram */
  for ( ch = &(parent_list[0]), ntables_excl=0; **ch; ch++ ){
    p = zzs_get(*ch);
    assert(p);

    if ( exclude ) {
      Table_lookup[p->id].nchild_cats = -1;
    }
    else {
      Table_lookup[p->id].nchild_cats = 0;
    }

    ntables_excl++;
  }


  /* Find all edges in graph. First-pass analysis, by symbol table entry */

  for (i = 0; i < size; i++) {
    for (q = table[i]; q; q = q->next) {

      if ( ! q->parent.rec || q->true_name.rec ) continue;

      p_lookup = q->parent.rec->category.rec->id;
      c_lookup = q->category.rec->id;
      edge = &(Edge_matrix[c_lookup][p_lookup]);
      
      
      if ( exclude ) {
	/* If parent category is excluded, exclude this category as well */
	if ( Table_lookup[p_lookup].nchild_cats < 0 ) {
	  if ( Table_lookup[c_lookup].nchild_cats >= 0 ) {
	    Table_lookup[c_lookup].nchild_cats = -1;
	    ntables_excl++;
	  }
	    continue;
	}
      }
      else {
	/* If parent category is included, include this category as well */
	if ( Table_lookup[p_lookup].nchild_cats >= 0 ) {
	  if ( Table_lookup[c_lookup].nchild_cats < 0 ) {
	    Table_lookup[c_lookup].nchild_cats = 0;
	    ntables_excl++;
	  }
	}
	else 
	  continue;
      }
	
      edge->edge = 1;
      
      /* Look for parent in array Parent */
      for (j = 0; j < edge->nparents; j++)
	if ( edge->Parent[j] == q->parent.rec ) break;
      if ( j == edge->nparents ) {
	edge->nparents++;
	edge->Parent = (Sym **) realloc(edge->Parent, edge->nparents * sizeof (Sym *));
	assert (edge->Parent);
	edge->Parent[j] = q->parent.rec;
	
	if ( q->parent.rec->key.rec ) {
	  edge->nkey_parent++;
	  }
	else {
	  edge->nother_parent++;
	}
	
      }
      
      /* Set up child-parent link */
      edge->Link = (struct _link *) realloc(edge->Link, (edge->nlinks + 1) * sizeof (struct _link) );
      assert (edge->Link);
      edge->Link[edge->nlinks].child = q;
      edge->Link[edge->nlinks].parent = j;
      edge->nlinks++;
      
      if ( q->key.rec ) {
	edge->nkey_child++;
      }  
      else {
	  edge->nother_child++;
      }



      
      /* Mark parent category, if it is not ultimate parent... */
      if ( q->parent.rec->parent.rec ) {

#if 0
	for ( j = 0; j < Table_lookup[p_lookup].nitems_in_port; j++ )
	  if (Table_lookup[p_lookup].items_in_port[j] == q->parent.rec) break;
	
	if ( j == Table_lookup[p_lookup].nitems_in_port ) {
	  j = ++Table_lookup[p_lookup].nitems_in_port;
	  Table_lookup[p_lookup].items_in_port = 
	    (Sym **) realloc(Table_lookup[p_lookup].items_in_port, j * sizeof (Sym *));
	  assert(Table_lookup[p_lookup].items_in_port);
	  Table_lookup[p_lookup].items_in_port[j-1] = q->parent.rec;
	}
#else
	ADD_TO_UNIQUE_ITEM_LIST(j, q->parent.rec, Table_lookup[p_lookup]., items_in_port);
#endif
	
	  edge->parent_port = 2;  /* port ":t" */
	  
      }
      else {   /* ...otherwise add to list of parent items in parent category */
	ADD_TO_UNIQUE_ITEM_LIST(j, q->parent.rec, Table_lookup[p_lookup]., links_parent_end);
      }

      /* Always add to list of items in child category. This isn't quite correct - may fix it
	 sometime if I think it is important enough */
      ADD_TO_UNIQUE_ITEM_LIST(j, q, Table_lookup[c_lookup]., links_child_end);
      
      nedges++;
      
      
    }    /* for (p = table[i]; p; p = p->next) */
  }   /*  for (i = 0, ntables = 0; i < size; i++) */
  

  /* Second pass analysis, by edge */
  for ( child = 0; child < ntables; child++ ) {
    for ( parent = 0; parent < ntables; parent++ ) {

      edge = &(Edge_matrix[child][parent]);
      if ( !edge->edge ) continue;
      
      Table_lookup[child].nparent_cats++;
      Table_lookup[parent].nchild_cats++;
      
      if ( Table_lookup[child].nitems_in_port ) edge->child_port = 1;
      for ( i = 0; i < Table_lookup[child].nitems_in_port; i++ ) {
	for ( j = 0; j < edge->nlinks; j++ ) {
	  if ( edge->Link[j].child == Table_lookup[child].items_in_port[i] )
	    edge->child_port = 2;
	}
      }

      if ( Table_lookup[parent].nitems_in_port && !edge->parent_port ) 
	edge->parent_port = 1;
      
    }
  }



  if ( exclude ) {
    fprintf(stderr, "...%d nodes, %d edges....",ntables - ntables_excl, nedges);
  }
  else {
    fprintf(stderr, "...%d nodes, %d edges....",ntables_excl, nedges);
  }
  
  
}




/* Outputs source for dot/dotty to filename */

void dot_dump (char *filename){

  FILE *ptrsout;
  int i,j, row=0, column=0, c_lookup, p_lookup, 
    child, parent, nchild, nparent, key_to_key;

  char **ch;
  Sym *p, *q, *r;
  struct _edge_matrix *edge;


  ptrsout = fopen(filename, "w");
  assert(ptrsout != NULL);
  fprintf(ptrsout, dot_header);
  fprintf(ptrsout, pointer_key);

  /* output all node specifiers */
  for ( i = 0; i < ntables; i++ ) {
    if ( Table_lookup[i].nchild_cats < 0 ) continue; /* excluded node */

#if 1   /* Don't output orphrans */
    if ( !Table_lookup[i].nparent_cats && !Table_lookup[i].nchild_cats ) {
      ntables_excl++;
      fprintf(stderr, "Category \"%s\" contains no child or parent items\n",
	      zzs_recname(Table_lookup[i].p));
      continue;
    }
#endif

#ifdef ALL_LINKED_ITEMS
    fprintf(ptrsout, "   %1$s [shape=record, label=\"<c> %1$s", zzs_recname(Table_lookup[i].p) );
    
    if ( Table_lookup[i].nitems_in_port ) {
      fprintf (ptrsout, " | <t> Child/parent items");
      for ( j = 0; j < Table_lookup[i].nitems_in_port; j++)
	fprintf (ptrsout, "\\n%s", 
		   strchr(zzs_recname(Table_lookup[i].items_in_port[j]), '.') + 1 );
    }
      
    if ( Table_lookup[i].nlinks_parent_end ) {
      fprintf(ptrsout, " | <p> Parent items");
      for ( j = 0; j < Table_lookup[i].nlinks_parent_end; j++ )
	fprintf(ptrsout, "\\n%s",
		strchr(zzs_recname(Table_lookup[i].links_parent_end[j]), '.') + 1);
    }
    
    if ( Table_lookup[i].nlinks_child_end ) {
	fprintf(ptrsout, " | <ch> Child items");
	for ( j = 0; j < Table_lookup[i].nlinks_child_end; j++ )
	  fprintf(ptrsout, "\\n%s",
		  strchr(zzs_recname(Table_lookup[i].links_child_end[j]), '.') + 1);
    }
    
    fprintf(ptrsout, "\"];\n");
#else
    
    if ( Table_lookup[i].nitems_in_port ) {
      fprintf (ptrsout, "   %1$s [shape=record, label=\"<c> %1$s | <t> %2$s",
	       zzs_recname(Table_lookup[i].p), 
	       strchr(zzs_recname(Table_lookup[i].items_in_port[0]), '.') + 1 );
      for ( j = 1; j < Table_lookup[i].nitems_in_port; j++)
	fprintf (ptrsout, "\\n%s", 
		 strchr(zzs_recname(Table_lookup[i].items_in_port[j]), '.') + 1 );
      fprintf(ptrsout, "\"];\n");
    }
    else {
      fprintf(ptrsout, "   %s [shape=plaintext];\n", zzs_recname(Table_lookup[i].p) ); 
      
    }
  
#endif
} /* for ( i = 0; i < ntables; i++ ) */

      /* Loop over all edges, outputting each one with appropriate
	 visual properties */
  for ( child = 0; child < ntables; child++ ) {
    q = Table_lookup[child].p;
    for (parent = 0; parent < ntables; parent++ ) {

      edge = &(Edge_matrix[child][parent]);
      if ( !edge->edge ) continue;
      
      p = Table_lookup[parent].p;
      
      if ( ( Table_lookup[parent].nkey == edge->nkey_parent )
	   && ( Table_lookup[child].nkey == edge->nkey_child ) ) {

	fprintf(ptrsout, "   %s%s -> %s%s [style=bold,color=red];\n", 
		zzs_recname(q), ports[edge->child_port],
		zzs_recname(p), ports[edge->parent_port] );
	key_to_key = 1;
	nchild = edge->nother_child;
	nparent = edge->nother_parent;
      }
      else {
	key_to_key = 0;
	nchild = edge->nother_child + edge->nkey_child;
	nparent = edge->nother_parent + edge->nkey_parent;
      }
     

      if ( nchild && nparent ) {
	fprintf(ptrsout, "   %s%s -> %s%s [",
		zzs_recname(q), ports[edge->child_port],
		zzs_recname(p), ports[edge->parent_port] );
	
	if ( Table_lookup[child].nkey == Edge_matrix[child][parent].nkey_child ) {
	  fprintf(ptrsout, "style=solid");
	}
	else {
	  fprintf(ptrsout, "style=dashed"); 
	}
	
	if ( Table_lookup[parent].nkey == Edge_matrix[child][parent].nkey_parent ) {
	  fprintf(ptrsout, ", color=red");
	    }
	else {
	  fprintf(ptrsout, ", color=black");
	}
	
	if ( nchild != nparent ) {
	  fprintf(ptrsout, ", decorate=false, fontsize=8, label=\"%d-%d\"", nchild, nparent);
	}
	
	fprintf(ptrsout,"];\n");
	
      }
    }
  }
  
 
  fprintf(ptrsout, "}\n");
  fclose(ptrsout);
  
}


/* Outputs the CREATE TABLE instruction corresponding to category p to
   the open file unit sqlout */

void cat2sql ( FILE * sqlout, Sym *p ) {
  
  Sym *q;
  char *n, *cnl;

  int lookup;

  assert ( sqlout );
  assert ( p && p->category.rec == p );

  lookup = p->id;

  fprintf(sqlout, "\nCREATE TABLE %s (", (zzs_recname(p) ) );


  /* Attribute list */
  for ( q = p->item_list.rec, cnl = "\n"; 
	q ;
	q = q->item_list.rec ,cnl = ",\n") {
    n = strchr(zzs_recname(q), '.');
    assert(n != NULL);
    fprintf(sqlout, "%s      %s%s       CHAR(18) %s", cnl, Table_lookup[lookup].charid, n+1,
	    (q->key.rec || q->mandatory_code == yes || q->mandatory_code == implicit) ?
	       "NOT NULL" : "");
  }

  /* Primary key list */
  fprintf(sqlout, ",\n      PRIMARY KEY (");
  for ( q = p->key.rec, cnl = "";
	q != p;
	q = q->key.rec, cnl = "," ) {
    n = strchr(zzs_recname(q), '.');
    assert(n != NULL);
    fprintf(sqlout, "%s %s%s", cnl, Table_lookup[lookup].charid, n+1);
  }
  fprintf(sqlout, ")");

  /* Foreign key list */
  for ( q = p->item_list.rec;
	q;
	q = q->item_list.rec ) {
    if ( ! q->parent.rec ) continue;
    n = strchr(zzs_recname(q), '.');
    assert(n);
    fprintf(sqlout, ",\n      FOREIGN KEY (%s%s)", Table_lookup[lookup].charid, n+1);
    n = strchr(zzs_recname(q->parent.rec), '.');
    assert(n);
    fprintf(sqlout, "\n         REFERENCES %s (%s%s)", 
	    zzs_recname(q->parent.rec->category.rec),
	    Table_lookup[q->parent.rec->category.rec->id].charid,
	    n+1);
  }
  fprintf(sqlout, "\n);\n");
}


/* Calls cat2sql for all categories in the graph representation.
   Ensures that where a foreign key reference exists, the referenced
   table has already been created.

   Won't cope with indirect cycles
*/

void sql_dump (char *filename) {

  FILE *sqlout;
  int tables_out=0, i,j;

  sqlout = fopen(filename, "w");
  assert(sqlout);
  
  /* First, deal with direct cycles (this code won't cope with indirect ones) */
  for ( i = 0; i < ntables; i++ )
    if ( Edge_matrix[i][i].edge ) 
      Table_lookup[i].nparent_cats--;
  
  /* Pass repeatedly over all tables, dumping SQL for all those which
     depend on tables already dumped, until no more remain */
  do {
    for ( i = 0, tables_out = 0; i < ntables; i++) {
      if ( !Table_lookup[i].nparent_cats ) {
	cat2sql(sqlout, Table_lookup[i].p);
	tables_out++;
	for ( j = 0; j < ntables; j++ ) {
	  if ( i == j ) continue;
	  if ( Edge_matrix[j][i].edge ) Table_lookup[j].nparent_cats--;
	}
	Table_lookup[i].nparent_cats--;
      }
    }

    fprintf (stderr,"%d tables this pass\n", tables_out);

  } while ( tables_out );

  fclose(sqlout);
}

void zzs_print_keys (void) {

  int i;
  Sym *cat, *item;


  printf("--- Start: Key items per category ---\n");

  /* Loop over all symbol table entries */
  for (i = 0; i < size; i++) {
    for (cat = table[i]; cat; cat = cat->next) {
      
      /* If current one is a category, write out all key items */
      if ( cat->category.rec == cat ) {
	printf("%s ",zzs_recname(cat));
	/* Loop over all key items */
	for ( item = cat->key.rec; item && item != cat; item = item->key.rec )
	  printf("%s ",zzs_recname(item));

	printf("\n");
      }
    }
  }
  printf("--- End: Key items per category ---\n");
}
	       
