/*
  Copyright (c) Navraj Pannu       1998 - 2005
  Copyright (c) Leiden University  2002 - 2005

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

#include <ctype.h>
#include "afroinput.h"

Afroinput::Afroinput(Model &model, Crystal &xtl, Afrolikelihood &likelihood,
		     Minimizer &minimizer)
  : Input(model, xtl, likelihood, minimizer), like(likelihood)
{
  // parse 
  CCP4parse();
  
  // clean up parser array
  CCP4::ccp4_parse_end(parser);

  check();

  if (getenv("HKLIN") != NULL)
    xtal.setwithmtz("ALL");
  else
    xtal.setwithmanymtz("ALL");  
  
  print();
}

void Afroinput::check() 
{
  // checks if input is okay
  Input::check();

  if (mdl.atom.size() <= 0)
    Bp3Error("Afroinput::check", "No atoms defined");
  
  for (unsigned d = 0; d < xtal.sf.size(); d++)
  {    
    if ( (xtal.sf[d].Getsdatap() == "") || (xtal.sf[d].Getsdevp() == "") )
      Bp3Error("Afroinput::check", "Dataset declared without F and/or SIGF columns");
    if (xtal.sf[d].Gettype() != "AMPLITUDE")
      Bp3Error("Afrolikelihood::setup","Input amplitudes only");
  }
}

void Afroinput::CCP4parse()
{
  // The setup of reading keyword input and it follows pretty much
  // exactly Peter Briggs' documentation of his parser.

  // read lines from stdin until END or EOF is reached
  bool cont       = true;
  char line[201];

  while (cont)
  {
    // Blank line before calling ccp4_parser implies reading from stdin
    line[0]       = '\0';
    unsigned ntok = CCP4::ccp4_parser(line, 200, parser, 1);
    
    if (ntok      < 1)                                            // reached EOF
      break;
    else   // perform keyword analysis and interpretation
    {
      if (Input::xtalkeywords())
	continue;
      else if (Input::datakeywords())
	continue;
      else if (Input::modelkeywords())
	continue;
      else if (Input::minimizerkeywords())
	continue;
      else if (Input::likelihoodkeywords())
	continue;
      else if (afrokeywords())
	continue;
      else if (CCP4::ccp4_keymatch("END", parser->keyword))
	break;
      else
	Bp3Error("CCP4parse","Keyword not recognised");
    }
  }  
}

bool Afroinput::afrokeywords()
{
  if (CCP4::ccp4_keymatch("ASCI", parser->keyword))       // mode options
    ascikeyword();
    else if (CCP4::ccp4_keymatch("HIRE", parser->keyword))
    hirekeyword();
  else if (CCP4::ccp4_keymatch("EVAL", parser->keyword))
    evalkeyword();
  else if (CCP4::ccp4_keymatch("DFFE", parser->keyword))
    dffekeyword();
  else if (CCP4::ccp4_keymatch("GIAC", parser->keyword))
    giackeyword();
  else if (CCP4::ccp4_keymatch("DELT", parser->keyword))
    deltkeyword();
  else
    return false;

  return true;
}

void Afroinput::ascikeyword()
{
  // check to see if we have no more than one token
  if (parser->ntokens > 1)
    Bp3Warning("Afroinput::ascikeyword", "Extra text in ASCIi keyword will be ignored");
  
  like.Setascii(true);
}

void Afroinput::hirekeyword()
{
  // check to see if we have no more than one token
  if (parser->ntokens > 2)
    Bp3Warning("Afroinput::hirekeyword", "Extra text in HIREs keyword will be ignored");

  if (parser->token[1].isnumber)
    like.Setsubhires(parser->token[1].value);
  else
    Bp3Error("Afroinput::hirekeyword", "A number is expected after the HIREs keyword");

}

void Afroinput::evalkeyword()
{
  // check to see if we have no more than one token
  if (parser->ntokens > 1)
    Bp3Warning("Afroinput::evalkeyword", "Extra text in EVALue keyword will be ignored");
  
  like.Setmode("EVALUES");
}

void Afroinput::dffekeyword()
{
  // check to see if we have no more than one token
  if (parser->ntokens > 1)
    Bp3Warning("Afroinput::dffekeyword", "Extra text in DFFE keyword will be ignored");
  
  like.Setmode("DIFFE");
}

void Afroinput::giackeyword()
{
  // check to see if we have no more than one token
  if (parser->ntokens > 1)
    Bp3Warning("Afroinput::giackeyword", "Extra text in GIAC keyword will be ignored");
  
  like.Setmode("GIAC");
}

void Afroinput::deltkeyword()
{
  // check to see if we have no more than one token
  if (parser->ntokens > 1)
    Bp3Warning("Afroinput::deltkeyword", "Extra text in DELT keyword will be ignored");
  
  like.Setmode("DELTA");
}
