/*
  Copyright (c) Navraj Pannu       1998 - 2007
  Copyright (c) Leiden University  2002 - 2007

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

#include <math.h>
#include <time.h>
#include "afrolikelihood.h"
#include "csymlib.h"
#include "gauss.h"

Afrolikelihood::Afrolikelihood(Model &model, Crystal &crystal)
  : Likelihood(model, crystal)
{
  // defaults
  likelihood    = ZERO; cd = 0;
  mode          = "DELTA";
  title         = "EA values from AFRO";
  allin         = stats = false;
  subhires      = 1000.0;
  interpolate   = true;
  output        = "heavy";
  ascii         = false;
  refineluzzati = refineocc  = false;       
  refinescale   = refinexyz  = false;
  refinefp      = refinefpp  = false;
  refinebfac    = refinenumb = true;
  sadweight.resize(35);
  lweight.resize(100);
}

void Afrolikelihood::setup(const bool refine)
{
  // Determine how many isomorphism and atomic parameters there are  
  // to refine and allocate memory for variables storing gradients.

  checkdata();

  resize();

  checkdata();

  Setresocutoff();
  
  xtal.Setselected("REFINE");

  checkdata();
  
  xtal.Setnormalization().SetLuzzati();

  checkdata();

  if (refine && ( (mode == "DIFFE") || (mode == "GIAC") ) )
  {
    unsigned nscalepars(0);
    unsigned nluzzatipars(0);
    unsigned natmpars(0);

    for (unsigned d             = 0; d < xtal.sf.size(); d++)
    {
      if (xtal.sf[d].refinek && refinescale)
	nscalepars++;
      if (xtal.sf[d].refineb && refinescale)
	nscalepars++;

      if (xtal.sf[d].refined && refineluzzati)
	nluzzatipars         += xtal.sf[d].nbins;
      if (xtal.sf[d].refinee && refineluzzati)
	nluzzatipars         += xtal.sf[d].nbins;             
      if (xtal.sf[d].refinead && refineluzzati)
	nluzzatipars         += xtal.sf[d].nbins;
      if (xtal.sf[d].refinesd && refineluzzati)
	nluzzatipars         += xtal.sf[d].nbins;      
    }

    // Atom parameters

    for (unsigned a             = 0; a < mdl.atom.size()  ; a++)
    {
      if (mdl.atom[a].Getrefinebfac() && refinebfac)
	if (mdl.atom[a].Getisotropic() )
	  natmpars++;
	else  
	  natmpars             += 6;
      if (mdl.atom[a].Getrefinenumb() && refinenumb)
	natmpars++;      
    }

    for (unsigned f             = 0; f < mdl.form.size()  ; f++)
      for (unsigned w           = 0; w < mdl.form[f].Getnwave(); w++)
      {
	if (mdl.form[f].Getrefinefp(w) && refinefp)
	  natmpars++;
	if (mdl.form[f].Getrefinefpp(w) && refinefpp)
	  natmpars++;
      }

    unsigned npars(natmpars + nluzzatipars + nscalepars);
    printf("Number of parameters being refined: %u\n", npars);
    printf("Consisting of...\n");
    if (nscalepars)
      printf("Scale parameters: %u\n", nscalepars);
    if (nluzzatipars)
      printf("Luzzati parameters: %u\n", nluzzatipars);
    if (natmpars)
      printf("Atomic parameters: %u\n\n", natmpars);

    Settype();

    // check for reflections with low-likelihood
    grad(true);
  }
  
  xtal.print();

  checkdata();

}

void Afrolikelihood::resize()
{
  fa.resize(xtal.maxref, (double)NOTUSED);
  sigfa.resize(xtal.maxref, (double)NOTUSED);
  /*
  sigmfa.resize(xtal.maxref, (double)ZERO);
  ssfpp.resize(xtal.maxref, (double)ZERO);
  ssfpfpp.resize(xtal.maxref, (double)ZERO);
  dLdssfpp.resize(xtal.maxref, (double)ZERO);
  dLdssfpfpp.resize(xtal.maxref, (double)ZERO);
  */
  
  if ( (mode            == "DIFFE") || (mode == "GIAC") )
  {
        // ***NSP - should centralize resizing

    // inverse/eigenvalue filtering
    unsigned row(3);
    evalues.resize(row);

    lwork                = 1 + 6*row + 2*row*row;
    liwork               = 3 + 5*row;
    lawork.resize(lwork);
    iwork.resize(liwork);

    // Sad function
    Gauss gausslegendre("Legendre", sadweight.size());
    vector<double> snode = gausslegendre.Getnode();
    sadweight            = gausslegendre.Getweight();
    sadsin.resize(sadweight.size());
    sadcos.resize(sadweight.size());
    // Precompute sin, cos
    for (unsigned i      = 0; i < sadweight.size(); i++)
    {
      sadcos[i]          = tab.tcos(PI*snode[i]);
      sadsin[i]          = tab.tsin2(PI*snode[i]);
    }
    
    Gauss gausshermite("Hermite", lweight.size());
    lnode                = gausshermite.Getnode();
    lweight              = gausshermite.Getweight();

    gausshermite.print();
    
    // ***NSP - should centralize with likelihood.C/h
    sumfpp.resize(xtal.sf.size()), sumfpfpp.resize(xtal.sf.size());
    difffpp.resize(xtal.sf.size()), difffp.resize(xtal.sf.size());

    for (unsigned d      = 0; d < xtal.sf.size(); d++)
    {
      sumfpp[d].resize(xtal.sf.size());
      sumfpfpp[d].resize(xtal.sf.size());
      difffpp[d].resize(xtal.sf.size());
      difffp[d].resize(xtal.sf.size());
      for (unsigned d1   = 0; d1 < xtal.sf.size(); d1++)
      {
        sumfpp[d][d1].resize(xtal.sf[d].nbins, ZERO);
        sumfpfpp[d][d1].resize(xtal.sf[d].nbins, ZERO);
        difffpp[d][d1].resize(xtal.sf[d].nbins, ZERO);
        difffp[d][d1].resize(xtal.sf[d].nbins, ZERO);
      }
    }
    // derivatives for refinement

    // scale parameters
    dLdkscale.resize(xtal.sf.size(), ZERO);
    dLdbscale.resize(xtal.sf.size(), ZERO);

    // Real and imaginary F+ heavy atom structure factor (or just F if F+=F=)

    // Wilson's sigmaH parameter
    dLdsigmah.resize(xtal.sf.size());  
    dLdsumfpp.resize(xtal.sf.size());
    dLdsumfpfpp.resize(xtal.sf.size());
  
    for (unsigned d        = 0; d < xtal.sf.size(); d++)
    {
      dLdsigmah[d].resize(xtal.sf[d].nbins, ZERO);
      dLdsumfpp[d].resize(xtal.sf[d].nbins, ZERO);
      dLdsumfpfpp[d].resize(xtal.sf[d].nbins, ZERO);
    }
    
    dLdnumb.resize(mdl.atom.size(), ZERO);
    dLdbiso.resize(mdl.atom.size(), ZERO);
    dLduaniso.resize(mdl.atom.size());
    
    for (unsigned a      = 0; a < mdl.atom.size(); a++)
      dLduaniso[a].resize(6, ZERO);

    dLdfp.resize(mdl.form.size());
    dLdfpp.resize(mdl.form.size());

    for (unsigned f      = 0; f < mdl.form.size(); f++)
    {
      dLdfp[f].resize(mdl.form[f].nwave, ZERO);
      dLdfpp[f].resize(mdl.form[f].nwave, ZERO);
    }
  }
}

void Afrolikelihood::Setresocutoff()
{
    // determine highest-resolution data set and set that to reference
  double subtemp(1000.0);
  
  for (unsigned d             = 0; d < xtal.sf.size(); d++)
    if (xtal.sf[d].hires      < subtemp)
      for (unsigned a         = 0; a < mdl.atom.size(); a++)
	if (xtal.sf[d].nxtal == mdl.atom[a].crystal)
	  subtemp             = xtal.sf[d].hires;

  if (subhires                > 999.0)
  {
    printf("HALF plus highest resolution data set limit is %.2f\n\n",subtemp+0.5);
    subhires                  = subtemp + HALF;
    if (xtal.mad)
    {
      double suboverthirty    = calcanomcorrel();
      if ( (suboverthirty     > ZERO) && (suboverthirty < 6.0) )
      {
	printf("Anomalous signed correlation coefficient under 0.3 at resolution: %.2f\n\n",suboverthirty);
	subhires              = suboverthirty;
      }
    }
  }

  printf("Using substructure detection resolution cut-off: %.2f\n\n",subhires);
}

void Afrolikelihood::calcsigmah()
{

  for (unsigned a                    = 0; a < mdl.atom.size(); a++)
  {  
    dLdnumb[a]                       = ZERO;
    dLdbiso[a]                       = ZERO;
    for (unsigned b                  = 0; b < 6; b++)
      dLduaniso[a][b]                = ZERO;
  }

  for (unsigned f                    = 0; f < mdl.form.size(); f++)
    for (unsigned w                  = 0; w < mdl.form[f].nwave; w++)
    {
      dLdfp[f][w]                    = ZERO;
      dLdfpp[f][w]                   = ZERO;
    }

  for (unsigned d                    = 0; d < xtal.sf.size()  ; d++)
    for (unsigned s                  = 0; s < xtal.sf[d].nbins; s++)
    {
      xtal.sf[d].sigmah[s]           = ZERO;
      for (unsigned d2               = 0; d2 < xtal.sf.size(); d2++)
      {
	sumfpp[d][d2][s]             = ZERO;
	sumfpfpp[d][d2][s]           = ZERO;
	difffpp[d][d2][s]            = ZERO;
	difffp[d][d2][s]             = ZERO;
      }
    }  

  for (unsigned r                    = 0; r < xtal.maxref; r++)
  {
    // ***NSP
    sigmfa[r]                        = ZERO;
    ssfpp[r]                         = ZERO;
    ssfpfpp[r]                       = ZERO;
    
    if (xtal.selected[r])
      for (unsigned a                = 0; a < mdl.atom.size(); a++)
      {	
	// the crystal the atom belongs to
	unsigned c(mdl.atom[a].crystal);
	double stolsq(xtal.cell[c].stolsq[r]);

	double bfactor;
	vector<double> coefbano(6, ZERO);

	if (mdl.atom[a].isotropic)
	  bfactor                    = tab.texp(-mdl.atom[a].biso*stolsq);
	else
	{
	  double h(xtal.miller[r][0]), k(xtal.miller[r][1]), l(xtal.miller[r][2]);
	  coefbano[0]                = h*h*xtal.cell[c].astar*xtal.cell[c].astar;
	  coefbano[1]                = TWO*h*k*xtal.cell[c].astar*xtal.cell[c].bstar;
	  coefbano[2]                = TWO*h*l*xtal.cell[c].astar*xtal.cell[c].cstar;
	  coefbano[3]                = k*k*xtal.cell[c].bstar*xtal.cell[c].bstar;
	  coefbano[4]                = TWO*k*l*xtal.cell[c].bstar*xtal.cell[c].cstar;
	  coefbano[5]                = l*l*xtal.cell[c].cstar*xtal.cell[c].cstar;
	  bfactor                    = tab.texp(-TWOPI2*std::max(coefbano[0]*
								 mdl.atom[a].uaniso[0] +
								 coefbano[1]*
								 mdl.atom[a].uaniso[1]+
								 coefbano[2]*
								 mdl.atom[a].uaniso[2]+
								 coefbano[3]*
								 mdl.atom[a].uaniso[3]+
								 coefbano[4]*
								 mdl.atom[a].uaniso[4]+
								 coefbano[5]*
								 mdl.atom[a].uaniso[5],
								 ZERO)); 
	}
	double btimeso(bfactor*mdl.atom[a].occ);	

	// Calculate the normal scattering factor for atom and resolution

	unsigned f(mdl.atom[a].nform);
	
	double normalscat((double)(mdl.form[f].a[0]*
				   tab.texp(-mdl.form[f].b[0]*stolsq) +
				   mdl.form[f].a[1]*
				   tab.texp(-mdl.form[f].b[1]*stolsq) +
				   mdl.form[f].a[2]*
				   tab.texp(-mdl.form[f].b[2]*stolsq) +
				   mdl.form[f].a[3]*
				   tab.texp(-mdl.form[f].b[3]*stolsq) +
				   mdl.form[f].c));

	// calculate anomalous scattering for each wavelength
	for (unsigned w              = 0; w < mdl.form[f].fp.size(); w++)
	{
 	  unsigned d(xtal.dataset[c][w]);
	  if (xtal.sf[d].use(r))
	  {
	    unsigned sh(xtal.bin(d,r));
	    double scat(normalscat + mdl.form[f].fp[w]); 
	    xtal.sf[d].sigmah[sh]   += scat*scat*btimeso*btimeso*mdl.atom[a].number;
	    if (xtal.sf[d].anomalous)
	    {
	      double fpp(mdl.form[f].fpp[w]);
	      double fp(mdl.form[f].fp[w]*btimeso);
	      fpp                   *= btimeso;

	      sigmfa[r]             += (scat*scat*btimeso*btimeso +
                                        fpp*fpp)*mdl.atom[a].number;
              ssfpp[r]              += TWO*fpp*fpp*mdl.atom[a].number;
              ssfpfpp[r]            += fpp*scat*btimeso*mdl.atom[a].number;

	      xtal.sf[d].sigmah[sh] += fpp*fpp*mdl.atom[a].number;
	      sumfpp[d][d][sh]      += TWO*fpp*fpp*mdl.atom[a].number;
	      sumfpfpp[d][d][sh]    += fpp*scat*btimeso*mdl.atom[a].number;

	      /*
	      double mult((double)(xtal.sg.NSYMP));
              dLdnumb[a]            += (dLdssfpp[r]*TWO*fpp*fpp +
                                        dLdssfpfpp[r]*fpp*scat*btimeso)*mult;
              if (mdl.atom[a].isotropic)
                dLdbiso[a]          -= (dLdssfpp[r]*TWO*fpp*fpp +
                                        dLdssfpfpp[r]*fpp*scat*btimeso
                                        )*TWO*stolsq*mdl.atom[a].number*mult;
              else
                for (unsigned u      = 0; u < 6; u++)
                  dLduaniso[a][u]   -= (dLdssfpp[r]*TWO*fpp*fpp +
                                        dLdssfpfpp[r]*fpp*scat*btimeso
                                        )*TWO*TWOPI2*mdl.atom[a].occ*coefbano[u]*mdl.atom[a].number*mult;
	      */
	      double mult((double)(xtal.sg.NSYMP)/
			  ((double)xtal.sf[d].anonshl[sh]));

	      dLdnumb[a]            += (dLdsumfpp[d][sh]*TWO*fpp*fpp +
					dLdsumfpfpp[d][sh]*fpp*scat*btimeso)*mult;
	      if (mdl.atom[a].isotropic)
		dLdbiso[a]          -= (dLdsumfpp[d][sh]*TWO*fpp*fpp +
					dLdsumfpfpp[d][sh]*fpp*scat*btimeso
					)*TWO*stolsq*mdl.atom[a].number*mult;
	      else
		for (unsigned u      = 0; u < 6; u++)
		  dLduaniso[a][u]   -= (dLdsumfpp[d][sh]*TWO*fpp*fpp +
		                        dLdsumfpfpp[d][sh]*fpp*scat*btimeso
					)*mdl.atom[a].number;
	      
	      for (unsigned w2       = w+1; w2 < mdl.form[f].fp.size(); w2++)
	      {
		unsigned d2(xtal.dataset[c][w2]);
		double fp2(mdl.form[f].fp[w2]*btimeso);
		double fpp2(mdl.form[f].fpp[w2]*btimeso);
		difffp[d][d2][sh]   += scat*(fp2 - fp)*btimeso*mdl.atom[a].number;
		difffpp[d][d2][sh]  += fpp*(fpp2 - fpp)*mdl.atom[a].number;
		sumfpp[d][d2][sh]   += fpp*(fpp2 + fpp)*mdl.atom[a].number;
	      }
	    }
	  }
	}
      }
  }
  
  for (unsigned d                    = 0; d < xtal.sf.size()   ; d++)
    for (unsigned s                  = 0; s < xtal.sf[d].nbins ; s++)
    {
      if (xtal.sf[d].nshl[s])
 	xtal.sf[d].sigmah[s]        *= ((double)(xtal.sg.NSYMP)/
					((double)(xtal.sf[d].nshl[s])));
      
      if (xtal.sf[d].anomalous)
 	if (xtal.sf[d].anonshl[s])
	  for (unsigned d2           = 0; d2 < xtal.sf.size()  ; d2++)
	  {
	    sumfpp[d][d2][s]        *= ((double)(xtal.sg.NSYMP)/
					((double)xtal.sf[d].anonshl[s]));
	    sumfpfpp[d][d2][s]      *= ((double)(xtal.sg.NSYMP)/
					((double)xtal.sf[d].anonshl[s]));
	    difffpp[d][d2][s]       *= ((double)(xtal.sg.NSYMP)/
					((double)xtal.sf[d].anonshl[s]));
	    difffp[d][d2][s]        *= ((double)(xtal.sg.NSYMP)/
					((double)xtal.sf[d].anonshl[s]));
	  }
    }    
}

void Afrolikelihood::outputeas()
{

  if (mode == "DELTA")
  {
    calcanodiff();
    calcdeltafa();
  }
  else
  {
    calcsigmah();
    xtal.print();
    mdl.print();
    if (mode == "GIAC")
      giacovazzo();
    else
      sadfunction();
  }

  downweighoutliers();
  calcsigmafa();

  if (xtal.sf[0].sfmodel.size())
    calccorrel();

  renormalize();

  writemtz();
  if (xtal.sf[0].sfmodel.size())
    calccorrel();

  if (ascii)
    writedrear();
}

void Afrolikelihood::calccorrel()
{
  unsigned d(0);
  vector<double> sumx(xtal.sf[d].nbins, ZERO);
  vector<double> sumx2(xtal.sf[d].nbins, ZERO);
  vector<double> sumy(xtal.sf[d].nbins, ZERO);
  vector<double> sumy2(xtal.sf[d].nbins, ZERO);
  vector<double> sumz(xtal.sf[d].nbins, ZERO);
  vector<double> sumz2(xtal.sf[d].nbins, ZERO);
  vector<double> sumxy(xtal.sf[d].nbins, ZERO);
  vector<double> sumxz(xtal.sf[d].nbins, ZERO);
  vector<double> sumyz(xtal.sf[d].nbins, ZERO);
  vector<double> sum(xtal.sf[d].nbins, ZERO);
  
  double totalsumx(ZERO), totalsumx2(ZERO), totalsumy(ZERO), totalsumy2(ZERO);
  double totalsumz(ZERO), totalsumz2(ZERO), totalsumxy(ZERO), totalsumxz(ZERO);
  double totalsumyz(ZERO), totalsum(ZERO);

  unsigned rstart(0);
  
  for (unsigned r = 0; r < xtal.maxselref; r++)
    if ( (fa[r]   > NOTUSED) && xtal.sf[d].anouse(r) &&
	 (xtal.Getres(d,r) > subhires) )
    {
      totalsum   += ONE;
      totalsumx   = fa[r];
      totalsumy   = fabs(xtal.sf[d].dano(r));
      totalsumz   = xtal.sf[d].fmodel[r];
      rstart      = r+1;
      break;
    }

  for (unsigned r = rstart; r < xtal.maxselref; r++)
    if ( (fa[r]   > NOTUSED) && xtal.sf[d].anouse(r) &&
	 (xtal.Getres(d,r) > subhires) )
    {
      unsigned s(xtal.bin(d,r)); 
      sumx[s]    += fa[r];
      sumy[s]    += fabs(xtal.sf[d].dano(r));
      sumz[s]    += xtal.sf[d].fmodel[r]; 
      sumxy[s]   += fa[r]*fabs(xtal.sf[d].dano(r));
      sumxz[s]   += fa[r]*xtal.sf[d].fmodel[r];
      sumyz[s]   += fabs(xtal.sf[d].dano(r))*xtal.sf[d].fmodel[r];
      sumx2[s]   += fa[r]*fa[r]; 
      sumy2[s]   += xtal.sf[d].dano(r)*xtal.sf[d].dano(r);
      sumz2[s]   += xtal.sf[d].fmodel[r]*xtal.sf[d].fmodel[r];
      sum[s]     += ONE;

      totalsum   += ONE;
      double temp2((totalsum - ONE)/totalsum);
      double deltax(fa[r] - totalsumx), deltay(fabs(xtal.sf[d].dano(r)) - totalsumy);
      double deltaz(xtal.sf[d].fmodel[r] - totalsumz);
      totalsumxy += deltax*deltay*temp2;
      totalsumxz += deltax*deltaz*temp2;
      totalsumyz += deltay*deltaz*temp2;
      totalsumx2 += deltax*deltax*temp2;
      totalsumy2 += deltay*deltay*temp2;
      totalsumz2 += deltaz*deltaz*temp2;
      totalsumx  += deltax/totalsum;
      totalsumy  += deltay/totalsum;
      totalsumz  += deltaz/totalsum;
    }

  double totals(ZERO), totalcor(ZERO), totalcor1(ZERO), correl(ZERO);

  printf("Correlation of multivariate FA with Fcalc\n\n");
  printf(" Bin   HiRes  LoRes        N     Correlation\n");
  totals          = totalcor = ZERO;
  for (unsigned s = 0; s < xtal.sf[d].nbins; s++)
  {
    if (sum[s]    > ZERO)
      correl      = ((sum[s]*sumxz[s] - sumx[s]*sumz[s])/
                     (sqrt(sum[s]*sumx2[s] - sumx[s]*sumx[s])*
                      sqrt(sum[s]*sumz2[s] - sumz[s]*sumz[s])));
    else
      correl      = ZERO;

    printf("%3d   %6.2f %6.2f     %5.0f     %6.4f\n",s+1,xtal.sf[d].Getlores(s),
	   xtal.sf[d].Gethires(s),sum[s],correl);
    totalcor     += correl*sum[s];
    totals       += sum[s];
  }
  double t1(totalsumxz/totals), t2(sqrt(totalsumx2/totals)), t3(sqrt(totalsumz2/totals));
  
  totalcor1       = t1/(t2*t3);

  printf("TOTAL %6.2f %6.2f     %5.0f     %6.4f %6.4f\n\n",xtal.sf[d].Getlores(0),
	 xtal.sf[d].Gethires(xtal.sf[d].nbins-1),totals,totalcor/totals,totalcor1);

  printf("Correlation of Fcalc with DeltaF\n\n");
  printf(" Bin   HiRes  LoRes        N     Correlation\n");
  totals          = totalcor = correl = ZERO;
  for (unsigned s = 0; s < xtal.sf[d].nbins; s++)
  {
    if (sum[s]    > ZERO)
      correl      = ((sum[s]*sumyz[s] - sumy[s]*sumz[s])/
                     (sqrt(sum[s]*sumy2[s] - sumy[s]*sumy[s])*
                      sqrt(sum[s]*sumz2[s] - sumz[s]*sumz[s])));
    else
      correl      = ZERO;
    printf("%3d   %6.2f %6.2f     %5.0f     %6.4f\n",s+1,xtal.sf[d].Getlores(s),
	   xtal.sf[d].Gethires(s),sum[s],correl);

    totalcor     += correl*sum[s];
    totals       += sum[s];
  }

  double t4(totalsumyz/totals), t5(sqrt(totalsumy2/totals)), t6(sqrt(totalsumz2/totals));
  
  totalcor1       = t4/(t5*t6);

  printf("TOTAL %6.2f %6.2f     %5.0f     %6.4f %6.4f\n\n",xtal.sf[d].Getlores(0),
	 xtal.sf[d].Gethires(xtal.sf[d].nbins-1),totals,totalcor/totals,totalcor1);
}

double Afrolikelihood::calcanomcorrel() const
{
  const double FPPCUT(1.5);
  
  double resover                = 0.0;
  
  if (xtal.sf.size()            > 1)
    for (unsigned d1            = 0; d1 < xtal.sf.size(); d1++)
      for (unsigned d2          = d1+1; d2 < xtal.sf.size(); d2++)
	if (xtal.sf[d1].anomalous && xtal.sf[d2].anomalous &&
	    (xtal.sf[d1].nxtal == xtal.sf[d2].nxtal) )
	{
	  bool lowfpp(false);
	  unsigned d            = ( (xtal.sf[d1].hires > xtal.sf[d2].hires) ?
				    d1 : d2);
	  
 	  vector<double> sumx(xtal.sf[d].nbins, ZERO);
	  vector<double> sumx2(xtal.sf[d].nbins, ZERO);
	  vector<double> sumy(xtal.sf[d].nbins, ZERO);
	  vector<double> sumy2(xtal.sf[d].nbins, ZERO);
	  vector<double> sumxy(xtal.sf[d].nbins, ZERO);
	  vector<double> sum(xtal.sf[d].nbins, ZERO);

	  for (unsigned f       = 0; f < mdl.form.size(); f++)
	    for (unsigned w     = 0; w < mdl.form[f].nwave; w++)
	    {
	      unsigned c(mdl.form[f].Getcrystal());
	      if ( (mdl.form[f].fpp[w] < FPPCUT) &&
		   ( (xtal.dataset[c][w] == d1) || (xtal.dataset[c][w] == d2) ) )
		lowfpp          = true;
	    }
	  
	  
	  for (unsigned r       = 0; r < xtal.maxselref; r++)
	    if (xtal.sf[d1].anouse(r) && xtal.sf[d2].anouse(r))
	    {
	      unsigned s(xtal.bin(d,r));

	      sumx[s]          += xtal.sf[d1].dano(r);
	      sumy[s]          += xtal.sf[d2].dano(r);
	      sumxy[s]         += xtal.sf[d1].dano(r)*xtal.sf[d2].dano(r);
	      sumx2[s]         += xtal.sf[d1].dano(r)*xtal.sf[d1].dano(r); 
	      sumy2[s]         += xtal.sf[d2].dano(r)*xtal.sf[d2].dano(r);
 	      sum[s]           += ONE;
	    }

 	  double totals(ZERO), totalcor(ZERO), correl(ZERO);
	  printf("\n $TABLE: Signed anomalous difference correlation: %s and %s\n",xtal.sf[d1].name.c_str(),
		 xtal.sf[d2].name.c_str());
	  printf("$GRAPHS: Signed anomalous difference correlation vs. 1/Res^2 :N:4,6:\n");
	  printf("$$\n Bin   HiRes  LoRes  1/Res^2   Refls  Correlation $$\n$$\n");
	  double lastcorrel(ZERO), lastreso(ZERO);
	  bool underthirty(false);
	  for (unsigned s       = 0; s < xtal.sf[d].nbins; s++)
	  {
	    if (sum[s]          > ZERO)
	      correl            = ((sum[s]*sumxy[s] - sumx[s]*sumy[s])/
				   (sqrt(sum[s]*sumx2[s] - sumx[s]*sumx[s])*
				    sqrt(sum[s]*sumy2[s] - sumy[s]*sumy[s])));
 	    else
	      correl            = ZERO;

	    double reso         = HALF/sqrt(xtal.sf[d].astolsq[s]);
	      
	    if ( (correl        < 0.300) && (lastcorrel > 0.300) && (!underthirty)
		 && (!lowfpp) )
	    {
	      double tmp        = (lastreso-reso)/(lastcorrel-correl)*(0.3-correl) + reso;
	      underthirty       = true;
	      if ( (resover     > tmp) || (resover == ZERO) )
		resover         = tmp;
	    }
	    
	    lastcorrel          = correl;
	    lastreso            = reso;

 	    printf(" %2d  %6.2f  %6.2f  %7.5f  %5.0f  %7.5f\n",s+1,xtal.sf[d].Getlores(s),
		   xtal.sf[d].Gethires(s), xtal.sf[d].astolsq[s],sum[s],correl);
 	    totalcor           += correl*sum[s];
	    totals             += sum[s];
	  }
	  if (!underthirty     && (resover == ZERO) && !(lowfpp))
	    if (correl         >= 0.3)
	      resover           = xtal.sf[d].hires;
	  printf("$$\nTOTAL\n");
	  printf("     %6.2f  %6.2f           %5.0f  %7.5f\n\n",xtal.sf[d].Getlores(0),
		 xtal.sf[d].Gethires(xtal.sf[d].nbins-1),totals,totalcor/totals);
	}
		   return resover;
}

void Afrolikelihood::checkdata()
{
  for (unsigned r = 0; r < xtal.maxref; r++)
    if (xtal.selected[r])
      for (unsigned d = 0; d < xtal.sf.size(); d++)
	if (xtal.sf[d].anomalous)
	  if (xtal.sf[d].anouse(r))
	    if (fabs(xtal.sf[d].dano(r)) > 1000.0)
	      printf("%u %u %u %u %f %f\n",xtal.miller[r][0],xtal.miller[r][1], xtal.miller[r][2],cd,xtal.sf[cd].dano(r),xtal.sf[cd].sigdano(r));
}

void Afrolikelihood::calcanodiff()
{
  // determine whether anomalous differences are to be used from which data set and a cutoff

  vector<vector<double> > anosignal(xtal.sf.size()), isosignal(xtal.sf.size());
  vector<vector<double> > anomean(xtal.sf.size()), isomean(xtal.sf.size());
  vector<vector<double> > anomoment(xtal.sf.size()), isomoment(xtal.sf.size());
  vector<vector<unsigned> > anoninbin(xtal.sf.size()), isoninbin(xtal.sf.size());
  vector<double> totalanosignal(xtal.sf.size(), ZERO);
  vector<double> totalisosignal(xtal.sf.size(), ZERO);
  
  for (unsigned d               = 0; d < xtal.sf.size(); d++)
  {
    anosignal[d].resize(xtal.sf[d].nbins, ZERO);
    isosignal[d].resize(xtal.sf[d].nbins, ZERO);
    anomean[d].resize(xtal.sf[d].nbins, ZERO);
    isomean[d].resize(xtal.sf[d].nbins, ZERO);
    anomoment[d].resize(xtal.sf[d].nbins, ZERO);
    isomoment[d].resize(xtal.sf[d].nbins, ZERO);
    anoninbin[d].resize(xtal.sf[d].nbins, 0);
    isoninbin[d].resize(xtal.sf[d].nbins, 0);
  }

  for (unsigned r               = 0; r < xtal.maxref; r++)
    if (xtal.selected[r])
    {
      vector<double> sigman(xtal.sf.size(), ZERO);
      double eps((double) xtal.epsilon[r]);

      for (unsigned d           = 0; d < xtal.sf.size(); d++)
	if (xtal.sf[d].use(r))
	{
	  unsigned sa(xtal.bin(d,r)), sb(sa);
	  sigman[d]             = xtal.sf[d].sigman[sa];
	  double dluz(xtal.sf[d].dluz[sa]);
	  double wa(ONE), wb(ZERO);
      
	  if (interpolate)
	  {
	    xtal.binweights(d, r, sa, sb, wa, wb);
	    sigman[d]           = wa*xtal.sf[d].sigman[sa] + wb*xtal.sf[d].sigman[sb];
	    dluz                = wa*xtal.sf[d].dluz[sa] + wb*xtal.sf[d].dluz[sb];
	  }

 	  if ( (d               > 0) && xtal.sf[0].use(r) ) 
	  {
	    double emean(xtal.sf[d].datamean(r)/sqrt(eps*sigman[d]));
	    double emean0(xtal.sf[0].datamean(r)/sqrt(eps*sigman[0]));
	    double temp(emean - dluz*sqrt(sigman[d]/sigman[0])*emean0);
	    isomean[d][sa]     += temp;
	    isomoment[d][sa]   += temp*temp;
	    double sd           = (xtal.sf[d].devmean(r)*
				   xtal.sf[d].devmean(r) +
				   dluz*dluz*sigman[d]/sigman[0]*
				   xtal.sf[0].devmean(r)*
				   xtal.sf[0].devmean(r));
	    if (sd              > ZERO)
	      isosignal[d][sa] += (fabs(xtal.sf[d].datamean(r) - dluz*
		 		        xtal.sf[0].datamean(r))/sqrt(sd));
  	    isoninbin[d][sa]++;
	  }
	  
 	  if (xtal.sf[d].anomalous)
	    if (xtal.sf[d].anouse(r) && !xtal.centric[r])
	    {
 	      double eplus(xtal.sf[d].datap[r]/sqrt(eps*sigman[d]));
	      double eminus(xtal.sf[d].datam[r]/sqrt(eps*sigman[d]));
	      double temp(eplus - eminus);
	      anosignal[d][sa] += fabs(xtal.sf[d].dano(r))/xtal.sf[d].sigdano(r);
	      anomean[d][sa]   += temp;
	      anomoment[d][sa] += temp*temp;
	      anoninbin[d][sa]++;
	    }
	}
    }

  for (unsigned d               = 0; d < xtal.sf.size(); d++)
  {
    printf("\n");
    printf(" $TABLE: Dataset %s Signal to noise ratio vs. ASTOL^2:\n",
           xtal.sf[d].name.c_str());    
    printf("$GRAPHS: Diso/SigDiso    vs.  STOL^2:N:4,6:\n");
                                         
    if (xtal.sf[d].anomalous)            
      printf("       : Dano/SigDano vs.  STOL^2:N:4,8:\n");
                                         
    printf("$$\n Bin  HiRes  LoRes   STOL^2    IsoRefls   Diso/SigDiso");
    if (xtal.sf[d].anomalous)
      printf("   AnoRefls    Dano/SigDano");
    printf(" $$\n$$\n");

    double totalastolsq(ZERO);
    unsigned totalisorefs(0), totalanorefs(0);

    for (unsigned s             = 0; s < xtal.sf[d].nbins; s++)
    {
      totalastolsq             += xtal.sf[d].astolsq[s];
      if (anoninbin[d][s])
      {
	totalanosignal[d]      += anosignal[d][s];
	totalanorefs           += anoninbin[d][s];
	anosignal[d][s]        /= anoninbin[d][s];
	anomean[d][s]          /= anoninbin[d][s];
	anomoment[d][s]        /= anoninbin[d][s];
      }
      if (isoninbin[d][s])
      {
	totalisosignal[d]      += isosignal[d][s];
	totalisorefs           += isoninbin[d][s];
	isosignal[d][s]        /= isoninbin[d][s];
	isomean[d][s]          /= isoninbin[d][s];
	isomoment[d][s]        /= isoninbin[d][s];
      }
      printf(" %2d  %6.2f %6.2f  %7.5f    %5i     %10.3f",
             s+1, xtal.sf[d].Getlores(s), xtal.sf[d].Gethires(s), xtal.sf[d].astolsq[s],
             isoninbin[d][s], isosignal[d][s]);
      if (xtal.sf[d].anomalous)
        printf("      %5i     %10.3f\n",anoninbin[d][s], anosignal[d][s]);
      else
        printf("\n");      
    }
    if (totalisorefs)
      totalisosignal[d]        /= (double) totalisorefs;

    if (totalanorefs)
      totalanosignal[d]        /= (double) totalanorefs;
    printf("$$\nTOTAL\n");
    printf("     %6.2f %6.2f  %7.5f    %5i     %10.3f",xtal.sf[d].Getlores(0),
           xtal.sf[d].Gethires(xtal.sf[d].nbins-1), totalastolsq/xtal.sf[d].nbins,
           totalisorefs, totalisosignal[d]);
      if (xtal.sf[d].anomalous)
        printf("      %5i     %10.3f\n\n",totalanorefs, totalanosignal[d]);
      else
        printf("\n\n");

    fflush(stdout);    
  }
  
  // determine whether anomalous or isomorphous differences should be used and
  // in the case of anomalous, which wavelength
  anodiff                       = true;
  for (unsigned d               = 0; d < xtal.sf.size(); d++)
    if (xtal.sf[d].anomalous)
      if (totalanosignal[d]     > totalanosignal[cd])
	cd                      = d;
  // bool reset(false);
  if ((xtal.sf.size()          == 2) && !xtal.mad)
    for (unsigned d             = 0; d < xtal.sf.size(); d++)
      if (d                    == 1)
	if ( ( (totalisosignal[d]
		                > 1.75*totalanosignal[d])
	       && (totalanosignal[d] < ONE) ) ||
	     (totalisosignal[d]   > FOUR*totalanosignal[d]) )
	{
	  anodiff               = false;
	  cd                    = 1;
	}
  
  double danocutoff(xtal.sf[cd].anosigmacut);
  double disocutoff(xtal.sf[cd].isosigmacut);
  
  if (anodiff)
  {
    printf("Anomalous differences from dataset %s will be used\n\n",
	   xtal.sf[cd].name.c_str());

    if (danocutoff                  < ZERO)
    {
      if (totalanosignal[cd]        > 2.0)  
	danocutoff                  = 0.5;
      else if (totalanosignal[cd]  <= 0.2)  
	danocutoff                  = DSMALL;
      else
	danocutoff                  = 0.1;
      if (xtal.verbose)
        printf("SANO is being set to %.2f\n\n",danocutoff);

      for (unsigned d               = 0; d < xtal.sf.size(); d++)
	xtal.sf[d].anosigmacut      = danocutoff;
    }
  }
  else
  {
    printf("Isomorphous differences from first and second data sets will be used\n\n"); 
    if (disocutoff                  < ZERO)
    {
      if (totalisosignal[cd]        > 2.5) 
	disocutoff                  = 0.5;
      else
	disocutoff                  = 0.1;
      printf("SISO is being set to %.2f\n\n",disocutoff);

      for (unsigned d               = 0; d < xtal.sf.size(); d++)
	xtal.sf[d].isosigmacut      = disocutoff;
    }
  }
}

void Afrolikelihood::calcdeltafa()
{
  vector<double> dluz(xtal.sf.size(), ZERO);
  vector<double> sigman(xtal.sf.size(), ZERO);
  vector<double> sigmanref(xtal.sf.size(), ZERO);

  for (unsigned r               = 0; r < xtal.maxref; r++)
  {
    fa[r]                       = NOTUSED;
    sigfa[r]                    = NOTUSED;

    if (xtal.selected[r])
      if (xtal.sf[cd].use(r) && (xtal.Getres(cd,r) > subhires))
 	if (anodiff && xtal.sf[cd].anouse(r))
	{
	  double fano(fabs(xtal.sf[cd].dano(r)));
	  double sigfano(xtal.sf[cd].sigdano(r));
	  if ( (fano            > ZERO) && (sigfano > ZERO) )
	    if ((fano/sigfano  >= xtal.sf[cd].anosigmacut) && !xtal.centric[r])
	    {
	      fa[r]             = fano;
	      sigfa[r]          = sigfano;
 	    }
	}
	else
	  if ( (cd              > 0) && xtal.sf[0].use(r) )
	  {
	    for (unsigned d     = 0; d < xtal.sf.size(); d++)
	      if (xtal.sf[d].use(r))
	      {
		unsigned sa(xtal.bin(d,r)), sb(sa);
		double wa(ONE), wb(ZERO);
		sigman[d]       = xtal.sf[d].sigman[sa];
		sigmanref[d]    = xtal.sf[d].sigmanref[sa];
		dluz[d]         = xtal.sf[d].dluz[sa];

		if (interpolate)
		{
		  xtal.binweights(d, r, sa, sb, wa, wb);
		  sigman[d]     = wa*xtal.sf[d].sigman[sa]   + wb*xtal.sf[d].sigman[sb];
		  sigmanref[d]  = (wa*xtal.sf[d].sigmanref[sa] +
			 	   wb*xtal.sf[d].sigmanref[sb]);
		  dluz[d]       = wa*xtal.sf[d].dluz[sa]     + wb*xtal.sf[d].dluz[sb];
		} 	 
	      }
	    
	    double signal       = fabs(xtal.sf[cd].datamean(r) - dluz[cd]*
		  		       sqrt(sigman[cd]/sigman[0])*
				       xtal.sf[0].datamean(r));
  	    double noise        = sqrt(xtal.sf[cd].devmean(r)*
		  		       xtal.sf[cd].devmean(r) +
				       dluz[cd]*dluz[cd]*sigman[cd]/sigman[0]*
				       xtal.sf[0].devmean(r)*
				       xtal.sf[0].devmean(r));
 	    if (noise           > ZERO)
	      if (signal/noise >= xtal.sf[cd].isosigmacut)
	      {
		fa[r]           = signal;
		sigfa[r]        = noise;
	      }	   
	  }
  }
}

void Afrolikelihood::downweighoutliers()
{
  // If an FA value is greater than maxval*sd, turn it to maxval*sd
  unsigned downweight(0), toosmall(0);
  const double maxval(5.0);
  unsigned d(cd);
  vector<double> mean(xtal.sf[d].nbins, ZERO);
  vector<double> moment(xtal.sf[d].nbins, ZERO);
  vector<double> ninbin(xtal.sf[d].nbins, ZERO);  

  for (unsigned r        = 0; r < xtal.maxref; r++)
    if (xtal.selected[r])
      if (fa[r]         != NOTUSED)
      {
	unsigned s(xtal.bin(d,r));

	mean[s]         += fa[r];
	moment[s]       += fa[r]*fa[r];
	ninbin[s]       += ONE;
      }	  

  vector<double> sdev(xtal.sf[d].nbins, ZERO);
  for (unsigned s        = 0; s < xtal.sf[d].nbins; s++)
    if (ninbin[s]        > ZERO)
    {
      sdev[s]            = sqrt(std::max((moment[s] -
					  mean[s]*mean[s]/ninbin[s])/
					 (ninbin[s] - ONE), ZERO));
      mean[s]           /= ninbin[s];
    }
    
  for (unsigned r        = 0; r < xtal.maxref; r++)
    if (xtal.selected[r])
      if (fa[r]         != NOTUSED)
      {
	unsigned s(xtal.bin(d,r));
	if (fa[r]        > mean[s] + maxval*sdev[s])
	{
	  fa[r]          = mean[s] + maxval*sdev[s];
	  downweight++;
	}
	else if (fa[r] < (mean[s] - maxval*sdev[s]))
	{
	  fa[r]          = NOTUSED;
	  toosmall++;
	}
	
      }
  
  if (downweight && xtal.verbose)
    printf("Number of reflections downweighted %u\n\n",downweight);

  if (toosmall)
    printf("Number of reflections with too small Fa estimates %u\n\n",toosmall);
}

void Afrolikelihood::calcsigmafa()
{
  unsigned d(cd);
  sigmafa.resize(xtal.sf[d].nbins, ZERO);
  vector<double> sumweight(xtal.sf[d].nbins, ZERO);
  for (unsigned s              = 0; s < xtal.sf[d].nbins; s++)
    sigmafa[s]                 = ZERO;
  
  for (unsigned r              = 0; r < xtal.maxref; r++)
    if (xtal.selected[r]      && (fa[r] != NOTUSED) )
    {
      double eps((double) (xtal.epsilon[r]));
      double weight((xtal.centric[r]) ? ONE : TWO);
      unsigned s(xtal.bin(d,r));
      sigmafa[s]              += weight*fa[r]*fa[r]/eps;
      sumweight[s]            += weight;
    }    

  for (unsigned s              = 0; s < xtal.sf[d].nbins; s++)
    if (sumweight[s]           > ZERO)
      sigmafa[s]              /= sumweight[s];

  printf("$$\n Bin  HiRes  LoRes   STOL2     Refls    SigmaF $$\n$$\n");
  for (unsigned s              = 0; s < xtal.sf[d].nbins; s++)
    printf(" %2d  %6.2f %6.2f  %7.5f   %5i  %10.2f\n",s+1,xtal.sf[d].Getlores(s),
	   xtal.sf[d].Gethires(s), xtal.sf[d].astolsq[s],xtal.sf[d].nshl[s],sigmafa[s]);
  printf("$$\n\n");
}

void Afrolikelihood::renormalize()
{
  unsigned d(cd);
  bool converge(false);
  unsigned cycles(0);
  do
  {
    vector<double> esquared(xtal.sf[d].nbins, ZERO);
    vector<double> total(xtal.sf[d].nbins, ZERO);

    for (unsigned r   = 0; r < xtal.maxref; r++)
      if (xtal.selected[r]   && (fa[r] != NOTUSED))
      {
	double eps((double) (xtal.epsilon[r]));
	unsigned sa(xtal.bin(d,r)), sb(sa);
	double wa(ONE), wb(ZERO), sfa(ZERO);      
	if (interpolate)
	{
	  xtal.binweights(d, r, sa, sb, wa, wb);
	  sfa         = wa*sigmafa[sa] + wb*sigmafa[sb];
	}
	else
	  sfa         = sigmafa[sa];

	esquared[sa] += fa[r]*fa[r]/(eps*sfa);
	total[sa]    += ONE;
      }

    converge          = true;

    for (unsigned s   = 0; s < xtal.sf[d].nbins; s++)
      if (total[s]    > 0)
      {
	double rescale(esquared[s]/total[s]);
	sigmafa[s]   *= rescale;
	if ( (rescale > 1.005) || (rescale < 0.995) )
 	  converge    = false;
      }
    cycles++;
    
  } while (!converge || (cycles > 100));
  printf("Number of renormalization cycles %u\n\n",cycles);
}

void Afrolikelihood::writemtz()
{
  unsigned d(0);
  CMtz::MTZ *MTZOUT            = CMtz::MtzMalloc(0,0);
  CMtz::MTZ *MTZIN             = CMtz::MtzGet("HKLIN",0);
  setupmtz(MTZOUT,MTZIN);

  unsigned inc                 = (allin) ? colin.size() : 3;
  unsigned columns(4);

  for (unsigned r              = 0; r < xtal.maxref; r++)
  {
    // Default value for columns to be written out is MNF
    vector<float> fdata(columns+inc, CCP4::ccp4_nan().f);
      
    storeinitialcolumns(MTZIN, fdata, inc, r);

    if (xtal.selected[r]      && (fa[r] != NOTUSED) &&
	(xtal.Getres(d,r)      > subhires))
    {
      double eps((double)(xtal.epsilon[r]));
      unsigned sa(xtal.bin(d,r)), sb(sa);
      double wa(ONE), wb(ZERO), sfa(ZERO);      

      if (interpolate)
      {
	xtal.binweights(d, r, sa, sb, wa, wb);
	sfa                    = wa*sigmafa[sa] + wb*sigmafa[sb];
      }
      else
	sfa                    = sigmafa[sa];

      fdata[inc    ]           = (float) fa[r];
      fdata[inc + 1]           = (float) sigfa[r];
      fdata[inc + 2]           = (float) (fa[r]/sqrt(eps*sfa));
      fdata[inc + 3]           = (float) (sigfa[r]/sqrt(eps*sfa));

      // store ea in fa[r] now - for sorting in drear file
      fa[r]                   /= sqrt(eps*sfa);
      sigfa[r]                /= sqrt(eps*sfa);
    } else
      fa[r]                    = sigfa[r] = NOTUSED;

    CMtz::ccp4_lwrefl(MTZOUT, &fdata[0], &colout[0], fdata.size(), r+1); 
  }

  if (allin)
    CMtz::MtzFree(MTZIN);
  CMtz::MtzPut(MTZOUT, " ");
  CMtz::MtzFree(MTZOUT);
}

void Afrolikelihood::writedrear()
{
  // method assumes vector fa stores Ea values
  unsigned d(0);
  FILE *dfile(NULL);  
  string filename             = output + ".drear";
  dfile                       = fopen(filename.c_str(),"w");

  for (unsigned r             = 0; r < xtal.maxref; r++)
  {
      unsigned pr(0);
      // find current maximal ea value
      for (unsigned rr        = 0; rr < xtal.maxref; rr++)
        if (xtal.selected[rr] && (fa[rr] > NOTUSED) &&
	    (xtal.Getres(d,rr) > subhires) )
        {
	  if (fa[rr]          > fa[pr])
	    pr                = rr;
        }

      double eps((double)(xtal.epsilon[pr]));
      unsigned sa(xtal.bin(d,pr)), sb(sa);
      double wa(ONE), wb(ZERO), sfa(ZERO);      
      
      if (interpolate)
      {
        xtal.binweights(d, pr, sa, sb, wa, wb);
        sfa                   = wa*sigmafa[sa] + wb*sigmafa[sb];
      }
      else
	sfa                   = sigmafa[sa];

      if (fa[pr]             != NOTUSED )
        fprintf(dfile, "%5d%5d%5d%9.2f%9.2f%9.3f%9.3f\n",
	        xtal.miller[pr][0], xtal.miller[pr][1], xtal.miller[pr][2],
	        fa[pr]*sqrt(eps*sfa), sigfa[pr]*sqrt(eps*sfa), fa[pr], sigfa[pr]);
      else
        break;
    
      fa[pr]                  = NOTUSED;  
    }
  
  fclose(dfile);
}

void Afrolikelihood::finitedifftest()
{
  // Perform finite difference tests on various derivatives calculated:
  double likelihood2;
  double difftol(1.0e-7);
  printf("\nATOMIC PARAMETERS\n\n");    

  unsigned id1            = mdl.atom.size()-1;

  double value            = grad();

  for (unsigned i         = 0; i < mdl.atom.size(); i++)
    if (mdl.atom[id1].isotropic)
      printf(" Grad[%u][B]   = %f\n",i, dLdbiso[i]);
    else
    {
      printf(" Grad[%u][U1]   = %f\n",i, dLduaniso[i][0]);
      printf(" Grad[%u][U2]   = %f\n",i, dLduaniso[i][1]);
      printf(" Grad[%u][U3]   = %f\n",i, dLduaniso[i][2]);
      printf(" Grad[%u][U4]   = %f\n",i, dLduaniso[i][3]);
      printf(" Grad[%u][U5]   = %f\n",i, dLduaniso[i][4]);
      printf(" Grad[%u][U6]   = %f\n",i, dLduaniso[i][5]);
    }
      /*
      printf(" Grad[" << i << "][f']  = " << dLdfp[i]   << "n";
      printf(" Grad[" << i << "][f''] = " << dLdfpp[i]  << "n";
      */
  
  if (mdl.atom[id1].isotropic)
    mdl.atom[id1].Setbiso(mdl.atom[id1].biso + difftol);
  else
    mdl.atom[id1].Setuaniso(mdl.atom[id1].uaniso[0] + difftol, 0);

  calcsigmah();
  
  likelihood2             = function();

  printf("Likelihood from function = %f\n", likelihood2);
  printf("Likelihood from gradient = %f\n", value);
  printf("Finite diff              = %f\n", (likelihood2 - value)/difftol);
  
  value                   = grad();
    
  for (unsigned i         = 0; i < mdl.atom.size(); i++)
    printf(" Grad[%u][N] = %f\n",i,dLdnumb[i]);
  
  mdl.atom[id1].Setnumb(mdl.atom[id1].number+difftol);
  
  calcsigmah();
  
  likelihood2             = function();

  printf("Likelihood from function = %f\n", likelihood2);
  printf("Likelihood from gradient = %f\n", value);
  printf("Finite diff              = %f\n", (likelihood2 - value)/difftol);
}

void Afrolikelihood::outputevalues()
{
  CMtz::MTZ *MTZOUT        = CMtz::MtzMalloc(0,0);
  CMtz::MTZ *MTZIN         = CMtz::MtzGet("HKLIN",0);
  setupmtz(MTZOUT,MTZIN);
  FILE *dfile(NULL);

  unsigned inc             = (allin) ? colin.size() : 3;
  
  unsigned columns(0);
  for (unsigned d          = 0; d < xtal.sf.size(); d++)
    if (xtal.sf[d].anomalous)
      columns             += 8;
    else
      columns             += 4;

  if (ascii)
  {
    string filename        = output + ".txt";
    dfile                  = fopen(filename.c_str(),"w");
  }  

  for (unsigned r          = 0; r < xtal.maxref; r++)
  {    
    // Default value for columns to be written out is MNF
    vector<float> fdata(columns+inc, CCP4::ccp4_nan().f);
 
    storeinitialcolumns(MTZIN, fdata, inc, r);

    double eps((double) (xtal.epsilon[r]));
    double sigman(ZERO);
    int c(-1);
    for (unsigned d        = 0; d < xtal.sf.size(); d++)
    {
      unsigned sa(xtal.bin(d,r));
      unsigned sb(sa);
      double wa(ONE), wb(ZERO);
      
      if (interpolate)
      {
	xtal.binweights(d, r, sa, sb, wa, wb);
        sigman             = wa*xtal.sf[d].sigman[sa]   + wb*xtal.sf[d].sigman[sb];
      }
      else
	sigman             = xtal.sf[d].sigman[sa];
      
      if (xtal.sf[d].anomalous)
      {
	if (xtal.sf[d].anouse(r) && !xtal.centric[r])
	{
	  fdata[inc + ++c] = xtal.sf[d].datap[r];
	  fdata[inc + ++c] = xtal.sf[d].devp[r];
	  fdata[inc + ++c] = xtal.sf[d].datam[r];
	  fdata[inc + ++c] = xtal.sf[d].devm[r];
	  fdata[inc + ++c] = xtal.sf[d].datap[r]/sqrt(eps*sigman);
	  fdata[inc + ++c] = xtal.sf[d].devp[r]/sqrt(eps*sigman);
	  fdata[inc + ++c] = xtal.sf[d].datam[r]/sqrt(eps*sigman);
	  fdata[inc + ++c] = xtal.sf[d].devm[r]/sqrt(eps*sigman);
	}
	else
	  c               += 8;
      }
      else
	if (xtal.sf[d].use(r))
	{
	  fdata[inc + ++c] = xtal.sf[d].datamean(r);
	  fdata[inc + ++c] = xtal.sf[d].devmean(r);
	  fdata[inc + ++c] = xtal.sf[d].datamean(r)/sqrt(eps*sigman);
	  fdata[inc + ++c] = xtal.sf[d].devmean(r)/sqrt(eps*sigman);
	}
	else
	  c               += 4;
    }
    
    CMtz::ccp4_lwrefl(MTZOUT, &fdata[0], &colout[0], fdata.size(), r+1);
    if (ascii             && xtal.sf[0].anouse(r) && !xtal.centric[r])
      fprintf(dfile, "%5d%5d%5d%9.3f%9.3f%9.3f%9.3f%9.3f%9.3f%9.3f%9.3f\n",
	      xtal.miller[r][0], xtal.miller[r][1], xtal.miller[r][2],
	      xtal.sf[0].datap[r], xtal.sf[0].devp[r],
	      xtal.sf[0].datam[r], xtal.sf[0].devm[r],
	      xtal.sf[0].datap[r]/sqrt(eps*sigman),
	      xtal.sf[0].devp[r]/sqrt(eps*sigman),
	      xtal.sf[0].datam[r]/sqrt(eps*sigman),
	      xtal.sf[0].devm[r]/sqrt(eps*sigman));
  }

  if (allin)
    CMtz::MtzFree(MTZIN);
  if (ascii)
    fclose(dfile);
  CMtz::MtzPut(MTZOUT, " ");
  CMtz::MtzFree(MTZOUT);
}
