/*
  Copyright (c) Navraj Pannu       1998 - 2005
  Copyright (c) Leiden University  2002 - 2005

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

#include <ctype.h>
#include "bp3input.h"
#include "ccp4_parser.h"                  // Peter Briggs' parser in c libraries

Bp3input::Bp3input(Model &model, Crystal &xtl, Bp3likelihood &likelihood,
		   Minimizer &minimizer)
  : Input(model, xtl, likelihood, minimizer), like(likelihood)
{
  norefocc             = norefxyz = norefbfac = false;
  
  // parse
  CCP4parse();

  CCP4::ccp4_parse_end(parser); // clean up parser array
  
  check();

  if (getenv("HKLIN") != NULL)
    xtal.setwithmtz();
 else
   xtal.setwithmanymtz();

  // some checking of input ie. should coordinates not be refined due
  // to special positions or polar spacegroups.

  /*
  if ( (like.Getmode()   == "PHASE") || (like.Getmode()   == "CHECK") )
    for (unsigned d       = 0; d < xtal.sf.size(); d++)
      xtal.sf[d].Setdefaultsdluz(0.65);
   */

  like.checkocc();  

  printf("Making sure atoms have fractional coordinates less than ONE\n\n");
  like.checkfrac(frac);
  
  // fix x, y, and/or z for one atom if spacegroup is polar
  for (unsigned i         = 0; i < 3; i++)
    if (xtal.sg.Getpolar(i))
      if (mdl.site.size() > 0)
	mdl.site[0].Setrefinex(i,false);

  for (unsigned i         = 0; i < mdl.atom.size(); i++)
  {
    unsigned s(mdl.atom[i].Getnsite());
    
    double x[3]           = {mdl.site[s].Getx(0),
			     mdl.site[s].Getx(1),
			     mdl.site[s].Getx(2)};

    if (norefxyz)
      mdl.site[s].Setrefinex(0,false).Setrefinex(1,false).Setrefinex(2,false);

    if (norefocc)
      mdl.atom[i].Setrefineocc(false);

    if (norefbfac)
      mdl.atom[i].Setrefinebfac(false);

    double occ            = mdl.atom[i].Getocc();
    
    if (xtal.sg.specialpos(x[0], x[1], x[2], occ))
    {
      for (unsigned c     = 0; c < 3; c++)
	mdl.site[s].Setx(c, x[c]);
      mdl.atom[i].Setocc(occ);
      printf("Warning: Site %d is on a special position\n", i+1);
      printf("Its occupancy is being divided by the multiplicity.\n\n");
      
    }
  }

  print();
}

void Bp3input::check() 
{
  // checks if input is okay

  Input::check();

  for (unsigned d = 0; d < xtal.sf.size(); d++)
  {
    if ( (xtal.sf[d].Getsdatap() == "") || (xtal.sf[d].Getsdevp() == "") )
      Bp3Error("Bp3Input::check", "Dataset declared without F and/or SIGF columns");
    if ( (xtal.sf[d].Gettype()  != "AMPLITUDE") && (like.Getmode() != "OUTPUTHKL") &&
	 (like.Getmode()        != "OUTPUTSCALEPACK") )
      Bp3Error("Bp3likelihood::setup","input amplitudes only");    
  }
 
  if ( !((like.Getmode() == "INTENSITY") || (like.Getmode() == "OUTPUTHKL") || 
	(like.Getmode()  == "OUTPUTSCALEPACK") ) )
  {  
    if (mdl.site.size() <= 0)
    {
      Bp3Warning("Bp3Input::check", "no atoms have been defined");
      Bp3Warning("Bp3Input::check", "BP3 will output intensities");
      if (like.Getmode() != "OUTPUTSCALEPACK") 
        like.Setmode("INTENSITY");
    }

    if ( (xtal.sf.size()    == 1) && (xtal.sf[0].Getsdatam() == "") )
    {
      Bp3Warning("Bp3Input::check", "only one nonanomalous data set is given");
      Bp3Warning("Bp3Input::check", "BP3 will output intensities"); 
      if (like.Getmode() != "OUTPUTSCALEPACK") 
        like.Setmode("INTENSITY");
    }
  }
}

void Bp3input::CCP4parse()
{
  // The setup of reading keyword input and it follows pretty much
  // exactly Peter Briggs' documentation of his parser.

  // read lines from stdin until END or EOF is reached
  bool cont       = true;
  char line[201];

  while (cont)
  {
    // Blank line before calling ccp4_parser implies reading from stdin
    line[0]       = '\0';
    unsigned ntok = CCP4::ccp4_parser(line, 200, parser, 1);
    
    if (ntok      < 1)                                            // reached EOF
      break;
    else   // perform keyword analysis and interpretation
    {
      if (Input::xtalkeywords())
	continue;
      else if (Input::datakeywords())
	continue;
      else if (Input::modelkeywords())
	continue;
      else if (Input::minimizerkeywords())
	continue;
      else if (Input::likelihoodkeywords())
	continue;
      else if (bp3keywords())
	continue;
      else if (CCP4::ccp4_keymatch("END", parser->keyword))
	break;
      else
	Bp3Error("CCP4parse","Keyword not recognised");
    }
  }  
}

bool Bp3input::bp3keywords()
{
  if (CCP4::ccp4_keymatch("REFI", parser->keyword))       // mode options
    refikeyword();
  else if (CCP4::ccp4_keymatch("PHAS", parser->keyword))            
    phaskeyword();
  else if (CCP4::ccp4_keymatch("CHEC", parser->keyword))            
    checkeyword();
  else if (CCP4::ccp4_keymatch("HAND", parser->keyword))            
    handkeyword();
  else if (CCP4::ccp4_keymatch("NOHA", parser->keyword))            
    nohakeyword();
  else if (CCP4::ccp4_keymatch("DIFF", parser->keyword))            
    diffkeyword();
  else if (CCP4::ccp4_keymatch("NORE", parser->keyword))            
    norekeyword();
  else if (CCP4::ccp4_keymatch("REFA", parser->keyword))  // parameters to refine
    refakeyword();
  else if (CCP4::ccp4_keymatch("SHEL", parser->keyword))            
    shelkeyword();
  else if (CCP4::ccp4_keymatch("MOCC", parser->keyword))  // maximum parameter shifts  
    mocckeyword();                                           
  else if (CCP4::ccp4_keymatch("MXYZ", parser->keyword))            
    mxyzkeyword();
  else if (CCP4::ccp4_keymatch("MBFA", parser->keyword))            
    mbfakeyword();
  else if (CCP4::ccp4_keymatch("MLUZ", parser->keyword))            
    mluzkeyword();
  else if (CCP4::ccp4_keymatch("FHOU", parser->keyword))            
    fhoukeyword();
  else if (CCP4::ccp4_keymatch("THRE", parser->keyword))            
    threkeyword();  
  else
    return false;
  
  return true;
}

void Bp3input::refikeyword() const
{
  // check to see if we have no more than one token
  if (parser->ntokens > 1)
    Bp3Warning("Bp3input::refikeyword", "Extra text in REFIne keyword will be ignored");
  
  like.Setmode("REFINE");
}

void Bp3input::phaskeyword() const
{
  // check to see if we have no more than one token
  if (parser->ntokens > 2)
    Bp3Warning("Bp3input::phaskeyword", "Extra text in PHASe keyword will be ignored");
  
  like.Setmode("PHASE");
  like.Setmaxocc(HALF).Setmaxluzz(ONE);

  if (parser->ntokens == 2)
    if (parser->token[1].isnumber)
      mini.Setiterations((unsigned)parser->token[1].value);  
    else
      Bp3Error("Bp3input::phaskeyword", "A numerical value is expected after PHAS");
}

void Bp3input::checkeyword() const
{
  // check to see if we have no more than one token
  if (parser->ntokens > 2)
    Bp3Warning("Bp3input::checkeyword", "Extra text in CHECk keyword will be ignored");
  
  like.Setmode("CHECK");
  like.Setmaxocc(HALF).Setmaxluzz(ONE);

  if (parser->ntokens == 2)
    if (parser->token[1].isnumber)
      mini.Setiterations((unsigned)parser->token[1].value);  
    else
      Bp3Error("Bp3input::checkeyword", "A numerical value is expected after CHECk");
}

void Bp3input::handkeyword() const
{
  Bp3Warning("Bp3input::handkeyword", "This keyword is no longer needed - it is a default!");

  // check to see if we have no more than one token
  if (parser->ntokens > 1)
    Bp3Warning("Bp3input::handkeyword", "Extra text in HAND keyword will be ignored");
  
  like.Sethand(1);
}

void Bp3input::nohakeyword() const
{
  // check to see if we have no more than one token
  if (parser->ntokens > 1)
    Bp3Warning("Bp3input::nohakeyword", "Extra text in NOHAnd keyword will be ignored");
  
  like.Sethand(0);
} 

void Bp3input::diffkeyword() const
{
  // check to see if we have no more than one token
  if (parser->ntokens  > 2)
    Bp3Warning("Bp3input::diffkeyword", "Extra text in DIFF keyword will be ignored");

  like.Setmode("DIFF");
  
  if (parser->ntokens == 2)
    if (parser->token[1].isnumber)
      like.Setdifftol((double)parser->token[1].value);
    else
      Bp3Error("Bp3input::diffkeyword", "A numerical value is expected after DIFF");
}

void Bp3input::norekeyword()
{
  if (parser->ntokens > 1)
  {
    for (int i        = 1; i < parser->ntokens; i++)
      if (CCP4::ccp4_keymatch("XYZ",parser->token[i].word))
	norefxyz      = true;
      else if (CCP4::ccp4_keymatch("OCCU",parser->token[i].word))
 	norefocc      = true;
      else if (CCP4::ccp4_keymatch("BFAC",parser->token[i].word))
 	norefbfac     = true;
      else
	Bp3Error("Bp3input::norekeyword", "Subkeyword of NOREf not recognized");
  }
  else
    Bp3Error("Bp3input::norekeyword", "XYZ, OCCU, and/or BFAC is expected after NOREf");

} 

void Bp3input::refakeyword() const
{
  // check to see if we have no more than one token
  if (parser->ntokens > 1)
    Bp3Warning("Bp3input::refakeyword", "Extra text in REFAll keyword will be ignored");

  like.Setrefineall(true);
}

void Bp3input::shelkeyword() const
{
  // check to see if we have no more than one token
  if (parser->ntokens > 2)
    Bp3Warning("Bp3input::shelkeyword", "Extra text in SHELdrick keyword will be ignored");
  if (parser->ntokens == 2)
    if (parser->token[1].isnumber)
      like.Setshelthres((unsigned)parser->token[1].value);
    else
      Bp3Error("Bp3input::shelkeyword", "A numerical value is expected after SHEL");
  else
    like.Setshelthres(0);
}

void Bp3input::mocckeyword() const
{
  // check to see if we have no more than one token
  if (parser->ntokens > 2)
    Bp3Warning("Bp3input::mocckeyword", "Extra text in MOCC keyword will be ignored");

  if (!parser->token[1].isnumber && (parser->token[1].value <= ZERO) )
    Bp3Error("Bp3input::mocckeyword", "A positive numerical value is expected for MOCC");

  like.Setmaxocc((double)(parser->token[1].value));
}

void Bp3input::mluzkeyword() const
{
  // check to see if we have no more than one token
  if (parser->ntokens > 2)
    Bp3Warning("Bp3input::mluzkeyword", "Extra text in MLUZ keyword will be ignored");

  if (!parser->token[1].isnumber && (parser->token[1].value <= ZERO) )
    Bp3Error("Bp3input::mluzkeyword", "A positive numerical value is expected for MLUZ");

  like.Setmaxluzz((double)(parser->token[1].value));
}

void Bp3input::mbfakeyword() const
{
  // check to see if we have no more than one token
  if (parser->ntokens > 2)
    Bp3Warning("Bp3input::mbfakeyword", "Extra text in MBFA keyword will be ignored");

  if (!parser->token[1].isnumber && (parser->token[1].value <= ZERO) )
    Bp3Error("Bp3input::mbfakeyword", "A positive numerical value is expected for MBFA");

  like.Setmaxbfac((double)(parser->token[1].value));
}

 void Bp3input::mxyzkeyword() const
{
  // check to see if we have no more than one token
  if (parser->ntokens > 2)
    Bp3Warning("Bp3input::mxyzkeyword", "Extra text in MXYZ keyword will be ignored");

  if (!parser->token[1].isnumber && (parser->token[1].value <= ZERO) )
    Bp3Error("Bp3input::mxyzkeyword", "A positive numerical value is expected for MXYZ");

  like.Setmaxxyz((double)(parser->token[1].value));
}

void Bp3input::fhoukeyword() const
{
  // check to see if we have no more than one token
  if (parser->ntokens  > 1)
    Bp3Warning("Bp3input::fhoukeyword", "Extra text in FHOUt keyword will be ignored");

  like.Setoutputhcalc(true);
}
 
void Bp3input::threkeyword() const
{
  // check to see if we have no more than one token
  if (parser->ntokens  > 2)
    Bp3Warning("Bp3input::threkeyword", "Extra text in THRE keyword will be ignored");

  if (parser->token[1].isnumber)
    like.Setthreshold((double)parser->token[1].value);
  else
    Bp3Error("Bp3input::threkeyword", "A numerical value is expected for the THREshold");
}
