/*
  Copyright (c) Navraj Pannu       1998 - 2005
  Copyright (c) Leiden University  2002 - 2005

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

#include <math.h>
#include "gauss.h"
#include "bp3likelihood.h"
#include "matrix.h"
#include "csymlib.h"


Bp3likelihood::Bp3likelihood(Model &model, Crystal &crystal)
  : Likelihood(model, crystal)
{
  // defaults
  likelihood      = ZERO;
  mode            = "REFINE";
  target          = "UNCO";
  title           = "Phasing from BP3";
  threshold       = FOUR;
  difftol         = 1.0e-7;
  verbose         = cd             = 0;
  shelthres       = 50;
  refineluzzati   = refineocc      = true;       
  refinescale     = refinexyz      = refinebfac  = false;
  refinefp        = refinefpp      = refinenumb  = false;
  minvar          = EPSILON;
  allin           = interpolate    = stats = sheldrick = refineall = false;
  output          = "heavy";
  outputhcalc     = false;
  // default size for Gaussian integration parameters
  cweight.resize(3);
  pweight.resize(30);
  afweight.resize(3);
  sadweight.resize(30);
}

double Bp3likelihood::grad(const bool check, const bool outputmtz)
{
  // Evaluates the function and the gradient

  // Calculation derivatives of requested function
  double val(ZERO);

  if (target      == "UNCO")
    val            = uncorrelatedgradient(check,outputmtz);
  else if (target == "SAD")
    val            = sadgradient(check,outputmtz);
  else if (target == "MAD")
    val            = uncorrelatedgradient(check,outputmtz);
  
  // The above routine only calculate d{Target}/d{Structure Factor}.
  // The routine below calculates d{Target}/d{atom parameter}, if we
  // are refining atomic parameters.
  directsfcalc(true,updatesigmah);
  
  return val;
}

double Bp3likelihood::function(const bool check)
{
  double val(ZERO);

  if (target      == "UNCO")
    val            = uncorrelatedfunction(check);
  else if (target == "SAD")
    val            = sadfunction(check);
  else if (target == "MAD")
    val            = uncorrelatedfunction(check);

  if (check)
    directsfcalc(false,updatesigmah);
  
  return val;
}

void Bp3likelihood::setup(const bool first, const bool refine, const bool fast)
{
  // Determine how many isomorphism and atomic parameters there are  
  // to refine and allocate memory for variables storing gradients.

  if (mode                     == "CHECK")
  {
    if (xtal.sf[0].anomalous)
    {
      target                    = "SAD";
      cd                        = 0;

      if (xtal.sf.size()        > 1)
      {
	vector<double> totaldano(xtal.sf.size(), ZERO);
	vector<double> total(xtal.sf.size(), ZERO);	
	for (unsigned d         = 0; d < xtal.sf.size(); d++)
	  for (unsigned s       = 0; s < xtal.sf[d].nbins; s++)
	  {
	     totaldano[d]      += xtal.sf[d].danoovers[s]*xtal.sf[d].anonshl[s];
	     total[d]          += (double)xtal.sf[d].anonshl[s];
	  }

	for (unsigned d         = 0; d < xtal.sf.size(); d++)
	  if (totaldano[d]/total[d] > totaldano[cd]/total[cd])
	    cd                  = d;
      }
      
      for (unsigned d           = 0; d < xtal.sf.size(); d++)
	if (d                  != cd)
	  xtal.sf[d].refinesd   = false;
    }    
  }
  else if (xtal.mad)  // ***NSP
  {
    target                      = "UNCO";
    if (!refine)
      madphasingdefaults();
  }
  
  if (xtal.sf.size()           == 1)
    if (xtal.sf[0].anomalous)
      target                    = "SAD";

  refinescale                   = false; 
  refinexyz                     = refineluzzati = refinebfac = !first;

  if (sheldrick)
  {
    refinexyz                   = refinebfac    = false;
    refineluzzati               = true;
  }

  if (fast)
    refinebfac                  = false;
  
  resize();
  
  // Setup Gaussian variables
  Setgauss();

  if (!refine)
    xtal.Setselected("PHASE");
    
  directsfcalc(false, true);
  
  xtal.Setnormalization().SetLuzzati();

  if (!xtal.mad && (target     == "UNCO") )
    for (unsigned d             = 1; d < xtal.sf.size(); d++)
      for (unsigned s           = 0; s < xtal.sf[d].nbins; s++)
	if (xtal.sf[d].var(ONE, s) <= ONE && (xtal.sf[d].sigman[s] > xtal.sf[d].sigmah[s]))
	  xtal.sf[d].dluz[s]    = 0.99*sqrt((xtal.sf[d].sigman[s] -
					     xtal.sf[d].sigmah[s])/
					    xtal.sf[d].sigmanref[s]);
 
    
  if ((xtal.sf[0].sigmah[0]    <= ZERO) && xtal.mad)
      Bp3Error("Bp3likelihood::setup", "define the native data set last in mad");

  if (refine)
  {
    unsigned nscalepars(0);
    unsigned nluzzatipars(0);
    unsigned natmpars(0);

    for (unsigned d             = 0; d < xtal.sf.size(); d++)
    {
      if (target               == "UNCO")
      {
	if (xtal.sf[d].refinek && refinescale)
	  nscalepars++;
	if (xtal.sf[d].refineb && refinescale)
	  nscalepars++;

	if (xtal.sf[d].refined && refineluzzati)
	  nluzzatipars         += xtal.sf[d].nbins;
	if (xtal.sf[d].refinee && refineluzzati)
	  nluzzatipars         += xtal.sf[d].nbins;             
	if (xtal.sf[d].refinead && refineluzzati)
	  nluzzatipars         += xtal.sf[d].nbins;
      }
      else if ( (target        == "SAD") || (target == "MAD") )
	if (xtal.sf[d].refinesd && refineluzzati)
	  nluzzatipars         += xtal.sf[d].nbins;      
    }

    // Atom parameters
    if (refinexyz)
      for (unsigned s           = 0; s < mdl.site.size(); s++)
	for (unsigned k         = 0; k < 3              ; k++)
	  if (mdl.site[s].Getrefinex(k))
	    natmpars++;
    
    for (unsigned a             = 0; a < mdl.atom.size()  ; a++)
    {
      if (mdl.atom[a].Getrefinebfac() && refinebfac)
	if (mdl.atom[a].Getisotropic() )
	  natmpars++;
	else  
	  natmpars             += 6;
      if (mdl.atom[a].Getrefineocc() && refineocc)
	natmpars++;      
    }

    for (unsigned f             = 0; f < mdl.form.size()  ; f++)
      for (unsigned w           = 0; w < mdl.form[f].Getnwave(); w++)
      {
	if (mdl.form[f].Getrefinefp(w) && refinefp)
	  natmpars++;
	if (mdl.form[f].Getrefinefpp(w) && refinefpp)
	  natmpars++;
      }

    npars                          = natmpars + nluzzatipars + nscalepars;
    printf("Number of parameters being refined: %u\n", npars);
    printf("Consisting of...\n");
    if (nscalepars)
      printf("Scale parameters: %u\n", nscalepars);
    if (nluzzatipars)
      printf("Luzzati parameters: %u\n", nluzzatipars);
    if (natmpars)
      printf("Atomic parameters: %u\n\n", natmpars);

    Settype();

    // check for reflections with low-likelihood
    grad(true);
  }
  xtal.print();
}

void Bp3likelihood::checkocc()
{
  if ( ( (mode                   != "PHASE") || (target == "UNCO") )  )
  {
    bool setocc(false);
    // checks it occupancies are too high at user input
    for (unsigned i               = 0; i < mdl.atom.size(); i++)
      if ( (mdl.atom[i].Getocc()  > 0.5) && (xtal.sf.size() != 1) )
        setocc                    = true;

    if (verbose                  && setocc)
      Bp3Warning("Bp3likelihood::checkocc","Occupancies are high at user input, lowering them");
  
    if (setocc)
      for (unsigned i             = 0; i < mdl.atom.size(); i++)
        if (mdl.atom[i].Getocc()  > 0.5) 
          mdl.atom[i].Setocc(0.4);    
  }
}

void Bp3likelihood::checkfrac(const vector<bool> &frac)
{
  // transform to fractional coordinates if orthogonal were given
  for (unsigned c                     = 0; c < xtal.cell.size(); c++)
    if (!frac[c])
      for (unsigned s                 = 0; s < mdl.site.size(); s++)
	// get crystal for the particular site
	for (unsigned a               = 0; a < mdl.atom.size(); a++)
	  if ( (mdl.atom[a].nsite    == s) && (mdl.atom[a].crystal  == c) )
	  {
	    vector<double> f(3, ZERO);
	    for (unsigned i           = 0; i < f.size(); i++)
	    {
	      for (unsigned j         = 0; j < f.size(); j++)
		f[i]                 += xtal.cell[c].or2frac[i][j]*mdl.site[s].x[j];
 	      mdl.site[s].Setx(i,f[i]);
	    }
	  }
    
  // make sure atoms are -1 < x < 1

  for (unsigned i                     = 0; i < mdl.site.size(); i++)
    for (unsigned x                   = 0; x < 3; x++)
      do
      {
	if (mdl.site[i].Getx(x)      >= 0.99)
	  mdl.site[i].Setx(x, mdl.site[i].Getx(x) - ONE);
	else if (mdl.site[i].Getx(x) <= -ONE)
	  mdl.site[i].Setx(x, mdl.site[i].Getx(x) + ONE);
      } while ( (mdl.site[i].Getx(x) >= 0.99) || (mdl.site[i].Getx(x) <= - ONE) );
  
}

unsigned Bp3likelihood::Setiterations() const
{

  unsigned iterations(3);
  
  if (mode               != "CHECK" )
  { 
    unsigned wavelengths(0);

    for (unsigned d     = 0; d < xtal.sf.size(); d++)
      if (xtal.sf[d].nwave)
        wavelengths       = std::max(xtal.sf[d].nwave+1, wavelengths);
  
    if (wavelengths      == 2)
      iterations          = 8;
    else if (wavelengths == 3)
      iterations          = 10;
    else if (wavelengths == 4)
      iterations          = 12;

  }

  printf("Setting the number of iterations to %u\n\n", iterations);
  
  return iterations;
}

void Bp3likelihood::checkatoms()
{

  // if occupancies are low, do not refine
  for (unsigned i               = 0; i < mdl.atom.size(); i++)
    if (mdl.atom[i].Getocc()    < 0.05)
    {
      mdl.site[mdl.atom[i].nsite].Setrefinex(0,false).Setrefinex(1,false).Setrefinex(2,false);
      mdl.atom[i].Setrefineocc(false).Setrefinebfac(false);
      printf("Not refining parameters for Atom %u\n",i+1);
    }
  printf("\n");
}

void Bp3likelihood::madphasingdefaults()
{
  if (xtal.Getmad())
    xtal.Setonlyacentrics(false).Setonlycentrics(false);
}

void Bp3likelihood::resize()
{
  // Allocate memory for variables storing gradients.

  if ( (target          == "UNCO") && (dLddluz.size() != xtal.sf.size()) )
  {
    // Non-Isomorphism parameter(s) 
    dLddluz.resize(xtal.sf.size());

    // Error in heavy atom model
    dLdeluz.resize(xtal.sf.size());

    // Error in anomalous difference
    dLdadluz.resize(xtal.sf.size());

    for (unsigned d      = 0; d < xtal.sf.size(); d++)
    {
      dLddluz[d].resize(xtal.sf[d].nbins, ZERO);
      dLdeluz[d].resize(xtal.sf[d].nbins, ZERO);
      dLdadluz[d].resize(xtal.sf[d].nbins, ZERO);
    }
  }

  if ( ( (target        == "MAD") || (target == "SAD") ) &&
       (dLdsdluz.size() != xtal.sf.size()) )
  {
    // Luzzati Errors 
    dLddluz.resize(xtal.sf.size());
    dLdsdluz.resize(xtal.sf.size());

    // Matrices
    covmodel.resize(xtal.sf[0].nbins);
    covinvmodel.resize(xtal.sf[0].nbins);

    unsigned matrixsize(0);

    if (target          == "SAD")
      matrixsize         = 2;
    else
      for (unsigned d    = 0; d < xtal.sf.size(); d++)
	if (xtal.sf[d].anomalous)
	  matrixsize    += 2;
	else
	  matrixsize++;

    sumfpp.resize(xtal.sf.size());
    difffpp.resize(xtal.sf.size()), difffp.resize(xtal.sf.size());
    
    for (unsigned d      = 0; d < xtal.sf.size(); d++)
    {
      sumfpp[d].resize(xtal.sf.size());
      difffpp[d].resize(xtal.sf.size());
      difffp[d].resize(xtal.sf.size());

      dLddluz[d].resize(xtal.sf[d].nbins, ZERO);
      dLdsdluz[d].resize(xtal.sf[d].nbins, ZERO);
      for (unsigned s    = 0; s < xtal.sf[d].nbins; s++)
      {
	covmodel[s].resize(matrixsize);
	covinvmodel[s].resize(matrixsize);
	for (unsigned d1 = 0; d1 < xtal.sf.size(); d1++)
	{
	  sumfpp[d][d1].resize(xtal.sf[d].nbins, ZERO);
	  difffpp[d][d1].resize(xtal.sf[d].nbins, ZERO);
	  difffp[d][d1].resize(xtal.sf[d].nbins, ZERO);
	}
      }
    }
    unsigned row(2*matrixsize);
    recov.resize(row), recovinv.resize(row);
    redAdsd.resize(row), redAdsigh.resize(row);
    redAdsfpp.resize(row);

    // inverse/eigenvalue filtering
    evalues.resize(row);

    lwork                = 1 + 6*row + 2*row*row;
    liwork               = 3 + 5*row;
    lawork.resize(lwork);
    iwork.resize(liwork);    
  }

   // scale parameters
  dLdkscale.resize(xtal.sf.size(), ZERO);
  dLdbscale.resize(xtal.sf.size(), ZERO);

  // Real and imaginary F+ heavy atom structure factor (or just F if F+=F=) 

  if (dLdAp.size()      != xtal.sf.size() )
  {
    dLdAp.resize(xtal.sf.size());
    dLdBp.resize(xtal.sf.size());

    for (unsigned d      = 0; d < xtal.sf.size(); d++)
    {
      dLdAp[d].resize(xtal.maxref, ZERO);
      dLdBp[d].resize(xtal.maxref, ZERO);
    }  

    // Real and imaginary F- heavy atom structure factor.
    // Memory is only allocated if anomalous scattering is present

    dLdAm.resize(xtal.sf.size());
    dLdBm.resize(xtal.sf.size());
    
    for (unsigned d        = 0; d < xtal.sf.size(); d++)
      if (xtal.sf[d].anomalous)
      {	
	 dLdAm[d].resize(xtal.maxref, ZERO);
	 dLdBm[d].resize(xtal.maxref, ZERO);	
      }

    // Wilson's sigmaH parameter
    dLdsigmah.resize(xtal.sf.size());  
    dLdsumfpp.resize(xtal.sf.size());

    for (unsigned d        = 0; d < xtal.sf.size(); d++)
    {
      dLdsigmah[d].resize(xtal.sf[d].nbins, ZERO);
      dLdsumfpp[d].resize(xtal.sf[d].nbins, ZERO);
    }
  
    // Atomic parameters
    dLdx.resize(mdl.site.size());

    for (unsigned s        = 0; s < mdl.site.size(); s++)
      dLdx[s].resize(3, ZERO);  

    dLdocc.resize(mdl.atom.size(), ZERO);
  
    dLdbiso.resize(mdl.atom.size(), ZERO);
    dLduaniso.resize(mdl.atom.size());
    
    for (unsigned a        = 0; a < mdl.atom.size(); a++)
       dLduaniso[a].resize(6, ZERO);

    dLdfp.resize(mdl.form.size());
    dLdfpp.resize(mdl.form.size());

    for (unsigned f        = 0; f < mdl.form.size(); f++)
    {
      dLdfp[f].resize(mdl.form[f].nwave, ZERO);
      dLdfpp[f].resize(mdl.form[f].nwave, ZERO);
    }  

    // fom
    afom.resize(xtal.sf[0].nbins);
    cfom.resize(xtal.sf[0].nbins);
  
    // variables needed in structure factor calculation
    hrot.resize(xtal.sg.NSYMP);
    htran.resize(xtal.sg.NSYMP);
    for (unsigned s        = 0; s < hrot.size(); s++)
      hrot[s].resize(3);
  }
}

Bp3likelihood &Bp3likelihood::Setgauss()
{
  // Setup Gaussian variables

  if (target            == "UNCO")
  {
    Gauss gausscentrichermite("Hermite", cweight.size());
    cnode                = gausscentrichermite.Getnode();
    cweight              = gausscentrichermite.Getweight();

    for (unsigned i      = 0; i < cweight.size(); i++)
      cweight[i]        *= exp(cnode[i]*cnode[i]);
    
    if (verbose          > 1)
      gausscentrichermite.print();

    Gauss gausslegendre("Legendre", pweight.size());
    vector<double> pnode = gausslegendre.Getnode();
    pweight              = gausslegendre.Getweight();

    if (verbose          > 1)
      gausslegendre.print();

    gsin.resize(pweight.size());
    gcos.resize(pweight.size());  
    // Precompute sin, cos                                             
    for (unsigned i      = 0; i < pweight.size(); i++)
    {
      gcos[i]            = tab.tcos(PI*pnode[i]);
      gsin[i]            = tab.tsin2(PI*pnode[i]);
    }

    Gauss gausshermite("Hermite", afweight.size());
    afnode               = gausshermite.Getnode();
    afweight             = gausshermite.Getweight();

    for (unsigned i      = 0; i < afweight.size(); i++)
      afweight[i]       *= exp(afnode[i]*afnode[i]);

    if (verbose          > 1)
      gausshermite.print();
    
    sadweight.resize(0);
  }
  else if ( (target     == "SAD") || (target == "MAD") )
  {
    // Sad function
    Gauss gausslegendre("Legendre", sadweight.size());
    vector<double> snode = gausslegendre.Getnode();
    sadweight            = gausslegendre.Getweight();

    sadsin.resize(sadweight.size());
    sadcos.resize(sadweight.size());  
    // Precompute sin, cos                                             
    for (unsigned i      = 0; i < sadweight.size(); i++)
    {
      sadcos[i]          = tab.tcos(PI*snode[i]);
      sadsin[i]          = tab.tsin2(PI*snode[i]);
    }
    cweight.resize(0);
    pweight.resize(0);
    afweight.resize(0);

    if (verbose          > 1)
      gausslegendre.print();
  }
  return *this;
}

void Bp3likelihood::otherhand()
{
  unsigned enantinumber(xtal.sg.enantiomorph());
  output   += "-oh";
  hand      = 2;
  if (enantinumber)
    xtal.sg = Spacegroup(enantinumber);
  mdl.inverthand(enantinumber);
}

void Bp3likelihood::writescript() const
{
  string scriptname         = output + ".sh";

  FILE *sfile               = fopen(scriptname.c_str(),"w");

  fprintf(sfile,"#!/bin/sh\n\n");

  fprintf(sfile, "bp3 HKLIN %s HKLOUT nextrun.mtz << eof-bp3 > nextrun.log\n\n",getenv("HKLIN"));

  for (unsigned s           = 0; s < mdl.site.size() ; s++) 	
  {
    fprintf(sfile,"SITE %3d %5.3f %5.3f %5.3f  ", s + 1,
 	    mdl.site[s].x[0], mdl.site[s].x[1], mdl.site[s].x[2]);

    if (!mdl.site[s].Getrefinex(0) || !mdl.site[s].Getrefinex(1) ||
	!mdl.site[s].Getrefinex(2))
    {
      fprintf(sfile, " NOREf ");
      if (!mdl.site[s].Getrefinex(0))
	fprintf(sfile, " X ");
      if (!mdl.site[s].Getrefinex(1))
	fprintf(sfile, " Y ");
      if (!mdl.site[s].Getrefinex(2))
	fprintf(sfile, " Z ");
    }
    fprintf(sfile, "\n");
  }

  fprintf(sfile, "\n\n");
  
  for (unsigned c           = 0; c < xtal.cell.size(); c++)
  {
    fprintf(sfile, "XTAL %s\n", xtal.cell[c].name.c_str());
    int w                   = -1;
    fprintf(sfile, "  CELL %8.3f %8.3f %8.3f %7.3f %7.3f %7.3f\n",
 	    xtal.cell[c].a, xtal.cell[c].b, xtal.cell[c].c,
	    xtal.cell[c].alpha, xtal.cell[c].beta,
	    xtal.cell[c].gamma);

    for (unsigned a         = 0; a < mdl.atom.size() ; a++)
      if (c                == mdl.atom[a].crystal)
      {
	fprintf(sfile, "  ATOM %s SITE %u\n", mdl.atom[a].name.c_str(),
	        mdl.atom[a].nsite + 1);
	fprintf(sfile, "    OCCU %7.3f", mdl.atom[a].occ);
	if (!mdl.atom[a].Getrefineocc())
	  fprintf(sfile, " NOREf");
	fprintf(sfile, "\n");
	if (mdl.atom[a].isotropic)
 	  fprintf(sfile, "    BISO %7.3f", mdl.atom[a].biso);
	else
	{
	  fprintf(sfile, "    UANO ");
	  for (unsigned b   = 0; b < 6; b++)
	    fprintf(sfile, "%7.3f ",mdl.atom[a].uaniso[b]);
	}
	if (!mdl.atom[a].Getrefinebfac())
	  fprintf(sfile, " NOREf");
	fprintf(sfile, "\n");
      }

    for (unsigned d         = 0; d < xtal.sf.size()   ; d++)
      if (c                == xtal.sf[d].nxtal)
      {
	w++;
	fprintf(sfile, "  DNAMe %s\n", xtal.sf[d].name.c_str());
	if (xtal.sf[d].anomalous)
	  fprintf(sfile, "    COLUmn  F+= %s  SF+= %s  F-= %s  SF-= %s\n",
		  xtal.sf[d].sdatap.c_str(), xtal.sf[d].sdevp.c_str(),
		  xtal.sf[d].sdatam.c_str(), xtal.sf[d].sdevm.c_str());
	else
	  fprintf(sfile, "    COLUmn  F= %s  SF= %s\n",
		  xtal.sf[d].sdatap.c_str(),  xtal.sf[d].sdevp.c_str());
	
	fprintf(sfile, "    RESOlution %7.2f %7.2f\n", 
	        xtal.sf[d].hires, xtal.sf[d].lowres);
	fprintf(sfile, "    BINS %u\n", xtal.sf[d].nbins);
	if (d              != 0)
	{
	  fprintf(sfile, "    ISOE ");
	  for (unsigned s   = 0; s < xtal.sf[d].nbins; s++)
	    fprintf(sfile, "%9.6f ",xtal.sf[d].dluz[s]);
	  fprintf(sfile, "\n");
	}
	if (xtal.sf[d].refinee)
	{
	  fprintf(sfile, "    CORE ");
	  for (unsigned s   = 0; s < xtal.sf[d].nbins; s++)
	    fprintf(sfile, "%9.6f ",xtal.sf[d].eluz[s]);
	  fprintf(sfile, "\n");
	}
	
	if (xtal.sf[d].anomalous)
	{
	  if (target       == "UNCO")
	  {
	    fprintf(sfile, "    ANOE ");
	    for (unsigned s = 0; s < xtal.sf[d].nbins; s++)
	      fprintf(sfile, "%9.6f ",xtal.sf[d].adluz[s]);
	    fprintf(sfile, "\n");
	  }
	  else if (target  == "SAD")
	  {
	    fprintf(sfile, "    SDLU ");
	    for (unsigned s = 0; s < xtal.sf[d].nbins; s++)
	      fprintf(sfile, "%9.6f ",xtal.sf[d].sdluz[s]);
	    fprintf(sfile, "\n");
	  }
	}
	for (unsigned f     = 0; f < mdl.form.size(); f++)
	  if (d            == xtal.dataset[mdl.form[f].Getcrystal()][w])
 	    fprintf(sfile, "    FORM  %s FP= %7.3f  FPP= %7.3f\n",
		   mdl.form[f].Getname().c_str(), mdl.form[f].Getfp(w), 
		   mdl.form[f].Getfpp(w));
      }
    fprintf(sfile, "\n\n");
  }
  fprintf(sfile, "# Refine all parameters\n");
  fprintf(sfile, "REFAll\n");
  fprintf(sfile, "\n\n");
  fprintf(sfile, "eof-bp3\n");  
  fclose(sfile);
}

void Bp3likelihood::finitedifftest()
{
  // Perform finite difference tests on various derivatives calculated:
  double likelihood2;

  if (target               == "UNCO")
  {
    printf("\nSCALE PARAMETERS\n");

    double value            = grad();

    for (unsigned d         = 0; d < xtal.sf.size(); d++)
    {
      printf("dLdkscale   = %f\n", dLdkscale[d]);
      printf("dLdbscale   = %f\n", dLdbscale[d]);
    }

    unsigned de             = xtal.sf.size()-1;

    xtal.sf[de].biso       += difftol;
    
    likelihood2             = function();

    printf("Likelihood from function = %f\n", likelihood2);
    printf("Likelihood from gradient = %f\n", value);
    printf("Finite diff              = %f\n", (likelihood2 - value)/difftol);

    value                   = grad();

    for (unsigned d         = 0; d < xtal.sf.size(); d++)
    {
      printf("dLdkscale   = %f\n", dLdkscale[d]);
      printf("dLdbscale   = %f\n", dLdbscale[d]);
    }
    
    xtal.sf[de].kscale     += difftol;
    
    likelihood2 = function();    

    printf("Likelihood from function = %f\n", likelihood2);
    printf("Likelihood from gradient = %f\n", value);
    printf("Finite diff              = %f\n", (likelihood2 - value)/difftol);

    printf("\nNON-ISOMORPHISM PARAMETERS\n");    

    value                   = grad();
  
    for (unsigned d         = 0; d < xtal.sf.size()    ; d++)
    {
      for (unsigned s       = 0; s < xtal.sf[d].nbins; s++)
	printf("GradDluz[%u][%u]= %f \n",d,s,dLddluz[d][s]);
      printf("\n");
      for (unsigned s       = 0; s < xtal.sf[d].nbins; s++)
	printf("GradEluz[%u][%u]= %f \n",d,s,dLdeluz[d][s]);
      printf("\n");
    }

    unsigned sh             = xtal.sf[de].nbins -1;

    if (xtal.sf[de].refined)
      xtal.sf[de].dluz[sh] += difftol;
    else if (xtal.sf[de].refinee)
      xtal.sf[de].eluz[sh] += difftol;
    
    likelihood2             = function();

    printf("Likelihood from function = %f\n", likelihood2);
    printf("Likelihood from gradient = %f\n", value);
    printf("Finite diff              = %f\n", (likelihood2 - value)/difftol);

    if (xtal.sf[de].anomalous)
    {
      value                 = grad();
  
      for (unsigned d       = 0; d < xtal.sf.size()  ; d++)
      {
	for (unsigned s     = 0; s < xtal.sf[d].nbins; s++)
	  printf("GradADluz[%u][%u]= %f \n",d,s,dLdadluz[d][s]);
 	printf("\n");
      }
      
      xtal.sf[de].adluz[0] += difftol;

      likelihood2           = function();

      printf("Likelihood from function = %f\n", likelihood2);
      printf("Likelihood from gradient = %f\n", value);
      printf("Finite diff              = %f\n", (likelihood2 - value)/difftol);
    }
  }
  else if ( (target        == "SAD") || (target == "MAD") )
  {
    double value            = grad();
  
    for (unsigned d         = 0; d < xtal.sf.size()    ; d++)
    {
      for (unsigned s       = 0; s < xtal.sf[d].nbins; s++)
	printf("GradSDluz[%u][%u]= %f \n",d,s,dLdsdluz[d][s]);
      printf("\n");
    }
    
    unsigned s              =xtal.sf[0].nbins -1;
    
    if (xtal.sf[0].refinesd)
      xtal.sf[0].sdluz[s]  -= difftol;
    
    likelihood2             = function();

    printf("Likelihood from function = %f\n", likelihood2);
    printf("Likelihood from gradient = %f\n", value);
    printf("Finite diff              = %f\n", (value - likelihood2)/difftol);
  }

  directsfcalc(false,updatesigmah);  
  
  printf("\nATOMIC PARAMETERS\n");    

  unsigned id1            = mdl.site.size()-1;

  double value            = grad();

  for (unsigned i         = 0; i < mdl.site.size(); i++)
  {
    printf(" Grad[%u][X]   = %f\n", i, dLdx[i][0]);
    printf(" Grad[%u][Y]   = %f\n", i, dLdx[i][1]);
    printf(" Grad[%u][Z]   = %f\n", i, dLdx[i][2]);
  }

  mdl.site[id1].Setx(2,mdl.site[id1].x[2] + difftol);
  
  directsfcalc(false,updatesigmah);

  likelihood2             = function();

  printf("Likelihood from function = %f\n", likelihood2);
  printf("Likelihood from gradient = %f\n", value);
  printf("Finite diff              = %f\n", (likelihood2 - value)/difftol);

  value                   = grad();
  
  for (unsigned i         = 0; i < mdl.site.size(); i++)
  {
    printf(" Grad[%u][X]   = %f\n", i, dLdx[i][0]);
    printf(" Grad[%u][Y]   = %f\n", i, dLdx[i][1]);
    printf(" Grad[%u][Z]   = %f\n", i, dLdx[i][2]);
  }
    
  mdl.site[id1].Setx(0,mdl.site[id1].x[0] + difftol);
  
  directsfcalc(false,updatesigmah);

  likelihood2             = function();

  printf("Likelihood from function = %f\n", likelihood2);
  printf("Likelihood from gradient = %f\n", value);
  printf("Finite diff              = %f\n", (likelihood2 - value)/difftol); 

  value                   = grad();

  for (unsigned i         = 0; i < mdl.atom.size(); i++)
  {
    printf(" Grad[%u][O]   = %f\n",i,dLdocc[i]);
    if (mdl.atom[id1].isotropic)
      printf(" Grad[%u][B]   = %f\n",i, dLdbiso[i]);
    else
    {
      printf(" Grad[%u][U1]   = %f\n",i, dLduaniso[i][0]);
      printf(" Grad[%u][U2]   = %f\n",i, dLduaniso[i][1]);
      printf(" Grad[%u][U3]   = %f\n",i, dLduaniso[i][2]);
      printf(" Grad[%u][U4]   = %f\n",i, dLduaniso[i][3]);
      printf(" Grad[%u][U5]   = %f\n",i, dLduaniso[i][4]);
      printf(" Grad[%u][U6]   = %f\n",i, dLduaniso[i][5]);
    }
      /*
      printf(" Grad[" << i << "][f']  = " << dLdfp[i]   << "n";
      printf(" Grad[" << i << "][f''] = " << dLdfpp[i]  << "n";
      */
  }
    
  if (mdl.atom[id1].isotropic)
    mdl.atom[id1].Setbiso(mdl.atom[id1].biso + difftol);
  else
    mdl.atom[id1].Setuaniso(mdl.atom[id1].uaniso[0] + difftol, 0);
  
  directsfcalc(false,updatesigmah);

  likelihood2             = function();

  printf("Likelihood from function = %f\n", likelihood2);
  printf("Likelihood from gradient = %f\n", value);
  printf("Finite diff              = %f\n", (likelihood2 - value)/difftol);
  
  value                   = grad();
    
  for (unsigned i         = 0; i < mdl.atom.size(); i++)
    printf(" Grad[%u][O] = %f\n",i,dLdocc[i]);
  
  mdl.atom[id1].Setocc(mdl.atom[id1].occ+difftol);
  
  directsfcalc(false,updatesigmah);
  
  likelihood2             = function();

  printf("Likelihood from function = %f\n", likelihood2);
  printf("Likelihood from gradient = %f\n", value);
  printf("Finite diff              = %f\n", (likelihood2 - value)/difftol);  
}

void Bp3likelihood::printfom() const
{
  printf(" $TABLE: Figure of merit vs. 1/Res^2-centric, acentric, overall:\n");
  printf("$GRAPHS: Figure of merit vs. 1/Res^2 :N:4,6,8,10:\n");

  double totalcfom(ZERO);
  unsigned totalcnshl(0);

  double totalafom(ZERO);
  unsigned totalanshl(0);

  printf("$$\n Bin   HiRes  LoRes  1/Res^2   Refls  Centr    Refls  Acentr   Refls   All  $$\n");

  printf("$$\n");

  for (unsigned s = 0; s < cfom.size(); s++)
  {
    double cfbin  = (xtal.sf[0].cnshl[s]) ? cfom[s]/(double)(xtal.sf[0].cnshl[s]) : ZERO;
    double afbin  = (xtal.sf[0].anshl[s]) ? afom[s]/(double)(xtal.sf[0].anshl[s]) : ZERO;
    double tfbin  = ( (xtal.sf[0].anshl[s]+xtal.sf[0].cnshl[s]) ?
		      (afom[s] + cfom[s])/
		      (double)(xtal.sf[0].anshl[s] + xtal.sf[0].cnshl[s]) : ZERO);
    
    printf(" %2d  %6.2f  %6.2f  %7.5f  %5d  %7.5f  %5d  %7.5f  %5d  %7.5f\n",
	   s+1, xtal.sf[0].Getlores(s), xtal.sf[0].Gethires(s), xtal.sf[0].astolsq[s],
	   xtal.sf[0].cnshl[s], cfbin, xtal.sf[0].anshl[s], afbin,
	   xtal.sf[0].anshl[s] + xtal.sf[0].cnshl[s], tfbin);
      
    totalcfom    += cfom[s];
    totalcnshl   += xtal.sf[0].cnshl[s];
    totalafom    += afom[s];
    totalanshl   += xtal.sf[0].anshl[s];
  }
  double totcfom  = (totalcnshl) ? totalcfom/(double)(totalcnshl) : ZERO;
  double totafom  = (totalanshl) ? totalafom/(double)(totalanshl) : ZERO;
  double totfom   = ( (totalanshl+ totalcnshl) ?
		      (totalafom+totalcfom)/(double)(totalanshl+ totalcnshl) : ZERO);

  printf("$$\nTOTAL\n");
  printf("     %6.2f  %6.2f           %5d  %7.5f  %5d  %7.5f  %5d  %7.5f\n\n",
	 xtal.sf[0].Getlores(0), xtal.sf[0].Gethires(xtal.sf[0].Getnbins()-1),
	 totalcnshl, totcfom, totalanshl, totafom, totalanshl+totalcnshl, totfom);

  // printf("\n\nThe overall FOM is %.3f\n\n",totfom);
  fflush(stdout);
}

void Bp3likelihood::print(const int cyc) const
{
  if (stats)
  {
    printfom();

    if (target       != "SAD")
      for (unsigned d = 0; d < xtal.sf.size(); d++)
	xtal.sf[d].printstat();
  }
  
  if ( (refinescale) && (target != "SAD") )
    xtal.printscale();

  if (refineluzzati)
    xtal.printLuzzati(target, cyc);

  mdl.print();
}
