/*
  Copyright (c) Navraj Pannu       1998 - 2005
  Copyright (c) Leiden University  2002 - 2005

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

#include "misc.h"
#include "bp3input.h"
#include "crystal.h"
#include "bp3likelihood.h"
#include "minimizer.h"
#include "model.h"
#include <string>

// local functions
void Banner(int, char**);
void GetInput(Model &, Crystal &, Bp3likelihood &, Minimizer &);
void Refine(Bp3likelihood &, Minimizer &);
void Phase(Bp3likelihood &, Minimizer &);
void Check(Bp3likelihood &, Minimizer &);
void Intensity(Bp3likelihood &);
void Outputhkl(Bp3likelihood &);
void Outputscalepack(Bp3likelihood &);
void Script(Bp3likelihood &);
void Diff(Bp3likelihood &);

int main(int argc, char** argv)
{
  Banner(argc, argv); 
  // initialize classes for our parameters and data
  Model mdl;
  Crystal xtal;
  Bp3likelihood likelihood(mdl, xtal);
  Minimizer minimizer(likelihood);
  
  GetInput(mdl, xtal, likelihood, minimizer);

  if (likelihood.Getmode()      == "REFINE")
    Refine(likelihood, minimizer);
  else if (likelihood.Getmode() == "DIFF")
    Diff(likelihood);
  else if (likelihood.Getmode() == "PHASE")
    Phase(likelihood, minimizer);
  else if (likelihood.Getmode() == "CHECK")
    Check(likelihood, minimizer);
  else if (likelihood.Getmode() == "INTENSITY")
    Intensity(likelihood);
  else if (likelihood.Getmode() == "OUTPUTHKL")
    Outputhkl(likelihood);
  else if (likelihood.Getmode() == "OUTPUTSCALEPACK")
    Outputscalepack(likelihood);
  
  Termination();
}

void Banner(int argc, char** argv)
{
  char version[] = "1.00";
  CCP4Banner(argc,argv,version);
  printf("                           BP3\n");
  printf(" Multivariate likelihood substructure refinement and phasing\n");
  printf("                       21 March 2007\n\n");
  printf("        http://www.bfsc.leidenuniv.nl/software/bp3\n\n");
}

void GetInput(Model &mdl, Crystal &xtal,
	      Bp3likelihood &like, Minimizer &minimizer)
{
  // gets input and stores it in data structures
  Bp3input input(mdl, xtal, like, minimizer);
}

void Refine(Bp3likelihood &likelihood, Minimizer &minimizer)
{
  if (!likelihood.Getrefineall())
  {
    printf("Refinement cycle\n\n");
    likelihood.Setminocc(0.0075).setup(true);
    minimizer.refine();
    likelihood.Setminocc(ZERO);
  }

  if (!likelihood.Getsheldrick())
  {    
    printf("Refinement cycle\n\n");
    // likelihood.checkatoms();
    likelihood.setup(false);
    if (likelihood.Getmad())
      minimizer.Setiterations((unsigned)(likelihood.Getnpars()*1.25));
    minimizer.refine();
  }

  printf("Generating phases and phase probability statistics\n\n");
  likelihood.setup(true,false);
  likelihood.grad(true,true);
  likelihood.Setstats(true);
  likelihood.print();

  Script(likelihood);

  if (likelihood.Gethand())
  {
    // printf("Refinement cycle for other hand (if needed)\n\n");
    likelihood.otherhand();
    likelihood.Setstats(false);

    likelihood.setup(false,true,true);
    /*
    if ( (likelihood.norm()     > 10000.0) && !likelihood.Getmad() )
    {
      likelihood.setup(true);
      minimizer.refine();
      likelihood.setup(false,true,true);
      minimizer.refine();
    }
    else if (likelihood.norm() > 250.0)
      minimizer.refine();
    */

    printf("Generating phases and phase probability statistics for the other hand\n\n");
    likelihood.setup(true,false);
    likelihood.grad(true,true);
    likelihood.Setstats(true);
    likelihood.print();

    Script(likelihood);
  }
}

void Diff(Bp3likelihood &likelihood)
{
  printf("Perform finite difference tests\n\n");
  likelihood.setup(true,true);
  likelihood.finitedifftest();
}

void Phase(Bp3likelihood &likelihood, Minimizer &minimizer)
{
  if (!minimizer.Getsetit())
    minimizer.Setiterations(likelihood.Setiterations());

  if (minimizer.Getiterations())
  {
    if (!likelihood.Getrefineall())
    {
      likelihood.Setminocc(0.025).setup(true);
      minimizer.refine();
      likelihood.Setsheldrick(true).Setminocc(ZERO).setup(true);
      minimizer.refine();
    }
    // likelihood.checkatoms();
    likelihood.Setsheldrick(false).setup(false,true,true);
    minimizer.refine();
  }
 
  printf("Generating phases and phase probability statistics\n\n");
  likelihood.setup(true,false);
  likelihood.grad(true,true);
  likelihood.Setstats(true);
  likelihood.print();

  Script(likelihood);

  if (likelihood.Gethand())
  {
    //   printf("Refinement cycle for other hand\n\n");
    likelihood.otherhand();
    likelihood.Setstats(false);
    likelihood.setup(false,true,true);
    /*
    if (likelihood.norm() > 250.0)
      minimizer.refine();
    */

    printf("Generating phases and phase probability statistics for the other hand\n\n");
    likelihood.setup(true,false);
    likelihood.grad(true,true);
    likelihood.Setstats(true);
    likelihood.print();

    Script(likelihood);
  }
}

void Check(Bp3likelihood &likelihood, Minimizer &minimizer)
{
  if (!minimizer.Getsetit())
    minimizer.Setiterations(likelihood.Setiterations());
  
  if (minimizer.Getiterations())
  {
    likelihood.Setminocc(0.025).setup(true);
    minimizer.refine();
    likelihood.Setsheldrick(true).Setminocc(ZERO).setup(true);
    minimizer.refine();
  }
 
  printf("Generating phases and phase probability statistics\n\n");
  likelihood.setup(true,false);
  likelihood.grad(true,true);
  likelihood.Setstats(true);
  likelihood.print();

  Script(likelihood);

}

void Intensity(Bp3likelihood &likelihood)
{
  printf("Outputting intensity file\n\n");
  likelihood.setup(true,false);
  likelihood.outputintensities();
}

void Outputhkl(Bp3likelihood &likelihood)
{
  printf("Outputting hkl file\n\n");
  likelihood.setup(true,false);
  likelihood.outputhklfile();
}

void Outputscalepack(Bp3likelihood &likelihood)
{
  printf("Outputting hkl file\n\n");
  likelihood.setup(true,false);
  likelihood.outputscafile();
}

void Script(Bp3likelihood &likelihood)
{
  printf("Writing out script for further refinement, crank xml and pdb file\n\n");
  likelihood.writescript();
  likelihood.writexml();
  likelihood.writepdb();
}
