/*
  Copyright (c) Navraj Pannu       1998 - 2005
  Copyright (c) Leiden University  2002 - 2005

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

#include <ctype.h>
#include "gcxinput.h"
#include "ccp4_parser.h"                  // Peter Briggs' parser in c libraries

Gcxinput::Gcxinput(Model &model, Crystal &xtl, Gcxlikelihood &likelihood,
		   Minimizer &minimizer)
  : Input(model, xtl, likelihood, minimizer), like(likelihood)
{

  // parse
  CCP4parse();

  CCP4::ccp4_parse_end(parser); // clean up parser array
  
  check();

  if (!like.Getnooutput())
    print();

  for (unsigned d = 0; d < xtal.sf.size(); d++)
    if (!xtal.sf[d].Getnbins())
      xtal.sf[d].Setusernbins(10);

  if (getenv("HKLIN") != NULL)
    xtal.setwithmtz("ALL");
  else
    xtal.setwithmanymtz("ALL",false);
}

void Gcxinput::check()
{
  Input::check();
}

void Gcxinput::CCP4parse()
{
  // The setup of reading keyword input and it follows pretty much
  // exactly Peter Briggs' documentation of his parser.

  // read lines from stdin until END or EOF is reached
  bool cont       = true;
  char line[201];

  while (cont)
  {
    // Blank line before calling ccp4_parser implies reading from stdin
    line[0]       = '\0';
    unsigned ntok = CCP4::ccp4_parser(line, 200, parser, 1);
    
    if (ntok      < 1)                                            // reached EOF
      break;
    else   // perform keyword analysis and interpretation
    {
      if (Input::xtalkeywords())
	continue;
      else if (Input::datakeywords())
	continue;
      else if (Input::modelkeywords())
	continue;
      else if (Input::minimizerkeywords())
	continue;
      else if (Input::likelihoodkeywords())
	continue;
      else if (gcxkeywords())
	continue;
      else if (CCP4::ccp4_keymatch("END", parser->keyword))
	break;
      else
	Bp3Error("CCP4parse","Keyword not recognised");
    }
  }  
}
bool Gcxinput::gcxkeywords()
{
  if (CCP4::ccp4_keymatch("SOLV", parser->keyword))       // mode options
    solvkeyword();
  else if (CCP4::ccp4_keymatch("NRES", parser->keyword))
    nreskeyword();
  else if (CCP4::ccp4_keymatch("NNUC", parser->keyword))
    nnuckeyword();
  else if (CCP4::ccp4_keymatch("NMON", parser->keyword))
    nmonkeyword();
  else if (CCP4::ccp4_keymatch("SADO", parser->keyword))
    sadokeyword();
  else if (CCP4::ccp4_keymatch("DISK", parser->keyword))
    diskkeyword();
  else if (CCP4::ccp4_keymatch("OBFA", parser->keyword))
    obfakeyword();
  else if (CCP4::ccp4_keymatch("PIPE", parser->keyword))
    pipekeyword();
  else if (CCP4::ccp4_keymatch("PREP", parser->keyword))
    prepkeyword();
  else if (CCP4::ccp4_keymatch("AFRO", parser->keyword))
    afrokeyword();
  else if (CCP4::ccp4_keymatch("CRUN", parser->keyword))
    crunkeyword();
  else if (CCP4::ccp4_keymatch("BP3", parser->keyword))
    bp3keyword();
  else if (CCP4::ccp4_keymatch("REFM", parser->keyword))
    refmkeyword();
  else if (CCP4::ccp4_keymatch("SOLO", parser->keyword))
    solokeyword();
  else if (CCP4::ccp4_keymatch("DM", parser->keyword))
    dmkeyword();
  else if (CCP4::ccp4_keymatch("PIRA", parser->keyword))
    pirakeyword();
  else if (CCP4::ccp4_keymatch("BUCC", parser->keyword))
    bucckeyword();
  else if (CCP4::ccp4_keymatch("REDM", parser->keyword))
    redmkeyword();
  else if (CCP4::ccp4_keymatch("REBU", parser->keyword))
    rebukeyword();
  else if (CCP4::ccp4_keymatch("WARP", parser->keyword))
    warpkeyword();
  else if (CCP4::ccp4_keymatch("SHLC", parser->keyword))
    shlckeyword();
  else if (CCP4::ccp4_keymatch("SHLD", parser->keyword))
    shldkeyword();
  else if (CCP4::ccp4_keymatch("SHLE", parser->keyword))
    shlekeyword();
  else if (CCP4::ccp4_keymatch("CHEC", parser->keyword))
    checkeyword();
  else
    return false;

  return true;
}

void Gcxinput::solvkeyword()
{
  // check to see if we have no more than one token
  if (parser->ntokens > 2)
    Bp3Warning("Gcxinput::solvkeyword", "Extra text in SOLV keyword will be ignored");

  if (!parser->token[1].isnumber)	
    Bp3Error("Gcxinput::solvkeyword", "A numerical value is expected for SOLV");
  
  like.Setsolvent((double)(parser->token[1].value));
}

void Gcxinput::nreskeyword()
{
  // check to see if we have no more than one token
  if (parser->ntokens > 2)
    Bp3Warning("Gcxinput::nreskeyword", "Extra text in NRES keyword will be ignored");

  if (!parser->token[1].isnumber)	
    Bp3Error("Gcxinput::nreskeyword", "A numerical value is expected for NRES");
  
  like.Setnresidues((unsigned)(parser->token[1].value));
}

void Gcxinput::nnuckeyword()
{
  // check to see if we have no more than one token
  if (parser->ntokens > 2)
    Bp3Warning("Gcxinput::nnuckeyword", "Extra text in NNUC keyword will be ignored");

  if (!parser->token[1].isnumber)	
    Bp3Error("Gcxinput::nnuckeyword", "A numerical value is expected for NNUC");
  
  like.Setnnucleotides((unsigned)(parser->token[1].value));
}

void Gcxinput::nmonkeyword()
{
  // check to see if we have no more than one token
  if (parser->ntokens > 2)
    Bp3Warning("Gcxinput::nmonkeyword", "Extra text in NMON keyword will be ignored");

  if (!parser->token[1].isnumber)	
    Bp3Error("Gcxinput::nmonkeyword", "A numerical value is expected for NMON");
  
  like.Setnmonomers((unsigned)(parser->token[1].value));
}

void Gcxinput::sadokeyword()
{
  // check to see if we have no more than one token
  if (parser->ntokens > 1)
    Bp3Warning("Gcxinput::sadokeyword", "Extra text in SADOnly keyword will be ignored");
  
  like.Setsadonly(true);
}

void Gcxinput::diskkeyword()
{
  // check to see if we have no more than one token
  for (int i        = 1; i < parser->ntokens; i++)
    if (CCP4::ccp4_keymatch("CLEA",parser->token[i].word))
      like.Setdiskspace("clean");
    else if (CCP4::ccp4_keymatch("COMP",parser->token[i].word))
      like.Setdiskspace("compress");
    else
      Bp3Error("Gcxinput::diskkeyword", "Text after DISK keyword must be CLEAn or COMPress");

}

void Gcxinput::obfakeyword()
{
  // check to see if we have no more than one token
  if (parser->ntokens > 2)
    Bp3Warning("Gcxinput::obfakeyword", "Extra text in OBFA keyword will be ignored");

  if (!parser->token[1].isnumber)	
    Bp3Error("Gcxinput::obfakeyword", "A numerical value is expected for OBFA");
  
  like.Setoverallbfactor((double)(parser->token[1].value));
}

void Gcxinput::pipekeyword()
{
  // check to see if we have no more than one token

  if (!parser->token[1].isnumber)	
    Bp3Error("Gcxinput::pipekeyword", "A 1, 2, 3, 4 or 5 is expected after PIPE");

  if (parser->ntokens > 2)
    for (int i        = 2; i < parser->ntokens; i++)
      if (CCP4::ccp4_keymatch("LAST",parser->token[i].word))
	if ( (i+1)    < parser->ntokens)
	  if (CCP4::ccp4_keymatch("DETE",parser->token[i+1].word))
	    like.Setlaststep("DETECT");
	  else if (CCP4::ccp4_keymatch("PHAS",parser->token[i+1].word))
	    like.Setlaststep("PHASE");
	  else if (CCP4::ccp4_keymatch("DM",parser->token[i+1].word))
	    like.Setlaststep("DM");
	  else if (CCP4::ccp4_keymatch("BUIL",parser->token[i+1].word))
	    like.Setlaststep("BUILD");
	  else
	    Bp3Error("Gcxinput::pipekeyword", "Text after LAST keyword must be DETEct, PHASe, DM or BUILd");
	else
	  Bp3Error("Gcxinput::pipekeyword", "Text after LAST keyword must be DETEct, PHASe, DM or BUILd");
      else if (CCP4::ccp4_keymatch("FIRS",parser->token[i].word))
	if ( (i+1)    < parser->ntokens)
	  if (CCP4::ccp4_keymatch("DETE",parser->token[i+1].word))
	    like.Setfirststep("DETECT");
	  else if (CCP4::ccp4_keymatch("PHAS",parser->token[i+1].word))
	    like.Setfirststep("PHASE");
  /*
	  else if (CCP4::ccp4_keymatch("DM",parser->token[i+1].word))
	    like.Setlaststep("DM");
	  else if (CCP4::ccp4_keymatch("BUIL",parser->token[i+1].word))
	    like.Setlaststep("BUILD");
  */
	  else
	    Bp3Error("Gcxinput::pipekeyword", "Text after FIRSt keyword must be DETEct or PHASe");
	else
	  Bp3Error("Gcxinput::pipekeyword", "Text after FIRSt keyword must be DETEct or PHASe");

  like.Setpipeline((unsigned)(parser->token[1].value));
}

void Gcxinput::prepkeyword()
{
  unsigned number(0);
  
  if (!parser->token[1].isnumber)	
    Bp3Error("Gcxinput::prepkeyword", "An integer value is expected for step");
  else
    number    = (unsigned)(parser->token[1].value);

  if (number  < 1)
    Bp3Error("Gcxinput::prepkeyword", "Step number must be a positive integer");
    
  
  if (like.Getstepsize()     ==  number - 1)
    like.Setstep("PREP");
  else if (like.Getstepsize() > number - 1)
    if (like.Getstep(number-1) != "PREP")
      Bp3Error("Gcxinput::prepkeyword", "PREP has not been set to this step");

  // at the moment, no prep options in parser

  if (parser->ntokens           > 2)
    Bp3Warning("Gcxinput::prepkeyword", "Extra text after PREP <num> will be ignored");
  
  //  for (int i                 = 1; i < parser->ntokens; i++)
}

void Gcxinput::afrokeyword()
{
  unsigned number(0);
  
  if (!parser->token[1].isnumber)	
    Bp3Error("Gcxinput::afrokeyword", "An integer value is expected for step");
  else
    number    = (unsigned)(parser->token[1].value);

  if (number  < 1)
    Bp3Error("Gcxinput::afrokeyword", "Step number must be a positive integer");
    
  
  if (like.Getstepsize()     ==  number - 1)
    like.Setstep("AFRO");
  else if (like.Getstepsize() > number - 1)
    if (like.Getstep(number-1) != "AFRO")
      Bp3Error("Gcxinput::afrokeyword", "AFRO has not been set to this step");

  for (int i                 = 2; i < parser->ntokens; i++)
  {
    if (CCP4::ccp4_keymatch("SIGF",parser->token[i].word))
      if ( (i+1)             < parser->ntokens)
	if (parser->token[i+1].isnumber)
	{
	  like.Setafrosigf((double)(parser->token[i+1].value));
	  i++;
	}
	else
	  Bp3Error("Gcxinput::afrokeyword","A positive number is expected for SIGF");
      else
	Bp3Error("Gcxinput::afrokeyword","A positive number is expected for SIGF");
    else if (CCP4::ccp4_keymatch("SANO",parser->token[i].word))
      if ( (i+1)             < parser->ntokens)
	if (parser->token[i+1].isnumber)
	{
	  like.Setafrosano((double)(parser->token[i+1].value));
	  i++;
	}
	else
	  Bp3Error("Gcxinput::afrokeyword","A positive number is expected for SANO");
      else
	Bp3Error("Gcxinput::afrokeyword","A positive number is expected for SANO");
    else if (CCP4::ccp4_keymatch("SISO",parser->token[i].word))
      if ( (i+1)             < parser->ntokens)
	if (parser->token[i+1].isnumber)
	{
	  like.Setafrosiso((double)(parser->token[i+1].value));
	  i++;
	}
	else
	  Bp3Error("Gcxinput::afrokeyword","A positive number is expected for SISO");
      else
	Bp3Error("Gcxinput::afrokeyword","A positive number is expected for SISO");
    else if (CCP4::ccp4_keymatch("HIRE",parser->token[i].word))
      if ( (i+1)             < parser->ntokens)
	if (parser->token[i+1].isnumber)
	{
	  like.Setsubhirescut((double)(parser->token[i+1].value));
	  like.Setautoreso(false);
	  i++;
	}
	else
	  Bp3Error("Gcxinput::afrokeyword","A positive number is expected for HIREs");
      else
	Bp3Error("Gcxinput::afrokeyword","A positive number is expected for HIREs");
    else
      Bp3Error("Gcxinput::afrokeyword","Subkeyword of AFRO not recognized");
      
  }
}

void Gcxinput::crunkeyword()
{
  unsigned number(0);
  
  if (!parser->token[1].isnumber)	
    Bp3Error("Gcxinput::crunkeyword", "A positive integer value is expected for step");
  else
    number    = (unsigned)(parser->token[1].value);

  if (number  < 1)
    Bp3Error("Gcxinput::crunkeyword", "Step number must be a positive integer");
    
  
  if (like.Getstepsize()     ==  number - 1)
    like.Setstep("CRUNCH2");
  else if (like.Getstepsize() > number - 1)
    if (like.Getstep(number-1) != "CRUNCH2")
      Bp3Error("Gcxinput::crunkeyword", "CRUNCH2 has not been set to this step");

  for (int i                 = 2; i < parser->ntokens; i++)
  {
    if (CCP4::ccp4_keymatch("TRIA",parser->token[i].word))
      if ( (i+1)             < parser->ntokens)
	if (parser->token[i+1].isnumber)
	{
	  like.Setcrunchtrials((unsigned)(parser->token[i+1].value));
	  i++;
	}
	else
	  Bp3Error("Gcxinput::crunkeyword","A positive number is expected for TRIAls");
      else
	Bp3Error("Gcxinput::crunkeyword","A positive number is expected for TRIAls");
    else if (CCP4::ccp4_keymatch("PTRI",parser->token[i].word))
      if ( (i+1)             < parser->ntokens)
	if (parser->token[i+1].isnumber)
	{
	  like.Setcrunchptrials((unsigned)(parser->token[i+1].value));
	  i++;
	}
	else
	  Bp3Error("Gcxinput::crunkeyword","A positive number is expected for PTRIals");
      else
	Bp3Error("Gcxinput::crunkeyword","A positive number is expected for PTRIals");
    else if (CCP4::ccp4_keymatch("THRE",parser->token[i].word))
      if ( (i+1)             < parser->ntokens)
	if (parser->token[i+1].isnumber)
	{
	  like.Setcrunchthres((double)(parser->token[i+1].value));
	  i++;
	}
	else
	  Bp3Error("Gcxinput::crunkeyword","A positive number is expected for THREsh");
      else
	Bp3Error("Gcxinput::crunkeyword","A positive number is expected for THREsh");
    else if (CCP4::ccp4_keymatch("BP3T",parser->token[i].word))
      if ( (i+1)             < parser->ntokens)
	if (parser->token[i+1].isnumber)
	{
	  like.Setcrunchbp3((unsigned)(parser->token[i+1].value));
	  i++;
	}
	else
	  Bp3Error("Gcxinput::crunkeyword","A number is expected for BP3T");
      else
	Bp3Error("Gcxinput::crunkeyword","A number is expected for BP3T");
    else if (CCP4::ccp4_keymatch("SPEC",parser->token[i].word))
      like.Setcrunchspec(true);
    else if (CCP4::ccp4_keymatch("DEVI",parser->token[i].word))
      if ( (i+1)             < parser->ntokens)
	if (parser->token[i+1].isnumber)
	{
	  like.Setcrunchdev((double)(parser->token[i+1].value));
	  i++;
	}
	else
	  Bp3Error("Gcxinput::crunkeyword","A positive number is expected for DEVIation");
      else
	Bp3Error("Gcxinput::crunkeyword","A positive number is expected for DEVIation");
    else
      Bp3Error("Gcxinput::crunkeyword","Subkeyword of CRUN not recognized");
  } 	
}

void Gcxinput::bp3keyword()
{
  unsigned number(0);
  
  if (!parser->token[1].isnumber)	
    Bp3Error("Gcxinput::bp3keyword", "A positive integer value is expected for step");
  else
    number    = (unsigned)(parser->token[1].value);

  if (number  < 1)
    Bp3Error("Gcxinput::bp3keyword", "Step number must be a positive integer");
    
  
  if (like.Getstepsize()     ==  number - 1)
    like.Setstep("BP3");
  else if (like.Getstepsize() > number - 1)
    if (like.Getstep(number-1) != "BP3")
      Bp3Error("Gcxinput::bp3keyword", "BP3 has not been set to this step");

  for (int i                 = 2; i < parser->ntokens; i++)
  {
    if (CCP4::ccp4_keymatch("STOP",parser->token[i].word))
      if ( (i+1)             < parser->ntokens)
	if (parser->token[i+1].isnumber)
	{
	  like.Setbp3stopfom((double)(parser->token[i+1].value));
	  i++;
	}
	else
	  Bp3Error("Gcxinput::bp3keyword","A positive number is expected for STOP");
      else
	Bp3Error("Gcxinput::bp3keyword","A positive number is expected for STOP");
    else if (CCP4::ccp4_keymatch("PHASE",parser->token[i].word))
      like.Setbp3phase("FAST");
    else
      Bp3Error("Gcxinput::bp3keyword","Subkeyword of BP3 not recognized");
  }
}

void Gcxinput::refmkeyword()
{
  unsigned number(0);
  
  if (!parser->token[1].isnumber)	
    Bp3Error("Gcxinput::refmkeyword", "A positive integer value is expected for step");
  else
    number    = (unsigned)(parser->token[1].value);

  if (number  < 1)
    Bp3Error("Gcxinput::refmkeyword", "Step number must be a positive integer");
    
  
  if (like.Getstepsize()     ==  number - 1)
    like.Setstep("REFMAC");
  else if (like.Getstepsize() > number - 1)
    if (like.Getstep(number-1) != "REFMAC")
      Bp3Error("Gcxinput::refmkeyword", "REFMAC has not been set to this step");

  for (int i                 = 2; i < parser->ntokens; i++)
  {
    if (CCP4::ccp4_keymatch("STOP",parser->token[i].word))
      if ( (i+1)             < parser->ntokens)
	if (parser->token[i+1].isnumber)
	{
	  like.Setbp3stopfom((double)(parser->token[i+1].value));
	  i++;
	}
	else
	  Bp3Error("Gcxinput::refmkeyword","A positive number is expected for STOP");
      else
	Bp3Error("Gcxinput::refmkeyword","A positive number is expected for STOP");
    else if (CCP4::ccp4_keymatch("SOLV",parser->token[i].word))
      like.Setrefmacsolvent(true);
    else
      Bp3Error("Gcxinput::refmkeyword","Subkeyword of REFMac not recognized");
  }
}

void Gcxinput::solokeyword()
{
  unsigned number(0);
  
  if (!parser->token[1].isnumber)	
    Bp3Error("Gcxinput::solokeyword", "A positive integer value is expected for step");
  else
    number    = (unsigned)(parser->token[1].value);

  if (number  < 1)
    Bp3Error("Gcxinput::solokeyword", "Step number must be a positive integer");
    
  
  if (like.Getstepsize()     ==  number - 1)
    like.Setstep("SOLOMON");
  else if (like.Getstepsize() > number - 1)
    if (like.Getstep(number-1) != "SOLOMON")
      Bp3Error("Gcxinput::solokeyword", "SOLOMON has not been set to this step");

  for (int i                 = 2; i < parser->ntokens; i++)
    if (CCP4::ccp4_keymatch("NOHA",parser->token[i].word))
      like.Setsolomonhand(false);
    else if (CCP4::ccp4_keymatch("OPTI",parser->token[i].word))
      like.Setsolomonopt(true);
    else if (CCP4::ccp4_keymatch("NODM",parser->token[i].word))
      like.Setsolomondenmod(false);
    else if (CCP4::ccp4_keymatch("NCYC",parser->token[i].word))
    {
      like.Setsolomondenmod(true);
      if ( (i+1)             < parser->ntokens)
	if (parser->token[i+1].isnumber)
	{
	  like.Setsolomonncycles((unsigned)(parser->token[i+1].value));
	  i++;
	}
	else
	  Bp3Error("Gcxinput::solokeyword", "A positive integer is expected after NCYCles");
      else
	Bp3Error("Gcxinput::solokeyword", "A positive integer is expected after NCYCles");
    }
    else if (CCP4::ccp4_keymatch("HCYC",parser->token[i].word))
    {
      like.Setsolomonhand(true);
      if ( (i+1)             < parser->ntokens)
	if (parser->token[i+1].isnumber)
	{
	  like.Setsolomonhcycles((unsigned)(parser->token[i+1].value));
	  i++;
	}
	else
	  Bp3Error("Gcxinput::solokeyword", "A positive integer is expected after HCYCles");
      else
	Bp3Error("Gcxinput::solokeyword", "A positive integer is expected after HCYCles");
    }
    else if (CCP4::ccp4_keymatch("OCYC",parser->token[i].word))
    {
      like.Setsolomonopt(true);
      if ( (i+1)             < parser->ntokens)
	if (parser->token[i+1].isnumber)
	{
	  like.Setsolomonocycles((unsigned)(parser->token[i+1].value));
	  i++;
	}
	else
	  Bp3Error("Gcxinput::solokeyword", "A positive integer is expected after OCYCles");
      else
	Bp3Error("Gcxinput::solokeyword", "A positive integer is expected after OCYCles");
    }
    else
      Bp3Error("Gcxinput::solokeyword","Subkeyword of SOLOmon not recognized");

}

void Gcxinput::dmkeyword()
{
  unsigned number(0);
  
  if (!parser->token[1].isnumber)	
    Bp3Error("Gcxinput::dmkeyword", "A positive integer value is expected for step");
  else
    number    = (unsigned)(parser->token[1].value);

  if (number  < 1)
    Bp3Error("Gcxinput::dmkeyword", "Step number must be a positive integer");
    
  
  if (like.Getstepsize()     ==  number - 1)
    like.Setstep("DM");
  else if (like.Getstepsize() > number - 1)
    if (like.Getstep(number-1) != "DM")
      Bp3Error("Gcxinput::dmkeyword", "DM has not been set to this step");

  for (int i                 = 2; i < parser->ntokens; i++)
    if (CCP4::ccp4_keymatch("NOHA",parser->token[i].word))
      like.Setdmhand(false);
    else if (CCP4::ccp4_keymatch("NOOP",parser->token[i].word))
      like.Setdmopt(false);
    else if (CCP4::ccp4_keymatch("NODM",parser->token[i].word))
      like.Setdmdenmod(false);
    else if (CCP4::ccp4_keymatch("NOHI",parser->token[i].word))
      like.Setdmhist(false);
    else
      Bp3Error("Gcxinput::dmkeyword","Subkeyword of DM not recognized");
}

void Gcxinput::pirakeyword()
{
  unsigned number(0);
  
  if (!parser->token[1].isnumber)	
    Bp3Error("Gcxinput::pirakeyword", "A positive integer value is expected for step");
  else
    number    = (unsigned)(parser->token[1].value);

  if (number  < 1)
    Bp3Error("Gcxinput::pirakeyword", "Step number must be a positive integer");
    
  
  if (like.Getstepsize()     ==  number - 1)
    like.Setstep("PIRATE");
  else if (like.Getstepsize() > number - 1)
    if (like.Getstep(number-1) != "PIRATE")
      Bp3Error("Gcxinput::pirakeyword", "PIRATE has not been set to this step");

  for (int i                 = 2; i < parser->ntokens; i++)
    if (CCP4::ccp4_keymatch("NONC",parser->token[i].word))
      like.Setpiratencs(false);
    else if (CCP4::ccp4_keymatch("PIHL",parser->token[i].word))
      like.Setpiratehl(false);
    else if (CCP4::ccp4_keymatch("NCYC",parser->token[i].word))
    {
      like.Setpiratedenmod(true);
      if ( (i+1)             < parser->ntokens)
	if (parser->token[i+1].isnumber)
	{
	  like.Setpiratencycles((unsigned)(parser->token[i+1].value));
	  i++;
	}
	else
	  Bp3Error("Gcxinput::pirakeyword", "A positive integer is expected after NCYCles");
      else
	Bp3Error("Gcxinput::pirakeyword", "A positive integer is expected after NCYCles");
    }
    else if (CCP4::ccp4_keymatch("INWE",parser->token[i].word))
    {
      like.Setpiratedenmod(true);
      if ( (i+1)             < parser->ntokens)
	if (parser->token[i+1].isnumber)
	{
	  like.Setpiratewihl((double)(parser->token[i+1].value));
	  i++;
	}
	else
	  Bp3Error("Gcxinput::pirakeyword", "A positive real is expected after INWEight");
      else
	Bp3Error("Gcxinput::pirakeyword", "A positive real is expected after INWEight");
    }

    else
      Bp3Error("Gcxinput::pirakeyword","Subkeyword of PIRAte not recognized");
}

void Gcxinput::bucckeyword()
{
  unsigned number(0);
  
  if (!parser->token[1].isnumber)	
    Bp3Error("Gcxinput::bucckeyword", "A positive integer value is expected for step");
  else
    number    = (unsigned)(parser->token[1].value);

  if (number  < 1)
    Bp3Error("Gcxinput::bucckeyword", "Step number must be a positive integer");
    
  
  if (like.Getstepsize()     ==  number - 1)
    like.Setstep("BUCCANEER");
  else if (like.Getstepsize() > number - 1)
    if (like.Getstep(number-1) != "BUCCANEER")
      Bp3Error("Gcxinput::pirakeyword", "BUCCANEER has not been set to this step");

  for (int i                 = 2; i < parser->ntokens; i++)
    if (CCP4::ccp4_keymatch("NORF",parser->token[i].word))
      like.Setbuccnofreer(true);
    else if (CCP4::ccp4_keymatch("FAST",parser->token[i].word))
      like.Setbuccfast(true);
    else if (CCP4::ccp4_keymatch("NOPD",parser->token[i].word))
      like.Setbuccusepdb(false);
    else if (CCP4::ccp4_keymatch("FRAG",parser->token[i].word))
    {
      if ( (i+1)             < parser->ntokens)
	if (parser->token[i+1].isnumber)
	{
	  like.Setbuccfragments((unsigned)(parser->token[i+1].value));
	  i++;
	}
	else
	  Bp3Error("Gcxinput::bucckeyword", "A positive integer is expected after FRAGment");
      else
	Bp3Error("Gcxinput::bucckeyword", "A positive integer is expected after FRAGment");
    }
    else if (CCP4::ccp4_keymatch("NCYC",parser->token[i].word))
    {
      if ( (i+1)             < parser->ntokens)
	if (parser->token[i+1].isnumber)
	{
	  like.Setbucccycles((unsigned)(parser->token[i+1].value));
	  i++;
	}
	else
	  Bp3Error("Gcxinput::bucckeyword", "A positive integer is expected after CYCLes");
      else
	Bp3Error("Gcxinput::bucckeyword", "A positive integer is expected after CYCLes");
    }
    else if (CCP4::ccp4_keymatch("RELI",parser->token[i].word))
    {
      if ( (i+1)             < parser->ntokens)
	if (parser->token[i+1].isnumber)
	{
	  like.Setbuccreli((double)(parser->token[i+1].value));
	  i++;
	}
	else
	  Bp3Error("Gcxinput::bucckeyword", "A positive real is expected after RELIable");
      else
	Bp3Error("Gcxinput::bucckeyword", "A positive real is expected after RELIable");
    }
    else if (CCP4::ccp4_keymatch("TARG",parser->token[i].word))
      if ( (i+1)             < parser->ntokens)
	if (parser->token[i+1].isstring)
	{
	  if (CCP4::ccp4_keymatch("DIRE",parser->token[i+1].word) ||
	      CCP4::ccp4_keymatch("SAD",parser->token[i+1].word))
	    like.Setrefmactarget("DIRE");
 	  else if (CCP4::ccp4_keymatch("MLHL",parser->token[i+1].word))
	    like.Setrefmactarget("MLHL");
	  else if (CCP4::ccp4_keymatch("RICE",parser->token[i+1].word))
	    like.Setrefmactarget("RICE");
	  i++;
	}    
	else
	  Bp3Error("Gcxinput::bucckeyword","A string is expected for TARG");
      else
	Bp3Error("Gcxinput::bucckeyword","A string is expected for TARG");
    else
      Bp3Error("Gcxinput::bucckeyword","Subkeyword of BUCCaneer not recognized");
}

void Gcxinput::redmkeyword()
{
  unsigned number(0);
  
  if (!parser->token[1].isnumber)	
    Bp3Error("Gcxinput::redmkeyword", "A positive integer value is expected for step");
  else
    number    = (unsigned)(parser->token[1].value);

  if (number  < 1)
    Bp3Error("Gcxinput::redmkeyword", "Step number must be a positive integer");
    
  
  if (like.Getstepsize()     ==  number - 1)
    like.Setstep("RESOLVEDM");
  else if (like.Getstepsize() > number - 1)
    if (like.Getstep(number-1) != "RESOLVEDM")
      Bp3Error("Gcxinput::redmkeyword", "RESOLVEDM has not been set to this step");

  for (int i                 = 2; i < parser->ntokens; i++)
    if (CCP4::ccp4_keymatch("NONC",parser->token[i].word))
      like.Setresolvedmncs(false);
    else if (CCP4::ccp4_keymatch("PIHL",parser->token[i].word))
      like.Setresolvehl(false);
    else
      Bp3Error("Gcxinput::redmkeyword","Subkeyword of REDM not recognized");
}

void Gcxinput::rebukeyword()
{
  unsigned number(0);
  
  if (!parser->token[1].isnumber)	
    Bp3Error("Gcxinput::rebukeyword", "A positive integer value is expected for step");
  else
    number    = (unsigned)(parser->token[1].value);

  if (number  < 1)
    Bp3Error("Gcxinput::rebukeyword", "Step number must be a positive integer");
    
  
  if (like.Getstepsize()     ==  number - 1)
    like.Setstep("RESOLVEBUILD");
  else if (like.Getstepsize() > number - 1)
    if (like.Getstep(number-1) != "RESOLVEBUILD")
      Bp3Error("Gcxinput::rebukeyword", "RESOLVEBUILD has not been set to this step");

  for (int i                 = 2; i < parser->ntokens; i++)
    if (CCP4::ccp4_keymatch("NONC",parser->token[i].word))
      like.Setresolvembncs(false);
    else if (CCP4::ccp4_keymatch("BUIL",parser->token[i].word))
      if ( (i+1)             < parser->ntokens)
	if (parser->token[i+1].isstring)
 	  if (CCP4::ccp4_keymatch("SUPE",parser->token[i+1].word))
	  {
	    like.Setresolvetype("superquick");
	    i++;
	  }
	  else if (CCP4::ccp4_keymatch("QUIC",parser->token[i+1].word))
	  {
	    like.Setresolvetype("quick");
	    i++;
	  }
	  else if (CCP4::ccp4_keymatch("THOR",parser->token[i+1].word))
	  {
	    like.Setresolvetype("thorough");
	    i++;
	  }
	  else if (CCP4::ccp4_keymatch("AGGR",parser->token[i+1].word))
	  {
	    like.Setresolvetype("aggressive");
	    i++;
	  }
	  else if (CCP4::ccp4_keymatch("CONS",parser->token[i+1].word))
	  {
	    like.Setresolvetype("conservative");
	    i++;
	  }
	  else
	    Bp3Error("Gcxinput::rebukeyword","The string after BUILd must be QUICk SUPErquick THORough AGGRessive or CONServative");
 	else
	  Bp3Error("Gcxinput::rebukeyword","A string is expected after BUILd");
      else
	Bp3Error("Gcxinput::rebukeyword","parsing error");
    else
      Bp3Error("Gcxinput::rebukeyword","Subkeyword of REBUild not recognized");
}

void Gcxinput::warpkeyword()
{
  unsigned number(0);
  
  if (!parser->token[1].isnumber)	
    Bp3Error("Gcxinput::warpkeyword", "A positive integer value is expected for step");
  else
    number    = (unsigned)(parser->token[1].value);

  if (number  < 1)
    Bp3Error("Gcxinput::warpkeyword", "Step number must be a positive integer");
    
  
  if (like.Getstepsize()     ==  number - 1)
    like.Setstep("ARPWARP");
  else if (like.Getstepsize() > number - 1)
    if (like.Getstep(number-1) != "ARPWARP")
      Bp3Error("Gcxinput::warpkeyword", "ARPWARP has not been set to this step");

  for (int i                 = 2; i < parser->ntokens; i++)
  {
    if (CCP4::ccp4_keymatch("BIG",parser->token[i].word))
      if ( (i+1)             < parser->ntokens)
	if (parser->token[i+1].isnumber)
	{
	  like.Setarpbigcycles((unsigned)(parser->token[i+1].value));
	  i++;
	}
	else
	  Bp3Error("Gcxinput::warpkeyword","A positive number is expected for BIG");
      else
	Bp3Error("Gcxinput::warpkeyword","A positive number is expected for BIG");
    else if (CCP4::ccp4_keymatch("SMAL",parser->token[i].word))
      if ( (i+1)             < parser->ntokens)
	if (parser->token[i+1].isnumber)
	{
	  like.Setarpsmallcycles((unsigned)(parser->token[i+1].value));
	  i++;
	}
	else
	  Bp3Error("Gcxinput::warpkeyword","A positive number is expected for SMAL");
      else
	Bp3Error("Gcxinput::warpkeyword","A positive number is expected for SMAL");
    else if (CCP4::ccp4_keymatch("USES",parser->token[i].word))
      like.Setarpusesix(true);
    else if (CCP4::ccp4_keymatch("NOPD",parser->token[i].word))
      like.Setarpusepdb(false);
    else if (CCP4::ccp4_keymatch("DOCK",parser->token[i].word))
      if ( (i+1)             < parser->ntokens)
	if (parser->token[i+1].isnumber)
	{
	  like.Setarpsideafter((unsigned)(parser->token[i+1].value));
	  i++;
	}
	else
	  Bp3Error("Gcxinput::warpkeyword","A positive number is expected for DOCK");
      else
	Bp3Error("Gcxinput::warpkeyword","A positive number is expected for DOCK");
    else if (CCP4::ccp4_keymatch("CYCL",parser->token[i].word))
      if ( (i+1)             < parser->ntokens)
	if (parser->token[i+1].isnumber)
	{
	  like.Setarpcycles((unsigned)(parser->token[i+1].value));
	  i++;
	}
	else
	  Bp3Error("Gcxinput::warpkeyword","A positive number is expected for CYCLe");
      else
	Bp3Error("Gcxinput::warpkeyword","A positive number is expected for CYCLe");
    else if (CCP4::ccp4_keymatch("TARG",parser->token[i].word))
      if ( (i+1)             < parser->ntokens)
	if (parser->token[i+1].isstring)
	{
	  if (CCP4::ccp4_keymatch("DIRE",parser->token[i+1].word) ||
	      CCP4::ccp4_keymatch("SAD",parser->token[i+1].word))
	    like.Setrefmactarget("DIRE");
	  else if (CCP4::ccp4_keymatch("MLHL",parser->token[i+1].word))
	    like.Setrefmactarget("MLHL");
	  else if (CCP4::ccp4_keymatch("RICE",parser->token[i+1].word))
 	    like.Setrefmactarget("RICE");
	  i++;
 	}    
	else
	  Bp3Error("Gcxinput::warpkeyword","A string is expected for TARG");
      else
	Bp3Error("Gcxinput::warpkeyword","A string is expected for TARG");
    else
      Bp3Error("Gcxinput::warpkeyword","Subkeyword of WARP not recognized");
  }
}

void Gcxinput::shlckeyword()
{
  unsigned number(0);
  
  if (!parser->token[1].isnumber)	
    Bp3Error("Gcxinput::shlckeyword", "A positive integer value is expected for step");
  else
    number    = (unsigned)(parser->token[1].value);

  if (number  < 1)
    Bp3Error("Gcxinput::shlckeyword", "Step number must be a positive integer");
    
  
  if (like.Getstepsize()     ==  number - 1)
    like.Setstep("SHELXC");
  else if (like.Getstepsize() > number - 1)
    if (like.Getstep(number-1) != "SHELXC")
      Bp3Error("Gcxinput::shlckeyword", "SHELXC has not been set to this step");

  for (int i                 = 2; i < parser->ntokens; i++)
    if (CCP4::ccp4_keymatch("HIRE",parser->token[i].word))
      if ( (i+1)             < parser->ntokens)
	if (parser->token[i+1].isnumber)
	{
	  like.Setsubhirescut((double)(parser->token[i+1].value));
	  like.Setautoreso(false);
	  i++;
	}
	else
	  Bp3Error("Gcxinput::shlckeyword","A positive number is expected for HIREs");
      else
	Bp3Error("Gcxinput::shlckeyword","A positive number is expected for HIREs");
    else
      Bp3Error("Gcxinput::shlckeyword","Subkeyword of SHLC not recognized");
}

void Gcxinput::shldkeyword()
{
  unsigned number(0);
  
  if (!parser->token[1].isnumber)	
    Bp3Error("Gcxinput::shldkeyword", "A positive integer value is expected for step");
  else
    number    = (unsigned)(parser->token[1].value);

  if (number  < 1)
    Bp3Error("Gcxinput::shldkeyword", "Step number must be a positive integer");
    
  
  if (like.Getstepsize()     ==  number - 1)
    like.Setstep("SHELXD");
  else if (like.Getstepsize() > number - 1)
    if (like.Getstep(number-1) != "SHELXD")
      Bp3Error("Gcxinput::shldkeyword", "SHELXD has not been set to this step");

  for (int i                 = 2; i < parser->ntokens; i++)
  {
    if (CCP4::ccp4_keymatch("TRIA",parser->token[i].word))
      if ( (i+1)             < parser->ntokens)
	if (parser->token[i+1].isnumber)
	{
	  like.Setshelxdtrials((unsigned)(parser->token[i+1].value));
	  i++;
	}
	else
	  Bp3Error("Gcxinput::shldkeyword","A positive number is expected for TRIAls");
      else
	Bp3Error("Gcxinput::shldkeyword","A positive number is expected for TRIAls");
    else if (CCP4::ccp4_keymatch("THRE",parser->token[i].word))
      if ( (i+1)             < parser->ntokens)
	if (parser->token[i+1].isnumber)
	{
	  like.Setshelxdthres((double)(parser->token[i+1].value));
	  i++;
	}
	else
	  Bp3Error("Gcxinput::shldkeyword","A positive number is expected for THREsh");
      else
	Bp3Error("Gcxinput::shldkeyword","A positive number is expected for THREsh");
    else if (CCP4::ccp4_keymatch("MIND",parser->token[i].word))
      if ( (i+1)             < parser->ntokens)
	if (parser->token[i+1].isnumber)
	{
	  like.Setshelxdmind((double)(parser->token[i+1].value));
	  i++;
	}
	else
	  Bp3Error("Gcxinput::shldkeyword","A positive number is expected for MIND");
      else
	Bp3Error("Gcxinput::shldkeyword","A positive number is expected for MIND");
    else if (CCP4::ccp4_keymatch("MDEQ",parser->token[i].word))
      if ( (i+1)             < parser->ntokens)
	if (parser->token[i+1].isnumber)
	{
	  like.Setshelxdmdeq((double)(parser->token[i+1].value));
	  i++;
	}
	else
	  Bp3Error("Gcxinput::shldkeyword","A number is expected for MDEQ");
      else
	Bp3Error("Gcxinput::shldkeyword","A number is expected for MDEQ");

    else if (CCP4::ccp4_keymatch("DSUL",parser->token[i].word))
      if ( (i+1)             < parser->ntokens)
	if (parser->token[i+1].isnumber)
	{
	  like.Setshelxddsul((unsigned)(parser->token[i+1].value));
	  i++;
	}
	else
	  Bp3Error("Gcxinput::shldkeyword","A positive number is expected for DSUL");
      else
	Bp3Error("Gcxinput::shldkeyword","A positive number is expected for DSUL");
    else
      Bp3Error("Gcxinput::shldkeyword","Subkeyword of SHLD not recognized");
  }
}

void Gcxinput::shlekeyword()
{
  unsigned number(0);
  
  if (!parser->token[1].isnumber)	
    Bp3Error("Gcxinput::shlekeyword", "A positive integer value is expected for step");
  else
    number    = (unsigned)(parser->token[1].value);

  if (number  < 1)
    Bp3Error("Gcxinput::shlekeyword", "Step number must be a positive integer");
    
  
  if (like.Getstepsize()     ==  number - 1)
    like.Setstep("SHELXE");
  else if (like.Getstepsize() > number - 1)
    if (like.Getstep(number-1) != "SHELXE")
      Bp3Error("Gcxinput::shlekeyword", "SHELXE has not been set to this step");

  for (int i                 = 2; i < parser->ntokens; i++)
      if (CCP4::ccp4_keymatch("NOHA",parser->token[i].word))
	like.Setshelxehand(false);
      else if (CCP4::ccp4_keymatch("OPTI",parser->token[i].word))
	like.Setshelxeopt(true);
      else if (CCP4::ccp4_keymatch("NODM",parser->token[i].word))
	like.Setshelxedenmod(false);
      else if (CCP4::ccp4_keymatch("FREE",parser->token[i].word))
      {
	like.Setfreelunch(true);
	if ( (i+1)             < parser->ntokens)
	  if (parser->token[i+1].isnumber)
	  {
	    like.Setfreelunchreso((double)(parser->token[i+1].value));
	    i++;
	  }
      }
      else if (CCP4::ccp4_keymatch("NCYC",parser->token[i].word))
      {
	like.Setshelxedenmod(true);
	if ( (i+1)             < parser->ntokens)
	  if (parser->token[i+1].isnumber)
	  {
	    like.Setshelxencycles((unsigned)(parser->token[i+1].value));
	    i++;
	  }
	  else
	    Bp3Error("Gcxinput::shlekeyword", "A positive integer is expected after NCYCles");
	else
	  Bp3Error("Gcxinput::shlekeyword", "A positive integer is expected after NCYCles");
      }
      else if (CCP4::ccp4_keymatch("HCYC",parser->token[i].word))
      {
	like.Setshelxehand(true);
	if ( (i+1)             < parser->ntokens)
	  if (parser->token[i+1].isnumber)
	  {
	    like.Setshelxehcycles((unsigned)(parser->token[i+1].value));
	    i++;
	  }
	  else
	    Bp3Error("Gcxinput::shlekeyword", "A positive integer is expected after HCYCles");
	else
	  Bp3Error("Gcxinput::shlekeyword", "A positive integer is expected after HCYCles");
      }
      else if (CCP4::ccp4_keymatch("OCYC",parser->token[i].word))
      {
	like.Setshelxeopt(true);
	if ( (i+1)             < parser->ntokens)
	  if (parser->token[i+1].isnumber)
	  {
	    like.Setshelxeocycles((unsigned)(parser->token[i+1].value));
	    i++;
	  }
	  else
	    Bp3Error("Gcxinput::shlekeyword", "A positive integer is expected after OCYCles");
	else
	  Bp3Error("Gcxinput::shlekeyword", "A positive integer is expected after OCYCles");
      }
      else
	Bp3Error("Gcxinput::shlekeyword","Subkeyword of SHLE not recognized");
}

void Gcxinput::checkeyword()
{
  like.Setcheck(true);

  for (int i                 = 1; i < parser->ntokens; i++)
  {
    if (CCP4::ccp4_keymatch("PDB",parser->token[i].word))
      if ( (i+1)             < parser->ntokens)
	if (parser->token[i+1].isstring)
	{
	  string filename = parser->token[i+1].fullstring;
	  like.Setfinalpdb(filename);
	  i++;
	}
	else
	  Bp3Error("Gcxinput::checkeyword","A PDB file name is expected");
      else
	Bp3Error("Gcxinput::checkeyword","A PDB file name is expected");
    else if (CCP4::ccp4_keymatch("NOTR",parser->token[i].word))
      like.Setnotrans(true);
    else if (CCP4::ccp4_keymatch("SUBS",parser->token[i].word))
	if (parser->token[i+1].isstring)
	{
	  string filename = parser->token[i+1].fullstring;
	  like.Setfinalsubstructure(filename);
	  i++;
	}
	else
	  Bp3Error("Gcxinput::checkeyword","A (substructure) PDB file name is expected");
    else
      Bp3Error("Gcxinput::checkeyword","Subkeyword of CHECk not recognized");
  }
}
