/*
  Copyright (c) Navraj Pannu       1998 - 2007
  Copyright (c) Leiden University  2002 - 2007

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

#include <fstream>
#include <math.h>
#include <string>
#include "gcxlikelihood.h"

#if defined _MVS || _MSC_VER 
const char dir('\\');
#else
const char dir('/');
#endif

Gcxlikelihood::Gcxlikelihood(Model &model, Crystal &crystal)
  : Likelihood(model, crystal)
{
  output         = "crank"; overallbfactor = weight              = ZERO;
  refd           = refc                    = nnucleotides        = 0;
  sadonly        = false;
  bp3refd        = 0;
  nmonomers      = nresidues               = 0; solvent          = ZERO;
  firststep      = "DETECT"; laststep      = "BUILD";
  autoreso       = true;    subhires       = 1000.0; subhirescut = 1000.0;
  refltype       = "MTZ";   version        = "121";
  // scaling
  rescalef       = 1;
  
  // fa values
  afrosigf       = TWO;     afrosiso       = afrosano            = -0.1;

  // substructure detection
  crunchspec     = false;   shelxeinphase  = false;
  crunchbp3      = 3;
  crunchtrials   = 20;      crunchptrials  = 150; shelxdtrials   = 500;
  crunchthres    = 1.000;   crunchdev      = 1.75; shelxdthres   = 30.0;
  shelxdmind     = 3.5;     shelxdmdeq     = -0.1; shelxddsul    = 0;
  shelxdmtrials  = 10;

  // phasing
  bp3stopfom     = 0.075;   bp3hand        = true;
  refmachand     = true;    refmacsolvent  = false;

  // density modification
  solomonhand    = solomondenmod       = true;
  solomonopt     = false;
  solomonnomtz   = solomonnomap            = true;
  shelxehand     = shelxedenmod            = true;
  shelxeopt      = false;
  dmhand         = dmopt                   = dmdenmod            = dmhist  = true;
  solomonocycles = 5; solomonhcycles       = 2;
  solomonncycles = 10;
  shelxeocycles  = 10; shelxehcycles       = 5;
  shelxencycles  = 100;
  piratehand     = pirateopt               = false;
  piratedenmod   = true; piratencs         = false;
  piratehl       = true;
  piratewihl     = 1.0; piratencycles      = 3;
  resolvedmncs   = resolvehl               = resolvembncs        = true;
  dmhandmargin   = 0.005;
  solomonhandmargin                        = 0.0025;
  freelunch      = false; freelunchreso    = ZERO;

  // model building
  arpusesix      = false;
  arpcycles      = 3;       arpsmallcycles = 5; arpbigcycles     = 10;
  arpsideafter   = 7;
  freertag       = "Unassigned";
  resolvetype    = "quick";
  buccusepdb     = arpusepdb               = true;
  buccreli       = 0.95;    buccfragments  = 20; buccnofreer     = false;
  bucccycles     = 3;       buccfast       = false;

  // model comparison
  check          = false;
  notrans        = false;
}

void Gcxlikelihood::setup()
{
  // determine highest-resolution data set and set that to reference
  double hires(1000.0);
  
  for (unsigned d             = 0; d < xtal.sf.size(); d++)
  {
    if (xtal.sf[d].hires      < hires)
    {
      hires                   = xtal.sf[d].hires;
      refc                    = xtal.sf[d].nxtal;
      refd                    = xtal.sf[d].nwave;
      bp3refd                 = d;
    }
    if (xtal.sf[d].hires      < subhires)
      for (unsigned a         = 0; a < mdl.atom.size(); a++)
	if (xtal.sf[d].nxtal == mdl.atom[a].crystal)
 	  subhires            = xtal.sf[d].hires;    
  }
  
  freelunchreso               = hires - 0.5;
  
  double subtemp(std::max(subhires,
			  std::min(subhires+HALF,
				   THREE)));

  if (!autoreso               && (subtemp <  999.0) )
    printf("Calculated substructure detection resolution cut off %.2f\n\n",subtemp);
    

  printf("Reference crystal %u\n\n",refc+1);
  printf("Reference dataset %u\n",refd+1);

  if (!autoreso              && (subhirescut > 999.0) )
    subhirescut               = subtemp;
  
  printf("\n");

  /*
  if (!(xtal.sf.size() == 1 && !xtal.sf[0].anomalous))
    xtal.Setselected("PHASE");
  */
  
  xtal.Setnormalization();
  xtal.SetLuzzati();

  if (!nooutput)
    xtal.print();

  if (getenv("SEQIN")        != NULL)
  {
    string temp(getenv("SEQIN"));
    if ( (temp[0]            != dir) && (temp[1] !=':'))
    {
      string t2(getenv("PWD"));
      temp                    =  t2 + dir + temp;
    }
    parseseqin(temp.c_str());
  }
  
  if ( (nresidues            == 0) && (nnucleotides == 0) )
    Bp3Error("Gcxlikelihood::setup","Either input a sequence, or set NRES and/or NNUC");

  xtal.Calculatematthews(nnucleotides, nresidues, weight, &nmonomers, &solvent);

  if (overallbfactor         == ZERO)
    overallbfactor            = xtal.Calculatebfactor(nnucleotides, nresidues, nmonomers, bp3refd);

  if (xtal.rscale)
  {
    printf("Writing a rescaled MTZ file\n\n");
    writemtzout();
  }
  
  if (!nooutput)
  {
    printf("Writing the Crank XML file\n\n");
    writexml();
  }
}

Gcxlikelihood &Gcxlikelihood::Setpipeline(const unsigned input)
{

  if ( (firststep    == "PHASE") && (laststep == "DETECT"))
    Bp3Error("Gcxlikelihood::Setpipeline","Cannot have last step as DETECT with first step PHASE");
    
  if (input          == 1)
  {
    // default pipeline
    bp3hand           = 1;
   
    step.push_back("PREP");
    if (firststep    == "DETECT")
    {  
      step.push_back("AFRO");
      step.push_back("CRUNCH2");
    }
    
    if (laststep     != "DETECT")
    {
      step.push_back("BP3");
      if (laststep   != "HREF")
      {
	step.push_back("SOLOMON");
	if (laststep != "DM")
	  step.push_back("ARPWARP");
      }
    }
  } 	
  else if (input     == 2)
  {
    step.resize(0);
    // shelx/bp3/solomon pipeline
    bp3hand           = 0;
    solomonhand       = false;
    solomonopt        = false;
    shelxedenmod      = false;
    
    step.push_back("PREP");

    if (firststep    == "DETECT")
    {
      step.push_back("SHELXC");
      step.push_back("SHELXD");
      step.push_back("SHELXE");
    }
    else
      bp3hand         = solomonhand = true;
    
    if (laststep     != "DETECT")
    {
      step.push_back("BP3");
      if (laststep   != "HREF")
      {
	step.push_back("SOLOMON");
	if (laststep != "DM")
	  step.push_back("ARPWARP");
      }
    }
  }
  else if (input     == 3)
  {  
    step.resize(0);

    step.push_back("PREP");
    step.push_back("SHELXC");

    if (firststep    == "DETECT")
      step.push_back("SHELXD");

    if (laststep     != "DETECT")
    {
      step.push_back("SHELXE");
      if ( (laststep != "HREF") && (laststep != "DM") )
	step.push_back("ARPWARP");
    }
  }
  else if (input     == 4)
  {
    // crunch2/bp3/dm 
    step.resize(0);
    bp3hand           = 1;
   
    step.push_back("PREP");
    if (firststep    == "DETECT")
    {
      step.push_back("AFRO");
      step.push_back("CRUNCH2");
    }
    if (laststep     != "DETECT")
    {
      step.push_back("BP3");
      if (laststep   != "HREF")
      {
	step.push_back("DM");
 	if (laststep != "DM")
	  step.push_back("ARPWARP");
      }
    }
  }
  else if (input     == 5)
  {
    // resolve pipeline
    bp3hand           = 1;
    solomondenmod     = 0;
   
    step.push_back("PREP");
    if (firststep    == "DETECT")
    {  
      step.push_back("AFRO");
      step.push_back("CRUNCH2");
    }
    
    if (laststep     != "DETECT")
    {
      step.push_back("BP3");
      if (laststep   != "HREF")
      {
	step.push_back("SOLOMON");
	step.push_back("RESOLVEDM");
	if (laststep != "DM")
	  step.push_back("RESOLVEBUILD");
      }
    }
  }
  else if (input     == 6)
  {
    // crunch2/bp3/solomon/buc
    bp3hand           = 1;
   
    step.push_back("PREP");
    if (firststep    == "DETECT")
    {
      step.push_back("AFRO");
      step.push_back("CRUNCH2");
    }
    if (laststep     != "DETECT")
    {
      step.push_back("BP3");
      if (laststep   != "HREF")
      {
	step.push_back("SOLOMON");
	if (laststep != "DM")
	  step.push_back("BUCCANEER");
      }
    }
  }
  else if (input     == 7)
  {
    step.resize(0);
    refmachand        = false;
    solomonhand       = solomonopt = false;
    shelxedenmod      = false;
    
    step.push_back("PREP");

    if (firststep    == "DETECT")
    {
      step.push_back("SHELXC");
      step.push_back("SHELXD");
      step.push_back("SHELXE");
    }
    else
      refmachand      = solomonhand = true;

    if (laststep     != "DETECT")
    {
      step.push_back("REFMAC");
      if (laststep   != "HREF")
      {
	step.push_back("SOLOMON");
	if (laststep != "DM")
	  step.push_back("ARPWARP");
      }
    }
  }
  else
    Bp3Error("Gcxlikelihood::Setpipeline","pipeline not known");
  
  return *this;
}

void Gcxlikelihood::writemtzout()
{
  CMtz::MTZ *MTZOUT            = CMtz::MtzMalloc(0,0);
  CMtz::MTZ *MTZIN             = CMtz::MtzGet("HKLIN",0);
  mode                         = "RESCALE"; allin = false;
  setupmtz(MTZOUT,MTZIN);

  unsigned inc                 = (allin) ? colin.size() : 3;

  unsigned columns(0);
  for (unsigned d              = 0; d < xtal.sf.size(); d++)
  {
    columns                   += 2;    
    if (xtal.sf[d].anomalous)
      columns                 += 4;
  }
  
  for (unsigned r              = 0; r < xtal.maxref; r++)
  {
    // Default value for columns to be written out is MNF
    vector<float> fdata(columns+inc, CCP4::ccp4_nan().f);

    storeinitialcolumns(MTZIN, fdata, inc, r);

    unsigned ncol(inc-1);

    for (unsigned d            = 0; d < xtal.sf.size(); d++)
    {
      if (xtal.sf[d].use(r))
      {
	fdata[++ncol]          = xtal.sf[d].datamean(r);
	fdata[++ncol]          = xtal.sf[d].devmean(r);
      }
      else
	ncol                  += 2;

      if (xtal.sf[d].anomalous)
      {
	if (xtal.sf[d].usep(r))
	{
	  fdata[++ncol]        = xtal.sf[d].datap[r];
	  fdata[++ncol]        = xtal.sf[d].devp[r];
	}
	else
	  ncol                += 2;

	if (xtal.sf[d].usem(r) && !xtal.centric[r])
	{
	  fdata[++ncol]        = xtal.sf[d].datam[r];
	  fdata[++ncol]        = xtal.sf[d].devm[r];
	}
	else
	  ncol                += 2;
      }
    }

    CMtz::ccp4_lwrefl(MTZOUT, &fdata[0], &colout[0], fdata.size(), r+1); 
  }

  if (allin)
    CMtz::MtzFree(MTZIN);
  CMtz::MtzPut(MTZOUT, " ");
  CMtz::MtzFree(MTZOUT);  
}
  

void Gcxlikelihood::writexml()
{
  string xmlname      = output + ".xml";
  xmlfile             = fopen(xmlname.c_str(),"w");

  fprintf(xmlfile, "<?xml version=\"1.0\" standalone=\"yes\"\?>\n");
  fprintf(xmlfile, "<crank>\n");

  for (unsigned i     = 0; i < step.size(); i++)
    if (step[i]      == "SHELXC")
      checkshelx();

  writeparameters();
  writeinputfile();
  writeoutputfile();
    
  fprintf(xmlfile, "<soap>\n");
  fprintf(xmlfile, "<run>\n");

  string directory(getenv("PWD"));
  directory           = directory + dir + output;
    
  fprintf(xmlfile, "<directory>%s</directory>\n\n",directory.c_str());

  if (step.size()    == 0)
    Setpipeline(1);
    
  for (unsigned i     = 0; i < step.size(); i++)
  {
    unsigned inc(i+1);
    fprintf(xmlfile, "<step id=\"%u\">", inc);
    if (step[i]      == "PREP")
      writeprep(inc);
    else if (step[i] == "AFRO")
      writeafro(inc);
    else if (step[i] == "CRUNCH2")
      writecrunch2(inc);
    else if (step[i] == "BP3")
      writebp3(inc);
    else if (step[i] == "REFMAC")
      writerefmac(inc);
    else if (step[i] == "SOLOMON")
      writesolomon(inc);
    else if (step[i] == "DM")
      writedm(inc);
    else if (step[i] == "PIRATE")
      writepirate(inc);
    else if (step[i] == "BUCCANEER")
      writebuccaneer(inc);
    else if (step[i] == "RESOLVEDM")
      writeresolvedm(inc);
    else if (step[i] == "RESOLVEBUILD")
      writeresolvebuild(inc);
    else if (step[i] == "ARPWARP")
      writearpwarp(inc);
    else if (step[i] == "SHELXC")
      writeshelxc(inc);
    else if (step[i] == "SHELXD")
      writeshelxd(inc);
    else if (step[i] == "SHELXE")
      writeshelxe(inc);
    else
      Bp3Error("gcxlikelihood::writexml","program name not known");
     
    fprintf(xmlfile, "</step>\n\n");     
  }

  if (check)
  {
    unsigned inc(step.size()+1);
    fprintf(xmlfile, "<step id=\"%u\">", inc);
    writecheck(inc);
    fprintf(xmlfile, "</step>\n\n");     
  }
  fprintf(xmlfile, "</run>\n");
  fprintf(xmlfile, "</soap>\n");

  fprintf(xmlfile, "</crank>\n");

  fclose(xmlfile);
}

void Gcxlikelihood::parseseqin(const char *seqin)
{
  std::ifstream pirfile(seqin, std::ios::in);

  if (!pirfile)
    Bp3Error("Gcxlikelihood::parseseqin","could not open SEQIN file");

  char ch;
  pirfile.get(ch);
  if (ch            != '>')
    Bp3Error("Gcxlikelihood::parseseqin","SEQIN does not appear to be a pir file (does not start with >)");

  unsigned linebreak(0);
  unsigned numres(0);
  double mass(ZERO);
  while (pirfile.get(ch))
  {
    if (ch          == '\n')
      linebreak++;

    if (linebreak   >= 2)
      if ( (ch      == 'A') || (ch == 'a') ) 
      {
	numres++;
	sequence.push_back('A');
	mass        += 89.09;
      }
      else if ( (ch == 'C') || (ch == 'c') )
      {
	numres++;
	sequence.push_back('C');
	mass        += 121.16;
      }
      else if ( (ch == 'D') || (ch == 'd') ) 
      {
	numres++;
	sequence.push_back('D');
	mass        += 133.1;
      }
      else if ( (ch == 'E') || (ch == 'e') )
      {
	numres++;
	sequence.push_back('E');
	mass        += 147.13;
      }
      else if ( (ch == 'F') || (ch == 'f') )
      {
	numres++;
	sequence.push_back('F');
	mass        += 165.19;
      }
      else if ( (ch == 'G') || (ch == 'g') )
      {
	numres++;
	sequence.push_back('G');
	mass        += 75.07;
      }
      else if ( (ch == 'H') || (ch == 'h') )
      {
	numres++;
	sequence.push_back('H');	
	mass        += 155.16;
      }
      else if ( (ch == 'I') || (ch == 'i') )
      {
	numres++;
	sequence.push_back('I');
	mass        += 131.18;
      }
      else if ( (ch == 'K') || (ch == 'k') )
      {
	numres++;
	sequence.push_back('K');
	mass        += 146.19;
      }
      else if ( (ch == 'L') || (ch == 'l') )
      {
	numres++;
	sequence.push_back('L');
	mass        += 131.18;
      }
      else if ( (ch == 'M') || (ch == 'm') )
      {
	numres++;
	sequence.push_back('M');
	mass        += 149.21;
      }
      else if ( (ch == 'N') || (ch == 'n') )
      {
	numres++;
	sequence.push_back('N');
	mass        += 132.12;
      }
      else if ( (ch == 'P') || (ch == 'p') )
      {
	numres++;
	sequence.push_back('P');
	mass        += 115.13;
      }
      else if ( (ch == 'Q') || (ch == 'q') )
      {
	numres++;
	sequence.push_back('Q');
	mass        += 146.15;
      }
      else if ( (ch == 'R') || (ch == 'r') )
      {
	numres++;
	sequence.push_back('R');
	mass        += 174.20;
      }
      else if ( (ch == 'S') || (ch == 's') )
      {
	numres++;
	sequence.push_back('S');
	mass        += 105.09;
      }
      else if ( (ch == 'T') || (ch == 't') )
      {
	numres++;
	sequence.push_back('T');
	mass        += 119.12;
      }
      else if ( (ch == 'V') || (ch == 'v') )
      {
	numres++;
	sequence.push_back('V');
	mass        += 117.15;
      }
      else if ( (ch == 'W') || (ch == 'w') )
      {
	numres++;
	sequence.push_back('W');
	mass        += 204.23;
      }
      else if ( (ch == 'Y') || (ch == 'y') )
      {
	numres++;
	sequence.push_back('Y');
	mass        += 181.19;
      }
      else if ( (ch == 'Z') || (ch == 'z') )  // seleno-met
      {
	numres++;
	sequence.push_back('Z');
	mass        += 185.21;
      }
      else if ( (ch == ' ') || (ch == '*') || (ch == '-') )
	continue;
      else if (ch   == '\n')
	linebreak++;
      else
      {
	Bp3Warning("Gcxlikelihood::parseqin","Unknown character/Amino Acid in SEQIN file - assuming glycine");
	sequence.push_back('G');
	numres++;
	mass        += 75.07;
      }
  }

  printf("The number of residues determined from SEQIN is %u\n\n",numres);
  printf("The molecular weight determined from SEQIN is %.2f\n\n",mass);

  if (weight        == ZERO)
    weight           = mass;
  
  if (nresidues     == 0)
    nresidues        = numres;
  else if (numres   != nresidues)
    Bp3Error("Gcxlikelihood::parseqin","Number of residues inputted with NRES does not equal number determined from SEQIN");
  
  pirfile.close();
}

  
void Gcxlikelihood::writeparameters() 
{
  fprintf(xmlfile, "<parameters>\n");
  fprintf(xmlfile, "<version>%s</version>\n",version.c_str());
  fprintf(xmlfile, "<code>crank</code>\n");
  if (diskspace.size())
    fprintf(xmlfile, "<disk_space>%s</disk_space>\n",diskspace.c_str());
  fprintf(xmlfile, "<solvent_content>%.3f</solvent_content>\n",solvent);
  fprintf(xmlfile, "<residues>%u</residues>\n",nresidues);
  fprintf(xmlfile, "<nucleotides>%u</nucleotides>\n",nnucleotides);
  fprintf(xmlfile, "<nmonomers>%u</nmonomers>\n",nmonomers);
  fprintf(xmlfile, "<bfactor>%.2f</bfactor>\n",overallbfactor);
  
  fprintf(xmlfile, "<ref_crystal>%u</ref_crystal>\n",refc+1);
  fprintf(xmlfile, "<ref_dataset>%u</ref_dataset>\n",refd+1);

  fprintf(xmlfile, "<input_spacegroup>\n");
  fprintf(xmlfile, "<name>%s</name>\n",xtal.sg.hmname.c_str());
  fprintf(xmlfile, "<number>%u</number>\n",xtal.sg.number);
  fprintf(xmlfile, "</input_spacegroup>\n");

  if (sequence.size())
  {
    fprintf(xmlfile, "<sequence>\n");
    unsigned i(0);
    while (i                < sequence.size())
      if ((i + 60)          < sequence.size())
      {
	fprintf(xmlfile, "%s\n",sequence.substr(i,60).c_str());
	i                  += 60;
      }
      else
      {
 	fprintf(xmlfile, "%s\n",sequence.substr(i,sequence.size()-i).c_str());
	i                   = sequence.size();
      }
    fprintf(xmlfile, "</sequence>\n");
  }
  unsigned inputintensity((xtal.sf[0].type == "INTENSITY") ? 1 : 0);
  fprintf(xmlfile, "<input_intensity>%u</input_intensity>\n",inputintensity); 

  for (unsigned c           = 0; c < xtal.cell.size(); c++)
  {
    int w(-1);
    fprintf(xmlfile, "<crystal id=\"%u\">\n",c+1);

    bool substructure(false);
    unsigned sites(0);
    for (unsigned a         = 0; a < mdl.atom.size(); a++)
      if (c                == mdl.atom[a].crystal)
      {
	substructure        = true;
	if (mdl.site.size() > 0)
	{
	  sites++;
	  subtag            = "IN";
	}
      }
    
    if (substructure)
    {
      fprintf(xmlfile, "<native>0</native>\n");
      fprintf(xmlfile, "<substructure>\n");
      fprintf(xmlfile, "<input>%u</input>\n",(sites > 0));
      fprintf(xmlfile, "<atom_name>%s</atom_name>\n", mdl.atom[0].name.c_str());
      if (sites            == 0)
	fprintf(xmlfile, "<natoms>%u</natoms>\n", (unsigned)(mdl.atom[0].Getnumber()*nmonomers));
      else
 	fprintf(xmlfile, "<natoms>%u</natoms>\n", (unsigned)mdl.atom.size());
	
      if ((mdl.site.size()  > 0) && (xtal.pdbfilename[c] == "") )
      {
	fprintf(xmlfile, "<coordinate_format>Fractional</coordinate_format>\n");
	
	for (unsigned a     = 0; a < mdl.atom.size(); a++)
	  if (c            == mdl.atom[a].crystal)
	  {
	    fprintf(xmlfile, "<atom id=\"%u\">\n", a+1);
	    fprintf(xmlfile, "<name>%s</name>\n", mdl.atom[a].name.c_str());
	    fprintf(xmlfile, "<x>%.3f</x>\n", mdl.site[mdl.atom[a].Getnsite()].Getx(0));
	    fprintf(xmlfile, "<y>%.3f</y>\n", mdl.site[mdl.atom[a].Getnsite()].Getx(1));
	    fprintf(xmlfile, "<z>%.3f</z>\n", mdl.site[mdl.atom[a].Getnsite()].Getx(2));
	    fprintf(xmlfile, "<noref_x>%u</noref_x>\n", !mdl.site[mdl.atom[a].Getnsite()].Getrefinex(0));
	    fprintf(xmlfile, "<noref_y>%u</noref_y>\n", !mdl.site[mdl.atom[a].Getnsite()].Getrefinex(1));
	    fprintf(xmlfile, "<noref_z>%u</noref_z>\n", !mdl.site[mdl.atom[a].Getnsite()].Getrefinex(2));
	    fprintf(xmlfile, "<occ>%.3f</occ>\n", mdl.atom[a].Getocc());
	    fprintf(xmlfile, "<bfac>%.3f</bfac>\n", mdl.atom[a].Getbiso());
	    fprintf(xmlfile, "<noref_occ>%u</noref_occ>\n", !mdl.atom[a].Getrefineocc());
	    fprintf(xmlfile, "<noref_bfac>%u</noref_bfac>\n", !mdl.atom[a].Getrefinebfac());
	    fprintf(xmlfile, "</atom>\n");
	  } 
      }
      fprintf(xmlfile, "</substructure>\n");
    }
    else
      fprintf(xmlfile, "<native>1</native>\n");

    fprintf(xmlfile, "<cell>\n");
    fprintf(xmlfile, "<cell_a>%.3f</cell_a>\n",xtal.cell[c].a);
    fprintf(xmlfile, "<cell_b>%.3f</cell_b>\n",xtal.cell[c].b);
    fprintf(xmlfile, "<cell_c>%.3f</cell_c>\n",xtal.cell[c].c);
    fprintf(xmlfile, "<cell_alpha>%.3f</cell_alpha>\n",xtal.cell[c].alpha);
    fprintf(xmlfile, "<cell_beta>%.3f</cell_beta>\n",xtal.cell[c].beta);
    fprintf(xmlfile, "<cell_gamma>%.3f</cell_gamma>\n",xtal.cell[c].gamma);
    fprintf(xmlfile, "</cell>\n");

    for (unsigned d         = 0; d < xtal.sf.size(); d++)
      if (c                == xtal.sf[d].nxtal)
      {
	w++;
	fprintf(xmlfile, "<data id=\"%u\">\n",w+1);
	fprintf(xmlfile, "<type>%s</type>\n",xtal.sf[d].name.c_str());
	fprintf(xmlfile, "<anomalous>%u</anomalous>\n", (xtal.sf[d].anomalous ? 1 : 0 ));
	for (unsigned f     = 0; f < mdl.form.size(); f++)
	  if (d            == xtal.dataset[mdl.form[f].Getcrystal()][w])
	  {
	    fprintf(xmlfile, "<fprime>%.3f</fprime>\n", mdl.form[f].Getfp(w));
	    fprintf(xmlfile, "<fprimeprime>%.3f</fprimeprime>\n", mdl.form[f].Getfpp(w));
	    fprintf(xmlfile, "<bfactor>%.2f</bfactor>\n", overallbfactor);
	  }
	fprintf(xmlfile, "</data>\n");
      }
    fprintf(xmlfile, "</crystal>\n"); 
  }

  fprintf(xmlfile, "</parameters>\n\n");
}

void Gcxlikelihood::writeinputfile() 
{
  fprintf(xmlfile, "<input_files>\n");
  fprintf(xmlfile, "<file id=\"1\">\n");
  fprintf(xmlfile, "<type>%s</type>\n",refltype.c_str());
  fprintf(xmlfile, "<label>IN</label>\n");
  if (getenv("HKLIN") != NULL)
  {
    string temp(getenv("HKLIN"));
    if (!temp.size())
      Bp3Error("Gcxlikelihood::writeinputfile","No reflection file given");
      
    if (temp[0]               != dir)
    {
      string t2(getenv("PWD"));
      temp                     =  t2 + dir + temp;
    }
    fprintf(xmlfile, "<name>%s</name>\n",temp.c_str());
  }

  for (unsigned c   = 0; c < xtal.cell.size(); c++)
  {
    int w           = -1;
    fprintf(xmlfile, "<crystal id=\"%u\">\n",c+1);
    for (unsigned d = 0; d < xtal.sf.size(); d++)
      if (c        == xtal.sf[d].nxtal)
      {
	w++;
        fprintf(xmlfile, "<data id=\"%u\">\n",w+1);
	if (getenv("HKLIN") == NULL)
	  if (xtal.sf[d].mtzin.size())
	    fprintf(xmlfile, "<name>%s</name>\n",xtal.sf[d].mtzin.c_str());
	  else if (xtal.sf[d].scain.size())
 	    fprintf(xmlfile, "<name>%s</name>\n",xtal.sf[d].scain.c_str());
	  else
	    Bp3Error("Gcxlikelihood::writeinputfile","No reflection file given");
        if (xtal.sf[d].anomalous)
        {
	  if (xtal.sf[d].type == "INTENSITY")
	  {
	    fprintf(xmlfile, "<i_plus>%s</i_plus>\n",xtal.sf[d].sdatap.c_str());
	    fprintf(xmlfile, "<sigi_plus>%s</sigi_plus>\n",xtal.sf[d].sdevp.c_str());
	    fprintf(xmlfile, "<i_minus>%s</i_minus>\n",xtal.sf[d].sdatam.c_str() );
	    fprintf(xmlfile, "<sigi_minus>%s</sigi_minus>\n",xtal.sf[d].sdevm.c_str());
	  }
	  else
	  {
	    fprintf(xmlfile, "<f_plus>%s</f_plus>\n",xtal.sf[d].sdatap.c_str());
	    fprintf(xmlfile, "<sigf_plus>%s</sigf_plus>\n",xtal.sf[d].sdevp.c_str());
	    fprintf(xmlfile, "<f_minus>%s</f_minus>\n",xtal.sf[d].sdatam.c_str() );
	    fprintf(xmlfile, "<sigf_minus>%s</sigf_minus>\n",xtal.sf[d].sdevm.c_str());
	  }  
        }
        else
        {
	  if (xtal.sf[d].type == "INTENSITY")
	  {
	    fprintf(xmlfile, "<i>%s</i>\n", xtal.sf[d].sdatap.c_str());
	    fprintf(xmlfile, "<sigi>%s</sigi>\n",xtal.sf[d].sdevp.c_str());
	  }
	  else
	  {
	    fprintf(xmlfile, "<f>%s</f>\n", xtal.sf[d].sdatap.c_str());
	    fprintf(xmlfile, "<sigf>%s</sigf>\n",xtal.sf[d].sdevp.c_str());
	  }
        }
        fprintf(xmlfile, "</data>\n");
      }
    fprintf(xmlfile, "</crystal>\n");
  }
  fprintf(xmlfile, "</file>\n");
  unsigned j(1);
  for (unsigned c            = 0; c < xtal.pdbfilename.size(); c++)
    if (xtal.pdbfilename[c] != "")
    {
      string temp            = xtal.pdbfilename[c];
      if (temp[0]           != dir)
      {
	string t2(getenv("PWD"));
	temp                 =  t2 + dir + temp;
      }
      j++;
      fprintf(xmlfile, "<file id=\"%u\">\n",j);
      fprintf(xmlfile, "<type>SUBSTRUCTURE</type>\n");
      fprintf(xmlfile, "<label>IN</label>\n");
      fprintf(xmlfile, "<xtal>%u</xtal>\n",c+1);  
      fprintf(xmlfile, "<name>%s</name>\n",temp.c_str());      
      fprintf(xmlfile, "</file>\n");
    }
  
  fprintf(xmlfile, "</input_files>\n\n");
}

void Gcxlikelihood::writeoutputfile()
{
  fprintf(xmlfile, "<output_files>\n");
  fprintf(xmlfile, "<mtz_out>\n");
  fprintf(xmlfile, "<name>%s</name>\n",getenv("HKLOUT"));
  fprintf(xmlfile, "</mtz_out>\n");
  if (getenv("XYZOUT") != NULL)
  {
    fprintf(xmlfile, "<pdb_out>\n");
    fprintf(xmlfile, "<name>%s</name>\n",getenv("XYZOUT"));
    fprintf(xmlfile, "</pdb_out>\n");
  }
  
  fprintf(xmlfile, "</output_files>\n\n");
}

void Gcxlikelihood::writeprep(const unsigned step) 
{
  fprintf(xmlfile, "PREP\n");
  fprintf(xmlfile, "<name>PREP</name>\n");
  fprintf(xmlfile, "<tag>%u%s</tag>\n",step,"_PREP");
  fprintf(xmlfile, "<step>%u</step>\n",step);
  fprintf(xmlfile, "<input>\n");
  fprintf(xmlfile, "<exp_columns>\n");
  for (unsigned c   = 0; c < xtal.cell.size(); c++)
  {
    int w           = -1;
    fprintf(xmlfile, "<crystal id=\"%u\">\n",c+1);
    for (unsigned d = 0; d < xtal.sf.size(); d++)
      if (c        == xtal.sf[d].nxtal)
      {
	w++;
        fprintf(xmlfile, "<data id=\"%u\">\n",w+1);
        fprintf(xmlfile, "<type>%s</type>\n",xtal.sf[d].name.c_str());
        if (xtal.sf[d].anomalous)
        {
	  if (xtal.sf[d].type == "INTENSITY")
	  {
	    fprintf(xmlfile, "<i_plus>%s</i_plus>\n",xtal.sf[d].sdatap.c_str());
	    fprintf(xmlfile, "<sigi_plus>%s</sigi_plus>\n",xtal.sf[d].sdevp.c_str());
	    fprintf(xmlfile, "<i_minus>%s</i_minus>\n",xtal.sf[d].sdatam.c_str() );
	    fprintf(xmlfile, "<sigi_minus>%s</sigi_minus>\n",xtal.sf[d].sdevm.c_str());
	  }
	  else
	  {
	    fprintf(xmlfile, "<f_plus>%s</f_plus>\n",xtal.sf[d].sdatap.c_str());
	    fprintf(xmlfile, "<sigf_plus>%s</sigf_plus>\n",xtal.sf[d].sdevp.c_str());
	    fprintf(xmlfile, "<f_minus>%s</f_minus>\n",xtal.sf[d].sdatam.c_str() );
	    fprintf(xmlfile, "<sigf_minus>%s</sigf_minus>\n",xtal.sf[d].sdevm.c_str());
	  }  
	  fprintf(xmlfile, "<wavelength></wavelength>\n");
	  fprintf(xmlfile, "<anomalous>1</anomalous>\n");
        }
        else
        {
	  if (xtal.sf[d].type == "INTENSITY")
	  {
	    fprintf(xmlfile, "<i>%s</i>\n", xtal.sf[d].sdatap.c_str());
	    fprintf(xmlfile, "<sigi>%s</sigi>\n",xtal.sf[d].sdevp.c_str());
	  }
	  else
	  {
	    fprintf(xmlfile, "<f>%s</f>\n", xtal.sf[d].sdatap.c_str());
	    fprintf(xmlfile, "<sigf>%s</sigf>\n",xtal.sf[d].sdevp.c_str());
	  }
          fprintf(xmlfile, "<wavelength></wavelength>\n");
          fprintf(xmlfile, "<anomalous>0</anomalous>\n");
        }
        fprintf(xmlfile, "</data>\n");
      }
    fprintf(xmlfile, "</crystal>\n");
  }
  fprintf(xmlfile, "</exp_columns>\n");
  fprintf(xmlfile, "</input>\n");
  fprintf(xmlfile, "<output>\n");
  char stepstring[3];
  sprintf(stepstring,"%1u",step);
  expcoltag        = stepstring;
  expcoltag       += "_PREP";
  freertag         = expcoltag;
  freertag        += "_FREER";
  writeexpcol();
  fprintf(xmlfile, "<freer_columns>\n");
  fprintf(xmlfile, "<freer>%s</freer>\n", freertag.c_str());
  fprintf(xmlfile, "</freer_columns>\n");
  fprintf(xmlfile, "</output>\n");
  fprintf(xmlfile, "<prep>\n");
  fprintf(xmlfile, "<input_rfree>0</input_rfree>\n");
  fprintf(xmlfile, "<rfree_label>Unassigned</rfree_label>\n");
  fprintf(xmlfile, "<rfree_fraction>0.05</rfree_fraction>\n");
  fprintf(xmlfile, "<rescale_f>%u</rescale_f>\n",rescalef);
  fprintf(xmlfile, "<ncyc>100</ncyc>\n");
  fprintf(xmlfile, "<lores>10</lores>\n");
  fprintf(xmlfile, "<hires>%.2f</hires>\n",subhires);
  fprintf(xmlfile, "<mode>ANISOTROPIC</mode>\n");
  fprintf(xmlfile, "<wilson>0</wilson>\n");
  fprintf(xmlfile, "<weight>0</weight>\n");
  fprintf(xmlfile, "</prep>\n");  
}

void Gcxlikelihood::writeafro(const unsigned step) 
{
  fprintf(xmlfile, "AFRO\n");
  fprintf(xmlfile, "<name>AFRO</name>\n");
  fprintf(xmlfile, "<tag>%u%s</tag>\n",step,"_AFRO");
  fprintf(xmlfile, "<step>%u</step>\n",step);
  fprintf(xmlfile, "<input>\n");
  writeexpcol("AMPLITUDE");
  fprintf(xmlfile, "</input>\n");
  fprintf(xmlfile, "<output>\n");
  char stepstring[3];
  sprintf(stepstring,"%1u",step);
  fatag        = stepstring;
  fatag       += "_AFRO";
  eatag        = fatag;
  writeeacol();
  writefacol();
  fprintf(xmlfile, "</output>\n");
  fprintf(xmlfile, "<afro>\n");
  double lores(ZERO);
  for (unsigned d = 0; d < xtal.sf.size(); d++)
    lores         = std::max((double)xtal.sf[d].lowres,lores);
  fprintf(xmlfile, "<manual_reso>>%u</manual_reso>\n",!autoreso);

  if (!autoreso)
  { 
    fprintf(xmlfile, "<lo_res>%.2f</lo_res>\n",lores);
    fprintf(xmlfile, "<hi_res>%.2f</hi_res>\n",subhirescut);
  }
  
  fprintf(xmlfile, "<sigf>%.1f</sigf>\n", afrosigf);
  if (afrosano > ZERO)
  {
    fprintf(xmlfile, "<manual_sigmas>1</manual_sigmas>\n");
    fprintf(xmlfile, "<sano>%.2f</sano>\n", afrosano);
  }
  else
    fprintf(xmlfile, "<manual_sigmas>0</manual_sigmas>\n");
  if (afrosiso > ZERO)
    fprintf(xmlfile, "<siso>%.2f</siso>\n", afrosiso);
  fprintf(xmlfile, "</afro>\n");
}

void Gcxlikelihood::writecrunch2(const unsigned step) 
{
  fprintf(xmlfile, "CRUNCH2\n");
  fprintf(xmlfile, "<name>CRUNCH2</name>\n");
  fprintf(xmlfile, "<tag>%u%s</tag>\n",step,"_CRUNCH2");
  fprintf(xmlfile, "<step>%u</step>\n",step);
  fprintf(xmlfile, "<input>\n");
  writeexpcol("AMPLITUDE");
  writeeacol();
  writefacol();  
  fprintf(xmlfile, "</input>\n");
  fprintf(xmlfile, "<output>\n");
  char stepstring[3];
  sprintf(stepstring,"%1u",step);
  subtag        = stepstring;
  subtag       += "_CRUNCH2";
  fprintf(xmlfile, "<substructure>%s</substructure>\n",subtag.c_str());  
  fprintf(xmlfile, "</output>\n");
  fprintf(xmlfile, "<crunch2>\n");
  fprintf(xmlfile, "<patt_search>1</patt_search>\n");
  fprintf(xmlfile, "<patt_trials>%u</patt_trials>\n",crunchptrials);
  fprintf(xmlfile, "<score_threshold>%.3f</score_threshold>\n",crunchthres);
  fprintf(xmlfile, "<deviation>%.2f</deviation>\n",crunchdev);
  fprintf(xmlfile, "<bp3_verify>%u</bp3_verify>\n",crunchbp3);
  fprintf(xmlfile, "<scattering_power>15</scattering_power>\n");
  if (crunchspec)
    fprintf(xmlfile, "<spec>1</spec>\n");
  fprintf(xmlfile, "<min_atoms>3</min_atoms>\n");
  fprintf(xmlfile, "<ntrials>%u</ntrials>\n",crunchtrials);
  fprintf(xmlfile, "<ncycles>400</ncycles>\n");
  fprintf(xmlfile, "</crunch2>\n");
}

void Gcxlikelihood::writebp3(const unsigned step) 
{
  fprintf(xmlfile, "BP3\n");
  fprintf(xmlfile, "<name>BP3</name>\n");
  fprintf(xmlfile, "<tag>%u%s</tag>\n",step,"_BP3");
  fprintf(xmlfile, "<step>%u</step>\n",step);
  fprintf(xmlfile, "<input>\n");
  if (subtag == "")
    Bp3Error("Gcxlikelihood::writebp3","A substructure must be input");
  fprintf(xmlfile, "<substructure>%s</substructure>\n",subtag.c_str());  
  writeexpcol("AMPLITUDE");
  fprintf(xmlfile, "</input>\n");
  fprintf(xmlfile, "<output>\n");
  char stepstring[3];
  sprintf(stepstring,"%u",step);
  subtag  = stepstring;
  subtag += "_BP3";
  hltag   = phasetag = lastbp3tag = subtag;
  fprintf(xmlfile, "<substructure>%s</substructure>\n",subtag.c_str());
  writephasecol();
  writehlcol();
  fprintf(xmlfile, "</output>\n");
  fprintf(xmlfile, "<bp3>\n");
  if (bp3phase      == "FAST")
    fprintf(xmlfile, "<phase>1</phase>\n");
  else if (bp3phase == "NORMAL")
    fprintf(xmlfile, "<phase>0</phase>\n");
  else
    fprintf(xmlfile, "<phase>%u</phase>\n",((unsigned)(xtal.Getmad()) && !sadonly));
  fprintf(xmlfile, "<hand>%u</hand>\n",(unsigned)bp3hand);
  fprintf(xmlfile, "<stop_fom>%.3f</stop_fom>\n",bp3stopfom);
  if (sadonly)
    fprintf(xmlfile, "<sad_only>1</sad_only>\n");
  fprintf(xmlfile, "</bp3>\n");
}

void Gcxlikelihood::writerefmac(const unsigned step) 
{
  fprintf(xmlfile, "REFMAC\n");
  fprintf(xmlfile, "<name>REFMAC</name>\n");
  fprintf(xmlfile, "<tag>%u%s</tag>\n",step,"_REFMAC");
  fprintf(xmlfile, "<step>%u</step>\n",step);
  fprintf(xmlfile, "<input>\n");
  if (subtag == "")
    Bp3Error("Gcxlikelihood::writerefmac","A substructure must be input");
  fprintf(xmlfile, "<substructure>%s</substructure>\n",subtag.c_str());  
  writeexpcol("AMPLITUDE");
  fprintf(xmlfile, "<freer_columns>\n");
  fprintf(xmlfile, "<freer>%s</freer>\n",freertag.c_str());  
  fprintf(xmlfile, "</freer_columns>\n");
  fprintf(xmlfile, "</input>\n");
  char stepstring[3];
  sprintf(stepstring,"%u",step);
  subtag  = stepstring;
  subtag += "_REFMAC";
  hltag   = phasetag = lastbp3tag = subtag;
  fprintf(xmlfile, "<output>\n");
  fprintf(xmlfile, "<substructure>%s</substructure>\n",subtag.c_str());
  writephasecol();
  writehlcol();
  fprintf(xmlfile, "</output>\n");
  fprintf(xmlfile, "<refmac>\n");
  fprintf(xmlfile, "<solvent>%u</solvent>\n",(unsigned)refmacsolvent);
  fprintf(xmlfile, "<stop_fom>%.3f</stop_fom>\n",bp3stopfom);
  fprintf(xmlfile, "<hand>%u</hand>\n",(unsigned)refmachand);
  fprintf(xmlfile, "</refmac>\n");
}

void Gcxlikelihood::writeshelxc(const unsigned step) 
{
  fprintf(xmlfile, "SHELXC\n");
  fprintf(xmlfile, "<name>SHELXC</name>\n");
  fprintf(xmlfile, "<tag>%u%s</tag>\n",step,"_SHELXC");
  fprintf(xmlfile, "<step>%u</step>\n",step);
  fprintf(xmlfile, "<input>\n");
  fprintf(xmlfile, "<exp_columns>\n");
  for (unsigned c   = 0; c < xtal.cell.size(); c++)
  {
    int w           = -1;
    fprintf(xmlfile, "<crystal id=\"%u\">\n",c+1);
    for (unsigned d = 0; d < xtal.sf.size(); d++)
      if (c        == xtal.sf[d].nxtal)
      {
	w++;
        fprintf(xmlfile, "<data id=\"%u\">\n",w+1);
        fprintf(xmlfile, "<type>%s</type>\n",xtal.sf[d].name.c_str());
	char junk[6];
 	sprintf(junk,"_X%u_D%u",c+1,w+1);
	string junk2(junk);
	string f("IN" + junk2 + "_F");
 	string sigf("IN" + junk2 + "_SIGF");	
        if (xtal.sf[d].anomalous)
        {
	  if (xtal.sf[d].type == "INTENSITY")
	  {
	    string ip("IN" + junk2 + "_I+");
	    string sigip("IN" + junk2 + "_SIGI+");
	    string im("IN" + junk2 + "_I-");
	    string sigim("IN" + junk2 + "_SIGI-");
	    fprintf(xmlfile, "<i_plus>%s</i_plus>\n",ip.c_str());
	    fprintf(xmlfile, "<sigi_plus>%s</sigi_plus>\n",sigip.c_str());
	    fprintf(xmlfile, "<i_minus>%s</i_minus>\n",im.c_str() );
	    fprintf(xmlfile, "<sigi_minus>%s</sigi_minus>\n",sigim.c_str());
	  }
	  else
	  {
 	    string fp("IN" + junk2 + "_F+");
	    string sigfp("IN" + junk2 + "_SIGF+");
	    string fm("IN" + junk2 + "_F-");
	    string sigfm("IN" + junk2 + "_SIGF-");	    
	    fprintf(xmlfile, "<f_plus>%s</f_plus>\n",fp.c_str());
	    fprintf(xmlfile, "<sigf_plus>%s</sigf_plus>\n",sigfp.c_str());
	    fprintf(xmlfile, "<f_minus>%s</f_minus>\n",fm.c_str() );
	    fprintf(xmlfile, "<sigf_minus>%s</sigf_minus>\n",sigfm.c_str());
	  }  
	  fprintf(xmlfile, "<wavelength></wavelength>\n");
	  fprintf(xmlfile, "<anomalous>1</anomalous>\n");
        }
        else
        {
	  if (xtal.sf[d].type == "INTENSITY")
	  {
	    string i("IN" + junk2 + "_I");
	    string sigi("IN" + junk2 + "_SIGI");
	    fprintf(xmlfile, "<i>%s</i>\n", i.c_str());
	    fprintf(xmlfile, "<sigi>%s</sigi>\n",sigi.c_str());
	  }
	  else
	  {
	    string f("IN" + junk2 + "_F");
	    string sigf("IN" + junk2 + "_SIGF");
	    fprintf(xmlfile, "<f>%s</f>\n", f.c_str());
	    fprintf(xmlfile, "<sigf>%s</sigf>\n",sigf.c_str());
	  }
          fprintf(xmlfile, "<wavelength></wavelength>\n");
          fprintf(xmlfile, "<anomalous>0</anomalous>\n");
        }
        fprintf(xmlfile, "</data>\n");
      }
    fprintf(xmlfile, "</crystal>\n");
  }
  fprintf(xmlfile, "</exp_columns>\n");
  fprintf(xmlfile, "</input>\n");
  fprintf(xmlfile, "<output>\n");
  char stepstring[3];
  sprintf(stepstring,"%1u",step);
  fatag        = stepstring;
  fatag       += "_SHELXC";
  alphatag     = fatag;
  writefacol();

  if (xtal.sf[0].type == "INTENSITY")
  {
    icoltag    = fatag;
    writeicol();
  }
  else
  {
    fcoltag    = fatag;
    writefcol();
  }
  alphatag     = alphatag + "_ALPHA";
  fprintf(xmlfile, "<alpha_columns>\n");
  fprintf(xmlfile, "<alpha>%s</alpha>\n", alphatag.c_str());
  fprintf(xmlfile, "</alpha_columns>\n");
  fprintf(xmlfile, "</output>\n");
  fprintf(xmlfile, "<shelx>\n");
  fprintf(xmlfile, "</shelx>\n");
}

void Gcxlikelihood::checkshelx()
{
  bool changelabels(false), mad(false), native(false);
  double maxfpp(ZERO), minfp(100.0), minfpp(100.0), maxfp(-100.0);
  unsigned peakd(0), infld(0), hremd(0), lremd(0), natived(0);
  
  // make sure labels are allowed
  for (unsigned c                       = 0; c < xtal.cell.size(); c++)
  {
    int w(-1);
    bool substructure(false);
    for (unsigned a                     = 0; a < mdl.atom.size(); a++)
      if (c                            == mdl.atom[a].crystal)
 	substructure                    = true;
    
    for (unsigned d                     = 0; d < xtal.sf.size(); d++)
      if (c                            == xtal.sf[d].nxtal)
      {
	w++;
	if (w                           > 0)
	  mad                           = true;
	if (!( (xtal.sf[d].Getname()   == "NAT")  ||
	       (xtal.sf[d].Getname()   == "SAD")  ||
	       (xtal.sf[d].Getname()   == "SIR")  ||
	       (xtal.sf[d].Getname()   == "SIRA") ||
	       (xtal.sf[d].Getname()   == "PEAK") ||
	       (xtal.sf[d].Getname()   == "INFL") ||
	       (xtal.sf[d].Getname()   == "HREM") ||
	       (xtal.sf[d].Getname()   == "LREM")   ))
	{
 	  if (!changelabels)
	    Bp3Warning("Gcxlikelihood::checkshelx","Data name must be NAT, SAD, SIR, SIRA, PEAK, INFL, HREM or LREM");
	  changelabels                  = true;
	}

	if (xtal.sf.size()             == 1)
	{
	  if (xtal.sf[0].Getname()     != "SAD")
	    Bp3Warning("Gcxlikelihood::checkshelx","Data name must be SAD - resetting it");
	  xtal.sf[0].Setname("SAD");
	  changelabels                  = false;
	}

	if (changelabels)
	{
	  for (unsigned f               = 0; f < mdl.form.size(); f++)
	    if (d                      == xtal.dataset[mdl.form[f].Getcrystal()][w])
	    {
	      printf("dataset %u fp = %f fpp = %f\n",d, mdl.form[f].Getfp(w), mdl.form[f].Getfpp(w));
	      printf("maxfp = %f maxfpp = %f minfp = %f minfpp = %f\n",maxfp,maxfpp,minfp,minfpp);
	      if (mdl.form[f].Getfpp(w) > maxfpp)
	      {
		peakd                   = d;
		maxfpp                  = mdl.form[f].Getfpp(w);
	      }
	      if (mdl.form[f].Getfpp(w) < minfpp)
	      {
		lremd                   = d;
		minfpp                  = mdl.form[f].Getfpp(w);
	      }
	      if (mdl.form[f].Getfp(w)  < minfp)
	      {
		infld                   = d;
		minfp                   = mdl.form[f].Getfp(w); 	
	      }
	      if (mdl.form[f].Getfp(w)  > maxfp)
	      {
		hremd                   = d;
		maxfp                   = mdl.form[f].Getfp(w);
	      }
	      printf("maxfp = %f maxfpp = %f minfp = %f minfpp = %f\n",maxfp,maxfpp,minfp,minfpp);
	    }
	  if (!substructure)
	  {
	    xtal.sf[d].Setname("NAT");
	    native                      = true;
	    natived                     = d;
	  }
	}
      }
  }
  
  bool siras(false);
  if (xtal.sf.size()                   == 2)
    if (!(xtal.sf[0].anomalous && xtal.sf[1].anomalous) && (xtal.cell.size() == 2))
      siras                             = true;

  if (changelabels)
  {
    if (xtal.sf.size()                 == 2)
    {
      if (siras)
	if (xtal.sf[0].anomalous)
	  xtal.sf[0].Setname("SIRA");
	else
	  xtal.sf[1].Setname("SIRA");
      else if (xtal.cell.size()        == 2)
	xtal.sf[1].Setname("SIR");
      else
	if ((peakd                      < 2) && (infld < 2) && (peakd != infld))
	{
	  xtal.sf[peakd].Setname("PEAK");
	  xtal.sf[infld].Setname("INFL");
	}
	else
	  Bp3Error("Gcxlikelihood::checkshelx","Error in assigning data set labels I - set them manually");
    }
    else if (mad)
    {
      printf("peakd = %u infld = %u hremd = %u lremd = %u\n",peakd,infld,hremd,lremd);
      for (unsigned d                   = 0; d < xtal.sf.size(); d++)
	if (d                          == peakd)
	  xtal.sf[d].Setname("PEAK");
	else if (d                     == infld)
 	  xtal.sf[d].Setname("INFL");
	else if (d                     == hremd)
  	  xtal.sf[d].Setname("HREM");
	else if (d                     == lremd)
  	  xtal.sf[d].Setname("LREM");
	else if (d                     != natived)
	  Bp3Error("Gcxlikelihood::checkshelx","Error in assigning data set labels II - set them manually");
    }
    else
      Bp3Error("Gcxlikelihood::checkshelx","Error in assigning data set labels III - set them manually");
  }
  
  bool usemlhl(false);
  for (unsigned i                       = 0; i < step.size(); i++)
    if (step[i]                        == "BP3")
      usemlhl                           = true;
  
  // assign target function if not assigned
  if (!refmactarget.size())
    if ( /* siras || */ ( ( (xtal.sf.size()     == 1) || (xtal.Getmad() && !usemlhl) ) && xtal.sf[refd].anomalous) )
      refmactarget                      = "DIRE";
    else if (usemlhl)
      refmactarget                      = "MLHL";
    else
      refmactarget                      = "RICE";
}

void Gcxlikelihood::writeshelxd(const unsigned step) 
{
  fprintf(xmlfile, "SHELXD\n");
  fprintf(xmlfile, "<name>SHELXD</name>\n");
  fprintf(xmlfile, "<tag>%u%s</tag>\n",step,"_SHELXD");
  fprintf(xmlfile, "<step>%u</step>\n",step);
  fprintf(xmlfile, "<input>\n");
  writefacol();
  fprintf(xmlfile, "<alpha_columns>\n");
  fprintf(xmlfile, "<alpha>%s</alpha>\n", alphatag.c_str());
  fprintf(xmlfile, "</alpha_columns>\n");
  fprintf(xmlfile, "</input>\n");
  fprintf(xmlfile, "<output>\n");
  char stepstring[3];
  sprintf(stepstring,"%1u",step);
  subtag        = stepstring;
  subtag       += "_SHELXD";
  lastshelxdtag = subtag;
  fprintf(xmlfile, "<substructure>%s</substructure>\n",subtag.c_str());
  fprintf(xmlfile, "</output>\n");
  fprintf(xmlfile, "<shelx>\n");
  fprintf(xmlfile, "<dsul>%u</dsul>\n",shelxddsul);
  fprintf(xmlfile, "<mind>%.2f</mind>\n",shelxdmind);
  fprintf(xmlfile, "<mdeq>%.2f</mdeq>\n",shelxdmdeq);
  fprintf(xmlfile, "<ntry>%u</ntry>\n",shelxdtrials);
  fprintf(xmlfile, "<ccweak_threshold>%.2f</ccweak_threshold>\n",shelxdthres);
  fprintf(xmlfile, "<min_trials>%u</min_trials>\n",shelxdmtrials);
  fprintf(xmlfile, "<pats>1</pats>\n");
  fprintf(xmlfile, "<lo_res>%.2f</lo_res>\n",xtal.sf[refd].lowres);
  if (!autoreso)
     fprintf(xmlfile, "<hi_res>%.2f</hi_res>\n",subhirescut);
  fprintf(xmlfile, "</shelx>\n");
}

void Gcxlikelihood::writeshelxe(const unsigned step) 
{
  fprintf(xmlfile, "SHELXE\n");
  fprintf(xmlfile, "<name>SHELXE</name>\n");
  fprintf(xmlfile, "<tag>%u%s</tag>\n",step,"_SHELXE");
  fprintf(xmlfile, "<step>%u</step>\n",step);
  fprintf(xmlfile, "<input>\n");
  if (subtag == "")
    Bp3Error("Gcxlikelihood::writeshelxe","A substructure must be input");
  fprintf(xmlfile, "<substructure>%s</substructure>\n",subtag.c_str());
  if (xtal.sf[0].type == "INTENSITY")
    writeicol();
  else
    writefcol();
  writefacol();
  fprintf(xmlfile, "<alpha_columns>\n");
  fprintf(xmlfile, "<alpha>%s</alpha>\n", alphatag.c_str());
  fprintf(xmlfile, "</alpha_columns>\n");
  if (shelxeinphase)
    writephasecol();
  fprintf(xmlfile, "</input>\n");
  fprintf(xmlfile, "<output>\n");
  char stepstring[3];
  sprintf(stepstring,"%1u",step);
  subtag          = stepstring;
  subtag         += "_SHELXE";
  phasetag        = subtag;
  fprintf(xmlfile, "<substructure>%s</substructure>\n",subtag.c_str());
  // writefcol();
  writephasecol();
  fprintf(xmlfile, "</output>\n");
  fprintf(xmlfile, "<shelxe>\n");
  // ***NSP  ** should i also output ha phase and f columns?
  unsigned generateheavy(1);
  fprintf(xmlfile, "<determine_hand>%u</determine_hand>\n",shelxehand);
  fprintf(xmlfile, "<hand_cycles>%u</hand_cycles>\n",shelxehcycles);
  fprintf(xmlfile, "<optimize_solvent>%u</optimize_solvent>\n",shelxeopt);
  fprintf(xmlfile, "<opt_cycles>%u</opt_cycles>\n",shelxeocycles);
  fprintf(xmlfile, "<density_modify>%u</density_modify>\n",shelxedenmod);
  fprintf(xmlfile, "<ncycles>%u</ncycles>\n",shelxencycles);
  fprintf(xmlfile, "<generate_heavy>%u</generate_heavy>\n",generateheavy);
  fprintf(xmlfile, "<free_lunch>%u</free_lunch>\n",freelunch);
  fprintf(xmlfile, "<free_lunch_reso>%.2f</free_lunch_reso>\n",freelunchreso);
  fprintf(xmlfile, "</shelxe>\n");
}

void Gcxlikelihood::writesolomon(const unsigned step) 
{
  fprintf(xmlfile, "SOLOMON\n");
  fprintf(xmlfile, "<name>SOLOMON</name>\n");
  fprintf(xmlfile, "<tag>%u%s</tag>\n",step,"_SOLOMON");
  fprintf(xmlfile, "<step>%u</step>\n",step);
  fprintf(xmlfile, "<input>\n");
  // set the input phase and hl columns to the last bp3 run
  hltag    = phasetag = lastbp3tag;
  writeexpcol("AMPLITUDE");
  writephasecol();
  writehlcol();
  fprintf(xmlfile, "<substructure>%s</substructure>\n",subtag.c_str());
  fprintf(xmlfile, "</input>\n");
  fprintf(xmlfile, "<output>\n");
  char stepstring[3];
  sprintf(stepstring,"%u",step);
  hltag    = stepstring;
  hltag   += "_SOLOMON";
  phasetag = hltag;
  subtag   = phasetag;
  fprintf(xmlfile, "<substructure>%s</substructure>\n",subtag.c_str());
  writephasecol();
  writehlcol();
  fprintf(xmlfile, "</output>\n");
  fprintf(xmlfile, "<solomon>\n");
  fprintf(xmlfile, "<trunc_low>0.4</trunc_low>\n");
  fprintf(xmlfile, "<trunc_high>1.0</trunc_high>\n");
  fprintf(xmlfile, "<radius>3</radius>\n");
  fprintf(xmlfile, "<determine_hand>%u</determine_hand>\n",solomonhand);
  fprintf(xmlfile, "<hand_cycles>%u</hand_cycles>\n",solomonhcycles);
  fprintf(xmlfile, "<margin>%.3f</margin>\n",solomonhandmargin);
  fprintf(xmlfile, "<optimize_solvent>%u</optimize_solvent>\n",solomonopt);
  fprintf(xmlfile, "<opt_cycles>%u</opt_cycles>\n",solomonocycles);
  fprintf(xmlfile, "<density_modify>%u</density_modify>\n",solomondenmod);
  fprintf(xmlfile, "<ncycles>%u</ncycles>\n",solomonncycles);  
  fprintf(xmlfile, "<keep_mtz>%u</keep_mtz>\n",!solomonnomtz);  
  fprintf(xmlfile, "<keep_map>%u</keep_map>\n",!solomonnomap);  
  fprintf(xmlfile, "</solomon>\n");
}

void Gcxlikelihood::writedm(const unsigned step) 
{
  fprintf(xmlfile, "DM\n");
  fprintf(xmlfile, "<name>DM</name>\n");
  fprintf(xmlfile, "<tag>%u%s</tag>\n",step,"_DM");
  fprintf(xmlfile, "<step>%u</step>\n",step);
  fprintf(xmlfile, "<input>\n");
  // set the input phase and hl columns to the last bp3 run
  hltag    = phasetag = lastbp3tag;
  writeexpcol("AMPLITUDE");
  writephasecol();
  writehlcol();
  fprintf(xmlfile, "</input>\n");
  fprintf(xmlfile, "<output>\n");
  char stepstring[3];
  sprintf(stepstring,"%u",step);
  hltag    = stepstring;
  hltag   += "_DM";
  phasetag = hltag;
  if (dmhand)
  {
    subtag   = phasetag;
    fprintf(xmlfile, "<substructure>%s</substructure>\n",subtag.c_str());
  }
  writephasecol();
  writehlcol();
  fprintf(xmlfile, "</output>\n");
  fprintf(xmlfile, "<dm>\n");
  fprintf(xmlfile, "<mode_solv>1</mode_solv>\n");
  fprintf(xmlfile, "<mode_hist>%u</mode_hist>\n",dmhist);
  fprintf(xmlfile, "<ncycles>AUTO</ncycles>\n");
  fprintf(xmlfile, "<determine_hand>%u</determine_hand>\n",dmhand);
  fprintf(xmlfile, "<margin>%.3f</margin>\n",dmhandmargin);
  fprintf(xmlfile, "<optimize_solvent>%u</optimize_solvent>\n",dmopt);
  fprintf(xmlfile, "<density_modify>%u</density_modify>\n",dmdenmod);
  fprintf(xmlfile, "</dm>\n");
}

void Gcxlikelihood::writepirate(const unsigned step) 
{
  fprintf(xmlfile, "PIRATE\n");
  fprintf(xmlfile, "<name>PIRATE</name>\n");
  fprintf(xmlfile, "<tag>%u%s</tag>\n",step,"_PIRATE");
  fprintf(xmlfile, "<step>%u</step>\n",step);
  fprintf(xmlfile, "<input>\n");
  // set the input phase and hl columns to the last bp3 run
  //  hltag    = phasetag = lastbp3tag;
  writeexpcol("AMPLITUDE");
  writehlcol();
  fprintf(xmlfile, "<freer_columns>\n");
  fprintf(xmlfile, "<freer>%s</freer>\n",freertag.c_str());  
  fprintf(xmlfile, "</freer_columns>\n");
  fprintf(xmlfile, "<substructure>%s</substructure>\n",subtag.c_str());
  fprintf(xmlfile, "</input>\n");
  fprintf(xmlfile, "<output>\n");
  char stepstring[3];
  sprintf(stepstring,"%u",step);
  hltag    = stepstring;
  hltag   += "_PIRATE";
  phasetag = hltag;
  writephasecol();
  writehlcol();
  fprintf(xmlfile, "</output>\n");
  fprintf(xmlfile, "<pirate>\n");
  fprintf(xmlfile, "<determine_hand>%u</determine_hand>\n",piratehand);
  fprintf(xmlfile, "<hlphasing_coefficients>%u</hlphasing_coefficients>\n",piratehl);
  fprintf(xmlfile, "<use_ncs>%u</use_ncs>\n",piratencs);
  //  fprintf(xmlfile, "<margin>%.3f</margin>\n",dmhandmargin);
  fprintf(xmlfile, "<optimize_solvent>%u</optimize_solvent>\n",pirateopt);
  fprintf(xmlfile, "<density_modify>%u</density_modify>\n",piratedenmod);
  fprintf(xmlfile, "<ncycles>%u</ncycles>\n",piratencycles);
  fprintf(xmlfile, "<inweight>%.3f</inweight>\n",piratewihl);
  fprintf(xmlfile, "</pirate>\n");
}

void Gcxlikelihood::writebuccaneer(const unsigned step) 
{
  fprintf(xmlfile, "BUCCANEER\n");
  fprintf(xmlfile, "<name>BUCCANEER</name>\n");
  fprintf(xmlfile, "<tag>%u%s</tag>\n",step,"_BUCCANEER");
  fprintf(xmlfile, "<step>%u</step>\n",step);
  fprintf(xmlfile, "<input>\n");
  writeexpcol("AMPLITUDE");
  writephasecol();
  writehlcol();
  if ( ( (xtal.sf.size() == 1) /* || ( (xtal.sf.size() == 2) && !(xtal.sf[0].anomalous && xtal.sf[1].anomalous)) */ ) && !refmactarget.size())
    refmactarget         = "DIRE";
  
  if (refmactarget      == "DIRE")
    fprintf(xmlfile, "<substructure>%s</substructure>\n",subtag.c_str());
  
  if (!buccnofreer)
  {
    fprintf(xmlfile, "<freer_columns>\n");
    fprintf(xmlfile, "<freer>%s</freer>\n",freertag.c_str());  
    fprintf(xmlfile, "</freer_columns>\n");
  }
  bool inputpdb(false);
  if (pdbtag.size())
  {
    fprintf(xmlfile, "<pdb>%s</pdb>\n",pdbtag.c_str());
    inputpdb             = true;
  }
  
  fprintf(xmlfile, "</input>\n");
  fprintf(xmlfile, "<output>\n");
  char stepstring[3];
  sprintf(stepstring,"%u",step);
  pdbtag                 = stepstring;
  pdbtag                += "_BUCCANEER";
  phasetag               = pdbtag;
  writephasecol();
  if (!buccnofreer)
  {
    fprintf(xmlfile, "<freer_columns>\n");
    fprintf(xmlfile, "<freer>%s</freer>\n",freertag.c_str());  
    fprintf(xmlfile, "</freer_columns>\n");
  }
  fprintf(xmlfile, "<pdb>%s</pdb>\n",pdbtag.c_str());
  fprintf(xmlfile, "</output>\n");
  fprintf(xmlfile, "<buccaneer>\n");
  if (refmactarget      == "DIRE")  
    fprintf(xmlfile, "<phase_restrain>DIRECT</phase_restrain>\n");
  else if (refmactarget != "RICE") 
    fprintf(xmlfile, "<phase_restrain>MLHL</phase_restrain>\n");    
  else
    fprintf(xmlfile, "<phase_restrain>NO</phase_restrain>\n");
  fprintf(xmlfile, "<nouse_freer>%u</nouse_freer>\n",buccnofreer);
  fprintf(xmlfile, "<no_use_pdb>%u</no_use_pdb>\n",!(buccusepdb && inputpdb));
  fprintf(xmlfile, "<fast_build>%u</fast_build>\n",buccfast);
  fprintf(xmlfile, "<fragments>%u</fragments>\n",buccfragments);
  fprintf(xmlfile, "<num_cycles>%u</num_cycles>\n",bucccycles);
  fprintf(xmlfile, "<sequence_match>%.3f</sequence_match>\n",buccreli);
  fprintf(xmlfile, "</buccaneer>\n");
}

void Gcxlikelihood::writeresolvedm(const unsigned step) 
{
  fprintf(xmlfile, "RESOLVEDM\n");
  fprintf(xmlfile, "<name>RESOLVEDM</name>\n");
  fprintf(xmlfile, "<tag>%u%s</tag>\n",step,"_RESOLVEDM");
  fprintf(xmlfile, "<step>%u</step>\n",step);
  fprintf(xmlfile, "<input>\n");
  writeexpcol("AMPLITUDE");
  writephasecol();
  writehlcol();
  fprintf(xmlfile, "<substructure>%s</substructure>\n",subtag.c_str());
  fprintf(xmlfile, "<freer_columns>\n");
  fprintf(xmlfile, "<freer>%s</freer>\n",freertag.c_str());  
  fprintf(xmlfile, "</freer_columns>\n");
  fprintf(xmlfile, "</input>\n");
  fprintf(xmlfile, "<output>\n");
  char stepstring[3];
  sprintf(stepstring,"%u",step);
  hltag    = stepstring;
  hltag   += "_RESOLVEDM";
  phasetag = hltag;
  writephasecol();
  writehlcol();
  fprintf(xmlfile, "</output>\n");
  fprintf(xmlfile, "<resolvedm>\n");
  fprintf(xmlfile, "<hlphasing_coefficients>%u</hlphasing_coefficients>\n",resolvehl);
  fprintf(xmlfile, "<ncs>%u</ncs>\n",resolvedmncs);
  fprintf(xmlfile, "</resolvedm>\n");
}

void Gcxlikelihood::writeresolvebuild(const unsigned step) 
{
  fprintf(xmlfile, "RESOLVEMB\n");
  fprintf(xmlfile, "<name>RESOLVEMB</name>\n");
  fprintf(xmlfile, "<tag>%u%s</tag>\n",step,"_RESOLVEMB");
  fprintf(xmlfile, "<step>%u</step>\n",step);
  fprintf(xmlfile, "<input>\n");
  writeexpcol("AMPLITUDE");
  writephasecol();
  writehlcol();
  fprintf(xmlfile, "<substructure>%s</substructure>\n",subtag.c_str());
  fprintf(xmlfile, "<freer_columns>\n");
  fprintf(xmlfile, "<freer>%s</freer>\n",freertag.c_str());  
  fprintf(xmlfile, "</freer_columns>\n");
  fprintf(xmlfile, "</input>\n");
  fprintf(xmlfile, "<output>\n");
  char stepstring[3];
  sprintf(stepstring,"%u",step);
  hltag    = stepstring;
  hltag   += "_RESOLVEMB";
  phasetag = pdbtag = hltag;
  writephasecol();
  writehlcol();
  fprintf(xmlfile, "<pdb>%s</pdb>\n",pdbtag.c_str());
  fprintf(xmlfile, "</output>\n");
  fprintf(xmlfile, "<resolvemb>\n");
  fprintf(xmlfile, "<type>%s</type>\n",resolvetype.c_str());
  fprintf(xmlfile, "<ncs>%u</ncs>\n",resolvembncs);
  fprintf(xmlfile, "</resolvemb>\n");
}

void Gcxlikelihood::writearpwarp(const unsigned step) 
{
  fprintf(xmlfile, "ARPWARP\n");
  fprintf(xmlfile, "<name>ARPWARP</name>\n");
  fprintf(xmlfile, "<tag>%u%s</tag>\n",step,"_ARPWARP");
  fprintf(xmlfile, "<step>%u</step>\n",step);
  fprintf(xmlfile, "<input>\n");
  writeexpcol("AMPLITUDE");
  writephasecol();

  if ( ( (xtal.sf.size() == 1) /* || ( (xtal.sf.size() == 2) && !(xtal.sf[0].anomalous && xtal.sf[1].anomalous)) */ ) && !refmactarget.size())
    refmactarget          = "DIRE";
  
  if (refmactarget       == "DIRE")
  {
    fprintf(xmlfile, "<substructure>%s</substructure>\n",subtag.c_str());
    fprintf(xmlfile, "<freer_columns>\n");
    fprintf(xmlfile, "<freer>%s</freer>\n",freertag.c_str());  
    fprintf(xmlfile, "</freer_columns>\n");
  }
  else if (refmactarget  != "RICE")
    writehlcol();
  bool inputpdb(false);
  if (pdbtag.size())
  {
    fprintf(xmlfile, "<pdb>%s</pdb>\n",pdbtag.c_str());
    inputpdb              = true;
  }
  
  fprintf(xmlfile, "</input>\n");
  fprintf(xmlfile, "<output>\n");
  char stepstring[3];
  sprintf(stepstring,"%u",step);
  phasetag                = stepstring;
  phasetag               += "_ARPWARP";
  pdbtag                  = phasetag;
  fprintf(xmlfile, "<pdb>%s</pdb>\n",pdbtag.c_str());
  writephasecol();
  fprintf(xmlfile, "<freer_columns>\n");
  fprintf(xmlfile, "<freer>%s</freer>\n",freertag.c_str());  
  fprintf(xmlfile, "</freer_columns>\n");
  fprintf(xmlfile, "</output>\n");
  fprintf(xmlfile, "<arpwarp>\n");

  fprintf(xmlfile, "<use_six>%u</use_six>\n",arpusesix);
  fprintf(xmlfile, "<no_use_pdb>%u</no_use_pdb>\n",!(arpusepdb && inputpdb));
  fprintf(xmlfile, "<side_after>%u</side_after>\n",arpsideafter);
  fprintf(xmlfile, "<small_cycles>%u</small_cycles>\n",arpsmallcycles);
  fprintf(xmlfile, "<big_cycles>%u</big_cycles>\n",arpbigcycles);
  fprintf(xmlfile, "<ncycles>%u</ncycles>\n",arpcycles);
  if (refmactarget       == "DIRE")  
  {
    fprintf(xmlfile, "<phase_restrain>DIRECT</phase_restrain>\n");
    fprintf(xmlfile, "<exclude_freer>Y</exclude_freer>\n");
  }
  else if (refmactarget  != "RICE") 
  {
    fprintf(xmlfile, "<phase_restrain>MLHL</phase_restrain>\n");    
    fprintf(xmlfile, "<exclude_freer>N</exclude_freer>\n");
  }
  else
  {
    fprintf(xmlfile, "<phase_restrain>NO</phase_restrain>\n");    
    fprintf(xmlfile, "<exclude_freer>N</exclude_freer>\n");    
  }
  fprintf(xmlfile, "<phase_blur>1.0</phase_blur>\n");
  fprintf(xmlfile, "<damp_p>0.99</damp_p>\n");
  fprintf(xmlfile, "<damp_b>0.99</damp_b>\n");
  fprintf(xmlfile, "<weight_mode>AUTO</weight_mode>\n");
  fprintf(xmlfile, "<wmat></wmat>\n");
  fprintf(xmlfile, "<scale>BULK</scale>\n");
  fprintf(xmlfile, "<scanis>Y</scanis>\n");
  fprintf(xmlfile, "<refmac_ref_set>Y</refmac_ref_set>\n");
  fprintf(xmlfile, "<solvent>0</solvent>\n");
  fprintf(xmlfile, "<tlsonoff>0</tlsonoff>\n");
  fprintf(xmlfile, "</arpwarp>\n");
}

void Gcxlikelihood::writecheck(const unsigned step) 
{
  fprintf(xmlfile, "CHECK\n");
  fprintf(xmlfile, "<name>CHECK</name>\n");
  fprintf(xmlfile, "<tag>%u%s</tag>\n",step,"_CHECK");
  fprintf(xmlfile, "<step>%u</step>\n",step);
  fprintf(xmlfile, "<input>\n");
  writephasecol();
  fprintf(xmlfile, "<substructure>%s</substructure>\n",subtag.c_str());
  if (pdbtag.size())
     fprintf(xmlfile, "<pdb>%s</pdb>\n",pdbtag.c_str());
  fprintf(xmlfile, "</input>\n");
  fprintf(xmlfile, "<check>\n");

  if (finalpdb.size())
  {
    string temp(finalpdb);

    if (finalpdb[0]          != dir)
    {
      string t2(getenv("PWD"));
      temp                    =  t2 + dir + finalpdb;
    }
    fprintf(xmlfile, "<final_pdb>%s</final_pdb>\n",temp.c_str());
  }
  fprintf(xmlfile, "<no_trans>%u</no_trans>\n",(unsigned)notrans);

  if (finalsubstructure.size())
  {
    string temp(finalsubstructure);

    if (finalsubstructure[0] != dir)
    {
      string t2(getenv("PWD"));
      temp                    =  t2 + dir + finalsubstructure;
    }
    fprintf(xmlfile, "<final_substructure>%s</final_substructure>\n",temp.c_str());
  }
  fprintf(xmlfile, "</check>\n");
}

void Gcxlikelihood::writeexpcol(const string type) const
{
  fprintf(xmlfile, "<exp_columns>\n");
  for (unsigned c   = 0; c < xtal.cell.size(); c++)
  {
    int w           = -1;
    fprintf(xmlfile, "<crystal id=\"%u\">\n",c+1);
    for (unsigned d = 0; d < xtal.sf.size(); d++)
      if (c        == xtal.sf[d].nxtal)
      {
	w++;
 	fprintf(xmlfile, "<data id=\"%u\">\n",w+1);
	char junk[6];
 	sprintf(junk,"_X%u_D%u",c+1,w+1);
	string junk2(junk);
	string f(expcoltag + junk2 + "_F");
 	string sigf(expcoltag + junk2 + "_SIGF");
	if (xtal.sf[d].anomalous)
	{
	  if ((type == "AMPLITUDE") || (type == "BOTH") )
	  {
 	    string fp(expcoltag + junk2 + "_F+");
	    string sigfp(expcoltag + junk2 + "_SIGF+");
	    string fm(expcoltag + junk2 + "_F-");
	    string sigfm(expcoltag + junk2 + "_SIGF-");

	    fprintf(xmlfile, "<f>%s</f>\n",f.c_str());
	    fprintf(xmlfile, "<sigf>%s</sigf>\n",sigf.c_str());
	    fprintf(xmlfile, "<f_plus>%s</f_plus>\n",fp.c_str());
	    fprintf(xmlfile, "<sigf_plus>%s</sigf_plus>\n",sigfp.c_str());
	    fprintf(xmlfile, "<f_minus>%s</f_minus>\n",fm.c_str());
	    fprintf(xmlfile, "<sigf_minus>%s</sigf_minus>\n",sigfm.c_str());
	  }
	  if ((type == "INTENSITY") || (type == "BOTH") &&
	      (xtal.sf[0].type != "AMPLITUDE"))
	    {
	      string ip(expcoltag + junk2 + "_I+");
	      string sigip(expcoltag + junk2 + "_SIGI+");
	      string im(expcoltag + junk2 + "_I-");
	      string sigim(expcoltag + junk2 + "_SIGI-");

	      fprintf(xmlfile, "<i_plus>%s</i_plus>\n",ip.c_str());
	      fprintf(xmlfile, "<sigi_plus>%s</sigi_plus>\n",sigip.c_str());
	      fprintf(xmlfile, "<i_minus>%s</i_minus>\n",im.c_str());
	      fprintf(xmlfile, "<sigi_minus>%s</sigi_minus>\n",sigim.c_str());
	    }
	}
	else
	{
	  if ((type == "AMPLITUDE") || (type == "BOTH"))
	  {
	    fprintf(xmlfile, "<f>%s</f>\n",f.c_str());
	    fprintf(xmlfile, "<sigf>%s</sigf>\n",sigf.c_str());
	  }
	  if ((type == "INTENSITY") || (type == "BOTH") &
	      (xtal.sf[0].type != "AMPLITUDE"))
	  {	  
	    string i(expcoltag + junk2 + "_I");
	    string sigi(expcoltag + junk2 + "_SIGI");
	    fprintf(xmlfile, "<i>%s</i>\n",i.c_str());
	    fprintf(xmlfile, "<sigi>%s</sigi>\n",sigi.c_str());
	  }
	}
	fprintf(xmlfile, "</data>\n");
      }
    fprintf(xmlfile, "</crystal>\n");
  }
  fprintf(xmlfile, "</exp_columns>\n");
}

void Gcxlikelihood::writefcol() const
{
  fprintf(xmlfile, "<f_columns>\n");
  string f(fcoltag + "_F");
  string sigf(fcoltag + "_SIGF");
  fprintf(xmlfile, "<f>%s</f>\n",f.c_str());
  fprintf(xmlfile, "<sigf>%s</sigf>\n",sigf.c_str());
  fprintf(xmlfile, "</f_columns>\n");
}

void Gcxlikelihood::writeicol() const
{
  fprintf(xmlfile, "<i_columns>\n");
  string i(icoltag + "_I");
  string sigi(icoltag + "_SIGI");
  fprintf(xmlfile, "<i>%s</i>\n",i.c_str());
  fprintf(xmlfile, "<sigi>%s</sigi>\n",sigi.c_str());
  fprintf(xmlfile, "</i_columns>\n");
}

void Gcxlikelihood::writeeacol() const
{
  fprintf(xmlfile, "<ea_columns>\n");
  string ea(eatag + "_EA");
  string sigea(eatag + "_SIGEA");
  fprintf(xmlfile, "<ea>%s</ea>\n",ea.c_str());
  fprintf(xmlfile, "<sigea>%s</sigea>\n",sigea.c_str());
  fprintf(xmlfile, "</ea_columns>\n");
}

void Gcxlikelihood::writefacol() const
{
  fprintf(xmlfile, "<fa_columns>\n");
  string fa(fatag + "_FA");
  string sigfa(fatag + "_SIGFA");
  fprintf(xmlfile, "<fa>%s</fa>\n",fa.c_str());
  fprintf(xmlfile, "<sigfa>%s</sigfa>\n",sigfa.c_str());
  fprintf(xmlfile, "</fa_columns>\n");
}

void Gcxlikelihood::writephasecol() const
{
  fprintf(xmlfile, "<phase_columns>\n");
  string f(phasetag + "_F");
  string sigf(phasetag + "_SIGF");
  string phib(phasetag + "_PHIB");
  string fom(phasetag + "_FOM");
  fprintf(xmlfile, "<f>%s</f>\n",f.c_str());
  fprintf(xmlfile, "<sigf>%s</sigf>\n",sigf.c_str());
  fprintf(xmlfile, "<phib>%s</phib>\n",phib.c_str());
  fprintf(xmlfile, "<fom>%s</fom>\n",fom.c_str());
  fprintf(xmlfile, "</phase_columns>\n");
}

void Gcxlikelihood::writehlcol() const
{
  fprintf(xmlfile, "<hl_columns>\n");
  string hla(hltag + "_HLA");
  string hlb(hltag + "_HLB");
  string hlc(hltag + "_HLC");
  string hld(hltag + "_HLD");
  fprintf(xmlfile, "<hla>%s</hla>\n",hla.c_str());
  fprintf(xmlfile, "<hlb>%s</hlb>\n",hlb.c_str());
  fprintf(xmlfile, "<hlc>%s</hlc>\n",hlc.c_str());
  fprintf(xmlfile, "<hld>%s</hld>\n",hld.c_str());
  fprintf(xmlfile, "</hl_columns>\n");
}
