/*
  Copyright (c) Navraj Pannu       1998 - 2005
  Copyright (c) Leiden University  2002 - 2005

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.  
*/

#include <complex>
#include <algorithm>
#if defined (__sgi) && defined (_MIPS_ISA)
extern "C" {
#include <ctype.h>
}
#else
#include <cctype>
#include <cmath>
#endif
#include "gauss.h"
#include "likelihood.h"
#include "matrix.h"
#include "ccp4_fortran.h"
#include "csymlib.h"

extern "C" void FORTRAN_CALL ( DSYEVD, dsyevd,
       (char*, char*, int*, double*, int *,double*, double*, int*, int*, int*, int*),
       (char*, char*, int*, double*, int *,double*, double*, int*, int*, int*, int*),
       (char*, char*, int*, double*, int *,double*, double*, int*, int*, int*, int*));

extern "C" void
FORTRAN_CALL( ZHEEVD, zheevd,
              (char*, char*, int*, std::complex<double>*, int*, double*, std::complex<double>*, int*, double*, int*, int*, int*, int*),
              (char*, char*, int*, std::complex<double>*, int*, double*, std::complex<double>*, int*, double*, int*, int*, int*, int*),
              (char*, char*, int*, std::complex<double>*, int*, double*, std::complex<double>*, int*, double*, int*, int*, int*, int*));

Likelihood::Likelihood(Model &model, Crystal &crystal)
  : mdl(model), xtal(crystal)
{
  // defaults
  likelihood = ZERO;
  verbose    = 0;
  hand       = 1;
  allin      = interpolate = stats = nooutput = false;

  // default size for Gaussian integration parameters
  cweight.resize(0);
  pweight.resize(0);
  afweight.resize(0);
  sadweight.resize(0);
  maxoccshift     = 0.5;
  maxnumbshift    = TEN;
  maxluzzshift    = 0.4;
  maxbfacshift    = 10.0;
  maxuanoshift    = 0.25;
  maxfpshift      = 0.5;
  maxfppshift     = 0.5;
  maxkscaleshift  = 0.25;
  maxbscaleshift  = 5.0;
  maxxyzshift     = 0.05;  // ***NSP - should make it in terms of Orthogonal...
  updatesigmah    = true;
}

void Likelihood::Settarget(const string targin)
{
  if (targin      == "UNCO")
    target         = targin;
  else if (targin == "SAD")
    target         = targin;
  else if (targin == "MAD")
    target         = targin;
  else
    Bp3Error("Crystal::Settarget", "Unknown target function");
}

void Likelihood::directsfcalc(const bool deriv, const bool calcsigmah)
{
  // Calculates structure factors and first derivatives of a TARGET function 
  // wrt parameters via direct summation.

  // The subroutine needs the derivative of the TARGET function 
  // wrt the calculated structure amplitude and pcalc and updated 
  // structure factor amplitude and pcalcs to calculate the required 
  // derivatives.

  // The atomic coordinates are assumed to be in Fractional coordinates.

  // only sum over primitive symmetry operations
  double nmult((double)(xtal.sg.NSYM/xtal.sg.NSYMP));

  if (deriv)
  {
    // initialize first order partial derivatives
    for (unsigned a                        = 0; a < mdl.atom.size(); a++)
    {  
      dLdocc[a]                            = ZERO;
      dLdbiso[a]                           = ZERO;
      for (unsigned b                      = 0; b < 6; b++)
	dLduaniso[a][b]                    = ZERO;
    }

    for (unsigned f                        = 0; f < mdl.form.size(); f++)
      for (unsigned w                      = 0; w < mdl.form[f].nwave; w++)
      {
	dLdfp[f][w]                        = ZERO;
	dLdfpp[f][w]                       = ZERO;
      }

    for (unsigned s                        = 0; s < mdl.site.size(); s++)
      for (unsigned k                      = 0; k < 3; k++)
	dLdx[s][k]                         = ZERO;
  }
  else if (calcsigmah)
    for (unsigned d                        = 0; d < xtal.sf.size(); d++)
      for (unsigned s                      = 0; s < xtal.sf[d].nbins; s++)
      {
	xtal.sf[d].sigmah[s]               = ZERO;
	if ( (target                      == "SAD") || (target == "MAD") ) 
	  for (unsigned d2                 = 0; d2 < xtal.sf.size(); d2++)
	  {
	    sumfpp[d][d2][s]               = ZERO;
	    difffpp[d][d2][s]              = ZERO;
	    difffp[d][d2][s]               = ZERO;
	  }
      }
  
  vector<double> acalc(xtal.sf.size());
  vector<double> bcalc(xtal.sf.size());
  vector<double> anoacalc(xtal.sf.size());
  vector<double> anobcalc(xtal.sf.size());

  for (unsigned r                          = 0; r < xtal.maxselref; r++)
    if (xtal.selected[r])
    {
      for (unsigned d                      = 0; d < xtal.sf.size(); d++)
      {
	acalc[d]                           = ZERO;
	bcalc[d]                           = ZERO;
	anoacalc[d]                        = ZERO;
	anobcalc[d]                        = ZERO;
      }
    
      double h(xtal.miller[r][0]), k(xtal.miller[r][1]), l(xtal.miller[r][2]);

      for (unsigned sym                    = 0; sym < xtal.sg.NSYMP; sym++)
      {
	hrot[sym][0]                       = ((double)(xtal.sg.symrot[sym][0][0])*h +
					      (double)(xtal.sg.symrot[sym][0][1])*k +
					      (double)(xtal.sg.symrot[sym][0][2])*l);
	
	hrot[sym][1]                       = ((double)(xtal.sg.symrot[sym][1][0])*h +
					      (double)(xtal.sg.symrot[sym][1][1])*k +
					      (double)(xtal.sg.symrot[sym][1][2])*l);	

	hrot[sym][2]                       = ((double)(xtal.sg.symrot[sym][2][0])*h +
					      (double)(xtal.sg.symrot[sym][2][1])*k +
					      (double)(xtal.sg.symrot[sym][2][2])*l);	

	htran[sym]                         = (xtal.sg.symtran[sym][0]*h +
					      xtal.sg.symtran[sym][1]*k +
					      xtal.sg.symtran[sym][2]*l);
      }

      for (unsigned a                      = 0; a < mdl.atom.size(); a++)
      {
	double costotal(ZERO), sintotal(ZERO);

	double dcosdx[3]                   = {ZERO, ZERO, ZERO};
	double dsindx[3]                   = {ZERO, ZERO, ZERO};

	// Atom's site number
	unsigned s(mdl.atom[a].nsite);

	for (unsigned sym                  = 0; sym < xtal.sg.NSYMP; sym++)
	{
	  // Generate all symmetry elements                                          
	  double arg((hrot[sym][0]*mdl.site[s].x[0] +
		      hrot[sym][1]*mdl.site[s].x[1] +
		      hrot[sym][2]*mdl.site[s].x[2] +
		      htran[sym])*TWOPI);

	  double cosarg(tab.tcos(arg));
	  double sinarg(tab.tsin2(arg));

	  costotal                        += cosarg;
	  sintotal                        += sinarg;

	  if (deriv)
	  {
	    dcosdx[0]                     -= sinarg*hrot[sym][0];
	    dsindx[0]                     += cosarg*hrot[sym][0];
	    dcosdx[1]                     -= sinarg*hrot[sym][1];
	    dsindx[1]                     += cosarg*hrot[sym][1];
	    dcosdx[2]                     -= sinarg*hrot[sym][2];
	    dsindx[2]                     += cosarg*hrot[sym][2];
	  }
	}
	
	// the crystal the atom belongs to
	unsigned c(mdl.atom[a].crystal);
	double stolsq(xtal.cell[c].stolsq[r]);

	double bfactor;
	double coefbano[6]                 = {ZERO, ZERO, ZERO, ZERO, ZERO, ZERO};
      
	if (mdl.atom[a].isotropic)
	  bfactor                          = tab.texp(-mdl.atom[a].biso*stolsq);
	else
	{
	  coefbano[0]                      = h*h*xtal.cell[c].astar*xtal.cell[c].astar;
	  coefbano[1]                      = TWO*h*k*xtal.cell[c].astar*xtal.cell[c].bstar;
	  coefbano[2]                      = TWO*h*l*xtal.cell[c].astar*xtal.cell[c].cstar;
	  coefbano[3]                      = k*k*xtal.cell[c].bstar*xtal.cell[c].bstar;
	  coefbano[4]                      = TWO*k*l*xtal.cell[c].bstar*xtal.cell[c].cstar;
	  coefbano[5]                      = l*l*xtal.cell[c].cstar*xtal.cell[c].cstar;
	  bfactor                          = tab.texp(-TWOPI2*std::max(coefbano[0]*
							 	       mdl.atom[a].uaniso[0] +
								       coefbano[1]*
								       mdl.atom[a].uaniso[1]+
								       coefbano[2]*
								       mdl.atom[a].uaniso[2]+
								       coefbano[3]*
								       mdl.atom[a].uaniso[3]+
								       coefbano[4]*
								       mdl.atom[a].uaniso[4]+
								       coefbano[5]*
								       mdl.atom[a].uaniso[5],
								       ZERO)); 
        }
	double btimeso(bfactor*mdl.atom[a].occ);
	
	costotal                          *= bfactor*nmult;
	sintotal                          *= bfactor*nmult;

	if (deriv)
	{
	  double temp(TWOPI*btimeso*nmult);
	  dcosdx[0]                       *= temp;
	  dsindx[0]                       *= temp;
	  dcosdx[1]                       *= temp;
	  dsindx[1]                       *= temp;
	  dcosdx[2]                       *= temp;
	  dsindx[2]                       *= temp;
	}

	// Calculate the normal scattering factor for atom and resolution

	unsigned f(mdl.atom[a].nform);
	
	double normalscat((double)(mdl.form[f].a[0]*
				   tab.texp(-mdl.form[f].b[0]*stolsq) +
				   mdl.form[f].a[1]*
				   tab.texp(-mdl.form[f].b[1]*stolsq) +
				   mdl.form[f].a[2]*
				   tab.texp(-mdl.form[f].b[2]*stolsq) +
				   mdl.form[f].a[3]*
				   tab.texp(-mdl.form[f].b[3]*stolsq) +
				   mdl.form[f].c));

	// calculate anomalous scattering for each wavelength
	for (unsigned w                    = 0; w < mdl.form[f].fp.size(); w++)
	{
	  unsigned d(xtal.dataset[c][w]);
	  if (xtal.sf[d].use(r))
	  {
	    unsigned sh(xtal.bin(d,r));
	    double scat(normalscat + mdl.form[f].fp[w]);
	    double dReFpdo(scat*costotal), dImFpdo(scat*sintotal);

	    if (!deriv)
	    {
	      acalc[d]                    += dReFpdo*mdl.atom[a].occ;
	      bcalc[d]                    += dImFpdo*mdl.atom[a].occ;
	      if (calcsigmah)
		xtal.sf[d].sigmah[sh]     += scat*scat*btimeso*btimeso;
	      
	      if (xtal.sf[d].anomalous)
	      {
		double fpp(mdl.form[f].fpp[w]);
		anoacalc[d]               -= fpp*sintotal*mdl.atom[a].occ;
		anobcalc[d]               += fpp*costotal*mdl.atom[a].occ;
		if (calcsigmah)
		{
		  double fp(mdl.form[f].fp[w]*btimeso);
		  fpp                     *= btimeso;
		  xtal.sf[d].sigmah[sh]   += fpp*fpp;
		  if ( (target            == "SAD") || (target == "MAD") ) 
		  {  
		    sumfpp[d][d][sh]     += TWO*fpp*fpp;
		    for (unsigned w2      = w+1; w2 < mdl.form[f].fp.size(); w2++)
		    {
		      unsigned d2(xtal.dataset[c][w2]);
		      double fp2(mdl.form[f].fp[w2]*btimeso);
		      double fpp2(mdl.form[f].fpp[w2]*btimeso);
		      difffp[d][d2][sh]  += scat*(fp2 - fp)*btimeso;
		      difffpp[d][d2][sh] += fpp*(fpp2 - fpp);
		      sumfpp[d][d2][sh]  += fpp*(fpp2 + fpp);
		    }
		  }
		} 	
	      }
	    }
	    else
	    {
	      double dReFdx[3]             = {scat*dcosdx[0], scat*dcosdx[1], scat*dcosdx[2]};
	      double dImFdx[3]             = {scat*dsindx[0], scat*dsindx[1], scat*dsindx[2]};

	      if (xtal.sf[d].anomalous)
	      {
		dReFpdo                   -= mdl.form[f].fpp[w]*sintotal;
		dImFpdo                   += mdl.form[f].fpp[w]*costotal;
		 for (unsigned i           = 0; i < 3; i++)
		 {
		   dReFdx[i]              -= mdl.form[f].fpp[w]*dsindx[i];
		   dImFdx[i]              += mdl.form[f].fpp[w]*dcosdx[i];
		 }
	      }

	      double mult(TWO*bfactor*btimeso*xtal.sg.NSYMP/
 			  ((double)(xtal.sf[d].nshl[sh])));

	      double temp(scat*scat*mult);
	      
	      dLdocc[a]                   += (dLdAp[d][r]*dReFpdo +
					      dLdBp[d][r]*dImFpdo +
					      dLdsigmah[d][sh]*temp);
	  
	      dLdx[s][0]                  += (dLdAp[d][r]*dReFdx[0] +
					      dLdBp[d][r]*dImFdx[0]);

	      dLdx[s][1]                  += (dLdAp[d][r]*dReFdx[1] +
					      dLdBp[d][r]*dImFdx[1]);

	      dLdx[s][2]                  += (dLdAp[d][r]*dReFdx[2] +
					      dLdBp[d][r]*dImFdx[2]);

	      if (mdl.atom[a].isotropic)
		dLdbiso[a]                -= mdl.atom[a].occ*stolsq*(dLdAp[d][r]*dReFpdo+
								     dLdBp[d][r]*dImFpdo+
 								     dLdsigmah[d][sh]*
								     temp);
	      else
		for (unsigned u            = 0; u < 6; u++)
		  dLduaniso[a][u]         -= (TWOPI2*mdl.atom[a].occ*coefbano[u]*
					      (dLdAp[d][r]*dReFpdo +
					       dLdBp[d][r]*dImFpdo +
					       dLdsigmah[d][sh]*temp));
	      
	      if (xtal.sf[d].anomalous)
	      {
		double amult(TWO*bfactor*btimeso*
			     ((double)xtal.sg.NSYMP)/
			     ((double)(xtal.sf[d].anonshl[sh])));

		double fpp(mdl.form[f].fpp[w]);
		double dReFmdo(scat*costotal + fpp*sintotal);
		double dImFmdo(scat*sintotal - fpp*costotal);

		double atemp(fpp*fpp*mult);

		dLdocc[a]                 += (dLdAm[d][r]*dReFmdo +
					      dLdBm[d][r]*dImFmdo +
					      fpp*fpp*dLdsumfpp[d][sh]*amult +
 					      dLdsigmah[d][sh]*atemp);	
	  
		dLdx[s][0]                += (dLdAm[d][r]*(scat*dcosdx[0]+
							   fpp*dsindx[0])+
					      dLdBm[d][r]*(scat*dsindx[0]-
							   fpp*dcosdx[0]));

		dLdx[s][1]                += (dLdAm[d][r]*(scat*dcosdx[1]+
							   fpp*dsindx[1])+
					      dLdBm[d][r]*(scat*dsindx[1]-
							   fpp*dcosdx[1]));

		dLdx[s][2]                += (dLdAm[d][r]*(scat*dcosdx[2]+
							   fpp*dsindx[2])+
					      dLdBm[d][r]*(scat*dsindx[2]-
							   fpp*dcosdx[2]));

		if (mdl.atom[a].isotropic)
		  dLdbiso[a]              -= (mdl.atom[a].occ*stolsq*
					      (dLdAm[d][r]*dReFmdo +
					       dLdBm[d][r]*dImFmdo +
					       fpp*fpp*dLdsumfpp[d][sh]*amult +
					       dLdsigmah[d][sh]*atemp));
		else
		  for (unsigned u          = 0; u < 6; u++)
		    dLduaniso[a][u]       -= (TWOPI2*mdl.atom[a].occ*coefbano[u]*
					      (dLdAm[d][r]*dReFmdo +
					       dLdBm[d][r]*dImFmdo +
					       fpp*fpp*dLdsumfpp[d][sh]*amult +
					       dLdsigmah[d][sh]*atemp));
		}
	    }
	  }
	}
      }

      if (!deriv) // store amplitude and phase
	for (unsigned d                    = 0; d < xtal.sf.size(); d++)
	{
	  double atemp(acalc[d] + anoacalc[d]), btemp(bcalc[d] + anobcalc[d]);
      
	  xtal.sf[d].fcalcp[r]             = sqrt(atemp*atemp + btemp*btemp);
	  xtal.sf[d].pcalcp[r]             = atan2(btemp, atemp);

	  if (xtal.sf[d].anomalous)
	  {
	    atemp                          = acalc[d] - anoacalc[d];
	    btemp                          = bcalc[d] - anobcalc[d];
	    xtal.sf[d].fcalcm[r]           = sqrt(atemp*atemp + btemp*btemp);
	    xtal.sf[d].pcalcm[r]           = atan2(btemp, atemp);
	  }
	}
    }

  if (!deriv)
    for (unsigned d                        = 0; d < xtal.sf.size()   ; d++)
      for (unsigned s                      = 0; s < xtal.sf[d].nbins ; s++)
      {
	if (xtal.sf[d].nshl[s]            && calcsigmah)
	  xtal.sf[d].sigmah[s]            *= ((double)(xtal.sg.NSYMP)/
					      ((double)(xtal.sf[d].nshl[s])));

	if (xtal.sf[d].anomalous          && calcsigmah &&
	    ( (target                     == "SAD") || (target == "MAD") ) )
	  if (xtal.sf[d].anonshl[s])
	    for (unsigned d2               = 0; d2 < xtal.sf.size()  ; d2++)
	    {
	      sumfpp[d][d2][s]            *= ((double)(xtal.sg.NSYMP)/
					      ((double)xtal.sf[d].anonshl[s]));
	      difffpp[d][d2][s]           *= ((double)(xtal.sg.NSYMP)/
					      ((double)xtal.sf[d].anonshl[s]));
	      difffp[d][d2][s]            *= ((double)(xtal.sg.NSYMP)/
					      ((double)xtal.sf[d].anonshl[s]));
	    }
      }    
} 	

void Likelihood::shift(const double stepsize, const vector<double> &pars,
	 	       const vector<double> &direction)
{
  int i                   = -1;
  
  // Apply direction and step size to the parameters

  for (unsigned d         = 0; d < xtal.sf.size()          ; d++)
    if (target           == "UNCO")
    {

      if (xtal.sf[d].refinek && refinescale)
      {
        ++i; xtal.sf[d].Setkscale(pars[i] + xsshift(stepsize*direction[i], "KSCALE"));
      }
      if (xtal.sf[d].refineb && refinescale)
      {
	++i; xtal.sf[d].Setbiso(pars[i] + xsshift(stepsize*direction[i], "BSCALE"));
      }
      
      if (xtal.sf[d].refined  && refineluzzati)
	for (unsigned s   = 0; s < xtal.sf[d].nbins ; s++)
	  {
	    ++i; xtal.sf[d].Setdluz(s,pars[i] + xsshift(stepsize*direction[i], "LUZZ"));
	  }

      if (xtal.sf[d].refinee  && refineluzzati)
	for (unsigned s   = 0; s < xtal.sf[d].nbins ; s++)
	{
	  ++i; xtal.sf[d].Seteluz(s,pars[i] + xsshift(stepsize*direction[i], "LUZZ"));
	}

      if (xtal.sf[d].refinead && refineluzzati)
	for (unsigned s   = 0; s < xtal.sf[d].nbins ; s++)
	{
	  ++i; xtal.sf[d].Setadluz(s,pars[i] + xsshift(stepsize*direction[i], "LUZZ"));
	}
      }
    else if ( (target    == "SAD") || (target == "MAD") )
	if (xtal.sf[d].refinesd & refineluzzati)
	  for (unsigned s = 0; s < xtal.sf[d].nbins ; s++)
	  {
	    ++i; xtal.sf[d].Setsdluz(s,pars[i] + xsshift(stepsize*direction[i], "LUZZ"));
	  }
  
  if (refinexyz)
    for (unsigned s       = 0; s < mdl.site.size()  ; s++)
      for (unsigned k     = 0; k < 3                ; k++)
	if (mdl.site[s].Getrefinex(k))
	{
	  ++i; mdl.site[s].Setx(k, pars[i] + xsshift(stepsize*direction[i], "XYZ"));
	}
    
  for (unsigned a         = 0; a < mdl.atom.size()  ; a++)
  {
    if (mdl.atom[a].Getrefinebfac() && refinebfac)
      if (mdl.atom[a].Getisotropic() )
      {
	++i; mdl.atom[a].Setbiso(pars[i] + xsshift(stepsize*direction[i], "BFAC"));
      }
      else
	for (unsigned k   = 0; k < 6                ; k++)
	{
	  ++i; mdl.atom[a].Setuaniso(pars[i] + xsshift(stepsize*direction[i], "UANO"), k);  
	}
      
    if (mdl.atom[a].Getrefineocc() && refineocc)
    {
      ++i; mdl.atom[a].Setocc(pars[i] + xsshift(stepsize*direction[i], "OCCU"));
    }

    if (mdl.atom[a].Getrefinenumb() && refinenumb)
    {
      ++i; mdl.atom[a].Setnumb(pars[i] + xsshift(stepsize*direction[i], "NUMB"));
    }

  }

  for (unsigned f         = 0; f < mdl.form.size()       ; f++)
    for (unsigned w       = 0; w < mdl.form[f].Getnwave(); w++)
    {
      if (mdl.form[f].Getrefinefp(w) && refinefp)
      {
	++i; mdl.form[f].Setfp(w, pars[i] + xsshift(stepsize*direction[i], "FP"));
      }
	
      if (mdl.form[f].Getrefinefpp(w) && refinefpp)
      {
	++i; mdl.form[f].Setfpp(w, pars[i] + xsshift(stepsize*direction[i], "FPP"));
      }
    }
}

double Likelihood::xsshift(const double &shift, const string &type) const
{
  // dampen shifts - if required
  
  if (type     == "OCCU")
    return ( (fabs(shift) > maxoccshift)    ? sign(shift)*maxoccshift    : shift);
  else if (type == "NUMB")
    return ( (fabs(shift) > maxnumbshift)   ? sign(shift)*maxnumbshift   : shift);
  else if (type == "LUZZ")
    return ( (fabs(shift) > maxluzzshift)   ? sign(shift)*maxluzzshift   : shift);
  else if (type == "BFAC")
    return ( (fabs(shift) > maxbfacshift)   ? sign(shift)*maxbfacshift   : shift);
  else if (type == "XYZ")
    return ( (fabs(shift) > maxxyzshift)    ? ZERO                       : shift);
  else if (type == "UANO")
    return ( (fabs(shift) > maxuanoshift)   ? sign(shift)*maxuanoshift   : shift);
  else if (type == "FP")
    return ( (fabs(shift) > maxfpshift)     ? sign(shift)*maxfpshift     : shift);
  else if (type == "FPP")
    return ( (fabs(shift) > maxfppshift)    ? sign(shift)*maxfppshift    : shift);
  else if (type == "KSCALE")
    return ( (fabs(shift) > maxkscaleshift) ? sign(shift)*maxkscaleshift : shift);
  else if (type == "BSCALE")
    return ( (fabs(shift) > maxbscaleshift) ? sign(shift)*maxbscaleshift : shift);
  else
    Bp3Error("Bp3likelihood::xsshift", "Type unknown");
  
  return ZERO;
}

vector<double> Likelihood::Getgradient() const
{
  // place gradients in a vector

  vector<double> grad;

  for (unsigned d           = 0; d < xtal.sf.size()        ; d++)
    if (target             == "UNCO")
    {
       if (xtal.sf[d].refinek  && refinescale)
          grad.push_back(dLdkscale[d]);
        if (xtal.sf[d].refineb && refinescale)
          grad.push_back(dLdbscale[d]);
	
      if (xtal.sf[d].refined  && refineluzzati)
	for (unsigned s     = 0; s < xtal.sf[d].nbins      ; s++)
	  grad.push_back(dLddluz[d][s]);
      if (xtal.sf[d].refinee  && refineluzzati)
	for (unsigned s     = 0; s < xtal.sf[d].nbins      ; s++)
	  grad.push_back(dLdeluz[d][s]);
      if (xtal.sf[d].refinead && refineluzzati)
	for (unsigned s     = 0; s < xtal.sf[d].nbins      ; s++)
	  grad.push_back(dLdadluz[d][s]);
    }
    else if ( (target      == "SAD") || (target == "MAD") )
	if (xtal.sf[d].refinesd && refineluzzati)
	  for (unsigned s   = 0; s < xtal.sf[d].nbins      ; s++)
	    grad.push_back(dLdsdluz[d][s]);
  
  if (refinexyz)
    for (unsigned s         = 0; s < mdl.site.size()       ; s++)
      for (unsigned k       = 0; k < 3                     ; k++)
	if (mdl.site[s].Getrefinex(k))
	  grad.push_back(dLdx[s][k]);
    
    for (unsigned a         = 0; a < mdl.atom.size()       ; a++)
    {
      if (mdl.atom[a].Getrefinebfac() && refinebfac)
	if (mdl.atom[a].Getisotropic() )
	  grad.push_back(dLdbiso[a]);
	else
	{
	  // ***NSP - check to see if anisotropic matrix is positive definite 
	  if (!mdl.atom[a].posdefu() )
	    printf("Atom %u does not have a positive definite anisotropic matrix\n",a+1);
	
	  for (int b        = 0; b < 6                     ; b++)
	    grad.push_back(dLduaniso[a][b]);
	}
      
      if (mdl.atom[a].Getrefineocc() && refineocc)
	grad.push_back(dLdocc[a]);

      if (mdl.atom[a].Getrefinenumb() && refinenumb)
	grad.push_back(dLdnumb[a]);

    }
 
    for (unsigned f         = 0; f < mdl.form.size()       ; f++)
      for (unsigned w       = 0; w < mdl.form[f].Getnwave(); w++)
      {
	if (mdl.form[f].Getrefinefp(w)  && refinefp)
	  grad.push_back(dLdfp[f][w]);
	if (mdl.form[f].Getrefinefpp(w) && refinefpp)
	  grad.push_back(dLdfpp[f][w]);
      }

  return grad;
}

double Likelihood::Getstepmax(const vector<double> &direction) const
{
  double step(ZERO);
  for (unsigned i          = 0; i < direction.size(); i++)
    if (fabs(direction[i]) > EPSILON)
    {
      if (type[i]         == "LUZZ")
	step               = std::max(step, fabs(maxluzzshift/direction[i]));
      if (type[i]         == "XYZ")
	step               = std::max(step, fabs(maxxyzshift/direction[i]));
      /*
      if (type[i]         == "BFAC")
	step               = std::max(step, fabs(maxbfacshift/direction[i]));
      */
      if (type[i]         == "OCCU")
	step               = std::max(step, fabs(maxoccshift/direction[i]));
      if (type[i]         == "NUMB")
	step               = std::max(step, fabs(maxnumbshift/direction[i]));
      if (type[i]         == "KSCALE")
	step               = std::max(step, fabs(maxkscaleshift/direction[i]));
      if (type[i]         == "BSCALE")
	step               = std::max(step, fabs(maxbscaleshift/direction[i]));
    }
  
  step                     = between(ZERO,step,ONE);

  return step;
}

void Likelihood::Settype()
{
  // place type of parameter in vector

  for (unsigned d           = 0; d < xtal.sf.size()        ; d++)
    if (target             == "UNCO")
    {
      if (xtal.sf[d].refinek && refinescale)
	type.push_back("KSCALE");
      if (xtal.sf[d].refineb && refinescale)
	type.push_back("BSCALE");
      
      if (xtal.sf[d].refined && refineluzzati)
	for (unsigned s     = 0; s < xtal.sf[d].nbins      ; s++)
	  type.push_back("LUZZ");
      if (xtal.sf[d].refinee && refineluzzati)
	for (unsigned s     = 0; s < xtal.sf[d].nbins      ; s++)
	  type.push_back("LUZZ");
      if (xtal.sf[d].refinead && refineluzzati)
	for (unsigned s     = 0; s < xtal.sf[d].nbins      ; s++)
	  type.push_back("BFAC");  // ***NSP - so it doesn't effect step size!!
    }
    else if ( (target      == "SAD") || (target == "MAD") )
      if (xtal.sf[d].refinesd && refineluzzati)
	for (unsigned s     = 0; s < xtal.sf[d].nbins      ; s++)
	  type.push_back("LUZZ");
  
  if (refinexyz)
    for (unsigned s         = 0; s < mdl.site.size()       ; s++)
      for (unsigned k       = 0; k < 3                     ; k++)
	if (mdl.site[s].Getrefinex(k))
 	  type.push_back("XYZ");
    
    for (unsigned a         = 0; a < mdl.atom.size()       ; a++)
    {
      if (mdl.atom[a].Getrefinebfac() && refinebfac)
	if (mdl.atom[a].Getisotropic() )
	  type.push_back("BFAC");
	else
	{
	  // ***NSP - check to see if anisotropic matrix is positive definite 
	  if (!mdl.atom[a].posdefu() )
	  {
	    printf("Atom %u", a+1);
	    printf(" does not have a positive definite anisotropic matrix\n");
	  }
	
	  for (int b        = 0; b < 6                     ; b++)
	    type.push_back("UANO");
	}
      
      if (mdl.atom[a].Getrefineocc() && refineocc)
	type.push_back("OCCU");

      if (mdl.atom[a].Getrefinenumb() && refinenumb)
	type.push_back("NUMB");
    }
 
    for (unsigned f         = 0; f < mdl.form.size()       ; f++)
      for (unsigned w       = 0; w < mdl.form[f].Getnwave(); w++)
      {
	if (mdl.form[f].Getrefinefp(w)  && refinefp)
	  type.push_back("FP");
	if (mdl.form[f].Getrefinefpp(w) && refinefpp)
 	  type.push_back("FPP");
      }
}

vector<double> Likelihood::Setpars() const
{
  // Place original/refineable parameters in array pars

  vector<double> pars;
  
  for (unsigned d           = 0; d < xtal.sf.size()     ; d++)
    if (target             == "UNCO")
    {
      if (xtal.sf[d].refinek && refinescale)
	pars.push_back(xtal.sf[d].Getkscale());

      if (xtal.sf[d].refineb && refinescale)
	pars.push_back(xtal.sf[d].Getbiso());
      
      if (xtal.sf[d].refined  && refineluzzati)
	for (unsigned s     = 0; s < xtal.sf[d].nbins   ; s++)
	  pars.push_back(xtal.sf[d].Getdluz(s));
      
      if (xtal.sf[d].refinee  && refineluzzati)
	for (unsigned s     = 0; s < xtal.sf[d].nbins; s++)
	  pars.push_back(xtal.sf[d].Geteluz(s));

      if (xtal.sf[d].refinead && refineluzzati)
	for (unsigned s     = 0; s < xtal.sf[d].nbins   ; s++)
	  pars.push_back(xtal.sf[d].Getadluz(s));
    }
    else if ( (target      == "SAD") || (target == "MAD") )
      if (xtal.sf[d].refinesd && refineluzzati)
	for (unsigned s     = 0; s < xtal.sf[d].nbins   ; s++)
	  pars.push_back(xtal.sf[d].Getsdluz(s));
  
  if (refinexyz)
    for (unsigned s         = 0; s < mdl.site.size()    ; s++)
      for (unsigned k       = 0; k < 3                  ; k++)
	if (mdl.site[s].Getrefinex(k))
	  pars.push_back(mdl.site[s].Getx(k));
    
    for (unsigned a         = 0; a < mdl.atom.size()    ; a++)
    {
      if (mdl.atom[a].Getrefinebfac() && refinebfac)
	if (mdl.atom[a].Getisotropic() )
	  pars.push_back(mdl.atom[a].Getbiso());
	else
	  for (unsigned b   = 0; b < 6               ; b++)
	    pars.push_back(mdl.atom[a].Getuaniso(b));

      if (mdl.atom[a].Getrefineocc() && refineocc)
	pars.push_back(mdl.atom[a].Getocc());      

      if (mdl.atom[a].Getrefinenumb() && refinenumb)
	pars.push_back(mdl.atom[a].Getnumb());      
    }

    for (unsigned f         = 0; f < mdl.form.size()    ; f++)
      for (unsigned w       = 0; w < mdl.form[f].Getnwave(); w++)
      {
	if (mdl.form[f].Getrefinefp(w)  && refinefp)
	  pars.push_back(mdl.form[f].Getfp(w));
	if (mdl.form[f].Getrefinefpp(w) && refinefpp)
	  pars.push_back(mdl.form[f].Getfpp(w));
      }
  return pars;
}

void Likelihood::writepdb() const
{
  for (unsigned c                      = 0; c < xtal.cell.size(); c++)
  {
    bool atomincrystal(false);
    for (unsigned a                    = 0; a < mdl.atom.size(); a++)
      if (c                           == mdl.atom[a].crystal)
	atomincrystal                  = true;

    if (atomincrystal)
    {
      char pdbnumber[3];
    
      if (c                            < 9)
	sprintf(pdbnumber,"%1u",c+1);
      else
	sprintf(pdbnumber,"%2u",c+1);

      string pdbname                   = output + pdbnumber + ".pdb";

      FILE *pdbfile                    = fopen(pdbname.c_str(),"w");

      fprintf(pdbfile,"CRYST1%9.3f%9.3f%9.3f%7.2f%7.2f%7.2f %10s\n",
	      xtal.cell[c].a, xtal.cell[c].b, xtal.cell[c].c,
	      xtal.cell[c].alpha, xtal.cell[c].beta, xtal.cell[c].gamma,
	      xtal.sg.Gethmname().c_str());
      fprintf(pdbfile,"SCALE1    %10.6f%10.6f%10.6f     %10.5f\n",
	      xtal.cell[c].or2frac[0][0], xtal.cell[c].or2frac[0][1],
	      xtal.cell[c].or2frac[0][2],0.0);
      fprintf(pdbfile,"SCALE2    %10.6f%10.6f%10.6f     %10.5f\n",
	      xtal.cell[c].or2frac[1][0], xtal.cell[c].or2frac[1][1],
	      xtal.cell[c].or2frac[1][2],0.0);
      fprintf(pdbfile,"SCALE3    %10.6f%10.6f%10.6f     %10.5f\n",
	      xtal.cell[c].or2frac[2][0], xtal.cell[c].or2frac[2][1],
	      xtal.cell[c].or2frac[2][2],0.0);

      for (unsigned a                  = 0; a < mdl.atom.size() ; a++)
	if (c                         == mdl.atom[a].crystal)
	{
	  vector<double> ortho = mdl.site[mdl.atom[a].nsite].ortho(xtal.cell[c].frac2or);
	  if (mdl.atom[a].name.size() == 1)
	    fprintf(pdbfile,"HETATM %4d  %1s   %2s    %3d    %8.3f%8.3f%8.3f%6.2f%6.2f\n",
		    a+1,mdl.atom[a].name.substr(0,1).c_str(),
		    mdl.atom[a].name.substr(0,2).c_str(),a+1,
		    ortho[0],ortho[1],ortho[2],mdl.atom[a].occ,mdl.atom[a].biso);
	  else
	    fprintf(pdbfile,"HETATM %4d  %2s  %2s    %3d    %8.3f%8.3f%8.3f%6.2f%6.2f\n",
		    a+1,mdl.atom[a].name.substr(0,2).c_str(),
		    mdl.atom[a].name.substr(0,2).c_str(),a+1,
		    ortho[0],ortho[1],ortho[2],mdl.atom[a].occ,mdl.atom[a].biso);
	}
      fprintf(pdbfile,"END\n");
      fclose(pdbfile);
    }
  }
}

void Likelihood::writexml() const
{
   string xmlname  = output + ".xml";

   FILE *xmlfile   = fopen(xmlname.c_str(),"w");

   fprintf(xmlfile, "<coordinate_data>\n");
   for (unsigned a = 0; a < mdl.atom.size() ; a++)
   {
     fprintf(xmlfile, "\t<atom id=\"%d\">\n",a+1);
     fprintf(xmlfile, "\t\t<atom_id>%s%d</atom_id>\n",
         mdl.atom[a].name.c_str(),a+1);
     fprintf(xmlfile, "\t\t<element_name>%s</element_name>\n",
         mdl.atom[a].name.c_str());
     unsigned s  = mdl.atom[a].nsite;
     fprintf(xmlfile, "\t\t<x>%7.4f</x>\n",
         mdl.site[s].x[0]);
     fprintf(xmlfile, "\t\t<y>%7.4f</y>\n",
         mdl.site[s].x[1]);
     fprintf(xmlfile, "\t\t<z>%7.4f</z>\n",
         mdl.site[s].x[2]);
     fprintf(xmlfile, "\t\t<occupancy>%7.3f</occupancy>\n",
         mdl.atom[a].occ);
     fprintf(xmlfile, "\t\t<bfactor>%7.3f</bfactor>\n",
         mdl.atom[a].biso);
     fprintf(xmlfile, "\t</atom>\n");
   }
   fprintf(xmlfile, "</coordinate_data>\n");
   fprintf(xmlfile, "\n\n");

   fclose(xmlfile);
}

bool Likelihood::hermitianinverse(Matrix &reori, Matrix &imori,
                                  Matrix &reinv, Matrix &iminv, double &det)
{
  // pseudoinverse for a hermitian matrix
  vector<std::complex<double> > cevectors(reori.size());
  evalues.resize(reori.rsize());

  for (unsigned i        = 0; i < reori.size(); i++)
    cevectors[i]         = std::complex<double> (reori.vec(i), imori.vec(i));

  char jobz('V'), uplo('U');
  int n(reori.rsize()), lda(reori.rsize());
  int clwork(n*2 + n*n), clrwork(1 + 5*n + 2*n*n), cliwork(3 + 5*n);
  vector<std::complex<double> > cwork(clwork);
  vector<double> crwork(clrwork, ZERO);
  vector<int> ciwork(cliwork, 0);

  FORTRAN_CALL ( ZHEEVD, zheevd,
                 (&jobz, &uplo, &n, &cevectors[0], &lda, &evalues[0], &cwork[0],
                  &clwork, &crwork[0], &clrwork,  &ciwork[0], &cliwork, &info),
                 (&jobz, &uplo, &n, &cevectors[0], &lda, &evalues[0], &cwork[0],
                  &clwork, &crwork[0], &clrwork,  &ciwork[0], &cliwork, &info),
                 (&jobz, &uplo, &n, &cevectors[0], &lda, &evalues[0], &cwork[0],
                  &clwork, &crwork[0], &clrwork,  &ciwork[0], &cliwork, &info));

  for (unsigned i        = 0; i < reori.rsize(); i++)
    for (unsigned j      = i; j < reori.csize(); j++)
    {
      double recomp(ZERO), recompinv(ZERO);
      double imcomp(ZERO), imcompinv(ZERO);
      for (unsigned k    = 0; k < reori.rsize(); k++)
        if (evalues[k]   > 0.1)
        {
          double temp    = (std::real(cevectors[i+reori.rsize()*k])*
                            std::real(cevectors[j+reori.rsize()*k]) +
                            std::imag(cevectors[i+reori.rsize()*k])*
                            std::imag(cevectors[j+reori.rsize()*k]));
          recomp        += temp*evalues[k];
          recompinv     += temp/evalues[k];
          temp           = (std::real(cevectors[i+reori.rsize()*k])*
                            std::imag(cevectors[j+reori.rsize()*k]) -
                            std::imag(cevectors[i+reori.rsize()*k])*
                            std::real(cevectors[j+reori.rsize()*k]));
          imcomp        -= temp*evalues[k];
          imcompinv     -= temp/evalues[k];
        }
      reori(i,j)         = recomp;
      reori(j,i)         = recomp;
      reinv(i,j)         = recompinv;
      reinv(j,i)         = recompinv;
      imori(i,j)         = imcomp;
      imori(j,i)         = -imcomp;
      iminv(i,j)         = imcompinv;
      iminv(j,i)         = -imcompinv;
    }

  det                    = ONE;

  bool filter(false);

  for (unsigned i        = 0; i < reori.rsize(); i++)
  {
    if (evalues[i]       < 0.1)
      filter             = true;
    det                 *= (evalues[i] < 0.1) ? 0.1 : evalues[i];
  }
  return filter;
}

bool Likelihood::inverse(Matrix &ori, Matrix &inv, double &det)
{
  // pseudoinverse for real symmetric matrix
  evectors.resize(ori.size());
  evalues.resize(ori.rsize());
  
  for (unsigned i        = 0; i < evectors.size(); i++)
    evectors[i]          = ori.vec(i);
  
  char jobz('V'), uplo('U');
  int n(ori.rsize()), lda(ori.rsize());
  
  FORTRAN_CALL ( DSYEVD, dsyevd,
                 (&jobz, &uplo, &n, &evectors[0], &lda, &evalues[0], &lawork[0],
                  &lwork, &iwork[0], &liwork, &info),
                 (&jobz, &uplo, &n, &evectors[0], &lda, &evalues[0], &lawork[0], 
                  &lwork, &iwork[0], &liwork, &info),
                 (&jobz, &uplo, &n, &evectors[0], &lda, &evalues[0], &lawork[0], 
		  &lwork, &iwork[0], &liwork, &info));

  for (unsigned i        = 0; i < ori.rsize(); i++)
    for (unsigned j      = i; j < ori.csize(); j++)
    {
      double recomp(ZERO), recompinv(ZERO);
      for (unsigned k    = 0; k < ori.rsize(); k++)
	if (evalues[k]   > MINEIG)
	{
	  double temp    = evectors[i+ori.rsize()*k]*evectors[j+ori.rsize()*k];
	  recomp        += temp*evalues[k];
	  recompinv     += temp/evalues[k];
	}
      ori(i,j)           = recomp;
      ori(j,i)           = recomp;
      inv(i,j)           = recompinv;
      inv(j,i)           = recompinv;
    }

  det                    = ONE;

  bool filter(false);

  for (unsigned i        = 0; i < ori.rsize(); i++)
  {
    if (evalues[i]       < MINEIG)
      filter             = true;
    det                 *= (evalues[i] < MINEIG) ? MINEIG : evalues[i];
  }
  return filter;  
}

void Likelihood::inverse2by2(Matrix &ori, Matrix &inv, double &det)
{
  // ***NSP error when data[0] == data[3] !!
  // returns the positive semi-definite generalized inverse of a 2x2 Hermitian matrix
  // and corrects the original matrix to maintain positive definiteness

  // returns the positive semi-definite generalized inverse of a 2x2 Hermitian matrix
  // and corrects the original matrix to maintain positive definiteness

  double sumsquared  = ori(0,1)*ori(0,1);

  double temp        = sqrt(FOUR*sumsquared + pow(ori(0,0) - ori(1,1), 2));

  // ensure only positive eigenvalues
  double eigenvalue0 = std::max((ori(0,0) + ori(1,1) - temp)/TWO, MINEIG);
  double eigenvalue1 = std::max((ori(0,0) + ori(1,1) + temp)/TWO, MINEIG);

  det                = eigenvalue0*eigenvalue1;

  double mu0         = (ori(0,0) - ori(1,1) - temp)/TWO;
  double mu1         = (ori(0,0) - ori(1,1) + temp)/TWO;

  double arg0(ZERO), arg1(ZERO), invarg0(ZERO), invarg1(ZERO);

  double temp0       = mu0*mu0 + sumsquared;
  if ( (temp0        > EPSILON) && (eigenvalue0 > MINEIG) )
  {
    arg0             = ONE/temp0;
    invarg0          = arg0/eigenvalue0;
    arg0            *= eigenvalue0;
  }

  double temp1       = mu1*mu1 + sumsquared;
  if ( (temp1        > EPSILON) && (eigenvalue1 > MINEIG) )
  {
    arg1             = ONE/temp1;
    invarg1          = arg1/eigenvalue1;
    arg1            *= eigenvalue1;
  }
  
  inv(0,0)           = mu0*mu0*invarg0 + mu1*mu1*invarg1;
  inv(0,1)           = ori(0,1)*(mu0*invarg0 + mu1*invarg1);
  inv(1,0)           = inv(0,1);

  ori(0,0)           = mu0*mu0*arg0 + mu1*mu1*arg1;
  ori(0,1)          *= mu0*arg0 + mu1*arg1;
  ori(1,0)           = ori(0,1);

  if (sumsquared     < EPSILON)
    if (ori(1,1)     > EPSILON)
      inv(1,1)       = ONE/ori(1,1);
    else
    {
      inv(1,1)       = ZERO;
      ori(1,1)       = ZERO;
    }
  else
  {
    inv(1,1)         = sumsquared*(invarg0 + invarg1);
    ori(1,1)         = sumsquared*(arg0 + arg1);
  }  
}

void Likelihood::matrixprod(Matrix &reout, const Matrix &reain, const double *recin)
{
  // return reout = A*C*A for real symmetric square matrices A,C
  for (unsigned i              = 0; i < reain.rsize(); i++)
    for (unsigned j            = i; j < reain.csize(); j++)
    {
      double recomp(ZERO);
      for (unsigned l          = 0; l < reain.rsize(); l++)
	for (unsigned k        = 0; k < reain.csize(); k++)
 	  recomp              += reain(i,l)*recin[l+reain.rsize()*k]*reain(k,j);
      reout(i,j)               = recomp;
      reout(j,i)               = recomp;
    }
}

void Likelihood::storeinitialcolumns(CMtz::MTZ *MTZIN, vector<float> &fdata,
				     const unsigned inc, const unsigned r)
{
  // store initial columns needed for output mtz
  
  if (allin)
  {
    vector<int> flags(fdata.size());
    float res;
    CMtz::ccp4_lrreff(MTZIN, &res, &fdata[0], &flags[0], (const CMtz::MTZCOL**)
		      &colin[0], colin.size(), r+1);
    for (unsigned i           = 0; i < colin.size(); i++)
      if (flags[i])
	fdata[i]              = CCP4::ccp4_nan().f;
  }
  else
  {
    fdata[0]                  = (float) xtal.miller[r][0];
    fdata[1]                  = (float) xtal.miller[r][1];
    fdata[2]                  = (float) xtal.miller[r][2];
  }
  
  if (xtal.sf[0].use(r)      && (mode != "INTENSITY") && (mode != "DIFFE") && (mode != "GIAC") && (mode != "DELTA") && (mode != "RESCALE") )
  {
    if ( (target             == "SAD") && (xtal.sf[0].usep(r)) )
    {
      fdata[inc]              = (float) xtal.sf[0].datap[r];
      fdata[inc + 1]          = (float) xtal.sf[0].devp[r];
    }
    else
    {
      fdata[inc]              = (float) xtal.sf[0].datamean(r);
      fdata[inc + 1]          = (float) xtal.sf[0].devmean(r);
    }
  }
}

void Likelihood::setupmtz(CMtz::MTZ *MTZOUT, CMtz::MTZ *MTZIN)
{
  colin.resize(0);
  colout.resize(0);

  CMtz::ccp4_lwtitl(MTZOUT, title.c_str(), 0);
  MTZOUT->refs_in_memory             = 0;

  MTZOUT->mtzsymm.spcgrp             = xtal.sg.number;
  MTZOUT->mtzsymm.nsym               = xtal.sg.NSYM;
  MTZOUT->mtzsymm.nsymp              = xtal.sg.NSYMP;
  MTZOUT->mtzsymm.symtyp             = xtal.sg.lattice;
  CSym::ccp4spg_to_shortname(MTZOUT->mtzsymm.spcgrpname,
			     xtal.sg.Gethmname().substr(0,10).c_str());
  strncpy(MTZOUT->mtzsymm.pgname, xtal.sg.Getpointgroup().substr(0,10).c_str(), 11);
  for (unsigned s                    = 0; s < xtal.sg.NSYM; s++ )
  {
    for (unsigned i                  = 0; i < 3; i++ )
      for (unsigned j                = 0; j < 3; j++ )
	MTZOUT->mtzsymm.sym[s][i][j] = xtal.sg.symrot[s][j][i];
    for (unsigned i                  = 0; i < 3; i++ )
      MTZOUT->mtzsymm.sym[s][i][3]   = xtal.sg.symtran[s][i];
  }

  string mtzoutname(getenv("HKLOUT"));
  if (hand                           > 1)
    mtzoutname                      += "-oh";
  
  MTZOUT->fileout                    = CMtz::MtzOpenForWrite(mtzoutname.c_str());

  if (allin) // ***NSP - this may have to change for the hand problem!!!
    for (int x                       = 0; x < CMtz::MtzNxtal(MTZIN); x++) 
    {
      CMtz::MTZXTAL *xtl             = CMtz::MtzIxtal(MTZIN,x);
      CMtz::MTZXTAL *xo              = CMtz::MtzAddXtal(MTZOUT, xtl->xname, xtl->pname,
						        xtl->cell);
      for (int s                     = 0; s < CMtz::MtzNsetsInXtal(xtl); s++) 
      {
	CMtz::MTZSET *set            = CMtz::MtzIsetInXtal(xtl,s);
	CMtz::MTZSET *so             = CMtz::MtzAddDataset(MTZOUT, xo, set->dname,
						           set->wavelength);
	for (int c                   = 0; c < CMtz::MtzNcolsInSet(set); c++)
	{
	  colin.push_back(CMtz::MtzIcolInSet(set,c));
	  colout.push_back(CMtz::MtzAddColumn(MTZOUT, so,
					      CMtz::MtzIcolInSet(set,c)->label,
					      CMtz::MtzIcolInSet(set,c)->type));
	}
      }
    }

  if (mode                          == "INTENSITY")
  {
    for (unsigned d                  = 0; d < xtal.sf.size(); d++)
    {
      // ***NSP
      string junk                    = "BP3I";
      
      CMtz::MTZXTAL *xtl             = CMtz::MtzAddXtal(MTZOUT, junk.c_str(),
							junk.c_str(),
							&xtal.Getcell(d)[0]);

      CMtz::MTZSET *set              = CMtz::MtzAddDataset(MTZOUT, xtl,
			 			           xtl->xname,
							   0.0);      

      if (!allin &&              (d == 0))
      {
	// h column setup
	colout.push_back(CMtz::MtzAddColumn(MTZOUT, set, "H", "H"));
	// k column setup
	colout.push_back(CMtz::MtzAddColumn(MTZOUT, set, "K", "H"));
	// l column setup
	colout.push_back(CMtz::MtzAddColumn(MTZOUT, set, "L", "H"));
      }

      char datanumber[3];
      sprintf(datanumber,"%1d",d+1);
      string i("I"), sigi("SIGI");

      if (xtal.sf[d].anomalous)
      {
	string label                 = i    + datanumber + "(+)";
	colout.push_back(CMtz::MtzAddColumn(MTZOUT, set, label.c_str(), "K"));
	label                        = sigi + datanumber + "(+)";
	colout.push_back(CMtz::MtzAddColumn(MTZOUT, set, label.c_str(), "M"));
	label                        = i    + datanumber + "(-)";
	colout.push_back(CMtz::MtzAddColumn(MTZOUT, set, label.c_str(), "K"));
	label                        = sigi + datanumber + "(-)";
	colout.push_back(CMtz::MtzAddColumn(MTZOUT, set, label.c_str(), "M"));
      }
      else
      {
	string label                 = i    + datanumber;
 	colout.push_back(CMtz::MtzAddColumn(MTZOUT, set, label.c_str(), "J"));
	label                        = sigi + datanumber;
	colout.push_back(CMtz::MtzAddColumn(MTZOUT, set, label.c_str(), "Q"));
      }
    }
  }
  else if (mode                     == "EVALUES")
  {
    for (unsigned d                  = 0; d < xtal.sf.size(); d++)
    {
      // ***NSP
      string junk                    = "BP3I";
      
      CMtz::MTZXTAL *xtl             = CMtz::MtzAddXtal(MTZOUT, junk.c_str(),
							junk.c_str(),
							&xtal.Getcell(d)[0]);

      CMtz::MTZSET *set              = CMtz::MtzAddDataset(MTZOUT, xtl,
			 			           xtl->xname,
							   0.0);      

      if (!allin &&              (d == 0))
      {
	// h column setup
	colout.push_back(CMtz::MtzAddColumn(MTZOUT, set, "H", "H"));
	// k column setup
	colout.push_back(CMtz::MtzAddColumn(MTZOUT, set, "K", "H"));
	// l column setup
	colout.push_back(CMtz::MtzAddColumn(MTZOUT, set, "L", "H"));
      }

      char datanumber[3];
      sprintf(datanumber,"%1d",d+1);
      string f("F"), sigf("SIGF");
      string e("E"), sige("SIGE");

      if (xtal.sf[d].anomalous)
      {
	string label                 = f    + datanumber + "(+)";
	colout.push_back(CMtz::MtzAddColumn(MTZOUT, set, label.c_str(), "G"));
	label                        = sigf + datanumber + "(+)";
	colout.push_back(CMtz::MtzAddColumn(MTZOUT, set, label.c_str(), "L"));
	label                        = f    + datanumber + "(-)";
	colout.push_back(CMtz::MtzAddColumn(MTZOUT, set, label.c_str(), "G"));
	label                        = sigf + datanumber + "(-)";
	colout.push_back(CMtz::MtzAddColumn(MTZOUT, set, label.c_str(), "L"));
	label                        = e    + datanumber + "(+)";
	colout.push_back(CMtz::MtzAddColumn(MTZOUT, set, label.c_str(), "R"));
	label                        = sige + datanumber + "(+)";
	colout.push_back(CMtz::MtzAddColumn(MTZOUT, set, label.c_str(), "Q"));
	label                        = e    + datanumber + "(-)";
	colout.push_back(CMtz::MtzAddColumn(MTZOUT, set, label.c_str(), "R"));
	label                        = sige + datanumber + "(-)";
	colout.push_back(CMtz::MtzAddColumn(MTZOUT, set, label.c_str(), "Q"));
      }
    }
  }
  else if (mode                     == "RESCALE")
  {
    for (unsigned d                  = 0; d < xtal.sf.size(); d++)
    {
      // ***NSP
      string junk                    = "RESCALE";
      
      CMtz::MTZXTAL *xtl             = CMtz::MtzAddXtal(MTZOUT, junk.c_str(),
				 			junk.c_str(),
							&xtal.Getcell(d)[0]);

      CMtz::MTZSET *set              = CMtz::MtzAddDataset(MTZOUT, xtl,
			 	 	 	           xtl->xname,
							   0.0);
	  
      if (!allin &&              (d == 0))
      {
	// h column setup
	colout.push_back(CMtz::MtzAddColumn(MTZOUT, set, "H", "H"));
	// k column setup
	colout.push_back(CMtz::MtzAddColumn(MTZOUT, set, "K", "H"));
	// l column setup
	colout.push_back(CMtz::MtzAddColumn(MTZOUT, set, "L", "H"));
      }

      string scoldatap("G"), scoldevp("L"), scoldata("F"), scoldev("Q");
      if (xtal.sf[d].type           == "INTENSITY")
      {
	scoldatap                    = "K";
	scoldevp                     = "M";
	scoldata                     = "J";
	scoldev                      = "Q";	    
      }

      colout.push_back(CMtz::MtzAddColumn(MTZOUT, set, xtal.sf[d].sdata.c_str(), scoldata.c_str()));
      colout.push_back(CMtz::MtzAddColumn(MTZOUT, set, xtal.sf[d].sdev.c_str(), scoldev.c_str()));

      if (xtal.sf[d].anomalous)
      {
	colout.push_back(CMtz::MtzAddColumn(MTZOUT, set, xtal.sf[d].sdatap.c_str(), scoldatap.c_str()));
	colout.push_back(CMtz::MtzAddColumn(MTZOUT, set, xtal.sf[d].sdevp.c_str(), scoldevp.c_str()));
	colout.push_back(CMtz::MtzAddColumn(MTZOUT, set, xtal.sf[d].sdatam.c_str(), scoldatap.c_str()));
	colout.push_back(CMtz::MtzAddColumn(MTZOUT, set, xtal.sf[d].sdevm.c_str(), scoldevp.c_str()));
      }
    }
  }
  else if ( (mode                   == "DIFFE") || (mode == "GIAC") || (mode == "DELTA") )
  {
    unsigned d(0);

    // ***NSP
    string junk                      = "AFRO";
      
    CMtz::MTZXTAL *xtl               = CMtz::MtzAddXtal(MTZOUT, junk.c_str(),
							junk.c_str(),
							&xtal.Getcell(d)[0]);

    CMtz::MTZSET *set                = CMtz::MtzAddDataset(MTZOUT, xtl,
			 		 	           xtl->xname,
							   ZERO);
    if (!allin)
    {
      // h column setup
      colout.push_back(CMtz::MtzAddColumn(MTZOUT, set, "H", "H"));
      // k column setup
      colout.push_back(CMtz::MtzAddColumn(MTZOUT, set, "K", "H"));
      // l column setup
      colout.push_back(CMtz::MtzAddColumn(MTZOUT, set, "L", "H"));
    }

    string f("FA"), sigf("SIGFA");
    string e("EA"), sige("SIGEA");

    colout.push_back(CMtz::MtzAddColumn(MTZOUT, set, f.c_str(), "F"));
    colout.push_back(CMtz::MtzAddColumn(MTZOUT, set, sigf.c_str(), "Q"));
    colout.push_back(CMtz::MtzAddColumn(MTZOUT, set, e.c_str(), "R"));
    colout.push_back(CMtz::MtzAddColumn(MTZOUT, set, sige.c_str(), "Q"));
  }
  else
  {
    // ***NSP - I am not sure what I should call the new project/crystal/data set
    // that is to be created.

    string junk                    = "BP3-PHASED";
      
    CMtz::MTZXTAL *xtl             = CMtz::MtzAddXtal(MTZOUT,
						      junk.c_str(),
						      junk.c_str(),
						      &xtal.Getcell(0)[0]);

    CMtz::MTZSET *set              = CMtz::MtzAddDataset(MTZOUT, xtl,
							 xtl->xname,
							 ZERO);
    if (!allin)
    {
      // h column setup
      colout.push_back(CMtz::MtzAddColumn(MTZOUT, set, "H", "H"));
      // k column setup
      colout.push_back(CMtz::MtzAddColumn(MTZOUT, set, "K", "H"));
      // l column setup
      colout.push_back(CMtz::MtzAddColumn(MTZOUT, set, "L", "H"));
    }
  
    // observed amplitude column setup
    colout.push_back(CMtz::MtzAddColumn(MTZOUT, set, "FPHASED", "F"));
    // observed sigma column setup
    colout.push_back(CMtz::MtzAddColumn(MTZOUT, set, "SIGFPHASED", "Q"));
    // best amplitude column setup
    colout.push_back(CMtz::MtzAddColumn(MTZOUT, set, "FB", "F"));
    // best phase column setup
    colout.push_back(CMtz::MtzAddColumn(MTZOUT, set, "PHIB", "P"));
    // FOM column setup
    colout.push_back(CMtz::MtzAddColumn(MTZOUT, set, "FOM", "W"));
    // HLA column setup
    colout.push_back(CMtz::MtzAddColumn(MTZOUT, set, "HLA", "A"));
    // HLB column setup
    colout.push_back(CMtz::MtzAddColumn(MTZOUT, set, "HLB", "A"));
    // HLC column setup
    colout.push_back(CMtz::MtzAddColumn(MTZOUT, set, "HLC", "A"));
    // HLD column setup
    colout.push_back(CMtz::MtzAddColumn(MTZOUT, set, "HLD", "A"));
    // Difference MAP coefficient
    if (target                    == "SAD")
      colout.push_back(CMtz::MtzAddColumn(MTZOUT, set, "FDIFF", "F"));
  }
}

void Likelihood::outputintensities()
{
  double maxi(ZERO), maxs(ZERO);
  for (unsigned r          = 0; r < xtal.maxref; r++)
    for (unsigned d        = 0; d < xtal.sf.size(); d++)
      if (xtal.sf[d].use(r))
      {
	maxi               = std::max(maxi, (double)xtal.sf[d].datamean(r)*
				      xtal.sf[d].datamean(r));
	maxs               = std::max(maxs, (double)xtal.sf[d].devmean(r)*
				      sqrt(FOUR*xtal.sf[d].datamean(r)*
					   xtal.sf[d].datamean(r) +
					   TWO*xtal.sf[d].devmean(r)*
					   xtal.sf[d].devmean(r)));
      }

  double maxis             = std::max((double)maxi, (double)maxs);
  double scale             = (maxis >= 100000.0) ? 25000.0/maxis : ONE;
  
  CMtz::MTZ *MTZOUT        = CMtz::MtzMalloc(0,0);
  CMtz::MTZ *MTZIN         = CMtz::MtzGet("HKLIN",0);
  setupmtz(MTZOUT,MTZIN);

  unsigned inc             = (allin) ? colin.size() : 3;
  
  unsigned columns(0);
  for (unsigned d          = 0; d < xtal.sf.size(); d++)
    if (xtal.sf[d].anomalous)
      columns             += 4;
    else
      columns             += 2;

  for (unsigned r          = 0; r < xtal.maxref; r++)
  {    
    // Default value for columns to be written out is MNF
    vector<float> fdata(columns+inc, CCP4::ccp4_nan().f);
 
    storeinitialcolumns(MTZIN, fdata, inc, r);
    
    int c(-1);
    for (unsigned d        = 0; d < xtal.sf.size(); d++)
      if (xtal.sf[d].anomalous)
      {
	if (xtal.sf[d].usep(r))
	{
	  fdata[inc + ++c] = scale*xtal.sf[d].datap[r]*xtal.sf[d].datap[r];
	  fdata[inc + ++c] = (scale*xtal.sf[d].devp[r]*
			      sqrt(FOUR*xtal.sf[d].datap[r]*xtal.sf[d].datap[r] +
				   TWO*xtal.sf[d].devp[r]*xtal.sf[d].devp[r]));
	}
	else
	  c               += 2;
	
	if (xtal.sf[d].usem(r) && !xtal.centric[r])
	{
	  // ***NSP - CCP4 (truncate) default is to have F- = ABSENT for centric!
	  fdata[inc + ++c] = scale*xtal.sf[d].datam[r]*xtal.sf[d].datam[r];
	  fdata[inc + ++c] = (scale*xtal.sf[d].devm[r]*
			      sqrt(FOUR*xtal.sf[d].datam[r]*xtal.sf[d].datam[r] +
				   TWO*xtal.sf[d].devm[r]*xtal.sf[d].devm[r]));
	}
	else
	  c               += 2;
      }
      else
	if (xtal.sf[d].use(r))
	{
	  fdata[inc + ++c] = scale*xtal.sf[d].datamean(r)*xtal.sf[d].datamean(r);
	  fdata[inc + ++c] = (scale*xtal.sf[d].devmean(r)*
			      sqrt(FOUR*xtal.sf[d].datamean(r)*xtal.sf[d].datamean(r) +
				   TWO*xtal.sf[d].devmean(r)*
				   xtal.sf[d].devmean(r)));
	}
	else
	  c               += 2;
    CMtz::ccp4_lwrefl(MTZOUT, &fdata[0], &colout[0], fdata.size(), r+1);
  }

  if (allin)
    CMtz::MtzFree(MTZIN);
  CMtz::MtzPut(MTZOUT, " ");
  CMtz::MtzFree(MTZOUT);
}

void Likelihood::outputhklfile() const
{
  if (xtal.sf.size() > 1)
    Bp3Error("Likelihood::outputhklfile","can only handle one data set at the moment");
  
  double maxd(ZERO), maxs(ZERO);
  for (unsigned r          = 0; r < xtal.maxref; r++)
    for (unsigned d        = 0; d < xtal.sf.size(); d++)
      if (xtal.sf[d].use(r))
      {
	maxd               = std::max(maxd, (double)xtal.sf[d].datamean(r));
	maxs               = std::max(maxs, (double)xtal.sf[d].devmean(r));
      }

  double maxds             = std::max((double)maxd, (double)maxs);
  double scale             = (maxds >= 9999.9) ? 9999.9/maxds : ONE;

  FILE *dfile(NULL);
  string filename           = output + ".hkl";
  dfile                     = fopen(filename.c_str(),"w");
  /*
  fprintf(dfile,"TITLE SHELX format from BP3\n");
  fprintf(dfile, "CELL %7.5f%8.2f%8.2f%8.2f%8.2f%8.2f%8.2f\n",
	  1.54178,xtal.cell[0].a, xtal.cell[0].b, xtal.cell[0].c,
	  xtal.cell[0].alpha,xtal.cell[0].beta,xtal.cell[0].gamma);
  fprintf(dfile,"ZERR    1.00    0.001   0.001   0.001   0.00   0.00   0.00\n");
  fprintf(dfile,"LATT -7\n");
  fprintf("SYMM -X, Y, -Z\n");
  fprintf(dfile,"HKLF  3\n");
  */
  for (unsigned r          = 0; r < xtal.maxref; r++)
    for (unsigned d        = 0; d < xtal.sf.size(); d++)
      if (xtal.sf[d].use(r))
	if (xtal.sf[d].anomalous && !xtal.centric[r])
	{
	  if (xtal.sf[d].usep(r))
 	    fprintf(dfile, "%4d%4d%4d%8.2f%8.2f\n",
		    xtal.miller[r][0], xtal.miller[r][1], xtal.miller[r][2],
		    scale*xtal.sf[d].datap[r], scale*xtal.sf[d].devp[r]);
	  if (xtal.sf[d].usem(r))
	  {
	    int h = (xtal.miller[r][0] == 0) ? 0 : -xtal.miller[r][0];
	    int k = (xtal.miller[r][1] == 0) ? 0 : -xtal.miller[r][1];
	    int l = (xtal.miller[r][2] == 0) ? 0 : -xtal.miller[r][2];
	    
	    fprintf(dfile, "%4d%4d%4d%8.2f%8.2f\n",
		    h,k,l,scale*xtal.sf[d].datam[r], scale*xtal.sf[d].devm[r]);
	  }
	}
	else
	  fprintf(dfile, "%4d%4d%4d%8.2f%8.2f\n",
		  xtal.miller[r][0], xtal.miller[r][1], xtal.miller[r][2],
		  scale*xtal.sf[d].datap[r], scale*xtal.sf[d].devp[r]);

  fclose(dfile);
}

void Likelihood::outputscafile() const
{
  if (xtal.sf.size() > 1)
    Bp3Error("Likelihood::outputscafile","can only handle one data set at the moment");

  if (xtal.sf[0].type     != "INTENSITY")
    Bp3Error("Likelihood::outputscafile","Intensities must be input");
    
  double maxd(ZERO), maxs(ZERO);
  for (unsigned r          = 0; r < xtal.maxref; r++)
    for (unsigned d        = 0; d < xtal.sf.size(); d++)
      if (xtal.sf[d].use(r))
      {
	maxd               = std::max(maxd, (double)xtal.sf[d].datamean(r));
	maxs               = std::max(maxs, (double)xtal.sf[d].devmean(r));
      }

  double maxds             = std::max((double)maxd, (double)maxs);
  double scale             = (maxds >= 99999.9) ? 99999.9/maxds : ONE;

  FILE *dfile(NULL);
  string filename           = output + ".sca";
  dfile                     = fopen(filename.c_str(),"w");
  fprintf(dfile,"    1\n");
  fprintf(dfile," -987\n");

  string name;
  for (unsigned i          = 0; i < xtal.sg.hmname.size(); i++)
    if (!isspace(xtal.sg.hmname[i]))
      name.push_back(tolower(xtal.sg.hmname[i]));
  
  fprintf(dfile,"%10.3f%10.3f%10.3f%10.3f%10.3f%10.3f %s\n",xtal.cell[0].a,
	  xtal.cell[0].b, xtal.cell[0].c, xtal.cell[0].alpha,
	  xtal.cell[0].beta,xtal.cell[0].gamma,name.c_str());

  for (unsigned r          = 0; r < xtal.maxref; r++)
    for (unsigned d        = 0; d < xtal.sf.size(); d++)
      if (xtal.sf[d].use(r))
	if (xtal.sf[d].anomalous && !xtal.centric[r])
	{
	  if (xtal.sf[d].anouse(r))
	    fprintf(dfile, "%4d%4d%4d%8.1f%8.1f%8.1f%8.1f\n",
		    xtal.miller[r][0], xtal.miller[r][1], xtal.miller[r][2],
		    scale*xtal.sf[d].datap[r], scale*xtal.sf[d].devp[r],
		    scale*xtal.sf[d].datam[r], scale*xtal.sf[d].devm[r]);
	  else if (xtal.sf[d].usep(r))
 	    fprintf(dfile, "%4d%4d%4d%8.1f%8.1f\n",
		    xtal.miller[r][0], xtal.miller[r][1], xtal.miller[r][2],
		    scale*xtal.sf[d].datap[r], scale*xtal.sf[d].devp[r]);
 	  else if (xtal.sf[d].usem(r))
 	    fprintf(dfile, "%4d%4d%4d%8.1f%8.1f\n",
		    xtal.miller[r][0], xtal.miller[r][1], xtal.miller[r][2],
		    scale*xtal.sf[d].datam[r], scale*xtal.sf[d].devm[r]);
	}
	else
	  fprintf(dfile, "%4d%4d%4d%8.1f%8.1f\n",
		  xtal.miller[r][0], xtal.miller[r][1], xtal.miller[r][2],
		  scale*xtal.sf[d].datap[r], scale*xtal.sf[d].devp[r]);
  
  
  fclose(dfile);
}
