/*
  Copyright (c) Navraj Pannu       1998 - 2005
  Copyright (c) Leiden University  2002 - 2005

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.  
*/

#ifndef LIKELIHOOD_H
#define LIKELIHOOD_H	1

#include "misc.h"
#include "cmtzlib.h"
#include "crystal.h"
#include "matrix.h"
#include "model.h"
#include "tab.h"

class Likelihood
{
  // methods needed for minimizer
 public:
  Likelihood(Model &, Crystal &);  // to use minimizer, the Likelihood class doesn't
  virtual ~Likelihood(){;}         // need Model or Crystal classes!

  // returns the maximum step to take in the line search - set to 1.0 if you
  // are unsure.

  double Getstepmax(const vector<double> &) const;

  // returns the (double) parameters that will be refined in a vector  

  vector<double> Setpars() const;  

  // returns the function value and gradient vector of the parameters

  virtual double gradient (const vector<double> &, vector<double> &) = 0;
  
  // given a stepsize, the parameters and a search direction, it returns the function
  // values

  virtual double function1dim (const double,
			       const vector<double> &,
			       const vector<double> &)               = 0;


  // given a stepsize, the parameters and a search direction, it returns the function
  // value, the directional derivative (dfunc = grad dot_prod direction) and the
  // gradient vector (grad)

  virtual double gradient1dim (double &, vector<double> &,
			       const double,
			       const vector<double> &,
			       const vector<double> &d)              = 0;

  
  virtual void print(const int = -1) const                           = 0;
  void Setverbose(const unsigned input)  {verbose = input;   }
  unsigned Getverbose()  const           {return verbose;    }
  bool     Getnooutput() const           {return nooutput;   }
  double   Getfunction() const           {return likelihood; }
  vector<double> Getgradient() const;
  
 protected:
  double likelihood;
  // the likelhood variable holds the current function value in the likelihood class
  unsigned verbose;
  // verbose flag. 0 implies minimal messages, 1 implies more output, > 1 even more. 

  // methods needed for crystallography
 public:
  void Settitle(const string input)           {title        = input;   }
  void Settarget(const string input);
  void Setstats(const bool input)             {stats        = input;   }
  void Setallin(const bool input)             {allin        = input;   }
  void Setinterpolate(const bool input)       {interpolate  = input;   }
  void Setoutput(const string input)          {output       = input;   }
  void Setnooutput(const bool input)          {nooutput     = true;    }
  void Setmode(const string input)            {mode         = input;   }
  void Sethand(const unsigned input)          {hand         = input;   }
  void Setcweightsize(const unsigned input)   {cweight.resize(input);  }
  void Setpweightsize(const unsigned input)   {pweight.resize(input);  }
  void Setafweightsize(const unsigned input)  {afweight.resize(input); }
  void Setsadweightsize(const unsigned input) {sadweight.resize(input);}
  void Setupdatesigmah(const bool input)      {updatesigmah = input;   }
  
  string Gettitle()  const                    {return title;           }
  string Getmode()   const                    {return mode;            }
  string Gettarget() const                    {return target;          }
  unsigned Gethand() const                    {return hand;            }

  void writepdb() const;
  void writexml() const;
  void outputintensities();
  void outputscafile() const;
  void outputhklfile() const;
  
 protected:
  Model &mdl;
  Crystal &xtal;
  Tab<double> tab;
  bool allin;                              // give input mtz columns in output
  bool interpolate, stats, nooutput;
  string mode;
  string target;                           // target function requested
  string title;                            // title for MTZ file
  string output;
  unsigned hand;                           // run other hand (or not)

  // linking minimizer and likelihood function
  vector<string> type;
  void shift(const double, const vector<double> &, const vector<double> &);
  double xsshift(const double &, const string &) const;
  double maxoccshift, maxluzzshift, maxxyzshift, maxbfacshift, maxuanoshift;
  double maxfpshift, maxfppshift, maxkscaleshift, maxbscaleshift, maxnumbshift;
  void Settype();

  // gaussian integration variables
  vector<double> gsin, gcos, pweight;      // phase integral - acentric 1D
  vector<double> cnode, cweight;           // amplitude integral - centric
  vector<double> afnode, afweight;         // amplitude integral - acentric
  vector<double> sadsin, sadcos, sadweight;// sad integral

  // variables for matrix inversion
  vector<double> evectors, evalues, lawork;
  vector<int> iwork;
  int lwork, liwork, info;    
  bool hermitianinverse(Matrix &, Matrix &, Matrix &, Matrix &, double &);
  bool inverse(Matrix &, Matrix &, double &);
  void inverse2by2(Matrix &, Matrix &, double &);
  void matrixprod(Matrix &, const Matrix &, const double *);

  // structure factor calculation
  void directsfcalc(const bool, const bool);
  vector<vector<work> > hrot;
  vector<work> htran;

  // flags indicating parameter type(s) whose derivatives are required
  bool refineluzzati, refineocc, refinescale, refinexyz, refinebfac;
  bool refinefp, refinefpp, refinenumb, updatesigmah; 

  // derivatives
  vector<vector<vector<double> > > sumfpp, difffpp, difffp, sumfpfpp;

  // Note! Derivatives of structure factors are indexed [d][r] to save space if
  // there is no anomalous scattering.
  vector<vector<double> > dLdAp;           // Re(F+) - or Re(F) if F+=F=
  vector<vector<double> > dLdBp;           // Im(F+) - or Im(F) if F+=F-
  
  vector<vector<double> > dLdAm;           // Re(F-) (if anomalous)
  vector<vector<double> > dLdBm;           // Im(F-) (if anomalous)

  vector<vector<double> > dLdsigmah;       
  vector<vector<double> > dLdsumfpp;       
  vector<vector<double> > dLdsumfpfpp;       
  vector<double>          dLdocc;          // heavy atom parameters
  vector<double>          dLdnumb;
  vector<double>          dLdbiso;
  vector<vector<double> > dLduaniso;
  vector<vector<double> > dLdfp;
  vector<vector<double> > dLdfpp;
  vector<vector<double> > dLdx;

  vector<vector<double> > dLddluz;         // Luzzati parameters
  vector<vector<double> > dLdeluz;
  vector<vector<double> > dLdadluz;
  vector<vector<double> > dLdsdluz;
   
  vector<double> dLdkscale;                // Scale parameters
  vector<double> dLdbscale;
  
  // mtz writing
  vector<CMtz::MTZCOL * > colin, colout;  
  void setupmtz(CMtz::MTZ *, CMtz::MTZ *);
  void storeinitialcolumns(CMtz::MTZ *, vector<float> &,
			   const unsigned , const unsigned );
};
#endif  
