/*
  Copyright (c) Navraj Pannu       1998 - 2005
  Copyright (c) Leiden University  2002 - 2005

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.    
*/

#include <math.h>
#include "minimizer.h"

void Minimizer::refine(const bool savehessian)
{
  // Compute function and gradient
  pars                  = likelihood.Setpars();
  firstfunc             = likelihood.gradient(pars,grad);

  if (grad.size()      <= 0)
  {
    Bp3Warning("Minimizer::refine","no parameters to refine");
    return;
  }
  
  resize();
  func                  = firstfunc; 
  norm                  = ZERO;
  // initialise Hessian to identity matrix
  negativegamma         = 0;
  if (savehessian)
  {
    for (unsigned j     = 0; j < grad.size(); j++)
    {
      direction[j]      = ZERO;
      norm             += grad[j]*grad[j];
      for (unsigned k   = 0; k < grad.size(); k++)
	direction[j]   -= hessian[j][k]*grad[k];
    }
  }
  else
    for (unsigned   i   = 0; i < grad.size(); i++)
    {
      for (unsigned j   = 0; j < grad.size(); j++)
	hessian[i][j]   = ZERO;
  
      hessian[i][i]     = ONE;
      direction[i]      = -grad[i];
      norm             += grad[i]*grad[i];
    }	
  norm                  = sqrt(norm);

  minit                 = std::max(unsigned(0.6*grad.size()), (unsigned)10);

  unsigned cyc(maxit);

  if (!setit)
    cyc                 = std::max((unsigned)(1.5*grad.size()), minit);

  printf("Maximum number of iterations: %u\n", cyc);
  
  // Main loop   
  for (iter             = 0; iter < cyc ; iter++)
  {
    print();

    samegrad            = false;
    oldgrad             = grad;
    oldpars             = pars;
    double oldfunc      = func;
    lineminimizer();

    func                = (samegrad) ? newfunc : likelihood.gradient(pars, grad);

    if ( (reinit        > 1) || ( ( (oldfunc - func) < DSMALL) && ( (iter < 2) || (iter > grad.size()) ) ) )   
    {
      Bp3Warning("Minimizer::refine", "minimization without convergence");
      break;
    }

    // test for convergence
    double partest(ZERO);
    norm                = ZERO;
    for (unsigned j     = 0; j < grad.size(); j++)
    {
      norm             += grad[j]*grad[j];
      double temp       = fabs(pars[j] - oldpars[j])/std::max(ONE, oldpars[j]);
      partest           = (temp > partest) ? temp : partest;
    }
    norm                = sqrt(norm);

    if (norm          < ONE)
      break;
    /*
    if (norm          < FIVE) 
      break;

    if ( (norm          < TEN) && (iter > minit ) )
      break;

    if ( ( (norm        < 50.0)     || (partest < partol) ) && (iter > minit) &&
	 (norm          < 7500.0)   && (fabs(oldfunc - func) < 0.01) )
      break;
      */

    // update step direction...

    if (negativegamma   > 0) 
    {
      if (negativegamma > 1)
      {
        reinitializehessian();
	reinit++;
	negativegamma   = 0;
      }
      for (unsigned j   = 0; j < grad.size(); j++)
	direction[j]    = -grad[j];
    }
    else
      if (search       == "DFP")
	dfp();
      else 
	bfgs();
  }
  if (iter             == cyc)
    Bp3Warning("Minimizer::refine", "maximum number of cycles with no convergence");
  else
    iter++;
  
  print();
}

void Minimizer::resize()
{
  // Allocate memory for variables needed by the local minimizers
  direction.resize(grad.size());

  hessian.resize(grad.size());
  for (unsigned i = 0; i < grad.size(); i++)
    hessian[i].resize(grad.size());
}

void Minimizer::print() const
{
  //  Print out information detailing the minimization progress 
  
  printf("Cycle # %u\n", iter+1);
  printf("Log-likelihood      = %10.3f\n", func);
  printf("Log-likelihood gain = %10.3f\n", firstfunc - func);
  printf("Norm of gradient    = %10.3f\n\n", norm);
  fflush(stdout);
  
  likelihood.print(iter+1);
}

void Minimizer::bfgs()
{
  // updates the bfgs step direction
  vector<double> diffgrad(grad.size(), ZERO);

  for (unsigned j       = 0; j < grad.size(); j++)
  {
    diffgrad[j]         = grad[j] - oldgrad[j];
    direction[j]        = pars[j] - oldpars[j];
  }

  vector<double> hdg(grad.size(), ZERO);
  
  for (unsigned k       = 0; k < grad.size(); k++)
    for (unsigned j     = 0; j < grad.size(); j++)
      hdg[k]           += hessian[k][j]*diffgrad[j];

  double gamma(ZERO), delta(ZERO);

  for (unsigned j       = 0; j < grad.size(); j++)
  {
    gamma              += diffgrad[j]*direction[j];
    delta              += diffgrad[j]*hdg[j];
  }
  
  if (gamma             > EPSILON)
  {
    negativegamma       = 0;
    for (unsigned j     = 0; j < grad.size(); j++)
      for (unsigned k   = j; k < grad.size(); k++)
      {
 	hessian[j][k]  += ((direction[j]*direction[k]*(delta/gamma + ONE))/gamma - 
			   (hdg[j]*direction[k] + direction[j]*hdg[k])/gamma);
	hessian[k][j]   = hessian[j][k];
      }
  }
  else
  {
    negativegamma++;
    if (likelihood.Getverbose() > 1)
      Bp3Warning("Minimizer::bfgs", "gamma <= EPSILON");
  }
  
  for (unsigned j       = 0; j < grad.size(); j++)
  {
    direction[j]        = ZERO;
    for (unsigned k     = 0; k < grad.size(); k++)
      direction[j]     -= hessian[j][k]*grad[k];
  }
}

void Minimizer::dfp()
{
  // updates the dfp step direction
  vector<double> diffgrad(grad.size());

  for (unsigned j      = 0; j < grad.size(); j++)
  {
    diffgrad[j]        = grad[j] - oldgrad[j];
    direction[j]       = pars[j] - oldpars[j];
  }

  vector<double> hdg(grad.size());
  
  for (unsigned k      = 0; k < grad.size(); k++)
  {
    hdg[k]             = ZERO;
    for (unsigned j    = 0; j < grad.size(); j++)
      hdg[k]          += hessian[k][j]*diffgrad[j];
  }

  double gamma(ZERO), delta(ZERO);

  for (unsigned j      = 0; j < grad.size(); j++)
  {
    gamma             += direction[j]*diffgrad[j];
    delta             += diffgrad[j]*hdg[j];
  }

  if ( (gamma          > EPSILON) && (fabs(delta) > EPSILON) )
  {
    for (unsigned j    = 0; j < grad.size(); j++)
      for (unsigned k  = j; k < grad.size(); k++)
      {
	hessian[j][k] += (direction[j]*direction[k]/gamma -
			  hdg[j]*hdg[k]/delta);
	hessian[k][j]  = hessian[j][k];
      }
  }
  else
    if (likelihood.Getverbose() > 1)
      Bp3Warning("Minimizer::dfp", "gamma <= EPSILON or fabs(delta) <= EPSILON");
  
  for (unsigned j      = 0; j < grad.size(); j++)
  {
    direction[j]       = ZERO;
    for (unsigned k    = 0; k < grad.size(); k++)
      direction[j]    -= hessian[j][k]*grad[k];
  }
}

void Minimizer::lineminimizer()
{
  lineconverge        = false;
  nfe                 = 0;
  double step(ONE);
  if ( (iter          > ((unsigned)grad.size()/2) ) || (norm < 150000) )
    step              = likelihood.Getstepmax(direction);
  else
    step              = std::max(250.0/norm,0.0000001);

  stepmax             = step;   // ***NSP
  dfunc               = dot_prod(grad,direction);
  
  if (dfunc          >= ZERO)
  {
    if (likelihood.Getverbose() > 1)
      Bp3Warning("Minimizer::lineminimizer", 
		 "switching to steepest descent");
    for (unsigned i   = 0; i < grad.size(); i++)
      direction[i]    = -grad[i];
  }  

  if (line           == "MORE")
  {
    if (!more(step))
    {
      if ( (fbest     < newfunc) && (fabs(dfbest)   < TEN*fabs(dfunc)) ) 
      {
	step          = stepbest;
	newfunc       = fbest;
	dnewfunc      = dfbest;
      }
    }
    else 
      lineconverge    = true;
    if (fabs(newfunc  - likelihood.Getfunction()) < EPSILON)
      samegrad        = true;
  }
  else
  {
    newfunc           = likelihood.function1dim(step,pars,direction); nfe++;
      
    if (newfunc       > (func + alpha*step*dfunc))
    { 
      double diff     = fabs((newfunc - func)/func);
      step           /= std::max(ONE,std::min(diff*10.0, 1000.0));
      newfunc         = brent(step);
      lineconverge    = (step > DSMALL);
    }
  }

  if (step            < SMALLESTD)
  {
    reinit++;
    reinitializehessian();
    likelihood.function1dim(0,oldpars,direction);
  }
  else
  {
    if (fabs(newfunc  - likelihood.Getfunction()) > 0.0001)
        likelihood.function1dim(step,pars,direction);
    pars              = likelihood.Setpars();
    reinit            = 0;
  }
  
  if (likelihood.Getverbose())
  {
    printf("Line minimization\n");
    printf("Maximum step size:      %g\n", stepmax);
    printf("Final step size:        %g\n", step);
    printf("Function evaluations:   %u\n\n", nfe);
  }
}

bool Minimizer::more(double &step)
{
  // Reference:
  // J. J. More' and D. J. Thuente,
  //  Line search algorithms with guaranteed sufficient decrease,
  //  ACM Transactions on  Mathematical Software, 20, (1994), 286--307

  bool interval(false), converge(false);
  double width(stepmax-stepmin), width1(width*TWO);
  double gtest(alpha*dfunc), xtol(0.001);
  double stmin(ZERO);
  double stmax(std::min(stepmax*HALF + stepmax*TWO, ONE)); 
  double fx(func), fy(func), gx(dfunc), gy(dfunc);
  double stx(ZERO), sty(ZERO);
  fbest           = func;
  dfbest          = dfunc;
  stepbest        = ZERO;
  bracket         = false;
  
  for (unsigned i = 0; i < maxlineit; i++)
  {
    newfunc       = likelihood.gradient1dim(dnewfunc, grad, step, pars, direction);
    nfe++;
   
    /*
    if ( (fabs(newfunc - fbest) < 0.0001) && (i > 4) )
    {
      if (likelihood.Getverbose() > 1)
	Bp3Warning("Minimizer::more", "Function not decreasing anymore");
      converge    = true;
      break;
    }
    */

    if (newfunc   < fbest)
    {
      stepbest    = step;
      fbest       = newfunc;
      dfbest      = dnewfunc;
    }

    double ftest  = func + step * gtest;
    if (!interval && newfunc <= ftest && dnewfunc >= ZERO)
      interval    = true;

    if (bracket && (step <= stmin || step >= stmax))
    {
      if (likelihood.Getverbose() > 1)
	Bp3Warning("Minimizer::more", "Roundings errors prevent minimization");
      break;
    }
    if (step     == stepmax && newfunc <= ftest && dnewfunc <= gtest)
    {
      converge    = true;
      break;
    }    
    if (bracket && stmax - stmin <= xtol * stmax)
    {
      converge    = true;
      break;
    }        
    //  Test for convergence.
    if (newfunc  <= ftest && fabs(dnewfunc) <= beta * (-dfunc))
    {
      converge    = true;
      break;
    }
    
    if (!interval && newfunc <= fx && newfunc > ftest) 
    {
      double fm   = newfunc - step * gtest;
      double fxm  = fx - stx * gtest;
      double fym  = fy - sty * gtest;
      double gm   = dnewfunc - gtest;
      double gxm  = gx - gtest;
      double gym  = gy - gtest;
      choosestep(stx, fxm, gxm, sty, fym, gym, step, fm, gm, stmin, stmax);
      fx          = fxm + stx * gtest;
      fy          = fym + sty * gtest;
      gx          = gxm + gtest;
      gy          = gym + gtest;
    } 
    else
      choosestep(stx, fx, gx, sty, fy, gy, step, newfunc, dnewfunc, stmin, stmax);
    
    if (bracket) 
    {
      if (fabs(sty - stx) >= width1 * TWOTHIRD)
	step      = stx + (sty - stx) * HALF;
      width1      = width;
      width       = fabs(sty - stx);
    }
    if (bracket)  
    {
      stmin       = std::min(stx,sty);
      stmax       = std::max(stx,sty);
    } 
    else 
    {
      stmin       = step + (step - stx) * 1.1;
      stmax       = step + (step - stx) * FOUR;
    }
    step          = std::min(std::max(step,stepmin), stepmax);
    if ((bracket && (step <= stmin || step >= stmax)) || 
 	(bracket && stmax - stmin <= xtol * stmax))
      step        = stx;
  }
  return converge;
} 

void Minimizer::choosestep(double &stx,  double &fx, double &dx, 
			   double &sty,  double &fy, double &dy,
			   double &step, double &fp, double &dp,
			   const double stmin, const double stmax)
{
  
  double stepf(ZERO), stepc(ZERO), stepq(ZERO);

  double sgnd        = dp * (dx / fabs(dx));

  if (fp             > fx)
  {
    double theta     = (fx - fp) * THREE / (step - stx) + dx + dp;
    double s         = std::max(std::max(fabs(theta),fabs(dx)), fabs(dp));
    double gamma     = s * sqrt((theta*theta)/(s * s) - dx / s * (dp / s));
    if (step         < stx) 
      gamma         *= -ONE;
    double p         = gamma - dx + theta;
    double q         = gamma - dx + gamma + dp;
    double r         = p / q; 
    stepc            =  stx + r * (step - stx);
    stepq            = stx + dx / ((fx - fp) / (step - stx) + dx) / TWO * (step - stx);
    stepf            = ( (fabs(stepc - stx) < fabs(stepq - stx)) ? 
			 stepc : stepc + (stepq - stepc)/TWO);
    bracket          = true;
  } 
  else if (sgnd      < ZERO) 
  {
    double theta     = (fx - fp) * THREE / (step - stx) + dx + dp;
    double s         = std::max(std::max(fabs(theta), fabs(dx)), fabs(dp));
    double gamma     = s * sqrt(theta*theta/(s*s) - dx / s * (dp / s));
    if (step         > stx) 
      gamma         *= -ONE;
    double p         = gamma - dp + theta;
    double q         = gamma - dp + gamma + dx;
    double r         = p / q;
    stepc            = step + r * (stx - step);
    stepq            = step + dp / (dp - dx) * (stx - step);
    stepf            = (fabs(stepc - step) > fabs(stepq - step)) ? stepc : stepq;
    bracket          = true;
  } 
  else if (fabs(dp)  < fabs(dx)) 
  {
    double theta     = (fx - fp) * THREE / (step - stx) + dx + dp;
    double s         = std::max(std::max(fabs(theta),fabs(dx)) ,fabs(dp));
    double gamma     = s * sqrt((std::max(ZERO,
					  theta*theta/(s*s) - dx / s * (dp / s))));
    if (step         > stx)
      gamma          = -ONE;
    double p         = gamma - dp + theta;
    double q         = gamma + (dx - dp) + gamma;
    double r         = p / q;
    if (r            < ZERO && gamma != ZERO)  
      stepc          = step + r * (stx - step);
    else if (step    > stx) 
      stepc          = stmax;
    else
      stepc          = stmin;
    stepq            = step + dp / (dp - dx) * (stx - step);
    if (bracket) 
    {
      stepf          = (fabs(stepc - step) < fabs(stepq - step)) ? stepc : stepq;
      if (step       > stx) 
	stepf        = std::min(step + (sty - step) * TWOTHIRD,stepf);
      else 		 	
	stepf        = std::max(step + (sty - step) * TWOTHIRD,stepf);
    } 
    else
    {
      stepf          = (fabs(stepc - step) > fabs(stepq - step) ) ? stepc : stepq;
      stepf          = std::max(std::min(stmax,stepf), stmin);
    }
  } 
  else 
  {
    if (bracket) 
    {
      double theta   = (fp - fy) * THREE / (sty - step) + dy + dp;
      double s       = std::max(std::max(fabs(theta),fabs(dy)),fabs(dp));
      double gamma   = s * sqrt(theta*theta/(s*s) - dy / s * (dp / s));
      if (step       > sty) 
	gamma        = -ONE;
      double p       = gamma - dp + theta;
      double q       = gamma - dp + gamma + dy;
      double r       = p / q;
      double stepc   = step + r * (sty - step);
      stepf          = stepc;
    } 
    else if (step    > stx) 
      stepf          = stmax;
    else
      stepf          = stmin;    
  }
  if (fp             > fx) 
  { 
    sty              = step;
    fy               = fp;
    dy               = dp;
  } 
  else 
  {
    if (sgnd         < ZERO) 
    {
      sty            = stx;
      fy             = fx;
      dy             = dx;
    }
    stx              = step;
    fx               = fp;
    dx               = dp;
  }
  step               = stepf;
}

double Minimizer::brent(double &x)
{
  double a          = stepmin;
  double b          = stepmax;
  double v(ZERO);
  double w          = v;
  x                 = v;
  double fx         = func;
  double fv         = fx;
  double fw         = fx;

  double d(ZERO), e(ZERO), u(ZERO), fu(ZERO);

  for (unsigned i   = 0; i < 100; i++) 
  {
    double xm       = HALF*(a + b);
    double eps      = 0.2*fabs(x) + 1.0e-10;
    double tol      = eps*TWO;

    if (fabs(x-xm) <= tol - (b - a)*HALF) 
      break;

    bool change(true);
    
    if (fabs(e)     > eps) 
    {
      double r      = (x - w)*(fx - fv);
      double q      = (x - v)*(fx - fw);
      double p      = (x - v)*q - (x - w)*r;
      q             = (q - r)*TWO;
      if (q         > ZERO)
        p          *= -ONE;
      q             = fabs(q);
      double temp   = e;
      e             = d;
      change        = (fabs(p) >= fabs(q*HALF*temp) ||
		       p <= q * (a - x) || p >= q * (b - x));
      if (!change)
      {
	d           = p / q;
	u           = x + d;
	if (u - a   < tol || b - u < tol) 
	  d         = dsign(eps, xm - x);
      }
    }

    if (change)
    {
      if (x        >= xm) 
	e           = a - x;
      else 
	e           = b - x;
      d             = e * 0.381966;
    }
    
    if (fabs(d)    >= eps) 
      u             = x + d;
    else 
      u             = x + dsign(eps, d);
    fu              = likelihood.function1dim(u,pars,direction); nfe++;
    if (fu         <= fx) 
    {
      if (u        >= x) 
        a           = x;
      else 
        b           = x;
      v             = w;
      fv            = fw;
      w             = x;
      fw            = fx;
      x             = u;
      fx            = fu;
    } 
    else 
    {
      if (u         < x) 
        a           = u;
      else 
        b           = u;
      if (fu       <= fw || w == x) 
      {
        v           = w;
        fv          = fw;
        w           = u;
        fw          = fu;
      } 
      else if (fu  <= fv || v == x || v == w) 
      {
        v           = u;
        fv          = fu;
      }
    }
  }
  return fx;
}
