/*
  Copyright (c) Navraj Pannu       1998 - 2005
  Copyright (c) Leiden University  2002 - 2005
  
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

#include <math.h>
#include "bp3likelihood.h"

double Bp3likelihood::uncorrelatedfunction(const bool check)
{
  // Calculates likelihood assuming uncorrelated errors.

  double REJECT                    = (check) ? DSMALL : SMALLESTD;

  likelihood                       = ZERO;

  unsigned flagged(0);

  unsigned dmax(xtal.sf.size());

  // use only datasets from one crystal in phasing for mad
  if (xtal.mad &&
      xtal.dataset.size()          > 1)
    dmax                           = xtal.dataset[1][0];
  
  if (check)
    xtal.Setselected("REFINE");

  for (unsigned r                  = 0; r < xtal.maxselref; r++)
  {
    int mders(-1);
    unsigned counts(0);
    for (unsigned d                = 0; d < dmax; d++)
      if (xtal.sf[d].use(r))
      {
	mders++;
        counts++;
        if (xtal.sf[d].anomalous  && !xtal.centric[r])
          if (xtal.sf[d].anouse(r))
            counts++;
      }
    
    if ( (counts                   > 1)  && xtal.selected[r])
    {
      unsigned sd(0);
      for (unsigned d              = 0; ( ((sd=d) < dmax) &&
                                          !(xtal.sf[d].use(r)) ); d++);
      double fos(xtal.sf[sd].foversigma(r));
      bool onedim(xtal.sf[0].use(r) && (fos > threshold) && !xtal.heavyref);
      unsigned ind                 = (onedim) ? 1 : sd;

      vector<double> koverall(dmax, ONE);

      double integral(ZERO), product(ONE);

      double eps((double)xtal.epsilon[r]);
      vector<double> var(dmax, ZERO);
      if (!(xtal.centric[r] || xtal.onlycentrics))
      {
        vector<double> weight, node;
        if (onedim)
        {
          node.push_back(ZERO);
          weight.push_back(ONE);
        }
        else
        {
          weight                   = afweight;
          node                     = afnode;
        }

        vector<double> besselarg(dmax), avar(dmax);
        vector<double> cosp(dmax), sinp(dmax);
        vector<double> cosm(dmax), sinm(dmax);

        for (unsigned d            = ind; d < dmax; d++)
          if (xtal.sf[d].use(r))
          {
            unsigned s(xtal.bin(d,r));
            var[d]                 = pow(koverall[d]*xtal.sf[d].devmean(r), 2);
	    if (xtal.sf[d].noniso && (d != sd) )
	    {
 	      double temp(eps*xtal.sf[d].var(koverall[d],s,xtal.heavyref));
	      var[d]              += fabs(temp);
	    }
            besselarg[d]           = TWO*koverall[d]*xtal.sf[d].datamean(r)/var[d];
            cosp[d]                = tab.tcos(xtal.sf[d].pcalcp[r]);
            sinp[d]                = tab.tsin2(xtal.sf[d].pcalcp[r]);
            if (mders)
              product             *= besselarg[d];
            if (xtal.sf[d].anomalous)
            {
              cosm[d]              = tab.tcos(xtal.sf[d].pcalcm[r]);
              sinm[d]              = tab.tsin2(xtal.sf[d].pcalcm[r]);
              if (xtal.sf[d].anouse(r))
              {
                avar[d]            = (std::max(ZERO, eps*(koverall[d]*koverall[d]*
                                                          xtal.sf[d].sigmadano[s]*
                                                          (ONE -
                                                           xtal.sf[d].adluz[s]*
                                                           xtal.sf[d].adluz[s]))) +
                                      pow(koverall[d]*xtal.sf[d].sigdano(r), 2));
                product           /= sqrt(PI*avar[d]);
              }
            }
          }

	for (unsigned i            = 0; i < pweight.size(); i++)
	{
          double phasesum(ZERO);
	  for (unsigned a          = 0; a < weight.size(); a++)
	  {
	    double afvarnode(xtal.sf[sd].devmean(r)*node[a] +
			     xtal.sf[sd].datamean(r));
	    double afvarnode1(ZERO);

 	    if (xtal.heavyref)
	    {
	      double delcosp(gcos[i]*cosp[sd] + gsin[i]*sinp[sd]);
	      afvarnode1           = std::max(sqrt(fabs(afvarnode*afvarnode -
		 					xtal.sf[sd].fcalcp[r]*
							xtal.sf[sd].fcalcp[r]*
							(ONE - delcosp*delcosp))),
					      DSMALL);
   	      afvarnode            = afvarnode1 - xtal.sf[sd].fcalcp[r]*delcosp;
 	    }

            double expsum(ZERO), prob(ONE);

	    for (unsigned d        = ind; d < dmax; d++)
	      if (xtal.sf[d].use(r))
	      {
		double delcosp(gcos[i]*cosp[d] + gsin[i]*sinp[d]);
		unsigned s(xtal.bin(d,r));
		double df(xtal.sf[d].dluz[s]*afvarnode);
		double efcalcp(xtal.sf[d].fcalcp[r]);
		double fjcalc(sqrt(std::max(df*df + efcalcp*efcalcp +
					    TWO*df*efcalcp*delcosp, EPSILON)));
		double fjcalcm(ZERO);
		if (xtal.sf[d].anomalous)
		{
		  double delcosm(gcos[i]*cosm[d] + gsin[i]*sinm[d]);
		  double efcalcm(xtal.sf[d].fcalcm[r]);
		  fjcalcm          = (sqrt(std::max(df*df + efcalcm*efcalcm +
						  TWO*df*efcalcm*delcosm, EPSILON)));
 		  if (xtal.sf[d].anouse(r))
		  {
		    double adiff(koverall[d]*xtal.sf[d].dano(r) - (fjcalc - fjcalcm));
		    fjcalc         = (fjcalc + fjcalcm)*HALF;
		    expsum        -= adiff*adiff/avar[d];
		  }
		  else if (xtal.sf[d].usem(r))
		    fjcalc         = fjcalcm;
		}
		if (mders)
		{
		  double diff(koverall[d]*xtal.sf[d].datamean(r) - fjcalc);
		  expsum          -= diff*diff/var[d];
		  prob            *= tab.ti0e(besselarg[d]*fjcalc);
		}
	      }
 	    phasesum              += prob*tab.texp(expsum)*weight[a];
	  }
	  integral                += phasesum*pweight[i];
        }
        product                   *= integral*PI;

        if (!onedim)
          product                 *= xtal.sf[sd].devmean(r);
      }
      else if (xtal.centric[r]    && !xtal.onlyacentrics)    // Centric case
      {
        vector<double> weight, node;
        if (onedim)
        {
          node.push_back(ZERO);
          weight.push_back(ONE);
        }
        else
        {
          weight                   = cweight;
          node                     = cnode;
        }

        vector<double> samephase(dmax, ONE), cosharg(dmax);
        for (unsigned d            = ind; d < dmax; d++)
          if(xtal.sf[d].use(r))
          {
            unsigned s(xtal.bin(d,r));
            var[d]                 = pow(koverall[d]*xtal.sf[d].devmean(r), 2);
	    if (xtal.sf[d].noniso && (d != sd) )
	    {
	      double temp(eps*xtal.sf[d].var(koverall[d],s,xtal.heavyref));
	      if (temp             > ZERO)
		var[d]            += temp;
	      else
		likelihood        += 100.0;
	    }
	    if (tab.tcos(xtal.centricphase[r] - xtal.sf[d].pcalcp[r]) < ZERO)
	      samephase[d]         = -ONE;
	    product               /= (TWO*PI*var[d]);
            cosharg[d]             = koverall[d]*xtal.sf[d].datamean(r)/var[d];
          }

        for (unsigned i            = 0; i < weight.size(); i++)
        {
	  double cvarnode(xtal.sf[sd].devmean(r)*node[i] +
			  xtal.sf[sd].datamean(r));
	  double cvarnode1(cvarnode), cvarnode2(cvarnode);	  

          double prob1(ONE), expsum1(ZERO), prob2(ONE), expsum2(ZERO);

	  if (xtal.heavyref)
	  {
	    cvarnode1          = (cvarnode - samephase[sd]*xtal.sf[sd].fcalcp[r]);
	    cvarnode2          = (cvarnode + samephase[sd]*xtal.sf[sd].fcalcp[r]);
	  }

          for (unsigned d          = ind; d < dmax; d++)
            if (xtal.sf[d].use(r))
            {
              unsigned s(xtal.bin(d,r));
	      
	      double fjcalc1(fabs(xtal.sf[d].dluz[s]*cvarnode1 +
				  samephase[d]*xtal.sf[d].fcalcp[r]));
	      
	      double fjcalc2(fabs(xtal.sf[d].dluz[s]*cvarnode2 -
				  samephase[d]*xtal.sf[d].fcalcp[r]));
	      
              double diff1(koverall[d]*xtal.sf[d].datamean(r) - fjcalc1);
              double diff2(koverall[d]*xtal.sf[d].datamean(r) - fjcalc2);

	      if (d               != sd)
	      {
		expsum1           -= diff1*diff1/(TWO*var[d]);
		expsum2           -= diff2*diff2/(TWO*var[d]);
	      }

	      prob1               *= ONE + tab.texp(-TWO*cosharg[d]*fjcalc1);
	      prob2               *= ONE + tab.texp(-TWO*cosharg[d]*fjcalc2);
	    }
	  integral                += (prob1*tab.texp(expsum1)+
                                      prob2*tab.texp(expsum2))*weight[i];
        }
        product                    = sqrt(product)*integral;
        if (!onedim)
          product                 *= xtal.sf[sd].devmean(r);
      }

      if (product                  > REJECT)
        likelihood                -= log(product);
      else // A very low likelihood, so as not to confuse the minimizer
        likelihood                += 700.0;
    }
  }

  if (xtal.verbose > 1)
    printf("LIKELIHOOD = %f\n",likelihood);

  if (check && flagged)
    printf("Number of reflections with low likelihood values: %d\n", flagged);

  return likelihood;
}

double Bp3likelihood::uncorrelatedgradient(const bool check, const bool outputmtz)
{
  // Calculates likelihood assuming uncorrelated errors.

  likelihood                               = ZERO;

  double REJECT((check && !outputmtz) ? DSMALL : SMALLESTD);
  
  const bool deriv(true);
  
  unsigned flagged(0), nstop(0);
  double stopfom(ZERO);
  unsigned dmax(xtal.sf.size());

  // use only datasets from one crystal in phasing for mad
  if (xtal.mad && xtal.dataset.size()      > 1)
      dmax                                 = xtal.dataset[1][0];

  if (check)
    xtal.Setselected( (outputmtz ? "PHASE" : "REFINE") );

  CMtz::MTZ *MTZIN(NULL), *MTZOUT(NULL);

  if (outputmtz)
  {
    MTZOUT                                 = CMtz::MtzMalloc(0,0);
    if (allin)
    {
      if (getenv("HKLIN")                 != NULL)
	MTZIN                              = CMtz::MtzGet("HKLIN",0);
      else if (xtal.sf[0].mtzin.size())
	MTZIN                              = CMtz::MtzGet(xtal.sf[0].mtzin.c_str(),0);

      if (!MTZIN)
	Bp3Error("Bp3likelihood::uncorrelatedgradient","Can not read input mtz");
    }
    
    setupmtz(MTZOUT,MTZIN);
  }

  unsigned inc((allin) ? colin.size() : 3);

  for (unsigned d                          = 0; d < dmax  ; d++)
  {
    for (unsigned s                        = 0; s < xtal.sf[d].nbins; s++)
    {
      dLdsigmah[d][s]                      = dLdsumfpp[d][s] = ZERO;
      dLddluz[d][s]                        = dLdadluz[d][s]  = ZERO;

      if (d                               == 0)
	afom[s]                            = cfom[s]         = ZERO;
    }
    xtal.sf[d].initstat();
  }
  
  vector<double> koverall(dmax, ONE), var(dmax, ZERO);
  vector<double> besselarg(dmax), avar(dmax);
  vector<double> fcalc(dmax), fcalcp(dmax), fcalcm(dmax);
  vector<double> diff(dmax, ZERO), adiff(dmax);
  vector<double> tsim(dmax, ZERO);
  vector<double> cosp(dmax), sinp(dmax);
  vector<double> cosm(dmax), sinm(dmax);
  vector<double> cosharg(dmax);
  vector<double> fcalc2(dmax), diff2(dmax, ZERO);
  vector<double> nfcalc1(dmax), nfcalc2(dmax);
  
  vector<double> dLdd(dmax, ZERO),  dLdvar(dmax, ZERO);
  vector<double> dLdad(dmax, ZERO), dLdavar(dmax, ZERO);
  vector<double> dLdFp(dmax, ZERO), dLdPp(dmax, ZERO);
  vector<double> dLdFm(dmax, ZERO), dLdPm(dmax, ZERO);

  vector<double> dphasedfp(dmax, ZERO), dphasedpp(dmax, ZERO);
  vector<double> dphasedfm(dmax, ZERO), dphasedpm(dmax, ZERO);
  vector<double> dphaseddluz(dmax, ZERO), dphasedvar(dmax, ZERO);
  vector<double> dphasedadluz(dmax, ZERO), dphasedavar(dmax, ZERO);
  vector<double> delcosp(dmax), delcosm(dmax);

  vector<double> isolofc(dmax, ZERO), alofc(dmax, ZERO);
  vector<double> iphaselofc(dmax, ZERO), aphaselofc(dmax, ZERO);
  vector<double> ecosdiff(dmax, ZERO), ephasecosdiff(dmax, ZERO);

  for (unsigned r                          = 0; r < xtal.maxselref; r++)
  {
    // Default value for columns to be written out is MNF
    vector<float> fdata(9+inc, CCP4::ccp4_nan().f);

    if (outputmtz)
      storeinitialcolumns(MTZIN, fdata, inc, r);

    double eps((double) xtal.epsilon[r]);
    
    unsigned counts(0);
    int mders(-1);
    for (unsigned d                        = 0; d < dmax; d++)
      if (xtal.sf[d].use(r))
      {
 	counts++;
	mders++;
	if (xtal.sf[d].anomalous          && !xtal.centric[r])
	  if (xtal.sf[d].anouse(r))
	    counts++;
      }
    
    if (outputmtz                         && (counts == 1))
      for (unsigned d                      = 0; d < dmax; d++)
	if (xtal.sf[d].use(r))
	  counts                           = (xtal.sf[d].sigmah[0] > ZERO) ? 1 : 0;
    
    if ( (counts                           > 1) && xtal.selected[r])
    {
      unsigned sd(0);
      for (unsigned d                      = 0; ( ((sd=d) < dmax) && 
						  !(xtal.sf[d].use(r)) ); d++); 

      double fos(xtal.sf[sd].foversigma(r));

      bool onedim(xtal.sf[0].use(r) && (fos > threshold) && !xtal.heavyref);

      unsigned ind((onedim) ? 1 : sd);
      
      for (unsigned d                      = 0; d < dmax; d++)
      {
	dLdAp[d][r]                        = dLdBp[d][r] = ZERO;
	if (xtal.sf[d].anomalous)
	  dLdAm[d][r]                      = dLdBm[d][r] = ZERO;

	dLdFp[d]                           = dLdPp[d]  = dLdFm[d] = dLdPm[d]   = ZERO;
	dLdd[d]                            = dLdvar[d] = dLdad[d] = dLdavar[d] = ZERO;

	if (xtal.sf[d].use(r)  || (d      == 0) )
	  if (xtal.centric[r])
	    xtal.sf[d].cnshl[xtal.bin(d,r)]++;
	  else
	    xtal.sf[d].anshl[xtal.bin(d,r)]++;
	isolofc[d]                         = alofc[d]  = ecosdiff[d]           = ZERO;
      }
      double totcos(ZERO), totsin(ZERO), totamp(ZERO), totampcos(ZERO), totampsin(ZERO);
      double totlogcos(ZERO), totlogsin(ZERO), totlogcos2(ZERO), totlogsin2(ZERO);
      
      double integral(ZERO), product(ONE);
      vector<double> dervar(dmax, ONE);
      vector<double> weight, node;

      if (!(xtal.centric[r] || xtal.onlycentrics))
      {
	if (onedim)
	{
	  node.push_back(ZERO);
	  weight.push_back(ONE);
	}
	else
	{
	  weight                           = afweight;
	  node                             = afnode;
	}
	
	vector<bool> deravar(dmax, true);

	for (unsigned d                    = ind; d < dmax; d++)
	  if (xtal.sf[d].use(r))
	  {
	    unsigned s(xtal.bin(d,r));
	    var[d]                         = pow(koverall[d]*xtal.sf[d].devmean(r), 2);
	    
	    if (xtal.sf[d].noniso         && (d != sd) )
	    {
	      double temp(eps*xtal.sf[d].var(koverall[d],s,xtal.heavyref));
	      dervar[d]                    = sign(temp);
	      if (temp                     > ZERO)
                var[d]                    += temp;
	      else
		dervar[d]                  = ZERO;
	    }
 	    besselarg[d]                   = TWO*koverall[d]*xtal.sf[d].datamean(r)/var[d];
	    cosp[d]                        = tab.tcos(xtal.sf[d].pcalcp[r]);
	    sinp[d]                        = tab.tsin2(xtal.sf[d].pcalcp[r]);
	    if (mders)
	      product                     *= besselarg[d];
	    xtal.sf[d].asumfph[s]         += koverall[d]*xtal.sf[d].datamean(r);
	    xtal.sf[d].asumfh[s]          += xtal.sf[d].fcalcp[r];
	    if (xtal.sf[0].use(r))
	      xtal.sf[d].asumdiff[s]      += fabs(koverall[d]*xtal.sf[d].datamean(r) -
						  koverall[0]*xtal.sf[0].datamean(r));
 	    if (xtal.sf[d].anomalous)
	    {
	      cosm[d]                      = tab.tcos(xtal.sf[d].pcalcm[r]);
	      sinm[d]                      = tab.tsin2(xtal.sf[d].pcalcm[r]);
	      if (xtal.sf[d].anouse(r))
	      {
		avar[d]                    = std::max(eps*(koverall[d]*koverall[d]*
							   xtal.sf[d].sigmadano[s]*
							   (ONE-xtal.sf[d].adluz[s]*
							    xtal.sf[d].adluz[s])),ZERO);
		deravar[d]                 = (avar[d] > ZERO);
		avar[d]                   += pow(koverall[d]*xtal.sf[d].sigdano(r), 2);
		product                   /= sqrt(PI*avar[d]);
		xtal.sf[d].sumimfh[s]     += fabs(xtal.sf[d].fcalcp[r]*
						  tab.tsin(xtal.sf[d].pcalcp[r]));
		xtal.sf[d].sumdano[s]     += fabs(xtal.sf[d].dano(r));
		xtal.sf[d].anosumfph[s]   += fabs(xtal.sf[d].datamean(r));
		xtal.sf[d].sanonshl[s]++;
	      }
	    }
	  }

	for (unsigned i                    = 0; i < pweight.size(); i++)
	{
	  double afvarnode(ZERO), afvarnode1(ZERO), saf1(ONE);
	  double phasesum(ZERO), phaseamp(ZERO);

	  for (unsigned d                  = ind; d < dmax; d++)
 	    if (xtal.sf[d].use(r))
	    {
	      delcosp[d]                   = gcos[i]*cosp[d] + gsin[i]*sinp[d];
	      if (xtal.sf[d].anomalous)
		delcosm[d]                 = gcos[i]*cosm[d] + gsin[i]*sinm[d];
	      dphasedfp[d]                 = dphasedpp[d]   = dphasedfm[d]     = ZERO;
	      dphasedpm[d]                 = dphaseddluz[d] = dphasedvar[d]    = ZERO;
	      dphasedadluz[d]              = dphasedavar[d] = ZERO;
	      
	      iphaselofc[d]                = aphaselofc[d]  = ephasecosdiff[d] = ZERO;
	    }

	  for (unsigned a                  = 0; a < weight.size(); a++)
	  { 
 	    double expsum(ZERO), prob(ONE);

	    afvarnode                      = (xtal.sf[sd].devmean(r)*node[a] +
					      xtal.sf[sd].datamean(r));
	    if (xtal.heavyref)
	    {
	      afvarnode1                   = (afvarnode*afvarnode -
					      (xtal.sf[sd].fcalcp[r]*
					       xtal.sf[sd].fcalcp[r]*
					       (ONE - delcosp[sd]*
						delcosp[sd])));
	      saf1                         = sign(afvarnode1);
	      afvarnode1                   = std::max(sqrt(fabs(afvarnode1)), DSMALL);
	      afvarnode                    = (afvarnode1 -
                                              xtal.sf[sd].fcalcp[r]*delcosp[sd]);
	      afvarnode                    = (afvarnode1 -
					      xtal.sf[sd].fcalcp[r]*delcosp[sd]);
 	    }

	    for (unsigned d                = ind; d < dmax; d++)
	      if (xtal.sf[d].use(r))
	      {
		double df(xtal.sf[d].dluz[xtal.bin(d,r)]*afvarnode);
		fcalc[d] = fcalcp[d]       = sqrt(std::max(df*df+
					 		   xtal.sf[d].fcalcp[r]*
							   xtal.sf[d].fcalcp[r]+
							   TWO*df*delcosp[d]*
							   xtal.sf[d].fcalcp[r],
							   EPSILON));
		if (xtal.sf[d].anomalous)
		{
		  fcalcm[d]                = sqrt(std::max(df*df+
				 			   xtal.sf[d].fcalcm[r]*
							   xtal.sf[d].fcalcm[r]+
							   TWO*df*delcosm[d]*
							   xtal.sf[d].fcalcm[r],
							   EPSILON));
		  if (xtal.sf[d].anouse(r))
		  {
		    fcalc[d]               = (fcalcp[d] + fcalcm[d])*HALF;
		    adiff[d]               = (koverall[d]*xtal.sf[d].dano(r) - 
 					      (fcalcp[d] - fcalcm[d]));
		    expsum                -= adiff[d]*adiff[d]/avar[d];
		  }
		  else if (xtal.sf[d].usem(r))
		    fcalc[d]               = fcalcm[d];  
		}
		if (mders)
		{
		  diff[d]                  = koverall[d]*xtal.sf[d].datamean(r) - fcalc[d];
		  expsum                  -= diff[d]*diff[d]/var[d];
		  double io(tab.ti0e(besselarg[d]*fcalc[d]));
		  tsim[d]                  = tab.tsim(besselarg[d]*fcalc[d], io);
		  prob                    *= io;
		}
	      }
	    prob                          *= tab.texp(expsum);

	    if (outputmtz                 && xtal.heavyref)
	      prob                        *= (fabs(xtal.sf[sd].devmean(r)*node[a] +
			 			   xtal.sf[sd].datamean(r))/afvarnode1);
	    double ppweight(prob*weight[a]);
	    phasesum                      += ppweight;
	    phaseamp                      += afvarnode*ppweight;

	    for (unsigned d                = ind; d < dmax; d++)
 	      if (xtal.sf[d].use(r))
	      {
		unsigned s(xtal.bin(d,r));
		double dfjcalc(ZERO), dadiffdfp(ZERO);

		double temp                = ((tsim[d]*koverall[d]*xtal.sf[d].datamean(r)
					       - fcalc[d])*ppweight/var[d]);

		if (xtal.sf[d].anomalous)
		{
		  double dfjcalcm(ZERO), dadiffdfm(ZERO);

		  if (xtal.sf[d].anouse(r))
		  {
		    if (mders)
		      dfjcalc = dfjcalcm   = temp/TWO;

		    dadiffdfp              = adiff[d]*ppweight/avar[d];
		    dadiffdfm              = adiff[d]*ppweight/avar[d];
		    dphasedavar[d]        += adiff[d]*adiff[d]*ppweight;
		  }
		  else if (xtal.sf[d].usem(r)  && mders)
		    dfjcalcm               = temp;
		  else if (mders)
		    dfjcalc                = temp;

		  dphaseddluz[d]          += ((xtal.sf[d].dluz[s]*afvarnode +
					       xtal.sf[d].fcalcm[r]*
					       delcosm[d])/fcalcm[d]*
					      (dfjcalcm - dadiffdfm))*afvarnode;
		  dphasedfm[d]            += ((xtal.sf[d].fcalcm[r]+xtal.sf[d].dluz[s]*
					       afvarnode*delcosm[d])/fcalcm[d]*
					      (dfjcalcm - dadiffdfm));
		  dphasedpm[d]            += ((dfjcalcm - dadiffdfm)/fcalcm[d]*
					      afvarnode);

		  if (xtal.heavyref)
		  {
		    double dfcdf((xtal.sf[d].dluz[s]*
				  (xtal.sf[d].dluz[s]*afvarnode +
				   xtal.sf[d].fcalcm[r]*delcosm[d])
				  )*(dfjcalcm - dadiffdfm)/fcalcm[d]);
		    if (afvarnode1         > DSMALL)
		    {
		      dphasedfp[sd]       -= ((xtal.sf[sd].fcalcp[r]*saf1/afvarnode1*
                                               (ONE - delcosp[sd]*delcosp[sd])
                                               + delcosp[sd])*dfcdf);
		      dphasedpp[sd]       += ((delcosp[sd]*xtal.sf[sd].fcalcp[r]*saf1
                                               /afvarnode1 - ONE)*dfcdf);
		    }
		    else
		    {
		      dphasedfp[sd]       -= delcosp[sd]*dfcdf;
		      dphasedpp[sd]       -= dfcdf;
		    }
 		  }
		}
		else if (mders)
		  dfjcalc                  = temp;

		dphasedvar[d]             += (diff[d]*diff[d]/var[d] +
					      besselarg[d]*fcalc[d]*
					      (ONE - tsim[d]))*ppweight;
		dphaseddluz[d]            += ((xtal.sf[d].dluz[s]*afvarnode +
					       xtal.sf[d].fcalcp[r]*delcosp[d])
					      /fcalcp[d]*
					      (dfjcalc + dadiffdfp))*afvarnode;
		dphasedfp[d]              += ((xtal.sf[d].fcalcp[r]+xtal.sf[d].dluz[s]*
					       afvarnode*delcosp[d])/fcalcp[d]*
					      (dfjcalc + dadiffdfp));
		dphasedpp[d]              += ((dfjcalc + dadiffdfp)/fcalcp[d]*
					      afvarnode);

		if (xtal.heavyref)
		{
		  double dfcdf(xtal.sf[d].dluz[s]*
			       (xtal.sf[d].dluz[s]*afvarnode +
				xtal.sf[d].fcalcp[r]*delcosp[d]
				)*(dfjcalc + dadiffdfp)/fcalcp[d]);
		  if (afvarnode1           > DSMALL)
		  {
		    dphasedfp[sd]         -= (xtal.sf[sd].fcalcp[r]*saf1/afvarnode1*
                                              (ONE - delcosp[sd]*delcosp[sd]) +
                                              delcosp[sd])*dfcdf;
		    dphasedpp[sd]         += (delcosp[sd]*xtal.sf[sd].fcalcp[r]*saf1
                                              /afvarnode1 - ONE)*dfcdf;
		  }
		  else
		  {
		    dphasedfp[sd]         -= delcosp[sd]*dfcdf;
		    dphasedpp[sd]         -= dfcdf;		    
		  }
		}
	      }

	    for (unsigned d               = ind; d < dmax; d++)
	      if (xtal.sf[d].use(r))
	      {
		ephasecosdiff[d]          += delcosp[d]*ppweight;
		iphaselofc[d]             += fabs(diff[d])*ppweight;
		if (xtal.sf[d].anomalous)
		  if (xtal.sf[d].anouse(r))
		    aphaselofc[d]         += fabs(adiff[d])*ppweight;  
	      }
	  }
	  integral                        += phasesum*pweight[i];
	  totcos                          += gcos[i]*phasesum*pweight[i];
	  totsin                          += gsin[i]*phasesum*pweight[i];
	  totamp                          += phaseamp*pweight[i];
	  totampcos                       += gcos[i]*phaseamp*pweight[i];
	  totampsin                       += gsin[i]*phaseamp*pweight[i];
	  if (phasesum                     > SMALLESTD)
	  {
	    double temp(log(phasesum)*pweight[i]);
	    totlogcos                     += gcos[i]*temp;
	    totlogsin                     += gsin[i]*temp;
	    totlogcos2                    += ((gcos[i]*gcos[i] - gsin[i]*gsin[i])*temp);
	    totlogsin2                    += (TWO*gcos[i]*gsin[i]*temp);
	  } 

	  for (unsigned d                  = ind; d < dmax; d++)
	    if (xtal.sf[d].use(r))
	    {
	      dLdd[d]                     += dphaseddluz[d]*pweight[i];
	      dLdvar[d]                   += dphasedvar[d]*pweight[i];
	      dLdFp[d]                    += dphasedfp[d]*pweight[i];
	      dLdPp[d]                    += (dphasedpp[d]*pweight[i]*
					      (gsin[i]*cosp[d] - gcos[i]*sinp[d]));
	      if (xtal.sf[d].anomalous)
	      {
		dLdavar[d]                += dphasedavar[d]*pweight[i];
		dLdFm[d]                  += dphasedfm[d]*pweight[i];
		dLdPm[d]                  += (dphasedpm[d]*pweight[i]*
					      (gsin[i]*cosm[d] - gcos[i]*sinm[d]));
	      }
	    }

	  for (unsigned d                  = ind; d < dmax; d++)
	    if (xtal.sf[d].use(r))
	    {
	      ecosdiff[d]                 += ephasecosdiff[d]*pweight[i];
	      isolofc[d]                  += iphaselofc[d]*pweight[i];
	      if (xtal.sf[d].anomalous)
		if (xtal.sf[d].anouse(r))
		  alofc[d]                += aphaselofc[d]*pweight[i];
	    }
	}
	
	if (integral                       > REJECT)
	  for (unsigned d                  = ind; d < dmax; d++)
	    if (xtal.sf[d].use(r))
	    {
	      unsigned s                   = xtal.bin(d,r);
	      double denom(ONE/(integral*var[d]));
	      dLdvar[d]                    = dervar[d]*(ONE/var[d] - dLdvar[d]*denom);
	      dLddluz[d][s]               -= TWO*dLdd[d]/integral;

	      if (mders                   && xtal.sf[d].noniso)
	      {
		dLddluz[d][s]             -= (TWO*eps*xtal.sf[d].dluz[s]*
					      dLdvar[d]*xtal.sf[d].sigmanref[s]);
		if (updatesigmah          && !xtal.heavyref)
		  dLdsigmah[d][s]         -= eps*dLdvar[d];
 	      }
	      denom                        = TWO/integral;
	      if (xtal.sf[d].anomalous)
	      {
		if (xtal.sf[d].anouse(r))
		{
		  dLdavar[d]               = (HALF - dLdavar[d]/(integral*avar[d])
					      )/avar[d];
		  if (deravar[d])
		    dLdadluz[d][s]        -= (xtal.sf[d].adluz[s]*dLdavar[d]*
					      eps*koverall[d]*koverall[d]*
					      xtal.sf[d].sigmadano[s])*TWO;
		}
		dLdAm[d][r]                = -(dLdFm[d]*cosm[d] -
				 	       dLdPm[d]*sinm[d]*xtal.sf[d].dluz[s])*denom;
		dLdBm[d][r]                = -(dLdFm[d]*sinm[d] +
					       dLdPm[d]*cosm[d]*xtal.sf[d].dluz[s])*denom;
	      }
	      dLdAp[d][r]                  = -(dLdFp[d]*cosp[d] -
					       dLdPp[d]*sinp[d]*xtal.sf[d].dluz[s])*denom;
	      dLdBp[d][r]                  = -(dLdFp[d]*sinp[d] +
					       dLdPp[d]*cosp[d]*xtal.sf[d].dluz[s])*denom;
	    }
	product                           *= integral*PI;

	if (!onedim)
	  product                         *= xtal.sf[sd].devmean(r);
      }
      else if (xtal.centric[r]            && !xtal.onlyacentrics)   // Centric case
      {
	if (onedim)
	{
	  node.push_back(ZERO);
	  weight.push_back(ONE);
	}
	else
	{
	  weight                           = cweight;
	  node                             = cnode;
	}

	vector<double> samephase(dmax, ONE);

	for (unsigned d                    = ind; d < dmax; d++)
	  if(xtal.sf[d].use(r))
	  {
	    unsigned s                     = xtal.bin(d,r);
	    var[d]                         = pow(koverall[d]*xtal.sf[d].devmean(r), 2);
	    if (xtal.sf[d].noniso         && (d != sd))
	    {
	      double temp(eps*xtal.sf[d].var(koverall[d],s,xtal.heavyref));
	      dervar[d]                    = (temp > ZERO);
	      if (temp                     > ZERO)
 		var[d]                    += temp;
	      else
		likelihood                += 100.0;
	    }
	    if (tab.tcos(xtal.centricphase[r] - xtal.sf[d].pcalcp[r]) < ZERO)
	      samephase[d]                 = -ONE;
	    product                       /= (TWO*PI*var[d]);
	    cosharg[d]                     = koverall[d]*xtal.sf[d].datamean(r)/var[d];
	    xtal.sf[d].csumfph[s]         += koverall[d]*xtal.sf[d].datamean(r);
	    xtal.sf[d].csumfh[s]          += xtal.sf[d].fcalcp[r];
	    if (xtal.sf[0].use(r))
	      xtal.sf[d].csumdiff[s]      += fabs(koverall[d]*xtal.sf[d].datamean(r) -
						  koverall[0]*xtal.sf[0].datamean(r));
	  }	  

	for (unsigned i                    = 0; i < weight.size(); i++)
	{
	  double cvarnode(xtal.sf[sd].devmean(r)*node[i] +
		 	  xtal.sf[sd].datamean(r));
	  
	  double cvarnode1(cvarnode), cvarnode2(cvarnode);

	  if (xtal.heavyref)
	  {
	    cvarnode1                  = (cvarnode - samephase[sd]*xtal.sf[sd].fcalcp[r]);
	    cvarnode2                  = (cvarnode + samephase[sd]*xtal.sf[sd].fcalcp[r]);
	  }

	  double prob1(ONE),  expsum1(ZERO),   prob2(ONE),  expsum2(ZERO);
	  
	  for (unsigned d                  = ind; d < dmax; d++)
	    if (xtal.sf[d].use(r))
	    {
	      unsigned s(xtal.bin(d,r));	      
	      
	      nfcalc1[d]                   = (xtal.sf[d].dluz[s]*cvarnode1 +
					      samephase[d]*xtal.sf[d].fcalcp[r]);
	      nfcalc2[d]                   = (xtal.sf[d].dluz[s]*cvarnode2 -
					      samephase[d]*xtal.sf[d].fcalcp[r]);

	      fcalc[d]                     = fabs(nfcalc1[d]);
	      fcalc2[d]                    = fabs(nfcalc2[d]);

	      diff[d]                      = (koverall[d]*xtal.sf[d].datamean(r)
					      - fcalc[d]);
	      diff2[d]                     = (koverall[d]*xtal.sf[d].datamean(r)
					      - fcalc2[d]);
	      if (d                       != sd)
	      {
		expsum1                   -= diff[d]*diff[d]/(TWO*var[d]);
		expsum2                   -= diff2[d]*diff2[d]/(TWO*var[d]);
	      }
	      prob1                       *= ONE + tab.texp(-TWO*cosharg[d]*fcalc[d]);
	      prob2                       *= ONE + tab.texp(-TWO*cosharg[d]*fcalc2[d]);
	    }

	  prob1                           *= tab.texp(expsum1);
	  prob2                           *= tab.texp(expsum2);
	  
	  integral                        += (prob1 + prob2)*weight[i];
	  totcos                          += (prob1 - prob2)*weight[i];
	  totampcos                       += (prob1*cvarnode1 -
					      prob2*cvarnode2)*weight[i];
	  
	  for (unsigned d                  = ind; d < dmax; d++)
	    if (xtal.sf[d].use(r))
	      isolofc[d]                  += (prob1*fabs(diff[d]) + 
					      prob2*fabs(diff2[d]))*weight[i];

	  if (deriv)
	    for (unsigned d                = ind; d < dmax; d++)
	      if (xtal.sf[d].use(r))
	      {
		double tanharg1(tanh(cosharg[d]*fcalc[d]));
		double tanharg2(tanh(cosharg[d]*fcalc2[d]));

		double derivfjcalc1(ZERO), derivfjcalc2(ZERO);

		if (d                     != sd)
		{
		  derivfjcalc1             = prob1*(koverall[d]*xtal.sf[d].datamean(r)*
						    tanharg1 - fcalc[d])*weight[i];

		  derivfjcalc2             = prob2*(koverall[d]*xtal.sf[d].datamean(r)*
		 		 		    tanharg2 - fcalc2[d])*weight[i];
		}
		
		dLdvar[d]                 += (prob1*(diff[d]*diff[d]/var[d] +
						     TWO*fcalc[d]*koverall[d]*
						     xtal.sf[d].datamean(r)/var[d]*
						     (ONE - tanharg1)) +
					      prob2*(diff2[d]*diff2[d]/var[d] +
						     TWO*fcalc2[d]*koverall[d]*
						     xtal.sf[d].datamean(r)/var[d]*
						     (ONE - tanharg2)))*weight[i];
 		dLdd[d]                   += (sign(nfcalc1[d])*cvarnode1*derivfjcalc1 +
					      sign(nfcalc2[d])*cvarnode2*derivfjcalc2);

		double temp(sign(nfcalc1[d])*derivfjcalc1 -
			    sign(nfcalc2[d])*derivfjcalc2);
		  
  		dLdFp[d]                  += temp*samephase[d];
		
		if (xtal.heavyref)
		  dLdFp[sd]               -= xtal.sf[d].dluz[xtal.bin(d,r)]*temp*samephase[sd];
	      }
	  
	}

	product                            = sqrt(product)*integral;
	if (!onedim)
	  product                         *= xtal.sf[sd].devmean(r);
	
	if (deriv && (integral             > REJECT))
	  for (unsigned d                  = ind; d < dmax; d++)
	    if (xtal.sf[d].use(r))
	    {
	      unsigned s(xtal.bin(d,r));
	      double temp(ONE/(integral*var[d]));
	      dLdvar[d]                    = (ONE - dLdvar[d]/integral)*HALF;

	      dLddluz[d][s]               -= dLdd[d]*temp;

	      if (xtal.sf[d].noniso       && (d != sd) )
	      {
		  dLddluz[d][s]           -= (TWO*eps*xtal.sf[d].dluz[s]*dLdvar[d]*
					      xtal.sf[d].sigmanref[s]/var[d]);
		if (updatesigmah          && !xtal.heavyref)
		  dLdsigmah[d][s]         -= dLdvar[d]*eps/var[d];
	      }
	      dLdAp[d][r]                  = (-dLdFp[d]*tab.tcos(xtal.sf[d].pcalcp[r])
					      *temp);
	      dLdBp[d][r]                  = (-dLdFp[d]*tab.tsin2(xtal.sf[d].pcalcp[r])
					      *temp);
	    }
	}
      
      if (product                          > REJECT)
	likelihood                        -= log(product);
      else // A very low likelihood, so as not to confuse the minimizer
      {
	likelihood                        += 700.0;
	if (check)
	{
	  flagged++;
	  xtal.selected[r]                 = false;
	}
      }
    
      if (product                          > REJECT)
      {
	unsigned s0                        = xtal.bin(0,r);
	double fom                         = sqrt(totcos*totcos +
 						  totsin*totsin)/integral;
 	double fb                          = sqrt(totampcos*totampcos +
						  totampsin*totampsin)/integral;
	
	double phib(ZERO), hla(ZERO), hlb(ZERO);
	if (xtal.centric[r])
	{
	  phib                             = ((totcos >= ZERO) ? xtal.centricphase[r] :
					      xtal.centricphase[r] + PI);
	  cfom[s0]                        += fom;
	  for (unsigned d                  = 0; d < dmax  ; d++)
	  {
	    unsigned s                     = xtal.bin(d,r);
	    xtal.sf[d].cisolofc[s]        += isolofc[d]/integral;
	  }

	  if (outputmtz)
	  {
	    double temp                    = atanh(fom);
	    hla                            = temp*tab.tcos(phib);
	    hlb                            = temp*tab.tsin2(phib);
	  }
	}
	else
	{
	  phib                             = atan2(totsin, totcos);
	  afom[s0]                        += fom;
	  for (unsigned d                  = 0; d < dmax  ; d++)
	  {
	    unsigned s                     = xtal.bin(d,r);
	    xtal.sf[d].aisolofc[s]        += isolofc[d]/integral;  
	    if (xtal.sf[d].anomalous)
	      if (xtal.sf[d].anouse(r))
		xtal.sf[d].anolofc[s]     += alofc[d]/integral;
	  }
	  if (outputmtz)
	  {
	    hla                            = totlogcos;
	    hlb                            = totlogsin;
	  }
	}
	stopfom                           += fom;
	nstop++;
	
	if (outputmtz)
	{
	  fdata[inc + 2]                   = (float) fb;
	  fdata[inc + 3]                   = (float) (phib/DEGREEtoRAD);
	  fdata[inc + 4]                   = (float) fom;
	  fdata[inc + 5]                   = (float) hla;
	  fdata[inc + 6]                   = (float) hlb;   
	  fdata[inc + 7]                   = (float) totlogcos2;
	  fdata[inc + 8]                   = (float) totlogsin2;
	}
      }
    }
    else if (outputmtz && (counts         == 1) && xtal.selected[r])
    {
      for (unsigned d                      = 0; d < dmax; d++)
	if (xtal.sf[d].use(r))
	{
	  double fobs, arg(ZERO), fom(CCP4::ccp4_nan().f);
	  unsigned s                       = xtal.bin(d,r);
	  double var                       = std::max(eps*(xtal.sf[d].sigman[s] - 
		   					   xtal.sf[d].sigmah[s]), EPSILON);
   	  if (xtal.sf[d].usep(r))
	  {
	    fobs                           = xtal.sf[d].datap[r];
	    var                           += xtal.sf[d].devp[r]*xtal.sf[d].devp[r];
	    arg                            = TWO*fobs*xtal.sf[d].fcalcp[r]/var;
	  }
	  else
	  {
	    fobs                           = xtal.sf[d].datam[r];
	    var                           += xtal.sf[d].devm[r]*xtal.sf[d].devm[r];
	    arg                            = TWO*fobs*xtal.sf[d].fcalcm[r]/var;
	  }
	
	  if (xtal.centric[r])
	  {
	    fom                            = tanh(arg/TWO);
	    cfom[s]                       += fom;
	    xtal.sf[d].cnshl[s]++;
	  } 	
	  else
	  {
	    fom                            = sim(arg);
	    afom[s]                       += fom;
	    xtal.sf[d].anshl[s]++;
	  }

	  double phib                      = xtal.sf[d].pcalcp[r];

	  fdata[inc + 2]                   = (float) fom*fobs;
	  fdata[inc + 3]                   = (float) (phib/DEGREEtoRAD);
	  fdata[inc + 4]                   = (float) fom;
	  double temp                      = atanh(fom);
	  fdata[inc + 5]                   = (float) temp*cos(phib);
	  fdata[inc + 6]                   = (float) temp*sin(phib);
	  fdata[inc + 7]                   = (float) ZERO;
	  fdata[inc + 8]                   = (float) ZERO;
	}
    }
    if (outputmtz)
      CMtz::ccp4_lwrefl(MTZOUT, &fdata[0], &colout[0], fdata.size(), r+1);
  }

  if (outputmtz)
  {
    if (allin)
      CMtz::MtzFree(MTZIN);

    if (nstop)
      printf("The overall FOM is %.3f\n\n",stopfom/((double)nstop));
    
    CMtz::MtzPut(MTZOUT, " ");
    CMtz::MtzFree(MTZOUT);
  }  
  
  if (xtal.verbose > 1)
    printf("LIKELIHOOD = %f\n", likelihood);
  
  if (check && flagged)
    printf("Number of reflections with low likelihood values: %u\n", flagged);
  
  return likelihood;
}
