C
C     BULKING
C     Copyright (C) 2003 A. Fokine
C
C     This code is distributed under the terms and conditions of the
C     CCP4 Program Suite Licence Agreement as a CCP4 Application.
C     A copy of the CCP4 licence can be obtained by writing to the
C     CCP4 Secretary, Daresbury Laboratory, Warrington WA4 4AD, UK.
C
C#################################################################################
C
C  Program description:
C
C  Patch for AmoRe (CCP4 v4.1)
C  to make 
C  BULK SOLVENT CORRECTION DURING TRANSLATION SEARCH IN MOLECULAR REPLACEMENT
C 
C  Program calculates solvent mask from atomic coordinates,then it calculates 
C  solvent structure factors as:
C                      Fsolv=ksol*exp(-Bsol/(4d**2))*FT(mask)
C  and adds Fsolv to the structure factors calculated from atomic model     
C
C#################################################################################
C#################################################################################
C  A. Fokine, G. Capinani, M.G. Grutter, A. Urzhumtsev  
C version  25.07.02
C######################
C
C
C  Program can be run in dialog or in batch mode.
C
C
C  Input parameters: 
C                    1: inp_coord_file - file with atomic coordinates (x,y,z,B)
C                       example: model.pdb 
C
C                    2: inp_F_file - file with structure factors calculated 
C                       from the atomic model by AmoRe 
C                        example: search.tab 
C                     
C                    3: out_F_corrected_file - file with corrected structure 
C                                               factors Fcorr = Fatom + Fsolvent
C                        example: search.tabs
C
C                    4: resolution - resolution for mask calculation (max grid 
C                                    step will be 1/3*resolution
C                       
C
C                    5: ksol 
C                    6: Bsol - solvent parameters for flat solvent model
C                              it is recomended to use ksol=0.35 e/A3, Bsol=50 A2
C
C                    7: rsolv - radius of the solvent molecule 
C 
C                    8: xtrans,ytrans,ztrans - vector to translate molecule such
C                                              that the center of mass will be
C                                              at the origin
C
C                    9: theta1,theta2,theta3 - euler angles to rotate model such
C                                              that the axes of inertia will
C                                              parallel to coordinate axes.
C                     
C                     
C            Output:
C                    1: out_F_corrected_file - file with corrected structure 
C                                               factors Fcorr = Fatom + Fsolvent                       
C
C
C
C
C
C##################################################################################
C
C    Algorithm:
C               Subroutine READ_INPUT_DATA reads input parameters
C               ( inp_coord_file, inp_F_file, out_F_corrected_file,
C                 resolution,
C                 ksol,Bsol,rsolv,
C                 xtrans,ytrans,ztrans,theta1,theta2,theta3 ) 
C                 from STANDARD INPUT.
C               FROM the file inp_F_file it takes unit cell parameters,
C               and maximal reflection indexes ( hsup,ksup,lsup ) 
C               
C
C               Subroutine DEFINE_GRID defines the grid numbers NX NY NZ in 
C               accordance with resolution, grid and cell parameters. 
C               NX will be even, the maximal multiplier of each number will not be
C               greater 19 (for FFT)      
C 
C               Subroutine READ_PDB reads atomic coord. of the model. format of coord
C               file must be done in its header (for ex. see xyz#.d)
C
C               Subroutine TRANSLATE_ROTATE transforms the model coordinate such that 
C               the center of mass of the model will be in the origin and the axes of 
C               inertia of the model will be parallel to the coord. axes
C
C               Subroutine ATOMIC_RADII sets the VDW radii of the atoms
C
C               
C               Subroutine CARTHESIANS_TO_FRACTIONAL calculates fractional coordinates 
C               from carthesian coordinates.
C
C               Subroutine MOVE_ATOMS_TO_UNIT_CELL moves atoms to the unit cell 
C               (fractional coordinates will belong to the interval 0..1)
C
C               Subroutine ACCESSIBLE_MASK_CALCULATION calculates solvent mask (all
C               points in the solvent region will be set to 1, all points in the 
C               molecule region will be set to 0
C 
C               Subroutine CALL SOLVENT_STRUCTURE_FACTOR_CALCULATION calculates 
C               solvent structure factors by Fourier transform of the mask 
C               (array af structure factors (-1:hsup,-ksup:ksup,-lsup:lsup))
C
C               Subroutine READ_MODEL_STR_FACTORS reads structure factors calculated
C               by AmoRe for the atomic model (and puts them to 'mask' array)
C
C               Subroutine  ADD_FSOLV_TO_FMODEL adds solvent to structure factors 
C               of atomic model
C 
C               Subroutine WRITE_REFLECTIONS write file with corrected structure 
C               factors
C               
C 
C##################################################################################       
C##################################################################################
C##################################################################################



	IMPLICIT NONE


        integer maxNgrid              ! Max. possible number of grid points     
        parameter (maxNgrid=20000000) ! in x, y, and z directions 
                                     
                                     
        integer maxrefl               ! max. possible number 
        parameter (maxrefl=3000000)   ! of reflections
                                      

        integer maxatm                ! max possible number of atoms
        parameter (maxatm=100000)     ! 
                                      
        
        real mask(maxNgrid)           ! array with solvent mask, it will be
                                      ! also used to store the reflections
                                                                                

        real Fsolv(maxrefl)                      ! array of modules of solvent
                                                 ! structure factors
                                                 
        real PHASEsolv(maxrefl)                  ! array of solvent phases

 
        integer nrefl                            ! number of reflections


        integer ih(maxrefl),ik(maxrefl),il(maxrefl)   !arrays with reflection
                                                        !reflection indexes

        integer hsup,lsup,ksup                          !max reflection indexes 
 
                                                          
        real x(maxatm),y(maxatm),z(maxatm)              !arrays of atomic coord.
                                                           

        character*4 atmname(maxatm)    ! array of atomic names
                                       ! symbols 13:16 in PDB file
                                       ! atom types are in atmname(i)(2:2)


        real atmrad(maxatm)            ! array of atomic radii 

        integer natm                   ! number of atoms


        real a,b,c,alpha,beta,gamma   ! cell parameters


        real resolution                ! resolution for mask calculation

        real grid
        parameter (grid=0.3333333)     ! max grid step will be 
                                       ! grid*resolution

        real ksol, Bsol                ! bulk solvent parameters
                                       ! ( see Jiang and Brunger 
                                       !   J.Mol.Biol 1994, 243, 100-115 )


        real rsolv                     ! parameter for solvent mask calculation
                                       ! ( see Jiang and Brunger 
                                       !   J.Mol.Biol 1994, 243, 100-115 )
        
 
        integer NX, NY, NZ             ! number of grid points in x,y and z 
                                       ! directions (ALL U. CELL)


        real xtrans,ytrans,ztrans      !translation vector to put center mass of
                                       !the molecule at the origin (for AmoRe) 

        real theta1,theta2,theta3      !euler angles (around Z, newY, newZ) to
                                       !rotate molecule such that after rotaion the
                                       !inertia axes be parallel to coord. axes  

        character*200  inp_coord_file         ! name of the file with model

        character*200  inp_F_file             ! name of file with str. factors
                                             ! Amore             


        character*200  out_F_corrected_file   ! File with str. factors corrected
                                             ! by solvent contribution - result 
                                             ! of this program 


        integer ncode                  ! ort. code note used (for AmoRe)


        real  sqhsup                   ! not used (for AmoRe) 

        integer lenstr


C_____________________________________________________________________________

      call ccpfyp
      call ccprcs (6, 'BULKING', '$Date: 2008/05/14 09:16:16 $')

      print *
      print *
      print *
      print *
      print *,'######################################################'
      print *,'######################################################'
      print *,'######################################################'
      print *,'######################################################'
      print *,'######################################################'
      print *
      print *
      print *,'                   Program BULK                      '
      print *
      print *,'######################################################'
      print *,'version 1   25.07.02'
      print *,'Authors:   A. Fokine*, G. Capitani+,'
      print *,'           M.G. Grutter+, A. Urzhumtsev*'
      print *,'*University of Nancy 1, France'
      print *,'+University of Zurich, Switzerland'
      print *,'Contacts : fokine@lcm3b.uhp-nancy.fr'
      print *,'           capitani@bioc.unizh.ch'
      print *,'           sacha@lcm3b.uhp-nancy.fr'
      print *,'######################################'
      print *
      print *,'  Program BULK realises the bulk solvent correction '
      print *,
     * 'in the translation search and in the rigid body refinement'  
      print *,'in AMoRe (Navaza, J. (1994). Acta Cryst. A50,157-163) '
      print *
      print *



C      Reading input data

       CALL READ_INPUT_DATA(inp_coord_file,inp_F_file,
     * out_F_corrected_file,
     * ksol,Bsol,rsolv,
     * xtrans,ytrans,ztrans,theta1,theta2,theta3,
     * resolution,
     * a,b,c,alpha,beta,gamma,
     * hsup,ksup,lsup,ncode,nrefl,
     * maxrefl)
       
 

C     Calculation of NX, NY, NZ  for the grid

        CALL DEFINE_GRID(a,b,c,grid,resolution,NX,NY,NZ,maxNgrid)



C      Reading atomic COORDINATES, B-factors (not used), and atom NAMES 

        CALL READ_PDB(inp_coord_file,maxatm,atmname,x,y,z,
     * natm )



C      Translate and Rotate model. After translation coordinates of model 
C      center of mass of the model will be at (0 0 0). After rotation the 
C      inertia axes of molecule will be parallel to coordinate axes
C         !!! rotation matrix and translation vector will be taken    
C         !!!  from the file tabl#

        CALL TRANSLATE_ROTATE(x,y,z,natm,
     * xtrans,ytrans,ztrans,theta1,theta2,theta3 )
 


C      Setting atomic radii 
 
        CALL ATOMIC_RADII(natm, atmname, atmrad)

 
C      Calculation of fractional atomic coordinates 

        CALL CARTHESIANS_TO_FRACTIONAL(a,b,c,alpha,beta,gamma,
     * natm, x, y, z) 



C      Modification of fractional atomic coordinates to put all atoms
C      in the unit cell (all atoms will have coordinates between 0 an 1)  

        CALL MOVE_ATOMS_TO_UNIT_CELL(natm, x, y, z)



C      Calculation of the mask bounded by ACCESSIBLE SURFACE 
C      All grid points within the a distance (VDW(i) + Rsolv) to any
C      NON HYDROGEN atom will be set to 0. All other points will be set to 1
C       MASK will  be calculated in the calculated in the ASYMMETRIC UNIT
C       other elements of mask array will be meaningless

   
        CALL ACCESSIBLE_MASK_CALCULATION(a,b,c,alpha,beta,gamma,
     * rsolv,natm,x,y,z,atmname,atmrad,
     * NX,NY,NZ,mask)



C       Calculation of tre solvent structure factors 
C       Fsolv=ksol*exp(-Bsol/(4d**2))*FT(mask)

        CALL SOLVENT_STR_FACTOR_CALCULATION(
     * a,b,c,alpha,beta,gamma,ksol,Bsol,mask,
     * NX,NY,NZ,Fsolv,PHASEsolv,
     * nrefl,hsup,ksup,lsup,ih,ik,il)


C      Reading model str. factors (calculated by AmoRe) 

       CALL READ_MODEL_STR_FACTORS(inp_F_file,nrefl,mask)


C      Adding solvent structure factors to Fmodel

       CALL ADD_FSOLV_TO_FMODEL(mask,Fsolv,PHASEsolv,nrefl)


C      Write file with corrected structure factors

      sqhsup = 0.0
       CALL WRITE_REFLECTIONS(mask,a,b,c,alpha,beta,gamma,
     * ncode,hsup,ksup,lsup,sqhsup,nrefl,
     * out_F_corrected_file)

      print *
      print *,'#######################################'
      print *,'corrected structure factors are written'
      print *,'in the file ', 
     +        out_F_corrected_file(1:lenstr(out_F_corrected_file))
      print *,'Use it as TABLE1 in the traing step'
      print *,'#######################################'
      print *
      print *

        
      
       CALL CCPERR(0,'*** NORMAL TERMINATION ***')
       END



C      ######################################################################
C      ######################################################################
C      ######################################################################






C      ######################################################################
C
C       Subroutine READ_INPUT_DATA: 
C
C       inp_coord_file - name of model coor. file 
C       inp_F_file name of model str. factor file (calculated by AmoRe)
C
C       resolution - resolution and grid for mask calculation
C       ksol,Bsol - solvent parameters
C       rsolv - solvent radius
C       trans1,trans2,trans3,th1,th2,th3 translation vector and euler angles 
C                                        to transform model coordinates
C
C       nrefl    number of reflection
C       maxrefl  maximal possible number of reflections
C       a,b,c,alpha,beta,gamma - cell parameters
C       hsup,ksup,lsup - maximal indexes
C
C
C       inp_coord_file, inp_F_file, ksol,Bsol,
C       rsolv, resolution 
C       trans1,trans2,trans3 
C       th1,th2,th3  - will be read from STANDARD INPUT
C
C
C       a,b,c,alpha,beta,gamma, hsup,ksup,lsup - will be read
C                                                 from inp_F_file


       SUBROUTINE READ_INPUT_DATA(inp_coord_file,inp_F_file,
     * out_F_corrected_file,
     * ksol,Bsol,rsolv,
     * trans1,trans2,trans3,th1,th2,th3,
     * resolution,
     * a,b,c,alpha,beta,gamma,
     * hsup,ksup,lsup,ncode,nrefl,
     * maxrefl)



       IMPLICIT NONE

       character*200  inp_coord_file, inp_F_file
       character*200  out_F_corrected_file
       real          ksol,Bsol,rsolv
       real          trans1,trans2,trans3,th1,th2,th3
       real          resolution
       integer       nrefl
       real          a,b,c,alpha,beta,gamma
       integer       ncode,hsup,ksup,lsup
       real          sqhsup
       integer       maxrefl

       character*200 errline
       integer lenstr
        

        read (5,'(a)',err=10) inp_coord_file
        print *
        print *,'File with atomic coordinates: ',
     +   inp_coord_file(1:lenstr(inp_coord_file))


        read (5,'(a)') inp_F_file
        print *
        print *,'File with structure factors calculated by AmoRe; '
     * ,inp_F_file(1:lenstr(inp_F_file))
    

        read (5,'(a)',err=10) out_F_corrected_file
        print *
        print *,'Name of file with corrected structure factors: '
     * ,out_F_corrected_file(1:lenstr(out_F_corrected_file))


        read (5,*,err=10) resolution
        print *
        print *,'Resolution for mask calculation ',resolution
        print *,'Max. grid step will be 0.333*resolution'


        read (5,*,err=10) ksol
        read (5,*,err=10) Bsol
        print *
        print *,'Solvent parameters ksol (in e/A3), Bsol (in A2) '
        print *,
     * ' as defined in Jiang and Brunger J.Mol.Biol 1994, 243, 100-115'
        print *, 'ksol= ',ksol
        print *, 'Bsol= ',Bsol


        read (5,*,err=10) rsolv
        print *
        print *,'Rsolv - parameter for solvent mask calculation'
        print *,'Rsolv=',rsolv


        read (5,*,err=10) trans1,trans2,trans3
        print *          
        print *, 'model will be translated by vector :'
        print *, trans1,trans2,trans3 
 

        read (5,*,err=10) th1,th2,th3
        print *
        print *, 'model will be rotated by euler angles :'
        print *,th1,th2,th3
         

        print *, 'euler angle definition: th1 - around Z'
        print *, '                    th2 - around new Y'
        print *, '                    th2 - around new Z'


         open(unit=1,file=inp_F_file, status='old',form='unformatted')

         read (1,err=20) a,b,c,alpha,beta,gamma,ncode,hsup,
     * ksup,lsup,sqhsup

         nrefl = (hsup+1+1)*(ksup+ksup+1)*(lsup+lsup+1)

         print *
         print *,'Reading structure factors calculated by AmoRE:'
         print *,'File name:',inp_F_file(1:lenstr(inp_F_file))
         print *,'Unit cell Parameters:'
         print *,a,b,c
         print *,alpha,beta,gamma
         print *
         print *,' ncode=',ncode
         print *,' hsup,ksup,lsup:',hsup,ksup,lsup
         print *,' 1.0/sqrt(sqhsup):',1.0/sqrt(sqhsup)
         print *,' Number of reflections:',nrefl
         print *



         if(nrefl.gt.maxrefl) then

          print *,'NUMBER OF REFLECTIONS EXCEEDS',maxrefl
          print *,'PLEASE CHANGE THE HIGH RESOLUTION LIMIT'
          print *,'OF THE SEARCH OR OPEN THE FILE bulking.f,'
          print *,'CHANGE THE PARAMETER maxrefl AND RECOMPILE'

          CALL CCPERR(1,'No. of reflections exceeds PARAMETER maxrefl')

         end if            



         RETURN

10       CALL CCPERR(1,'ERROR in program input')
 
         
20       write(errline,'(A20,A80)') 
     +        'ERROR: reading file ',inp_F_file
         CALL CCPERR(1,errline)
         
 
        END

C      #################################################################








C      #################################################################
C      SUBROUTINE READ_PDB(inpfname,maxatm,atmname,x,y,z,natm)
C
C      Reading atomic COORDINATES, B-factors and atom NAMES 
C 
C      INPUT parameters:
C                        inpfname - PDB file name
C                        maxatm   - maximal possible number of atoms
C                         
C      OUTPUT parameters:  
C      atmname - array of atomic names (symbols 13:16 in PDB file) 
C                atom types are in atmname(i)(2:2). 
C      x, y, z - arrays of CARTHESIAN atomic coordinates 
C      B  - arrayof B-factors 
C      natm - number of atoms in PDB file 
C
C         Only string starting by 'ATOM' or 'HETATM' will be read      
C      _______________________________________________________________________



        SUBROUTINE READ_PDB(inpfname,maxatm,atmname,x,y,z,natm)

        IMPLICIT NONE 

         character*200 inpfname
         integer maxatm
         real x(maxatm), y(maxatm), z(maxatm)
         real xcurr, ycurr, zcurr

         character*4 atmname(maxatm), namecurr
         integer natm

         character*200 errline
         character*6 firstchar
C        character*80 FORMAT_coor_file

C        __________________________         


          open(unit=1,file=inpfname, status='old', form='formatted')  

     
          firstchar='REMARK' 

          dowhile ( (firstchar.eq.'REMARK'.or.firstchar.eq.
     * 'Remark'.or.firstchar.eq.'remark') )
           read(1,'(a6)',end=10,err=20) firstchar
          end do

          natm=0

          dowhile (.true.)
 
          read (1,'(12X,A4,14X,3F8.3,6X,F6.2)',end=10,err=20) 
     * namecurr,xcurr,ycurr,zcurr

          if(namecurr.ne.'    ') then
           natm=natm+1
           atmname(natm)=namecurr
           x(natm)=xcurr
           y(natm)=ycurr
           z(natm)=zcurr
          end if                       
         
      
          end do


10	   close(unit=1)


C          print *,'subroutine READ_PDB normal termination'
          print *
          print *,'number of atoms found in coord. file:'
          print *,natm

          if(natm.le.0) then
           print *,'NO ATOMS FOUND IN COORD. FILE', inpfname
           print *,'PROGRAM TERMINATION'
           CALL CCPERR(1,'No atoms found in coord. file.')
          end if


          if (natm.gt.maxatm) then

           print *, 'NUMBER OF ATOMS IN COORDINATE FILE IS GREATER THAN'
           print *,maxatm
           print *, 'PLEASE OPEN FILE bulking.f,'
           print *,'RESET THE PARAMETER maxatm'
           print *,'AND RECOMPILE'

           CALL CCPERR(1,'No. of atoms exceeds PARAMETER maxatm')

          end if



C         print *,'ATOM LIST'
C         do i=1,natm
C          print *
C          print *,i,atmname(i),x(i),y(i),z(i)
C          print *
C         end do 


           RETURN
          
20         write(errline,'(A31,A80)') 
     +        'ERROR: reading coordinate file ',inpfname
           CALL CCPERR(1,errline)

         END 

C       ########################################################################        
                  







C      ########################################################################
C
C      Translation and Rotation of the model. After translation coordinates  
C      of model center of mass of the model will be at (0 0 0). After rotation 
C      the  inertia axes of molecule will be parallel to coordinate axes
C         !!! rotation matrix and translation vector will be taken    
C         !!!  from the file 'tabl#'
C      
C 
C      INPUT parameters:
C                        x,y,z - atomic coordinates
C                        natm   - number of atoms
C                        vector_trans - translation vector
C                        th1, th2, th3 - euler angles of rotation
C                                     th1 - along Z
C                                     th2 - along new Y  !!!!!  
C                                     th3 - along new Z      
C
C      OUTPUT parameters:  
C                        x,y,z - modified coordinates  
C   
C
C      Internal varianles;
C                         matr_rot(3,3) - rotation matrix 
C                         x_rotated(natm) - array of rotated coord.  
C      _______________________________________________________________________




        SUBROUTINE TRANSLATE_ROTATE(x,y,z,natm,tr1,tr2,tr3,th1,th2,th3)

        IMPLICIT NONE 

         integer natm
         real x(natm), y(natm), z(natm)
         real tr1,tr2,tr3, th1,th2,th3
 
 
         integer i
         real x_rotated, y_rotated, z_rotated 
         real matr_rot(3,3) 
         real pi
C        __________________________         


         pi=3.1415926536

         th1=th1*pi/180
         th2=th2*pi/180
         th3=th3*pi/180

         matr_rot(1,1) = cos(th1)*cos(th2)*cos(th3) 
     *                   - sin(th1)*sin(th3)

         matr_rot(1,2) = (-1)*cos(th1)*cos(th2)*sin(th3)
     *                   - sin(th1)*cos(th3)

         matr_rot(1,3) = cos(th1)*sin(th2)

         matr_rot(2,1) = sin(th1)*cos(th2)*cos(th3) 
     *                   + cos(th1)*sin(th3)

         matr_rot(2,2) = (-1)*sin(th1)*cos(th2)*sin(th3)
     *                    + cos(th1)*cos(th3)

         matr_rot(2,3) = sin(th1)*sin(th2)

         matr_rot(3,1) = (-1)*sin(th2)*cos(th3)

         matr_rot(3,2) = sin(th2)*sin(th3)

         matr_rot(3,3) = cos(th2)


         do i=1,natm
          x(i)=x(i) + tr1
          y(i)=y(i) + tr2
          z(i)=z(i) + tr3         
         end do

           
         do i=1,natm
          
          x_rotated = x(i)*matr_rot(1,1) + y(i)*matr_rot(1,2)
     *                  + z(i)*matr_rot(1,3)
 
          y_rotated = x(i)*matr_rot(2,1) + y(i)*matr_rot(2,2)
     *                  + z(i)*matr_rot(2,3)
 
          z_rotated = x(i)*matr_rot(3,1) + y(i)*matr_rot(3,2) 
     *                  + z(i)*matr_rot(3,3)

          x(i) = x_rotated
          y(i) = y_rotated
          z(i) = z_rotated
 
         end do  

C         print *,tr1,tr2,tr3, th1,th2,th3

C         print *, 'Rotate translate: + termination'
           

         END 

C       ########################################################################        
   



        


C       ########################################################################
C       SUBROUTINE CARTHESIANS_TO_FRACTIONAL(a,b,c,alpha,beta,gamma,
C                                       natm,xort,yort,zort,xfrac,yfrac,zfrac)
C 
C       Calculation of fractional atomic coordinates 
C
C       INPUT parameters:
C           a,b,c,alpha,beta,gamma   - cell parameters
C           natm - number of atoms
C           xort,yort,zort - arrays with CARTHESIAN atomic coordinates
C                         
C       OUTPUT parameters:  
C           xfrac,yfrac,zfrac - arrays with FRACTIONAL atomic coordinates
C
C       Subroutine calls subroutine MATORT written by A. Urzhumtsev
C      _______________________________________________________________________




         SUBROUTINE CARTHESIANS_TO_FRACTIONAL(a,b,c,alpha,beta,gamma,
     * natm,x,y,z)

          IMPLICIT NONE
        
          real a,b,c,alpha,beta,gamma
          integer natm
          real x(natm), y(natm), z(natm)
          real xcurr, ycurr, zcurr

          integer i
C         __________________________________________

        
          do i=1,natm

 
           CALL MATORT(1,a,b,c,alpha,beta,gamma,
     * x(i),y(i),z(i), xcurr, ycurr, zcurr)

           x(i)=xcurr
           y(i)=ycurr
           z(i)=zcurr

          end do



C          print *,'Subroutine CARTHESIANS_TO_FRACTIONAL: + termination'
C          print *,'number of atoms:',i-1

         END 

C       #######################################################################    







C       ########################################################################
C       SUBROUTINE MOVE_ATOMS_TO_UNIT_CELL(natm,xfrac,yfrac,zfrac)
C
C       Subroutine changes fractional atomic coordinate to move all atoms in the
C       same cell. After the procedure all atomic fractional coordinates will be
C       between 0.0 and 1.0 
C  
C      INPUT parameters:
C          natm - number of atoms 
C          xfrac,yfrac,zfrac - arrays with FRACTIONAL atomic coordinates
C                     
C      OUTPUT parameters:
C         xfrac,yfrac,zfrac - modified arrays with FRACTIONAL atomic coordinates
C
C      _______________________________________________________________________          




         SUBROUTINE MOVE_ATOMS_TO_UNIT_CELL(natm,xfrac,yfrac,zfrac)

          IMPLICIT NONE 

          integer natm
          real xfrac(natm), yfrac(natm), zfrac(natm)

          integer i

C         ________________________________________



          do i=1,(natm)

           if(xfrac(i).ge.1.0) xfrac(i)= xfrac(i) - int(xfrac(i))
           if(xfrac(i).lt.0.0) xfrac(i)= 1.0 + xfrac(i) - int(xfrac(i))
          
           if(yfrac(i).ge.1.0) yfrac(i)= yfrac(i) - int(yfrac(i))
           if(yfrac(i).lt.0.0) yfrac(i)= 1.0 + yfrac(i) - int(yfrac(i))
          
           if(zfrac(i).ge.1.0) zfrac(i)= zfrac(i) - int(zfrac(i))
           if(zfrac(i).lt.0.0) zfrac(i)= 1.0 + zfrac(i) - int(zfrac(i))
         

          end do  

         
C         print *,'SUBROUTINE MOVE_ATOMS_TO_UNIT_CELL: + termination'

       
         END        


C       ###################################################################






C       ####################################################################
C       SUBROUTINE ATOMIC_RADII(natm, atmname,atmrad)
C 
C       Set atomic radii
C
C       INPUT parameters:
C          natm - number of atoms 
C          atmname- arrays of atimic names
C                     
C       OUTPUT parameters:
C          atmrad - array of atomic radii
C
C        VDW radii for N,O and S are taken from 
C        CNS file 'protein_rep.param' .
C        radius for C atoms was calculated as:
C                   (vdw_radius_C + vdw_radius_C_SP2)/2
C        where vdw_radius_C_SP2 - radius of SP2 hybridised carbon     
C        and vdw_radius_C - radius of SP3 hybridised carbon
C
C        For other atoms VDWrad=1.6 A  ?????!!!!!
C
C      It should be noted that the radii are quite different from those
C      proposed in Jiang and Brunger J.Mol.Biol 1994, 243, 100-115.
C      
C      _______________________________________________________________________          
       



        SUBROUTINE ATOMIC_RADII(natm, atmname,atmrad)

         IMPLICIT NONE

         integer natm
         character*4 atmname(natm)
         real atmrad(natm)
         integer numH, numO, numN, numC, numS, numother
 
         integer i

C       ________________________________________________


         numH=0 
         numO=0
         numN=0
         numC=0
         numS=0
         numother=0

         do i=1,natm

          if (atmname(i)(2:2).eq.'H') then 
            atmrad(i)=1.2
            numH=numH+1
          else if (atmname(i)(2:2).eq.'C') then 
            atmrad(i)=1.775
            numC=numC+1
          else if (atmname(i)(2:2).eq.'O') then
            atmrad(i)=1.45
            numO=numO+1
          else if (atmname(i)(2:2).eq.'N') then
            atmrad(i)=1.5
            numN=numN+1
          else if (atmname(i)(2:2).eq.'S') then
            atmrad(i)=1.8            
            numS=numS+1
          else
            atmrad(i)=1.6
            numother=numother+1 
          end if
  
         end do 

C         print *,'Subroutine ATOMIC_RADII: + termination'
C         print *
C         print *,'Numbers of different atoms:'
C         print *,' numH=',numH,' numC=',numC,' numO',numO,
C     * ' numS=',numS,' numN=',numN,' numother=',numother 


        END 
          
C     #####################################################################







C     ######################################################################
C
C       SUBROUTINE ACCESSIBLE_MASK_CALCULATION(a,b,c,alpha,beta,gamma,
C     * rsolv,natm,xfrac,yfrac,zfrac,atmname,atmrad,
C     * NX,NY,NZ,mask)
C
C      Calculation of the mask bounded by ACCESSIBLE SURFACE 
C      All grid points within the 
C      d<=(VDW(i)+Rsolv) to any NON HYDROGEN atom will be set to 0; 
C      all other points will be set to 1.0
C
C       INPUT parameters:
C 
C        a,b,c,alpha,beta,gamma - unit cell parameters 
C        rsolv - solvent radius
C        natm - number of atoms (in all unit cell
C        xfrac,yfrac,zfrac - arrays with fractional atomic coordinates 
C        atmname - arrray with atomic names
C        atmrad - array with atomic radii
C        NX,NY,NZ - number of grid points in x,y and z directions (ALL CELL)
C
C       OUTPUT parameters: 
C       
C        mask - array with ACCESSIBLE_MASK    
C
C
C       ____________________________________________________________________ 

 


       SUBROUTINE ACCESSIBLE_MASK_CALCULATION(a,b,c,alpha,beta,gamma,
     * rsolv,natm,xfrac,yfrac,zfrac,atmname,atmrad,
     * NX,NY,NZ,mask)


        IMPLICIT NONE

        real a,b,c,alpha,beta,gamma
        real rsolv
        integer natm
        real xfrac(natm),yfrac(natm),zfrac(natm)
        character*4 atmname(natm)
        real atmrad(natm)
        integer NX,NY,NZ
   
        real mask(NX,NY,NZ)

        integer kx,ky,kz,mx,my,mz
        integer x1box,x2box,y1box,y2box,z1box,z2box

        real ascale,bscale,cscale
        real cosa,cosb,cosg,sina,sinb,sing
        real cutoff
        real dist  
 
        real pi
        parameter (pi=3.1415926536)  


        integer i,counter
        real percentage
        real xdelta, ydelta, zdelta
        logical point_in_unit_cell



C       ________________________________


C        print *,a,b,c,alpha,beta,gamma

        cosa=cos(alpha*pi/180)
        cosb=cos(beta*pi/180)
        cosg=cos(gamma*pi/180)
        sina=sin(alpha*pi/180)
        sinb=sin(beta*pi/180)
        sing=sin(gamma*pi/180)


C       Parameters which will be used to calculation of boxes      
         
        ascale=sqrt(1.0-cosa**2-cosb**2-cosg**2+2.0*cosa*cosb*cosg)/sina
        bscale=sqrt(1.0-cosa**2-cosb**2-cosg**2+2.0*cosa*cosb*cosg)/sinb
        cscale=sqrt(1.0-cosa**2-cosb**2-cosg**2+2.0*cosa*cosb*cosg)/sing


C       _______________________________________________________________________
C      Calculation of ACCESSIBLE mask in the ASYMMETRIC UNIT. Each grid point of
C      the mask within a distance VDWrad to any non-hydrogen atom will
C      be set to -1.0 and points within a distance VDWrad<d<VDWrad+Rsolv to 0.0 
C      _______________________________________________________________________

C      _______________________________________________________________________       
C      SETUP
C            all array mask ! 

       do kx=0,NX-1
         do ky=0,NY-1
          do kz=0,NZ-1
            mask(kx+1,ky+1,kz+1)=1.0
          end do
         end do
        end do 

 
        do i=1,natm

         if( atmname(i)(2:2).ne.'H' ) then
    
          cutoff=atmrad(i)+rsolv

C         Define box around the atom
         
           x1box=NINT( NX*(xfrac(i)-cutoff/a/ascale) ) - 1 
           x2box=NINT( NX*(xfrac(i)+cutoff/a/ascale) ) + 1

           y1box=NINT( NY*(yfrac(i)-cutoff/b/bscale) ) - 1 
           y2box=NINT( NY*(yfrac(i)+cutoff/b/bscale) ) + 1

           z1box=NINT( NZ*(zfrac(i)-cutoff/c/cscale) ) - 1 
           z2box=NINT( NZ*(zfrac(i)+cutoff/c/cscale) ) + 1


C         Check all points in the box

           do kx=x1box,x2box
            do ky=y1box,y2box
             do kz=z1box,z2box


               xdelta= xfrac(i) - float(kx)/NX 
               ydelta= yfrac(i) - float(ky)/NY
               zdelta= zfrac(i) - float(kz)/NZ
 

               dist=sqrt(   
     *    a*a*xdelta*xdelta 
     * +  b*b*ydelta*ydelta  
     * +  c*c*zdelta*zdelta
     * + 2*a*b*cosg * xdelta * ydelta
     * + 2*a*c*cosb * xdelta * zdelta
     * + 2*b*c*cosa * ydelta * zdelta  
     *                )
 
C     Set all neigbour points of the mask within the 
C     atmrad(i) distance from the atomic center to -1.0 
C     and the points within atmrad(i)<d<atmrad(i)+rsolv to 0.0    

              if (dist.le.cutoff) then     
     
               point_in_unit_cell=(
     *   kx.lt.NX .and. kx.ge.0
     *  .and. ky.lt.NY .and. ky.ge.0
     *  .and. kz.lt.NZ .and. kz.ge.0 )


              if( point_in_unit_cell ) then               
               mask(kx+1,ky+1,kz+1)=0.0                              
              end if

C            ______ 

              if ( .not.point_in_unit_cell ) then  !the point is out 
                                                   !of the unit cell                  
                mx=kx
                my=ky
                mz=kz

                if (mx.ge.NX) mx= mx - int(mx/NX)*NX
                if (mx.lt.0)  mx= NX + mx - int(mx/NX)*NX
               
                if (my.ge.NY) my= my - int(my/NY)*NY
                if (my.lt.0)  my= NY + my - int(my/NY)*NY

                if (mz.ge.NZ) mz= mz - int(mz/NZ)*NZ
                if (mz.lt.0)  mz= NZ + mz - int(mz/NZ)*NZ 
  
                 mask(mx+1,my+1,mz+1)=0.0    

               end if                   !end the point is out u. cell       
 
              end if                   !end dist le cutoff

             end do
            end do                      !end cycles other box
           end do         


         end if                        !end atom is not hydrogen

        end do                         !cycle other all atoms


C      Calculation of the persantage  volume bounded
C      by accessible surface

        counter=0

        do kx=0,NX-1
         do ky=0,NY-1
          do kz=0,NZ-1
           if (mask(kx+1,ky+1,kz+1).eq.1.0) counter=counter+1
          end do
         end do
        end do

        percentage=float(counter)
     * /(NX*NY*NZ)

C        print *,'SUBROUTINE ACCESSIBLE_MASK_CALCULATION: +termination'
C        print *,'% of volume occupied by solvent:'
c     * ,percentage


        counter=0

        do kx=0,NX-1
         do ky=0,NY-1
          do kz=0,NZ-1
           if (mask(kx+1,ky+1,kz+1).eq.0.0) counter=counter+1
          end do
         end do
        end do

        percentage=float(counter)
     * /(NX*NY*NZ)

C        print *,'SUBROUTINE ACCESSIBLE_MASK_CALCULATION: +termination'
C        print *,'% of volume occupied by atoms:'
C     * ,percentage


       END

C     ########################################################################

 







C######################################################################
C       SUBROUTINE SOLVENT_STR_FACTOR_CALCULATION(
C     * a,b,c,alpha,beta,gamma,ksol,Bsol,mask,
C     * NX,NY,NZ,Fsolv,PHASEsolv,
C     * nrefl,hsup,ksup,lsup,ih,ik,il) 
C
C
C    Calculation of solvent structure Factors (group P1) 
C      All reflections in the resolution range inf - resolution will
C      be calculated (not only uniques reflections). 
C                    F = ksol*FFT(mask)*exp(-Bsol*s*s/4) 
C      
C  Subroutine SFP1 written by V.Yu. Lunin is used to calculate FFT(mask)
C
C      INPUT parameters:
C         a,b,c,alpha,beta,gamma - cell parameters
C         ksol,Bsol - bulk solvent parameters
C         mask - array with solvent mask
C         NX,NY,NZ - number of grid points in x,y,z directions
C         hsup,ksup,lsup - maximal reflection indexes
C
C      OUTPUT parameters:
C         ih,ik,il - arrays containing reflection indexes
C         Fsolv,PHASEsolv - arrays with structure factor modules
C                            and phases
C________________________________________________________________________________________ 



       SUBROUTINE SOLVENT_STR_FACTOR_CALCULATION(
     * a,b,c,alpha,beta,gamma,ksol,Bsol,mask,
     * NX,NY,NZ,Fsolv,PHASEsolv,
     * nrefl,hsup,ksup,lsup,ih,ik,il)


       IMPLICIT NONE
     
        real a,b,c,alpha,beta,gamma
        real ksol,Bsol
        integer NX,NY,NZ
        integer hsup,ksup,lsup

        real mask((NX+2)*NY*NZ)
         
        integer nrefl
         
        integer ih(nrefl), ik(nrefl), il(nrefl)
       
        real Fsolv(nrefl),PHASEsolv(nrefl)
      

        real V
     
        real cosa,cosb,cosg,sina,sinb,sing
       
        real pi, s2

        integer i,h,k,l
        
C       _______________________________________________

        pi=3.1415926536


        cosa=cos(alpha*pi/180)
        cosb=cos(beta*pi/180)
        cosg=cos(gamma*pi/180)
        sina=sin(alpha*pi/180)
        sinb=sin(beta*pi/180)
        sing=sin(gamma*pi/180)
             
        do i=1,nrefl
         Fsolv(i)=0.0
         PHASEsolv(i)=0.0
        end do


C        print *,'STR.fac.cal.',hsup,ksup,lsup,nrefl

        i=0

        do l=-lsup, lsup
         do k=-ksup, ksup    
          do h=-1, hsup

          i=i+1 

          ih(i)=h
          ik(i)=k
          il(i)=l
        
          end do
        end do
       end do   

C      print *, 'program terminated',i
         

 
        V=a*b*c*sqrt( 1.0-cosa**2-cosb**2-cosg**2
     * +2*cosa*cosb*cosg )

            
C       Call Lunin's subroutine

C        open(unit=2,file='lunin-out.txt',
C     *status='unknown',form='formatted')

C        print *,'Lunin input',((NX+2)*NY*NZ),
C     * NX,NY,NZ,V,nrefl


        CALL SFP1(mask,((NX+2)*NY*NZ),
     * NX,NY,NZ,V,
     * nrefl,ih,ik,il,Fsolv,PHASEsolv,2)


C       print *,'SUBROUTINE SFP1: +termination'



C       Scaling solvent structure factors 

        do i=1,nrefl

        h=ih(i)
        k=ik(i)
        l=il(i)

         s2=(h**2*(b*c*sina)**2 + k**2*(a*c*sinb)**2 + 
     * l**2*(a*b*sing)**2 + 
     * 2*h*k*a*b*c**2*(cosa*cosb-cosg) +
     * 2*k*l*b*c*a**2*(cosb*cosg-cosa) +
     * 2*l*h*a*c*b**2*(cosg*cosa-cosb))/(V**2)
     
          
         Fsolv(i)=ksol*Fsolv(i)*exp(-Bsol*s2/4.0)

        end do     

C        print *,'SOLVENT_STRUCTURE_FACTOR_CALCULATION +termination'  

       do i=1,nrefl 
        if(ih(i).eq.0.and.ik(i).eq.0.and.il(i).eq.0) then
C         print *,'F(0,0,0)=',Fsolv(i),PHASEsolv(i)
        end if 
       end do 

       END

C###################################################################








C###################################################################
C      SUBROUTINE READ_MODEL_STR_FACTORS reads structure factors 
C      calculated for the model by AmoRe to array mask
C
C      inp_F_file - file with model str. factors
C      nrefl - number of reflections


       SUBROUTINE READ_MODEL_STR_FACTORS (inp_F_file,nrefl,mask)

    
        IMPLICIT NONE

        character*200   inp_F_file
        integer        nrefl
        complex        mask(nrefl)


        real      a,b,c,alpha,beta,gamma
        integer   ncode,hsup,ksup,lsup
        real      sqhsup

        integer lenstr

        open(unit=1,file=inp_F_file, status='old',form='unformatted')

        read (1) a,b,c,alpha,beta,gamma,ncode,hsup,ksup,lsup,sqhsup

        print *,'Reading structure factors calculated by AmoRE'
        print *,'File name:',inp_F_file(1:lenstr(inp_F_file))
        print *,'Unit cell Parameters:'
        print *,a,b,c
        print *,alpha,beta,gamma
        print *
        print *,'ncode=',ncode
        print *,'hsup,ksup,lsup:',hsup,ksup,lsup
        print *,'1.0/sqrt(sqhsup):',1.0/sqrt(sqhsup)


        read (1) mask

   
C        print *, 'SUBROUTINE READ_F_FILE: +termination'


       END


C     ##################################################################







C     ##################################################################
C      SUBROUTINE ADD_FSOLV_TO_FMODEL adds solvent str. factors to model
C      str. factors
C
C      mask - array with model str. factors
C      Fsolv,PHASEsolv - arrays with moduls and phases of solvent str.fact
C      nrefl - number of reflections

 
       SUBROUTINE ADD_FSOLV_TO_FMODEL(mask,Fsolv,PHASEsolv,nrefl)


        IMPLICIT NONE

        integer nrefl
        complex mask(nrefl)
        real Fsolv(nrefl),PHASEsolv(nrefl)
        real rF, imF 
        complex cF
        integer i


        do i=1,nrefl

         rF=Fsolv(i)*cos(PHASEsolv(i))
         imF=Fsolv(i)*sin(PHASEsolv(i))

         cF=cmplx(rF,imF)
 
         if( mask(i) .ne. (0.0, 0.0) ) then
          mask(i)=mask(i) + cF 
         end if

        end do 

C       print *,'ADD_REFLECTION: + termination'

 
       END


C##################################################################







C##################################################################
C            Write reflection file in AmoRe format
C
                          


       SUBROUTINE WRITE_REFLECTIONS(mask,a,b,c,alpha,beta,gamma,
     * ncode,hsup,ksup,lsup,sqhsup,nrefl,
     * out_F_corrected_file)
       
       IMPLICIT NONE
      

       integer  nrefl,hsup,ksup,lsup,ncode
       complex mask(nrefl)
       character*200 out_F_corrected_file
       real a,b,c,alpha,beta,gamma
       real sqhsup


       integer h,k,l, i
       real F,phase,rF,imF
       
C      __________________________________________



C       open(unit=1,file='fsolv',
C     * status='unknown',form='formatted')


       open(unit=2,file=out_F_corrected_file,
     * status='unknown',form='unformatted')


C       write(1,'(a27)') 'REMARKS  h, k, l, F,  phase'
C       write(1,'(a37,a18)') 'DECLARE name=FPART domain=RECIPROCAL ',
C     * 'type=COMPLEX  end'
           
 
       i=0

        do l=-lsup, lsup
         do k=-ksup, ksup    
          do h=-1, hsup

           i=i+1

           rF=real(mask(i))
           imF=aimag(mask(i)) 

           if(abs(rf).gt.0.0 .or. abs(imf).gt.0.0) then

            F=sqrt(rF*rF + imF*imF)
            phase=atan2(imF,rF)

           else 

            F=0.0
            phase=0.0

           end if

C           print *,F, phase


C           if(.not.(h.eq.0.and.k.eq.0.and.l.eq.0)) then

C            write(1,'(a6,3I5,1X,a6,2F10.3)')  ' INDE ', h, k, l,
C     * 'FPART=', F, phase*180/3.1415926536              

C           end if

          end do
         end do
        end do   


       write (2) a,b,c,alpha,beta,gamma,
     * ncode,hsup,ksup,lsup,sqhsup

   
       write (2) mask 
   

C       print *,'WRITE_REFLECTIONS +termination'
 
      END


C##################################################################







C##################################################################
c****************************************************************	
c 	CARTESIAN TO FRACTIONAL COOR
c****************************************************************
c     *   ' ------------------------------------------------- '/
c     *   ' 1 for X // a    , Z // c* '/
c     *   ' 2 for X // b    , Z // a* '/
c     *   ' 3 for X // c    , Z // b* '/
c     *   ' 4 for X // (a+b), Z // c* '/
c     *   ' 5 for X // a*   , Z // c  '/
c     *   ' 6 for X // a*   , Y // b  ')
C===============================================================
      SUBROUTINE matort(north,acell,bcell,ccell,alpha,beta,gamma,
     *                      xx,yy,zz,xcurr,ycurr,zcurr)
c
c      calculation of the orthogonalisation/deorthogonalisation
c      matrices (kept by columns)
c
	IMPLICIT NONE
	integer north
        real acell,bcell,ccell,alpha,beta,gamma,xx,yy,zz
        real xcurr,ycurr,zcurr
	real*8 orth(9),deorth(9),grrad
        real*8 chb,shb,chg,shg
	real*8 alpr,betr,gamr,calr,salr,cber,sber,cgar,sgar,cha,sha
	real*8 chi,cchi,schi,cdg,sdg,cdb,sdb


C       print *, 'MATORT CALLED'
C       print *, north,acell,bcell,ccell,alpha,beta,gamma,
C     *xx,yy,zz,xcurr,ycurr,zcurr

c
c      data preparation
c
       grrad=3.14159265/180.0
c
       alpr=alpha*grrad
       betr=beta*grrad
       gamr=gamma*grrad
c
c      deorthogonalization matrix (Angstroms -> fract.coord)
c
       calr=cos(alpr)
       salr=sin(alpr)
       cber=cos(betr)
       sber=sin(betr)
       cgar=cos(gamr)
       sgar=sin(gamr)
c
       cha=(calr-cber*cgar)/(sber*sgar)
       if(cha.gt. 1.0) cha= 1.0
       if(cha.lt.-1.0) cha=-1.0
       sha=sqrt(1.-cha*cha)
c
       chb=(cber-calr*cgar)/(salr*sgar)
       if(chb.gt. 1.0) chb= 1.0
       if(chb.lt.-1.0) chb=-1.0
       shb=sqrt(1.-chb*chb)
c
       chg=(cgar-calr*cber)/(salr*sber)
       if(chg.gt. 1.0) chg= 1.0
       if(chg.lt.-1.0) chg=-1.0
       shg=sqrt(1.-chg*chg)
c
c      calcualtion of the matrices
c
       if(north.eq.1) then
c
         deorth(1)=1./acell
         deorth(2)=0.0
         deorth(3)=0.0
         deorth(4)=-deorth(1)*cgar/sgar
         deorth(5)=1./(bcell*sgar)
         deorth(6)=0.0
         deorth(7)=-deorth(1)*salr*chb/(sber*sgar*sha)
         deorth(8)=-deorth(5)*cha/sha
         deorth(9)=1./(ccell*sber*sha)
c
c        orthogonalization matrix (fract.coord -> angstroms)
c
         orth(1)=acell
         orth(2)=0.0
         orth(3)=0.0
         orth(4)=bcell*cgar
         orth(5)=bcell*sgar
         orth(6)=0.0
         orth(7)=ccell*cber
         orth(8)=ccell*sber*cha
         orth(9)=ccell*sber*sha
c
       else if(north.eq.2) then
c
         deorth(1)=0.0
         deorth(2)=1./bcell
         deorth(3)=0.0
         deorth(4)=0.0
         deorth(5)=-deorth(2)*calr/salr
         deorth(6)=1./(ccell*salr)
         deorth(7)=1./(acell*sgar*shb)
         deorth(8)=-deorth(2)*sber*chg/(salr*sgar*shb)
         deorth(9)=-deorth(6)*chb/shb
c
c        orthogonalization matrix (fract.coord -> angstroms)
c
         orth(1)=acell*cgar
         orth(2)=acell*sgar*chb
         orth(3)=acell*sgar*shb
         orth(4)=bcell
         orth(5)=0.0
         orth(6)=0.0
         orth(7)=ccell*calr
         orth(8)=ccell*salr
         orth(9)=0.0
c
       else if(north.eq.3) then
c
         deorth(1)=0.0
         deorth(2)=0.0
         deorth(3)=1./ccell
         deorth(4)=1./(acell*sber)
         deorth(5)=0.0
         deorth(6)=-deorth(3)*cber/sber
         deorth(7)=-deorth(4)*chg/shg
         deorth(8)=1./(bcell*salr*shg)
         deorth(9)=-deorth(3)*sgar*cha/(salr*sber*shg)
c
c        orthogonalization matrix (fract.coord -> angstroms)
c
         orth(1)=acell*cber
         orth(2)=acell*sber
         orth(3)=0.0
         orth(4)=bcell*calr
         orth(5)=bcell*salr*chg
         orth(6)=bcell*salr*shg
         orth(7)=ccell
         orth(8)=0.0
         orth(9)=0.0
c
       else if(north.eq.4) then
c
         chi=atan2(bcell*sgar,acell+bcell*cgar)
         cchi=cos(chi)
         schi=sin(chi)
         cdg=cos(gamr-chi)
         sdg=sin(gamr-chi)
         cdb=cos(betr-chi)
         sdb=sin(betr-chi)
c
         deorth(1)=sdg/(acell*sgar)
         deorth(2)=schi/(bcell*sgar)
         deorth(3)=0.0
         deorth(4)=-cdg/(acell*sgar)
         deorth(5)=cchi/(bcell*sgar)
         deorth(6)=0.0
         deorth(7)=-salr*chb/(acell*sber*sgar*sha)
         deorth(8)=-1.*cha/(bcell*sgar*sha)
         deorth(7)=1./(ccell*sber*sha)
c
c        orthogonalization matrix (fract.coord -> angstroms)
c
         orth(1)=acell*cchi
         orth(2)=-acell*schi
         orth(3)=0.0
         orth(3)=bcell*cdg
         orth(4)=bcell*sdg
         orth(6)=0.0
         orth(7)=ccell*(cdb    -sber*schi*(1.-cha))
         orth(8)=ccell*(sdb*cha-cber*schi*(1.-cha))
         orth(9)=ccell* sber*sha
c
       else if(north.eq.5) then
c
         deorth(7)=0.0
         deorth(8)=0.0
         deorth(9)=1./ccell
         deorth(4)=0.0
         deorth(5)=1./(bcell*salr)
         deorth(6)=-deorth(9)*calr/salr
         deorth(1)=1./(acell*sber*shg)
         deorth(2)=-deorth(5)*chg/shg
         deorth(3)=-deorth(9)*sgar*chb/(salr*sber*shg)
c
c        orthogonalization matrix (fract.coord -> angstroms)
c
         orth(7)=0.0
         orth(8)=0.0
         orth(9)=ccell
         orth(4)=0.0
         orth(5)=bcell*salr
         orth(6)=bcell*calr
         orth(1)=acell*sber*shg
         orth(2)=acell*sber*chg
         orth(3)=acell*cber
c
       else if(north.eq.6) then
c
         deorth(4)=0.0
         deorth(5)=1./bcell
         deorth(6)=0.0
         deorth(7)=0.0
         deorth(8)=-deorth(5)*calr/salr
         deorth(9)=1./(ccell*salr)
         deorth(1)=1./(acell*sgar*shb)
         deorth(2)=-deorth(5)*sber*chg/(salr*sgar*shb)
         deorth(3)=-deorth(9)*chb/shb
c
c        orthogonalization matrix (fract.coord -> angstroms)
c
         orth(1)=acell*sgar*shb
         orth(2)=acell*cgar
         orth(3)=acell*sgar*chb
         orth(4)=0.0
         orth(5)=bcell
         orth(6)=0.0
         orth(7)=0.0
         orth(8)=ccell*calr
         orth(9)=ccell*salr
c
       endif
c
	xcurr = deorth(1)*xx + deorth(4)*yy + deorth(7)*zz
	ycurr = deorth(2)*xx + deorth(5)*yy + deorth(8)*zz
	zcurr = deorth(3)*xx + deorth(6)*yy + deorth(9)*zz
c

C       print *,'MATORT FINISHED'

       return
       end
c***********************************************************************
C#######################################################################







C########################################################################


       SUBROUTINE DEFINE_GRID(a,b,c,
     * grid,resolution,NX,NY,NZ,maxNgrid)


      real a,b,c
      real grid,resolution
      integer NX, NY, NZ
      integer maxNgrid

      dimension NPRIME(160)
  
      DATA NPRIME / 23, 29, 31, 37, 41, 43, 47, 53, 59, 61,
     =              67, 71, 73, 79, 83, 89, 97,101,103,107,
     =             109,113,127,131,137,139,149,151,157,163,
     =             167,173,179,181,191,193,197,199,211,223,
     =             227,229,233,239,241,251,257,263,269,271,
     =             277,281,283,293,307,311,313,317,331,337,
     =             347,349,353,359,367,373,379,383,389,397,
     =             401,409,419,421,431,433,439,443,449,457,
     =             461,463,467,479,487,491,499,503,509,521,
     =             523,541,547,557,563,569,571,577,587,593,
     =             599,601,607,613,617,619,631,641,643,647,
     =             653,659,661,673,677,683,691,701,709,719,
     =             727,733,739,743,751,757,761,769,773,787,
     =             797,809,811,821,823,827,829,839,853,857,
     =             859,863,877,881,883,887,907,911,919,929,
     =             937,941,947,953,967,971,977,983,991,997/
C
      DATA NFGI,NFLGI,NFRGI,NFRQI,NFRRI,NFRFI,NFDATA,NFGOUT,NFLOUT
     =   /  11 ,  12 ,  13 ,  14 ,  15 ,  16 ,  10  ,  17  ,  18   /
      DATA LPRIME / 160 /
      DATA NFTIME /  30 /
      DATA NFPRNT / 6 /
      DATA LTITLE / 18 /
      DATA LTYPE,LTYPAR / 10 , 5 /

      NX=NINT(a/(grid*resolution))
      NY=NINT(b/(grid*resolution))
      NZ=NINT(c/(grid*resolution))


      if(2*INT(NX/2).ne.NX) NX=NX+1       !NX must be even for FFT


C      print *, 'INITIAL NX,NY,NZ:'
C      print *, NX, NY, NZ

      CALL  GRIDPN(NX,NY,NZ,2,1,1,1,nprime,lprime)

      print *
      print *,' NX, NY, NZ for mask calculation:'
C      print *,'FINAL NX,NY,NZ:'
      print *, NX, NY, NZ

      if(((NX+2)*NY*NZ) .gt. maxNgrid) then

       print *
       print *,'THE NUMBER OF GRID POINTS EXCEEDS ',maxNgrid  
       print *,'PLEASE CHANGE THE HIGH RESOLUTION'
       print *,'LIMIT OF THE SEARCH OR OPEN THE FILE bulking.f,'
       print *,'CHANGE THE PARAMETER maxNgrid AND RECOMPILE'

       CALL CCPERR(1,'No. of grid points exceeds PARAMETER maxNgrid')

      end if

      END



C---------------------------------------------------------------------
      SUBROUTINE GRIDPN(N1,N2,N3,NS1,NS2,NS3,JCORR,nprime,lprime)
C
C                   F R O G 2
C     FIND NUMBERS OF GRID POINTS
C
C     VERSION   1.1        CORRECTION  20.02.1987
C
C**********************************************************************
C
C     PARAMETERS :
C
C     N1,N2,N3    - INITIAL NUMBERS
C     NS1,NS2,NS3 - STEPS FOR NUMBER INCREMENT
C     JCORR       - FLAG OF NUMBER CORRECTION
C
C*******************
C
C     PARAMETERS FROM COMMON BLOCKS
C
C     INTX   - BUFFER ARRAY
C     LX     - ITS LENGTH
C     NPRIME - ARRAY OF PRIME NUMBERS LESS THAN 1000
C     LPRIME - ITS LENGTH
C
C**********************************************************************
C
        dimension INTX(1000)
        dimension nprime(lprime)
c
        MPMAX=NPRIME(LPRIME)
        LPMAX=0
C
C     FIND FORBIDDEN NUMBERS FOR GRID POINTS
C
C     ******
C     *
          DO 300 I=1,MPMAX
C     *
            DO 200 K=1,LPRIME
              KPRIME=NPRIME(K)
              IF(KPRIME.GT.I) GO TO 300
              IF((I/KPRIME)*KPRIME.EQ.I) GO TO 250
200         CONTINUE
            GO TO 300
C     *
C     *     FORBIDDEN INTEGER NUMBER
C     *
250         CONTINUE
            LPMAX=LPMAX+1
            INTX(LPMAX)=I
C     *
300       CONTINUE
C     *
C     ******
C
C     TEST AND CORRECT NUMBERS FOR GRID POINTS
C
      CALL PRIMET(N1,NS1,INTX,LPMAX,JCORR)
      CALL PRIMET(N2,NS2,INTX,LPMAX,JCORR)
      CALL PRIMET(N3,NS3,INTX,LPMAX,JCORR)
C
C**********************************************************************
C
      RETURN
      END
C---------------------------------------------------------------------
      SUBROUTINE PRIMET(N,NSTEP,NFORBD,LFORBD,JCORR)
C
C                   F R O G 2
C     CORRECT POSSIBLE NUMBER OF GRID POINTS
C
C     VERSION   1.0        CORRECTION  14.01.1987
C
C**********************************************************************
C
C     PARAMETERS :
C
C     N      - INITIAL NUMBER
C     NSTEP  - STEP FOR NUMBER SELECTION
C     JCORR  - FLAG OF NUMBER CORRECTION
C     NFORBD - ARRAY OF FORBIDEN NUMBERS
C     LFORBD - ITS LENGTH
C**********************************************************************
C
      DIMENSION NFORBD(LFORBD)
C
C**********************************************************************
C
        I1=1
C
C ************************
C *
100     CONTINUE
C *
C *   ******
C *   *
          DO 200 I=I1,LFORBD
            IF(NFORBD(I).GT.N) GO TO 500
            IF(NFORBD(I).EQ.N) GO TO 300
200       CONTINUE
C *   *
C *   ******
C *
          GO TO 400
C *
C *         NUMBER IS INVALID
C *
300       CONTINUE
C *
          N=N+NSTEP
          I1=I
C          JCORR=1
          GO TO 100
C *
C *********************
C
400       CONTINUE
C
C          WRITE(NFPRNT,2000) N
          CALL CCPERR(1,'s/r PRIMET: number is invalid.')
C
C**********************
C
500       CONTINUE
C
C**********************************************************************
C
2000  FORMAT(' GRID POINTS NUMBER IS TOO LARGE ',I10)
C
C**********************************************************************
C
      RETURN
      END
C
C#######################################################################







C#######################################################################
c*********************************************************************
c** SFP1 **                                           18 September 1997
c**********                                           V.Yu.Lunin
c
c           Structure factors calculation for P1 space group
c
      SUBROUTINE SFP1(x,lbuf,nx,ny,nz,scale,nref,ih,ik,il,f,phi,mes)
      dimension x(lbuf)
      dimension ih(nref),ik(nref),il(nref),f(nref),phi(nref)
c
c        The program calculates inverse Fourier transform
C
C          F(H,K,L)=(SCALE/(NX*NY*NZ))*
C                                                   H*R K*S L*T
C                   * SUMMA  (RO(R,S,T)*EXP(2*PI*I*(---+---+---)) )
C                     R=0,NX-1                      NX  NY   NZ
C                     S=0,NY-1
C                     T=0,NZ-1
C
c        for specified set of indexes, supposing that
c        max!h!<=nx/2-1, max!k!<=ny/2-1, max!l!<nz/2-1
c
c          INPUT
c               It is supposed that the NX*NY*NZ array containing
c               electron density values is a part of a larger array;
c               the rest of this extended array will be used as
c               additional memory when calculating FFT;
c
c               x(.) - the extended electron density array;
c                      its first nx*ny*nz elements are considered
c                      as REAL nx*ny*nz array containing density
c                      values;
c                      !!! THESE VALUES WILL BE DESTROYED !!!
c               lbuf - real length of x-array; it  must be
c                      (nx+2)*ny*nz at least;
c                      !!! THE FIRST (nx+2)*ny*nz ELEMENTS OF x-ARRAY
c                          WILL BE DESTROYED IN THE SUBROUTINE !!!
c               nx,ny,nz - grid dimensions;
c               scale - scale factor;
c               nref  - number of structure factors be calculated;
c               ih(.),ik(.),il(.) - their indexes;
c               mes - file for messages;
c          OUTPUT
c               f(.),phi(.) - modules and phases of desired
c                      structure factors;
c               x(.) - when it's finished, the input values are replaced
c                      in this array for structure factors ones;
c                      thei array is considered as
c                      COMPLEX  (maxh+1)*ny*nz array;
c                      the value F(h,k,l) is in X(h+1,k+1,l+1)
C                      (assuming that  F(H,K,-L)=F(H,K,NZ-L),
C                                    F(H,-K,L)=F(H,NY-K,L)  )
c                      the only defined elements are ones with
c                      0<=h<=maxh, -maxk<=k<=maxk, -maxl<=l<=maxl.
C
c        ATTENTION !
c           - the real size of X-array must be (Nx+2)*Ny*Nz at least !
c           - Nx must be even;
c           - pay your attention to F-data placing in X-array
c             (especially for negative h and k indexes)
c
c    SUBROUTINES USED : SF1, SF1A, SF1B - are present in this batch;
c                       1-dimensional FFT kernel;
c.................................................
c
      twopi=6.283185
      maxh=iabs(ih(1))
      maxk=iabs(ik(1))
      maxl=iabs(il(1))
      do j=1,nref
        iw=iabs(ih(j))
        if(iw.gt.maxh) maxh=iw
        iw=iabs(ik(j))
        if(iw.gt.maxk) maxk=iw
        iw=iabs(il(j))
        if(iw.gt.maxl) maxl=iw
      enddo
c
      if(maxh.gt.nx/2-1 .or. maxk.gt.ny/2-1 .or. maxl.gt.nz/2-1) then
        write(mes,1000) nx,ny,nz,maxh,maxk,maxl
1000    format(1x,'** SFP1 ** RO-array dimensions :',3i5/
     *         12x,'are too small for max!indexes! :',3i5)
        CALL CCPERR(1,'s/r SFP1: RO-array dimensions too small.')
      endif
      if((nx/2)*2.ne.nx) then
        write(mes,*)'** SFP1 **  NX must be even, NX=',nx
        CALL CCPERR(1,'s/r SFP1: NX must be even')
      endif
      iw=(nx+2)*ny*nz
      if(iw.gt.lbuf) then
        write(mes,*)'** SFP1 ** Too small x-array; its length must be',
     *               iw
        CALL CCPERR(1,'s/r SFP1: x-array too small.')
      endif
c
      CALL SF1(x,nx,ny,nz,maxh,maxk,maxl,scale,lbuf)
c
      n1=2*(maxh+1)
      n2=n1*ny
      do j=1,nref
        jjh=ih(j)
        jh=jjh
        jk=ik(j)
        jl=il(j)
        if(jjh.lt.0) then
          jh=-jh
          jk=-jk
          jl=-jl
        endif
        if(jk.lt.0) jk=ny+jk
        if(jl.lt.0) jl=nz+jl
        ind=2*jh+jk*n1+jl*n2+1
c
        fre=x(ind)
        fim=x(ind+1)
        if(jjh.lt.0) fim=-fim
c
        fw=sqrt(fre*fre+fim*fim)
        pw=0.
        if(fw.ne.0.) pw=atan2(fim,fre)
        if(pw.lt.0) pw=pw+twopi
        f(j)=fw
        phi(j)=pw
      enddo
c
      RETURN
      END
c-----------------------------------------------------------
C      SF1                                           18.09.97
c                                                    V.Yu.Lunin
C
c        The program calculates inverse Fourier transform
C
C          F(H,K,L)=(SCALE/(NX*NY*NZ))*
C                                                   H*R K*S L*T
C                   * SUMMA  (RO(R,S,T)*EXP(2*PI*I*(---+---+---)) )
C                     R=0,NX-1                      NX  NY   NZ
C                     S=0,NY-1
C                     T=0,NZ-1
C
C        for  0<=H<=MAXH,-MAXK<=K<=MAXK,0<=L<=MAXL
C
c        INPUT DATA
c        X - containes RO(r,s,t) values; X-array is considered as
c            three-dimensional with dimensions Nx Ny Nz;
c            the value RO(r,s,t) is in X(r+1,s+1,t+1);
c        Nx,Ny,Nz - grid numbers (dimensions of RO-array);
c        maxh,maxk,maxl - define the set of necessary F-values;
C	 SCALE - scale-value;
c
c        OUTPUT DATA
c        X - containes calculated F(h,k,l)-values; it is considered
c            as three-dimensional array with dimensions
c            (maxh+1) Ny Nz; the value F(h,k,l) is in X(h+1,k+1,l+1)
C                    (assuming that  F(H,K,-L)=F(H,K,NZ-L),
C                                    F(H,-K,L)=F(H,NY-K,L)  )
c            the only defined elements are ones with
c            0<=h<=maxh, -maxk<=k<=maxk, -maxl<=l<=maxl.
C
c        ATTENTION !
c           - the real size of X-array must be (Nx+2)*Ny*Nz at least !
c           - must be
C                              2*MAXH+1<=NX
C                              2*MAXK+1<=NY
C                              2*MAXL+2<=NZ
C                              NX - even;
c           - the axes order : input data x y z ,
c                              ouput data h k l ;
c           - pay your attention to F-data placing in X-array
c             (especially for negative h and k indexes)
c
c        SUBROUTINES USED
c           One-dimensional Fast Fourier Transform package of
c           L.F.Ten Eyck;
c           SF1A, SF1B.
C
      SUBROUTINE SF1(X,NX,NY,NZ,MAXH,MAXK,MAXL,SCALE,nsize)
C
      DIMENSION X(nsize),K(5)
C
c   decompressing X-array; The first dimension will be Nx+2
c
      CALL SF1A(X,NX,NY,NZ,X,NX+2,NY,NZ)
C
      NTOT=(NX+2)*NY*NZ
      K(1)=NTOT
      K(2)=2
      K(3)=NTOT
      K(4)=NTOT
      K(5)=NX+2
      CALL REALFT(X(1),X(2),NX/2,K,nsize)
C
C          compressing  X - array;
C          the first dimension will be equal to  2*(MAXH+1)
C
      N3=2*(MAXH+1)
      CALL SF1B(X,NX+2,NY,NZ,  X,N3,NY,NZ)
C
C
C
      K(1)=2*(MAXH+1)*NY*NZ
      K(2)=2*(MAXH+1)
      K(3)=2*(MAXH+1)*NY
      K(4)=2*(MAXH+1)
      K(5)=2
      CALL CMPLFT(X(1),X(2),NY,K,nsize)
C
C
C
      K(1)=2*(MAXH+1)*NY*NZ
      K(2)=2*(MAXH+1)*NY
      K(3)=2*(MAXH+1)*NY*NZ
      K(4)=2*(MAXH+1)*(MAXK+1)
      K(5)=2
      CALL CMPLFT(X(1),X(2),NZ,K,nsize)
C
      K(4)=2*(MAXH+1)*MAXK
      JA=(NY-MAXK)*2*(MAXH+1)+1
      CALL CMPLFT(X(JA),X(JA+1),NZ,K,nsize)
C
C
C          scaling and transition to complex conjugate
C
      CONST=SCALE/(NX*NY*NZ)
      N1=2*(MAXH+1)*NY*(MAXL+1)
      W=-CONST
      DO 10 J=1,N1
      W=-W
      X(J)=W*X(J)
10    CONTINUE
C
      N2=2*(MAXH+1)*NY*(NZ-MAXL)+1
      N3=2*(MAXH+1)*NY*NZ
      W=-CONST
      DO 20 J=N2,N3
      W=-W
      X(J)=W*X(J)
20    CONTINUE
C
C
C
      RETURN
      END
c
c-----------------------------------------------------------
c
      SUBROUTINE SF1A(X,n1,n2,n3,Y,k1,k2,k3)
c
c	Decompressing X-array into Y-array;
c       must be n1<=k1, n2<=k2, n3<=k3
c
      dimension X(n1,n2,n3), Y(k1,k2,k3)
c
      do 50 i3=1,n3
        m3=n3-i3+1
        do 40 i2=1,n2
          m2=n2-i2+1
          do 30 i1=1,n1
            m1=n1-i1+1
            Y(m1,m2,m3)=X(m1,m2,m3)
30        continue
40      continue
50    continue
c
c
      return
      end
c
      SUBROUTINE SF1B(X,n1,n2,n3,Y,k1,k2,k3)
      dimension X(n1,n2,n3), Y(k1,k2,k3)
c
c      compressing array X into Y begining from the first element
c      must be n1>=k1, n2>=k2, n3>=k3
c
      do 50 i3=1,k3
        do 40 i2=1,k2
          do 30 i1=1,k1
            Y(i1,i2,i3)=X(i1,i2,i3)
30        continue
40      continue
50    continue
c
      return
      end
c********************************************************************


c********************************************************************
c
c       ONE - DIMENSIONAL FAST FOURIER TRANSFORM
c       by L.Ten Eyck
c
c       modified by N.Lunina to satisfy too clever compilers
c********************************************************************
C     fft033
c
      SUBROUTINE HERM FT (X, Y, N, DIM,nsize)
      INTEGER N
      INTEGER DIM(5)
      REAL X(nsize), Y(nsize)
C
C     HERMITIAN SYMMETRIC FOURIER TRANSFORM
C
C     GIVEN THE UNIQUE TERMS OF A HERMITIAN SYMMETRIC SEQUENCE OF LENGTH
C     2N THIS SUBROUTINE CALCULATES THE 2N REAL NUMBERS WHICH ARE ITS
C     FOURIER TRANSFORM.  THE EVEN NUMBERED ELEMENTS OF THE TRANSFORM
C     (0, 2, 4, . . ., 2N-2) ARE RETURNED IN X AND THE ODD NUMBERED
C     ELEMENTS (1, 3, 5, . . ., 2N-1) IN Y.
C
C     A FINITE HERMITIAN SEQUENCE OF LENGTH 2N CONTAINS N + 1 UNIQUE
C     REAL NUMBERS AND N - 1 UNIQUE IMAGINARY NUMBERS.  FOR CONVENIENCE
C     THE REAL VALUE FOR X(N) IS STORED AT Y(0).
C
      REAL A, B, C, D, E, F, ANGLE, CO, SI, TWO N, TWO PI
      INTEGER NT, D2, D3, D4, D5
      INTEGER I, J, K, N OVER 2, I0, I1, I2
      INTEGER K1
C
      TWO PI = 6.283185
      TWO N = FLOAT(2*N)
C
      NT = DIM(1)
      D2 = DIM(2)
      D3 = DIM(3)
      D4 = DIM(4) - 1
      D5 = DIM(5)
C
      DO 100 I0 = 1, NT, D3
      I1 = I0 + D4
      DO 100 I = I0, I1, D5
      A = X(I)
      B = Y(I)
      X(I) = A + B
      Y(I) = A - B
  100 CONTINUE
C
      N OVER 2 = N/2 + 1
      IF (N OVER 2 .LT. 2) GO TO 500
      DO 400 I0 = 2, N OVER 2
      ANGLE = TWO PI*FLOAT(I0-1)/TWO N
      CO = COS(ANGLE)
      SI = SIN(ANGLE)
      K = (N + 2 - 2*I0)*D2
      K1 = (I0 - 1)*D2 + 1
      DO 300 I1 = K1, NT, D3
      I2 = I1 + D4
      DO 200 I = I1, I2, D5
      J = I + K
      A = X(I) + X(J)
      B = X(I) - X(J)
      C = Y(I) + Y(J)
      D = Y(I) - Y(J)
      E = B*CO + C*SI
      F = B*SI - C*CO
      X(I) = A + F
      X(J) = A - F
      Y(I) = E + D
      Y(J) = E - D
  200 CONTINUE
  300 CONTINUE
  400 CONTINUE
C
      CALL CMPL FT (X, Y, N, DIM,nsize)
C
  500 CONTINUE
      RETURN
C
      END
c
c--------------------------------------------------------------------
C     FFT034
C
      SUBROUTINE REAL FT (EVEN, ODD, N, DIM,nsize)
      INTEGER N
      INTEGER DIM(5)
      REAL EVEN(nsize), ODD(nsize)
C
C     REAL FOURIER TRANSFORM
C
C     GIVEN A REAL SEQUENCE OF LENGTH 2N THIS SUBROUTINE CALCULATES THE
C     UNIQUE PART OF THE FOURIER TRANSFORM.  THE FOURIER TRANSFORM HAS
C     N + 1 UNIQUE REAL PARTS AND N - 1 UNIQUE IMAGINARY PARTS.  SINCE
C     THE REAL PART AT X(N) IS FREQUENTLY OF INTEREST, THIS SUBROUTINE
C     STORES IT AT X(N) RATHER THAN IN Y(0).  THEREFORE X AND Y MUST BE
C     OF LENGTH N + 1 INSTEAD OF N.  NOTE THAT THIS STORAGE ARRANGEMENT
C     IS DIFFERENT FROM THAT EMPLOYED BY THE HERMITIAN FOURIER TRANSFORM
C     SUBROUTINE.
C
C     FOR CONVENIENCE THE DATA IS PRESENTED IN TWO PARTS, THE FIRST
C     CONTAINING THE EVEN NUMBERED REAL TERMS AND THE SECOND CONTAINING
C     THE ODD NUMBERED TERMS (NUMBERING STARTING AT 0).  ON RETURN THE
C     REAL PART OF THE TRANSFORM REPLACES THE EVEN TERMS AND THE
C     IMAGINARY PART OF THE TRANSFORM REPLACES THE ODD TERMS.
C
      REAL A, B, C, D, E, F, ANGLE, CO, SI, TWO PI, TWO N
      INTEGER NT, D2, D3, D4, D5
      INTEGER I, J, K, L, N OVER 2, I0, I1, I2
C
      TWO PI = 6.283185
      TWO N = FLOAT(2*N)
C
      CALL CMPL FT (EVEN, ODD, N, DIM,nsize)
C
      NT = DIM(1)
      D2 = DIM(2)
      D3 = DIM(3)
      D4 = DIM(4) - 1
      D5 = DIM(5)
      N OVER 2 = N/2 + 1
C
      IF (N OVER 2 .LT. 2) GO TO 400
      DO 300 I = 2, N OVER 2
      ANGLE = TWO PI*FLOAT(I-1)/TWO N
      CO = COS(ANGLE)
      SI = SIN(ANGLE)
      I0 = (I - 1)*D2 + 1
      J = (N + 2 - 2*I)*D2
      DO 200 I1 = I0, NT, D3
      I2 = I1 + D4
      DO 100 K = I1, I2, D5
      L = K + J
      A = (EVEN(L) + EVEN(K))/2.0
      C = (EVEN(L) - EVEN(K))/2.0
      B = (ODD(L) + ODD(K))/2.0
      D=(ODD(L)-ODD(K))/2.0
      E = C*SI + B*CO
      F = C*CO - B*SI
      EVEN(K) = A + E
      EVEN(L) = A - E
      ODD(K) = F - D
      ODD(L) = F + D
  100 CONTINUE
  200 CONTINUE
  300 CONTINUE
C
  400 CONTINUE
      IF (N .LT. 1) GO TO 600
      J = N*D2
      DO 500 I1 = 1, NT, D3
      I2 = I1 + D4
      DO 500 K = I1, I2, D5
      L = K + J
      EVEN(L) = EVEN(K) - ODD(K)
      ODD(L) = 0.0
      EVEN(K) = EVEN(K) + ODD(K)
      ODD(K) = 0.0
  500 CONTINUE
C
  600 CONTINUE
      RETURN
C
      END
C--------------------------------------------------------------------
c
C     FFT038
C
      SUBROUTINE CMPL FT (X, Y, N, D,nsize)
      INTEGER N
      INTEGER D(5)
      REAL X(nsize), Y(nsize)
C
C     COMPLEX FINITE DISCRETE FOURIER TRANSFORM
C     TRANSFORMS ONE DIMENSION OF MULTI-DIMENSIONAL DATA
C     MODIFIED BY L. F. TEN EYCK FROM A ONE-DIMENSIONAL VERSION WRITTEN
C     BY G. T. SANDE, 1969.
C
C     THIS PROGRAM CALCULATES THE TRANSFORM
C               (X(T) + I*Y(T))*(COS(2*PI*T/N) - I*SIN(2*PI*T/N))
C
C     INDEXING -- THE ARRANGEMENT OF THE MULTI-DIMENSIONAL DATA IS
C     SPECIFIED BY THE INTEGER ARRAY D, THE VALUES OF WHICH ARE USED AS
C     CONTROL PARAMETERS IN DO LOOPS.  WHEN IT IS DESIRED TO COVER ALL
C     ELEMENTS OF THE DATA FOR WHICH THE SUBSCRIPT BEING TRANSFORMED HAS
C     THE VALUE I0, THE FOLLOWING IS USED.
C
C               I1 = (I0 - 1)*D(2) + 1
C               DO 100 I2 = I1, D(1), D(3)
C               I3 = I2 + D(4) - 1
C               DO 100 I = I2, I3, D(5)
C                  .
C                  .
C           100 CONTINUE
C
C     WITH THIS INDEXING IT IS POSSIBLE TO USE A NUMBER OF ARRANGEMENTS
C     OF THE DATA, INCLUDING NORMAL FORTRAN COMPLEX NUMBERS (D(5) = 2)
C     OR SEPARATE STORAGE OF REAL AND IMAGINARY PARTS.
C
      LOGICAL ERROR
      INTEGER P MAX, P SYM, TWO GRP
      INTEGER FACTOR(15), SYM(15), UN SYM(15)
      character*200 errline
C
C     P MAX IS THE LARGEST PRIME FACTOR THAT WILL BE TOLERATED BY THIS
C     PROGRAM.
C     TWO GRP IS THE LARGEST POWER OF TWO THAT IS TREATED AS A SPECIAL
C     CASE.
C
      P MAX = 19
      TWO GRP = 8
C
      IF (N .LE. 1) GO TO 100
      CALL SRFP (N, P MAX, TWO GRP, FACTOR, SYM, P SYM, UN SYM, ERROR,
     *             nsize )
      IF (ERROR) GO TO 200
C
      CALL MDFTKD (N, FACTOR, D, X, Y,nsize)
      CALL DIPRP (N, SYM, P SYM, UN SYM, D, X, Y,nsize)
C
  100 CONTINUE
      RETURN
C
  200 CONTINUE

      write(errline,'(A43,I10)') 
     +  'INVALID NUMBER OF POINTS FOR CMPL FT.  N = ',N
      call ccperr(1,errline)
C
      END
c
c--------------------------------------------------------------------
C     FFT039
C
      SUBROUTINE SRFP (PTS,PMAX,TWO GRP,FACTOR,SYM,P SYM,UN SYM,ERROR,
     *                   nsize)
C     SYMMETRIZED REORDERING FACTORING PROGRAMME
C
      LOGICAL ERROR
      INTEGER PTS,PMAX,TWO GRP,P SYM
      INTEGER FACTOR (10), SYM (10), UN SYM (10)
C
      INTEGER PP(14), QQ (7)
      INTEGER F,J,JJ,N,NEST,P,P TWO,Q,R
      character*200 errline
C
      NEST=14
C
      N=PTS
      P SYM=1
      F=2
      P=0
      Q=0
  100 CONTINUE
      IF (N.LE.1) GO TO 500
      DO 200 J=F,PMAX
      IF (N.EQ.(N/J)*J) GO TO 300
  200 CONTINUE
      GO TO 1400
  300 CONTINUE
      IF (2*P+Q.GE.NEST) GO TO 1600
      F=J
      N=N/F
      IF (N.EQ.(N/F)*F) GO TO 400
      Q=Q+1
      QQ(Q)=F
      GO TO 100
  400 CONTINUE
      N=N/F
      P=P+1
      PP(P)=F
      P SYM=P SYM*F
      GO TO 100
C
  500 CONTINUE
      R=1
      IF (Q.EQ.0) R=0
      IF (P.LT.1) GO TO 700
      DO 600 J=1,P
      JJ=P+1-J
      SYM(J)=PP(JJ)
      FACTOR(J)=PP(JJ)
      JJ=P+Q+J
      FACTOR(JJ)=PP(J)
      JJ=P+R+J
      SYM(JJ)=PP(J)
  600 CONTINUE
  700 CONTINUE
      IF (Q.LT.1) GO TO 900
      DO 800 J=1,Q
      JJ=P+J
      UN SYM(J)=QQ(J)
      FACTOR(JJ)=QQ(J)
  800 CONTINUE
      SYM(P+1)=PTS/P SYM**2
  900 CONTINUE
      JJ=2*P+Q
      FACTOR(JJ+1)=0
      P TWO=1
      J=0
 1000 CONTINUE
      J=J+1
      IF (FACTOR(J).EQ.0) GO TO 1200
      IF (FACTOR(J).NE.2) GO TO 1000
      P TWO=P TWO*2
      FACTOR(J)=1
      IF (P TWO.GE.TWO GRP) GO TO 1100
      IF (FACTOR(J+1).EQ.2) GO TO 1000
 1100 CONTINUE
      FACTOR(J)=P TWO
      P TWO=1
      GO TO 1000
 1200 CONTINUE
      IF (P.EQ.0) R=0
      JJ=2*P+R
      SYM(JJ+1)=0
      IF (Q.LE.1) Q=0
      UN SYM(Q+1)=0
      ERROR=.FALSE.
C
 1300 CONTINUE
      RETURN
C
 1400 CONTINUE
      write(errline,'(A23,I3,A7,I6,A1)') 
     +  'LARGEST FACTOR EXCEEDS ', PMAX, '.  N = ', PTS, '.'
      call ccperr(1,errline)
      ERROR=.TRUE.
      GO TO 1300
C
 1600 CONTINUE
      write(errline,'(A21,I3,A7,I6,A1)') 
     +  'FACTOR COUNT EXCEEDS ', NEST, '.  N = ', PTS, '.'
      call ccperr(1,errline)
      ERROR=.TRUE.
      RETURN
C
      END
c
c--------------------------------------------------------------------
C     FFT040
C
      SUBROUTINE D IP RP (PTS,SYM,P SYM,UN SYM,DIM,X,Y,nsize)
C     DOUBLE IN PLACE REORDERING PROGRAMME
C
      INTEGER PTS,P SYM
      REAL X(nsize), Y(nsize)
      INTEGER SYM(10), UN SYM(10), DIM(5)
C
      REAL T
      LOGICAL ONE MOD
      INTEGER MODULO (14)
      INTEGER DK,JJ,KK,LK,MODS,MULT,NEST,P UN SYM,TEST
      INTEGER NT, SEP, DELTA, P, P0, P1, P2, P3, P4, P5, SIZE
C
      INTEGER S (14), U (14)
      INTEGER A,B,C,D,E,F,G,H,I,J,K,L,M,N
      INTEGER BS,CS,DS,ES,FS,GS,HS,IS,JS,KS,LS,MS,NS
      INTEGER AL,BL,CL,DL,EL,FL,GL,HL,IL,JL,KL,LL,ML,NL
      EQUIVALENCE           (AL,U(1)),(BS,S(2)),(BL,U(2))
      EQUIVALENCE (CS,S(3)),(CL,U(3)),(DS,S(4)),(DL,U(4))
      EQUIVALENCE (ES,S(5)),(EL,U(5)),(FS,S(6)),(FL,U(6))
      EQUIVALENCE (GS,S(7)),(GL,U(7)),(HS,S(8)),(HL,U(8))
      EQUIVALENCE (IS,S(9)),(IL,U(9)),(JS,S(10)),(JL,U(10))
      EQUIVALENCE (KS,S(11)),(KL,U(11)),(LS,S(12)),(LL,U(12))
      EQUIVALENCE (MS,S(13)),(ML,U(13)),(NS,S(14)),(NL,U(14))
C
      NEST=14
C
      NT = DIM(1)
      SEP = DIM(2)
      P2 = DIM(3)
      SIZE = DIM(4) - 1
      P4 = DIM(5)
      IF (SYM(1).EQ.0) GO TO 500
      DO 100 J=1,NEST
      U(J)=1
      S(J)=1
  100 CONTINUE
      N=PTS
      DO 200 J=1,NEST
      IF (SYM(J).EQ.0) GO TO 300
      JJ=NEST+1-J
      U(JJ)=N
      S(JJ)=N/SYM(J)
      N=N/SYM(J)
  200 CONTINUE
  300 CONTINUE
C
      JJ=0
      DO 400 A=1,AL
      DO 400 B=A,BL,BS
      DO 400 C=B,CL,CS
      DO 400 D=C,DL,DS
      DO 400 E=D,EL,ES
      DO 400 F=E,FL,FS
      DO 400 G=F,GL,GS
      DO 400 H=G,HL,HS
      DO 400 I=H,IL,IS
      DO 400 J=I,JL,JS
      DO 400 K=J,KL,KS
      DO 400 L=K,LL,LS
      DO 400 M=L,ML,MS
      DO 400 N=M,NL,NS
      JJ=JJ+1
      IF (JJ.GE.N) GO TO 400
      DELTA = (N-JJ)*SEP
      P1 = (JJ-1)*SEP + 1
      DO 350 P0 = P1, NT, P2
      P3 = P0 + SIZE
      DO 350 P = P0, P3, P4
      P5 = P + DELTA
      T = X(P)
      X(P) = X(P5)
      X(P5) = T
      T = Y(P)
      Y(P) = Y(P5)
      Y(P5) = T
  350 CONTINUE
  400 CONTINUE
C
  500 CONTINUE
      IF (UN SYM(1).EQ.0) GO TO 1900
      P UN SYM=PTS/P SYM**2
      MULT=P UN SYM/UN SYM(1)
      TEST=(UN SYM(1)*UN SYM(2)-1)*MULT*P SYM
      LK=MULT
      DK=MULT
      DO 600 K=2,NEST
      IF (UN SYM(K).EQ.0) GO TO 700
      LK=LK*UN SYM(K-1)
      DK=DK/UN SYM(K)
      U(K)=(LK-DK)*P SYM
      MODS=K
  600 CONTINUE
  700 CONTINUE
      ONE MOD=MODS.LT.3
      IF (ONE MOD) GO TO 900
      DO 800 J=3,MODS
      JJ=MODS+3-J
      MODULO(JJ)=U(J)
  800 CONTINUE
  900 CONTINUE
      MODULO(2)=U(2)
      JL=(P UN SYM-3)*P SYM
      MS=P UN SYM*P SYM
C
      DO 1800 J=P SYM,JL,P SYM
      K=J
C
 1000 CONTINUE
      K=K*MULT
      IF (ONE MOD) GO TO 1200
      DO 1100 I=3,MODS
      K=K-(K/MODULO(I))*MODULO(I)
 1100 CONTINUE
 1200 CONTINUE
      IF (K.GE.TEST) GO TO 1300
      K=K-(K/MODULO(2))*MODULO(2)
      GO TO 1400
 1300 CONTINUE
      K=K-(K/MODULO(2))*MODULO(2)+MODULO(2)
 1400 CONTINUE
      IF (K.LT.J) GO TO 1000
C
      IF (K.EQ.J) GO TO 1700
      DELTA = (K-J)*SEP
      DO 1600 L=1,P SYM
      DO 1500 M=L,PTS,MS
      P1 = (M+J-1)*SEP + 1
      DO 1500 P0 = P1, NT, P2
      P3 = P0 + SIZE
      DO 1500 JJ = P0, P3, P4
      KK = JJ + DELTA
      T=X(JJ)
      X(JJ)=X(KK)
      X(KK)=T
      T=Y(JJ)
      Y(JJ)=Y(KK)
      Y(KK)=T
 1500 CONTINUE
 1600 CONTINUE
 1700 CONTINUE
 1800 CONTINUE
C
 1900 CONTINUE
      RETURN
      END
c
c--------------------------------------------------------------------
C     FFT041
C
      SUBROUTINE MD FT KD (N, FACTOR, DIM, X, Y,nsize)
C     MULTI-DIMENSIONAL COMPLEX FOURIER TRANSFORM KERNEL DRIVER
C
      INTEGER N
      INTEGER FACTOR (10), DIM(5)
      REAL X(nsize), Y(nsize)
C
      INTEGER F, M, P, R, S
C
      S = DIM(2)
      F = 0
      M = N
  100 CONTINUE
      F = F + 1
      P = FACTOR(F)
      IF (P.EQ.0) RETURN
      M = M/P
      R = M*S
      IF (P.GT.8) GO TO 700
      GO TO (100, 200, 300, 400, 500, 800, 700, 600), P
      GO TO 800
C
  200 CONTINUE
      CALL R2 CFTK (N, M, X(1), Y(1), X(R+1), Y(R+1), DIM,nsize)
      GO TO 100
C
  300 CONTINUE
      CALL R3 CFTK (N, M, X(1), Y(1), X(R+1), Y(R+1), X(2*R+1), Y(2*R+1)
     ., DIM,nsize)
      GO TO 100
C
  400 CONTINUE
      CALL R4 CFTK (N, M, X(1), Y(1), X(R+1), Y(R+1), X(2*R+1), Y(2*R+1)
     ., X(3*R+1), Y(3*R+1), DIM,nsize)
      GO TO 100
C
  500 CONTINUE
      CALL R5 CFTK (N, M, X(1), Y(1), X(R+1), Y(R+1), X(2*R+1), Y(2*R+1)
     ., X(3*R+1), Y(3*R+1), X(4*R+1), Y(4*R+1), DIM,nsize)
      GO TO 100
C
  600 CONTINUE
      CALL R8 CFTK (N, M, X(1), Y(1), X(R+1), Y(R+1), X(2*R+1), Y(2*R+1)
     ., X(3*R+1), Y(3*R+1), X(4*R+1), Y(4*R+1), X(5*R+1), Y(5*R+1),
     .X(6*R+1), Y(6*R+1), X(7*R+1), Y(7*R+1), DIM,nsize)
      GO TO 100
C
  700 CONTINUE
      CALL RP CFTK (N, M, P, R, X, Y, DIM,nsize)
      GO TO 100
C
  800 CONTINUE
      CALL CCPERR(1,'TRANSFER ERROR DETECTED IN MDFTKD')
      RETURN
      END
c
c--------------------------------------------------------------------
C     FFT042
C
      SUBROUTINE R2 CFTK (N, M, X0, Y0, X1, Y1, DIM,nsize)
C     RADIX 2 MULTI-DIMENSIONAL COMPLEX FOURIER TRANSFORM KERNEL
C
      INTEGER N, M, DIM(5)
      REAL X0(nsize), Y0(nsize), X1(nsize), Y1(nsize)
C
      LOGICAL FOLD,ZERO
      INTEGER J,K,K0,M2,M OVER 2
      INTEGER K1, K2, KK, L, L1, MM2, NT, SIZE, SEP
      INTEGER NS
      REAL ANGLE,C,IS,IU,RS,RU,S,TWOPI
      REAL FJM1, FM2
      DATA TWO PI/6.283185/
C
      NT = DIM(1)
      SEP = DIM(2)
      L1 = DIM(3)
      SIZE = DIM(4) - 1
      K2 = DIM(5)
      NS = N*SEP
      M2=M*2
      FM2 = FLOAT(M2)
      M OVER 2=M/2+1
      MM2 = SEP*M2
C
      FJM1 = -1.0
      DO 600 J=1,M OVER 2
      FOLD=J.GT.1 .AND. 2*J.LT.M+2
      K0 = (J-1)*SEP + 1
      FJM1 = FJM1 + 1.0
      ANGLE = TWO PI*FJM1/FM2
      ZERO=ANGLE.EQ.0.0
      IF (ZERO) GO TO 200
      C=COS(ANGLE)
      S=SIN(ANGLE)
      GO TO 200
  100 CONTINUE
      FOLD=.FALSE.
      K0 = (M+1-J)*SEP + 1
      C=-C
  200 CONTINUE
C
      DO 500 KK = K0, NS, MM2
      DO 440 L = KK, NT, L1
      K1 = L + SIZE
      DO 420 K = L, K1, K2
      RS=X0(K)+X1(K)
      IS=Y0(K)+Y1(K)
      RU=X0(K)-X1(K)
      IU=Y0(K)-Y1(K)
      X0(K)=RS
      Y0(K)=IS
      IF (ZERO) GO TO 300
      X1(K)=RU*C+IU*S
      Y1(K)=IU*C-RU*S
      GO TO 400
  300 CONTINUE
      X1(K)=RU
      Y1(K)=IU
  400 CONTINUE
  420 CONTINUE
  440 CONTINUE
  500 CONTINUE
      IF (FOLD) GO TO 100
  600 CONTINUE
C
      RETURN
      END
c
c--------------------------------------------------------------------
C     FFT043
C
      SUBROUTINE R3 CFTK (N, M, X0, Y0, X1, Y1, X2, Y2, DIM,nsize)
C     RADIX 3 MULTI-DIMENSIONAL COMPLEX FOURIER TRANSFORM KERNEL
C
      INTEGER N, M, DIM(5)
      REAL X0(nsize),Y0(nsize),X1(nsize),Y1(nsize),
     * X2(nsize), Y2(nsize)
C
      LOGICAL FOLD,ZERO
      INTEGER J,K,K0,M3,M OVER 2
      INTEGER K1, K2, KK, L, L1, MM3, NT, SIZE, SEP
      INTEGER NS
      REAL ANGLE,A,B,C1,C2,S1,S2,T,TWOPI
      REAL I0,I1,I2,IA,IB,IS,R0,R1,R2,RA,RB,RS
      REAL FJM1, FM3
      DATA TWO PI/6.2831853/,A/-0.5/, B/0.86602540/
C
      NT = DIM(1)
      SEP = DIM(2)
      L1 = DIM(3)
      SIZE = DIM(4) - 1
      K2 = DIM(5)
      NS = N*SEP
      M3=M*3
      FM3 = FLOAT(M3)
      MM3 = SEP*M3
      M OVER 2=M/2+1
C
      FJM1 = -1.0
      DO 600 J=1,M OVER 2
      FOLD=J.GT.1 .AND. 2*J.LT.M+2
      K0 = (J-1)*SEP + 1
      FJM1 = FJM1 + 1.0
      ANGLE = TWO PI*FJM1/FM3
      ZERO=ANGLE.EQ.0.0
      IF (ZERO) GO TO 200
      C1=COS(ANGLE)
      S1=SIN(ANGLE)
      C2=C1*C1-S1*S1
      S2=S1*C1+C1*S1
      GO TO 200
  100 CONTINUE
      FOLD=.FALSE.
      K0 = (M+1-J)*SEP + 1
      T=C1*A+S1*B
      S1=C1*B-S1*A
      C1=T
      T=C2*A-S2*B
      S2=-C2*B-S2*A
      C2=T
  200 CONTINUE
C
      DO 500 KK = K0, NS, MM3
      DO 440 L = KK, NT, L1
      K1 = L + SIZE
      DO 420 K = L, K1, K2
      R0=X0(K)
      I0=Y0(K)
      RS=X1(K)+X2(K)
      IS=Y1(K)+Y2(K)
      X0(K)=R0+RS
      Y0(K)=I0+IS
      RA=R0+RS*A
      IA=I0+IS*A
      RB=(X1(K)-X2(K))*B
      IB=(Y1(K)-Y2(K))*B
      IF (ZERO) GO TO 300
      R1=RA+IB
      I1=IA-RB
      R2=RA-IB
      I2=IA+RB
      X1(K)=R1*C1+I1*S1
      Y1(K)=I1*C1-R1*S1
      X2(K)=R2*C2+I2*S2
      Y2(K)=I2*C2-R2*S2
      GO TO 400
  300 CONTINUE
      X1(K)=RA+IB
      Y1(K) = IA-RB
      X2(K)=RA-IB
      Y2(K) = IA + RB
  400 CONTINUE
  420 CONTINUE
  440 CONTINUE
  500 CONTINUE
      IF (FOLD) GO TO 100
  600 CONTINUE
C
      RETURN
      END
c
c--------------------------------------------------------------------
C     FFT044
C
      SUBROUTINE R4 CFTK (N, M, X0, Y0, X1, Y1, X2, Y2, X3, Y3, DIM,
     *                    nsize)
C     RADIX 4 MULTI-DIMENSIONAL COMPLEX FOURIER TRANSFORM KERNEL
C
      INTEGER N, M, DIM(5)
      REAL X0(nsize), Y0(nsize), X1(nsize), Y1(nsize)
      REAL X2(nsize), Y2(nsize), X3(nsize), Y3(nsize)
C
      LOGICAL FOLD,ZERO
      INTEGER J,K,K0,M4,M OVER 2
      INTEGER K1, K2, KK, L, L1, MM4, NT, SIZE, SEP
      INTEGER NS
      REAL ANGLE,C1,C2,C3,S1,S2,S3,T,TWOPI
      REAL I1,I2,I3,IS0,IS1,IU0,IU1,R1,R2,R3,RS0,RS1,RU0,RU1
      REAL FJM1, FM4
      DATA TWO PI/6.283185/
C
      NT = DIM(1)
      SEP = DIM(2)
      L1 = DIM(3)
      SIZE = DIM(4) - 1
      K2 = DIM(5)
      NS = N*SEP
      M4=M*4
      FM4 = FLOAT(M4)
      MM4 = SEP*M4
      M OVER 2=M/2+1
C
      FJM1 = -1.0
      DO 600 J=1,M OVER 2
      FOLD=J.GT.1 .AND. 2*J.LT.M+2
      K0 = (J-1)*SEP + 1
      FJM1 = FJM1 + 1.0
      ANGLE = TWO PI*FJM1/FM4
      ZERO=ANGLE.EQ.0.0
      IF (ZERO) GO TO 200
      C1=COS(ANGLE)
      S1=SIN(ANGLE)
      C2=C1*C1-S1*S1
      S2=S1*C1+C1*S1
      C3=C2*C1-S2*S1
      S3=S2*C1+C2*S1
      GO TO 200
  100 CONTINUE
      FOLD=.FALSE.
      K0 = (M+1-J)*SEP + 1
      T=C1
      C1=S1
      S1=T
      C2=-C2
      T=C3
      C3=-S3
      S3=-T
  200 CONTINUE
C
      DO 500 KK = K0, NS, MM4
      DO 440 L = KK, NT, L1
      K1 = L + SIZE
      DO 420 K = L, K1, K2
      RS0=X0(K)+X2(K)
      IS0=Y0(K)+Y2(K)
      RU0=X0(K)-X2(K)
      IU0=Y0(K)-Y2(K)
      RS1=X1(K)+X3(K)
      IS1=Y1(K)+Y3(K)
      RU1=X1(K)-X3(K)
      IU1=Y1(K)-Y3(K)
      X0(K)=RS0+RS1
      Y0(K)=IS0+IS1
      IF (ZERO) GO TO 300
      R1=RU0+IU1
      I1=IU0-RU1
      R2=RS0-RS1
      I2=IS0-IS1
      R3=RU0-IU1
      I3=IU0+RU1
      X2(K)=R1*C1+I1*S1
      Y2(K)=I1*C1-R1*S1
      X1(K)=R2*C2+I2*S2
      Y1(K)=I2*C2-R2*S2
      X3(K)=R3*C3+I3*S3
      Y3(K)=I3*C3-R3*S3
      GO TO 400
  300 CONTINUE
      X2(K)=RU0+IU1
      Y2(K)=IU0-RU1
      X1(K)=RS0-RS1
      Y1(K)=IS0-IS1
      X3(K)=RU0-IU1
      Y3(K)=IU0+RU1
  400 CONTINUE
  420 CONTINUE
  440 CONTINUE
  500 CONTINUE
      IF (FOLD) GO TO 100
  600 CONTINUE
C
      RETURN
      END
c
c--------------------------------------------------------------------
C     FFT045
C
      SUBROUTINE R5 CFTK (N, M, X0, Y0, X1, Y1, X2, Y2, X3, Y3, X4, Y4,
     . DIM,nsize)
C     RADIX 5 MULTI-DIMENSIONAL COMPLEX FOURIER TRANSFORM KERNEL
C
      INTEGER N, M, DIM(5)
      REAL X0(nsize), Y0(nsize), X1(nsize), Y1(nsize),
     *   X2(nsize), Y2(nsize)
      REAL X3(nsize), Y3(nsize), X4(nsize), Y4(nsize)
C
      LOGICAL FOLD,ZERO
      INTEGER J,K,K0,M5,M OVER 2
      INTEGER K1, K2, KK, L, L1, MM5, NT, SIZE, SEP
      INTEGER NS
      REAL ANGLE,A1,A2,B1,B2,C1,C2,C3,C4,S1,S2,S3,S4,T,TWOPI
      REAL R0,R1,R2,R3,R4,RA1,RA2,RB1,RB2,RS1,RS2,RU1,RU2
      REAL I0,I1,I2,I3,I4,IA1,IA2,IB1,IB2,IS1,IS2,IU1,IU2
      REAL FJM1, FM5
      DATA TWO PI/6.283185/, A1/0.30901699/, B1/0.95105652/,
     .      A2/-0.80901699/, B2/0.58778525/
C
      NT = DIM(1)
      SEP = DIM(2)
      L1 = DIM(3)
      SIZE = DIM(4) - 1
      K2 = DIM(5)
      NS = N*SEP
      M5=M*5
      FM5 = FLOAT(M5)
      MM5 = SEP*M5
      M OVER 2=M/2+1
C
      FJM1 = -1.0
      DO 600 J=1,M OVER 2
      FOLD=J.GT.1 .AND. 2*J.LT.M+2
      K0 = (J-1)*SEP + 1
      FJM1 = FJM1 + 1.0
      ANGLE = TWO PI*FJM1/FM5
      ZERO=ANGLE.EQ.0.0
      IF (ZERO) GO TO 200
      C1=COS(ANGLE)
      S1=SIN(ANGLE)
      C2=C1*C1-S1*S1
      S2=S1*C1+C1*S1
      C3=C2*C1-S2*S1
      S3=S2*C1+C2*S1
      C4=C2*C2-S2*S2
      S4=S2*C2+C2*S2
      GO TO 200
  100 CONTINUE
      FOLD=.FALSE.
      K0 = (M+1-J)*SEP + 1
      T=C1*A1+S1*B1
      S1=C1*B1-S1*A1
      C1=T
      T=C2*A2+S2*B2
      S2=C2*B2-S2*A2
      C2=T
      T=C3*A2-S3*B2
      S3=-C3*B2-S3*A2
      C3=T
      T=C4*A1-S4*B1
      S4=-C4*B1-S4*A1
      C4=T
  200 CONTINUE
C
      DO 500 KK = K0, NS, MM5
      DO 440 L = KK, NT, L1
      K1 = L + SIZE
      DO 420 K = L, K1, K2
      R0=X0(K)
      I0=Y0(K)
      RS1=X1(K)+X4(K)
      IS1=Y1(K)+Y4(K)
      RU1=X1(K)-X4(K)
      IU1=Y1(K)-Y4(K)
      RS2=X2(K)+X3(K)
      IS2=Y2(K)+Y3(K)
      RU2=X2(K)-X3(K)
      IU2=Y2(K)-Y3(K)
      X0(K)=R0+RS1+RS2
      Y0(K)=I0+IS1+IS2
      RA1=R0+RS1*A1+RS2*A2
      IA1=I0+IS1*A1+IS2*A2
      RA2=R0+RS1*A2+RS2*A1
      IA2=I0+IS1*A2+IS2*A1
      RB1=RU1*B1+RU2*B2
      IB1=IU1*B1+IU2*B2
      RB2=RU1*B2-RU2*B1
      IB2=IU1*B2-IU2*B1
      IF (ZERO) GO TO 300
      R1=RA1+IB1
      I1=IA1-RB1
      R2=RA2+IB2
      I2=IA2-RB2
      R3=RA2-IB2
      I3=IA2+RB2
      R4=RA1-IB1
      I4=IA1+RB1
      X1(K)=R1*C1+I1*S1
      Y1(K)=I1*C1-R1*S1
      X2(K)=R2*C2+I2*S2
      Y2(K)=I2*C2-R2*S2
      X3(K)=R3*C3+I3*S3
      Y3(K)=I3*C3-R3*S3
      X4(K)=R4*C4+I4*S4
      Y4(K)=I4*C4-R4*S4
      GO TO 400
  300 CONTINUE
      X1(K)=RA1+IB1
      Y1(K)=IA1-RB1
      X2(K)=RA2+IB2
      Y2(K)=IA2-RB2
      X3(K)=RA2-IB2
      Y3(K)=IA2+RB2
      X4(K)=RA1-IB1
      Y4(K)=IA1+RB1
  400 CONTINUE
  420 CONTINUE
  440 CONTINUE
  500 CONTINUE
      IF (FOLD) GO TO 100
  600 CONTINUE
C
      RETURN
      END
c
c--------------------------------------------------------------------
C     FFT046
C
      SUBROUTINE R8 CFTK (N, M, X0, Y0, X1, Y1, X2, Y2, X3, Y3, X4, Y4,
     . X5, Y5, X6, Y6, X7, Y7, DIM,nsize)
C     RADIX 8 MULTI-DIMENSIONAL COMPLEX FOURIER TRANSFORM KERNEL
C
      INTEGER N, M, DIM(5)
      REAL X0(nsize),Y0(nsize), X1(nsize),
     *    Y1(nsize), X2(nsize), Y2(nsize)
      REAL X3(nsize), Y3(nsize), X4(nsize),
     *   Y4(nsize), X5(nsize), Y5(nsize)
      REAL X6(nsize), Y6(nsize), X7(nsize), Y7(nsize)
C
      LOGICAL FOLD,ZERO
      INTEGER J,K,K0,M8,M OVER 2
      INTEGER K1, K2, KK, L, L1, MM8, NT, SIZE, SEP
      INTEGER NS
      REAL ANGLE,C1,C2,C3,C4,C5,C6,C7,E,S1,S2,S3,S4,S5,S6,S7,T,TWOPI
      REAL R1,R2,R3,R4,R5,R6,R7,RS0,RS1,RS2,RS3,RU0,RU1,RU2,RU3
      REAL I1,I2,I3,I4,I5,I6,I7,IS0,IS1,IS2,IS3,IU0,IU1,IU2,IU3
      REAL RSS0,RSS1,RSU0,RSU1,RUS0,RUS1,RUU0,RUU1
      REAL ISS0,ISS1,ISU0,ISU1,IUS0,IUS1,IUU0,IUU1
      REAL FJM1, FM8
      DATA TWO PI/6.283185/, E/0.70710678/
C
      NT = DIM(1)
      SEP = DIM(2)
      L1 = DIM(3)
      SIZE = DIM(4) - 1
      K2 = DIM(5)
      NS = N*SEP
      M8=M*8
      FM8 = FLOAT(M8)
      MM8 = SEP*M8
      M OVER 2=M/2+1
C
      FJM1 = -1.0
      DO 600 J=1,M OVER 2
      FOLD=J.GT.1 .AND. 2*J.LT.M+2
      K0 = (J-1)*SEP + 1
      FJM1 = FJM1 + 1.0
      ANGLE = TWO PI*FJM1/FM8
      ZERO=ANGLE.EQ.0.0
      IF (ZERO) GO TO 200
      C1=COS(ANGLE)
      S1=SIN(ANGLE)
      C2=C1*C1-S1*S1
      S2=S1*C1+C1*S1
      C3=C2*C1-S2*S1
      S3=S2*C1+C2*S1
      C4=C2*C2-S2*S2
      S4=S2*C2+C2*S2
      C5=C4*C1-S4*S1
      S5=S4*C1+C4*S1
      C6=C4*C2-S4*S2
      S6=S4*C2+C4*S2
      C7=C4*C3-S4*S3
      S7=S4*C3+C4*S3
      GO TO 200
  100 CONTINUE
      FOLD=.FALSE.
      K0 = (M+1-J)*SEP + 1
      T=(C1+S1)*E
      S1=(C1-S1)*E
      C1=T
      T=S2
      S2=C2
      C2=T
      T=(-C3+S3)*E
      S3=(C3+S3)*E
      C3=T
      C4=-C4
      T=-(C5+S5)*E
      S5=(-C5+S5)*E
      C5=T
      T=-S6
      S6=-C6
      C6=T
      T=(C7-S7)*E
      S7=-(C7+S7)*E
      C7=T
  200 CONTINUE
C
      DO 500 KK = K0, NS, MM8
      DO 440 L = KK, NT, L1
      K1 = L + SIZE
      DO 420 K = L, K1, K2
      RS0=X0(K)+X4(K)
      IS0=Y0(K)+Y4(K)
      RU0=X0(K)-X4(K)
      IU0=Y0(K)-Y4(K)
      RS1=X1(K)+X5(K)
      IS1=Y1(K)+Y5(K)
      RU1=X1(K)-X5(K)
      IU1=Y1(K)-Y5(K)
      RS2=X2(K)+X6(K)
      IS2=Y2(K)+Y6(K)
      RU2=X2(K)-X6(K)
      IU2=Y2(K)-Y6(K)
      RS3=X3(K)+X7(K)
      IS3=Y3(K)+Y7(K)
      RU3=X3(K)-X7(K)
      IU3=Y3(K)-Y7(K)
      RSS0=RS0+RS2
      ISS0=IS0+IS2
      RSU0=RS0-RS2
      ISU0=IS0-IS2
      RSS1=RS1+RS3
      ISS1=IS1+IS3
      RSU1=RS1-RS3
      ISU1=IS1-IS3
      RUS0=RU0-IU2
      IUS0=IU0+RU2
      RUU0=RU0+IU2
      IUU0=IU0-RU2
      RUS1=RU1-IU3
      IUS1=IU1+RU3
      RUU1=RU1+IU3
      IUU1=IU1-RU3
      T=(RUS1+IUS1)*E
      IUS1=(IUS1-RUS1)*E
      RUS1=T
      T=(RUU1+IUU1)*E
      IUU1=(IUU1-RUU1)*E
      RUU1=T
      X0(K)=RSS0+RSS1
      Y0(K)=ISS0+ISS1
      IF (ZERO) GO TO 300
      R1=RUU0+RUU1
      I1=IUU0+IUU1
      R2=RSU0+ISU1
      I2=ISU0-RSU1
      R3=RUS0+IUS1
      I3=IUS0-RUS1
      R4=RSS0-RSS1
      I4=ISS0-ISS1
      R5=RUU0-RUU1
      I5=IUU0-IUU1
      R6=RSU0-ISU1
      I6=ISU0+RSU1
      R7=RUS0-IUS1
      I7=IUS0+RUS1
      X4(K)=R1*C1+I1*S1
      Y4(K)=I1*C1-R1*S1
      X2(K)=R2*C2+I2*S2
      Y2(K)=I2*C2-R2*S2
      X6(K)=R3*C3+I3*S3
      Y6(K)=I3*C3-R3*S3
      X1(K)=R4*C4+I4*S4
      Y1(K)=I4*C4-R4*S4
      X5(K)=R5*C5+I5*S5
      Y5(K)=I5*C5-R5*S5
      X3(K)=R6*C6+I6*S6
      Y3(K)=I6*C6-R6*S6
      X7(K)=R7*C7+I7*S7
      Y7(K)=I7*C7-R7*S7
      GO TO 400
  300 CONTINUE
      X4(K)=RUU0+RUU1
      Y4(K)=IUU0+IUU1
      X2(K)=RSU0+ISU1
      Y2(K)=ISU0-RSU1
      X6(K)=RUS0+IUS1
      Y6(K)=IUS0-RUS1
      X1(K)=RSS0-RSS1
      Y1(K)=ISS0-ISS1
      X5(K)=RUU0-RUU1
      Y5(K)=IUU0-IUU1
      X3(K)=RSU0-ISU1
      Y3(K)=ISU0+RSU1
      X7(K)=RUS0-IUS1
      Y7(K)=IUS0+RUS1
  400 CONTINUE
  420 CONTINUE
  440 CONTINUE
  500 CONTINUE
      IF (FOLD) GO TO 100
  600 CONTINUE
C
      RETURN
      END
c
c--------------------------------------------------------------------
C     FFT047
C
      SUBROUTINE RP CFTK (N, M, P, R, X, Y, DIM,nsize)
C     RADIX PRIME MULTI-DIMENSIONAL COMPLEX FOURIER TRANSFORM KERNEL
C
CMDW: Note, routine works beyond the nominal bounds of X and Y.
C     We think this is deliberate, so don't be tempted to "fix" it.
C     This routine is therefore incompatible with "bounds-checking" 
C     options of compilers.
C
      INTEGER N, M, P, R, DIM(5)
c      REAL X(R,P), Y(R,P)
      real x(r,p),y(r,p)
C
      LOGICAL FOLD,ZERO
      REAL ANGLE,IS,IU,RS,RU,T,TWOPI,XT,YT
      REAL FU, FP, FJM1, FMP
      INTEGER J,JJ,K0,K,M OVER 2,MP,PM,PP,U,V
      INTEGER K1, K2, KK, L, L1, MMP, NT, SIZE, SEP
      INTEGER NS
C
      REAL AA (9,9), BB (9,9)
      REAL A (18), B (18), C (18), S (18)
      REAL IA (9), IB (9), RA (9), RB (9)
      data two pi/6.283185/
c
      NT = DIM(1)
      SEP = DIM(2)
      L1 = DIM(3)
      SIZE = DIM(4) - 1
      K2 = DIM(5)
      NS = N*SEP
      M OVER 2=M/2+1
      MP=M*P
      FMP = FLOAT(MP)
      MMP = SEP*MP
      PP=P/2
      PM=P-1
      FP = FLOAT(P)
      FU = 0.0
      DO 100 U=1,PP
      FU = FU + 1.0
      ANGLE = TWO PI*FU/FP
      JJ=P-U
      A(U)=COS(ANGLE)
      B(U)=SIN(ANGLE)
      A(JJ)=A(U)
      B(JJ)=-B(U)
  100 CONTINUE
      DO 300 U=1,PP
      DO 200 V=1,PP
      JJ=U*V-U*V/P*P
      AA(V,U)=A(JJ)
      BB(V,U)=B(JJ)
  200 CONTINUE
  300 CONTINUE
C
      FJM1 = -1.0
      DO 1500 J=1,M OVER 2
      FOLD=J.GT.1 .AND. 2*J.LT.M+2
      K0 = (J-1)*SEP + 1
      FJM1 = FJM1 + 1.0
      ANGLE = TWO PI*FJM1/FMP
      ZERO=ANGLE.EQ.0.0
      IF (ZERO) GO TO 700
      C(1)=COS(ANGLE)
      S(1)=SIN(ANGLE)
      DO 400 U=2,PM
      C(U)=C(U-1)*C(1)-S(U-1)*S(1)
      S(U)=S(U-1)*C(1)+C(U-1)*S(1)
  400 CONTINUE
      GO TO 700
  500 CONTINUE
      FOLD=.FALSE.
      K0 = (M+1-J)*SEP + 1
      DO 600 U=1,PM
      T=C(U)*A(U)+S(U)*B(U)
      S(U)=-S(U)*A(U)+C(U)*B(U)
      C(U)=T
  600 CONTINUE
  700 CONTINUE
C
      DO 1400 KK = K0, NS, MMP
      DO 1340 L = KK, NT, L1
      K1 = L + SIZE
      DO 1320 K = L, K1, K2
      XT=X(K,1)
      YT=Y(K,1)
      RS=X(K,2)+X(K,P)
      IS=Y(K,2)+Y(K,P)
      RU=X(K,2)-X(K,P)
      IU=Y(K,2)-Y(K,P)
      DO 800 U=1,PP
      RA(U)=XT+RS*AA(U,1)
      IA(U)=YT+IS*AA(U,1)
      RB(U)=RU*BB(U,1)
      IB(U)=IU*BB(U,1)
  800 CONTINUE
      XT=XT+RS
      YT=YT+IS
      DO 1000 U=2,PP
      JJ=P-U
      RS=X(K,U+1)+X(K,JJ+1)
      IS=Y(K,U+1)+Y(K,JJ+1)
      RU=X(K,U+1)-X(K,JJ+1)
      IU=Y(K,U+1)-Y(K,JJ+1)
      XT=XT+RS
      YT=YT+IS
      DO 900 V=1,PP
      RA(V)=RA(V)+RS*AA(V,U)
      IA(V)=IA(V)+IS*AA(V,U)
      RB(V)=RB(V)+RU*BB(V,U)
      IB(V)=IB(V)+IU*BB(V,U)
  900 CONTINUE
 1000 CONTINUE
      X(K,1)=XT
      Y(K,1)=YT
      DO 1300 U=1,PP
      JJ=P-U
      IF (ZERO) GO TO 1100
      XT=RA(U)+IB(U)
      YT=IA(U)-RB(U)
      X(K,U+1)=XT*C(U)+YT*S(U)
      Y(K,U+1)=YT*C(U)-XT*S(U)
      XT=RA(U)-IB(U)
      YT=IA(U)+RB(U)
      X(K,JJ+1)=XT*C(JJ)+YT*S(JJ)
      Y(K,JJ+1)=YT*C(JJ)-XT*S(JJ)
      GO TO 1200
 1100 CONTINUE
      X(K,U+1)=RA(U)+IB(U)
      Y(K,U+1)=IA(U)-RB(U)
      X(K,JJ+1)=RA(U)-IB(U)
      Y(K,JJ+1)=IA(U)+RB(U)
 1200 CONTINUE
 1300 CONTINUE
 1320 CONTINUE
 1340 CONTINUE
 1400 CONTINUE
      IF (FOLD) GO TO 500
 1500 CONTINUE
C
      RETURN
      END


