/*
     c_contour.c: 2d contouring functions
     Copyright (C) 2001  CCLRC, Peter Briggs

     This code is distributed under the terms and conditions of the
     CCP4 Program Suite Licence Agreement as a CCP4 Library.
     A copy of the CCP4 licence can be obtained by writing to the
     CCP4 Secretary, Daresbury Laboratory, Warrington WA4 4AD, UK.
*/

/*******************************************************************/

/* c_contour.c

   2d contouring functions

   contour_main constitutes the "user interface" to the contouring
   functionality.
*/

/*------------------------------------------------------------------------------*/
/* Include headers */
/*------------------------------------------------------------------------------*/
#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include "c_contour.h"

/*------------------------------------------------------------------------------*/
/* Functions */
/*------------------------------------------------------------------------------*/

/* contour_main

   Driver Function for Contouring
   
   nx and ny are the dimensions of the grid of density values
   stored in farray.
   farray should be a 2d array of scalar values [nx][ny].
   ncontours is the number of isovalues stores in fcontour.
   fcontour should be a 1d array of scalar values [ncontours]

   outfunc is a user supplied function which is called each time a
   complete contour line is traced. Its arguments are:
   isovalue - the current contour level
   n - the number of points in the contour line
   x - 1d array of n "x" points on a unit grid
   y - 1d array of n "y" points on a unit grid
*/
int contour_main(int nx, int ny, float **farray, int ncontours, float *fcontour,
		 int outfunc(float isovalue, int n, float *x, float *y))
{
  int i,nwork,ncell;
  int *workspace;
  float isovalue;

  if (DIAG) puts("> In contour_main");
  if (DIAG) {
    printf("contour_main: input parameters:\n");
    printf("contour_main: nx = %d, ny = %d, ncontours = %d\n",nx,ny,ncontours);
    printf("contour_main: levels are");
    for (i=0; i<ncontours; i++) printf(" %f (%i)",fcontour[i],i);
    printf("\n");
  }

  /* Set up */
  ncell = nx*ny;
  nwork = ncell*2;

  /* Workspace array required for book-keeping */
  if (DIAG) puts("contour_main: setting up workspace array");
  workspace = (int *) malloc(nwork*sizeof(int));
  if (!workspace) {
    puts("contour_main: could not allocate workspace array");
    return 0;
  }
  if (DIAG) puts("Done");
  if (ncontours<=0) {
    puts("contour_main: must have at least one contour level");
    return 0;
  }

  /* Main loop over all isovalues */
  if (DIAG) puts("Starting main loop");
  for (i=0; i<ncontours; i++) {
    isovalue = fcontour[i];
    if (DIAG) printf("Level %d - is %f (%f)\n",i,fcontour[i],isovalue);
    if (DIAG) puts("Calling find_contours");
    find_contours(isovalue,nx,ny,farray,ncell,workspace,outfunc);
    if (DIAG) puts("Done");
  }
  /* All done - free temporary workspace */
  if (workspace) free(workspace);
  if (DIAG) puts("Done contour_main");
  return 1;
}

/*------------------------------------------------------------------------------*/

/* find_contours

   Find all contours for the specified isovalue

   Search for a cell with a previously unexamined contour line
   passing through it, and follow it around until it closes or
   reaches the boundaries of the section. Then look for the next
   contour etc etc until all cells have been examined.
*/
int find_contours(float isovalue, int nx, int ny, float **farray, int ncell,
		  int *workspace,
		  int outfunc(float isovalue, int n, float *x, float *y))
{
  int i,ix,iy,ncoords;
  float *xcoordlist=NULL,*ycoordlist=NULL;

  if (DIAG) puts(">> Starting find_contours");

  /* Reset the workspace array */
  if (DIAG) puts("find_contours: clearing workspace array");
  for (i=0; i<ncell*2; i++) workspace[i] = 0;
  if (DIAG) puts("find_contours: cleared");

  /* Search over all cells in turn */
  if (DIAG) puts("find_contours: starting search over all cells");
  for (ix=0; ix<(nx-1); ix++) {
    for (iy=0; iy<(ny-1); iy++) {
      if (DIAG) printf("find_contours: current cell ix = %d, iy = %d\n",ix,iy);

      /* Check: are there unexamined contours passing through
	 this cell? */
      if (examine_cell(ix,iy,nx,ny,isovalue,farray,ncell,workspace)) {
	if (DIAG) puts("find_contours: cell has unmarked contours, starting new line");

	/* Initialise new contour line */
	/*start_new_contour(&ncoords,&xcoordlist,&ycoordlist);*/
	ncoords = 0;

	/* Now follow the contour around */
	if (DIAG) puts("find_contours: returned - call follow_contour");
	follow_contour(ix,iy,isovalue,nx,ny,farray,
		       ncell,workspace,&ncoords,&xcoordlist,
		       &ycoordlist);
	if (DIAG) puts("find_contours: done - now call function to output contour");

	/* Dump the coordinates for the list contour points */
	if (ncoords) {
	  outfunc(isovalue,ncoords,xcoordlist,ycoordlist);
	  if (DIAG) puts("find_contours: output of contour completed");
	} else {
	  if (DIAG) puts("find_contours: no points in line, outfunc not invoked");
	}

	/* Release dynamically allocated memory */
	if (xcoordlist) {
	  free(xcoordlist);
	  xcoordlist = NULL;
	}
	if (ycoordlist) {
	  free(ycoordlist);
	  ycoordlist = NULL;
	}
      }
      /* Go on to examine next cell */
    }
  }
  /* Finished */
  return 0;
}

/*------------------------------------------------------------------------------*/

/* examine_cell

   For cell ix,iy check whether the cell edges have already been crossed or
   otherwise ruled out

   Return 1 if there is an unexamined edge, 0 otherwise.
*/
int examine_cell(int ix, int iy, int nx, int ny, float isovalue, float **farray,
		 int ncell, int *workspace)
{
  int   cindex;
  float vert1,vert2;

  /* First check the edges belonging to this cell
     (edges 0 and 3) */
  cindex = cell_index(ix,iy,nx);
  if (DIAG) puts("  >>> examine_cell: edge 0");
  if (!workspace[cindex]) {
    vert1 = farray[ix][iy] - isovalue;
    vert2 = farray[ix+1][iy] - isovalue;
    if (DIAG)
      printf("  >>> examine_cell: unchecked edge, vertices %f and %f\n",vert1,vert2);
    if (vert1*vert2 <= 0.0) return 1;
    workspace[cindex] = 1;
  }
  if (DIAG) puts("  >>> examine_cell: edge 3");
  if (!workspace[cindex+ncell]) {
    vert1 = farray[ix][iy] - isovalue;
    vert2 = farray[ix][iy+1] - isovalue;
    if (DIAG) 
      printf("  >>> examine_cell: unchecked edge, vertices %f and %f\n",vert1,vert2);
    if (vert1*vert2 <= 0.0) return 1;
    workspace[cindex+ncell] = 1;
  }
  /* Check the other two edges (if they exist) */
  /* Cell to the right (edge 1) */
  if (ix+1 <= nx) {
  if (DIAG) puts("  >>> examine_cell: edge 1");
    cindex = cell_index(ix+1,iy,nx);
    if (!workspace[cindex+ncell]) {
      vert1 = farray[ix+1][iy] - isovalue;
      vert2 = farray[ix+1][iy+1] - isovalue;
      if (DIAG) 
	printf("  >>> examine_cell: unchecked edge, vertices %f and %f\n",vert1,vert2);
      if (vert1*vert2 <= 0.0) return 1;
      workspace[cindex+ncell] = 1;
    }
  }
  /* Cell below (edge 2) */
  if (DIAG) puts("  >>> examine_cell: edge 2");
  if (iy+1 <= ny) {
    cindex = cell_index(ix,iy+1,nx);
    if (!workspace[cindex]) {
      vert1 = farray[ix+1][iy+1] - isovalue;
      vert2 = farray[ix][iy+1] - isovalue;
      if (DIAG) 
	printf("  >>> examine_cell: unchecked edge, vertices %f and %f\n",vert1,vert2);
      if (vert1*vert2 <= 0.0) return 1;
      workspace[cindex] = 1;
    }
  }
  /* Everything already set - no suitable free edges */
  return 0;
}

/*------------------------------------------------------------------------------*/

/* cell_index

   Return the index for cell ix,iy
*/
int cell_index(int ix, int iy, int nx)
{
  return (iy*nx+ix);
}

/*------------------------------------------------------------------------------*/

/* test_cell

   Check cell ix,iy and determine whether there is a contour line
   passing through one of the cell boundaries.
   iedge is the edge through which the contour enters this cell, or
   -1 if the search is beginning "from scratch".

   The return status will be either CROSSING or NO_EXIT:
   
   CROSSING if there is a crossing into a neighbouring cell
   NO_EXIT  if there is no way out of the cell

   If the return status is CROSSING then xpos,ypos are set to have
   the coordinates of the cross-over point, and nextx, nexty are set to
   the grid coordinates of the next cell to examine, and nextedge to the
   edge index where the contour line enters that cell.
*/
int test_cell(int ix, int iy, int iedge,
	      float isovalue, int nx, int ny, float **farray,
	      int ncell, int *workspace,
	      float *xpos, float *ypos, int *nextx, int *nexty, int *nextedge)
{
  /* Jumps to get next edge in a sequence
     Searching sequence is: check neighbouring edges first, opposite
     last */
  int next_edge[4]={0,1,2,3};

  /* Offsets to obtain neighbouring gridpoints making up cell
     vertices
     Element [i][j] gives the x(j=0),y(j=1) offsets to from the
     current position to find vertex i */
  int gridpoint[4][2]={{0,0},{1,0},{1,1},{0,1}};

  /* Associate vertices and edges
     Edge i is between vertices[i][1] and vertices[i][2] */
  int vertices[4][2]={{0,1},{1,2},{2,3},{3,0}};

  int   i,j0,j1;
  int   this_edge,this_corner,edge,status,istart,eindex;
  float vert[4];

  if (DIAG) puts(">>> Entered test_cell");
  if (DIAG) printf("  >>> test_cell: isovalue is %f\n",isovalue);
  if (DIAG) printf("  >>> test_cell: examining cell %d %d\n",ix,iy);
  if (iedge < 0) {
    if (DIAG) puts("  >>> test_cell: new search, no prior information");
  } else {
    if (DIAG) printf("  >>> test_cell: starting edge is %d\n",iedge);
  }

  /* Initialise crossover points to nonesense values */
  *xpos = -1.0;
  *ypos = -1.0;

  /* Store the scalar values at the vertices of the cell
     relative to the current contour level */
  if (DIAG) puts("  >>> test_cell: storing relative values at each vertex");
  for (i=0; i<4; i++) {
    if (DIAG) printf("  >>> test_cell: vertex %d = gridpoint %d %d\n",i,ix,iy);
    if (DIAG) printf("  >>> test_cell: offset %d %d\n",
		     (ix+gridpoint[i][0]),(iy+gridpoint[i][1]));
    if (DIAG) printf("  >>> test_cell: absolute value = %f\n",
		     farray[(ix+gridpoint[i][0])][(iy+gridpoint[i][1])]);
    vert[i] = farray[(ix+gridpoint[i][0])][(iy+gridpoint[i][1])] - isovalue;
    if (DIAG) printf("  >>> test_cell: relative value = %f\n",vert[i]);
  }
  if (DIAG)
    printf("  >>> test_cell: vertices in this cell:\n  >>>  %4f**0**%4f\n  >>>     3            1\n  >>>  %4f**2**%4f\n",vert[0]+isovalue,vert[1]+isovalue,vert[3]+isovalue,vert[2]+isovalue);

  /* Initialise for edge searching */
  if (iedge<0) {
    istart = 0;
    this_edge = 0;
  } else {
    istart = 1;
    this_edge = iedge;
  }

  /* Check each edge in turn
     The search sequence is left, then right, then opposite cell edges */
  for (i=istart; i<4; i++) {
    this_edge = this_edge + next_edge[i];
    while (this_edge > 3) this_edge = this_edge - 4;
    if (DIAG) printf("  >>> test_cell: examining edge number %d\n",this_edge);

    eindex = cell_edge_index(ix,iy,this_edge,nx,ny,ncell);

    if(workspace[eindex] == 0) {
      if (DIAG) puts("  >>> test_cell: previously unexamined edge");
      j0 = vertices[this_edge][0];
      j1 = vertices[this_edge][1];
      if (DIAG) printf("  >>> test_cell: >>> edge defined by vertices %d and %d\n",
	     j0,j1);

      /* Check for special case of a corner point */
      if (vert[j0] == 0.0 && vert[j1] == 0.0) {
	/* Two corner points on one edge
	   This cannot have a contour passing through it*/
	if (DIAG) puts("  >>> test_cell: >>> two corner points! no contour line");
	edge = 0;
	/* Flip the workspace bit to say this edge has now been examined */
	if (eindex<ncell*2) {
	  workspace[eindex] = 1;
	} else {
	  if (DIAG) printf("WARNING: index %d is beyond the limits of work (%d)\n",
			   eindex,ncell*2);
	}
      } else if (vert[j0] == 0.0 || vert[j1] == 0.0) {
	/* Single corner point on this edge */
	if (DIAG) puts("  >>> test_cell: >>> one corner point, contour line");
	edge = 1;
	/* We need to flip workspace bits and find the next cell
	   Both performed by get_neighbouring_cell_for_corner */
	if (vert[j0]==0.0) {
	  this_corner = j0;
	} else {
	  this_corner = j1;
	}
	status = get_neighbouring_cell_for_corner(ix,iy,this_corner,
						  nx,ny,isovalue,farray,
						  workspace,nextx,nexty);
	/* Don't even think about assigning a "next edge" */
	*nextedge = -1;
      } else {
	/* Normal edge */
	if (DIAG) puts("  >>> test_cell: >>> normal edge, no corner points");
	edge = (vert[j0]*vert[j1]<0.0);
	/* Find the next cell */
	status = get_neighbouring_cell_for_edge(ix,iy,this_edge,nx,ny,nextx,nexty);
	/* Flip the workspace bit to say this edge has now been examined */
	if (eindex<ncell*2) {
	  workspace[eindex] = 1;
	} else {
	  if (DIAG) printf("WARNING: index %d is beyond the limits of work (%d)\n",
			   eindex,ncell*2);
	}
	/* Get the next edge */
	*nextedge = this_edge + 2;
	while (*nextedge > 3) *nextedge = *nextedge - 4;
      }

      if (edge) {
	/* Found an edge intersected by a contour */
	/* Determine the coordinates of the crossover */
	get_crossover_position(ix,iy,this_edge,vert[j0],vert[j1],xpos,ypos);
	if (DIAG) printf("  >>> test_cell: crossing point is %f %f\n",*xpos,*ypos);
	return status;
      }
    } else {
      if (DIAG) puts("  >>> test_cell: edge already examined");
    }
  }
  /* Can't get out of this cell */
  if (DIAG) puts("   >>> test_cell: no way out of this cell");
  return NO_EXIT;
}

/*------------------------------------------------------------------------------*/

/* Get index in workspace array for cell edge

   ix,iy are the indices for the cell of interest
   iedge is the index of the edge of interest (0...3)
   nx is the number of "x" grid points in the 2d array
   ncell is the total number of grid points (nx*ny)
*/
int cell_edge_index(int ix, int iy, int iedge, int nx, int ny, int ncell)
{
  /* Edges "owned" by cells
     The north and west edges belong to cell ix,iy
     The south edge is the north edge of cell ix+1,iy
     The east edge is the west edge of cell ix,iy+1

     Element [i][j] in the look-up table gives the x(j=0), y(j=1)
     and ncell offset (j=2) required to locate the appropriate
     workspace array element corresponding to edge i=0...3
  */
  int edge_belongs_to[4][3]={{0,0,0},{1,0,1},{0,1,0},{0,0,1}};
  int dx,dy,shift;

  /* Look-up the values */
  dx    = edge_belongs_to[iedge][0];
  dy    = edge_belongs_to[iedge][1];
  shift = edge_belongs_to[iedge][2];

  if (DIAG) {
    printf("  >>> cell_edge_index: >>> cell %d,%d ",ix,iy);
    switch (iedge) {
    case 0:
      printf("north");
      break;
    case 1:
      printf("east");
      break;
    case 2:
      printf("south");
      break;
    case 3:
      printf("west");
    }
    printf(" edge\n");
  }

  /* Valid cell? */
  if (ix+dx<0 || ix+dx>=nx || iy+dy<0 || iy+dy>=ny) {
    if (DIAG) puts("  >>> cell_edge_index: >>> cell out of range");
    return -1;
  }

  if (dx || dy)
    if (DIAG) printf("  >>> cell_edge_index: >>> edge belongs to next cell %d %d\n",
	   ix+dx,iy+dy);
  else
    if (DIAG) puts("  >>> cell_edge_index: >>> edge belongs to this cell");

  return cell_index(ix+dx,iy+dy,nx) + ncell*shift;
}

/*------------------------------------------------------------------------------*/

/* Get the neighbouring cell, given:
   ix iy - indicies of the current cell,
   iedge - index of the shared edge

   nx, ny are the dimensions of the section.

   Return the status - BOUNDARY (next cell is out of bounds) or
   CROSSING (next cell is okay).
*/
int get_neighbouring_cell_for_edge(int ix, int iy, int iedge, int nx, int ny,
				   int *jx, int *jy)
{
  /* Look up table for x,y shifts for each cell edge
     Element [i][j] gives the x(j=0), y(j=1) shifts required to
     reach the neighbouring cell associated with edge i=0...3 */
  int neighbouring_cell[4][2]={{0,-1},{1,0},{0,1},{-1,0}};

  *jx = ix+neighbouring_cell[iedge][0];
  *jy = iy+neighbouring_cell[iedge][1];

  /* Check if next cell is out of bounds */
  if (*jx<0 || *jx>=(nx-1) || *jy<0 || *jy>=(ny-1)) {
    return NO_EXIT;
  }
  return CROSSING;
}

/*------------------------------------------------------------------------------*/

/* Get the neighbouring cell, given:
   ix iy - indicies of the current cell,
   icorner - index of the shared corner

   nx, ny are the dimensions of the section.

   This function also flips the workspace bits corresponding to the
   corner point.

   Return the status - BOUNDARY (next cell is out of bounds) or
   CROSSING (next cell is okay).
*/
int get_neighbouring_cell_for_corner(int ix, int iy, int icorner, int nx, int ny,
				     float isovalue, float **farray,
				     int *workspace, int *jx, int *jy)
{
  int   ncell = nx*ny;
  int   i,cx,cy,ccx,ccy,eindex[4];

  /* Next cell offsets */
  int next_cell[4][4]={{0,-1,0,1},{-1,-1,0,3},{-1,0,2,3},{0,0,1,2}};
  
  if (DIAG) puts("get_neighbouring_cell_for_corner: starting");

  /* Find "absolute" cell index for corner point */
  switch (icorner) {
  case 0:
    cx = ix;
    cy = iy;
    break;
  case 1:
    cx = ix + 1;
    cy = iy;
    break;
  case 2:
    cx = ix + 1;
    cy = iy + 1;
    break;
  case 3:
    cx = ix;
    cy = iy + 1;
    break;
  default:
    if (DIAG) printf("get_neighbouring_cell_for_corner: bad corner index %d - returning\n",icorner);
    return NO_EXIT;
  }
  if (DIAG) printf("get_neighbouring_cell_for_corner: absolute cell is %d,%d\n",cx,cy);

  /* Get edge indices for all neighbouring edges
     and flip workspace bits */
  /* Indices for current cell, north and west */
  eindex[0] = cell_edge_index(cx,cy,0,nx,ny,ncell);
  eindex[1] = cell_edge_index(cx,cy,3,nx,ny,ncell);
  /* Index for north edge of one neighbour */
  eindex[2] = cell_edge_index(cx-1,cy,0,nx,ny,ncell);
  /* Index for west edge of other neighbour */
  eindex[3] = cell_edge_index(cx,cy-1,3,nx,ny,ncell);
  /* Flip the bits */
  for (i=0;i<4;i++)
    if (eindex[i]!=-1) {
      if (DIAG) printf("get_neighbouring_cell_for_corner: flipping bit %d\n",
		       eindex[i]);
      if (eindex[i]<ncell*2) {
	workspace[eindex[i]] = 1;
      } else {
	if (DIAG) printf("WARNING: index %d is beyond the limits of work (%d)\n",
			 eindex,ncell*2);
      }
    }

  /* Find the next cell - there are 3 possibilities (dicounting the
     current cell) */
  /* For each cell we need to check the "outermost" edges only */
  for (i=0;i<4;i++) {
    /* ccx, ccy is the neighbouring cell we wish to check */
    ccx = cx + next_cell[i][0];
    ccy = cy + next_cell[i][1];
    if (DIAG) printf("get_neighbouring_cell_for_corner: examining cell %d,%d\n",
		     ccx,ccy);
    if (ccx>=0 && ccx<(nx-1) && ccy>=0 && ccy<(ny-1)) {
      if (examine_cell(ccx,ccy,nx,ny,isovalue,farray,ncell,workspace)) {
	if (DIAG) puts("get_neighbouring_cell_for_corner: found next cell, returning");
	/* Found next cell */
	*jx = ccx;
	*jy = ccy;
	return CROSSING;
      }
    } else {
      if (DIAG) puts("get_neighbouring_cell_for_corner: this cell is out-of-bounds");
    }
    /* Finish loop over neighbours */
  }
  /* No way out */
  if (DIAG) puts("get_neighbouring_cell_for_corner: nowhere else to go from here");
  return NO_EXIT;
}

/*------------------------------------------------------------------------------*/

/* Determine the coordinates of the intersection between a contour
   line and a cell edge

   ix,iy are the cell indices
   iedge is the index of the cell edge
   rho1, rho2 are the function values at each end of the
   edge in question

   The coordinates of the intersection are returned in xpos,ypos
 */
int get_crossover_position(int ix, int iy, int iedge, float rho1, float rho2,
			   float *xpos, float *ypos)
{
  float offset;

  /* Determine an "offset" - distance from leading point */
  offset = linear_interp(rho1,rho2);

  /* Generate coordinates directly from the cell indicies */
  *xpos = (float) ix;
  *ypos = (float) iy;

  /* Apply the offset */
  if (iedge==0) {
    *xpos = *xpos + offset;
  } else if (iedge==1) {
    *xpos = *xpos + ((float) 1.0);
    *ypos = *ypos + offset;
  } else if (iedge==2) {
    *xpos = *xpos + ((float) 1.0) - offset;
    *ypos = *ypos + ((float) 1.0);
  } else {
    *ypos = *ypos + ((float) 1.0) - offset;
  }

  return 1;
}

/*------------------------------------------------------------------------------*/

/* Perform linear interpolation between two values */
float linear_interp(float v0, float v1)
{
  /* Assume that the distance between the two points is
     1 unit */
  if (v0==v1)
    return 0.5;
  else
    return (v0/(v0-v1));
}

/*------------------------------------------------------------------------------*/

/* initial_point_is_corner

   Check if the "first" (actually the last, since the order of coordinates
   should have been reversed before calling this routine) point in the list
   of coordinates lies on the corner of a cell.

   Return 1 (true) or 0 (false)
 */
int initial_point_is_corner(int ncoords, float **xptr, float **yptr)
{
  /* xptr, yptr are pointers to the address of xcoord[0], ycoord[0] */
  float *xcoordlist=*xptr,*ycoordlist=*yptr;
  double i;
  if (!ncoords) return 0;
  if (modf((double) xcoordlist[ncoords-1],&i)==0.0 && 
      modf((double) ycoordlist[ncoords-1],&i)==0.0) return 1;
  return 0;
}

/*------------------------------------------------------------------------------*/

/* start_new_contour

   Initialise arrays to start a new contour line
 */
int start_new_contour(int *nptr, float **xptr, float **yptr)
{
  if (DIAG) puts(">>> Entering start_new_contour");

  /* Set up a new list of coordinates */
  *nptr = 0;
  if (DIAG) puts(">>> Leaving start_new_contour");
  return 1;
}

/*------------------------------------------------------------------------------*/

/* follow_contour

   Given a "seed cell" init_x,init_y, trace a complete contour line
*/
int follow_contour(int init_x, int init_y,
		   float isovalue, int nx, int ny, float **farray,
		   int ncell, int *workspace,
		   int *ncoords, float **xcoordlist, float **ycoordlist)
{
  int   ix,iy,iedge,icorner,cell_status,cstatus=CONTINUE,flipped=0;
  int   nextx,nexty,nextedge;
  float xpos,ypos;

  if (DIAG) printf(">>> follow_contour : seed cell is %d,%d\n",init_x,init_y);
  ix = init_x;
  iy = init_y;
  iedge = -1;

  while (cstatus == CONTINUE) {

    cell_status = test_cell(ix,iy,iedge,isovalue,nx,ny,
			    farray,ncell,workspace,
			    &xpos,&ypos,&nextx,&nexty,&nextedge);

    /* Write the recovered coordinate pair to the list */
    if (xpos>=0.0 && ypos>=0.0) {
      if (DIAG) puts("Extending contour...");
      extend_contour(xpos,ypos,ncoords,xcoordlist,ycoordlist);
      if (DIAG) puts("Extended contour.");
    } else {
      if (DIAG) puts("Returned nonsense coordinates - not extending contour.");
    }
      
    if (cell_status!=NO_EXIT) {
      /* Move to the next cell */
      if (DIAG) puts("Move to next cell");
      ix = nextx;
      iy = nexty;
      iedge = nextedge;
    }
    else
      {
	/* Reached the end of the contour */
	if (DIAG) puts("Reached the end of the contour");
	if (!flipped) {
	  /* Reverse the list and look for the rest of
	     the contour line */
	  if (DIAG) puts("Flipping contour...");
	  flip_contour(*ncoords,xcoordlist,ycoordlist);
	  if (DIAG) puts("Flipped contour.");
	  flipped = 1;
	  /* Reset next cell to initial cell to start
	     searching again */
	  if (!initial_point_is_corner(*ncoords,xcoordlist,ycoordlist)) {
	    if (DIAG) puts("follow_contour: initial point not a corner");
	    ix = init_x;
	    iy = init_y;
	  } else {
	    /* Corners are a special case - there may not
	       be an exit from the original cell
	       Look for a suitable neighbouring cell */
	    if (DIAG) puts("follow_contour: the initial point was a corner!");
	    /* Find which corner */
	    if (farray[init_x][init_y]==isovalue)
	      icorner = 0;
	    else if (farray[init_x+1][init_y]==isovalue)
	      icorner = 1;
	    else if (farray[init_x+1][init_y+1]==isovalue)
	      icorner = 2;
	    else if (farray[init_x][init_y+1]==isovalue)
	      icorner = 3;
	    if (DIAG) printf("follow_contour: it was corner %d of cell %d,%d\n",
			     icorner,init_x,init_y);
	    /* Search for a possible way out */
	    if (get_neighbouring_cell_for_corner(init_x,init_y,icorner,nx,ny,
						 isovalue,farray,workspace,&ix,&iy)
		!=NO_EXIT) {
	      if (DIAG) printf("follow_contour: try next cell as %d,%d\n",ix,iy);
	    } else {
	      if (DIAG)
		puts("follow_contour: can't find a way back, must really be the end");
	      cstatus = STOP;
	    }
	  }
	  iedge = -1;
	} else {
	  /* Assume that we've found both ends
	     Time to stop looping */
	  if (DIAG) puts("This is the second time for this contour - stop!");
	  cstatus = STOP;
	}
      }
  }
  
  /* Finished tracing contour
     Need to check if it is closed or open */
  if (DIAG) puts("Finished tracing contour");
  if (DIAG) puts("Check if contoured should be closed ...");
  close_contour(ncoords,xcoordlist,ycoordlist,nx,ny);
  if (DIAG) puts("...Done closing contour");

  /* Contour line finished */
  return 1;
}

/*------------------------------------------------------------------------------*/

/* Add another point to an existing contour line */
int extend_contour(float xpos,float ypos,int *nptr,
		   float **xptr, float **yptr)
{ 
  /* xptr, yptr are pointers to the address of xcoord[0], ycoord[0] */
  float *xcoordlist=*xptr,*ycoordlist=*yptr;
  /* nptr is a pointer to the address of ncoords */
  int   i,ncoords=*nptr;

  if (DIAG) {
    puts(">>> Entering extend_contour");
    printf("  >>> extend_contour: there are %d points\n",ncoords);
    if (ncoords > 0) {
      for (i=0; i<ncoords; i++) printf("%f ",xcoordlist[i]);
      printf("\n");
      for (i=0; i<ncoords; i++) printf("%f ",ycoordlist[i]);
      printf("\n");
    }
  }

  /* Make space for new coordinates */
  xcoordlist = (float *) realloc(xcoordlist,(ncoords+1)*sizeof(float));
  ycoordlist = (float *) realloc(ycoordlist,(ncoords+1)*sizeof(float));
  if (DIAG) puts("  >>> extend_contour: done reallocs");
  if (!xcoordlist) {
      puts("extend_contour: cannot allocate memory for y coordinate list");
      return 0;
  }
  if (!ycoordlist) {
      puts("extend_contour: cannot allocate memory for y coordinate list");
      return 0;
  }

  /* Add the next coordinate */
  if (DIAG) printf("  >>> extend_contour: About to add to list...new coords %d: %f,%f\n",
		   ncoords+1,xpos,ypos);
  xcoordlist[ncoords] = xpos;
  ycoordlist[ncoords] = ypos;
  ncoords++;

  /* Update the pointers */
  *xptr = xcoordlist;
  *yptr = ycoordlist;
  *nptr = ncoords;

  if (DIAG) {
    printf("  >>> extend_contour: now there are %d points\n",ncoords);
    if (ncoords > 0) {
      for (i=0; i<ncoords; i++) printf("%f ",xcoordlist[i]);
      printf("\n");
      for (i=0; i<ncoords; i++) printf("%f ",ycoordlist[i]);
      printf("\n");
    }
  }

  if (DIAG) puts(">>> Leaving extend_contour");
  return 1;
}

/*------------------------------------------------------------------------------*/

/* Reverse the order of the contour points */
int flip_contour(int ncoords,float **xptr,float **yptr)
{
  /* xptr, yptr are pointers to the address of xcoord[0], ycoord[0] */
  float *xcoordlist=*xptr,*ycoordlist=*yptr;
  int i=0,j=(ncoords-1);
  float xtemp,ytemp;

  if (DIAG) puts(">>> Entering flip_contour");

  if (!ncoords) {
    if (DIAG) puts("No coordinates - nothing to do, returning");
    return 1;
  } 

  if (DIAG) printf("Initially there are %d pairs...\n",ncoords);
  if (DIAG) {
    printf("xcoords = ");
    for (i=0; i<ncoords; i++) {
      printf("%f ",xcoordlist[i]);
    }
    printf("\nycoords = ");
    for (i=0; i<ncoords; i++) {
      printf("%f ",ycoordlist[i]);
    }
    printf("\n");
  }

  i = 0;
  while(i<j) {
    if (DIAG) printf("  >>> flip_contour: swapping elements %d and %d\n",i,j);
    xtemp = xcoordlist[j];
    ytemp = ycoordlist[j];
    xcoordlist[j] = xcoordlist[i];
    ycoordlist[j] = ycoordlist[i];
    xcoordlist[i] = xtemp;
    ycoordlist[i] = ytemp;
    i++;
    j--;
  }
  /* Update the pointers */
  if (*xptr != xcoordlist) {
      puts("  >>> flip_contour: x pointer has moved?");
      *xptr = xcoordlist;
  }
  if (*yptr != ycoordlist) {
      puts("  >>> flip_contour: y pointer has moved?");
      *yptr = ycoordlist;
  }

  if (DIAG) puts("Finally...");
  if (DIAG) {
    printf("xcoords = ");
    for (i=0; i<ncoords; i++) {
      printf("%f ",xcoordlist[i]);
    }
    printf("\nycoords = ");
    for (i=0; i<ncoords; i++) {
      printf("%f ",ycoordlist[i]);
    }
    printf("\n");
  }
  if (DIAG) puts(">>> Leaving flip_contour");
  return 1;
}

/*------------------------------------------------------------------------------*/

/* close_contour

   Once a complete contour has been traced, check whether it should
   be open or closed; if the latter then make the last point of a closed
   contour equal to the first
*/
int close_contour(int *nptr, float **xptr, float **yptr, int nx, int ny)
{
  /* xptr, yptr are pointers to the address of xcoord[0], ycoord[0] */
  float *xcoordlist=*xptr,*ycoordlist=*yptr;
  int   n=(*nptr-1);

  if (DIAG) puts(">>> Entered close_contour");

  /* Do we have any points in the contour line? */
  if (!*nptr) {
    if (DIAG) puts("No points in contour - nothing to do");
    return 1;
  }

  /* Check if first point lies on a boundary */
  if (xcoordlist[0]==0.0 || xcoordlist[0]==(nx-1) ||
      ycoordlist[0]==0.0 || ycoordlist[0]==(ny-1)) {
    if (DIAG) printf(">>> close_contour: first point %f,%f is on a boundary\n",
	   xcoordlist[0],ycoordlist[0]);
    /* Check if last point lies on a boundary */
    if  (xcoordlist[n]==0.0 || xcoordlist[n]==(nx-1) ||
	 ycoordlist[n]==0.0 || ycoordlist[n]==(ny-1)) {
      if (DIAG) printf(">>> close_contour: last point %f,%f is on a boundary\n",
	     xcoordlist[n],ycoordlist[n]);
      if (DIAG) puts(">>> close_contour: this is an open contour, exiting");
      return 1;
    }
  }

  /* Assume closed contour */
  if (DIAG) puts(">>> close_contour: extending contour...");
  extend_contour(xcoordlist[0],ycoordlist[0],nptr,xptr,yptr);
  if (DIAG) puts(">>> Leaving close_contour");
  return 1;
}
	
