/*
     ccp4mapwish_canvas.c: Functions for handling Tk canvases
     Copyright (C) 2001 CCLRC, Peter Briggs

     This code is distributed under the terms and conditions of the
     CCP4 Program Suite Licence Agreement as a CCP4 Application.
     A copy of the CCP4 licence can be obtained by writing to the
     CCP4 Secretary, Daresbury Laboratory, Warrington WA4 4AD, UK.
*/

/*******************************************************************/

/* ccp4mapwish_canvas.c */

/*******************************************************************/

/* Functions for handling canvases used to display 2d contoured
   sections */

/*******************************************************************/

/* Include header file */

/*******************************************************************/

#include "ccp4mapwish.h"

/*******************************************************************/

/* Private local variables */

/*******************************************************************/

static CCP4CONTOURCANVASARRAY ccp4mapwish_canvasarray;

/*******************************************************************/

/* Canvas-structure storage and retrieval functions */

/*******************************************************************/

/* This must be called once to set up the array of canvases stored in
   memory
*/
int init_ccp4canvasarray(void) {
  ccp4mapwish_canvasarray.ncanvas = 0;
  ccp4mapwish_canvasarray.canvas_ptr = NULL;
  return 1;
}

/* Returns the pointer to the canvas structure with handle canvas, or
   NULL if no such handle can be found */
CCP4CONTOURCANVAS* fetch_canvas_ptr(char *canvas)
{
  int i,ncanvas;
  CCP4CONTOURCANVAS *canvas_ptr;

  if (!canvas || ccp4mapwish_canvasarray.ncanvas == 0) return NULL;
  ncanvas = ccp4mapwish_canvasarray.ncanvas;
  for (i=0; i<ncanvas; i++) {
    canvas_ptr = ccp4mapwish_canvasarray.canvas_ptr[i];
    if (canvas_ptr) {
      if (strcmp(canvas,canvas_ptr->canvas) == 0) return canvas_ptr;
    }
  }
  return NULL;
}

/* Sets up a new canvas structure in the array with the correct
   handle canvas, and returns the pointer to that structure
   (or NULL if there was an error) */
CCP4CONTOURCANVAS* new_canvas_ptr(char *canvas)
{
  int ncanvas,lenid;
  char *id_ptr=NULL;
  CCP4CONTOURCANVAS *canvas_ptr=NULL;
  CCP4CONTOURCANVAS **arr_ptr;

  if (!canvas) return NULL;
  if (CCP4MAPWISH_CANVAS_DIAG) printf("new_canvas_ptr: canvas name is \"%s\"\n",canvas);

  /* Check that the canvas exists */
  if (!canvas_exists(canvas)) {
    printf("new_canvas_ptr: canvas \"%s\" doesn't exist\n",canvas);
    return NULL;
  }

  /* Allocate memory for a new canvas */
  canvas_ptr = (CCP4CONTOURCANVAS *) malloc(sizeof(CCP4CONTOURCANVAS));
  if (!canvas_ptr) return NULL;
  if (CCP4MAPWISH_CANVAS_DIAG) 
    puts("new_canvas_ptr: memory allocated for new canvas_ptr");

  /* Set the canvasid for the new canvas */
  lenid = strlen(canvas);
  if (CCP4MAPWISH_CANVAS_DIAG)
    printf("new_canvas_ptr: length of id string is %d\n",lenid);
  id_ptr = (char *) malloc((lenid+1)*sizeof(char));
  if (!id_ptr) {
    if (canvas_ptr) free(canvas_ptr);
    return NULL;
  }
  if (CCP4MAPWISH_CANVAS_DIAG)
    puts("new_canvas_ptr: memory allocated for new id string");
  strncpy(id_ptr,canvas,lenid);
  id_ptr[lenid] = '\0';
  if (CCP4MAPWISH_CANVAS_DIAG)
    printf("new_canvas_ptr: id string copied, \"%s\"\n",id_ptr);

  /* Add a pointer to the new canvas in the canvas array */
  ncanvas = ccp4mapwish_canvasarray.ncanvas;
  if (CCP4MAPWISH_CANVAS_DIAG) 
    printf("new_canvas_ptr: number of canvas currently stored = %d\n",ncanvas);
  arr_ptr = ccp4mapwish_canvasarray.canvas_ptr;
  ncanvas++;
  arr_ptr = (CCP4CONTOURCANVAS **)
    realloc(arr_ptr,(ncanvas*sizeof(CCP4CONTOURCANVAS *)));
  if (!arr_ptr) {
    if (canvas_ptr) free(canvas_ptr);
    if (id_ptr) free(id_ptr);
    return NULL;
  }
  if (CCP4MAPWISH_CANVAS_DIAG)
    puts("new_canvas_ptr: reallocated memory for array of canvas");
  ccp4mapwish_canvasarray.ncanvas = ncanvas;
  ccp4mapwish_canvasarray.canvas_ptr = arr_ptr;
  ccp4mapwish_canvasarray.canvas_ptr[ncanvas-1] = canvas_ptr;
  canvas_ptr->canvas = id_ptr;
  if (CCP4MAPWISH_CANVAS_DIAG)
    puts("new_canvas_ptr: set all pointers, finished");
  return canvas_ptr;
}

/* Remove a canvas pointer with canvas path name "canvas" from the
   canvas array
*/
int zap_canvas_ptr(char *canvas)
{
  int ncanvas,i,j;
  CCP4CONTOURCANVAS *canvas_ptr;

  if (!canvas || ccp4mapwish_canvasarray.ncanvas == 0) return 0;
  ncanvas = ccp4mapwish_canvasarray.ncanvas;
  j = ncanvas - 1;
  for (i=0; i<ncanvas; i++) {
    canvas_ptr = ccp4mapwish_canvasarray.canvas_ptr[i];
    if (canvas_ptr) {
      if (CCP4MAPWISH_CANVAS_DIAG)
	printf("zap_canvas_ptr: path name for canvas %d is \"%s\"\n",
	       i,canvas_ptr->canvas);
      if (strcmp(canvas,canvas_ptr->canvas) == 0) {
	if (canvas_ptr->canvas) free(canvas_ptr->canvas);
	/* Overwrite this position with the canvas info from the
	   end of the array */
	if (i!=j)
	  ccp4mapwish_canvasarray.canvas_ptr[i] = ccp4mapwish_canvasarray.canvas_ptr[j];
	ncanvas--;
	ccp4mapwish_canvasarray.canvas_ptr = (CCP4CONTOURCANVAS **)
	  realloc(ccp4mapwish_canvasarray.canvas_ptr,
		  (sizeof(CCP4CONTOURCANVAS *))*ncanvas);
	ccp4mapwish_canvasarray.ncanvas = ncanvas;
	return 1;
      }
    }
  }
  if (CCP4MAPWISH_CANVAS_DIAG)
    printf("zap_canvas_ptr: path name \"%s\" not found\n",canvas);
  return 1;
}

/*******************************************************************/

/* Functions for interacting with a Tk canvas  */

/*******************************************************************/

/* canvas_draw_line

   This does the dirty business of generating and executing a Tcl command
   which will draw lines directly onto the canvas.

   canvas is the name of the canvas variable
   x,y are lists of coordinates npts in length and are the points which are
   to be plotted.

   tags is a string of tags to be associated with the line.

   arrow is NULL, unless arrow-head(s) are to be drawn in which case it can
   be one of "first", "last", "both" or "none" (see the Tk canvas man page).

   units is one of ' ' 'm' 'c' 'i' denoting coordinates in pixels,
   millimetres, centimetres or inches respectively (see the Tk canvas man
   page).
*/
int canvas_draw_line(char *canvas, int npts, float *x, float *y, char *tags,
		     char *arrow, char units)
{
  int  i;
  char *command=NULL,tmpstr[CCP4_LTMPSTR];
  Tcl_Interp *interp;

  if (CCP4MAPWISH_CANVAS_DIAG) puts("canvas_draw_line: starting");

  /* Check canvas name is set */
  if (!canvas) {
    puts("canvas_draw_line: canvas name is unset");
    return 0;
  }
  if (CCP4MAPWISH_CANVAS_DIAG)
    printf("canvas_draw_line: canvas name is \"%s\"\n",canvas);

  /* Fetch the pointer to the Tcl interpreter */
  interp = get_tcl_interp_ptr(NULL);
  if (CCP4MAPWISH_CANVAS_DIAG)
    puts("canvas_draw_line: fetched pointer to interpreter");
  if (!interp) {
    puts("canvas_draw_line: serious error, interpreter is null!");
    return 0;
  }

  /* Need at least two points to make a line */
  if (npts < 2) {
    if (CCP4MAPWISH_CANVAS_DIAG)
      puts("canvas_draw_line: only one point supplied for line, returning");
    return 1;
  }

  /* Build the Tcl command */
  /* Always start with canvas_name create line\0 */
  if (CCP4MAPWISH_CANVAS_DIAG)
    puts("canvas_draw_line: about to build draw command");
  if ((strlen(canvas)+12) >= CCP4_LTMPSTR) {
    puts("canvas_draw_line: tmprstr too small for initial canvas command");
    return 0;
  }
  sprintf(tmpstr,"%s create line",canvas);
  command = (char *) malloc((strlen(tmpstr)+1)*sizeof(char));
  if (!command) {
    puts("canvas_draw_line: couldn't allocate memory for command");
    return 0;
  }
  strcpy(command,tmpstr);
  if (CCP4MAPWISH_CANVAS_DIAG)
    printf("canvas_draw_line: command so far:\"%s\"\n",command);

  /* Add each pair of coordinates to the command */
  if (CCP4MAPWISH_CANVAS_DIAG)
    puts("canvas_draw_line: adding coordinates ...");
  for (i=0; i<npts; i++) {
    sprintf(tmpstr," %12f%c %12f%c",x[i],units,y[i],units);
    command = append_tcl_command(command,tmpstr);
    if (!command) {
      puts("canvas_draw_line: couldn't reallocate memory for points in command");
      return 0;
    }
  }
  if (CCP4MAPWISH_CANVAS_DIAG)
    puts("canvas_draw_line: ...done adding coordinates");

  /* Add a fill colour */
  if (CCP4MAPWISH_CANVAS_DIAG)
    puts("canvas_draw_line: adding -fill option");
  command = append_tcl_command(command," -fill \"black\"");
  if (CCP4MAPWISH_CANVAS_DIAG)
    puts("canvas_draw_line: ...done adding -fill option");

  /* Add the tags */
  if (CCP4MAPWISH_CANVAS_DIAG) puts("canvas_draw_line: adding tags ...");
  if (strlen(tags)>0) {
    sprintf(tmpstr," -tags { %s }",tags);
    command = append_tcl_command(command,tmpstr);
    if (!command) {
      puts("canvas_draw_line: couldn't reallocate memory for tags in command");
      return 0;
    }
  }
  if (CCP4MAPWISH_CANVAS_DIAG)
    puts("canvas_draw_line: ...done adding tags");

  /* Add the arrowheads, if requested */
  if (arrow) {
    if (CCP4MAPWISH_CANVAS_DIAG)
      puts("canvas_draw_line: arrow requested, adding command ...");
    sprintf(tmpstr," -arrow %s",arrow);
    command = append_tcl_command(command,tmpstr);
    if (!command) {
      puts("canvas_draw_line: couldn't reallocate memory for arrow in command");
      return 0;
    }
    if (CCP4MAPWISH_CANVAS_DIAG) 
      puts("canvas_draw_line: ...done adding arrow command");
  }

  /* Execute the command which draws the contours*/
  if (CCP4MAPWISH_CANVAS_DIAG) {
    printf("canvas_draw_line: command is \"%s\"\n",command);
    puts("canvas_draw_line: about to evaluate the command...");
  }
  if (Tcl_Eval(interp,command) != TCL_OK) {
    printf("canvas_draw_line: error from Tcl_Eval: %s",interp->result);
    return 0;
  }
  if (CCP4MAPWISH_CANVAS_DIAG)
    puts("canvas_draw_line: command evaluated - freeing memory");

  /* Free memory used for the command */
  if(command) free(command);
  if (CCP4MAPWISH_CANVAS_DIAG)
    puts("canvas_draw_line: memory freed - returning");

  return 1;
}

/* canvas_draw_rectangle

   This does the dirty business of generating and executing a Tcl command
   which will draw rectangles directly onto the canvas.
   It is essentially a copy of canvas_draw_line

   Nb although in grid coordinates the original shape will be a rectangle,
   coordinate transforms may distort it into a parallelogram.
   Thus this function interfaces to "canvas polygon" rather than "canvas
   rectangle".

   canvas is the name of the canvas variable
   x,y are lists of coordinates npts in length and are the points which are
   to be plotted.
   npts must be 4, and x and y must hold the positions of the vertices of
   the polygon in order.

   tags is a string of tags to be associated with the rectangle.

   color is a color stringe acceptable to Tk_GetColor (see the Tk canvas 
   man page).

   units is one of ' ' 'm' 'c' 'i' denoting coordinates in pixels,
   millimetres, centimetres or inches respectively (see the Tk canvas man
   page).
*/
int canvas_draw_rectangle(char *canvas, int npts, float *x, float *y,
			  char *tags, char *color, char units)
{
  int   i;
  char  *command=NULL,tmpstr[CCP4_LTMPSTR];
  Tcl_Interp *interp;

  if (CCP4MAPWISH_CANVAS_DIAG) puts("canvas_draw_rectangle: starting");

  /* Check canvas name is set */
  if (!canvas) {
    puts("canvas_draw_rectangle: canvas name is unset");
    return 0;
  }
  if (CCP4MAPWISH_CANVAS_DIAG)
    printf("canvas_draw_rectangle: canvas name is \"%s\"\n",canvas);

  /* Fetch the pointer to the Tcl interpreter */
  interp = get_tcl_interp_ptr(NULL);
  if (CCP4MAPWISH_CANVAS_DIAG)
    puts("canvas_draw_rectangle: fetched pointer to interpreter");
  if (!interp) {
    puts("canvas_draw_rectangle: serious error, interpreter is null!");
    return 0;
  }

  /* Need exactly two points to define a rectangle */
  if (npts != 4) {
    if (CCP4MAPWISH_CANVAS_DIAG)
      puts("canvas_draw_rectangle: need exactly four points for rectangle, returning");
    return 1;
  }

  /* Build the Tcl command */
  /* Always start with canvas_name create rectangle\0 */
  if (CCP4MAPWISH_CANVAS_DIAG)
    puts("canvas_draw_rectangle: about to build draw command");
  if ((strlen(canvas)+18) >= CCP4_LTMPSTR) {
    puts("canvas_draw_rectangle: tmprstr too small for initial canvas command");
    return 0;
  }
  sprintf(tmpstr,"%s create polygon",canvas);
  command = (char *) malloc((strlen(tmpstr)+1)*sizeof(char));
  if (!command) {
    puts("canvas_draw_rectangle: couldn't allocate memory for command");
    return 0;
  }
  strcpy(command,tmpstr);
  if (CCP4MAPWISH_CANVAS_DIAG)
    printf("canvas_draw_rectangle: command so far:\"%s\"\n",command);

  /* Add each pair of coordinates to the command */
  if (CCP4MAPWISH_CANVAS_DIAG)
    puts("canvas_draw_rectangle: adding coordinates ...");
  for (i=0; i<npts; i++) {
    sprintf(tmpstr," %12f%c %12f%c",x[i],units,y[i],units);
    command = append_tcl_command(command,tmpstr);
    if (!command) {
      puts("canvas_draw_rectangle: couldn't reallocate memory for points in command");
      return 0;
    }
  }
  if (CCP4MAPWISH_CANVAS_DIAG)
    puts("canvas_draw_rectangle: ...done adding coordinates");

  /* Add a fill colour */
  if (CCP4MAPWISH_CANVAS_DIAG)
    puts("canvas_draw_rectangle: adding -fill option");
  command = append_tcl_command(command," -fill ");
  command = append_tcl_command(command,color);
  if (CCP4MAPWISH_CANVAS_DIAG)
    puts("canvas_draw_rectangle: ...done adding -fill option");

  /* Set outline to blank */
  command = append_tcl_command(command," -outline {}");

  /* Add the tags */
  if (CCP4MAPWISH_CANVAS_DIAG) puts("canvas_draw_rectangle: adding tags ...");
  if (strlen(tags)>0) {
    sprintf(tmpstr," -tags { %s }",tags);
    command = append_tcl_command(command,tmpstr);
    if (!command) {
      puts("canvas_draw_rectangle: couldn't reallocate memory for tags in command");
      return 0;
    }
  }
  if (CCP4MAPWISH_CANVAS_DIAG)
    puts("canvas_draw_rectangle: ...done adding tags");

  /* Execute the command which draws the contours*/
  if (CCP4MAPWISH_CANVAS_DIAG) {
    printf("canvas_draw_rectangle: command is \"%s\"\n",command);
    puts("canvas_draw_rectangle: about to evaluate the command...");
  }
  if (Tcl_Eval(interp,command) != TCL_OK) {
    printf("canvas_draw_rectangle: error from Tcl_Eval: %s",interp->result);
    printf("canvas_draw_rectangle:\ncanvas = \"%s\"\ntags = \"%s\"\ncolor = \"%s\"\nunits = \"%c\"\n",canvas,tags,color,units);
    return 0;
  }
  if (CCP4MAPWISH_CANVAS_DIAG)
    puts("canvas_draw_rectangle: command evaluated - freeing memory");

  /* Free memory used for the command */
  if(command) free(command);
  if (CCP4MAPWISH_CANVAS_DIAG)
    puts("canvas_draw_rectangle: memory freed - returning");

  return 1;
}

/*------------------------------------------------------------------------------*/

/* canvas_draw_label

   Place text on the canvas at the specified point.
*/
int canvas_draw_label(char *canvas, char *text, float x, float y,
		      char *anchor, char *tags, char units)
{
  char *command=NULL,tmpstr[CCP4_LTMPSTR];
  Tcl_Interp *interp;

  /* Check canvas name is set */
  if (!canvas) {
    puts("canvas_draw_label: canvas name is unset");
    return 0;
  }

  /* Fetch the pointer to the Tcl interpreter */
  interp = get_tcl_interp_ptr(NULL);
  if (CCP4MAPWISH_CANVAS_DIAG) puts("canvas_draw_label: fetched pointer to interpreter");
  if (!interp) {
    puts("canvas_draw_label: serious error, interpreter is null!");
    return 0;
  }

  /* Build the Tcl command */
  /* Always start with canvas_name create text\0 */
  if ((strlen(canvas)+12) >= CCP4_LTMPSTR) {
    puts("canvas_draw_label: tmprstr too small for initial canvas command");
    return 0;
  }
  sprintf(tmpstr,"%s create text",canvas);
  command = (char *) malloc((strlen(tmpstr)+1)*sizeof(char));
  if (!command) {
    puts("canvas_draw_label: couldn't allocate memory for command");
    return 0;
  }
  strcpy(command,tmpstr);

  /* Add the coordinate pair to the command */
  sprintf(tmpstr," %12f%c %12f%c",x,units,y,units);
  command = append_tcl_command(command,tmpstr);
  if (!command) {
    puts("canvas_draw_label: couldn't reallocate memory for points in command");
    return 0;
  }
  
  /* Add the text for the label to the command
     This is " -text <label>" */
  command = append_tcl_command(command," -text {");
  if (!command) {
    puts("canvas_draw_label: couldn't reallocate memory for -text in command");
    return 0;
  }
  command = append_tcl_command(command,text);
  if (!command) {
    puts("canvas_draw_label: couldn't reallocate memory for text string in command");
    return 0;
  }
  command = append_tcl_command(command,"}");
  if (!command) {
    puts("canvas_draw_label: couldn't reallocate memory for closing quote in command");
    return 0;
  }

  /* Add the anchor */
  if (CCP4MAPWISH_CANVAS_DIAG)
    printf("canvas_draw_label: anchor position is %s\n",anchor);
  sprintf(tmpstr," -anchor %s",anchor);
  command = append_tcl_command(command,tmpstr);
  if (!command) {
    puts("canvas_draw_label: couldn't reallocate memory for anchor in command");
    return 0;
  }

  /* Add the tags */
  if (strlen(tags)>0) {
    sprintf(tmpstr," -tags { %s }",tags);
    command = append_tcl_command(command,tmpstr);
    if (!command) {
      puts("canvas_draw_label: couldn't reallocate memory for tags in command");
      return 0;
    }
  }

  /* Execute the command which draws the label */
  if (CCP4MAPWISH_CANVAS_DIAG) printf("canvas_draw_label: command is: %s\n",command);
  if (Tcl_Eval(interp,command) != TCL_OK) {
    printf("canvas_draw_label: error from Tcl_Eval: %s\n",interp->result);
    return 0;
  }

  /* Free memory used for the command */
  if(command) free(command);

  return 1;

}

/*------------------------------------------------------------------------------*/

/* canvas_scale_canvas

   Scale elements with "tag" in the canvas by the specified factor.

   Return 1 on success, 0 if there was an error.
*/
int canvas_scale_canvas(char *canvas, char *tag, float scale)
{
  char *command=NULL,tmpstr[CCP4_LTMPSTR];
  Tcl_Interp *interp;

  /* Check the canvas name is real */
  if (!canvas) {
    puts("canvas_scale_canvas: canvas name is unset");
    return 0;
  }

  /* Check the tag is real */
  if (!tag) {
    puts("canvas_scale_canvas: tag is unset");
    return 0;
  }

  /* Fetch the pointer to the Tcl interpreter */
  interp = get_tcl_interp_ptr(NULL);
  if (CCP4MAPWISH_CANVAS_DIAG)
    puts("canvas_scale_canvas: fetched pointer to interpreter");
  if (!interp) {
    puts("canvas_scale_canvas: serious error, interpreter is null!");
    return 0;
  }

  /* Build the scale command
     <canvas> scale <tag> 0 0 <xscale> <yscale>
  */
  /* Always start with canvas_name scale\0 */
  if ((strlen(canvas)+8) >= CCP4_LTMPSTR) {
    puts("canvas_scale_canvas: tmprstr too small for initial canvas command");
    return 0;
  }
  sprintf(tmpstr,"%s scale",canvas);
  command = (char *) malloc((strlen(tmpstr)+1)*sizeof(char));
  if (!command) {
    puts("canvas_scale_canvas: couldn't allocate memory for command");
    return 0;
  }
  strcpy(command,tmpstr);

  /* Add the tag */
  if (strlen(tag)>0) {
    sprintf(tmpstr," %s",tag);
    command = append_tcl_command(command,tmpstr);
    if (!command) {
      puts("canvas_scale_canvas: couldn't reallocate memory for tag in command");
      return 0;
    }
  }

  /* Add the origin and scale factors */
  sprintf(tmpstr," 0 0 %f %f",scale,scale);command = append_tcl_command(command,tmpstr);
  if (!command) {
    puts("canvas_scale_canvas: couldn't reallocate memory for scale factors in command");
    return 0;
  }

  /* Execute the command which scales the canvas */
  if (CCP4MAPWISH_CANVAS_DIAG) printf("canvas_scale_canvas: command is: %s\n",command);
  if (Tcl_Eval(interp,command) != TCL_OK) {
    printf("canvas_scale_canvas: error from Tcl_Eval: %s\n",interp->result);
    return 0;
  }

  /* Free memory used for the command */
  if(command) free(command);

  return 1;
}

/*------------------------------------------------------------------------------*/

/* canvas_shift_canvas

   Move and resize the canvas so that all elements are visible.
   This requires execution of a "canvas scale" command followed by
   a "canvas move" command.

   Return 1 on success, 0 on error.
*/
int canvas_shift_canvas(char *canvas)
{
  char *command=NULL,tmpstr[CCP4_LTMPSTR];
  Tcl_Interp *interp;
  int rheight,rwidth,shiftx,shifty,limits[4];
  CCP4CONTOURCANVAS *canvasptr;

  /* Check the canvas name is real */
  if (!canvas) {
    puts("canvas_shift_canvas: canvas name is unset");
    return 0;
  }
  /* Fetch the pointer to the appropriate canvas structure */
  canvasptr = fetch_canvas_ptr(canvas);
  if (!canvasptr) {
    puts("canvas_shift_canvas: serious error, cannot retrieve canvas information");
    return 0;
  }

  /* Fetch the pointer to the Tcl interpreter */
  interp = get_tcl_interp_ptr(NULL);
  if (CCP4MAPWISH_CANVAS_DIAG)
    puts("canvas_shift_canvas: fetched pointer to interpreter");
  if (!interp) {
    puts("canvas_shift_canvas: serious error, interpreter is null!");
    return 0;
  }

  /* Fetch the limits of the canvas */
  if (!canvas_get_limits(canvas,limits)) {
    puts("canvas_shift_canvas: error in trying to get the canvas limits");
    return 0;
  }

  /* Determine the new height and width of the canvas */
  rheight = limits[3] - limits[1];
  rwidth  = limits[2] - limits[0];

  /* Set the shifts required to reset the origin */
  shiftx  = -limits[0];
  shifty  = -limits[1];

  /* Build the resize command
     <canvas> configure -height <rheight> -width <rwidth>
  */
  /* Always start with canvas_name configure\0 */
  if ((strlen(canvas)+11) >= CCP4_LTMPSTR) {
    puts("canvas_shift_canvas: tmprstr too small for initial canvas command");
    return 0;
  }
  sprintf(tmpstr,"%s configure",canvas);
  command = (char *) malloc((strlen(tmpstr)+1)*sizeof(char));
  if (!command) {
    puts("canvas_move_canvas: couldn't allocate memory for command");
    return 0;
  }
  strcpy(command,tmpstr);

  /* Add the height and width tags */
  sprintf(tmpstr," -height %d -width %d",rheight,rwidth);
  command = append_tcl_command(command,tmpstr);
  if (!command) {
    puts("canvas_shift_canvas: couldn't reallocate memory for height/width in command");
    return 0;
  }

  /* Execute the command which resizes the canvas */
  if (CCP4MAPWISH_CANVAS_DIAG) printf("canvas_shift_canvas: command is: %s\n",command);
  if (Tcl_Eval(interp,command) != TCL_OK) {
    printf("canvas_shift_canvas: error from Tcl_Eval: %s\n",interp->result);
    return 0;
  }

  /* Free memory used for the command */
  if(command) free(command);

  /* Shift the origin */
  if (!canvas_shift_element(canvas,"all",shiftx,shifty)) {
    printf("canvas_shift_canvas: error returned from canvas_shift_element\n");
    return 0;
  }

  /* Store the shifts */
  canvasptr->xoffset = shiftx;
  canvasptr->yoffset = shifty;

  return 1;
}

/*------------------------------------------------------------------------------*/

/* canvas_shift_element

   Move the canvas element(s) denoted by the specified tags by the specified
   number of pixels.
   This requires execution of the "canvas move" command.

   Return 1 on success, 0 on error.
*/
int canvas_shift_element(char *canvas, char *tag, int shiftx, int shifty)
{
  char *command=NULL,tmpstr[CCP4_LTMPSTR];
  Tcl_Interp *interp;

  /* Check the canvas name is real */
  if (!canvas) {
    puts("canvas_shift_element: canvas name is unset");
    return 0;
  }

  /* Fetch the pointer to the Tcl interpreter */
  interp = get_tcl_interp_ptr(NULL);
  if (CCP4MAPWISH_CANVAS_DIAG)
    puts("canvas_shift_element: fetched pointer to interpreter");
  if (!interp) {
    puts("canvas_shift_element: serious error, interpreter is null!");
    return 0;
  }

  /* Build the command to apply the shift:
     <canvas> move <tag> <x> <y> */
  if ((strlen(canvas)+10) >= CCP4_LTMPSTR) {
    puts("canvas_shift_canvas: tmprstr too small for initial canvas command");
    return 0;
  }
  sprintf(tmpstr,"%s move",canvas);
  command = (char *) malloc((strlen(tmpstr)+1)*sizeof(char));
  if (!command) {
    puts("canvas_shift_element: couldn't allocate memory for command");
    return 0;
  }
  strcpy(command,tmpstr);

  /* Add the tag */
  sprintf(tmpstr," %s",tag);
  command = append_tcl_command(command,tmpstr);
  if (!command) {
    puts("canvas_shift_element: couldn't reallocate memory for offsets in command");
    return 0;
  }

  /* Add the x and y shifts */
  sprintf(tmpstr," %d %d",shiftx,shifty);
  command = append_tcl_command(command,tmpstr);
  if (!command) {
    puts("canvas_shift_element: couldn't reallocate memory for offsets in command");
    return 0;
  }

  /* Execute the command which resizes the canvas */
  if (CCP4MAPWISH_CANVAS_DIAG) printf("canvas_shift_canvas: command is: %s\n",command);
  if (Tcl_Eval(interp,command) != TCL_OK) {
    printf("canvas_shift_element: error from Tcl_Eval: %s\n",interp->result);
    return 0;
  }

  /* Free memory used for the command */
  if(command) free(command);

  return 1;
}

/*------------------------------------------------------------------------------*/

/* canvas_delete

   Deletes the item(s) identified with "tag" on the specified canvas,
   by issuing a "canvas delete ..." command.

   Returns 1 on success, 0 on error.
*/
int canvas_delete(char *canvas, char *tag)
{
  char *command=NULL,tmpstr[CCP4_LTMPSTR];
  Tcl_Interp *interp;

  /* Check the canvas name is real */
  if (!canvas) {
    puts("canvas_delete: canvas name is unset");
    return 0;
  }

  /* Check the tag is real */
  if (!tag) {
    puts("canvas_delete: \"tag\" is unset");
    return 0;
  }

  /* Fetch the pointer to the Tcl interpreter */
  interp = get_tcl_interp_ptr(NULL);
  if (CCP4MAPWISH_CANVAS_DIAG) puts("canvas_delete: fetched pointer to interpreter");
  if (!interp) {
    puts("canvas_delete: serious error, interpreter is null!");
    return 0;
  }

  /* Build the delete command
     <canvas> delete <tag>
  */
  if ((strlen(canvas)+8) >= CCP4_LTMPSTR) {
    puts("canvas_delete: tmprstr too small for initial canvas command");
    return 0;
  }
  sprintf(tmpstr,"%s delete",canvas);
  command = (char *) malloc((strlen(tmpstr)+1)*sizeof(char));
  if (!command) {
    puts("canvas_delete: couldn't allocate memory for command");
    return 0;
  }
  strcpy(command,tmpstr);

  /* Add the tag */
  if (strlen(tag)>0) {
    sprintf(tmpstr," %s",tag);
    command = append_tcl_command(command,tmpstr);
    if (!command) {
      puts("canvas_delete: couldn't reallocate memory for tag in command");
      return 0;
    }
  }

  /* Execute the command which deletes the tagged items */
  if (CCP4MAPWISH_CANVAS_DIAG) printf("canvas_delete: command is: %s\n",command);
  if (Tcl_Eval(interp,command) != TCL_OK) {
    printf("canvas_scale_canvas: error from Tcl_Eval: %s\n",interp->result);
    return 0;
  }

  /* Free memory used for the command */
  if(command) free(command);
  
  return 1;
}

/*------------------------------------------------------------------------------*/

/* canvas_exists

   Test whether canvas exists

   Return 1 if it does, 0 otherwise
*/

int canvas_exists(char *canvas)
{
  int istatus;
  char tmpstr[CCP4_LTMPSTR],*endptr;
  Tcl_Interp *interp;

  /* Fetch the pointer to the Tcl interpreter */
  interp = get_tcl_interp_ptr(NULL);
  if (!interp) return 0;

  sprintf(tmpstr,"winfo exists %s",canvas);
  if (Tcl_Eval(interp,tmpstr) != TCL_OK) {
    printf("canvas_exists: error from Tcl_Eval: %s",interp->result);
    return 0;
  }
  /* Extract a long int using strtol
     If no conversion takes place then "endptr" will be equal
     to "result" */
  istatus = (int) strtol(interp->result,&endptr,10);
  if (strcmp(interp->result,endptr) == 0) {
    printf("canvas_exists: bad return status from winfo command \"%s\"\n",interp->result);
    return 0;
  }
  if (!istatus) {
    printf("canvas_exists: canvas \"%s\" doesn't exist\n",canvas);
    return 0;
  }
  return 1;
}

/*------------------------------------------------------------------------------*/

/* canvas_get_tk_scaling

   Return the scaling of pixels to <unit> where unit is one of m,c,i,p
   (millimetres, centimetres, inches, points)
*/

float canvas_get_tk_scaling(char *canvas, char units)
{
  float tkscale,tkscale_inch,tkscale_cm,tkscale_mm;
  char *endptr;
  Tcl_Interp *interp;

  /* Fetch the pointer to the Tcl interpreter */
  interp = get_tcl_interp_ptr(NULL);
  if (!interp) return (float) 0.0;

  /* Execute the tk scaling command */
  if (Tcl_Eval(interp,"tk scaling") != TCL_OK) {
    printf("canvas_get_tk_scaling: error from Tcl_Eval: %s",interp->result);
    return (float) 0.0;
  }

  /* Extract a float from the result */
  tkscale = ((float) strtod(interp->result,&endptr));
  if (strcmp(interp->result,endptr) == 0) {
    printf("canvas_get_tk_scaling: couldn't extract scale factor\n");
    return (float) 0.0;
  }
  
  /* tkscale is the number of pixels per point
     Need to do conversion to other units */

  /* Inches
     There are 72 points to an inch */
  tkscale_inch = tkscale * 72.0;

  /* Centimetres
     There are 2.54cm to an inch */
  tkscale_cm = tkscale_inch/2.54;

  /* Millimetres
     There are 10 mm to a cm */
  tkscale_mm = tkscale_cm/10.0;

  /* Return the number of pixels per requested unit */
  switch (units) {
    /* Printers points */
  case 'p':
    return tkscale;
    break;
    /* Inches */
  case 'i':
    return tkscale_inch;
    break;
  case 'c':
    /* Centimetres */
    return tkscale_cm;
    break;
  case 'm':
    /* Millimetres */
    return tkscale_mm;
    break;
  default:
    /* Assume pixels for anything else */
    return (float) 1.0;
  }
}

/*------------------------------------------------------------------------------*/

/* canvas_get_limits

   Fetch the limits of the canvas in pixels and put them into the
   limits array.

   Return 1 on success, 0 if there was an error.
*/

int canvas_get_limits(char *canvas, int limits[4])
{
  Tcl_Interp *interp;
  int  i;
  char *p,command[CCP4_LTMPSTR];

  /* Fetch the pointer to the Tcl interpreter */
  interp = get_tcl_interp_ptr(NULL);
  if (CCP4MAPWISH_CANVAS_DIAG)
    puts("canvas_get_limits: fetched pointer to interpreter");
  if (!interp) {
    puts("canvas_get_limits: serious error, interpreter is null!");
    return 0;
  }

  /* Check that the canvas name is not null */
  if (!canvas) {
    puts("No canvas name - can't get the limits");
    return 0;
  }

  /* Build the command for getting the canvas limits
     This is canvas bbox all */
  if ((strlen(canvas)+10) > CCP4_LTMPSTR) {
    puts("canvas_get_limits: command too long for buffer");
    return TCL_ERROR;
  }
  sprintf(command,"%s bbox all",canvas);

  /* Evaluate the command */
  if (Tcl_Eval(interp,command) != TCL_OK) {
    printf("canvas_get_limits: error from Tcl_Eval: %s\n",interp->result);
    return 0;
  }

  /* Process the result - it should consist of 4 integers */
  /* Use strtok which can split the string based on tokens
     set by the programmer - in this case a space */
  if (!interp->result) {
    puts("canvas_get_limits: failed to fetch canvas limits");
    return TCL_ERROR;
  }
  i = 0;
  p = strtok(interp->result," ");
  if (p == NULL) {
    return 0;
  }
  do {
    limits[i] = atoi(p);
    ++i;
  } while ((p = strtok(NULL," ")) != NULL && i<4);

  /* Finished */
  return 1;
}

/*------------------------------------------------------------------------------*/
