/*
     ccp4mapwish_contour.c: Functions for drawing contoured sections
     Copyright (C) 2001  CCLRC, Peter Briggs

     This code is distributed under the terms and conditions of the
     CCP4 Program Suite Licence Agreement as a CCP4 Application.
     A copy of the CCP4 licence can be obtained by writing to the
     CCP4 Secretary, Daresbury Laboratory, Warrington WA4 4AD, UK.
*/

/*******************************************************************/

/* ccp4mapwish_contour.c */

/*******************************************************************/

/* Functions for drawing the elements of the contoured section
   in a Tk canvas
*/

/*******************************************************************/

/* Include header files */

/*******************************************************************/

#include "ccp4mapwish.h"
#include "c_contour.h"

/*******************************************************************/

/* Private local variables and accessor functions */

/*******************************************************************/

static char        *ccp4mapwish_contour_canvas;
static CCP4SECTION *ccp4mapwish_contour_section;
static char         ccp4mapwish_contour_units;
static int          ccp4mapwish_contour_level;
static float        ccp4mapwish_contour_isolevel;

/* Store a pointer to the name of the canvas */
int init_contour_canvas(char *canvas, CCP4SECTION *section, char units)
{
  if (canvas && section) {
    ccp4mapwish_contour_canvas   = canvas;
    ccp4mapwish_contour_section  = section;
    ccp4mapwish_contour_units    = units;
    ccp4mapwish_contour_level    = -1;
    ccp4mapwish_contour_isolevel = 0.0;
    return 1;
  }
  return 0;
}

/* Retrieve a pointer to the name of the canvas */
char* fetch_contour_canvas(void)
{
  return ccp4mapwish_contour_canvas;
}

/* Retrieve a pointer to the current section */
CCP4SECTION* fetch_contour_section(void)
{
  return ccp4mapwish_contour_section;
}

/* Retrive the units for drawing raw coordinates */
char fetch_contour_units(void)
{
  return ccp4mapwish_contour_units;
}

/*******************************************************************/

/* Drawing functions */

/*******************************************************************/

int greyscale_section(CCP4SECTION *section)
{
  float   **rho,*data,x[4],y[4];
  float   minrho,maxrho;
  double  colorval;
  int     i,j,ii,jj,this_sec,icolorval;
  int     iaxis,iu,iv,npts=4;
  int     u_start,v_start,u_end,v_end,u_ext,v_ext;
  char    *canvas,units,color[10];
  CCP4MAP *map;

  /* Valid section pointer? */
  if (!section) {
    puts("greyscale_section: section pointer is null");
    return 0;
  }

  /* Fetch the pointer to the map */
  map = fetch_map_ptr(section->mapid);
  if (!map) {
    puts("greyscale_section: couldn't fetch map pointer");
    return 0;
  }

  /* Retrieve a pointer to the name of the current canvas */
  canvas = fetch_contour_canvas();
  if (!canvas) {
    puts("greyscale_section: couldn't retrieve the canvas name");
    return 0;
  }

  /* Retrieve the units for drawing to the canvas */
  units = fetch_contour_units();

  /* Don't do greyscale display for slabs */
  if (section->slab_start != section->slab_end) {
    puts("greyscale_section: nothing to do for slabs");
    return 1;
  }
  
  /* Set up index parameters */
  iaxis = section->iaxis;
  iu    = section->iu;
  iv    = section->iv;

  /* Limits in the map
     Need to offset by the start position in the map */
  u_start = section->ulimits[0] - map->start[iu];
  u_end   = section->ulimits[1] - map->start[iu];
  v_start = section->vlimits[0] - map->start[iv];
  v_end   = section->vlimits[1] - map->start[iv];

  u_ext = u_end - u_start + 1;
  v_ext = v_end - v_start + 1;

  /* i points to the "absolute" section number from the mapfile
     This must be reset relative to the section numbers stored in
     memory */
  this_sec = section->nsec - map->start[iaxis];
    
  /* Cache the section within the limits */
  rho = (float **) malloc((u_ext*sizeof(float*)) + (u_ext*v_ext*sizeof(float)));
  if (!rho) {
    puts("contour_section: failed to allocate array for density");
    return 0;
  }
  /* Internal pointers to set up storage mapping function */
  data = (float *) (rho + u_ext);
  for (i=0; i<u_ext; ++i, data += v_ext) {
    rho[i] = data;
    for (j=0; j<v_ext; ++j)
      data[j] = 0;
  }
  maxrho = -99999.9;
  minrho =  99999.9;
  /* Populate the array and find max & min */
  for (i=0; i<u_ext; i++) {
    ii = i + u_start;
    for (j=0; j<v_ext; j++) {
      jj = j + v_start;
      if (iaxis == 0)
	rho[i][j] = map->map[this_sec][ii][jj];
      else if (iaxis == 1)
	rho[i][j] = map->map[jj][this_sec][ii];
      else if (iaxis == 2)
	rho[i][j] = map->map[ii][jj][this_sec];
      /* Max/min */
      if (rho[i][j] > maxrho)
	maxrho = rho[i][j];
      if (rho[i][j] < minrho)
	minrho = rho[i][j];
    }
  }
  /* Generate the greyscale display */
  for (i=0; i<u_ext; i++) {
    for (j=0; j<v_ext; j++) {
      /* Generate the positions of the cell corners (grid coordinates)
	 Nb correct for the boundaries */
      if (i==0)
	x[0] = ((float) i);
      else
	x[0] = ((float) i) - 0.5;
      if (i==(u_ext-1))
	x[2] = ((float) i);
      else
	x[2] = ((float) i) + 0.5;
      if (j==0)
	y[0] = ((float) j);
      else
	y[0] = ((float) j) - 0.5;
      if (j==(v_ext-1))
	y[2] = ((float) j);
      else
	y[2] = ((float) j) + 0.5;

      x[1] = x[0];
      y[1] = y[2];
      x[3] = x[2];
      y[3] = y[0];

      /* Transform the coordinates from grid to orthogonal coordinates */
      if (!transform_coordinates(section,npts,x,y)) {
	puts("greyscale_section: transform_coordinates returned an error");
	return 0;
      }
      /* Generate the fill color as an RGB string */
      colorval = (rho[i][j]-minrho)/(maxrho-minrho);
      colorval = colorval*255.0e0;
      icolorval = (int) colorval;
      sprintf(color,"#%02x%02x%02x",icolorval,icolorval,icolorval);
      color[8]='\0';
      /* Draw a box and fill with the scaled colour */
      if (!canvas_draw_rectangle(canvas,npts,x,y,"scaledcol",color,units)) {
	puts("greyscale_section: draw_rectangle returned an error");
	printf("greyscale_section: colorval = %f, icolorval = %d\n",
	       colorval,icolorval);
	printf("greyscale_section: rho = %f, min(rho) = %f, max(rho) = %f\n",
	       rho[i][j],minrho,maxrho);
	return 0;
      }
    }   
  }

  /* Release memory */
  if (rho) free(rho);

  /* Finished */
  return 1;
}

/* contour_section

   This function generates the contours for all sections
   in the section structure

   Each section is cached, and then the contouring routine is
   called to draw the lines on the specified canvas

   Returns 1 on success, or 0 if there was an error.
*/
int contour_section(CCP4SECTION *section)
{
  float   **rho,*data,*levels;
  int     i,j,k,ii,jj,this_sec,step;
  int     iaxis,iu,iv,nlevels;
  int     u_start,v_start,u_end,v_end,u_ext,v_ext;
  CCP4MAP *map;

  /* Valid section pointer? */
  if (!section) {
    puts("contour_section: section pointer is null");
    return 0;
  }

  /* Fetch the pointer to the map */
  map = fetch_map_ptr(section->mapid);
  if (!map) {
    puts("contour_section: could fetch map pointer");
    return 0;
  }

  /* Check there are contour levels */
  nlevels = section->nlevels;
  if (nlevels <= 0) {
    /* This is not an error */
    if (CCP4MAPWISH_CONTOUR_DIAG)
      puts("contour_section: no contour levels set, nothing to do here");
    return 1;
  }
  levels = section->levels;
  if (!levels) {
    /* This is an error */
    puts("contour_section: empty pointer to array of contour levels");
    return 0;
  }
     
  /* Set up index parameters */
  iaxis = section->iaxis;
  iu    = section->iu;
  iv    = section->iv;

  /* Limits in the map
     Need to offset by the start position in the map */
  u_start = section->ulimits[0] - map->start[iu];
  u_end   = section->ulimits[1] - map->start[iu];
  v_start = section->vlimits[0] - map->start[iv];
  v_end   = section->vlimits[1] - map->start[iv];

  u_ext = u_end - u_start + 1;
  v_ext = v_end - v_start + 1;

  /* Step size */
  step = section->slab_step;
  if (step < 1) {
    printf("contour_section: WARNING slab_step = %d\n",step);
    step = 1;
  }

  /* Loop over sections in the slab */
  for (k=section->slab_start; k<=section->slab_end; k = k+step) {

    if (CCP4MAPWISH_CONTOUR_DIAG)
      printf("contour_section: starting on section %d\n",k);

    /* i points to the "absolute" section number from the mapfile
       This must be reset relative to the section numbers stored in
       memory */
    this_sec = k - map->start[iaxis];

    /* Cache the section within the limits */
    rho = (float **) malloc((u_ext*sizeof(float*)) + (u_ext*v_ext*sizeof(float)));
    if (!rho) {
      puts("contour_section: failed to allocate array for density");
      return 0;
    }
    /* Internal pointers to set up storage mapping function */
    data = (float *) (rho + u_ext);
    for (i=0; i<u_ext; ++i, data += v_ext) {
      rho[i] = data;
      for (j=0; j<v_ext; ++j)
	data[j] = 0;
    }
    /* Populate the array */
    for (i=0; i<u_ext; i++) {
      ii = i + u_start;
      for (j=0; j<v_ext; j++) {
	jj = j + v_start;
	if (iaxis == 0)
	  rho[i][j] = map->map[this_sec][ii][jj];
	else if (iaxis == 1)
	  rho[i][j] = map->map[jj][this_sec][ii];
	else if (iaxis == 2)
	  rho[i][j] = map->map[ii][jj][this_sec];
      }
    }

    /* Pass the array into the contouring routines */
    if (!contour_main(u_ext,v_ext,rho,nlevels,levels,write_contour_line)) {
      puts("contour_section: error from contour_main");
      return 0;
    }

    /* Next section in slab? */
  }

  /* Release memory */
  if (rho) free(rho);

  /* Finished */
  return 1;
}

/*------------------------------------------------------------------------------*/

/* write_contour_line

   This function is supplied as an argument to contour_main
   and is called each time a complete contour line has been traced

   It takes the output of the 2d contouring functions and
   transforms these to the canvas coordinates, then calls canvas_draw_line
   to plot the contour line directly on the current canvas widget.

   Returns 1 on success, or 0 if there was an error.
*/

int write_contour_line(float isovalue, int npts, float *x, float *y)
{
  char         *canvas,tags[CCP4_LTMPSTR],units;
  CCP4SECTION  *section;

  /* Retrieve the canvas name */
  canvas = fetch_contour_canvas();
  if (!canvas) {
    puts("write_contour_line: couldn't retrieve the canvas name");
    return 0;
  }

  /* Retrieve the section pointer */
  section = fetch_contour_section();
  if (!section) {
    puts("write_contour_line: couldn't retrieve the section pointer");
    return 0;
  }

  /* Retrieve the units for drawing to the canvas */
  units = fetch_contour_units();

  /* Transform the coordinates from grid to orthogonal coordinates */
  if (!transform_coordinates(section,npts,x,y)) {
    puts("write_contour_line: transform_coordinates returned an error");
    return 0;
  }

  /* Set the index number for the current contour level */
  if (ccp4mapwish_contour_level < 0) {
    ccp4mapwish_contour_level    = 0;
    ccp4mapwish_contour_isolevel = isovalue;
  } else if (isovalue!=ccp4mapwish_contour_isolevel) {
    ccp4mapwish_contour_level++;
    ccp4mapwish_contour_isolevel = isovalue;
  }

  /* Generate the tags for this line
     All contour lines are tagged "contours"
     Lines for contour level nnn are also tagged as "level_nnn"
     Negative contour lines are are further tagges as "negcontours"
  */
  sprintf(tags,"level_%-3d",ccp4mapwish_contour_level);
  strcat(tags," contours");
  if (isovalue < 0.0) strcat(tags," negcontours");

  if (CCP4MAPWISH_CONTOUR_DIAG)
    printf("write_contour_line: isovalue = %f, level is no. %d, tags are \"%s\"\n",
	   isovalue,ccp4mapwish_contour_level,tags);

  /* Draw the line on the canvas */
  if (!canvas_draw_line(canvas,npts,x,y,tags,NULL,units)) {
    puts("write_contour_line: draw_line returned an error");
    return 0;
  }
  if (CCP4MAPWISH_CONTOUR_DIAG)
    printf("write_contour_line: back from canvas_draw_line\n"); 
  
  return 1;
}

/*------------------------------------------------------------------------------*/

/* write_contour_boundingbox

   Function to draw the bounding box for the current section.
   It is supplied with the section limits (in grid points)
   which are converted into orthogonal coords and then drawn to
   the canvas.

   Return 1 on success, 0 if there was an error.
*/
int write_contour_boundingbox(CCP4SECTION *section)
{
  int         npts=5;
  float       x[5],y[5];
  char        *canvas,units;
  CCP4MAP     *map;

  if (CCP4MAPWISH_CONTOUR_DIAG) puts("write_contour_boundingbox: starting");

  /* Retrieve a pointer to the name of the current canvas */
  canvas = fetch_contour_canvas();
  if (!canvas) {
    puts("write_contour_boundingbox: couldn't retrieve the canvas name");
    return 0;
  }
  if (CCP4MAPWISH_CONTOUR_DIAG) 
    printf("write_contour_boundingbox: retrieved canvas name \"%s\"\n",canvas);

  /* Retrieve the current map name */
  map = fetch_map_ptr(section->mapid);
  if (!map) {
    puts("write_contour_boundingbox: couldn't retrieve pointer to the map");
    return 0;
  }

  /* Retrieve the units for drawing to the canvas */
  units = fetch_contour_units();

  /* Generate the corner points of the bounding box */
  x[0] = 0.0;
  y[0] = 0.0;
  x[1] = 0.0;
  y[1] = (float) section->vlimits[1] - section->vlimits[0];
  x[2] = (float) section->ulimits[1] - section->ulimits[0];
  y[2] = (float) section->vlimits[1] - section->vlimits[0];
  x[3] = (float) section->ulimits[1] - section->ulimits[0];
  y[3] = 0.0;
  x[4] = x[0];
  y[4] = y[0];

  if (CCP4MAPWISH_CONTOUR_DIAG) {
    printf("write_contour_boundingbox: limits are\n>>> %f %f\n>>> %f %f\n>>> %f %f\n>>> %f %f\n>>> %f %f\n",x[0],y[0],x[1],y[1],x[2],y[2],x[3],y[3],x[4],y[4]);
  }

  /* Transform the coordinates from grid to orthogonal coordinates */
  if (CCP4MAPWISH_CONTOUR_DIAG) puts("write_contour_boundingbox: calling transform_coordinates");
  if (!transform_coordinates(section,npts,x,y)) {
    puts("write_contour_boundingbox: transform_coordinates returned an error");
    return 0;
  }
  if (CCP4MAPWISH_CONTOUR_DIAG) {
    puts("write_contour_boundingbox: done transform_coordinates");
    printf("write_contour_boundingbox: limits are\n>>> %f %f\n>>> %f %f\n>>> %f %f\n>>> %f %f\n>>> %f %f\n",x[0],y[0],x[1],y[1],x[2],y[2],x[3],y[3],x[4],y[4]);
  }
    
  /* Draw the line on the canvas
     Bounding box is tagged with "bbox" */
  if (CCP4MAPWISH_CONTOUR_DIAG) puts("write_contour_boundingbox: calling draw_line");

  if (!canvas_draw_line(canvas,npts,x,y,"bbox",NULL,units)) {
    puts("write_contour_boundingbox: draw_line returned an error");
    return 0;
  }

  if (CCP4MAPWISH_CONTOUR_DIAG)
    puts("write_contour_boundingbox: done draw_line - returning");
  return 1;
}

/*------------------------------------------------------------------------------*/

/* write_contour_gridlines

   Function to draw the gridlines for the current section, given the
   limits of the section and the spacing (in fractional coordinates)
   between gridlines on the u and v axes.
*/
int write_contour_gridlines(CCP4SECTION *section)
{
  int   i,npts=2,gridlabels,gridlines;
  float u0,v0,u_start,u_stop,v_start,v_stop,ustep,vstep,x[2],y[2],uspacing,vspacing;
  char  *canvas,units;
  CCP4MAP *map;

  if (CCP4MAPWISH_CONTOUR_DIAG) puts("write_contour_gridlines: starting");

  /* Retrieve the name of the current canvas structure */
  canvas = fetch_contour_canvas();
  if (!canvas) {
    puts("write_contour_gridlines: couldn't retrieve the canvas name");
    return 0;
  }
  if (CCP4MAPWISH_CONTOUR_DIAG) {
    printf("write_contour_gridlines: fetched canvas name \"%s\"\n",canvas);
  }

  /* Retrieve pointer to the map */
  map = fetch_map_ptr(section->mapid);
  if (!map) {
    puts("write_contour_gridlines: couldn't retrieve the map pointer");
    return 0;
  }

  /* Retrieve the units for drawing to the canvas */
  units = fetch_contour_units();

  /* Get floating limits */
  u_start = (float) 0.0;
  v_start = (float) 0.0;
  u_stop  = (float) section->ulimits[1] - section->ulimits[0];
  v_stop  = (float) section->vlimits[1] - section->vlimits[0];

  /* Get spacings */
  uspacing = section->uspace;
  vspacing = section->vspace;

  /* Flags for which elements to draw */
  gridlabels = section->labels;
  gridlines  = section->grid;

  if (CCP4MAPWISH_CONTOUR_DIAG) {
    printf("write_contour_gridlines: u_start,v_start = %f,%f u_stop,v_stop = %f %f\n",
	   u_start,v_start,u_stop,v_stop);
  }

  /* Generate spacings in grid units */
  ustep = uspacing*((float) map->mxyz[section->iu]);
  vstep = vspacing*((float) map->mxyz[section->iv]);
  if (CCP4MAPWISH_CONTOUR_DIAG)
    printf("write_contour_gridlines: ustep = %f, vstep = %f\n",ustep,vstep);

  /* FIXME  the following looks quite inefficient and should
     probably be rewritten later? */

  /* Generate and draw vertical gridlines */
  if (CCP4MAPWISH_CONTOUR_DIAG) puts("write_contour_gridlines: starting vertical lines");

  i = 0;
  u0 = u_start;
  while (u0<=u_stop) {
    /* Generate the coordinates defining the vertical lines*/
    x[0] = u0;
    x[1] = u0;
    y[0] = v_start;
    y[1] = v_stop;
    if (CCP4MAPWISH_CONTOUR_DIAG)
      printf("write_contour_gridlines: vertical line #%d\n",i);
    /* Write the grid label */
    if (gridlabels) {
      if (CCP4MAPWISH_CONTOUR_DIAG) puts("write_contour_gridlines: writing gridlabel");
      if (!write_contour_gridlabel(section,canvas,x,y)) {
	puts("write_contour_gridlines: write_contour_gridlabel returned an error");
	return 0;
      }
    }

    if (CCP4MAPWISH_CONTOUR_DIAG)
      puts("write_contour_gridlines: done label... transforming coords ...");
    
    /* Prepare and draw the gridline */
    if (gridlines) {
      if (!transform_coordinates(section,npts,x,y)) {
	puts("write_contour_gridlines: transform_coordinates returned an error");
	return 0;
      }
      if (CCP4MAPWISH_CONTOUR_DIAG)
	puts("write_contour_gridlines: done coord transform ... drawing line ...");
      
      /* Grid lines are tagged with "grid" */
      if (!canvas_draw_line(canvas,npts,x,y,"grid",NULL,units)) {
	puts("write_contour_gridlines: canvas_draw_line returned an error");
	return 0;
      } 
    }
    if (CCP4MAPWISH_CONTOUR_DIAG)
      puts("write_contour_gridlines: done drawing line");
    /* Next gridline */
    u0 = u0 + ustep;
    i++;
  }

  /* Generate and draw horizontal gridlines */
  if (CCP4MAPWISH_CONTOUR_DIAG)
    puts("write_contour_gridlines: starting horizontal lines");

  i = 0;
  v0 = v_start;
  while (v0<=v_stop) {
    /* Generate the coordinates defining the vertical lines*/
    x[0] = u_start;
    x[1] = u_stop;
    y[0] = v0;
    y[1] = v0;
    if (CCP4MAPWISH_CONTOUR_DIAG)
      printf("write_contour_gridlines: horizontal line #%d\n",i);
    /* Write the grid label */
    if (gridlabels) {
      if (!write_contour_gridlabel(section,canvas,x,y)) {
	puts("write_contour_gridlines: write_contour_gridlabel returned an error");
	return 0;
      }
    }

    if (CCP4MAPWISH_CONTOUR_DIAG)
      puts("write_contour_gridlines: done label... transforming coords ...");

    /* Prepare and draw the gridline */
    if (gridlines) {
      if (!transform_coordinates(section,npts,x,y)) {
	puts("write_contour_gridlines: transform_coordinates returned an error");
	return 0;
      }
      /* Grid lines are tagged with "grid" */
      if (!canvas_draw_line(canvas,npts,x,y,"grid",NULL,units)) {
	puts("write_contour_gridlines: draw_line returned an error");
	return 0;
      }
    }
    if (CCP4MAPWISH_CONTOUR_DIAG)
      puts("write_contour_gridlines: done drawing line");
    /* Next gridline */
    v0 = v0 + vstep;
    i++;
  }

  /* Completed ok */
  return 1;
}

/*------------------------------------------------------------------------------*/

/* write_contour_gridlabel

   This generates text to label a gridline, and calls draw_text to
   write it to the specified position.

   The start and stop points of the gridlines (x and y) are supplied
   in grid units.

   Return 1 on success, 0 if there was an error.
*/
int write_contour_gridlabel(CCP4SECTION *section, char *canvas, float *x, float *y)
{
  char  label[20],anchor[3],units;
  int   i,iaxis,npts=2;
  int   iend,point;
  int   ihorizontal=0,ivertical=0,iorient;
  float xtmp[2],ytmp[2],xlabel,ylabel,nlabel;
  float ustart,vstart;
  CCP4MAP *map;

  if (CCP4MAPWISH_CONTOUR_DIAG) puts("write_contour_gridlabel: starting");

  /* Fetch a pointer to the map */
  map = fetch_map_ptr(section->mapid);
  if (!map) {
    puts("write_contour_gridlabel: couldn't retrive map pointer");
    return 0;
  }

  /* Retrieve the units for drawing to the canvas */
  units = fetch_contour_units();

  /* Fetch the section limits in grid units */
  ustart = (float) section->ulimits[0];
  vstart = (float) section->vlimits[0];

  /* Make a temporary store for the coordinates */
  for (i=0; i<2; i++) {
    xtmp[i] = x[i];
    ytmp[i] = y[i];
  }
  if (CCP4MAPWISH_CONTOUR_DIAG)
    printf("write_contour_gridlabel: xtmp = %f %f, ytmp = %f %f\n",
	   xtmp[0],xtmp[1],ytmp[0],ytmp[1]);

  /* Transform the temporary coordinates */
  if (CCP4MAPWISH_CONTOUR_DIAG)
    puts("write_contour_gridlabel: calling transform coordinates...");
  if (!transform_coordinates(section,npts,xtmp,ytmp)) {
    puts("write_contour_gridlabel: transform_coordinates returned an error");
    return 0;
  }
  
  /* Fetch the view orientation
     iorient == 0 for lhs view
                1 for rhs view
  */
  iorient = strcmp(section->view_orient,"lhs");

  /* Determine gridline orientation (horizontal or vertical)
     and set parameters accordingly
     NB: xlabel and ylabel are the coordinates to anchor the
     gridlabel to. */
  if (x[0] == x[1]) {
    /* Pseudo-horizontal orientation */
    iaxis  = section->iu;
    ylabel = ytmp[0];
    if (iorient) {
      /* Absolute orientation is vertical */
      ivertical   = 1;
      ihorizontal = 0;
      xlabel = xtmp[0];
    } else {
      /* Absolute orientation is horizontal */
      ivertical   = 0;
      ihorizontal = 1;
      xlabel = xtmp[1];
    }
    /* Fractional coordinates */
    nlabel = (x[0]+ustart)/((float) map->mxyz[iaxis]);
  } else if (y[0] == y[1]) {
    /* Pseudo-vertical orientation */
    iaxis  = section->iv;
    xlabel = xtmp[0];
    if (iorient) {
      /* Absolute orientation is horizontal */
      ivertical   = 0;
      ihorizontal = 1;
      ylabel = ytmp[0];
    } else {
      /* Absolute orientation is vertical */
      ivertical   = 1;
      ihorizontal = 0;
      ylabel = ytmp[1];
    }
    /* Fractional coordinates */
    nlabel = (y[0]+vstart)/((float) map->mxyz[iaxis]);
  }
  /* Check for errors */
  if (!ihorizontal && !ivertical) {
    /* We should never reach this point */
    printf("write_contour_gridlabel: serious error\n");
    return 0;
  } 

  /* Anchoring for gridline labels depends on the 
     absolute orientation */
  if (ihorizontal) {
    /* Horizontal gridline */
    strcpy(anchor,"s");
  } else if (ivertical) {
    /* Vertical gridline */
    strcpy(anchor,"e");
  }

  /* Generate label text */
  sprintf(label,"%f",nlabel);

  if (CCP4MAPWISH_CONTOUR_DIAG)
    printf("write_contour_gridlabel: nlabel = %f, label = \"%s\"\n",nlabel,label);

  /* Truncate to required precision (i.e. no of decimal places)
     if necessary */
  if (section->precision > 0) {
    i = 0;
    iend  = 0;
    point = 0;
    while (i<strlen(label) && !iend) {
      if (point || label[i] == '.') {
	point++;
	if (point > section->precision) {
	  label[i+1] = '\0';
	  iend = 1;
	}
      }
      i++;
    }
  }

  /* Write the label to the canvas */
  if (!canvas_draw_label(canvas,label,xlabel,ylabel,anchor,"gridlabel",units)) {
    puts("write_contour_gridlabel: draw_label returned an error");
    return 0;
  }
  return 1;
}

/*------------------------------------------------------------------------------*/

/* write_contour_axes

   Draw and label the axes for the current section

   Axes consist of two lines with arrowheads parallel to the bounding box,
   plus appropriate labels.

   "scale" is the scalefactor used on the canvas

   Returns 1 on success, 0 if there was an error.
*/
int write_contour_axes(CCP4SECTION *section, char *canvas, float scale)
{
  float axis_offset_x,axis_offset_y;
  float x[2],y[2];

  /* Hardcoded offsets (in pixels) */
  axis_offset_x = -50;
  axis_offset_y = -30;

  /* This defines the "vertical" (v) axis
     The axis length is hardcoded to be 0.25 x the length of
     the bounding box edge
  */
  x[0] = 0.0;
  y[0] = 0.0;
  x[1] = 0.0;
  y[1] = ((float) section->vlimits[1] - section->vlimits[0])/((float) 4.0);

  if (!write_contour_axis(section,canvas,x,y,section->iv,scale)) {
    puts("write_contour_axes: returned error from write_contour_axis (v axis)");
    return 0;
  }

  /* This defines the "horizontal" (u) axis
     The axis length is hardcoded to be 0.25 x the length of
     the bounding box edge
  */
  x[0] = 0.0;
  y[0] = 0.0;
  x[1] = ((float) section->ulimits[1] - section->ulimits[0])/((float) 4.0);
  y[1] = 0.0;

  if (!write_contour_axis(section,canvas,x,y,section->iu,scale)) {
    puts("write_contour_axes: returned error from write_contour_axis (u axis)");
    return 0;
  }

  /* Shift the axes by the specified offsets */
  if (!canvas_shift_element(canvas,"axes",axis_offset_x,axis_offset_y)) {
    puts("write_contour_axes: returned error from canvas_shift_element");
    return 0;
  }

  return 1;
}

/*------------------------------------------------------------------------------*/

/* write_contour_axis

   Draw a single axis on the specified canvas - called from write_contour_axes

   Returns 1 on success, 0 if there was an error.
*/
int write_contour_axis(CCP4SECTION *section, char *canvas, float *x, float*y,
		       int iaxis, float scale)
{
  char label[LEN_AXISLABEL],anchor[3],units;
  int  i,npts=2;
  int  ihorizontal,ivertical,iorient;
  
  /* Retrieve axis name for display */
  sprintf(label,"%s",section->axislabel[iaxis]);

  if (CCP4MAPWISH_CONTOUR_DIAG)
    printf("write_contour_axis: axis name (no. %d) is \"%s\"\n",iaxis,label);

  /* Retrieve the units for drawing to the canvas */
  units = fetch_contour_units();

  /* Fetch the view orientation
     iorient == 0 for lhs view
                1 for rhs view
  */
  iorient = strcmp(section->view_orient,"lhs");

  /* Determine axes line orientation (horizontal or vertical)
     and set parameters accordingly */
  if (x[0] == x[1]) {
    /* Pseudo-horizontal orientation */
    if (iorient) {
      /* Absolute orientation is vertical */
      ivertical   = 1;
      ihorizontal = 0;
    } else {
      /* Absolute orientation is horizontal */
      ivertical   = 0;
      ihorizontal = 1;
    }
  } else if (y[0] == y[1]) {
    /* Pseudo-vertical orientation */
    if (iorient) {
      /* Absolute orientation is horizontal */
      ivertical   = 0;
      ihorizontal = 1;
    } else {
      /* Absolute orientation is vertical */
      ivertical   = 1;
      ihorizontal = 0;
    }
  }

  /* Determine the anchor */
  if (ihorizontal) {
    sprintf(anchor,"%s","n");
  } else if (ivertical) {
    sprintf(anchor,"%s","w");
  } else {
    sprintf(anchor,"%s","");
  }

  /* Transform the coordinates */
  if (!transform_coordinates(section,npts,x,y)) {
    puts("write_contour_axis: transform_coordinates returned an error");
    return 0;
  }

  /* Apply the scale factor*/
  for (i=0; i<2; i++) {
    x[i] = x[i] * scale;
    y[i] = y[i] * scale;
  }

  /* Draw the axis line on the canvas
     NB Axes have an arrow at the end of the line and are tagged "axes" */
  if (!canvas_draw_line(canvas,npts,x,y,"axes","last",units)) {
    puts("write_contour_axis: canvas_draw_line returned an error");
    return 0;
  }

  /* Add the axis label at the end of the line */
  if (!canvas_draw_label(canvas,label,x[1],y[1],anchor,"axes",units)) {
    puts("write_contour_axis: canvas_draw_label for v axis label returned an error");
    return 0;
  }
  return 1;
}

/*------------------------------------------------------------------------------*/

/* write_contour_titles

   Draw the titles for the section on the canvas
   These are: the map title, the current axis label, the current section
   number(s) (and fractional equivalents), and the scaling.

   Returns 1 on success, 0 on error.
*/
int write_contour_titles(CCP4SECTION *section, char *canvas, float scale)
{
  char    tmpstr[CCP4_LTMPSTR],unit_label[7];
  int     limits[4];
  int     lhside,rhside,centre,bottom,units;
  float   frac_sec1,frac_sec2;
  CCP4MAP *map;

  /* Fetch a pointer to the current map */
  map = fetch_map_ptr(section->mapid);
  if (!map) {
    puts("write_contour_titles: couldn't fetch a valid map pointer");
    return 0;
  }

  /* Fetch the limits for the current canvas */
  if (!canvas_get_limits(canvas,limits)) {
    puts("write_contour_titles: couldn't get the canvas limits");
    return 0;
  }

  /* Set up various positions for the different titles */
  lhside = limits[0];
  centre = (limits[2] - limits[0])/2 + limits[0];
  rhside = limits[2];
  bottom = limits[3] + 20;
  /* Note that:
     top    = limits[1] - 20;
     but this value is not required in practice and is not explicitly coded
  */

  /* Write the map title at the bottom-centre of the canvas */
  if (!canvas_draw_label(canvas,map->title,centre,(bottom+40),"center","titles",' ')) {
    puts("write_contour_titles: canvas_draw_label returned an error for main title");
    return 0;
  }

  /* Generate the section or slab title */
  if (section->slab_start == section->slab_end) {
    /* Single section */
    frac_sec1 = ((float) section->nsec)/((float) map->mxyz[section->iaxis]);
    sprintf(tmpstr,"Section %4d (% 8.3f) on %c axis",
	    section->nsec,frac_sec1,section->axis);
  } else {
    /* Slab of sections */
    frac_sec1 = ((float) section->slab_start)/((float) map->mxyz[section->iaxis]);
    frac_sec2 = ((float) section->slab_end)/((float) map->mxyz[section->iaxis]);
    sprintf(tmpstr,"Sections %4d-%4d (% 8.3f-% 8.3f) on %c axis",
	    section->slab_start,section->slab_end,frac_sec1,frac_sec2,
	    section->axis);
  }
  
  /* Write the section title at the bottom-rhside of the canvas */
  if (!canvas_draw_label(canvas,tmpstr,rhside,(bottom+20),"e","titles",' ')) {
    puts("write_contour_titles: canvas_draw_label returned an error for section title");
    return 0;
  }

  /* Generate the scale title */
  units = fetch_contour_units();
  switch (units) {
  case 'm':
    sprintf(unit_label,"%s","mm");
    break;
  case 'c':
    sprintf(unit_label,"%s","cm");
    break;
  case 'i':
    sprintf(unit_label,"%s","in");
    break;
  default:
    sprintf(unit_label,"%s","pixels");
    break;
  }
  sprintf(tmpstr,"Scale = %f %s/A",scale,unit_label);

  /* Write the scale title at the bottom-lhside of the canvas */
  if (!canvas_draw_label(canvas,tmpstr,lhside,bottom,"w","titles",' ')) {
    puts("write_contour_titles: canvas_draw_label returned an error for scale title");
    return 0;
  }

  return 1;
}

/*------------------------------------------------------------------------------*/
