/*
     ccp4mapwish_map.c: functions for handling CCP4 map files
     Copyright (C) 2001  CCLRC, Peter Briggs

     This code is distributed under the terms and conditions of the
     CCP4 Program Suite Licence Agreement as a CCP4 Application.
     A copy of the CCP4 licence can be obtained by writing to the
     CCP4 Secretary, Daresbury Laboratory, Warrington WA4 4AD, UK.
*/

/*******************************************************************/

/* ccp4mapwish_map.c

   Functions for handling maps stored in memory

   CVS id string: $Id: ccp4mapwish_map.c,v 1.9 2005/09/08 13:53:26 mdw Exp $

   CHANGES:
   Sept 2002 - Switched from using Fortran wrappers to CMap functions
   to read in map information from file.
*/

/*******************************************************************/

/* Include header file */

/*******************************************************************/

#include "ccp4mapwish.h"

/*******************************************************************/

/* Private local variables */

/*******************************************************************/

static CCP4MAPARRAY ccp4mapwish_maparray;

/*******************************************************************/

/* Tcl MAP functions */

/*******************************************************************/

/* ReadMap  --
   Tcl command to load a map into memory from the specified file

   Usage:
   readmap <mapid> <mapfile>

   Arguments:
   mapid    "handle" string identfying the map in memory
   mapfile  file to read from

   0  = completed ok
   -1 = map already loaded
   -2 = error reading from map file
   -3 = couldn't allocate memory
*/

int ReadMap(ClientData clientData, Tcl_Interp *interp,
		   int objc, Tcl_Obj *CONST objv[])
{
  /* Declarations */

  /* For reading map */
  char  *filename,*mapid;
  int   i,j,k,lenname,lenid,nsize,mode;
  uint8 *buffer=NULL;

  /* From the mapfile */
  CMMFile *mapfil=NULL;
  char  *title=NULL,*tmp_label=NULL;
  int   nsec,lspgrp,lentitle=0;
  int   iuvw[3],mxyz[3];
  float rhmin,rhmax;
  double rhmean,rhrms;
  float cell[6];

  /* Derived quantities */
  int   fast,medium,slow,ext[3],tmpext[3],start[3],tmpstart[3],iorth=1;
  float *data,orth_to_frac[3][3],frac_to_orth[3][3];

  /* Handling string result of Tcl command */
  Tcl_Obj *resultPtr;

  /* Pointer to map structure */
  CCP4MAP     *map;

  /* Check number of args - should be 3 */
  if (objc!=3) {
    Tcl_WrongNumArgs(interp, 1, objv,"mapid mapfile");
    return TCL_ERROR;
  }

  /* Set the result pointer */
  resultPtr = Tcl_GetObjResult(interp);

  /* Get map id */
  mapid = Tcl_GetStringFromObj(objv[1],&lenid);
  if (CCP4MAPWISH_MAP_DIAG) printf("ReadMap: mapid = \"%s\"\n",mapid);

  /* Get map name */
  filename = Tcl_GetStringFromObj(objv[2],&lenname);
  if (CCP4MAPWISH_MAP_DIAG) printf("ReadMap: filename = \"%s\"\n",filename);

  /* Check whether a map with this handle already exists */
  map = fetch_map_ptr(mapid);
  if (map) {
    puts("ReadMap: Map handle already in use");
    return TCL_ERROR;
  }

  /* Open map file for read only */
  mapfil = ccp4_cmap_open(filename,0);
  if (!mapfil) {
    puts("ReadMap: failed to open map file for read");
    Tcl_SetIntObj(resultPtr,-2);
    return TCL_ERROR;
  }

  /* Fetch the mode (= datatype) stored in the map */
  mode = ccp4_cmap_get_datamode(mapfil);
  if (CCP4MAPWISH_MAP_DIAG) {
      printf("ReadMap: map mode = %d\n",mode);
  }

  /* Set up a map pointer for the new map */
  map = new_map_ptr(mapid);
  if (!map) {
    puts("ReadMap: couldn't get a pointer to a new map");
    ccp4_cmap_close(mapfil);
    return TCL_ERROR;
  }

  /* Retrieve the various bits of header information from the file */
  if (CCP4MAPWISH_MAP_DIAG) puts("ReadMap: retrieving map information");
  ccp4_cmap_get_grid(mapfil,mxyz);
  if (CCP4MAPWISH_MAP_DIAG) {
    printf("ReadMap: retrieved grid:");
    printf(" %d %d %d\n",mxyz[0],mxyz[1],mxyz[2]);
  }
  ccp4_cmap_get_order(mapfil,iuvw);
  if (CCP4MAPWISH_MAP_DIAG) {
    printf("ReadMap: retrieved ordering:");
    printf(" %d %d %d\n",iuvw[0],iuvw[1],iuvw[2]);
  }
  ccp4_cmap_get_mapstats(mapfil,&rhmin,&rhmax,&rhmean,&rhrms);
  if (CCP4MAPWISH_MAP_DIAG) {
    printf("ReadMap: retrieved statistics:");
    printf(" min:%f max:%f mean:%f rms:%f\n",rhmin,rhmax,rhmean,rhrms);
  }
  /* Starting grid position along each axis
     The numbers are in map order not xyz, which is what we need */
  ccp4_cmap_get_origin(mapfil,tmpstart);
  /* Reorder */
  for (i=0;i<3;++i) {
    start[iuvw[i]-1] = tmpstart[i];
  } 
  if (CCP4MAPWISH_MAP_DIAG) {
    printf("ReadMap: retrieved origin:");
    printf(" %d %d %d\n",start[0],start[1],start[2]);
  }
  /* Extent i.e. number of grid points along each axis
     The numbers are in map order not xyz, which is what we need */
  ccp4_cmap_get_dim(mapfil,tmpext);
  /* Reorder */
  for (i=0;i<3;++i) {
    ext[iuvw[i]-1] = tmpext[i];
  }

  if (CCP4MAPWISH_MAP_DIAG) {
    printf("ReadMap: retrieved extents:\n");
    printf(" Grid direct from file: %d %d %d\n",tmpext[0],tmpext[1],tmpext[2]);
    printf(" Reordered:             %d %d %d\n",ext[0],ext[1],ext[2]);
  }
  /* Spacegroup */
  lspgrp = ccp4_cmap_get_spacegroup(mapfil);
  if (CCP4MAPWISH_MAP_DIAG) {
    printf("ReadMap: retrieved spacegroup:");
    printf(" %d\n",lspgrp);
  }
  /* Cell */
  ccp4_cmap_get_cell(mapfil,cell);
  if (CCP4MAPWISH_MAP_DIAG) {
    printf("ReadMap: retrieved cell:");
    printf(" %f %f %f (%f %f %f)\n",cell[0],cell[1],cell[2],
	   cell[3],cell[4],cell[5]);
  }

  /* Title is a special case - adapted code from cmaplib_f.c
     It might be better if there was a specialised ccp4_cmap_get_title
     9/9/2002: There is now a ccp4_cmap_get_title ...*/
  if (ccp4_cmap_number_label(mapfil)) {
    tmp_label = ccp4_cmap_get_label(mapfil,0);
    if (tmp_label != NULL) {
      lentitle = strlen(tmp_label);
      title = (char *) malloc((lentitle+1)*sizeof(char));
      strcpy(title,tmp_label);
      title[lentitle] = '\0';
    }
  }

  if (CCP4MAPWISH_MAP_DIAG)
    printf("ReadMap: title from map is \"%s\"\n",title);

  /* Decode information about axis ordering etc
     iuvw holds fast, medium, slow axes (1=X, 2=Y, 3=Z)
     mxyz hold sampling intervals along whole cell on X, Y, Z
   */
  fast   = iuvw[0] - 1;
  medium = iuvw[1] - 1;
  slow   = iuvw[2] - 1;

  /* Number of sections in map */
  nsec = ext[slow];

  /* Fetch orthogonalising and fractionalising matrices */
  if (!get_scale_matrix(cell,iorth,frac_to_orth,orth_to_frac)) {
    printf("ReadMap: failed to get fractionalising/orthogonalising matrices\n");
    return TCL_ERROR;
  }
  if (CCP4MAPWISH_MAP_DIAG) printf("ReadMap: fractionalising matrix:\n");
  for (i=0; i<3; i++) {
    if (CCP4MAPWISH_MAP_DIAG) printf("         ");
    for (j=0; j<3; j++) {
      map->orth_to_frac[i][j] = orth_to_frac[i][j];
      if (CCP4MAPWISH_MAP_DIAG) printf("%f ",map->orth_to_frac[i][j]);
    }
    if (CCP4MAPWISH_MAP_DIAG) printf("\n");
  }
  if (CCP4MAPWISH_MAP_DIAG) printf("ReadMap: orthogonalising matrix:\n");
  for (i=0; i<3; i++) {
    if (CCP4MAPWISH_MAP_DIAG) printf("         ");
    for (j=0; j<3; j++) {
      map->frac_to_orth[i][j] = frac_to_orth[i][j];
      if (CCP4MAPWISH_MAP_DIAG) printf("%f ",map->frac_to_orth[i][j]);
    }
    if (CCP4MAPWISH_MAP_DIAG) printf("\n");
  }

  /* Allocate the memory for the map
     This chunk of code for the allocation stolen from solomon */
  map->map = (float ***) malloc((ext[0] * sizeof(float**)) +
                                (ext[0] * ext[1] * sizeof(float*)) +
				(ext[0] * ext[1] * ext[2] * sizeof(float)));
  if (!map->map) {
    Tcl_SetIntObj(resultPtr, -3);
    return TCL_OK;
  }
  data = (float *) (map->map + ext[0] * (1 + ext[1]));
  for (i = 0; i < ext[0]; ++i) {
    map->map[i] =  (float **) map->map + ext[0] + i * ext[1];
    for (j = 0; j < ext[1]; ++j, data += ext[2]) {
      map->map[i][j] = data;
      for (k = 0; k < ext[2]; ++k)
	data[k] = 0;
    }
  }

  /* Store the title */
  if (CCP4MAPWISH_MAP_DIAG) printf("ReadMap: length of title string is %d\n",lentitle);
  map->title = (char *) malloc((lentitle+1)*sizeof(char));
  if (!map->title) {
    Tcl_SetIntObj(resultPtr, -3);
    return TCL_OK;
  }
  strncpy(map->title,title,lentitle);
  map->title[lentitle] = '\0';
  if (CCP4MAPWISH_MAP_DIAG) printf("ReadMap: stored title \"%s\"\n",map->title);

  /* Store the header information in the map */
  map->filename = (char *) malloc((lenname+1) * sizeof(char));
  strcpy(map->filename,filename);
  for (i=0;i<6;i++) map->cell[i] = cell[i];
  map->rhmin  = rhmin;
  map->rhmax  = rhmax;
  map->rhmean = (float) rhmean;
  map->rhrms  = (float) rhrms;
  map->lspgrp = lspgrp;
  if (CCP4MAPWISH_MAP_DIAG) printf("ReadMap: stored spacegroup is \"%d\"\n",map->lspgrp);

  /* Store sampling intervals, extent and limits */
  for (i=0;i<3;i++) {
    map->mxyz[i]  = mxyz[i];
    map->ext[i]   = ext[i];
    map->start[i] = start[i];
  }
  if (CCP4MAPWISH_MAP_DIAG) printf("ReadMap: stored sampling intervals\n");

  /* Store the map mode */
  map->mode = mode;

  /* Get the density in the map and store in memory in x-y-z order
     (This was originally taken from solomon)
     Reading the map can be done in quite a number of ways 
     because of axis permutations */
  if (CCP4MAPWISH_MAP_DIAG) printf("ReadMap: fast = %d, medium = %d\n",fast,medium);
  nsize = ext[fast] * ext[medium];
  if (CCP4MAPWISH_MAP_DIAG) printf("ReadMap: nsize = %d x %d = %d\n",
				   ext[fast],ext[medium],nsize);

  /* Data is a 1d array of floats into which we read each of the
     sections in turn
  */
  data = (float *) malloc(nsize * sizeof(float));
  if (CCP4MAPWISH_MAP_DIAG) {
    if (data)
      printf("ReadMap: data is set ...\n");
    else
      printf("ReadMap: data is not set ...\n");
  }
  /* Don't need to worry about section numbering not starting at
     zero when using ccp4_cmap_seek_section
     Cmap numbers always start at zero (regardless of those assigned
     in the file)
  */
  if (CCP4MAPWISH_MAP_DIAG) printf("ReadMap: section");
  for (i = 0; i < nsec; ++i) {
    if (CCP4MAPWISH_MAP_DIAG) printf(" %d",i);
    if (ccp4_cmap_seek_section(mapfil,i,SEEK_SET) == -1) {
      puts("ReadMap: failed to seek section");
      Tcl_SetIntObj(resultPtr, -2);
      return TCL_OK;
    }
    /* The file can contain data in an arbitrary mode (= datatype)
       To deal with this, each section is read into a standard buffer
       and then translated into float afterwards.
       This is shortcut for FLOAT32 maps since these are most likely
       what we will be dealing with anyway.
       This code is adapted from that in MGULPR (cmaplib_f.c)
    */
    if (mode == FLOAT32)
	buffer = (uint8 *) data;
    else 
	buffer = (uint8 *) calloc(nsize,ccp4_file_itemsize((*mapfil).stream));
    
    if (ccp4_cmap_read_section(mapfil,buffer) == 0) {
	/* Report an error reading from the map file */
	puts("ReadMap: error reading map sections");
	Tcl_SetIntObj(resultPtr,-2);
	return TCL_ERROR;
    } else {
	ccp4_utils_translate_mode_float(data, buffer,nsize,mode);
    }
    /* Data is a 1d array of floats, n_medium * n_fast */
    /* Build the map in memory in x,y,z order */
    if ((fast == 0) && (medium == 1) && (slow == 2))
      for (j = 0; j < ext[medium]; ++j)
	for (k = 0; k < ext[fast]; ++k)
	  map->map[k][j][i] = data[k+j*ext[fast]];
    else if ((fast == 0) && (medium == 2) && (slow == 1))
      for (j = 0; j < ext[medium]; ++j)
	for (k = 0; k < ext[fast]; ++k)
	  map->map[k][i][j] = data[k+j*ext[fast]];
    else if ((fast == 1) && (medium == 0) && (slow == 2))
      for (j = 0; j < ext[medium]; ++j)
	for (k = 0; k < ext[fast]; ++k)
	  map->map[j][k][i] = data[k+j*ext[fast]];
    else if ((fast == 1) && (medium == 2) && (slow == 0))
      for (j = 0; j < ext[medium]; ++j)
	for (k = 0; k < ext[fast]; ++k)
	  map->map[i][k][j] = data[k+j*ext[fast]];
    else if ((fast == 2) && (medium == 1) && (slow == 0))
      for (j = 0; j < ext[medium]; ++j)
	for (k = 0; k < ext[fast]; ++k)
	  map->map[i][j][k] = data[k+j*ext[fast]];
    else if ((fast == 2) && (medium == 0) && (slow == 1))
      for (j = 0; j < ext[medium]; ++j)
	for (k = 0; k < ext[fast]; ++k)
	  map->map[j][i][k] = data[k+j*ext[fast]];
  }
  if (CCP4MAPWISH_MAP_DIAG) printf("\n");
  /* Close the map */
  if (mapfil) ccp4_cmap_close(mapfil);

  /* Free temporary memory */
  if (data) free(data);
  if (mode != FLOAT32 && buffer != NULL) free(buffer);

  /* Finished */
  Tcl_SetIntObj(resultPtr, 0);
  return TCL_OK;
}

/*----------------------------------------------------------------*/
      
/* MapInfo  --
   Tcl command to interrogate the map held in memory.

   Usage:
   mapinfo ?mapid ?keyword??

   mapid is the handle identifying the map stored in memory

   If no mapid is given then all the currently stored mapids are
   returned as a list

   A maximum of one keyword is allowed.

   Keywords:
   [none]     = returns 1 if the map is currently held in memory
   spacegroup = returns spacegroup number
   maximum    = returns maximum density in map
   minimum    = returns minimum density
   mean       = returns mean density
   rms        = returns rms density
   cell       = returns cell info as a list of 6 numbers
   limits     = return start...stop on x..y..z as a list of 6 numbers
   map        = writes out map...careful!
   title      = returns the title from the map header
   grid       = returns the number of grid points along x,y,z cell
                edges as list of 3 numbers
   type       = returns the type (map or mask)
   mode       = returns the orginal map mode
   
   Returns:
   -1 = no map loaded
   Otherwise returns requested info, see keywords.

*/

int MapInfo(ClientData clientData, Tcl_Interp *interp,
	    int objc, Tcl_Obj *CONST objv[])
{
  /* Declarations */
  Tcl_Obj *resultPtr;
  CCP4MAP     *map;
  char *mapid,*keyword,*resultstr=NULL,tmpstr[100];
  int  i,lenname,lenresult;

  /* Check number of args - should 3 or less */
  if (objc > 3) {
    Tcl_WrongNumArgs(interp, 1, objv,"?mapid ?keyword??");
    return TCL_ERROR;
  }

  /* Set pointer for result */
  resultPtr = Tcl_GetObjResult(interp);

  /* If no arguments to command then list all the mapids
     in memory
     Not sure how to do this at the moment */
  if (objc == 1) {
    printf("MapInfo: list of mapids in memory not implemented, sorry\n");
    Tcl_SetIntObj(resultPtr, -1);
    return TCL_OK;
  }

  /* Otherwise: information about a particular map
     Get the mapid */
  mapid = Tcl_GetStringFromObj(objv[1], &lenname);
  if (CCP4MAPWISH_MAP_DIAG) printf("MapInfo: mapid is \"%s\"\n",mapid);

  /* Retrieve map pointer */
  map = fetch_map_ptr(mapid);

  /* Check - does the map exist? */
  if (!map) {
    Tcl_SetIntObj(resultPtr, -1);
    return TCL_OK;
  }

  /* No keywords - just return 1 */
  if (objc == 2) {
    Tcl_SetIntObj(resultPtr, 1);
    return TCL_OK;
  }

  /* Get keyword */
  keyword = Tcl_GetStringFromObj(objv[2], &lenname);
  if (CCP4MAPWISH_MAP_DIAG) printf("MapInfo: keyword is \"%s\"\n",keyword);

  /* Process keywords */

  /* Spacegroup */
  if (strcmp(keyword,"spacegroup")==0) {
    Tcl_SetIntObj(resultPtr,map->lspgrp);
  }

  /* Maximum density */
  if (strcmp(keyword,"maximum")==0) {
    Tcl_SetDoubleObj(resultPtr,map->rhmax);
  }

  /* Minimum density */
  if (strcmp(keyword,"minimum")==0) {
    Tcl_SetDoubleObj(resultPtr,map->rhmin);
  }

  /* Mean density */
  if (strcmp(keyword,"mean")==0) {
    Tcl_SetDoubleObj(resultPtr,map->rhmean);
  }

  /* Rms density */
  if (strcmp(keyword,"rms")==0) {
    Tcl_SetDoubleObj(resultPtr,map->rhrms);
  }
  
  /* Cell
     Returns the cell parameters as a list of 6 numbers */
  if (strcmp(keyword,"cell")==0) {
    Tcl_ResetResult(interp);
    for (i=0; i<6; i++) {
      sprintf(tmpstr,"%f",map->cell[i]);
      Tcl_AppendElement(interp,tmpstr);
    }
  }

  /* Limits
     Returns the start and stop grid positions along x, y and z as a
     list of 6 numbers */
  if (strcmp(keyword,"limits")==0) {
    Tcl_ResetResult(interp);
    sprintf(tmpstr,"%d",map->start[0]);
    Tcl_AppendElement(interp,tmpstr);
    sprintf(tmpstr,"%d",(map->start[0]+map->ext[0]-1));
    Tcl_AppendElement(interp,tmpstr);
    sprintf(tmpstr,"%d",map->start[1]);
    Tcl_AppendElement(interp,tmpstr);
    sprintf(tmpstr,"%d",(map->start[1]+map->ext[1]-1));
    Tcl_AppendElement(interp,tmpstr);
    sprintf(tmpstr,"%d",map->start[2]);
    Tcl_AppendElement(interp,tmpstr);
    sprintf(tmpstr,"%d",(map->start[2]+map->ext[2]-1));
    Tcl_AppendElement(interp,tmpstr);
  }

  /* Title */
  /* Returns the title from the map */
  if (strcmp(keyword,"title")==0) {
    lenresult = (strlen(map->title));
    resultstr = (char *) malloc((lenresult+1)*sizeof(char));
    strncpy(resultstr,map->title,lenresult);
    resultstr[lenresult] = '\0';
    Tcl_SetStringObj(resultPtr,resultstr,strlen(resultstr));
  }

  /* Grid
     Returns the number of grid points along x,y,z as a list
     of 3 numbers */
  if (strcmp(keyword,"grid")==0) {
    Tcl_ResetResult(interp);
    for (i=0; i<3; i++) {
      sprintf(tmpstr,"%d",map->mxyz[i]);
      Tcl_AppendElement(interp,tmpstr);
    }
  }

  /* Type
     Returns the map type ("map", "mask" or "other") based on the
     mode of the original map file */
  if (strcmp(keyword,"type")==0) {
    Tcl_ResetResult(interp);
    switch (map->mode) {
    case 0:
      sprintf(tmpstr,"mask");
      break;
    case 2:
      sprintf(tmpstr,"map");
      break;
    default:
      sprintf(tmpstr,"other");
      break; 
    }
    Tcl_AppendElement(interp,tmpstr);
  }

  /* Mode
     Returns the mode from the original file */
  if (strcmp(keyword,"mode")==0) {
    Tcl_ResetResult(interp);
    sprintf(tmpstr,"%d",map->mode);
    Tcl_AppendElement(interp,tmpstr);
  }

  /* Map */
  /* This option is intended for testing purposes only! */
  if (strcmp(keyword,"map")==0) {
    puts("map keyword disabled - sorry!");
    /*for (k=0; k<map->ext[2]; ++k)
      {
      for (j=0; j<map->ext[1]; ++j)
      {
      for (i=0; i<map->ext[0]; ++i)
      printf(" %d ", (int) map->map[i][j][k]);
      printf("\n");
      }
      printf("\n");
      }*/
  }

  if (resultstr) free(resultstr);
  return TCL_OK;
}

/*----------------------------------------------------------------*/
      
/* DeleteMap  --
   Tcl command to release the memory holding the map and remove
   it from the array of stored maps.

   Usage:
   deletemap mapid

   Returns:
   -1 = no map currently stored
   0  = successful
*/

int DeleteMap(ClientData clientData, Tcl_Interp *interp,
	      int objc, Tcl_Obj *CONST objv[])
{
  /* Declarations */
  char    *mapid;
  int     lenname;
  Tcl_Obj *resultPtr;
  CCP4MAP     *map;

  /* Check number of args - should be 2 */
  if (objc!=2) {
    Tcl_WrongNumArgs(interp, 1, objv,"\0");
    return TCL_ERROR;
  }

  /* Get the mapid */
  mapid = Tcl_GetStringFromObj(objv[1], &lenname);
  if (CCP4MAPWISH_MAP_DIAG) printf("DeleteMap: mapid is \"%s\"\n",mapid);

  /* Retrieve map pointer */
  map = fetch_map_ptr(mapid);

  /* Check - does the map exist? */
  if (!map) {
    if (CCP4MAPWISH_MAP_DIAG) printf("DeleteMap: no map \"%s\" was found\n",mapid);
    resultPtr = Tcl_GetObjResult(interp);
    Tcl_SetIntObj(resultPtr, -1);
    return TCL_OK;
  }

  /* Remove the map from memory */
  zap_map_ptr(mapid);
  resultPtr = Tcl_GetObjResult(interp);
  Tcl_SetIntObj(resultPtr, 0);
  return TCL_OK;
}

/*******************************************************************/

/* Map storage and retrieval functions */

/*******************************************************************/

/* This must be called once to set up the array of maps stored in
   memory
*/
int init_ccp4maparray(void) {
  ccp4mapwish_maparray.nmaps = 0;
  ccp4mapwish_maparray.map_ptr = NULL;
  return 1;
}

/* Returns the pointer to the map with handle mapid, or
   NULL if no such handle can be found */
CCP4MAP* fetch_map_ptr(char *mapid)
{
  int i,nmaps;
  CCP4MAP *map_ptr;

  if (!mapid || ccp4mapwish_maparray.nmaps == 0) return NULL;
  nmaps = ccp4mapwish_maparray.nmaps;
  for (i=0; i<nmaps; i++) {
    map_ptr = ccp4mapwish_maparray.map_ptr[i];
    if (map_ptr) {
      /*
	printf("fetch_map_ptr: mapid for map %d is \"%s\"\n",i,map_ptr->mapid);
      */
      if (strcmp(mapid,map_ptr->mapid) == 0) return map_ptr;
    }
  }
  return NULL;
}

/* Sets up a new map structure in the array with the correct
   handle mapid, and returns the pointer to that structure
   (or NULL if there was an error) */
CCP4MAP* new_map_ptr(char *mapid)
{
  int nmaps,lenid;
  char *id_ptr=NULL;
  CCP4MAP *map_ptr=NULL;
  CCP4MAP **arr_ptr;

  if (!mapid) return NULL;
  if (CCP4MAPWISH_MAP_DIAG) printf("new_map_ptr: mapid id is \"%s\"\n",mapid);

  /* Allocate memory for a new map */
  map_ptr = (CCP4MAP *) malloc(sizeof(CCP4MAP));
  if (!map_ptr) return NULL;
  if (CCP4MAPWISH_MAP_DIAG) puts("new_map_ptr: memory allocated for new map_ptr");

  /* Set the mapid for the new map */
  lenid = strlen(mapid);
  if (CCP4MAPWISH_MAP_DIAG) printf("new_map_ptr: length of id string is %d\n",lenid);
  id_ptr = (char *) malloc((lenid+1)*sizeof(char));
  if (!id_ptr) {
    if (map_ptr) free(map_ptr);
    return NULL;
  }
  if (CCP4MAPWISH_MAP_DIAG) puts("new_map_ptr: memory allocated for new id string");
  strncpy(id_ptr,mapid,lenid);
  id_ptr[lenid] = '\0';
  if (CCP4MAPWISH_MAP_DIAG) printf("new_map_ptr: id string copied, \"%s\"\n",id_ptr);

  /* Add a pointer to the new map in the map array */
  nmaps = ccp4mapwish_maparray.nmaps;
  if (CCP4MAPWISH_MAP_DIAG) 
    printf("new_map_ptr: number of maps currently stored = %d\n",nmaps);
  arr_ptr = ccp4mapwish_maparray.map_ptr;
  nmaps++;
  arr_ptr = (CCP4MAP **) realloc(arr_ptr,(nmaps*sizeof(CCP4MAP *)));
  if (!arr_ptr) {
    if (map_ptr) free(map_ptr);
    if (id_ptr) free(id_ptr);
    return NULL;
  }
  if (CCP4MAPWISH_MAP_DIAG)
    puts("new_map_ptr: reallocated memory for array of maps");
  ccp4mapwish_maparray.nmaps = nmaps;
  ccp4mapwish_maparray.map_ptr = arr_ptr;
  ccp4mapwish_maparray.map_ptr[nmaps-1] = map_ptr;
  map_ptr->mapid = id_ptr;
  if (CCP4MAPWISH_MAP_DIAG) 
    puts("new_map_ptr: set all pointers, finished");
  return map_ptr;
}

/* Remove a map pointer with handle mapid from the map array
   Calls deletemap to free all the associated memory first
*/
int zap_map_ptr(char *mapid)
{
  int nmaps,i,j;
  CCP4MAP *map_ptr;

  if (!mapid || ccp4mapwish_maparray.nmaps == 0) return 0;
  nmaps = ccp4mapwish_maparray.nmaps;
  j = nmaps - 1;
  for (i=0; i<nmaps; i++) {
    map_ptr = ccp4mapwish_maparray.map_ptr[i];
    if (map_ptr) {
      if (CCP4MAPWISH_MAP_DIAG) 
	printf("zap_map_ptr: mapid for map %d is \"%s\"\n",i,map_ptr->mapid);
      if (strcmp(mapid,map_ptr->mapid) == 0) {
	/* Free memory associated with the map which was allocated
	   here previously */
	if (map_ptr->mapid) free(map_ptr->mapid);
	if (CCP4MAPWISH_MAP_DIAG) puts("zap_map_ptr: calling deletemap");
	deletemap(map_ptr);
	if (CCP4MAPWISH_MAP_DIAG) puts("zap_map_ptr: done deletemap");
	/* Overwrite this position with the map info from the
	   end of the array */
	if (i!=j) ccp4mapwish_maparray.map_ptr[i] = ccp4mapwish_maparray.map_ptr[j];
	nmaps--;
	ccp4mapwish_maparray.map_ptr = (CCP4MAP **)
	  realloc(ccp4mapwish_maparray.map_ptr,(sizeof(CCP4MAP *))*nmaps);
	ccp4mapwish_maparray.nmaps = nmaps;
	return 1;
      }
    }
  }
  if (CCP4MAPWISH_MAP_DIAG) printf("zap_map_ptr: mapid \"%s\" not found\n",mapid);
  return 1;
}

/* Remove a stored map and release all associated memory
*/
int deletemap(CCP4MAP *mapPtr)
{
  if (mapPtr) {
    /* Release character strings */
    if (mapPtr->filename) free(mapPtr->filename);
    if (mapPtr->title) free(mapPtr->title);
    /* Release memory used for the map values */
    if (mapPtr->map) free(mapPtr->map);
    /* Release memory used for the map structure */
    free(mapPtr);
  }
  return 1;
}

/*------------------------------------------------------------------------------*/
