/*
     ccp4mapwish_section.c: handling section information stored in memory
     Copyright (C) 2001  CCLRC, Peter Briggs

     This code is distributed under the terms and conditions of the
     CCP4 Program Suite Licence Agreement as a CCP4 Application.
     A copy of the CCP4 licence can be obtained by writing to the
     CCP4 Secretary, Daresbury Laboratory, Warrington WA4 4AD, UK.
*/

/*******************************************************************/

/* ccp4mapwish_section.c */

/*******************************************************************/

/* Functions for handling section information stored in memory */

/*******************************************************************/

/* Include header file */

/*******************************************************************/

#include "ccp4mapwish.h"

/*******************************************************************/

/* Private local variables */

/*******************************************************************/

static CCP4SECTIONARRAY ccp4mapwish_sectionarray;

/*******************************************************************/

/* Tcl SECTION functions */

/*******************************************************************/

/* Section_cmd
   This implements the Tcl command "section" for manipulating
   CCP4 section structures in memory

   The general form of the command is:

   section <name> option ?arg arg ...?

   <name> is the handle id for a section structure in memory. "Option"
   and the "arg"s determine the exact behaviour of the command.
   The following options are available:

   section <name> init <mapid> <axis> <section> ?<section2> ?<slab_step>??
         ?option value ...?
         Initialise a new section structure called <name>, with
	 section number <section> on axis <axis> from map <mapid>.

	 If <section2> is also specified then the section is actually
	 a slab of sections from <section> to <section2>.
	 An optional third number specifies the slab_step size between
	 the two limits.

   section <name> configure ?option? ?value? ?option value ...?
         Change the settings in the existing section structure <name>
	 If option is specified with no value then the current setting
	 is returned.

   section <name> contours <type> range <min> <max> <interval> ?option ...?
   section <name> contours <type> list value ?value ...? ?option ...?
   section <name> contours
         Set or update the contour levels for section <name>

	 <type> is one of abs, sigma or frac

	 range: contour levels will be set starting from <min> in steps
	        of <interval> upto <max>
	 list:  contour levels as an explicit list of values

	 Option can be one or more of:
	 -negcontours: generate equivalent negative contour levels
	 -meanoffset : generate contour levels starting from the mean
	               density in the map (default is from zero)

	 Without any arguments, return the contours as a list of 
	 absolute values.

   section <name> coords <canvas> x y ?-precision value?
         Return the fractional coordinates in section <name> corresponding
	 to the canvas coordinates x,y as rendered on <canvas>
	 The optional -precision argument specifies the number of
	 decimal places to return

   section <name> render <canvas> ?arg arg ...?
         Draw contours on <canvas> with the settings in <name>
	 -scale <scale> : sets the scaling to <scal><units>/Angstrom
	 if no units are specified then defaults to pixels

   section <name> delete
         Delete the section structure <name> from memory

   section <name> exists
         Check whether the section <name> currently exists
	 Returns 1 if section is registered, 0 otherwise

   section <name> info option
         Inquiry about general properties of the section <name>
         "option" can be one of:
         exists    : return 1 if <name> exists, 0 otherwise
         extent    : return limits of the section within the map
	             nb these may be different from those used in
		     the display (see config -limits)
	 mapid     : return the id of the parent map

   Allowed args for init and configure options are:
   -axis <axis>            : set axis from which to take section
   -bbox <boolean>         : set whether bounding box is drawn
   -grid <boolean>         : set whether grid lines are drawn
   -gridlabels <boolean>   : set whether grid labels are drawn
   -gridspacing <u> <v>    : set the spacing of gridlines on each axis
   -limits <x1> <x2> <y1> <y2> : set limits of the section in the display
	                     If <x1>...<y2> are all > 1 then the limits
			     are taken to be in grid units; otherwise they
			     are taken as fractional units.
   -precision <ndp>        : set the number of d.p. for gridlabels
   -section <sec> ?<sec2> ?<step>?? : change section number or slab
                             definition
*/

int SectionCmd(ClientData clientData, Tcl_Interp *interp,
	       int objc, Tcl_Obj *CONST objv[])
{
  char *sectionid,*option;
  int  lensec,lenopt;

  int  tcl_status;

  CCP4SECTION *section;
  
  /* Check the number of args - should be at least 3 */
  if (objc < 3) {
    Tcl_WrongNumArgs(interp,1,objv,"secid option ?args?");
    return TCL_ERROR;
  }

  /* First argument should be a section id */
  sectionid = Tcl_GetStringFromObj(objv[1],&lensec);
  if (CCP4MAPWISH_SECTION_DIAG) 
    printf("SectionCmd: sectionid = \"%s\"\n",sectionid);

  /* Second argument is the option */
  option = Tcl_GetStringFromObj(objv[2],&lenopt);

  /* Check whether this section id has already been registered */
  section = fetch_section_ptr(sectionid);
  if (CCP4MAPWISH_SECTION_DIAG) {
    if (section) {
      puts("SectionCmd: sectionid already registered");
    } else {
      puts("SectionCmd: sectionid not currently registered");
    }
  }

  /* Call function for specific option */
  if (strcmp(option,"init")==0) {
    if (CCP4MAPWISH_SECTION_DIAG) printf("SectionCmd: init option\n");
    tcl_status = section_init(sectionid,section,clientData,interp,objc,objv);
  } else if (strcmp(option,"configure")==0 || strcmp(option,"config")==0) {
    if (CCP4MAPWISH_SECTION_DIAG) printf("SectionCmd: configure option\n");
    tcl_status = section_config(section,clientData,interp,objc,objv);
  } else if (strcmp(option,"contours")==0) {
    if (CCP4MAPWISH_SECTION_DIAG) printf("SectionCmd: contours option\n");
    tcl_status = section_contours(section,clientData,interp,objc,objv);
  } else if (strcmp(option,"render")==0) {
    if (CCP4MAPWISH_SECTION_DIAG) printf("SectionCmd: render option\n");
    tcl_status = section_render(section,clientData,interp,objc,objv);
  } else if (strcmp(option,"delete")==0) {
    if (CCP4MAPWISH_SECTION_DIAG) printf("SectionCmd: delete option\n");
    tcl_status = section_delete(section,clientData,interp,objc,objv);
  } else if (strcmp(option,"coords")==0) {
    if (CCP4MAPWISH_SECTION_DIAG) printf("SectionCmd: coords option\n");
    tcl_status = section_coords(section,clientData,interp,objc,objv);
  } else if (strcmp(option,"exists")==0) {
    if (CCP4MAPWISH_SECTION_DIAG) printf("SectionCmd: exists option\n");
    tcl_status = section_exists(section,clientData,interp,objc,objv);
  } else if (strcmp(option,"info")==0) {
    if (CCP4MAPWISH_SECTION_DIAG) printf("SectionCmd: info option\n");
    tcl_status = section_info(section,clientData,interp,objc,objv);
  } else {
    printf("SectionCmd: option \"%s\" not recognised\n",option);
    return TCL_ERROR;
  }
  return tcl_status;
}

/*----------------------------------------------------------------*/

/* section_init
   Implements the "section ... init" command

   Syntax:
   section <name> init <mapid> <axis> <sec1> ?<sec2> ?<slab_step>?? ?args?
*/
int section_init(char *sectionid, CCP4SECTION *section, ClientData clientData,
		 Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[])
{
  char *mapid,*axis,*word,*endptr;
  int  lenmap,lenaxis,lenword;

  int  tcl_status;
  int  iaxis,iu,iv;
  int  this_sec,last_sec,slab_step=1;

  int  next_arg=0;

  /* Check number of arguments */
  if (objc < 6) {
    Tcl_WrongNumArgs(interp,1,objv,"secid init mapid axis sec1 ?sec2? ?args?");
    return TCL_ERROR;
  }

  /* If the section already exists then this is an error */
  if (section) {
    printf("SectionCmd(init): section already exists\n");
    return TCL_ERROR;
  }

  /* Process the remaining compulsory arguments */

  /* Map id */
  mapid = Tcl_GetStringFromObj(objv[3],&lenmap);
  if (CCP4MAPWISH_SECTION_DIAG) printf("SectionCmd(init): mapid = \"%s\"\n",mapid);

  /* Axis name */
  axis = Tcl_GetStringFromObj(objv[4],&lenaxis);
  if (get_axis_id(axis,&iaxis,&iu,&iv)) {
    if (CCP4MAPWISH_SECTION_DIAG)
      printf("SectionCmd(init): axis is \"%s\" (%d), iu=%d iv=%d\n",axis,iaxis,iu,iv);
  } else {
    printf("SectionCmd(init): axis \"%s\" unrecognised\n",axis);
    return TCL_ERROR;
  }

  /* Section number */
  if (Tcl_GetIntFromObj(interp, objv[5],&this_sec)!= TCL_OK) {
    puts("SectionCmd(init): invalid section specification");
    return TCL_ERROR;
  }
  if (CCP4MAPWISH_SECTION_DIAG)
    printf("SectionCmd(init): section number = %d\n",this_sec); 
  
  /* Look for a second section number, which specifies a slab */
  if (objc > 6) {
    word = Tcl_GetStringFromObj(objv[6],&lenword);
    /* Extract a long int using strtol
       If no conversion takes place then "endptr" will be equal
       to "word" */
    if (CCP4MAPWISH_SECTION_DIAG)
      printf("SectionCmd(init): possible second section is \"%s\"\n",word); 
    last_sec = (int) strtol(word,&endptr,10);
    if (strcmp(word,endptr) == 0) {
      if (CCP4MAPWISH_SECTION_DIAG) 
	puts("SectionCmd(init): No second section number - single section");
      last_sec = this_sec;
      next_arg = 6;
    } else {
      if (CCP4MAPWISH_SECTION_DIAG)
	printf("SectionCmd(init): second section number  = %d\n",last_sec);
      next_arg = 7;
    }
    /* Look for a third number, which specifies a slab step */
    if (objc > 7) {
      word = Tcl_GetStringFromObj(objv[7],&lenword);
    /* Extract a long int using strtol
       If no conversion takes place then "endptr" will be equal
       to "word" */
      if (CCP4MAPWISH_SECTION_DIAG)
	printf("SectionCmd(init): possible slab step is \"%s\"\n",word); 
      slab_step = (int) strtol(word,&endptr,10);
      if (strcmp(word,endptr) == 0) {
	if (CCP4MAPWISH_SECTION_DIAG) 
	  puts("SectionCmd(init): No slab step given - assume default");
	slab_step = 1;
	next_arg = 7;
      } else {
	if (CCP4MAPWISH_SECTION_DIAG) 
	  printf("SectionCmd(init): slab step = %d\n",slab_step);
	next_arg = 7;
      }
    }
  } else {
    /* Set second section number equal to first */
    last_sec = this_sec;
    next_arg = objc;
  }

  /* Set up a new section structure */
  section = init_section_strct(sectionid,mapid,axis,iaxis,iu,iv,this_sec,
			       last_sec,slab_step);
  if (!section) {
    puts("SectionCmd(init): unable to create new section");
    return TCL_ERROR;
  }

  /* Process any remaining optional arguments */
  if (next_arg < objc) {
    if (CCP4MAPWISH_SECTION_DIAG) 
      puts("SectionCmd(init): processing optional arguments");
    tcl_status = section_process_args("config",section,next_arg,clientData,
				      interp,objc,objv);
    if (tcl_status != TCL_OK) return tcl_status;
  }


  return TCL_OK;
}

/*----------------------------------------------------------------*/

/* section_config
   Implements the "section ... configure" command
*/
int section_config(CCP4SECTION *section, ClientData clientData, Tcl_Interp
		   *interp, int objc, Tcl_Obj *CONST objv[])
{
  int tcl_status;

  /* Check number of arguments */
  if (objc < 4) {
    Tcl_WrongNumArgs(interp,1,objv,"secid config option ?value? ?option value ...?");
    return TCL_ERROR;
  }

  /* If the section doesn't exist then this is an error */
  if (!section) {
    printf("SectionCmd(config): section doesn't exist\n");
    return TCL_ERROR;
  }

  /* If there are exactly four arguments then assume we are
     in "query" mode */
  if (objc == 4) {
    if (CCP4MAPWISH_SECTION_DIAG) printf("SectionCmd(config): \"query\" mode\n");
    tcl_status = section_process_args("query",section,3,clientData,
				      interp,objc,objv);
  } else {
    if (CCP4MAPWISH_SECTION_DIAG) printf("SectionCmd(config): \"config\" mode\n");
    tcl_status = section_process_args("config",section,3,clientData,
				      interp,objc,objv);
  }
  return tcl_status;
}

/*----------------------------------------------------------------*/

/* section_contours
   Implements the "section ... contours" command
*/
int section_contours(CCP4SECTION *section, ClientData clientData, Tcl_Interp
		     *interp, int objc, Tcl_Obj *CONST objv[])
{
  char  buffer[200];
  char  *contourtype,*option;
  int   lentype,lenopt;
  float *value_list;
  int   nvalues;
  int   i,scalemode,irange,ilist,inegcontours=0,imeanoffset=0,next_arg;
  float scale_pos,scale_neg;
  CCP4MAP *map;

  if (CCP4MAPWISH_SECTION_DIAG) 
    puts("SectionCmd(contours): started");

  /* Check number of arguments */
  if (objc < 3) {
    Tcl_WrongNumArgs(interp,1,objv,"secid contours ?units type args...?");
    return TCL_ERROR;
  }

  if (CCP4MAPWISH_SECTION_DIAG) 
    puts("SectionCmd(contours): about to check for section");

  /* If the section doesn't exist then this is an error */
  if (!section) {
    printf("SectionCmd(contours): section doesn't exist\n");
    return TCL_ERROR;
  }

  /* If the parent map doesn't exist then this is also an error */
  map = fetch_map_ptr(section->mapid);
  if (!map) {
    printf("SectionCmd(contours): parent map doesn't exist\n");
    return TCL_ERROR;
  } 

  /* If there are exactly three arguments then this is a query so
     return the contour levels as a list */
  if (objc == 3) {
    if (CCP4MAPWISH_SECTION_DIAG) 
      puts("SectionCmd(contours): request for information");
    if (section->nlevels == 0) {
      /* No contours stored - return an empty list */
      Tcl_AppendElement(interp,"");
      return TCL_OK;
    }
    /* Check we have some values to write out */
    if (!section->levels) {
      puts("SectionCmd(contours): levels is null");
      return TCL_ERROR;
    }
    /* Build a list of the contour levels as the result */
    for (i=0; i<section->nlevels; i++) {
      sprintf(buffer,"%f",section->levels[i]);
      Tcl_AppendElement(interp,buffer);
    }
    return TCL_OK;
  }

  /* Check the third argument - this should be the type of
     contour levels being supplied */
  contourtype = Tcl_GetStringFromObj(objv[3],&lentype);
  if (CCP4MAPWISH_SECTION_DIAG)
    printf("SectionCmd(contours): type is \"%s\"\n",contourtype);

  if (strcmp(contourtype,"abs")==0) {
    /* ABSolute scale option: scalemode 1
       The contours are assumed to be absolute in value so no
       additional scaling is applied */
    scalemode = 1;
  } else if (strcmp(contourtype,"frac")==0) {
    /* FRACtional scale option : scalemode 2
       The contours are assumed to be fractions of the maximum
       density (for positive levels) and of the (absolute)
       minimum density (for negative levels) */
    scalemode = 2;
    scale_pos = map->rhmax;
    scale_neg = (float) fabsf(map->rhmin);
  } else if (strcmp(contourtype,"sigma")==0) {
    /* SIGMA levels option : scalemode 3
       The contours are assumed to be multiples of the sigma
       level */
    scalemode = 3;
    scale_pos = map->rhrms;
    scale_neg = scale_pos;
  } else {
    /* Unrecognised contour type */
    printf("SectionCmd(contours): Unrecognised type: \"%s\"\n",contourtype);
    return TCL_ERROR;
  }
  if (CCP4MAPWISH_SECTION_DIAG)
    printf("SectionCmd(contours): scalemode is %d\n",scalemode);

  /* Fourth argument - specifies the method of
     supplying the contour levels */
  option = Tcl_GetStringFromObj(objv[4],&lenopt);
  if (CCP4MAPWISH_SECTION_DIAG)
    printf("SectionCmd(contours): option is \"%s\"\n",option);

  if (strcmp(option,"range") == 0) {
    irange = 1;
    ilist  = 0;
  } else if (strcmp(option,"list") == 0) {
    irange = 0;
    ilist  = 1;
  } else {
    printf("SectionCmd(contours): Unrecognised option: \"%s\"\n",option);
    return TCL_ERROR;
  }

  /* Extract the values from the remaining arguments */
  if (irange) {
    if (CCP4MAPWISH_SECTION_DIAG) 
      puts("SectionCmd(contours): extracting arguments for range");
    next_arg = section_process_contour_range_args(clientData,
						  interp,objc,objv,
						  5,&nvalues,&value_list);
  } else if (ilist) {
    if (CCP4MAPWISH_SECTION_DIAG) 
      puts("SectionCmd(contours): extracting arguments for list");
    next_arg = section_process_contour_list_args(clientData,
						 interp,objc,objv,
						 5,&nvalues,&value_list);
  }
  if (!value_list) {
    printf("SectionCmd(contours): pointer to list of values is null\n");
    return TCL_ERROR;
  }
  if (!next_arg) {
    puts("SectionCmd(contours): error extracting arguments for range");
    if (value_list) free(value_list);
    return TCL_ERROR;
  }
  if (CCP4MAPWISH_SECTION_DIAG) 
    printf("SectionCmd(contours): extracted %d values from list\n",nvalues);

  /* Process the remaining arguments */
  for (i=next_arg; i<objc; i++) {
    if (CCP4MAPWISH_SECTION_DIAG) 
      printf("SectionCmd(contours): examining argument %d\n",i);
    option = Tcl_GetStringFromObj(objv[i],&lenopt);
    if (CCP4MAPWISH_SECTION_DIAG) 
      printf("SectionCmd(contours): option is \"%s\"\n",option);

    if (strcmp(option,"-negcontours") == 0) {
      inegcontours = 1;
      if (CCP4MAPWISH_SECTION_DIAG) 
	puts("SectionCmd(contours): Negative equivalent contours specified");
    } else if (strcmp(option,"-meanoffset") == 0) {
      imeanoffset = 1;
      if (CCP4MAPWISH_SECTION_DIAG) 
	puts("SectionCmd(contours): Offset contour levels by mean density");
    } else {
      printf("SectionCmd(contours): unrecognised option \"%s\"\n",option);
      if (value_list) free(value_list);
      return TCL_ERROR;
    }
  }
  
  /* Generate the contour levels */
  if (inegcontours) {
    if (CCP4MAPWISH_SECTION_DIAG) 
      puts("SectionCmd(contours): about to generate negative contours");
    for (i=0; i<nvalues; i++) {
      if (value_list[i] > 0) {
	nvalues++;
	value_list = (float *) realloc(value_list,nvalues*sizeof(float));
	if (!value_list) {
	  puts("SectionCmd(contours): couldn't reset value_list pointer");
	  return TCL_ERROR;
	}
	value_list[nvalues-1] = -value_list[i];
      }
    }
  }

  /* Sort the contours into ascending order using qsort */
  qsort(value_list,nvalues,sizeof(float),asc_cmp);

  /* Convert the values depending on the scalemode */
  if (scalemode != 1) {
    for (i=0; i<nvalues; i++) {
      if (value_list[i] > 0) {
	value_list[i] = scale_pos*value_list[i];
      } else {
	value_list[i] = scale_neg*value_list[i];
      }
      /* Offset the values by the mean density,
	 if requested */
      if (imeanoffset) value_list[i] = value_list[i] + map->rhmean;
    }
  }
  

  if (CCP4MAPWISH_SECTION_DIAG) {
    printf("SectionCmd(contours): levels are");
    for (i=0; i<nvalues; i++) {
      printf(" %f",value_list[i]);
    }
    printf("\n");
  }

  /* Store the levels in the section structure */
  if (section->levels) free(section->levels);
  section->levels  = value_list;
  section->nlevels = nvalues;

  /* Finished */
  return TCL_OK;
}

/*----------------------------------------------------------------*/

/* section_render
   Implements the "section ... render" command

   Options:
   -scale      set the scale for the rendering (default 1px/Angstrom)
   -greyscale  draw a greyscale backdrop (default is not to draw greyscale)
   -nocontours don't display contour lines (default is to draw contours)

*/
int section_render(CCP4SECTION *section, ClientData clientData, Tcl_Interp
		   *interp, int objc, Tcl_Obj *CONST objv[])
{
  char    *canvas,*option,*word,*endptr;
  char    scaleunits,tmpstr[CCP4_LTMPSTR];
  int     i,lencanvas,lenopt,lenword;
  int     greyscale=0,contours=1;
  float   scale=1.0,tkscale;
  CCP4MAP *map;
  CCP4CONTOURCANVAS *canvasptr;

  /* If the section doesn't exist then this is an error */
  if (!section) {
    printf("SectionCmd(render): section doesn't exist\n");
    return TCL_ERROR;
  }

  /* If the parent map doesn't exist then this is also an error */
  map = fetch_map_ptr(section->mapid);
  if (!map) {
    printf("SectionCmd(render): parent map \"%s\" doesn't exist\n",
	   section->mapid);
    return TCL_ERROR;
  }

  /* Check the number of arguments */
  if (objc < 4) {
    Tcl_WrongNumArgs(interp,1,objv,"secid render canvas ?option ...?");
    return TCL_ERROR;
  }
  
  /* Get the name of the canvas widget to write to */
  canvas = Tcl_GetStringFromObj(objv[3],&lencanvas);
  if (CCP4MAPWISH_SECTION_DIAG) printf("SectionCmd(render): canvas is \"%s\"\n",canvas);

  /* Process any remaining arguments */
  if (objc > 4) {
    i = 4;
    while (i<objc) {
      option = Tcl_GetStringFromObj(objv[i],&lenopt);
      if (CCP4MAPWISH_SECTION_DIAG)
	printf("SectionCmd(render): option %d is \"%s\"\n",i,option);

      if (strcmp(option,"-scale") == 0) {
	/* Scale option
	   Set the scaling to use in units/Angstrom 
	   If a single number then it's pixels, otherwise a number
	   followed by m (=millimetres), i (=inches), c (=centimetres) */
	if (CCP4MAPWISH_SECTION_DIAG) printf("SectionCmd(render): scale option\n");
	i++;
	if (i<objc) {
	  word = Tcl_GetStringFromObj(objv[i],&lenword);
	  if (CCP4MAPWISH_SECTION_DIAG)
	    printf("SectionCmd(render): scale is \"%s\"\n",word);
	  /* Check for possible units */
	  switch (word[lenword-1]) {
	  case 'm':
	    if (CCP4MAPWISH_SECTION_DIAG) 
	      puts("SectionCmd(render): scale in millimetres");
	    scaleunits = 'm';
	    lenword--;
	    break;
	  case 'i':
	    if (CCP4MAPWISH_SECTION_DIAG) 
	      puts("SectionCmd(render): scale in inches");
	    scaleunits = 'i';
	    lenword--;
	    break;
	  case 'c':
	    if (CCP4MAPWISH_SECTION_DIAG) 
	      puts("SectionCmd(render): scale in centimetres");
	    scaleunits = 'c';
	    lenword--;
	    break;
	  default:
	    if (CCP4MAPWISH_SECTION_DIAG) 
	      puts("SectionCmd(render): scale assumed to be in pixels");
	    scaleunits = ' ';
	  }
	  strncpy(tmpstr,word,CCP4_LTMPSTR-1);
	  tmpstr[lenword] = '\0';
	  /* Try to extract a number for the scale factor */
	  scale = (float) strtod(tmpstr,&endptr);
	  if (strcmp(word,endptr) == 0) {
	    puts("SectionCmd(render): cannot read scale factor");
	    return TCL_ERROR;
	  }
	  /* If scale factor is zero or negative then ignore */
	  if (scale <= 0.0) {
	    if (CCP4MAPWISH_SECTION_DIAG) 
	      printf("SectionCmd(render): scale of %f is not possible - ignored\n",
		     scale);
	  }
	  if (CCP4MAPWISH_SECTION_DIAG) 
	    printf("SectionCmd(render): scaling set to %f units/Angstrom\n",scale);
	} else {
	  puts("SectionCmd(render): no value supplied for -scale option");
	  return TCL_ERROR;
	}
      } else if (strcmp(option,"-greyscale") == 0) {
	/* -greyscale
	   Turn on the display of greyscale background (single sections
	   only) */
	greyscale = 1;
      } else if (strcmp(option,"-nocontours") == 0) {
	/* -nocontours
	   Switch off the display of contour lines */
	contours = 0;
      } else {
	/* Unrecognised keyword or option */
	printf("SectionCmd(render): unrecognised option \"%s\"\n",option);
	return TCL_ERROR;
      }
      /* Increment counter to look at the next argument */
      i++;
    }
  }

  /* FIXME: you should check whether this is a valid
     canvas name */

  /* Store the canvas name and section pointer so they can be accessed
     by the drawing functions */
  init_contour_canvas(canvas,section,scaleunits);

  /* Fetch the Tk scaling factor */
  tkscale = canvas_get_tk_scaling(canvas,scaleunits);
  if (CCP4MAPWISH_SECTION_DIAG)
    printf("SectionCmd(render): tk scale factor = %f\n",tkscale);

  /* Register the canvas in the array of canvases, if it doesn't already
     exist there */
  canvasptr = fetch_canvas_ptr(canvas);
  if (!canvasptr) {
    if (CCP4MAPWISH_SECTION_DIAG)
      printf("SectionCmd(render): registering new canvas \"%s\"\n",canvas);
    canvasptr = new_canvas_ptr(canvas);
    if (!canvasptr) {
      puts("SectionCmd(render): couldn't register canvas");
      return TCL_ERROR;
    }
  } else {
    if (CCP4MAPWISH_SECTION_DIAG) 
      printf("SectionCmd(render): canvas \"%s\" already registered\n",canvas);
  }

  /* Update the registered value for the scaling */
  canvasptr->scale = scale*tkscale;
  if (CCP4MAPWISH_SECTION_DIAG) 
    printf("SectionCmd(render): overall canvas scale factor (px/A) = %f\n",
	   canvasptr->scale);

  if (CCP4MAPWISH_SECTION_DIAG) printf("SectionCmd(render): rendering section\n");

  /* Blitz all existing elements on the canvas first */
  if (!canvas_delete(canvas,"all")) {
    puts("SectionCmd(render): error trying to delete existing elements in canvas");
    return TCL_ERROR;
  }

  /* Do things in the following order:
     1. Draw bounding box (if requested)
     2. Draw gridlines and labels (if requested)
     3. Draw section(s)
     4. Scale everything so far
     5. Draw axes
     6. Shift everything
     7. Draw titles
*/

  /* Draw colourscaled image
     Do this first so other components appear on top */
  if (greyscale) {
    /* Greyscale aborted if we are dealing with a slab */
    if (section->slab_start == section->slab_end)
      if (!greyscale_section(section)) {
	printf("SectionCmd(render): error generating greyscale\n");
	return TCL_ERROR;
      }
  }

  /* Draw bounding box */
  if (section->bbox) {
    if (CCP4MAPWISH_SECTION_DIAG)
      printf("SectionCmd(render): drawing bounding box\n");
    if (!write_contour_boundingbox(section)) {
      puts("SectionCmd(render): error drawing bounding box");
      return TCL_ERROR;
    }
  }

  /* Draw grid lines and labels */
  if (section->grid || section->labels) {
    if (CCP4MAPWISH_SECTION_DIAG) 
      printf("SectionCmd(render): drawing gridlines and labels\n");
    if (!write_contour_gridlines(section)) {
      puts("SectionCmd(render): error drawing grid");
      return TCL_ERROR;
    }
  }

  /* Draw section/slab */
  if (contours) {
    if (CCP4MAPWISH_SECTION_DIAG) 
      printf("SectionCmd(render): generating contour lines\n");
    if (!contour_section(section)) {
      printf("SectionCmd(render): error generating contours\n");
      return TCL_ERROR;
    }
  }

  /* Scale the canvas */
  if (!canvas_scale_canvas(canvas,"all",scale)) {
    puts("SectionCmd(render): error scaling canvas");
    return TCL_ERROR;
  }

  /* Draw the axes */
  if (section->axes) {
    if (CCP4MAPWISH_SECTION_DIAG) printf("SectionCmd(render): drawing axes\n");
    if(!write_contour_axes(section,canvas,scale)) {
      puts("SectionCmd(render): error drawing axes");
      return TCL_ERROR;
    }
  }

  /* Draw the titles */
  if (section->titles) {
    if (CCP4MAPWISH_SECTION_DIAG) printf("SectionCmd(render): drawing titles\n");
    if(!write_contour_titles(section,canvas,scale)) {
      puts("SectionCmd(render): error drawing titles");
      return TCL_ERROR;
    }
  }

  /* Shift the canvas */
  if (!canvas_shift_canvas(canvas)) {
    puts("SectionCmd(render): error shifting canvas elements");
    return TCL_ERROR;
  }

  if (CCP4MAPWISH_SECTION_DIAG) printf("SectionCmd(render): finished\n");

  return TCL_OK;
}

/*----------------------------------------------------------------*/

/* section_delete
   Implements the "section ... delete" command
*/
int section_delete(CCP4SECTION *section, ClientData clientData, Tcl_Interp
		   *interp, int objc, Tcl_Obj *CONST objv[])
{
  /* If the section doesn't exist then this is an error */
  if (!section) {
    printf("SectionCmd(delete): section doesn't exist\n");
    return TCL_ERROR;
  }
  if (CCP4MAPWISH_SECTION_DIAG) printf("SectionCmd(delete): deleting section\n");
  zap_section_ptr(section->sectionid);
  return TCL_OK;
}

/*----------------------------------------------------------------*/

/* section_coords
   Implements the "section ... coords" command
*/
int section_coords(CCP4SECTION *section, ClientData clientData, Tcl_Interp
		   *interp, int objc, Tcl_Obj *CONST objv[])
{
  int   lencanvas,lennum,iu,iv,this_sec,iaxis;
  int   x_grd,y_grd;
  char  *canvas,*number,*endptr,tmpstr[CCP4_LTMPSTR];
  float x_canvas,y_canvas,x_ang,y_ang,x_frc,y_frc,rho;
  CCP4CONTOURCANVAS *canvasptr;
  CCP4MAP *map;

  if (CCP4MAPWISH_SECTION_DIAG) puts("SectionCmd(coords): started");

  /* Check number of arguments */
  if (objc != 6) {
    Tcl_WrongNumArgs(interp,1,objv,"secid coords canvas x y");
    return TCL_ERROR;
  }

  if (CCP4MAPWISH_SECTION_DIAG) puts("SectionCmd(coords): about to check for section");

  /* If the section doesn't exist then this is an error */
  if (!section) {
    printf("SectionCmd(coords): section doesn't exist\n");
    return TCL_ERROR;
  }

  /* Get a pointer to the parent map */
  if (!section->mapid) {
    printf("SectionCmd(coords): section doesn't have a parent map\n");
    return TCL_ERROR;
  }
  map = fetch_map_ptr(section->mapid);
  if (!map) {
    printf("SectionCmd(coords): cannot get a pointer to the parent map\n");
    return TCL_ERROR;
  }

  /* Process the remaining arguments */

  /* Canvas name */
  canvas = Tcl_GetStringFromObj(objv[3],&lencanvas);
  if (CCP4MAPWISH_SECTION_DIAG) 
    printf("SectionCmd(coords): canvas name is \"%s\"\n",canvas);

  /* Check that this canvas name is registered */
  canvasptr = fetch_canvas_ptr(canvas);
  if (!canvasptr) {
    printf("SectionCmd(coords): canvas is not registered\n");
    return TCL_ERROR;
  }

  /* X and Y coordinate positions in the canvas */
  number = Tcl_GetStringFromObj(objv[4],&lennum);
  x_canvas = ((float) strtod(number,&endptr));
  if (strcmp(number,endptr) == 0) {
    printf("SectionCmd(coords): bad x coordinate \"%s\"\n",number);
    return TCL_ERROR;
  }
  number = Tcl_GetStringFromObj(objv[5],&lennum);
  y_canvas = ((float) strtod(number,&endptr));
  if (strcmp(number,endptr) == 0) {
    printf("SectionCmd(coords): bad y coordinate \"%s\"\n",number);
    return TCL_ERROR;
  }

  if (CCP4MAPWISH_SECTION_DIAG)
    printf("SectionCmd(coords): x_canvas=%f, y_canvas=%f\n",x_canvas,y_canvas);

  /* Perform conversions from canvas coordinates to section
     coordinates */

  /* Correct canvas coordinates to account for offseting */
  x_canvas = x_canvas - canvasptr->xoffset;
  y_canvas = y_canvas - canvasptr->yoffset;

  if (CCP4MAPWISH_SECTION_DIAG)
    printf("SectionCmd(coords): adjusted canvas x=%f, y=%f\n",x_canvas,y_canvas);

  /* Do conversions to grid, fraction and orthogonal coords */
  inv_transform_coordinates(section,x_canvas,y_canvas,&x_grd,&y_grd,
			    &x_frc,&y_frc,&x_ang,&y_ang,canvasptr->scale);
  if (CCP4MAPWISH_SECTION_DIAG)
    printf("SectionCmd(coords): called inv_transform_coords...\n");

  /* Test whether these fall inside the current section limits */
  if (x_grd < section->ulimits[0] || x_grd > section->ulimits[1] ||
      y_grd < section->vlimits[0] || y_grd > section->vlimits[1]) {
    return TCL_OK;
  }

  /* Fetch axis information from section */
  iu = section->iu;
  iv = section->iv;
  if (CCP4MAPWISH_SECTION_DIAG)
    printf("SectionCmd(coords): axes iu = %d, iv = %d\n",iu,iv);

  /* If this is a single section then we can also return the density
     at this grid position */
  if (section->slab_start == section->slab_end) {
    if (CCP4MAPWISH_SECTION_DIAG)
      printf("SectionCmd(coords): looking to fetch density at this position\n");
    iaxis = section->iaxis;
    this_sec  = section->nsec - map->start[iaxis];
    /* Correct the indices for the section in memory */
    x_grd = x_grd - section->ulimits[0];
    y_grd = y_grd - section->vlimits[0];
    if (CCP4MAPWISH_SECTION_DIAG)
      printf("SectionCmd(coords): corrected grid coordinates are %d,%d\n",x_grd,y_grd);
    if (iaxis == 0) {
      rho = map->map[this_sec][x_grd][y_grd];
    } else if (iaxis == 1) {
      rho = map->map[y_grd][this_sec][x_grd];
    } else if (iaxis == 2) {
      rho = map->map[x_grd][y_grd][this_sec];
    } else {
      printf("SectionCmd(coords): bad axis \"%d\"\n",iaxis);
      return TCL_ERROR;
    }
    if (CCP4MAPWISH_SECTION_DIAG) printf("SectionCmd(coords): set density to %f\n",rho);
  }

  /* Build a Tcl list to return the values */
  sprintf(tmpstr,"%f %f",x_ang,y_ang);
  Tcl_AppendElement(interp,tmpstr);

  sprintf(tmpstr,"%f %f",x_frc,y_frc);
  Tcl_AppendElement(interp,tmpstr);

  if (section->slab_start == section->slab_end) {
    sprintf(tmpstr,"%f",rho);
    Tcl_AppendElement(interp,tmpstr);
  }

  if (CCP4MAPWISH_SECTION_DIAG) puts("SectionCmd(coords): returning");
  return TCL_OK;
}

/*----------------------------------------------------------------*/

/* section_exists
   Implements the "section ... exists" command
*/
int section_exists(CCP4SECTION *section, ClientData clientData, Tcl_Interp
		   *interp, int objc, Tcl_Obj *CONST objv[])
{
  Tcl_Obj *resultPtr;

  if (CCP4MAPWISH_SECTION_DIAG) puts("SectionCmd(exists): started");

  /* Set up the result pointer */
  resultPtr = Tcl_GetObjResult(interp);

  /* Check number of arguments */
  if (objc != 3) {
    Tcl_WrongNumArgs(interp,1,objv,"secid exists");
    return TCL_ERROR;
  }

  if (CCP4MAPWISH_SECTION_DIAG) puts("SectionCmd(exists): about to check for section");

  /* If the section doesn't exist then return 0, else return 1 */
  if (!section) {
    if (CCP4MAPWISH_SECTION_DIAG) printf("SectionCmd(exists): section doesn't exist\n");
    Tcl_SetIntObj(resultPtr,0);
  } else {
    if (CCP4MAPWISH_SECTION_DIAG) printf("SectionCmd(exists): section exists\n");
    Tcl_SetIntObj(resultPtr,1);
  }
  return TCL_OK;
}

/*----------------------------------------------------------------*/

/* section_info
   Implements the "section ... info" command
*/
int section_info(CCP4SECTION *section, ClientData clientData, Tcl_Interp
		 *interp, int objc, Tcl_Obj *CONST objv[])
{
  char    *option,tmpstr[CCP4_LTMPSTR];
  int     lenopt;
  CCP4MAP *map;
  Tcl_Obj *resultPtr;

  if (CCP4MAPWISH_SECTION_DIAG) puts("SectionCmd(info): started");

  /* Set up the result pointer */
  resultPtr = Tcl_GetObjResult(interp);

  /* Check number of arguments */
  if (objc != 4) {
    Tcl_WrongNumArgs(interp,1,objv,"secid info option");
    return TCL_ERROR;
  }

  if (CCP4MAPWISH_SECTION_DIAG) puts("SectionCmd(info): about to check for section");

  /* Check arguments - check for "info exists" first */
  option = Tcl_GetStringFromObj(objv[3],&lenopt);
  if (strcmp(option,"exists") == 0) {
    /* If the section doesn't exist then return 0, else return 1 */
    if (!section) {
      if (CCP4MAPWISH_SECTION_DIAG)
	printf("SectionCmd(info): exists - section doesn't exist\n");
      Tcl_SetIntObj(resultPtr,0);
    } else {
      if (CCP4MAPWISH_SECTION_DIAG)
	printf("SectionCmd(info): exists - section exists\n");
      Tcl_SetIntObj(resultPtr,1);
    }
    return TCL_OK;
  }

  /* Otherwise if section doesn't exist then it's an error */
  if (!section) return TCL_ERROR;

  /* Check option against remaining possibilities */
  if (strcmp(option,"mapid") == 0) {
    /* Mapid option */
    if (CCP4MAPWISH_SECTION_DIAG) 
      printf("SectionCmd(info): mapid option, mapid = \"%s\"\n",section->mapid);
    Tcl_SetStringObj(resultPtr,section->mapid,strlen(section->mapid));
    return TCL_OK;

  } else if (strcmp(option,"extent") == 0) {
    /* Extent option */
    if (CCP4MAPWISH_SECTION_DIAG) 
      printf("SectionCmd(info): extent option\n");
    map = fetch_map_ptr(section->mapid);
    if (!map) return TCL_ERROR;
    sprintf(tmpstr,"%d",map->start[section->iu]);
    Tcl_AppendElement(interp,tmpstr);
    sprintf(tmpstr,"%d",map->start[section->iu] + map->ext[section->iu] - 1);
    Tcl_AppendElement(interp,tmpstr);
    sprintf(tmpstr,"%d",map->start[section->iv]);
    Tcl_AppendElement(interp,tmpstr);
    sprintf(tmpstr,"%d",map->start[section->iv] + map->ext[section->iv] - 1);
    Tcl_AppendElement(interp,tmpstr);
    return TCL_OK;

  } else {
    /* Unrecognised option */
    printf("SectionCmd(info): unrecognised option \"%s\"\n",option);
    return TCL_ERROR;
  }
}

/*******************************************************************/

/* Section storage and retrieval functions */

/*******************************************************************/

/* This must be called once to set up the array of sections stored in
   memory
*/
int init_ccp4sectionarray(void) {
  ccp4mapwish_sectionarray.nsections = 0;
  ccp4mapwish_sectionarray.section_ptr = NULL;
  return 1;
}

/* Returns the pointer to the section with handle sectionid, or
   NULL if no such handle can be found */
CCP4SECTION* fetch_section_ptr(char *sectionid)
{
  int i,nsections;
  CCP4SECTION *section_ptr;

  if (!sectionid || ccp4mapwish_sectionarray.nsections == 0) return NULL;
  nsections = ccp4mapwish_sectionarray.nsections;
  for (i=0; i<nsections; i++) {
    section_ptr = ccp4mapwish_sectionarray.section_ptr[i];
    if (section_ptr) {
      /*
	printf("fetch_section_ptr: sectionid for section %d is \"%s\"\n",
	i,section_ptr->sectionid);
      */
      if (strcmp(sectionid,section_ptr->sectionid) == 0) return section_ptr;
    }
  }
  return NULL;
}

/*----------------------------------------------------------------*/

/* Sets up a new section structure in the array with the correct
   handle sectionid, and returns the pointer to that structure
   (or NULL if there was an error) */
CCP4SECTION* new_section_ptr(char *sectionid)
{
  int nsections,lenid;
  char *id_ptr=NULL;
  CCP4SECTION *section_ptr=NULL;
  CCP4SECTION **arr_ptr;

  if (!sectionid) return NULL;
  if (CCP4MAPWISH_SECTION_DIAG) 
    printf("new_section_ptr: sectionid id is \"%s\"\n",sectionid);

  /* Allocate memory for a new section */
  section_ptr = (CCP4SECTION *) malloc(sizeof(CCP4SECTION));
  if (!section_ptr) return NULL;
  if (CCP4MAPWISH_SECTION_DIAG) 
    puts("new_section_ptr: memory allocated for new section_ptr");

  /* Set the sectionid for the new section */
  lenid = strlen(sectionid);
  if (CCP4MAPWISH_SECTION_DIAG) 
    printf("new_section_ptr: length of id string is %d\n",lenid);
  id_ptr = (char *) malloc((lenid+1)*sizeof(char));
  if (!id_ptr) {
    if (section_ptr) free(section_ptr);
    return NULL;
  }
  if (CCP4MAPWISH_SECTION_DIAG) 
    puts("new_section_ptr: memory allocated for new id string");
  strncpy(id_ptr,sectionid,lenid);
  id_ptr[lenid] = '\0';
  if (CCP4MAPWISH_SECTION_DIAG) 
    printf("new_section_ptr: id string copied, \"%s\"\n",id_ptr);

  /* Add a pointer to the new section in the section array */
  nsections = ccp4mapwish_sectionarray.nsections;
  if (CCP4MAPWISH_SECTION_DIAG) 
    printf("new_section_ptr: number of sections currently stored = %d\n",nsections);
  arr_ptr = ccp4mapwish_sectionarray.section_ptr;
  nsections++;
  arr_ptr = (CCP4SECTION **) realloc(arr_ptr,(nsections*sizeof(CCP4SECTION *)));
  if (!arr_ptr) {
    if (section_ptr) free(section_ptr);
    if (id_ptr) free(id_ptr);
    return NULL;
  }
  if (CCP4MAPWISH_SECTION_DIAG) 
    puts("new_section_ptr: reallocated memory for array of sections");
  ccp4mapwish_sectionarray.nsections = nsections;
  ccp4mapwish_sectionarray.section_ptr = arr_ptr;
  ccp4mapwish_sectionarray.section_ptr[nsections-1] = section_ptr;
  section_ptr->sectionid = id_ptr;
  if (CCP4MAPWISH_SECTION_DIAG) 
    puts("new_section_ptr: set all pointers, finished");
  return section_ptr;
}

/*----------------------------------------------------------------*/

/* Set up a new section structure
   Return a pointer to a new section structure with handle
   sectionid
   Return null pointer if there was an error */
CCP4SECTION* init_section_strct(char *sectionid, char *mapid, char *axis,
				int iaxis, int iu, int iv, int first_sec,
				int last_sec, int slab_step)
{
  int lenmap,start,end;
  int nlevels,bbox,grid,titles,labels,axes,precision;
  int ulimits[2],vlimits[2];
  float frc_ulimits[2],frc_vlimits[2];
  float *levels,uspace,vspace;
  char axislabel[3][LEN_AXISLABEL],view_orient[4];
  CCP4SECTION *section;
  CCP4MAP *map;

  /* Check for valid mapid */
  map = fetch_map_ptr(mapid);
  if (!map) return NULL;

  /* Check that the requested section is within the map */  
  start = map->start[iaxis];
  end   = start + map->ext[iaxis];
  if (start > first_sec || end < first_sec || start > last_sec || end < last_sec) {
    printf("init_section_strct: specified section(s) (%d-%d) out of range for specified axis (%s)\n",first_sec,last_sec,axis);
    return NULL;
  }

  /* Set a section pointer for the new section */
  section = new_section_ptr(sectionid);
  if (!section) return NULL;

  /* Initialise the section structure */
  lenmap = strlen(mapid);
  section->mapid = (char *) malloc((lenmap+1)*sizeof(char));
  if (!section->mapid) {
    puts("init_section_strct: couldn't get a pointer to mapid");
    free(section);
    return NULL;
  }
  strncpy(section->mapid,mapid,lenmap);
  section->mapid[lenmap] = '\0';
  if (CCP4MAPWISH_SECTION_DIAG) 
    printf("init_section_strct: parent mapid registered as \"%s\"\n",section->mapid);

  /* Defaults for other attributes */

  /* Get section limits directly from the map */
  ulimits[0] = map->start[iu];
  ulimits[1] = ulimits[0] + map->ext[iu] - 1;
  vlimits[0] = map->start[iv];
  vlimits[1] = vlimits[0] + map->ext[iv] - 1;
  if (CCP4MAPWISH_SECTION_DIAG) 
    printf("init_section_strct: ulimits = %d %d, vlimits = %d %d\n",
	   ulimits[0],ulimits[1],vlimits[0],vlimits[1]);

  /* Generate initial fractional limits */
  frc_ulimits[0] = ((float) ulimits[0])/((float) map->mxyz[iu]);
  frc_ulimits[1] = ((float) ulimits[1])/((float) map->mxyz[iu]);
  frc_vlimits[0] = ((float) vlimits[0])/((float) map->mxyz[iv]);
  frc_vlimits[1] = ((float) vlimits[1])/((float) map->mxyz[iv]);

  /* Initially no contour levels set */
  nlevels = 0;
  levels  = NULL;

  /* Set precision of grid labels to be "automatic" */
  precision = 0;

  /* Display flags */
  bbox   = 1;
  grid   = 1;
  labels = 1;
  titles = 1;
  axes   = 1;

  /* Grid spacings */
  uspace = (float) 0.1;
  vspace = (float) 0.1;

  /* Default axis labels */
  strncpy(axislabel[0],"X",LEN_AXISLABEL);
  strncpy(axislabel[1],"Y",LEN_AXISLABEL);
  strncpy(axislabel[2],"Z",LEN_AXISLABEL);

  /* Default view */
  strncpy(view_orient,"lhs",4);

  /* Update the section structure with these values */
  if (!update_section_strct(section,axis,iu,iv,iaxis,first_sec,last_sec,slab_step,
			    ulimits,vlimits,frc_ulimits,frc_vlimits,nlevels,
			    levels,bbox,grid,uspace,vspace,precision,labels,
			    titles,axes,axislabel,view_orient)) {
    printf("init_section_strct: error updating section information\n");
  }
  return section;
}

/*----------------------------------------------------------------*/

/* Update values in an existing section structure
   Return 1 if successful, 0 on an error
*/
int update_section_strct(CCP4SECTION *section, char *axis, int iu, int iv,
			 int iaxis, int first_sec, int last_sec, int slab_step,
			 int ulimits[], int vlimits[], float frc_ulimits[],
			 float frc_vlimits[], int nlevels, float *levels, int bbox,
			 int grid, float uspace, float vspace, int precision,
			 int labels, int titles, int axes,
			 char axislabel[][LEN_AXISLABEL], char view_orient[])
{
  int map_upper_limit,map_lower_limit,j;
  CCP4MAP *map;

  if (CCP4MAPWISH_SECTION_DIAG) puts("update_section_strct: begun");

  if (!section) return 0;
  map = fetch_map_ptr(section->mapid);
  if (!map) return 0;

  /* Initialise the axes */
  section->axis = axis[0];
  section->iu   = iu;
  section->iv   = iv;
  section->iaxis= iaxis;

  /* Section/slab - check that they are in range
     and truncate if necessary */
  map_lower_limit = map->start[iaxis];
  map_upper_limit = map->start[iaxis] + map->ext[iaxis] - 1;
  if (first_sec < map_lower_limit)
    first_sec = map_lower_limit;
  else if (first_sec > map_upper_limit)
    first_sec = map_upper_limit;
  if (last_sec < map_lower_limit)
    last_sec = map_lower_limit;
  else if (last_sec > map_upper_limit)
    last_sec = map_upper_limit;
  

  section->nsec = first_sec;
  section->slab_end   = last_sec;
  section->slab_start = first_sec;
  section->slab_step  = slab_step;

  if (CCP4MAPWISH_SECTION_DIAG) puts("update_section_strct: done axis and section");

  /* Section limits */
  section->ulimits[0] = ulimits[0];
  section->ulimits[1] = ulimits[1];
  section->vlimits[0] = vlimits[0];
  section->vlimits[1] = vlimits[1];

  /* Fractional equivalents */
  section->frc_ulimits[0] = frc_ulimits[0];
  section->frc_ulimits[1] = frc_ulimits[1];
  section->frc_vlimits[0] = frc_vlimits[0];
  section->frc_vlimits[1] = frc_vlimits[1];

  if (CCP4MAPWISH_SECTION_DIAG) puts("update_section_strct: done section limits");

  /* Contour levels */
  section->nlevels = nlevels;
  section->levels  = levels;

  if (CCP4MAPWISH_SECTION_DIAG) puts("update_section_strct: done contour levels");

  /* Precision for grid labels */
  section->precision = precision;

  /* Display flags */
  section->bbox     = bbox;
  section->grid     = grid;
  section->labels   = labels;
  section->titles   = titles;
  section->axes     = axes;

  if (CCP4MAPWISH_SECTION_DIAG) puts("update_section_strct: done the display flags");

  /* Grid spacings */
  section->uspace   = uspace;
  section->vspace   = vspace;

  if (CCP4MAPWISH_SECTION_DIAG) puts("update_section_strct: done the grid spacings");

  /* Axis labels */
  for (j=0; j<3; j++)
    strncpy(section->axislabel[j],axislabel[j],LEN_AXISLABEL);

  if (CCP4MAPWISH_SECTION_DIAG) puts("update_section_strct: done the axis labels");

  /* View orientation */
  strncpy(section->view_orient,view_orient,4);

  if (CCP4MAPWISH_SECTION_DIAG) puts("update_section_strct: done the view orientation");

  return 1;
}

/*----------------------------------------------------------------*/

/* Remove a section pointer with handle sectionid from the map array
   Calls deletemap to free all the associated memory first
*/
int zap_section_ptr(char *sectionid)
{
  int nsections,i,j;
  CCP4SECTION *section_ptr;

  if (!sectionid || ccp4mapwish_sectionarray.nsections == 0) return 0;
  nsections = ccp4mapwish_sectionarray.nsections;
  j = nsections - 1;
  for (i=0; i<nsections; i++) {
    section_ptr = ccp4mapwish_sectionarray.section_ptr[i];
    if (section_ptr) {
      if (CCP4MAPWISH_SECTION_DIAG) 
	printf("zap_section_ptr: sectionid for section %d is \"%s\"\n",
	     i,section_ptr->sectionid);
      if (strcmp(sectionid,section_ptr->sectionid) == 0) {
	/* Free memory associated with the section which was allocated
	   here previously */
	if (section_ptr->sectionid) free(section_ptr->sectionid);
	deletesection(section_ptr);
	/* Overwrite this position with the section info from the
	   end of the array */
	if (i!=j) ccp4mapwish_sectionarray.section_ptr[i] =
		    ccp4mapwish_sectionarray.section_ptr[j];
	nsections--;
	ccp4mapwish_sectionarray.section_ptr = (CCP4SECTION **)
	  realloc(ccp4mapwish_sectionarray.section_ptr,
		  (sizeof(CCP4SECTION *))*nsections);
	ccp4mapwish_sectionarray.nsections = nsections;
	return 1;
      }
    }
  }
  printf("zap_section_ptr: sectionid \"%s\" not found\n",sectionid);
  return 1;
}

/*----------------------------------------------------------------*/

/* deletesection

   Frees all allocated memory in a CCP4SECTION structure before
   it is destroyed
*/
int deletesection(CCP4SECTION *section)
{
  if (section->mapid) free(section->mapid);
  if (section->levels) free(section->levels);
  return TCL_OK;
}


/*******************************************************************/

/* Section utility functions */

/*******************************************************************/

/* section_process_args
   
   Process the argument list passed from the Tcl interpreter
   to the section init and section config commands

   The function operates in one of two modes:
   query - a single argument is processed, the corresponding
           value is placed in the result buffer, and the
	   function exits
   config - multiple arguments are processed and the
           corresponding values are written to the section
	   structure

   This function scans through the list of arguments passed
   to the "section" command
   The arguments specify either a new value, or (if no value is
   given) a query for the current value.
*/
int section_process_args(char *mode, CCP4SECTION *section, int first_arg,
			 ClientData clientData, Tcl_Interp
			 *interp, int objc, Tcl_Obj *CONST objv[])
{
  Tcl_Obj *resultPtr;
  char buffer[200];
  char *option,*word,*endptr;
  int  i,j,iquery,iconfig,lenopt,lenword;
  CCP4MAP *map;

  /* New values for parameters */
  char *axis;
  char axislabel[3][LEN_AXISLABEL],view_orient[4];
  int  lenaxis;
  int  iaxis,iu,iv,first_sec,last_sec,slab_step,bbox,grid,labels,precision;
  int  ulimits[2],vlimits[2];
  float uspace,vspace,flimits[4],frc_ulimits[2],frc_vlimits[2];

  /* Flags for new values */
  int  new_axis=0,new_section=0,new_gridspacing=0,new_bbox=0,new_labels=0;
  int  new_grid=0,new_limits=0,new_limits_frac=0,new_precision=0;
  int  new_axis_labels=0,new_view_orient=0;
  
  /* Parameters which are not reset here */
  int  nlevels,titles,axes;
  float *levels;

  /* Identify the mode of operation */
  if (strcmp(mode,"query") == 0) {
    /* Query mode */
    if (CCP4MAPWISH_SECTION_DIAG) puts("section_process_args: query mode");
    iquery  = 1;
    iconfig = 0;
    /* Set up the result pointer */
    resultPtr = Tcl_GetObjResult(interp);
  } else if (strcmp(mode,"config") == 0) {
    /* Configuration mode */
    if (CCP4MAPWISH_SECTION_DIAG) puts("section_process_args: configuration mode");
    iquery  = 0;
    iconfig = 1;
  } else {
    printf("section_process_args: unrecognised mode \"%s\"\n",mode);
    return TCL_ERROR;
  }

  /* Need the map pointer in configuration mode */
  if (iconfig) {
    map = fetch_map_ptr(section->mapid);
    if (!map) {
      printf("section_process_args: parent map \"%s\" not found\n",section->mapid);
      return TCL_ERROR;
    }
  }

  /* Check we have a valid section pointer */
  if (!section) {
    puts("section_process_args: section pointer is null");
    return TCL_ERROR;
  }

  /* Loop over the arguments in objv[] */
  i = first_arg;
  while (i < objc) {

    /* Fetch the option */
    option = Tcl_GetStringFromObj(objv[i],&lenopt);
    if (CCP4MAPWISH_SECTION_DIAG) 
      printf("section_process_args: option %d is \"%s\"\n",i,option);

    /* Match against possible keywords */

    if (strcmp(option,"-axis") == 0) {
      if (CCP4MAPWISH_SECTION_DIAG) 
	printf("section_process_args: -axis\n");
      /* Axis keyword */
      if (iquery) {
	sprintf(buffer,"%c",section->axis);
	Tcl_SetStringObj(resultPtr,buffer,strlen(buffer));
	return TCL_OK;
      } else if (iconfig) {
	if (++i < objc) {
	  axis = Tcl_GetStringFromObj(objv[i],&lenaxis);
	  if (get_axis_id(axis,&iaxis,&iu,&iv)) {
	    if (CCP4MAPWISH_SECTION_DIAG) 
	      printf("section_process_args: axis is \"%s\" (%d), iu=%d iv=%d\n",
		     axis,iaxis,iu,iv);
	    new_axis = 1;
	  } else {
	    printf("section_process_args: axis \"%s\" unrecognised\n",axis);
	    return TCL_ERROR;
	  }
	  if (CCP4MAPWISH_SECTION_DIAG) 
	    printf("section_process_args: new axis is \"%s\"\n",axis);
	} else {
	  puts("section_process_args: insufficient arguments for -axis");
	  return TCL_ERROR;
	}
      }

    } else if (strcmp(option,"-section") == 0) {
      if (CCP4MAPWISH_SECTION_DIAG) printf("section_process_args: -section\n");
      /* Section keyword*/
      if (iquery) {
	if (section->slab_start != section->slab_end) {
	  sprintf(buffer,"%d %d %d",section->slab_start,section->slab_end,
		  section->slab_step);
	} else {
	  sprintf(buffer,"%d",section->nsec);
	}
	Tcl_SetStringObj(resultPtr,buffer,strlen(buffer));
	return TCL_OK;
      } else if (iconfig) {
	if (++i < objc) {
	  /* Try to extract first section */
	  word = Tcl_GetStringFromObj(objv[i],&lenword);
	  /* Extract a long int using strtol
	     If no conversion takes place then "endptr" will be equal
	     to "word" */
	  first_sec = (int) strtol(word,&endptr,10);
	  if (strcmp(word,endptr) == 0) {
	    if (CCP4MAPWISH_SECTION_DIAG) 
	      printf("section_process_args: bad section number \"%s\"\n",word);
	    return TCL_ERROR;
	  }
	  new_section = 1;
	  slab_step = section->slab_step;
	  last_sec = first_sec;
	  /* Look for possible next section number, for a slab
	     Use a similar technique but assume that an error
	     just means we've found the next keyword */
	  if (++i < objc) {
	    word = Tcl_GetStringFromObj(objv[i],&lenword);
	    last_sec = (int) strtol(word,&endptr,10);
	    if (strcmp(word,endptr) == 0) {
	      last_sec = first_sec;
	      i--;
	    } else {
	      /* Look for a possible slab_step */
	      if (++i < objc) {
		word = Tcl_GetStringFromObj(objv[i],&lenword);
		slab_step = (int) strtol(word,&endptr,10);
		if (strcmp(word,endptr) == 0) {
		  slab_step = section->slab_step;
		  i--;
		}
	      }
	    }
		
	  }
	  if (CCP4MAPWISH_SECTION_DIAG) 
	    printf("section_process_args: section is %d (%d)\n",first_sec,last_sec);
	} else {
	  printf("section_process_args: insufficient arguments for -section\n");
	  return TCL_ERROR;
	}
      }

    } else if (strcmp(option,"-limits") == 0) {
      if (CCP4MAPWISH_SECTION_DIAG) printf("section_process_args: -limits\n");
      /* Limits keyword */
      if (iquery) {
	sprintf(buffer,"%d %d %d %d",section->ulimits[0],section->ulimits[1],
		section->vlimits[0],section->vlimits[1]);
	Tcl_SetStringObj(resultPtr,buffer,strlen(buffer));
	return TCL_OK;
      } else if (iconfig) {
	/* Should be a list of four numbers
	   If all four are less than 1 then they are assumed to
	   specify fractional limits, otherwise they are assumed to
	   be in grid units */
	for (j=0; j<4; j++) {
	  i++;
	  if (i < objc) {
	    word = Tcl_GetStringFromObj(objv[i],&lenword);
	    flimits[j] = (float) strtod(word,&endptr);
	    if (strcmp(word,endptr) == 0) {
	      printf("section_process_args: bad limit \"%s\"\n",word);
	      return TCL_ERROR;
	    }
	  } else {
	    printf("section_process_args: insufficient arguments for -limits\n");
	    return TCL_ERROR;
	  }
	}
	/* Check whether we have fractional or grid limits supplied */
	if ((fabs(flimits[0]) <= 1.0 && fabs(flimits[1]) <= 1.0)
	    && (fabs(flimits[2]) <= 1.0) && (fabs(flimits[3]) <= 1.0)) {
	  /* Assume fractional limits */
	  new_limits_frac = 1;
	  if (CCP4MAPWISH_SECTION_DIAG) 
	    printf("section_process_args: limits are supplied in fractional units\n");
	}
	new_limits = 1;
	if (CCP4MAPWISH_SECTION_DIAG) 
	  printf("section_process_args: floating limits are %f-%f, %f-%f\n",
		 flimits[0],flimits[1],flimits[2],flimits[3]);
      }  

    } else if (strcmp(option,"-bbox") == 0) {
      if (CCP4MAPWISH_SECTION_DIAG) printf("section_process_args: -bbox\n");
      /* Bbox keyword */
      if (iquery) {
	Tcl_SetIntObj(resultPtr,section->bbox);
	return TCL_OK;
      } else if (iconfig) {
	if (i++ < objc) {
	  word = Tcl_GetStringFromObj(objv[i],&lenword);
	  bbox = ((int) strtol(word,&endptr,10) != 0);
	  if (strcmp(word,endptr) == 0) {
	    printf("section_process_args: bad argument to -bbox \"%s\"\n",word);
	    return TCL_ERROR;
	  } else {
	    if (CCP4MAPWISH_SECTION_DIAG)
	      printf("section_process_args: bbox is %d\n",bbox);
	    new_bbox = 1;
	  }
	} else {
	  printf("section_process_args: insufficient arguments for -bbox\n");
	  return TCL_ERROR;
	}
      }

    } else if (strcmp(option,"-grid") == 0) {
      if (CCP4MAPWISH_SECTION_DIAG) printf("section_process_args: -grid\n");
      /* Grid keyword */
      if (iquery) {
	Tcl_SetIntObj(resultPtr,section->grid);
	return TCL_OK;
      } else if (iconfig) {
	if (++i < objc) {
	  word = Tcl_GetStringFromObj(objv[i],&lenword);
	  grid = ((int) strtol(word,&endptr,10) != 0);
	  if (strcmp(word,endptr) == 0) {
	      printf("section_process_args: bad argument to -grid \"%s\"\n",word);
	    return TCL_ERROR;
	  } else {
	    if (CCP4MAPWISH_SECTION_DIAG) 
	      printf("section_process_args: grid is %d\n",grid);
	    new_grid = 1;
	  }
	} else {
	  printf("section_process_args: insufficient arguments for -grid\n");
	  return TCL_ERROR;
	}
      }

    } else if (strcmp(option,"-labels") == 0) {
      if (CCP4MAPWISH_SECTION_DIAG) printf("section_process_args: -labels\n");
      /* Labels keyword */
      if (iquery) {
	Tcl_SetIntObj(resultPtr,section->labels);
	return TCL_OK;
      } else if (iconfig) {
	if (++i < objc) {
	  word = Tcl_GetStringFromObj(objv[i],&lenword);
	  labels = ((int) strtol(word,&endptr,10) != 0);
	  if (strcmp(word,endptr) == 0) {
	    printf("section_process_args: bad argument to -labels \"%s\"\n",word);
	    return TCL_ERROR;
	  } else {
	    if (CCP4MAPWISH_SECTION_DIAG) 
	      printf("section_process_args: labels is %d\n",labels);
	    new_labels = 1;
	  }
	} else {
	  printf("section_process_args: insufficient arguments for -labels\n");
	  return TCL_ERROR;
	}
      }

    } else if (strcmp(option,"-gridspacing") == 0) {
      if (CCP4MAPWISH_SECTION_DIAG) printf("section_process_args: -gridspacing\n");
      /* Gridspacing keyword */
      if (iquery) {
	/* Return current settings as a list */
	sprintf(buffer,"%f %f",section->uspace,section->vspace);
	Tcl_SetStringObj(resultPtr,buffer,strlen(buffer));
	return TCL_OK;
      } else if (iconfig) {
	/* Should be a list of two floats */
	if (i+2 < objc) {
	  i++;
	  word = Tcl_GetStringFromObj(objv[i],&lenword);
	  uspace = ((float) strtod(word,&endptr));
	  if (strcmp(word,endptr) == 0) {
	    printf("section_process_args: bad argument to -gridspacing \"%s\"\n",word);
	    return TCL_ERROR;
	  }
	  i++;
	  word = Tcl_GetStringFromObj(objv[i],&lenword);
	  vspace = ((float) strtod(word,&endptr));
	  if (strcmp(word,endptr) == 0) {
	    printf("section_process_args: bad argument to -gridspacing \"%s\"\n",word);
	    return TCL_ERROR;
	  }
	} else {
	  printf("section_process_args: insufficient arguments for -gridspacing\n");
	  return TCL_ERROR;
	}
	new_gridspacing = 1;
	if (CCP4MAPWISH_SECTION_DIAG) 
	  printf("section_process_args: grid spacing is %f %f\n",uspace,vspace);
      }

    } else if (strcmp(option,"-precision") == 0) {
      if (CCP4MAPWISH_SECTION_DIAG) printf("section_process_args: -precision\n");
      /* Precision keyword */
      if (iquery) {
	/* Return current setting */
	Tcl_SetIntObj(resultPtr,section->precision);
	return TCL_OK;
      } else if (iconfig) {
	/* Should be followed by an integer */
	i++;
	if (i < objc) {
	  word = Tcl_GetStringFromObj(objv[i],&lenword);
	  precision = ((int) strtol(word,&endptr,10));
	  if (strcmp(word,endptr) == 0) {
	    printf("section_process_args: bad argument to -precision \"%s\"\n",word);
	    return TCL_ERROR;
	  }
	  /* If less than zero then set to zero i.e. no truncation is
	     performed */
	  if (precision < 0) precision = 0;
	} else {
	  printf("section_process_args: insufficient arguments for -precision\n");
	  return TCL_ERROR;
	}
	new_precision = 1;
	if (CCP4MAPWISH_SECTION_DIAG) 
	  printf("section_process_args: precision is %d\n",precision);
      }

    } else if (strcmp(option,"-axislabels") == 0) {
      if (CCP4MAPWISH_SECTION_DIAG) 
	printf("section_process_args: -axislabels\n");

      /* Axislabels keyword */
      if (iquery) {
	sprintf(buffer,"%s %s %s",
		section->axislabel[0],
		section->axislabel[1],
		section->axislabel[2]);
	Tcl_SetStringObj(resultPtr,buffer,strlen(buffer));
	return TCL_OK;
      } else if (iconfig) {
	/* Should be a list of three strings */
	for (j=0; j<3; j++) {
	  if (++i < objc) {
	    word = Tcl_GetStringFromObj(objv[i],&lenword);
	    strncpy(axislabel[j],word,LEN_AXISLABEL);
	    if (lenword >= LEN_AXISLABEL)
	      axislabel[j][LEN_AXISLABEL-1] = '\0';
	  } else {
	    printf("section_process_args: insufficient arguments for -axislabels\n");
	    return TCL_ERROR;
	  }
	}
	if (CCP4MAPWISH_SECTION_DIAG) {
	  printf("section_process_args: axis labels are");
	  for (j=0; j<3; j++) printf(" \"%s\"",axislabel[j]);
	  printf("\n");
	}
	new_axis_labels = 1;
      }

    } else if (strcmp(option,"-view") == 0) {
      if (CCP4MAPWISH_SECTION_DIAG) 
	printf("section_process_args: -view\n");

      /* View keyword */
      if (iquery) {
	sprintf(buffer,"%s",section->view_orient);
	Tcl_SetStringObj(resultPtr,buffer,strlen(buffer));
	return TCL_OK;
      } else if (iconfig) {
	/* Should be a single string argument ("lhs" or "rhs") */
	if (++i < objc) {
	  word = Tcl_GetStringFromObj(objv[i],&lenword);
	  if (strcmp(word,"lhs") == 0 || strcmp(word,"rhs") == 0) {
	    strncpy(view_orient,word,4);
	    if (lenword >= 4) view_orient[3] = '\0';
	  } else {
	    printf("section_process_args: unrecognised value for view \"%s\"\n",word);
	    return TCL_ERROR;
	  }
	} else {
	  printf("section_process_args: insufficient arguments for -view\n");
	  return TCL_ERROR;
	}
      
	if (CCP4MAPWISH_SECTION_DIAG)
	  printf("section_process_args: view orientation is \"%s\"\n",
		 view_orient);
	new_view_orient = 1;
      }

    } else {
      /* Unrecognised keyword */
      printf("section_process_args: unrecognised keyword \"%s\"\n",option);
      return TCL_ERROR;
    }

    /* Increment counter to examine next argument */
    i++;
  }
  /* No more arguments to process */
  if (CCP4MAPWISH_SECTION_DIAG)
    printf("section_process_args: finished processing options\n");

  /* Check and assign new values */
  if (!new_axis) {
    axis  = &section->axis;
    iaxis = section->iaxis;
    iu    = section->iu;
    iv    = section->iv;
  } else {
    /* Reset the limits to be the full section from the
       map unless overridden by user input */
    if (!new_limits) {
      flimits[0] = (float) map->start[iu];
      flimits[1] = (float) flimits[0] + map->ext[iu] - 1;
      flimits[2] = (float) map->start[iv];
      flimits[3] = (float) flimits[2] + map->ext[iv] - 1;
      new_limits = 1;
      if (CCP4MAPWISH_SECTION_DIAG) 
	printf("section_process_args: change of axis, setting limits to %f-%f, %f-%f\n",
	       flimits[0],flimits[1],flimits[2],flimits[3]);
    }
  }
  if (!new_section) {
    first_sec = section->slab_start;
    last_sec  = section->slab_end;
    slab_step = section->slab_step;
  }
  if (!new_limits) {
    ulimits[0] = section->ulimits[0];
    ulimits[1] = section->ulimits[1];
    vlimits[0] = section->vlimits[0];
    vlimits[1] = section->vlimits[1];
    frc_ulimits[0] = ((float) ulimits[0])/((float) map->mxyz[iu]);
    frc_ulimits[1] = ((float) ulimits[1])/((float) map->mxyz[iu]);
    frc_vlimits[0] = ((float) vlimits[0])/((float) map->mxyz[iv]);
    frc_vlimits[1] = ((float) vlimits[1])/((float) map->mxyz[iv]);
  } else {
    /* If limits are fractional then do a conversion
       to grid units first */
    if (new_limits_frac) {
      frc_ulimits[0] = flimits[0];
      frc_ulimits[1] = flimits[1];
      frc_vlimits[0] = flimits[0];
      frc_vlimits[1] = flimits[1];
      flimits[0] = flimits[0]*((float) map->mxyz[iu]);
      flimits[1] = flimits[1]*((float) map->mxyz[iu]);
      flimits[2] = flimits[2]*((float) map->mxyz[iv]);
      flimits[3] = flimits[3]*((float) map->mxyz[iv]);
    }
    ulimits[0] = (int) flimits[0];
    ulimits[1] = (int) flimits[1];
    vlimits[0] = (int) flimits[2];
    vlimits[1] = (int) flimits[3];
    frc_ulimits[0] = flimits[0]/((float) map->mxyz[iu]);
    frc_ulimits[1] = flimits[1]/((float) map->mxyz[iu]);
    frc_vlimits[0] = flimits[2]/((float) map->mxyz[iv]);
    frc_vlimits[1] = flimits[3]/((float) map->mxyz[iv]);
    /* Check these limits are in range, reset any which aren't */
    if (ulimits[0] < map->start[iu]) {
      ulimits[0]     = map->start[iu];
      frc_ulimits[0] = ((float) ulimits[0])/((float) map->mxyz[iu]);
    }
    if (ulimits[1] > map->start[iu] + map->ext[iu] - 1) {
      ulimits[1]     = map->start[iu] + map->ext[iu] - 1;
      frc_ulimits[1] =((float) ulimits[1])/((float) map->mxyz[iu]);
    }
    if (vlimits[0] < map->start[iv]) {
      vlimits[0]     = map->start[iv];
      frc_vlimits[0] =((float) vlimits[0])/((float) map->mxyz[iv]);
    }
    if (vlimits[1] > map->start[iv] + map->ext[iv] - 1) {
      vlimits[1]     = map->start[iv] + map->ext[iv] - 1;
      frc_vlimits[1] =((float) vlimits[1])/((float) map->mxyz[iv]);
    }
  }
  if (!new_gridspacing) {
    uspace = section->uspace;
    vspace = section->vspace;
  }
  if (!new_precision) precision = section->precision;
  if (!new_bbox) bbox = section->bbox;
  if (!new_grid) grid = section->grid;
  if (!new_labels) labels = section->labels;

  if (!new_axis_labels) {
    for (j=0; j<3; j++)
      strncpy(axislabel[j],section->axislabel[j],LEN_AXISLABEL);
  }

  /* View orientation */
  if (!new_view_orient) strncpy(view_orient,section->view_orient,4);

  /* Other parameters not affected here */
  nlevels = section->nlevels;
  levels  = section->levels;
  titles  = section->titles;
  axes    = section->axes;

  if (CCP4MAPWISH_SECTION_DIAG) printf("section_process_args: about to do update\n");

  /* Update the section structure with these values */
  if (!update_section_strct(section,axis,iu,iv,iaxis,first_sec,last_sec,slab_step,
			    ulimits,vlimits,frc_ulimits,frc_vlimits,
			    nlevels,levels,bbox,grid,uspace,vspace,precision,
			    labels,titles,axes,axislabel,view_orient)) {
    printf("section_process_args: unable to update section structure\n");
    return TCL_ERROR;
  }
  return TCL_OK;
}

/*----------------------------------------------------------------*/

/* section_process_contour_list_args
   
   Process the argument list passed from the Tcl interpreter
   to the section ... contours ... list command

   The function reads each argument (starting with first_arg) and
   tries to extract a float value which is then added to the list of
   values.

   The function returns upon encountering the end of the list, or an
   argument which cannot be converted. It returns the index number for
   the next argument after the last successfully processed argument, or
   zero if nothing could be processed.
*/
int section_process_contour_list_args(ClientData clientData, Tcl_Interp
				      *interp, int objc, Tcl_Obj *CONST objv[],
				      int first_arg, int *nvalues,
				      float **values_ptr)
{
  char  *word,*endptr;
  int   i,lenword;
  float value,*values;

  *nvalues = 0;
  values = NULL;

  for (i=first_arg; i<objc; i++) {
    word = Tcl_GetStringFromObj(objv[i],&lenword);
    value = (float) strtod(word,&endptr);
    if (strcmp(word,endptr) == 0) {
      /* Couldn't convert this argument to a number */
      if (i==first_arg) {
	*values_ptr = values;
	return 0;
      }
      *values_ptr = values;
      return i;
    }
    if (CCP4MAPWISH_SECTION_DIAG)
      printf("section_process_contour_list_args: value=%f\n",value);
    /* Got a value - add to the list */
    (*nvalues)++;
    values = (float *) realloc(values,(*nvalues)*sizeof(float));
    if (!values) {
      printf("section_process_contour_list_args: couldn't assign memory\n");
      *values_ptr = values;
      return 0;
    }
    values[(*nvalues)-1] = value;
    if (CCP4MAPWISH_SECTION_DIAG) 
      printf("section_process_contour_list_args: nvalues=%d, value_list=%f\n",
	     *nvalues,values[(*nvalues)-1]);
  }
  *values_ptr = values;
  return i;
}

/*----------------------------------------------------------------*/

/* section_process_contour_range_args
   
   Process the argument list passed from the Tcl interpreter
   to the section ... contours ... range command

   The function reads three arguments (starting with first_arg) and
   tries to extract a float value.
   The three floats represent a minimum, maximum and interval value
   respectively.

   The function returns upon extracting three numbers, or if there
   was an error (either an argument which cannot be converted, or running
   out of arguments). It returns the index number for the next argument
   after the last successfully processed argument, or zero if there was
   an error.
*/
int section_process_contour_range_args(ClientData clientData, Tcl_Interp
				       *interp, int objc, Tcl_Obj *CONST objv[],
				       int first_arg, int *nvalues,
				       float **values_ptr)
{
  char  *word,*endptr;
  int   i,lenword;
  float min,max,interval,value,*values;

  *nvalues = 0;
  values = NULL;

  if (first_arg+2 < objc) {
    /* Minimum */
    i=first_arg;
    word = Tcl_GetStringFromObj(objv[i],&lenword);
    min = (float) strtod(word,&endptr);
    if (strcmp(word,endptr) == 0) { 
      *values_ptr = values;
      return 0;
    }
    /* Maximum */
    i++;
    word = Tcl_GetStringFromObj(objv[i],&lenword);
    max = (float) strtod(word,&endptr);
    if (strcmp(word,endptr) == 0) {
      *values_ptr = values;
      return 0;
    }
    /* Interval */
    i++;
    word = Tcl_GetStringFromObj(objv[i],&lenword);
    interval = (float) strtod(word,&endptr);
    if (strcmp(word,endptr) == 0) return 0;
    /* Make a list of values */
    value = min;
    while (value <= max) {
      (*nvalues)++;
      values = (float *) realloc(values,(*nvalues)*sizeof(float));
      if (!values) {
	printf("section_process_contour_range_args: couldn't assign memory\n");
	*values_ptr = values;
	return 0;
      }
      values[(*nvalues)-1] = value;
      value = value + interval;
    }
    *values_ptr = values;
    /* Return next argument index*/
    return ++i;
  }
  return 0;
}

/*----------------------------------------------------------------*/
