/*
     ccp4mapwish_utils.c: miscellaneous transformation utilities
     Copyright (C) 2001  CCRLC, Peter Briggs

     This code is distributed under the terms and conditions of the
     CCP4 Program Suite Licence Agreement as a CCP4 Application.
     A copy of the CCP4 licence can be obtained by writing to the
     CCP4 Secretary, Daresbury Laboratory, Warrington WA4 4AD, UK.
*/

/*******************************************************************/

/* ccp4mapwish_utils.c */

/*******************************************************************/

/* Various miscellaneous utility functions for ccp4mapwish
   for getting scale and inverse scale matrices, matrix inversion,
   transforming coords etc
 */

/*******************************************************************/

/* Include header file */

/*******************************************************************/

#include "ccp4mapwish.h"

/*******************************************************************/

/* Utility functions */

/*******************************************************************/

/* get_scale_matrix

   Given the cell dimensions a b c and angles alpha beta gamma plus the
   orth code, generate the fractional->orthogonal and orthogonal->fractional
   transformation matrices.

   Return 1 on success, 0 if there was an error.
*/
int get_scale_matrix(float cell[6], int iorth, float frac_to_orth[3][3],
		     float orth_to_frac[3][3])
{
  int    i,j;
  double a,b,c,alpha,beta,gamma,conv,term;
  double sina,cosa,sinb,cosb,sing,cosg;
  double sinas,cosas,cosbs,cosgs;

  /* Only deal with orth code 1 for now... */
  if (iorth!=1) {
    puts("get_scale_matrix: cannot deal with orth code other than 1");
    return 0;
  }

  /*
    printf("get_scale_matrix: cell = %f %f %f (%f %f %f)\n",
    cell[0],cell[1],cell[2],cell[3],cell[4],cell[5]);
  */

  /* Initialisations */
  for (i=0; i<6; i++) {
    if (cell[i]<=0.0) {
      printf("get_scale_matrix: cell parameter %d (%f) is less than zero\n",i,cell[i]);
      return 0;
    }
  }

  /* Cell edges */
  a = (double) cell[0];
  b = (double) cell[1];
  c = (double) cell[2];
  
  /* Cell angles are supplied in degrees but computers like
     radians, so convert */
  conv = 3.14159/180.0;
  
  alpha = (double) cell[3]*conv;
  beta  = (double) cell[4]*conv;
  gamma = (double) cell[5]*conv;

  /* Sines and cosines of angles */
  sina = sin(alpha);
  cosa = cos(alpha);
  sinb = sin(beta);
  cosb = cos(beta);
  sing = sin(gamma);
  cosg = cos(gamma);

  /* Other useful quantities */
  cosas = (cosg*cosb-cosa)/(sinb*sing);
  term  = 1.0-cosas*cosas;
  if (term < 0.0) {
    printf("get_scale_matrix: warning - term is < 0.0 (%lf) at position 1\n",
	   term);
    term = (float) 0.0;
  }
  sinas = sqrt(term);
  cosbs = (cosa*cosg-cosb)/(sina*sing);
  cosgs = (cosa*cosb-cosg)/(sina*sinb);

  /* Zero the matrices */
  for (i=0; i<3; i++) {
    for (j=0; j<3; j++) {
      frac_to_orth[i][j] = (float) 0.0;
      orth_to_frac[i][j] = (float) 0.0;
    }
  }

  /* Calculate the orthogonalising matrix */

  /* ORTHCODE 1
     XO along a, ZO along c*
  */
  if (iorth == 1) {
    frac_to_orth[0][0] = (float) a;
    frac_to_orth[0][1] = (float) b*cosg;
    frac_to_orth[0][2] = (float) c*cosb;
    frac_to_orth[1][1] = (float) b*sing;
    frac_to_orth[1][2] = (float) -c*sinb*cosas;
    frac_to_orth[2][2] = (float) c*sinb*sinas;
  }

  /* Other orthcodes here (2...6) */

  /*
    printf("get_scale_matrix: orthogonalising matrix:\n");
    for (i=0; i<3; i++) {
    printf("         ");
    for (j=0; j<3; j++) {
    printf("%f ",frac_to_orth[i][j]);
    }
    printf("\n");
    }
    puts("get_scale_matrix: inverting orthogonalising matrix");
  */

  /* Calculate inverse */
  if (!invert_3matrix(frac_to_orth,orth_to_frac)) {
    puts("get_scale_matrix: error inverting orthogonalising matrix");
    return 0;
  }

  /*
    printf("get_scale_matrix: fractionalising matrix:\n");
    for (i=0; i<3; i++) {
    printf("         ");
    for (j=0; j<3; j++) {
    printf("%f ",orth_to_frac[i][j]);
    }
    printf("\n");
    }
  */
  return 1;
}

/*------------------------------------------------------------------------------*/

/* invert_3matrix

   Given a 3x3 matrix m, calculate the inverse matrix invm
   Return 1 for success, 0 if there was an error
*/
int invert_3matrix(float m[3][3], float invm[3][3])
{
  int   i,j,k,ii,jj,i1,j1;
  float determinant,sign,sum,eps=(float) 1.0e-8;
  float x[2][2],c[3][3],ident[3][3];
  double index;

  if (CCP4MAPWISH_UTILS_DIAG) {
    printf("invert_3matrix: starting\n");
    printf("invert_3matrix: input matrix:\n   %f  %f  %f\n   %f  %f  %f\n   %f  %f  %f\n",
	   m[0][0],m[0][1],m[0][2],
	   m[1][0],m[1][1],m[1][2],
	   m[2][0],m[2][1],m[2][2]);
  }

  /* Calculate the determinant */
  determinant = m[0][0]*(m[1][1]*m[2][2]-m[1][2]*m[2][1])
    - m[1][0]*(m[0][1]*m[2][2]-m[0][2]*m[2][1])
    + m[2][0]*(m[0][1]*m[1][2]-m[0][2]*m[1][1]);

  if (CCP4MAPWISH_UTILS_DIAG) printf("invert_3matrix: determinant = %f\n",determinant);

  if (fabsf(determinant) < eps) {
    puts("invert_3matrix: determinant is too small, cannot generate inverse");
    return 0;
  }

  /* Create adjoint matrix
     This is the transposed cofactor matrix */
  for (i=0; i<3; i++) {
    for (j=0; j<3; j++) {
      /* Get cofactor for element i,j */
      i1 = 0;
      for (ii=0; ii<3; ii++) {
	/* Ignore elements in the same row as the
	   current element */
	if (ii != i) {
	  j1 = 0;
	  for (jj=0; jj<3; jj++) {
	    /* Ignore elements in the same column */
	    if (jj != j) {
	      x[i1][j1] = m[ii][jj];
	      j1++;
	    }
	  }
	  i1++;
	}
      }
      /* Set the correct sign */
      index = (double) (i+j);
      if ((int) fmod(index,2.0)) {
	sign = -1.0;
      } else {
	sign = 1.0;
      }
      /* Calculate the cofactor and place in the transposed position */
      c[j][i] = sign * (x[0][0]*x[1][1] - x[0][1]*x[1][0]);

      /*
	if (CCP4MAPWISH_UTILS_DIAG) {
	printf("invert_3matrix: element %d %d\n",i,j);
	printf("                Associated 2-matrix should be:\n");
	printf("                %f   %f\n                %f   %f\n",
	x[0][0],x[0][1],x[1][0],x[1][1]);
	printf("                Index = %lf, int (fmod(...,2.0) = %d\n",
	index,((int) fmod(index,2.0)));
	printf("                Sign is %f\n",sign);
	printf("                Determinant of 2-matrix is %f\n",
	(x[0][0]*x[1][1] - x[0][1]*x[1][0]));
	printf("                c = %f\n",c[i][j]);
	}
      */
    }
  }

  if (CCP4MAPWISH_UTILS_DIAG) {
    printf("invert_3matrix: cofactor matrix:\n   %f  %f  %f\n   %f  %f  %f\n   %f  %f  %f\n",
	   c[0][0],c[0][1],c[0][2],
	   c[1][0],c[1][1],c[1][2],
	   c[2][0],c[2][1],c[2][2]);
  }
  
  /* Generate inverse */
  for (i=0; i<3; i++) {
    for (j=0; j<3; j++) {
      invm[i][j] = c[i][j]/determinant;
    }
  }
  
  /* Check! */
  for (i=0; i<3; i++) {
    for (j=0; j<3; j++) {
      sum = 0.0;
      for (k=0; k<3; k++) {
	sum = sum + m[i][k]*invm[k][j];
      }
      ident[i][j] = sum;
    }
  }
  
  if (CCP4MAPWISH_UTILS_DIAG) {
    printf("invert_3matrix: output matrix:\n   %f  %f  %f\n   %f  %f  %f\n   %f  %f  %f\n",
	   invm[0][0],invm[0][1],invm[0][2],
	   invm[1][0],invm[1][1],invm[1][2],
	   invm[2][0],invm[2][1],invm[2][2]);
    printf("invert_3matrix: check matrix:\n   %f  %f  %f\n   %f  %f  %f\n   %f  %f  %f\n",
	   ident[0][0],ident[0][1],ident[0][2],
	   ident[1][0],ident[1][1],ident[1][2],
	   ident[2][0],ident[2][1],ident[2][2]);
  }

  /* Check! */
  for (i=0; i<3; i++) {
    for (j=0; j<3; j++) {
      if (i == j) {
	if ((fabs(ident[i][j])-1.0) > 0.001) {
	  printf("invert_3matrix: error in matrix inversion\nDiagonal element of identity matrix (%d,%d) = %f\n",i,j,ident[i][j]);
	  return 0;
	}
      } else {
	if (fabs(ident[i][j]) > 0.001) {
	  printf("invert_3matrix: error in matrix inversion\nOff-diagonal element of identity matrix (%d,%d) = %f\n",i,j,ident[i][j]);
	  return 0;
	}
      }
    }
  }
  return 1;
}

/*------------------------------------------------------------------------------*/

/* transform_coordinates

   Given two lists of floats x and y, representing npts pairs of x and y
   coordinates in grid units, convert these to orthogonal (= canvas)
   coordinates.

   Returns 1 on success, 0 if there was an error
*/
int transform_coordinates(CCP4SECTION *section, int npts, float *x, float *y)
{
  int   i,iu,iv,iorient;
  float u=0,v=0,scale_x,scale_y;
  CCP4MAP *map;

  if (CCP4MAPWISH_UTILS_DIAG) puts("transform_coordinates: starting");

  /* Check for valid section */
  if (!section) return 0;

  /* Find parent map */
  map = fetch_map_ptr(section->mapid);
  if (!map) return 0;

  /* Fetch axis indices */
  iu = section->iu;
  iv = section->iv;

  /* Determine orientation
     0 = lhs view
     1 = rhs view */
  iorient = strcmp(section->view_orient,"lhs");

  /* Scaling from grid to fractional */
  scale_x = (float) map->mxyz[iu];
  scale_y = (float) map->mxyz[iv];

  if (CCP4MAPWISH_UTILS_DIAG) {
    printf("transform_coordinates: iu,iv = %d,%d\ntransform_coordinates: scale x,y = %f,%f\n",iu,iv,scale_x,scale_y);
    printf("transform_coordinates: frac_to_orth matrix is %f %f\ntransform_coordinates:                        %f %f\n",map->frac_to_orth[iu][iu],map->frac_to_orth[iu][iv],map->frac_to_orth[iv][iu],map->frac_to_orth[iv][iv]);
  }

  /* Loop over all coordinate pairs */
  for (i=0; i<npts; i++) {
  
    /* First generate fractional coordinates */
    if (CCP4MAPWISH_UTILS_DIAG) {
      printf("transform_coordinates: point #%d - scaling\n",i);
      printf("transform_coordinates: initial x,y = %f,%f\n",x[i],y[i]);
      printf("transform_coordinates: scalefactors are x,y = %f,%f\n",scale_x,scale_y);
    }
    x[i] = x[i]/scale_x;
    y[i] = y[i]/scale_y;
    if (CCP4MAPWISH_UTILS_DIAG) printf("transform_coordinates: scaled x,y = %f,%f\n",
				     x[i],y[i]);

    /* Transform to orthogonal/plotter coordinates */
    u = map->frac_to_orth[iu][iu]*x[i] + map->frac_to_orth[iu][iv]*y[i];
    v = map->frac_to_orth[iv][iu]*x[i] + map->frac_to_orth[iv][iv]*y[i];

    /* If orientation is rhs (iorient == 1) then flip the axes */
    if (iorient == 0) {
      x[i] = u;
      y[i] = v;
    } else {
      x[i] = v;
      y[i] = u;
    } 

    if (CCP4MAPWISH_UTILS_DIAG)
      printf("transform_coordinates: transformed x,y = %f,%f\n",x[i],y[i]);
  }
  return 1;
}

/*------------------------------------------------------------------------------*/

/* inv_transform_coordinates

   Given coordinates x,y in canvas units (corrected to account for any
   offset in displaying the section) return the equivalent grid, fractional
   and orthogonal coordinates

   Return 1 on success, 0 if there was an error
*/

int inv_transform_coordinates(CCP4SECTION *section, int xcanvas, int ycanvas,
			      int *xgrid, int *ygrid,
			      float *xfrac, float *yfrac,
			      float *xorth, float *yorth,
			      float scale)
{
  int   iu,iv,iorient;
  float x,y,u=0,v=0,scale_x,scale_y;
  CCP4MAP *map;

  /* Check for valid section pointer */
  if (!section) {
    puts("inv_transform_coordinates: serious error, section pointer is null!");
    return 0;
  }

  /* Fetch map pointer */
  map = fetch_map_ptr(section->mapid);
  if (!map) {
    puts("inv_transform_coordinates: serious error, parent map pointer is null!");
    return 0;
  }

  if (CCP4MAPWISH_UTILS_DIAG)
    printf("inv_transform_coordinates: initial values %d %d\n",xcanvas,ycanvas);
  if (CCP4MAPWISH_UTILS_DIAG)
    printf("inv_transform_coordinates: scale %f\n",scale);

  /* Fetch axis indices */
  iu = section->iu;
  iv = section->iv;
  if (CCP4MAPWISH_UTILS_DIAG)
    printf("inv_transform_coordinates: fetched axis indices %d %d\n",iu,iv);

  /* Determine orientation
     0 = lhs view
     1 = rhs view */
  iorient = strcmp(section->view_orient,"lhs");

  /* Scaling from grid to fractional */
  scale_x = (float) map->mxyz[iu];
  scale_y = (float) map->mxyz[iv];
  if (CCP4MAPWISH_UTILS_DIAG)
    printf("inv_transform_coordinates: collected scaling %f %f\n",scale_x,scale_y);

  /* Transform from orthogonal/plotter coordinates
     If orientation is rhs (iorient == 1) then flip the axes */
  if (iorient == 0) {
    /* lhs view */
    x = ((float) xcanvas)/((float) scale); 
    y = ((float) ycanvas)/((float) scale);
  } else {
    /* rhs view */
    x = ((float) ycanvas)/((float) scale); 
    y = ((float) xcanvas)/((float) scale);
  }
 
  u = map->orth_to_frac[iu][iu]*x + map->orth_to_frac[iu][iv]*y;
  v = map->orth_to_frac[iv][iu]*x + map->orth_to_frac[iv][iv]*y;

  if (CCP4MAPWISH_UTILS_DIAG)
    printf("inv_transform_coordinates: transformed to fractional coords %f %f\n",u,v);

  /* Pseudo-fractional coordinates to grid coordinates */
  u = u*scale_x;
  v = v*scale_y;
  if (CCP4MAPWISH_UTILS_DIAG)
    printf("inv_transform_coordinates: transformed to grid coords %f %f\n",u,v);

  /* Shift origin in grid coords */
  u = u + (float) section->ulimits[0];
  v = v + (float) section->vlimits[0];
  *xgrid = (int) u;
  *ygrid = (int) v;
  if (CCP4MAPWISH_UTILS_DIAG)
    printf("inv_transform_coordinates: true grid coords %d %d\n",*xgrid,*ygrid);

  /* Generate true fractional coordinates */
  u = ((float) *xgrid)/scale_x;
  v = ((float) *ygrid)/scale_y;
  *xfrac = u;
  *yfrac = v;
  if (CCP4MAPWISH_UTILS_DIAG)
    printf("inv_transform_coordinates: true fractional coords %f %f\n",*xfrac,*yfrac);

  /* Generate orthogonal coordinates */
  *xorth = map->frac_to_orth[iu][iu]*u + map->frac_to_orth[iu][iv]*v;
  *yorth = map->frac_to_orth[iv][iu]*u + map->frac_to_orth[iv][iv]*v;
  if (CCP4MAPWISH_UTILS_DIAG)
    printf("inv_transform_coordinates: true orthogonal coords %f %f\n",*xorth,*yorth);

  return 1;
}

/*------------------------------------------------------------------------------*/

/* blankstr

   Set a character array to contain a null-terminated "blank" string,
   i.e. a string made up entirely of spaces.
   The number of spaces is set by the argument nblanks

   Returns 1 on completion, or 0 if nblanks is less than 1.

   The calling application needs to ensure that the character array has
   sufficient space to accomodate the requested number of spaces plus the
   trailing null character.
*/
int blankstr(char blank_str[], int nblanks)
{
  int i;

  if (nblanks<1) return 0;
  for (i=0; i<nblanks; i++) blank_str[i] = ' ';
  blank_str[nblanks] = '\0';
  return 1;
}

/*----------------------------------------------------------------*/

/* asc_cmp
   Utility function used to do comparision of contour level values
   when qsorting into ascending order */
int asc_cmp(const void *vp, const void *vq)
{
  const float *p = vp;
  const float *q = vq;
  float       diff = *p - *q;
  return ((diff <= 0.0) ? ((diff <0.0) ? -1 : 0) : +1);
}

/*******************************************************************/

/* Conversion functions */

/*******************************************************************/

/* get_axis_id

   Given character string representation of an axis identifier ("x",
   "y" or "z"), set the integer representation (x=0, y=1, z=2) and
   the corresponding indices for the U and V axes

   Return 1 if successful, 0 if there is an error (or if the axis
   is unrecognised)
*/
int get_axis_id(char *axis, int *iaxis, int *iu, int *iv)
{
  if (strcmp(axis,"x") == 0) {
    *iaxis = 0;
    *iu    = 1;
    *iv    = 2;
  } else if (strcmp(axis,"y") == 0) {
    *iaxis = 1;
    *iu    = 2;
    *iv    = 0;
  } else if (strcmp(axis,"z") == 0) {
    *iaxis = 2;
    *iu    = 0;
    *iv    = 1;
  } else {
    /* Invalid axis specification */
    return 0;
  }
  return 1;
}

/*----------------------------------------------------------------*/

/* get_axis_name

   Given the index number of an axis (0, 1 or 2) return the corresponding
   name (X, Y or Z) as a single character.

   Returns null (\0) if the index is out of range.
*/
char get_axis_name(int iaxis) {

  switch (iaxis) {
  case 0: return 'X';
  case 1: return 'Y';
  case 2: return 'Z';
  default: return '\0';
  }
}

/*------------------------------------------------------------------------------*/
