C
C     GEOMCALC
C     Copyright (C) 1995 Phil Evans
C
C     This code is distributed under the terms and conditions of the
C     CCP4 Program Suite Licence Agreement as a CCP4 Application.
C     A copy of the CCP4 licence can be obtained by writing to the
C     CCP4 Secretary, Daresbury Laboratory, Warrington WA4 4AD, UK.
C
C*************************************************************
C
C
C**************************************************************
c
c  GEOMCALC
c
c    Read and store molecule, then do various geometry calculations,
c according to commands
c      - fit subset of atoms to plane
c      - calculate deviation of atoms from plane
c      - calculate atom - atom distances
c      - calculate bond angles
c      - calculate torsion angles
c
c  Commands:-
c
c The syntax of an atom-name is <chain-name>|<residue_number>:<atomname>,
c ie with the 3 components separated by the delimiters "|" and ":".
c If either  <chain-name> or <residue_number> are omitted, they default
c to the previously used value (initially " " and "1" respectively).
c Note that atom-specifiers are case-sensitive.
c
c---> READ
c  Read molecule from file & store. from the "logical name" XYZIN. Any 
C previously store molecule will be overwritten. On entry to the program, 
C it will always try to read XYZIN. 
c
c---> PLANE [<plane_number>] [+] <atom-names> | "ALL"
c  List atom names to define plane. If the first token is "+", these
c atoms will be added to a previously defined plane. Otherwise, a
c PLANE command will replace a previously defined plane.
c The direction of the plane normal is defined as approximately parallel
c to the vector (xyz2-xyz1) x (xyz3-xyz1), where
c xyz1, xyz2, xyz3 are the coordinates of atoms 1,2,3
c If "plane_number" is given, then this command defines this plane
c and makes it the current plane. If it is not set, the plane-number
c defaults to 1.
c
c If the keyword ALL is given instead of an atom list, all atoms
c will be used. This may be useful for finding the principal directions
c of a whole molecule
c
c---> FROMPLANE <atom-names>
c  Calculate perpendicular distance of listed atoms from previously defined
c current plane (ie the last one defined, or the one set with the
c CURRENT command).
c
c---> ATOM     <atom-name>
c  Print coordinates of specified atom
c
c---> DISTANCE <atom-name1> <atom-name2>
c  Calculate distance between named atoms 
c
c---> ANGLE <atom-name1> <atom-name2> <atom-name3>
c  Calculate angle between named atoms 
c
c---> TORSION <atom-name1> <atom-name2> <atom-name3>  <atom-name4>
c  Calculate torsion angle between named atoms 
c
c---> TRANSFORM
c  Transform all stored atoms into frame of previously defined plane,
c ie put origin at centre of plane, with the major axis along x and the
c plane normal along z
c  Note that all currently defined planes will also be transformed into
c the new frame
c
c---> XAXIS <atom-name1> <atom-name2>
c  Redefine transformation for the current plane to put the plane normal
c along z and the vector  vx = <atom-name1> -> <atom-name2> in the xz plane
c (ie the projection of vx on the plane is put along x). This will be then
c be used for any subsequent TRANSFORM commands. It allows for a consistent
c frame definition for a series of related but different planes.
c
c---> CENTRE <atom-name> | <x> <y> <z>
c  Redefine the centre of the current plane, either to the position of an atom,
c or to the given coordinates. This will shift the origin
c of subsequent TRANSFORM operations, but will mangle future FROMPLANE
c commands (unless the new centre lies in the plane)
c
c---> WRITE
c  Write out stored coordinates to file (normally after a TRANSFORM command)
c the "logical name" XYZOUT is used.
c
c---> CURRENT <plane_number>
c  Define current plane number: this plane must have been previously
c defined with a PLANE command
c
c---> FOLD <plane_number_1> <plane_number_2>
c  Calculate angle between two previously defined planes 
c
c---> TILT <atom-name1> <atom-name2> | <dx> <dy> <dz>
c  Calculate angle between vector and the normal to the current plane.
c  The vector may be defined either as <dx> <dy> <dz>,
c  or as <atom-name1> -> <atom-name2>
c
c---> SEPARATION <plane_number_1> <plane_number_1>
c Calculate the distance between the centres of two planes


c  Phil Evans, MRC LMB, September 1995
c     Plane fit algorithm stolen from Judd Fermi
c
      program geomcalc
c     ================
c
C Things for PARSER ----
      integer maxtok
      parameter (maxtok=100)
      character key*4,cvalue(maxtok)*4,line*400
      integer ibeg(maxtok),iend(maxtok),ityp(maxtok),
     .   idec(maxtok),ntok
      real    fvalue(maxtok)
      logical lend
c
c Plane
      integer maxpln, maxnpl
      parameter (maxpln=30000, maxnpl=50)
      real xyzpln(3,maxpln), wgtpln(maxpln),
     $   plane_normal(3,maxnpl),
     $   cofg(3,maxnpl), rmat(3,3,maxnpl), thickness(3),
     $   radii(3), radius
      integer natpln(maxnpl), numpln, iplane, iplane2
      character atompln(maxpln)*20
c Locals
      integer maxatm
      parameter (maxatm=100)
      character filein*80, fileout*80, chain*4, atname*4, atom*20, 
     $          atoms(maxatm)*20, errlin*80
      integer istat, natoms, nfilein, nfileout, idum, ierr, lunout, 
     $   iresidue, lflag, jplane, kplane
      integer i, j, n
      real q, b(6), xyz(3), distance, rms, xyzs(3,maxatm),
     $   v1(3), v2(3), v3(3), angle, tangle, rm(3,3), rm2(3,3)
      logical allatm
c
      logical ccponl
      integer lenstr
      real dot
      external ccponl, lenstr, dot, matmul
c
c Commands
      integer nkeys, ikey
      parameter (nkeys=15)
      character*16 keys(nkeys)
c
c
c ---- Data ----
      data keys/'PLANE', 'FROMPLANE','TRANSFORM','READ','WRITE',
     $   'DISTANCE', 'ANGLE', 'TORSION','CURRENT','FOLD','XAXIS',
     $   'CENTRE','ATOM','TILT','SEPA'/
c=============================================================
c
      call ccpfyp
      call ccprcs (6, 'GEOMCALC', '$Date: 2008/04/08 12:22:06 $')
      call XYZINIT
      lunout = 6
c
      allatm = .false.
      numpln = 0
      iplane = 0
      do 91, i = 1, maxnpl
         natpln(i) = 0
 91   continue 
      chain = ' '
      iresidue = 1
      atname = ' '
      natoms = 0
      nfileout = 0
      nfilein = 1
      filein = 'XYZIN'
      call read_mol(filein, istat)
      if (istat .gt. 0) then
         natoms = istat
      endif
c
c--------- Read input commands
 1    if (ccponl(idum)) then
         write (6, '(a)') '-->'
      endif
      line=' '
      ntok=maxtok
      call parser(
     .   key,line,ibeg,iend,ityp,fvalue,cvalue,idec,ntok,lend,.true.)
      if(lend) go to 9000
      if(ntok.eq.0) go to 1
      if (cvalue(1) .eq. '?') then
         write (6, '(a,/,20(8(2x,a),/))') ' Possible keywords:',
     $      (keys(i)(1:lenstr(keys(i))),i=1,nkeys)
         go to 1
      endif
c
      call ccpupc(key)
      call chkkey(key,keys,nkeys,ikey)
      if(ikey.eq.0) then
       call ccperr(3,' ** Unrecognized keyword **')
       if(ccponl(idum)) then
        goto 1
       else
        ierr=ierr+1
       endif
      else if(ikey.lt.0) then
       call ccperr(3,' ** Ambiguous keyword **')
       if(ccponl(idum)) then
        goto 1
       else
        ierr=ierr+1
       endif
      endif
c
      do 11, i=1,ntok
       call ccpupc(cvalue(i))
 11   continue
c
      go to (
     $   100, 200, 300, 400, 500, 600, 700, 800, 900, 1000,
     $   1100, 1200, 1300, 1400, 1500), ikey
c
c============================================================
c
c  PLANE [<plane_number>] [+] <atom-names> | "ALL"
c
c   List atom names to define plane
c
 100  n = 2
      if (natoms .le. 0) then
         write (6, '(a)') ' *** No molecule read yet ***'
         go to 1
      endif
c
      if (ityp(n) .eq. 2) then
c  A number as 1st argument is taken as plane number
         call gttint(n,iplane,lflag,ntok,ityp,fvalue)
         if (lflag .ne. 0) then
            call ccperr(2,' **** Illegal plane number ****')
            go to 1
         endif
         n = n+1
      else
         iplane = 1
      endif
      numpln = max(numpln, iplane)
      if (iplane .gt. maxnpl) then
         write (errlin, '(a,i4,a)')
     $      ' **** Too many planes, maximum = ',maxnpl,
     $      ' recompile increasing maxnpl ****'
         call ccperr(1,errlin)
      endif
c
      if (cvalue(n) .eq. '+') then
c "+" means add to existing plane
         n = n+1
      else
         natpln(iplane) = 0
         atname = ' '
      endif
c
      allatm = .false.
      call ccpupc(cvalue(n))
      if (cvalue(n) .eq. 'ALL') then
c--- Use all atoms in plane
         allatm = .true.
         if (natoms .gt. maxpln) then
            write (errlin, '(a,i8,a)')
     $         ' **** More than ',maxpln,
     $         ' atoms in plane, recompile ****'
            call ccperr(1,errlin)
         endif
         do 105, j = 1, natoms
            natpln(iplane) = natpln(iplane)+1
            call next_atom(j, chain, iresidue, atompln(natpln(iplane)),
     $         xyzpln(1,natpln(iplane)), b, q, istat)
            if (istat .ne. 0) then
               write (errlin,'(a,i8,a)')
     $            ' **** Failed to find atom ',j,' ****'
               call ccperr(2,errlin)
               go to 1
            endif
            if (q .gt. 0.0) then
               wgtpln(j) = 1.0
            else
               wgtpln(j) = 0.0
            endif
 105     continue 
      else
c--- Atom list
         do 101, i = n, ntok
            natpln(iplane) = natpln(iplane)+1
            if (natpln(iplane) .gt. maxpln) then
               write (errlin, '(a,i8,a)')
     $            ' **** More than ',maxpln,
     $            ' atoms in plane, recompile ****'
               call ccperr(1,errlin)
            endif
            atompln(natpln(iplane)) = line(ibeg(i):iend(i))
 101     continue
c
c Interpret atom names
         do 110, j = 1, natpln(iplane)
            call atom_spec(atompln(j), chain, iresidue, atname, istat)
            if (istat .ne. 0) then
               errlin = ' *** Illegal format for atom: '//atompln(j)
            call ccperr(2,errlin)
               go to 1
            endif
            call get_atom(chain, iresidue, atname,
     $         xyzpln(1,j), b, q, istat)
            if (istat .ne. 0) then
               errlin = ' *** Atom '//atompln(j)//' not found ***'
               call ccperr(2,errlin)
               go to 1
            endif
            wgtpln(j) = 1.0
 110     continue 
      endif
c
c Fit plane
      call fit_plane
     $   (xyzpln, wgtpln, natpln(iplane),
     $   plane_normal(1,iplane), cofg(1,iplane), rmat(1,1,iplane),
     $   thickness, radii, radius, istat)
      if (istat .ne. 0) then
         call ccperr(2,
     $      ' **** Plane fitting failed, < 3 accepted atoms ****')
         natpln(iplane) = -1
         go to 1
      endif
      write (6,'(/,a,i6,a,/)')
     $   ' Plane fitted to ', natpln(iplane),' atoms'
      write (6, '(a,/,a,/)')
     $   ' "Thickness" is RMS deviation along principal axis, ',
     $   ' Radii are in plane perpendicular to axis'
      write (6, '(a)')
     $   ' Axis           Axes          Centre Thickness Radii'
      write (6, '(1x,i3,3f8.4,3f8.3)')
     $   (i, (rmat(i,j,iplane), j=1,3), cofg(i,iplane),
     $   thickness(i), radii(i), i=1,3)
      write (6, '(/,a,f8.3,/)')
     $   ' Equivalent overall radius', radius
c
c RMS deviation
      rms = thickness(3)
      write (6,'(a,f8.3,/)') ' RMS deviation from plane: ', rms
c
c Equation of plane
      write (6, 6100)
     $   ' Equation of plane:',(plane_normal(i,iplane),i=1,3),
     $   dot(plane_normal(1,iplane), cofg(1,iplane))

 6100 format (a/6x,f8.4,' x  + ', f8.4,' y  + ', f8.4,' z = ',
     $    f10.4/)
c
      write (lunout, '(/,a,i4,a/)')
     $   ' Plane ',iplane,' defined'
      go to 1
c
c
c============================================================
c
c FROMPLANE <atoms>
c   Get distance of atoms from plane
c
 200  n = 2
      if (iplane .le. 0 .or. iplane .gt. numpln) then
        call ccperr(2,' *** Undefined plane number ***')
         go to 1
      endif
      if (natpln(iplane) .le. 0) then
         call ccperr(2,' *** No plane defined ***')
         go to 1
      endif
c Look for keyword POINT
      call ccpupc(cvalue(2))
      if (cvalue(2) .eq. 'POIN') then
c  Should be three coordinates
         if (ntok .ne. 5) then
            call ccperr(1,' *** Invalid POINT coordinate ***')
         endif
         n = 3
         do i = 1, 3
            call gttrea(n,v1(i),lflag,ntok,ityp,fvalue)
            if (lflag .ne. 0) then
               call ccperr(2,' *** Illegal coordinate ***')
               go to 1
            endif
            n = n+1
         enddo
         call from_plane(v1, plane_normal(1,iplane),
     $      cofg(1,iplane), distance)
         q = 0.0
         if (abs(rms) .gt. 0.001) q = abs(distance/rms)
         write (6, '(/a,3f8.2,a,f8.3,a,f8.2,a)') 'Distance of point ',
     $      v1, ' from plane: ', distance,
     $      ', = ',q,' * rms'
         go to 1
      endif 

      do 201, j = n, ntok
         atom = line(ibeg(j):iend(j))
c Interpret atom name
         call atom_spec(atom, chain, iresidue, atname, istat)
         if (istat .ne. 0) then
           errlin = ' *** Illegal format for atom: '//atom
           call ccperr(2,errlin)
            go to 1
         endif
         call get_atom(chain, iresidue, atname,
     $      xyz, b, q, istat)
         if (istat .ne. 0) then
            errlin = ' *** Atom '//atom//' not found ***'
            call ccperr(2,errlin)
            go to 1
         endif
         call from_plane(xyz, plane_normal(1,iplane),
     $      cofg(1,iplane), distance)
         q = 0.0
         if (abs(rms) .gt. 0.001) q = abs(distance/rms)
         write (6, '(a,a,a,f8.3,a,f8.2,a)') 'Distance of atom ',
     $      atom, ' from plane: ', distance,
     $      ', = ', q,' * rms'
 201  continue
      go to 1
c            
c============================================================
c
c  TRANSFORM
c    Apply transformation to put whole molecule into frame of plane
c (principal axes along x,y,z, normal along z)
c
 300  if (natpln(iplane) .le. 0) then
         call ccperr(2,' *** No plane fitted ***')
         go to 1
      endif
      if (natoms .le. 0) then
         call ccperr(2,' *** No molecule read yet ***')
         go to 1
      endif
      call transform_mol(rmat(1,1,iplane), cofg(1,iplane))
c
c  Transform all current planes into new frame
      n = 0
      call ccpmvr(v1, cofg(1,iplane), 3)
      call ccpmvr(rm, rmat(1,1,iplane), 9)
c   Transformation = [rm](v - v1)   where rm = rmat(,,iplane), v1 =cofg(,iplane)
      do 310, i = 1, numpln
         if (natpln(i) .gt. 0) then
            n = n+1
c  Plane normal, just rotate, no translation
            call matvec(v2, rm, plane_normal(1,i))
            call ccpmvr(plane_normal(1,i), v2, 3)
c  Centre, translate & rotate
            call vdif(v2, cofg(1,i), v1)
            call matvec(v3, rm, v2)
            call ccpmvr(cofg(1,i), v3, 3)
c  Rotation matrix, rotate
            call matmul(rm2, rm, rmat(1,1,i))
            call ccpmvr(rmat(1,1,i), rm2, 9)
         endif
 310  continue
      write (lunout, '(1x,i5,a,/)')
     $   n, ' planes transformed into new frame'
      go to 1
c            
c============================================================
c
c  READ
c    Read molecule from XYZIN
 400  if (ntok .le. 1) then
         filein = ' '
      else
         filein = line(ibeg(2):)
      endif
      call read_mol(filein, istat)
      if (istat .gt. 0) then
         natoms = istat
      endif
      go to 1
c============================================================
c
c  WRITE
c   Write molecule to XYZOUT
 500  if (natoms .le. 0) then
         write (6, '(a)') ' *** No molecule read yet ***'
         go to 1
      endif
      if (ntok .le. 1) then
         fileout = ' '
      else
         fileout = line(ibeg(2):)
      endif
      call write_mol(fileout, filein, istat)
      if (istat .ne. 0) then
         write (6, '(a)') ' *** File write failed ***'
      endif
      go to 1
c            
c============================================================
c
c  DISTANCE  <atom-name1> <atom-name12>
c  Calculate distance between named atoms 
c
 600  if (natoms .le. 0) then
         call ccperr(2,' *** No molecule read yet ***')
         go to 1
      endif
c
      if (ntok .lt. 3) then
         call ccperr(2,' *** Must define two atoms ***')
         go to 1
      endif
      n = 2
      j = 0
      do 601, i = n, n+1
         j = j+1
         atoms(j) = line(ibeg(i):iend(i))
 601  continue 
c
      do 602, j = 1,2
         call atom_info
     $      (atoms(j), chain, iresidue, atname, xyzs(1,j), q, b, istat)
         if (istat .ne. 0) go to 1
 602  continue 
c  Distance
      call vdif(v1, xyzs(1,1), xyzs(1,2))
      distance = sqrt(dot(v1,v1))
      write (6, '(/,a,a,a,a,a,f8.3)')
     $   ' Distance between atoms ',atoms(1)(1:lenstr(atoms(1))),
     $   ' and ', atoms(2)(1:lenstr(atoms(2))),' = ', distance
      go to 1
c            
c============================================================
c
c   ANGLE <atom-name1> <atom-name12> <atom-name3>
c  Calculate angle between named atoms 
c
 700  if (natoms .le. 0) then
         call ccperr(2,' *** No molecule read yet ***')
         go to 1
      endif
c
      if (ntok .lt. 4) then
         call ccperr(2,' *** Must define three atoms ***')
         go to 1
      endif
      n = 2
      j = 0
      do 701, i = n, n+2
         j = j+1
         atoms(j) = line(ibeg(i):iend(i))
 701  continue 
c
      do 702, j = 1,3
         call atom_info
     $      (atoms(j), chain, iresidue, atname, xyzs(1,j), q, b, istat)
         if (istat .ne. 0) go to 1
 702  continue 
c
      call bangle(xyzs(1,1), xyzs(1,2), xyzs(1,3), angle)
      write (6, '(/,7a,f8.2)')
     $   ' Angle between atoms ',atoms(1)(1:lenstr(atoms(1))),
     $   ', ',atoms(2)(1:lenstr(atoms(2))),
     $   ' and ', atoms(3)(1:lenstr(atoms(3))),' = ', angle
      go to 1
c            
c============================================================
c
c  TORSION <atom-name1> <atom-name12> <atom-name3>  <atom-name4>
c  Calculate torsion angle between named atoms 
c
 800  if (natoms .le. 0) then
         call ccperr(2,' *** No molecule read yet ***')
         go to 1
      endif
c
      if (ntok .lt. 5) then
         call ccperr(2,' *** Must define four atoms ***')
         go to 1
      endif
      n = 2
      j = 0
      do 801, i = n, n+3
         j = j+1
         atoms(j) = line(ibeg(i):iend(i))
 801  continue 
c
      do 802, j = 1,4
         call atom_info
     $      (atoms(j), chain, iresidue, atname, xyzs(1,j), q, b, istat)
         if (istat .ne. 0) go to 1
 802  continue 
c
      call torsion(xyzs(1,1), xyzs(1,2), xyzs(1,3), xyzs(1,4), tangle)
      write (6, '(/,9a,f8.2)')
     $   ' Torsion angle between atoms ',atoms(1)(1:lenstr(atoms(1))),
     $   ', ',atoms(2)(1:lenstr(atoms(2))),
     $   ', ',atoms(3)(1:lenstr(atoms(3))),
     $   ' and ', atoms(4)(1:lenstr(atoms(4))),' = ', tangle
      go to 1
c            
c============================================================
c
c  CURRENT <plane_number>
c  Define current plane number: this plane must have been previously
c defined with a PLANE command
c
 900  call gttint(2,iplane,lflag,ntok,ityp,fvalue)
      if (lflag .ne. 0 .or. iplane .le. 0 .or.
     $   iplane .gt. numpln) then
         write (errlin, '(a,i8,a)')
     $      ' **** Illegal plane number ',iplane,' ****'
         call ccperr(2,errlin)
         iplane = -1
         go to 1
      endif
      if (natpln(iplane) .le. 0) then
         write (errlin, '(a,i4,a)')
     $      ' **** Plane ',iplane, ' has not been defined ****'
         call ccperr(2,errlin)
         iplane = -1
         go to 1
      endif
      go to 1
c            
c============================================================
c
c  FOLD <plane_number_1> <plane_number_2>
c  Calculate angle between two previously defined planes 
 1000 call gttint(2,jplane,lflag,ntok,ityp,fvalue)
      if (lflag .ne. 0 .or. jplane .le. 0 .or.
     $   jplane .gt. numpln) then
         write (errlin, '(a,i8,a)')
     $      ' **** Illegal plane number ',jplane,' ****'
         call ccperr(2,errlin)
         go to 1
      endif
      call gttint(3,kplane,lflag,ntok,ityp,fvalue)
      if (lflag .ne. 0 .or. kplane .le. 0 .or.
     $   kplane .gt. numpln) then
         write (errlin, '(a,i8,a)')
     $      ' **** Illegal plane number ',kplane,' ****'
         call ccperr(2,errlin)
         go to 1
      endif
      call fold
     $   (jplane, kplane, natpln, plane_normal, numpln, angle, istat)
      write (lunout, '(/,a,i4,a,i4,a,f10.3,a)')
     $   ' Angle between planes', jplane, ' and', kplane,
     $   ' = ',angle,' degrees'
c
      go to 1
c            
c============================================================
c
c   XAXIS <atom-name1> <atom-name2>
c
 1100 if (natoms .le. 0) then
         call ccperr(2,' *** No molecule read yet ***')
         go to 1
      endif
c
      if (ntok .lt. 3) then
         call ccperr(2,' *** Must define two atoms ***')
         go to 1
      endif
      n = 2
      j = 0
      do 1101, i = n, n+1
         j = j+1
         atoms(j) = line(ibeg(i):iend(i))
 1101  continue 
c
      do 1102, j = 1,2
         call atom_info
     $      (atoms(j), chain, iresidue, atname, xyzs(1,j), q, b, istat)
         if (istat .ne. 0) go to 1
 1102  continue 
       call vdif(v1, xyzs(1,1), xyzs(1,2))

       call xaxmat(v1, plane_normal(1,iplane), rmat(1,1,iplane), istat)
       if (istat .ne. 0) then
         call ccperr(2,' *** Failed to define transformation ***')
       endif
       go to 1
c            
c============================================================
c
c CENTRE <atom-name> | <x> <y> <z>
c  Redefine the centre of the current plane, either to the position of an atom,
c or to the given coordinates. This will shift the origin
c of subsequent TRANSFORM operations, but will mangle future FROMPLANE
c commands (unless the new centre lies in the plane)
c 
 1200  if (natoms .le. 0) then
          call ccperr(2,' *** No molecule read yet ***')
          go to 1
       endif
c
      if (ntok .lt. 2) then
         call ccperr(2,' *** Must define centre ***')
         go to 1
      endif
      n = 2
c 
      if (ntok .eq. 2) then
c Atom name given
         j = 1
         atoms(j) = line(ibeg(n):iend(n))
c
         call atom_info
     $      (atoms(j), chain, iresidue, atname, v1, q, b, istat)
         if (istat .ne. 0) go to 1
      else
c Coordinates given
         do 1201, i = 1, 3
            call gttrea(n,v1(i),lflag,ntok,ityp,fvalue)
            if (lflag .ne. 0) then
               call ccperr(2,' *** Illegal coordinate ***')
               go to 1
            endif
            n = n+1
 1201    continue 
      endif
c Store as new centre
      call ccpmvr(cofg(1,iplane), v1, 3)
      write (lunout, '(/,a,i4,a,3f10.4)')
     $   ' Centre of plane ', iplane, ' has been moved to ',
     $   v1
      go to 1
c            
c============================================================
c
c  ATOM     <atom-name>
c  Print coordinates of specified atom
c
 1300 if (natoms .le. 0) then
         call ccperr(2,' *** No molecule read yet ***')
         go to 1
      endif
c
      if (ntok .lt. 2) then
         call ccperr(2,' *** Must define atom ***')
         go to 1
      endif
      n = 2
      j = 1
      atoms(j) = line(ibeg(n):iend(n))
c
      call atom_info
     $   (atoms(j), chain, iresidue, atname, xyzs(1,j), q, b, istat)
      if (istat .ne. 0) go to 1
c
      write (lunout, '(a,a,a,3f10.4,/)')
     $   ' Coordinates of atom ',atoms(j),': ',(xyzs(i,j),i=1,3)
      go to 1
c============================================================
c
c TILT <atom-name1> <atom-name2> | <dx> <dy> <dz>
c  Calculate angle between vector and the normal to the current plane.
c  The vector may be defined either as <dx> <dy> <dz>,
c  or as <atom-name1> -> <atom-name2>
c
 1400 if (natoms .le. 0) then
         call ccperr(2,' *** No molecule read yet ***')
         go to 1
      endif
      if (natpln(iplane) .le. 0) then
         call ccperr(2,' *** No plane defined ***')
         go to 1
      endif
c
      if (ntok .le. 2) then
         call ccperr(2,' *** Must define two atoms ***')
         go to 1
      endif
      n = 2
      if (ntok .eq. 3) then
         j = 0
         do 1401, i = n, n+1
            j = j+1
            atoms(j) = line(ibeg(i):iend(i))
 1401    continue 
c
         do 1402, j = 1,2
            call atom_info
     $         (atoms(j), chain, iresidue, atname, xyzs(1,j),
     $         q, b, istat)
            if (istat .ne. 0) go to 1
 1402    continue 
         call vdif(v1, xyzs(1,1), xyzs(1,2))
      else
c Coordinates given
         do 1411, i = 1, 3
            call gttrea(n,v1(i),lflag,ntok,ityp,fvalue)
            if (lflag .ne. 0) then
               call ccperr(2,' *** Illegal coordinate ***')
               go to 1
            endif
            n = n+1
 1411    continue 
      endif
c
      call tilt(v1, plane_normal(1,iplane), angle, istat)
      if (istat .ne. 0) then
        call ccperr(2,' **** Zero-length vector ****')
      else
         write (lunout, '(/,a,3f7.4,a,f6.1,a,/)')
     $      ' Angle between vector (',v1,') and plane normal =',
     $      angle, ' degrees'
      endif
      go to 1
c            
c============================================================
c
c SEPARATION <p1> <p2>
c  Distance between the centres of two planes, numbers p1 & p2 (previously defined)
c
 1500 call gttint(2,iplane,lflag,ntok,ityp,fvalue)
      if (lflag .ne. 0 .or. iplane .le. 0 .or.
     $   iplane .gt. numpln) then
         write (errlin, '(a,i8,a)')
     $      ' **** Illegal plane number ',iplane,' ****'
         call ccperr(2,errlin)
         iplane = -1
         go to 1
      endif
      if (natpln(iplane) .le. 0) then
         write (errlin, '(a,i4,a)')
     $      ' **** Plane ',iplane, ' has not been defined ****'
         call ccperr(2,errlin)
         iplane = -1
         go to 1
      endif
      call gttint(3,iplane2,lflag,ntok,ityp,fvalue)
      if (lflag .ne. 0 .or. iplane2 .le. 0 .or.
     $   iplane2 .gt. numpln) then
         write (errlin, '(a,i8,a)')
     $      ' **** Illegal plane number ',iplane2,' ****'
         call ccperr(2,errlin)
         iplane2 = -1
         go to 1
      endif
      if (natpln(iplane2) .le. 0) then
         write (errlin, '(a,i4,a)')
     $      ' **** Plane ',iplane2, ' has not been defined ****'
         call ccperr(2,errlin)
         iplane2 = -1
         go to 1
      endif
      call vdif(v1, cofg(1,iplane), cofg(1,iplane2))
      distance = sqrt(dot(v1,v1))
      write (6, '(/,a,i3,a,i3,a,f8.3)')
     $   ' Distance between planes ',iplane,
     $   ' and ', iplane2,' = ', distance
      go to 1

c============================================================
c
 9000 write (6,'(//)')
      call ccperr(0,'*** Normal termination***')
      end
c
c
c
      subroutine read_mol(file, istat)
c     ================================
c
c  Read molecule into common
c
c  On entry:
c    file    filename 
c
c  On exit:
c    istat   .gt. 0    OK, = number of atoms stored
c            =  -1 file not found
c
      character*(*) file
      integer istat
c
C&&*&& include  molec.fh
c
c  Store for molecule
c
      integer maxat
      parameter (maxat = 30000)
c
      common /molec/ xyz_mol(3,maxat), b_mol(6,maxat), q_mol(maxat),
     $   numres_mol(maxat), numatom_mol
      real  xyz_mol, b_mol, q_mol
      integer  numres_mol, numatom_mol
c
      common /cmolec/ atname_mol(maxat), restyp_mol(maxat),
     $   chain_mol(maxat)
      character atname_mol*4, restyp_mol*4, chain_mol*4
      
      common /dmolec/ segid_mol(maxat), id_mol(maxat)
      character segid_mol*4, id_mol*4
c
      save /molec/, /cmolec/, /dmolec/
c
c=========================================================================
      
C&&*&& end_include  molec.fh
c
      character errlin*100
      character atnam*4, resnam*4, resno*4, chnnam*1, id*4, inscod*1,
     $          altcod*1, segid*4
      integer nat, iser, i, ixyzin, iz
      real biso,xx,yy,zz, occ, b(6)
      logical ccpexs
      external ccpexs
c Set filename if not set
      if (file .eq. ' ') then
         file = 'XYZIN'
      endif

      if (.not. ccpexs(file)) then
        istat = -1
        return
      endif
c
      istat = +1
      ixyzin = 0
      call XYZOPEN(file,'INPUT',' ',IXYZIN,istat)
      if (istat .lt. 0) then 
         return
c        ======
      endif
c
      nat = 0
c
 100  call XYZADVANCE(IXYZIN,0,0,*500,*500)
      call XYZATOM(IXYZIN,iser,atnam,resnam,chnnam,iresn,resno,
     $              inscod,altcod,segid,iz,id)
      call XYZCOORD(IXYZIN,'O','B',xx,yy,zz,occ,biso,b)
c
c  If ANISO card then transfer b.
c
      if (b(2).ne.0.0 .or. b(3).ne.0.0) then
        if (xx.eq.0.0 .and. yy.eq.0.0 .and. zz.eq.0.0) then
          do 10 i=nat,1,-1
            if (chnnam .eq. chain_mol(i)(1:1)) then
              if (iresn .eq. numres_mol(i)) then
                if (atnam .eq. atname_mol(i)) then
                  do 20 j=1,6
                    b_mol(j,i) = b(j)
  20              continue
                  goto 100
                endif
              endif
            endif
  10      continue
          goto 100
        endif
      endif

      nat = nat + 1
      if (nat .gt.maxat) then
         write (errlin, '(a,i8,a)') ' **** More than ',maxat,
     $                         ' atoms, increase ****'
         call ccperr(1,errlin)
      endif

      atname_mol(nat) = atnam
      restyp_mol(nat) = resnam
      numres_mol(nat) = iresn
      xyz_mol(1,nat) = xx
      xyz_mol(2,nat) = yy
      xyz_mol(3,nat) = zz
      q_mol(nat) = occ
      do 30 i=1,6
        b_mol(i,nat) = b(i)
  30  continue
      segid_mol(nat) = segid
      id_mol(nat) = id
c
c
c Store chain as left-justified *4
      chain_mol(nat) = ' '
      chain_mol(nat)(1:1) = chnnam
c
      go to 100
c
 500  numatom_mol = nat
      istat = nat
      write (6, '(1x,i8,a)') nat,' atoms read'
      call XYZCLOSE(IXYZIN)
      return
      end
c
c
c
      subroutine write_mol(fileout, filein, istat)
c     ============================================
c
c  Write molecule into common
c
c  On entry:
c    fileout    filename for output
c    filein     filename of input file, for copying header stuff etc
c
c  On exit:
c    istat   = 0 OK, = -1 failed
c
      character*(*) filein, fileout
      integer istat
c
C&&*&& include  molec.fh
c
c  Store for molecule
c
      integer maxat
      parameter (maxat = 30000)
c
      common /molec/ xyz_mol(3,maxat), b_mol(6,maxat), q_mol(maxat),
     $   numres_mol(maxat), numatom_mol
      real  xyz_mol, b_mol, q_mol
      integer  numres_mol, numatom_mol
c
      common /cmolec/ atname_mol(maxat), restyp_mol(maxat),
     $   chain_mol(maxat)
      character atname_mol*4, restyp_mol*4, chain_mol*4
c
      common /dmolec/ segid_mol(maxat), id_mol(maxat)
      character segid_mol*4, id_mol*4
c      
      save /molec/, /cmolec/, /dmolec/
c
c=========================================================================
      
C&&*&& end_include  molec.fh
c
      integer i, ixyzin, ixyzout, k
      real xx, yy, zz, occ, b(6)
      character resno*4, inscod*1, altcod*1
      integer iz
c
      istat = +1

      if (filein .eq. ' ') then
         filein = 'XYZIN'
      endif
      if (fileout .eq. ' ') then
         fileout = 'XYZOUT'
      endif
      ixyzin = 0
      ixyzout = 0
      call XYZOPEN(filein,'INPUT',' ',IXYZIN,istat)
      call XYZOPEN(fileout,'OUTPUT',' ',IXYZOUT,istat)
      if (istat .lt. 0) then 
         return
c        ======
      endif
c
      do 10, k = 1, numatom_mol
         call XYZADVANCE(IXYZIN,IXYZOUT,0,*30,*30)
         xx = xyz_mol(1,k)
         yy = xyz_mol(2,k)
         zz = xyz_mol(3,k)
         occ = q_mol(k)
         do 20 i=1,6
           b(i) = b_mol(i,k)
 20      continue

         resno = ' '
         inscod = ' '
         altcod = ' '
         iz = 0

         call XYZATOM(IXYZOUT,k,atname_mol(k),restyp_mol(k),
     $    chain_mol(k),numres_mol(k),resno,inscod,altcod,segid_mol(k),
     $    iz,id_mol(k))
         call XYZCOORD(IXYZOUT,'O','B',xx,yy,zz,occ,biso,b)
         call XYZADVANCE(IXYZOUT,0,0,*30,*30)
 10   continue
      istat = 0
 30   call XYZCLOSE(IXYZIN)
      call XYZCLOSE(IXYZOUT)
      return 
      end
c
c
c
      subroutine get_atom(chain, iresidue, atname, xyz, b, q, istat)
c     ==============================================================
c
c  Extract atom from stored arrays
c
c On entry:
c   chain     chain ID
c   iresidue  residue number
c   atname    atomname
c
c On exit:
c   xyz(3)    coordinates
c   b         Bfactor
c   q         occupancy
c   istat     = 0 found, = -1 not found
c
      character chain*(*), atname*(*)
      real xyz(3), b(6), q
      integer iresidue, istat
c
C&&*&& include  molec.fh
c
c  Store for molecule
c
      integer maxat
      parameter (maxat = 30000)
c
      common /molec/ xyz_mol(3,maxat), b_mol(6,maxat), q_mol(maxat),
     $   numres_mol(maxat), numatom_mol
      real  xyz_mol, b_mol, q_mol
      integer  numres_mol, numatom_mol
c
      common /cmolec/ atname_mol(maxat), restyp_mol(maxat),
     $   chain_mol(maxat)
      character atname_mol*4, restyp_mol*4, chain_mol*4
c
      save /molec/, /cmolec/
c
c=========================================================================
      
C&&*&& end_include  molec.fh
c
      integer i,j
c
      do 10, j = 1, numatom_mol
         if (chain(1:4) .eq. chain_mol(j)) then
            if (iresidue .eq. numres_mol(j)) then
               if (atname(1:4) .eq. atname_mol(j)) then
                  call ccpmvr(xyz, xyz_mol(1,j), 3)
                  do 20 i=1,6
                    b(i) = b_mol(i,j)
 20               continue
                  q = q_mol(j)
                  istat = 0
                  return
               endif
            endif
         endif
 10   continue 
c Not found
      istat = -1
      return 
      end
c
c
c
      subroutine next_atom
     $   (iatom, chain, iresidue, atname, xyz, b, q, istat)
c     =====================================================
c
c  Extract atom from stored arrays
c
c On entry:
c   iatom     atom number required
c
c On exit:
c   chain     chain ID
c   iresidue  residue number
c   atname    atomname
c   xyz(3)    coordinates
c   b         Bfactor
c   q         occupancy
c   istat     = 0 found, = -1 not found
c
      character chain*(*), atname*(*)
      real xyz(3), b(6), q
      integer iresidue, istat, iatom
c
C&&*&& include  molec.fh
c
c  Store for molecule
c
      integer maxat
      parameter (maxat = 30000)
c
      common /molec/ xyz_mol(3,maxat), b_mol(6,maxat), q_mol(maxat),
     $   numres_mol(maxat), numatom_mol
      real  xyz_mol, b_mol, q_mol
      integer  numres_mol, numatom_mol
c
      common /cmolec/ atname_mol(maxat), restyp_mol(maxat),
     $   chain_mol(maxat)
      character atname_mol*4, restyp_mol*4, chain_mol*4
c
      save /molec/, /cmolec/
c
c=========================================================================
      
C&&*&& end_include  molec.fh
c
      integer j
c
      if (iatom .le. 0 .or. iatom .gt. numatom_mol) then
         istat = -1
      endif
      j = iatom
      chain = chain_mol(j)
      iresidue = numres_mol(j)
      atname = atname_mol(j)
      call ccpmvr(xyz, xyz_mol(1,j), 3)
      do 10 i=1,6
        b(i) = b_mol(i,j)
  10  continue
      q = q_mol(j)
      istat = 0
      return
      end
c
c
c
      subroutine atom_info
     $   (atom, chain, iresidue, atname, xyz, q, b, istat)
c     ====================================================
c
c  Get stored information for atom
c
c On entry:
c   atom      atom specifier
c
c On exit:
c   chain     chain ID
c   iresidue  residue number
c   atname    atom name
c   xyz(3)    coordinates
c   q         occupancy
c   b         Bfactor
c   istat     =  0 OK
c             = +1 illegal atom format
c             = +2 atom not found
c
      character*(*) atom, chain, atname
      character*100 errlin
      integer iresidue, istat
      real xyz(3), q, b(6)
c
c Interpret atom names
      call atom_spec(atom, chain, iresidue, atname, istat)
      if (istat .ne. 0) then
         errlin = ' *** Illegal format for atom: '//atom
         call ccperr(2,errlin)
         istat = +1
         go to 10
      endif
      call get_atom(chain, iresidue, atname,
     $   xyz, b, q, istat)
      if (istat .ne. 0) then
         errlin = ' *** Atom '//atom//' not found ***'
         call ccperr(2,errlin)
         istat = +2
      endif
 10   return 
      end
c
c
c
      subroutine transform_mol(rmat, centre)
c     ======================================
c
c Transform molecule by rotation rmat, translate to put centre at origin
c
c On entry:
c   rmat(3,3)   rotation matrix
c   centre(3)   translation (before rotation)
c
      real rmat(3,3), centre(3)
c
C&&*&& include  molec.fh
c
c  Store for molecule
c
      integer maxat
      parameter (maxat = 30000)
c
      common /molec/ xyz_mol(3,maxat), b_mol(6,maxat), q_mol(maxat),
     $   numres_mol(maxat), numatom_mol
      real  xyz_mol, b_mol, q_mol
      integer  numres_mol, numatom_mol
c
      common /cmolec/ atname_mol(maxat), restyp_mol(maxat),
     $   chain_mol(maxat)
      character atname_mol*4, restyp_mol*4, chain_mol*4
c
      save /molec/, /cmolec/
c
c=========================================================================
      
C&&*&& end_include  molec.fh
c
      integer i,j,k
      real v(3)
c
      do 10, k = 1, numatom_mol
         call vdif(v, xyz_mol(1,k), centre)
         call matvec(xyz_mol(1,k), rmat, v)
 10   continue 
c
      write (6, 6001)
     $   numatom_mol, ' atoms transformed by',
     $   ((rmat(i,j), j=1,3), centre(i), i=1,3)
 6001 format(/,1x,i8,a,//,
     $   (7x,'(',3f9.5,' ) (x)',5x,'(',f7.3,' )',/),
     $   (7x,'(',3f9.5,' ) (y)','  +  ','(',f7.3,' )',/),
     $   (7x,'(',3f9.5,' ) (z)',5x,'(',f7.3,' )',/))
      return 
      end
c
c
c
c
      subroutine atom_spec(spec, chain, iresidue, atname, istat)
c     ==========================================================
c
c  Decompose atom specification given as "chain|residue:atom"
c
c On entry:
c    spec*(*)     atom specification
c
c On exit:
c    chain        chain ID, left-justified, unchanged if absent
c    iresidue     residue number, , unchanged if absent
c    atname       atom name, left-justified
c    istat        = 0 OK, = -1 illegal format
c
      character*(*) spec, chain, atname
      integer iresidue, istat
c
      integer i,j,k
      real dummy
c
      integer intfp
      external intfp
c
      istat = 0
c Chain
      k = index(spec,'|')
      if (k. gt. 0) then
         if (k .gt. 1) then
            call ljstfy(chain, spec(1:k-1))
         else
            chain = ' '
         endif
         k = k+1
      else
         k = 1
      endif
c Residue
      j = index(spec,':')
      if (j .gt. 0) then
         if (j .gt. 1) then
            istat = intfp(spec(k:j-1), dummy, i)
            if (istat .eq. 1) then
               iresidue = i
               istat = 0
            else
               istat = -1
            endif
         endif
         j = j+1
      else
         j = k
      endif
c Atom
      atname = spec(j:)
      return
      end
c
c
c
      subroutine fit_plane
     $   (xyz, weight, nat,
     $   plane_normal, cofg, rmat, thickness, radii, radius, istat)
c     =============================================================
c
c  Fit atoms to plane, from principal axes of inertia matrix
c
c On entry:
c   xyz(3,nat)     atom coordinates
c   weight(nat)    weight for fit
c   nat            number of atoms
c
c On exit:
c   plane_normal(3)  plane normal vector (unit)
c                  sign of normal is defined to be approximately parallel
c                  to the vector (xyz2-xyz1) x (xyz3-xyz1), where
c                  xyz1, xyz2, xyz3 are the coordinates of atoms 1,2,3
c   cofg(3)        centre of gravity of plane
c   rmat(3,3)      rotation matrix to rotate into plane coordinates
c                  (plane normal along z)
c   thickness(3)   "thickness" = weighted rms deviation along principal axes
c   radii(3)       "radii" = weighted rms deviation in plane perpendicular
c                  to principal axes
c   radius         rms radius
c   istat          = 0 OK, = -1 failure (too few atoms)
c
      integer nat, istat
      real xyz(3,nat), weight(nat), plane_normal(3), cofg(3),
     $   thickness(3), radii(3), radius, rmat(3,3)
c
      integer i,j,k,n
      real wt, dik, djk, a(3,3), v1(3), v2(3), vn(3), resid(3),
     $   vmat(3,3), det
c
      real dot
      external dot
c
      istat = 0
      n = 0
c  Weighted centre of gravity cofg
      wt = 0.0
      call ccpzr(cofg, 3)
c
      do 1, k = 1, nat
         if (weight(k) .gt. 0.0) then
            n = n+1
            wt = wt + weight(k)
            do 2, i = 1,3
               cofg(i) = cofg(i) + weight(k) * xyz(i,k)
 2          continue
         endif
 1    continue
      if (n .lt. 3) then
         istat = -1
         return
      endif
      do 3, i = 1,3
         cofg(i) = cofg(i)/wt
 3    continue 
c
c---
      call ccpzr(a, 9)
c  Inertia matrix 
      wt = 0.0
      do 10, k = 1,nat
         if (weight(k) .gt. 0.0) then
            wt = wt + weight(k)
            do 11, j = 1,3
               djk = xyz(j,k) - cofg(j)
               do 12, i = 1,3
                  dik = xyz(i,k) - cofg(i)
                  a(i,j) = a(i,j) + weight(k) * dik * djk
 12            continue 
 11         continue 
         endif
 10   continue 
c
      call eign3(a, vmat, resid)
c resid      eigenvalues of residual matrix, in descending order
c            = sums of weighted squared residuals along principal directions
c vmat(i,j)  eigenvector matrix
c            column vmat(.,3) is the plane normal
c
c  Get RMS residuals = thickness
      do 20, i = 1,3
         thickness(i) = sign(sqrt(abs(resid(i)/wt)), resid(i))
 20   continue
c Radii
      radii(1) = sqrt(thickness(2)**2 + thickness(3)**2)
      radii(2) = sqrt(thickness(3)**2 + thickness(1)**2)
      radii(3) = sqrt(thickness(1)**2 + thickness(2)**2)
      radius   = (radii(1)*radii(2)*radii(3))**(1./3.)
c
c Sign of plane_normal defined by approximate normal along vector
c (xyz2-xyz1) x (xyz3-xyz1),
c where  xyz1, xyz2, xyz3 are the coordinates of atoms 1,2,3
c  v1 = xyz2 - xyz1
      call vdif(v1, xyz(1,2), xyz(1,1))
c  v2 = xyz3 - xyz1
      call vdif(v2, xyz(1,3), xyz(1,1))
c  vn = v1 x v2
      call cross(vn, v1, v2)
c
      if (dot(vn, vmat(1,3)) .lt. 0.0) then
c Invert normal to make parallel to vn
         do 21, i = 1,3
            vmat(i,3) = -vmat(i,3)
 21      continue 
      endif
c Invert vmat(.,2) if necessary to make RH set
      call cross(v1, vmat(1,1), vmat(1,2))
      if (dot(v1, vmat(1,3)) .lt. 0.0) then
         do 22, i = 1,3
            vmat(i,2) = -vmat(i,2)
 22      continue 
      endif
c Normal
      call ccpmvr(plane_normal, vmat(1,3), 3)
c Transformation matrix is inverse of vmat
      call minv3(rmat, vmat, det)
c
      return
      end
c
c
c
      subroutine from_plane(xyz, plane_normal, cofg, distance)
c     ========================================================
c
c  Get distance from plane
c
c On entry:
c  xyz(3)           point to get distance from
c  plane_normal(3)  plane normal (unit) 
c  cofg(3)          centre of gravity of plane (or any point on plane)
c
c On exit:
c  distance         distance from plane
c
      real xyz(3), plane_normal(3), cofg(3), distance
c
      real v(3)
c
      real dot
      external dot
c
      call vdif(v, xyz, cofg)
      distance = dot(plane_normal, v)/
     $   sqrt(dot(plane_normal,plane_normal))
      return
      end
c
c
c
      subroutine bangle(x1, x2, x3, angle)
c     ====================================
c
c  Calculate angle between vectors x1->x2 and x3->x2
c
c On entry:
c   x1(3), x2(3), x3(3)   coordinates
c
c On exit:
c   angle    angle in degrees
c
      real x1(3), x2(3), x3(3), angle
c
      real v1(3), v2(3), v3(3), sphi, cphi, rtod
c
      real dot
      external dot
c
      rtod = 45./atan(1.)
      call vdif(v1, x2, x1)
      call vdif(v2, x2, x3)
      cphi = dot(v1, v2)
      call cross(v3, v1, v2)
      sphi = sqrt(dot(v3,v3))
      angle = atan2(sphi, cphi) * rtod
      return
      end
c
c
c
      subroutine torsion(x1, x2, x3, x4, tangle)
c     ==========================================
c
c  Calculate torsion angle defined by points x1 -> x4
c
c On entry:
c   x1(3), x2(3), x3(3), x4(3)   coordinates
c
c On exit:
c   torsion    angle in degrees
c
      real x1(3), x2(3), x3(3), x4(3), tangle
c
      real v1(3), v2(3), v3(3), q(3), t(3), rtod
c
      real dot
      external dot
c
      rtod = 45./atan(1.)
      call vdif(v1, x2, x1)
      call vdif(v2, x3, x2)
      call vdif(v3, x4, x3)
      call cross(q, v1,v2)
      call cross(t, v2, v3)
      tangle = atan2(dot(q,v3)*sqrt(dot(v2,v2)),dot(q,t))*rtod
      return
      end
c
c
c
      subroutine fold
     $   (jplane, kplane, natpln, plane_normal, numpln, angle, istat)
c     ===============================================================
c
c  Get angle between two planes: angle in range 0 - 90 degrees
c
c On entry:
c  jplane, kplane          plane numbers of planes
c  natpln(numpln)          number of atoms (.le. 0 if not defined)
c  plane_normal(3,numpln)  plane normals (unit)
c  numpln                  maximum number of defined planes
c
c On exit:
c  angle                   angle in degrees
c  istat                   =  0  OK
c                          = -1  failed, undefined plane
c
      integer numpln
      integer jplane, kplane, natpln(numpln),  istat
      real  plane_normal(3,numpln), angle
c
      real rtod
      real dot
      external dot
c      
      rtod = 45./atan(1.)
c
c
      if (natpln(jplane) .le. 0) then
         write (6, '(/a,i4,a/)')
     $      ' **** Plane ',jplane, ' has not been defined ****'
         istat = -1
         go to 9
      endif
      if (natpln(kplane) .le. 0) then
         write (6, '(/a,i4,a/)')
     $      ' **** Plane ',kplane, ' has not been defined ****'
         istat = -1
         go to 9
      endif
c
      angle = acos(abs(dot(plane_normal(1,jplane),
     $   plane_normal(1,kplane))))*rtod
      istat = 0
 9    return
      end
c
c
c
      subroutine tilt(v, plane_normal, angle, istat)
c     =============================================
c
c Calculate angle between vector v and plane
c
c On entry:
c  v(3)             vector
c  plane_normal(3)  normal to plane
c
c On exit:
c  angle            angle in degrees between vector and plane
c  istat            =  0 OK, = -1 failure (zero length vector)
c
      real v(3), plane_normal(3), angle
      integer istat
c
      real rtod, vn(3), pn(3)
      real dot
      external dot
c      
      rtod = 45./atan(1.)
c
      call vnorm(vn, v, istat)
      if (istat .ne. 0) return
      call vnorm(pn, plane_normal, istat)
      if (istat .ne. 0) return
      angle = 90.0 - rtod * acos(abs(dot(vn, pn)))
      return 
      end
c
c
c
      subroutine xaxmat(vx, plane_normal, rmat, istat)
c     ================================================
c
c  Define transformation to put z along plane_normal & vx in xz plane
c
c On entry:
c  vx(3)      vector defining x axis direction
c            (when projected along plane_normal)
c  plane_normal(3)  vector defining z axis
c
c On exit:
c  rmat(3,3)  transformation matrix
c  istat      =  0  success
c             = -1  failure (vx parallel to plane_normal or zero length)
c
      integer istat
      real vx(3), plane_normal(3), rmat(3,3)
c
      real vmat(3,3), det
c
c v3 along plane_normal
      call vnorm(vmat(1,3), plane_normal, istat)
      if (istat .ne. 0) return
c v2 = v3 x v1, normalized
      call cross(vmat(1,2), vmat(1,3), vx)
      call vnorm(vmat(1,2), vmat(1,2), istat)
      if (istat .ne. 0) return
c v1 = v2 x v3
      call cross(vmat(1,1), vmat(1,2), vmat(1,3))
      call vnorm(vmat(1,1), vmat(1,1), istat)
      if (istat .ne. 0) return
c
c  We want R such that R V = I, ie R = V**-1
      call minv3(rmat, vmat, det)
      if (det .eq. 0.0) then
         istat = -1
      else
         istat = 0
      endif
      return
      end
c
c
c
      subroutine vnorm(a, b, istat)
c     ============================
c
c  Normalise 3 vector b -> a
c
c On entry:
c  b(3)  vector to normalize
c On exit:
c  a(3)  normalized vector
c  istat = 0 OK, = -1 length zero
c
      integer istat
      real a(3), b(3)
c
      integer i
      real s
c
      istat = -1
      s = b(1)**2 + b(2)**2 + b(3)**2
      if (s .gt. 0.0) then
         s = sqrt(s)
         do 1, i = 1,3
            a(i) = b(i)/s
 1       continue 
         istat = 0
      endif
      return
      end
c
      
c
c
c
      subroutine ljstfy(a,b)
c     ======================
c
c Left justify string b into a
c
      character*(*) a,b
      integer j,lenstr
      external lenstr
c
      do 1, j = 1,lenstr(b)
         if (b(j:j) .ne. ' ') go to 3
 1    continue 
      j = 1
 3    a = b(j:)
      return
      end

C+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
C
C***********************
C**     INTFP         **
C***********************
C
C
       INTEGER FUNCTION INTFP (CHSTR, FP, IV) 
C
C PURPOSE:    Decode a character string into a number (floating and integer)
C
C AUTHOR:     John W. Campbell
C
C RETURN:     = 0  Syntax error in the number
C             = 1  integer  (no decimal point present)
C             = 2  floating point (decimal point present or E format)
C             = -1 blank string
C                                             
C ARGUMENTS:
C
C      CHSTR    (R)     Character string containing the number to be decoded
C                       (Leading spaces ignored; further space terminates
C                        the number string interpreted)
C      FP       (W)     Returns the value as a floating point value (0.0 if
C                       syntax error)
C      IV       (W)     Returns the value as an integer (the nearest integer
C                       if value contained a decimal point) (0 if syntax error)
C
C
C Declarations
C ============
C
C LOCALS:
C
      CHARACTER*(*) CHSTR
      CHARACTER*15 DIG
      LOGICAL END, START
      DOUBLE PRECISION FPVAL, X, S
      real fp
      integer iv, iexp, m1, m2, i, imax, id, ifdec, j, iex
      DATA DIG /'eE0123456789.+-'/
C
C-------------------------------------------------------------------------------
C
C
C====== Initialisations
C
      FPVAL = 0.0
      FP = 0.0
      IV = 0
      END = .FALSE.
      START = .FALSE.
      IEXP = 0
      M1 = 3
      M2 = 15
      I = 0
      IMAX = LEN (CHSTR) 
      IF (CHSTR.EQ.' ') THEN
         INTFP = -1
         RETURN
      ENDIF
C
C====== Interpret Number
C
100   S = 1.0
      ID = 0
      IFDEC = 0
      X = 0.0
120   I = I + 1
      IF (I.GT.IMAX) THEN
         END = .TRUE.
         IF (IEXP.EQ.1) GO TO 600
         GO TO 500
      ENDIF
      IF (CHSTR(I:I).EQ.' ') THEN
         IF (START) I = IMAX
         GO TO 120
      ENDIF
      J = INDEX(DIG(M1:M2),CHSTR(I:I))
      IF (J.EQ.0) GO TO 800
      START = .TRUE.
      J = J + M1 - 1
      M1 = 1 + 2*IEXP
C
C====== Digit
C
      IF (J.GE.3.AND.J.LE.12) THEN
         ID = ID + IFDEC
         X = 10.0 * X + J - 3
         IF (M2.GT.13) M2 = 13
         GO TO 120
C
C====== Decimal point
C
      ELSE IF (J.EQ.13) THEN
         IF (IEXP.EQ.1) GO TO 800
         IFDEC = 1
         M2 = 12
         GO TO 120
C
C====== Sign
C
      ELSE IF (J.GE.14) THEN
         IF (J.EQ.15) S = -1.0
         M2 = 13
         GO TO 120
C
C====== Exponent
C
      ELSE
         IEXP = 1
         M1 = 3
         M2 = 15
      ENDIF
C
C====== Evaluate number
C
500   FPVAL = S * X / (10**ID)
      IF (END) THEN
         FP = FPVAL
         IV = NINT (FP)
         INTFP = 1
         IF (IFDEC.GT.0) INTFP = 2
         RETURN
      ENDIF
      GO TO 100
C
C====== Evaluate and apply exponent
C
600   IEX = NINT (S*X)
      FPVAL = FPVAL * 10.0**IEX
      FP = FPVAL
      IV = NINT (FP)
      INTFP = 2
      RETURN
C
C====== Error Condition
C
800   INTFP = 0
      RETURN
      END
C
C - - - - - - - - - - - -
      SUBROUTINE EIGN3(AMAT,VECS,VALS)
      real AMAT(3,3),VECS(3,3),VALS(3)
     $   ,WORK(5,3) 
C ONLY LOWER TRIANGLE OF AMAT(I,J) (IE,I.GE.J)   NEED BE PROVIDED 
C RETURNS EIG-VALS IN DESC ORDER IN VALS; 
C CORRESP  VECS AS COLS OF VECS 
C RETURNS AMAT AS SYMMETRIC FULL MATRIX 
C REQUIRES LIBRARY=HARWELL.FORTLIB FOR EA06C
c
      integer n, i1p, i1, ih, i, j
      real v
c
      CALL EA06C(AMAT,VALS,VECS,3,3,3,WORK) 
C RETURNS VALS & VECS IN ARBITRARY ORDER & SCREWS UP 
C UPPER TRIANGLE OF AMAT 
C ORDER VALS,VECS, THEN SYMMETRIZE AMAT
      N=3
      I1P=1
 10   I1=I1P
      I1P=I1P+1
      IF(I1P.GT.N) GO TO 60
      IH=I1
      V=VALS(IH)
      DO 20 I=I1P,N
        IF(VALS(I).LE.V) GO TO 20
        IH=I
        V=VALS(IH)
 20   CONTINUE
      IF(IH.EQ.I1) GO TO 10
      V=VALS(I1)
      VALS(I1)=VALS(IH)
      VALS(IH)=V
      DO 30 J=1,N
        V=VECS(J,I1)
        VECS(J,I1)=VECS(J,IH)
 30     VECS(J,IH)=V
      GO TO 10

 60   CONTINUE
      DO 70 I=1,N
         DO 70 J=1,I
            AMAT(J,I)=AMAT(I,J)
 70   CONTINUE 
      RETURN
      END
