C
C     ROTAMER
C     Copyright (C) 2002 Dirk Kostrewa
C
C     This code is distributed under the terms and conditions of the
C     CCP4 Program Suite Licence Agreement as a CCP4 Application.
C     A copy of the CCP4 licence can be obtained by writing to the
C     CCP4 Secretary, Daresbury Laboratory, Warrington WA4 4AD, UK.
C
C
C
C---------------------------------------------------------------------------
C
C     ROTAMER
C
C---------------------------------------------------------------------------
C
C This program reads a file with protein coordinates in PDB format, analyses
C for each amino acid residue how much the observed side chain torsion angles
C deviate from the nearest rotamer and lists outlier residues.
C The rotamers are taken from the "Penultimate Rotamer Library", Table I, "common-atom
C values" (Lovell, S.C., Word, J.M., Richardson, J.S. & Richardson, D.C. (2000)
C "The penultimate rotamer library", Proteins: Structure, Function and Genetics, Vol.40,
C 389-408).
C The nearest rotamer is determined in a hierarchical order:
C first, all rotamers deviating less than the user defined threshold in Chi(1)
C are chosen, then of those all rotamers that deviate less in Chi(2) are
C chosen, then of those all rotamers that deviate less in Chi(3) are chosen,
C and finally all rotamers that deviate less in Chi(4) are chosen. If there
C are no rotamers deviating less than the user defined threshold for a given
C Chi(i) then the previous rotamer subset for Chi(i-1)is chosen and the hierarchical
C search is terminated. If this happens for Chi(1) then all rotamers are chosen as subset.
C Second, the nearest rotamer is the one with the minimum sum over the absolute
C Chi angle deviation over the subset of rotamers.
C A residue is listed as outlier if it has at least one observed Chi(i) that deviates more
C than the user defined threshold from the equivalent Chi(i) of the nearest rotamer.
C The output flag has in its i_th position a "*" if |Delta-Chi(i)| > Threshold
C and a "o" otherwise.
C
C The actual distribution of Chi(2) angles for Phe and Tyr appears to be very symmetric
C with respect to the sign of Chi(2) (the twofold rotational symmetry Chi(2)=180-Chi(2)
C is a question of torsion angle definition). Thus, I've added for each Phe and Tyr rotamer
C Chi(1),Chi(2) a symmetric one Chi(1),-Chi(2).
C 
C Some remarks about the proline rotamers: I think it would be much better to check
C for possible ring pucker modes. Of the two listed Chi-angles, Chi(1) and Chi(2), I found
C the check against the two possible Chi(1)-values of some use, whereas the check against 
C the Chi(2)-value range rarely gave useful results. Therefore, I check only the Chi1 angle
C for prolines.
C
C Known limitations:
C ==================
C 1). The program simply stops if an atom needed for the calculation of a side chain 
C     torsion angle is missing.
C
C MDW: changed to warning message only
C
C 2). The calulated side chain torsion angles for residues with multiple side chain 
C     conformations depend on the order of the atoms: only the last atoms of all alternate
C     conformations are used in the calculation! Thus, in the case that the order of atoms 
C     does not vary among alternate conformations (i.e. for two atoms A1,B1,A2,B2 or A1,A2,B1,B2),
C     the reported torsion angles are those for the last conformation, whereas in the case
C     that the order of atoms does vary for alternate conformations (i.e. for two atoms A1,B1,B2,A2
C     or B1,A1,A2,B2) the reported torsion angles are nonsense.
C
C===========================================================================
C  Start main program
C===========================================================================

	program rotamer
C	===============

C===========================================================================
C  Declarations
C===========================================================================

	implicit none

	integer max_aa,max_rot
	parameter ( max_aa = 18, max_rot = 34 )

	character*4 atnam,resnam,resno,segid,atid
	character*1 chnnam,altcod,inscod
	character resnamprev*3,chnnamprev*1,inscodprev*1
	character name*3,flag*4
	integer ifail,iser,ixyzin,iz
	integer i,j,nres,nresprev,id,idprev,rota,nrota,nchi,chi(4)
	integer idistmin,idist,nearest_rota,jsubset,ideltachi,ithreshold
	integer idefault
	real xyz(3),occ,biso,u(6),cell(6)
	real xyz1(3),xyz2(3),xyz3(3),xyz4(3),xyz5(3),xyz6(3),xyz7(3),tor
	logical outlier,firstresidue,lastresidue,rotasubset,emptysubset
	logical cont
	dimension name(max_aa),nrota(max_aa),nchi(max_aa)
	dimension rota(max_aa,max_rot,4),rotasubset(0:4,max_rot)
	dimension outlier(4)

C===========================================================================
C  Data
C===========================================================================

C Amino acid names with rotamer data

	data name /'ARG','LYS','MET','GLU','GLN','ASP','ASN','ILE',
     &             'LEU','HIS','TRP','TYR','PHE','PRO','THR','VAL',
     &             'SER','CYS'/

	data nrota /34,27,13,8,9,5,7,7,5,8,7,8,8,2,3,3,3,3/

	data nchi /4,4,3,3,3,2,2,2,2,2,2,2,2,1,1,1,1,1/

C Arginine rotamer data (residue 1, 34 rotamers, 4 torsion angles)

	data ((rota(1,i,j), j=1,4), i=1,34)
     &	      /    62,  180,   65,   85,
     &	           62,  180,   65, -175,
     &	           62,  180,  180,   85,
     &	           62,  180,  180,  180,
     &	           62,  180,  180,  -85,
     &	           62,  180,  -65,  175,
     &	           62,  180,  -65,  -85,
     &	         -177,   65,   65,   85,
     &	         -177,   65,   65, -175,
     &	         -177,   65,  180,   85,
     &	         -177,   65,  180,  180,
     &	         -177,  180,   65,   85,
     &	         -177,  180,   65, -175,
     &	         -177,  180,   65, -105,
     &	         -177,  180,  180,   85,
     &	         -177,  180,  180,  180,
     &	         -177,  180,  180,  -85,
     &	         -177,  180,  -65,  105,
     &	         -177,  180,  -65,  175,
     &	         -177,  180,  -65,  -85,
     &	          -67,  180,   65,   85,
     &	          -67,  180,   65, -175,
     &	          -67,  180,   65, -105,
     &	          -67,  180,  180,   85,
     &	          -67,  180,  180,  180,
     &	          -67,  180,  180,  -85,
     &	          -67,  180,  -65,  105,
     &	          -67,  180,  -65,  175,
     &	          -67, -167,  -65,  -85,
     &	          -62,  -68,  180,   85,
     &	          -62,  -68,  180,  180,
     &	          -62,  -68,  180,  -85,
     &	          -62,  -68,  -65,  175,
     &	          -62,  -68,  -65,  -85 /

C Lysine rotamer data (residue 2, 27 rotamers, 4 torsion angles)

	data ((rota(2,i,j), j=1,4), i=1,27)
     &	      /    62,  180,   68,  180,
     &	           62,  180,  180,   65,
     &	           62,  180,  180,  180,
     &	           62,  180,  180,  -65,
     &	           62,  180,  -68,  180,
     &	         -177,   68,  180,   65,
     &	         -177,   68,  180,  180,
     &	         -177,   68,  180,  -65,
     &	         -177,  180,   68,   65,
     &	         -177,  180,   68,  180,
     &	         -177,  180,  180,   65,
     &	         -177,  180,  180,  180,
     &	         -177,  180,  180,  -65,
     &	         -177,  180,  -68,  180,
     &	         -177,  180,  -68,  -65,
     &	          -90,   68,  180,  180,
     &	          -67,  180,   68,   65,
     &	          -67,  180,   68,  180,
     &	          -67,  180,  180,   65,
     &	          -67,  180,  180,  180,
     &	          -67,  180,  180,  -65,
     &	          -67,  180,  -68,  180,
     &	          -67,  180,  -68,  -65,
     &	          -62,  -68,  180,   65,
     &	          -62,  -68,  180,  180,
     &	          -62,  -68,  180,  -65,
     &	          -62,  -68,  -68,  180 /

C Methionine rotamer data (residue 3, 13 rotamers, 3 torsion angles)

	data ((rota(3,i,j), j=1,3), i=1,13)
     &	      /    62,  180,   75,
     &	           62,  180,  -75,
     &	         -177,   65,   75,
     &	         -177,   65,  180,
     &	         -177,  180,   75,
     &	         -177,  180,  180,
     &	         -177,  180,  -75,
     &	          -67,  180,   75,
     &	          -67,  180,  180,
     &	          -67,  180,  -75,
     &	          -65,  -65,  103,
     &	          -65,  -65,  180,
     &	          -65,  -65,  -70 /

C Glutamate rotamer data (residue 4, 8 rotamers, 3 torsion angles)

	data ((rota(4,i,j), j=1,3), i=1,8)
     &	      /    62,  180,  -20,
     &	           70,  -80,    0,
     &	         -177,   65,   10,
     &	         -177,  180,    0,
     &	         -177,  -80,  -25,
     &	          -65,   85,    0,
     &	          -67,  180,  -10,
     &	          -65,  -65,  -40 /

C Glutamine rotamer data (residue 5, 9 rotamers, 3 torsion angles)

	data ((rota(5,i,j), j=1,3), i=1,9)
     &	      /    62,  180,   20,
     &	           70,  -75,    0,
     &	         -177,   65, -100,
     &	         -177,   65,   60,
     &	         -177,  180,    0,
     &	          -65,   85,    0,
     &	          -67,  180,  -25,
     &	          -65,  -65,  -40,
     &	          -65,  -65,  100 /

C Aspartate rotamer data (residue 6, 5 rotamers, 2 torsion angles)

	data ((rota(6,i,j), j=1,2), i=1,5)
     &	      /    62,  -10,
     &	           62,   30,
     &	         -177,    0,
     &	         -177,   65,
     &	          -70,  -15 /

C Asparagine rotamer data (residue 7, 7 rotamers, 2 torsion angles)

	data ((rota(7,i,j), j=1,2), i=1,7)
     &	      /    62,  -10,
     &	           62,   30,
     &	         -174,  -20,
     &	         -177,   30,
     &	          -65,  -20,
     &	          -65,  -75,
     &	          -65,  120 /

C Isoleucine rotamer data (residue 8, 7 rotamers, 2 torsion angles)

	data ((rota(8,i,j), j=1,2), i=1,7)
     &	      /    62,  100,
     &	           62,  170,
     &	         -177,   66,
     &	         -177,  165,
     &	          -65,  100,
     &	          -65,  170,
     &	          -57,  -60 /

C Leucine rotamer data (residue 9, 5 rotamers, 2 torsion angles)

	data ((rota(9,i,j), j=1,2), i=1,5)
     &	      /    62,   80,
     &	         -177,   65,
     &	         -172,  145,
     &	          -85,   65,
     &	          -65,  175 /

C Histidine rotamer data (residue 10, 8 rotamers, 2 torsion angles)

	data ((rota(10,i,j), j=1,2), i=1,8)
     &	      /    62,  -75,
     &	           62,   80,
     &	         -177, -165,
     &	         -177,  -80,
     &	         -177,   60,
     &	          -65,  -70,
     &	          -65,  165,
     &	          -65,   80 /

C Tryptophane rotamer data (residue 11, 7 rotamers, 2 torsion angles)

	data ((rota(11,i,j), j=1,2), i=1,7)
     &	      /    62,  -90,
     &	           62,   90,
     &	         -177, -105,
     &	         -177,   90,
     &	          -65,  -90,
     &	          -65,   -5,
     &	          -65,   95 /

C Tyrosine rotamer data (residue 12, 2*4 rotamers, 2 torsion angles)

	data ((rota(12,i,j), j=1,2), i=1,8)
     &	      /    62,   90,
     &	           62,  -90,
     &	         -177,   80,
     &	         -177,  -80,
     &	          -65,  -85,
     &	          -65,   85,
     &	          -65,  -30,
     &	          -65,   30 /

C Phenylalanine rotamer data (residue 13, 2*4 rotamers, 2 torsion angles)

	data ((rota(13,i,j), j=1,2), i=1,8)
     &	      /    62,   90,
     &	           62,  -90,
     &	         -177,   80,
     &	         -177,  -80,
     &	          -65,  -85,
     &	          -65,   85,
     &	          -65,  -30,
     &	          -65,   30 /

C Proline rotamer data (residue 14, 3 tabulated rotamers, but I use only two rotamers in Chi1)

	data ((rota(14,i,j), j=1,1), i=1,2)
     &	      /    30,
     &	          -30 /

C Threonine rotamer data (residue 15, 3 rotamers, 1 torsion angle)

	data ((rota(15,i,j), j=1,1), i=1,3)
     &	      /    62,
     &	         -175,
     &	          -65 /

C Valine rotamer data (residue 16, 3 rotamers, 1 torsion angle)

	data ((rota(16,i,j), j=1,1), i=1,3)
     &	      /    63,
     &	          175,
     &	          -60 /

C Serine rotamer data (residue 17, 3 rotamers, 1 torsion angle)

	data ((rota(17,i,j), j=1,1), i=1,3)
     &	      /    62,
     &	         -177,
     &	          -65 /

C Cysteine rotamer data (residue 18, 3 rotamers, 1 torsion angle)

	data ((rota(18,i,j), j=1,1), i=1,3)
     &	      /    62,
     &	         -177,
     &	          -65 /

C===========================================================================
C  General initializations
C===========================================================================

	call ccpfyp
	call ccprcs (6,'ROTAMER','$Date: 2005/09/06 11:25:23 $')


C===========================================================================
C  PDB file initializations
C===========================================================================

	call xyzinit ()
	ixyzin = 1
	ifail = 0
	call xyzopen ('XYZIN','INPUT',' ',ixyzin,ifail)

C===========================================================================
C  Variable initializations
C===========================================================================

	firstresidue=.true.
	lastresidue=.false.
	idefault=30
	ithreshold=idefault

C===========================================================================
C  Read keywords
C===========================================================================

 5	call memoparse (.true.)
	call parseint ('DELT',ithreshold)

	write (6,'(a,i3)') 'DELTachi set to ',ithreshold

	if (ithreshold.lt.0.or.ithreshold.gt.180) then
	  write (6,'(/2a)') 'WARNING: DELTachi out of allowed range',
     &'[0,180]'
	  ithreshold=idefault
	  write (6,'(a,i3/)') 'WARNING: DELTachi reset to ',idefault
	end if

	call parsediagnose (cont)
        if (cont) goto 5

C===========================================================================
C  Write header of output table
C===========================================================================

	write (6,'(/a,i3/)') 'Rotamer list with |Delta-Chi| = ',
     &	                     ithreshold

	write (6,'(2a)') '  Residue  Flag      Chi1         Chi2',
     &'         Chi3         Chi4'
	write (6,'(2a)') '==========================================',
     &'========================='

C===========================================================================
C  Read coordinate records and loop back to here until the end 
C===========================================================================

   10	continue

	call xyzadvance( ixyzin,0,0,*10,*100)
	call xyzatom (ixyzin,iser,atnam,resnam,chnnam,nres,
     &	     resno,inscod,altcod,segid,iz,atid)
	call xyzcoord (ixyzin,'O','U',xyz(1),xyz(2),xyz(3),occ,biso,u)

	if (u(2).ne.0.0.or.u(3).ne.0.0) then
	if (xyz(1).eq.0.0.and.xyz(2).eq.0.0.and.xyz(3).eq.0.0) goto 10
	end if

C skip amino acids without rotamer data

	do i=1,max_aa
	  if (resnam.eq.name(i)) then
	    id=i
	    goto 20
	  end if
	end do

	goto 10

C continue with valid amino acid

   20	continue

C get variables for first amino acid

	if (firstresidue) then
	  resnamprev=resnam(1:3)
	  chnnamprev=chnnam
	  nresprev=nres
	  inscodprev=inscod
	  idprev=id
	  do j=1,3
	    xyz1(j)=-9999.
	    xyz2(j)=-9999.
	    xyz3(j)=-9999.
	    xyz4(j)=-9999.
	    xyz5(j)=-9999.
	    xyz6(j)=-9999.
	    xyz7(j)=-9999.
	  end do
	  firstresidue=.false.
	end if

C if not a new amino acid get all relevant coordinates
C beware: the atnam from the ccp4 pdb coordinate read routine is left-shifted, i.e.
C 'CG1 ' instead of the pdb definition ' CG1'!

 31   if ((nres.eq.nresprev).and.
     &      (inscod.eq.inscodprev).and.
     &      (.not.lastresidue)) then

	  if (atnam.eq.'N   ') then
	    do i=1,3
	      xyz1(i)=xyz(i)
	    end do
	  else if (atnam.eq.'CA  ') then
	    do i=1,3
	      xyz2(i)=xyz(i)
	    end do
	  else if (atnam.eq.'CB  ') then
	    do i=1,3
	      xyz3(i)=xyz(i)
	    end do
	  else if (atnam(2:3).eq.'G '.or.atnam(2:3).eq.'G1') then
	    do i=1,3
	      xyz4(i)=xyz(i)
	    end do
	  else if (nchi(id).gt.1.and.
     &    (atnam(2:3).eq.'D '.or.atnam(2:3).eq.'D1')) then
	    do i=1,3
	      xyz5(i)=xyz(i)
	    end do
	  else if (nchi(id).gt.2.and.
     &    (atnam(2:3).eq.'E '.or.atnam(2:3).eq.'E1')) then
	    do i=1,3
	      xyz6(i)=xyz(i)
	    end do
	  else if (nchi(id).gt.3.and.atnam(2:3).eq.'Z ') then
	    do i=1,3
	      xyz7(i)=xyz(i)
	    end do
	  end if

C else: calculate torsion angles, find nearest rotamer
C and list any outlier(s) for that amino acid

	else

C also jump here for calculations for last amino acid

   30	  continue

C Chi1

	  do j=1,3
	    if (xyz1(j).eq.-9999. .or. xyz2(j).eq.-9999. .or. 
     &          xyz3(j).eq.-9999. .or. xyz4(j).eq.-9999.) then 
	      write (6,'(a,a1,i4,a1,a3,a)') 'Missing atom for residue ',
     &                    chnnamprev,nresprev,inscodprev,resnamprev,
     &                    ' - residue skipped'
              goto 50  
	    end if
	  end do
	  call torsion(xyz1,xyz2,xyz3,xyz4,tor)
	  chi(1)=nint(tor)

C Chi2

	  if (nchi(idprev).gt.1) then
	    do j=1,3
	      if (xyz5(j).eq.-9999.) then 
	      write (6,'(a,a1,i4,a1,a3,a)') 'Missing atom for residue ',
     &                    chnnamprev,nresprev,inscodprev,resnamprev,
     &                    ' - residue skipped'
              goto 50  
	      end if
	    end do
	    call torsion(xyz2,xyz3,xyz4,xyz5,tor)
	    chi(2)=nint(tor)
	  end if

C Chi3

	  if (nchi(idprev).gt.2) then
	    do j=1,3
	      if (xyz6(j).eq.-9999.) then 
	      write (6,'(a,a1,i4,a1,a3,a)') 'Missing atom for residue ',
     &                    chnnamprev,nresprev,inscodprev,resnamprev,
     &                    ' - residue skipped'
              goto 50  
	      end if
	    end do
	    call torsion(xyz3,xyz4,xyz5,xyz6,tor)
	    chi(3)=nint(tor)
	  end if

C Chi4

	  if (nchi(idprev).gt.3) then
	    do j=1,3
	      if (xyz7(j).eq.-9999.) then 
	      write (6,'(a,a1,i4,a1,a3,a)') 'Missing atom for residue ',
     &                    chnnamprev,nresprev,inscodprev,resnamprev,
     &                    ' - residue skipped'
              goto 50  
	      end if
	    end do
	    call torsion(xyz4,xyz5,xyz6,xyz7,tor)
	    chi(4)=nint(tor)
	  end if

C The nearest rotamer is determined in a hierarchical order:
C first, all rotamers deviating less than the user defined threshold in Chi(1)
C are chosen, then of those all rotamers that deviate less in Chi(2) are
C chosen, then of those all rotamers that deviate less in Chi(3) are chosen,
C and finally all rotamers that deviate less in Chi(4) are chosen. If there
C are no rotamers deviating less than the user defined threshold for a given
C Chi(i) then the previous rotamer subset for Chi(i-1)is chosen and the hierarchical
C search is terminated. If this happens for Chi(1) then all rotamers are chosen as subset.
C Second, the nearest rotamer is the one with the minimum sum over the absolute
C Chi angle deviation over the subset of rotamers.

C initialize the rotamer subsets

	  do i=1,max_rot
	    rotasubset(0,i)=.true.
	  end do
	  do j=1,4
	    do i=1,max_rot
	      rotasubset(j,i)=.false.
	    end do
	  end do

C set all rotamer of subset j-1 to TRUE in subset j if it deviates less than the threshold
C if the current subset j is empty, choose subset j-1 and terminate hierarchical search

	  do j=1,nchi(idprev)

	    emptysubset=.true.

	    do i=1,nrota(idprev)
	      if (rotasubset(j-1,i)) then
	        ideltachi=abs(chi(j)-rota(idprev,i,j))
	        if (ideltachi.gt.180) ideltachi=360-ideltachi
	        if (ideltachi.le.ithreshold) then
	          rotasubset(j,i)=.true.
	          emptysubset=.false.
	        end if
	      end if
	    end do

	    if (emptysubset) then
	      jsubset=j-1
	      goto 40
	    else
	      jsubset=j
	    end if

	  end do

C find nearest rotamer from subset of rotamers

   40	  continue
	  idistmin=9999

	  do i=1,nrota(idprev)
	    if (rotasubset(jsubset,i)) then
	      idist=0
	      do j=1,nchi(idprev)
	        ideltachi=abs(chi(j)-rota(idprev,i,j))
	        if (ideltachi.gt.180) ideltachi=360-ideltachi
	        idist=idist+ideltachi
	      end do
	      if (idist.lt.idistmin) then
	        idistmin=idist
	        nearest_rota=i
	      end if
	    end if
	  end do

C Flag every outlier torsion angle.
C The output flag has in its i_th position a "*" if |Delta-Chi(i)| > Threshold
C and a "o" otherwise.

	  flag='    '
	  do j=1,4
	    outlier(j)=.false.
	  end do

	  do j=1,nchi(idprev)
	    ideltachi=abs(chi(j)-rota(idprev,nearest_rota,j))
	    if (ideltachi.gt.180) ideltachi=360-ideltachi
	    if (ideltachi.gt.ithreshold) then
	      outlier(j)=.true.
	      flag(j:j)='*'
	    else
	      flag(j:j)='o'
	    end if
	  end do

C list amino acids with any outlier

	  if (outlier(1).or.outlier(2).or.outlier(3).or.outlier(4)) then
	    if (nchi(idprev).eq.1) then
	      write (6,'(a1,i4,a1,a3,2x,a4,1(2x,i4," (",i4,")"))')
     &          chnnamprev,nresprev,inscodprev,resnamprev,flag,
     &          (chi(j),(rota(idprev,nearest_rota,j)),j=1,nchi(idprev))
	    else if (nchi(idprev).eq.2) then
	      write (6,'(a1,i4,a1,a3,2x,a4,2(2x,i4," (",i4,")"))')
     &          chnnamprev,nresprev,inscodprev,resnamprev,flag,
     &          (chi(j),(rota(idprev,nearest_rota,j)),j=1,nchi(idprev))
	    else if (nchi(idprev).eq.3) then
	      write (6,'(a1,i4,a1,a3,2x,a4,3(2x,i4," (",i4,")"))')
     &          chnnamprev,nresprev,inscodprev,resnamprev,flag,
     &          (chi(j),(rota(idprev,nearest_rota,j)),j=1,nchi(idprev))
	    else if (nchi(idprev).eq.4) then
	      write (6,'(a1,i4,a1,a3,2x,a4,4(2x,i4," (",i4,")"))')
     &          chnnamprev,nresprev,inscodprev,resnamprev,flag,
     &          (chi(j),(rota(idprev,nearest_rota,j)),j=1,nchi(idprev))
	    end if
	  end if

 50	  continue

C set variables for next amino acid

	   resnamprev=resnam(1:3)
	   chnnamprev=chnnam
	   nresprev=nres
	   inscodprev=inscod
	   idprev=id

	   do j=1,3
	     xyz1(j)=-9999.
	     xyz2(j)=-9999.
	     xyz3(j)=-9999.
	     xyz4(j)=-9999.
	     xyz5(j)=-9999.
	     xyz6(j)=-9999.
	     xyz7(j)=-9999.
	   end do

	   call xyzbksp (ixyzin)

	end if

	if (.not.lastresidue) then
	  goto 10
	else
	  goto 110
	end if

C do calculations for last residue

  100	continue

	lastresidue=.true.
	goto 31

  110	continue

	write (6,'(/)')
	call xyzclose(ixyzin)
	call ccperr(0,' - Normal termination of ROTAMER')

	end

C===========================================================================
C  End main program
C===========================================================================


C===========================================================================
C  Start subroutine for calculating torsion angle
C===========================================================================

C This subroutine calculates the torsion angle t between atoms with orthogonal
C coordinate sets a(3),b(3),c(3),d(3)

	subroutine torsion(a,b,c,d,t)

	real a(3),b(3),c(3),d(3),t
	real ab(3),bc(3),cd(3),abxbc(3),bcxcd(3),cross(3)
	real lenabxbc,lenbcxcd,scalar,triple,cost,pi

	lenabxbc=0.
	lenbcxcd=0.
	scalar=0.
	triple=0.

	do i=1,3
	   abxbc(i)=0.
	   bcxcd(i)=0.
	   cross(i)=0.
	end do

	pi=3.141592654

C calculate bond vectors (AB),(BC),(CD)

	do i=1,3
	   ab(i)=b(i)-a(i)
	   bc(i)=c(i)-b(i)
	   cd(i)=d(i)-c(i)
	end do

C calculate vector products (AB)x(BC),(BC)x(CD)

	abxbc(1)=ab(2)*bc(3)-ab(3)*bc(2)
	abxbc(2)=ab(3)*bc(1)-ab(1)*bc(3)
	abxbc(3)=ab(1)*bc(2)-ab(2)*bc(1)
	bcxcd(1)=bc(2)*cd(3)-bc(3)*cd(2)
	bcxcd(2)=bc(3)*cd(1)-bc(1)*cd(3)
	bcxcd(3)=bc(1)*cd(2)-bc(2)*cd(1)

C calculate lengths of vector products (AB)x(BC),(BC)x(CD)

	do i=1,3
	   lenabxbc=lenabxbc+abxbc(i)*abxbc(i)
	   lenbcxcd=lenbcxcd+bcxcd(i)*bcxcd(i)
	end do

	lenabxbc=sqrt(lenabxbc)
	lenbcxcd=sqrt(lenbcxcd)
	
C calculate scalar product [(AB)x(BC)]*[(BC)x(CD)]

	do i=1,3
	   scalar=scalar+abxbc(i)*bcxcd(i)
	end do

C calculate vector product [(AB)x(BC)]x[(BC)x(CD)]

	cross(1)=abxbc(2)*bcxcd(3)-abxbc(3)*bcxcd(2)
	cross(2)=abxbc(3)*bcxcd(1)-abxbc(1)*bcxcd(3)
	cross(3)=abxbc(1)*bcxcd(2)-abxbc(2)*bcxcd(1)

C calculate triple product [(AB)x(BC)]x[(BC)x(CD)]*(BC)

	do i=1,3
	   triple=triple+cross(i)*bc(i)
	end do

C calculate cos(torsion)

	cost=scalar/(lenabxbc*lenbcxcd)

C calculate torsion angle with the sign from the triple product

	t=sign(180./pi*acos(cost),triple)

	return
	end

C===========================================================================
C  End subroutine
C===========================================================================
