C
C
C     This code is distributed under the terms and conditions of the
C     CCP4 Program Suite Licence Agreement as a CCP4 Application.
C     A copy of the CCP4 licence can be obtained by writing to the
C     CCP4 Secretary, Daresbury Laboratory, Warrington WA4 4AD, UK.
C
C Package LEXANA
C
C Lexical analysis of text, string or file.
C
C Definitions used in this package:
C
C Separators are characters without any syntactical meaning except
C for separating items. Delimiters are characters that have syntactical
C meaning, as well as delimiting words, such as '=' and ')' in Fortran.
C
C String delimiters enclose a string (which may contain any characters
C except string delimiters). A comment delimiter indicates the start
C of a comment, and the end-of-line is end of comment.
C
C An item is either a word between separators, a delimiter character,
C or a string of characters (string delimiters included). End-of-line
C and end-of-record in files are always separators, and strings may not
C contain such characters.
C
C Warning: results will be unpredictable if LXINIT is not called first.
C
C Note: ascii (tab) = 9, ascii (') = 39
C
C LXINIT s   set separator, delimiter, string and comment character sets
C LXCURR s   return current character sets
C LXLINE s   analyze whole line, find items
C LXFILE s   analyze formatted file, return items one by one
C LXNEXT s   find next item in line
C LXDATA o   character lexical type code table
C
C Uses no separately compiled procedures.
C
C Copyright (C) 1987, 1990 Per Kraulis
C
C Per Kraulis, Dept Molecular Biology, Uppsala University Sweden.
C   2-May-1987  main code written
C  27-Jun-1987  get and split actions into separate procedures
C  15-Sep-1987  rearrangement, only parse (not split), tokens added
C  13-Oct-1987  added removal of comment in line
C  21-Apr-1988  added parsing of formatted file
C  30-May-1988  terminology: this is lexical analysis, not parsing
C  18-Jun-1988  LXNEXT added and used in other procedures
C   2-Sep-1988  LXCURR added
C   9-Jan-1990  major rewrite; string, comment delimiter added;
C               LXSPLT and LXREMC removed; interface partly changed
C  26-Jun-1990  modified LXCURR to give 'NONE' when no characters in set
C
C
C----------------------------------------------------
      SUBROUTINE LXINIT (SEPARC, DELIMC, STRC, COMMC)
C
      CHARACTER*(*) SEPARC, DELIMC, STRC, COMMC
C
C SEPARC  (In)  string of separator characters, 'NONE' if none
C DELIMC  (In)  string of delimiter characters, 'NONE' if none
C STRC    (In)  string of string delimiter characters, 'NONE' if none
C COMMC   (In)  string of comment start characters, 'NONE' if none
C
C Character lexical set codes
C
      INTEGER    LXTOK, LXSEP, LXDEL, LXSTR, LXCOM
      PARAMETER (LXTOK = 0, LXSEP = 1, LXDEL = 2, LXSTR = 3, LXCOM = 4)
C
C Character set list
C
      INTEGER LXDSET (0:128)
C
      COMMON /LXDATA/ LXDSET
      SAVE   /LXDATA/
C
C Help variable
C
      INTEGER I
C
C All characters are ordinary token characters by default
C
      DO 10 I = 0, 128
        LXDSET (I) = LXTOK
10    CONTINUE
C
C Load character set codes
C
      IF (SEPARC .NE. 'NONE') THEN
        DO 100 I = 1, LEN (SEPARC)
          LXDSET (ICHAR (SEPARC (I:I))) = LXSEP
100     CONTINUE
      END IF
C
      IF (DELIMC .NE. 'NONE') THEN
        DO 200 I = 1, LEN (DELIMC)
          LXDSET (ICHAR (DELIMC (I:I))) = LXDEL
200     CONTINUE
      END IF
C
      IF (STRC .NE. 'NONE') THEN
        DO 300 I = 1, LEN (STRC)
          LXDSET (ICHAR (STRC (I:I))) = LXSTR
300     CONTINUE
      END IF
C
      IF (COMMC .NE. 'NONE') THEN
        DO 400 I = 1, LEN (COMMC)
          LXDSET (ICHAR (COMMC (I:I))) = LXCOM
400     CONTINUE
      END IF
C
      RETURN
      END

C
C
C---------------------------------------------------------
      SUBROUTINE LXCURR (SEPARC, SEPARN, DELIMC, DELIMN,
     $                   STRC, STRN, COMMC, COMMN, ERRCOD)
C
      CHARACTER*(*) SEPARC, DELIMC, STRC, COMMC
      INTEGER       SEPARN, DELIMN, STRN, COMMN, ERRCOD
C
C SEPARC  (Out) string of separator characters
C SEPARN  (Out) number of separator characters
C DELIMC  (Out) string of delimiter characters
C DELIMN  (Out) number of delimiter characters
C STRC    (Out) string of string delimiter characters
C STRN    (Out) number of string delimiter characters
C COMMC   (Out) string of comment start characters
C COMMN   (Out) number of comment start characters
C ERRCOD  (Out) error code
C     0 = no error
C     1 = some characters did not fit into string
C
C Character lexical set codes
C
      INTEGER    LXTOK, LXSEP, LXDEL, LXSTR, LXCOM
      PARAMETER (LXTOK = 0, LXSEP = 1, LXDEL = 2, LXSTR = 3, LXCOM = 4)
C
C Character set list
C
      INTEGER LXDSET (0:128)
C
      COMMON /LXDATA/ LXDSET
      SAVE   /LXDATA/
C
C Help variable
C
      INTEGER I
C
C Init
C
      SEPARN = 0
      DELIMN = 0
      STRN   = 0
      COMMN  = 0
C
C Loop through characters
C
      DO 10 I = 0, 128
C
        GOTO (100, 200, 300, 400) LXDSET (I)
C
        GOTO 10
C
  100   IF (SEPARN .GE. LEN (SEPARC)) GOTO 900
        SEPARN = SEPARN + 1
        SEPARC (SEPARN:SEPARN) = CHAR (I)
        GOTO 10
C
  200   IF (DELIMN .GE. LEN (DELIMC)) GOTO 900
        DELIMN = DELIMN + 1
        DELIMC (DELIMN:DELIMN) = CHAR (I)
        GOTO 10
C
  300   IF (STRN .GE. LEN (STRC)) GOTO 900
        STRN = STRN + 1
        STRC (STRN:STRN) = CHAR (I)
        GOTO 10
C
  400   IF (COMMN .GE. LEN (COMMC)) GOTO 900
        COMMN = COMMN + 1
        COMMC (COMMN:COMMN) = CHAR (I)
        GOTO 10
C
10    CONTINUE
C
C Check if any empty set
C
      IF (SEPARN .EQ. 0) THEN
        SEPARC = 'NONE'
        SEPARN = 4
      END IF
C
      IF (DELIMN .EQ. 0) THEN
        DELIMC = 'NONE'
        DELIMN = 4
      END IF
C
      IF (STRN .EQ. 0) THEN
        STRC = 'NONE'
        STRN = 4
      END IF
C
      IF (COMMN .EQ. 0) THEN
        COMMC = 'NONE'
        COMMN = 4
      END IF
C
      ERRCOD = 0
      RETURN
C
900   ERRCOD = 1
      RETURN
      END
C
C
C---------------------------------------------------------------------
      SUBROUTINE LXLINE (LINE, MAXITM, ITMPOS, ITMTYP, ITMNUM, ERRCOD)
C
      CHARACTER*(*) LINE
      INTEGER       MAXITM, ITMNUM, ERRCOD
      INTEGER       ITMPOS (2, MAXITM), ITMTYP (MAXITM)
C
C LINE    (In)  line to analyze
C MAXITM  (In)  maximum number of items
C ITMPOS  (Out) first and last character of items
C ITMTYP  (Out) item type codes
C     0 = ordinary token
C     1 = delimiter token
C     2 = string token
C ITMNUM  (Out) number of items
C ERRCOD  (Out) error code
C     0 = no error
C     1 = too many items in text line; the last skipped
C     2 = string not finished at end-of-line
C
C Help variables
C
      INTEGER LINLEN, ITYP
      INTEGER IPOS (2)
C
C Init
C
      LINLEN = LEN (LINE)
      IPOS (2) = 0
      ITMNUM = 0
C
C Get one item, and put its pointers and type into arrays
C
100   CALL LXNEXT (LINE, IPOS, ITYP, LINLEN, ERRCOD)
C
C No more items left in string; just return
C
      IF (ERRCOD .EQ. 1) THEN
        ERRCOD = 0
        RETURN
C
C String not finished
C
      ELSE IF (ERRCOD .EQ. 2) THEN
        RETURN
C
C If no more space in pointers array, return with failure
C
      ELSE IF (ITMNUM .GE. MAXITM) THEN
        ERRCOD = 1
        RETURN
      END IF
C
C Transfer pointers and item type to arrays
C
      ITMNUM = ITMNUM + 1
      ITMPOS (1, ITMNUM) = IPOS (1)
      ITMPOS (2, ITMNUM) = IPOS (2)
      ITMTYP (ITMNUM) = ITYP
C
C Return to get next item
C
      GOTO 100
C
      END
C
C
C--------------------------------------------------------------------
      SUBROUTINE LXFILE (LINE, IPOS, ITYP, LNUM, LINLEN, LUN, ERRCOD)
C
      CHARACTER*(*) LINE
      INTEGER       ITYP, LNUM, LUN, LINLEN, ERRCOD
      INTEGER       IPOS (2)
C
C LINE   (InOut) string to store latest line read from file
C IPOS   (InOut) first and last position of (latest) found item
C                IPOS (2) = 0 if no line read from file
C ITYP   (Out) item type
C     0 = ordinary token
C     1 = delimiter token
C     2 = string token
C LNUM   (Out) number of new lines read from file
C LINLEN (In)  length of string; limits record length readable from file
C LUN    (In)  logical unit to read from; must be opened formatted file
C ERRCOD (Out) error code
C     0 = no error
C     1 = end-of-file reached
C     2 = could not read from file
C     3 = string not finished at end-of-line
C
C Read new line from file if none read earlier
C
      IF (IPOS (2) .EQ. 0) THEN
        READ (LUN, 100, END=900, ERR=910) LINE
100     FORMAT (A)
        LNUM = 1
      ELSE
        LNUM = 0
      END IF
C
C Get next word or delimiter in string, and return if any found
C
200   CALL LXNEXT (LINE, IPOS, ITYP, LINLEN, ERRCOD)
C
C No more item in line; get next
C
      IF (ERRCOD .EQ. 1) THEN
        READ (LUN, 100, END=900, ERR=910) LINE
        LNUM = LNUM + 1
        IPOS (2) = 0
        GOTO 200
C
C String not finished in line
C
      ELSE IF (ERRCOD .EQ. 2) THEN
        ERRCOD = 3
        RETURN
      END IF
C
C Return if no error
C
      RETURN
C
900   ERRCOD = 1
      RETURN
910   ERRCOD = 2
      RETURN
      END
C
C
C---------------------------------------------------------
      SUBROUTINE LXNEXT (LINE, IPOS, ITYP, LINLEN, ERRCOD)
C
      CHARACTER*(*) LINE
      INTEGER       IPOS (2)
      INTEGER       ITYP, LINLEN, ERRCOD
C
C LINE    (In)  string to find next word or delimiter in
C IPOS    (InOut) first and last position of (latest) found item
C                 (0, 0) if new line to start processing
C ITYP    (Out) item type
C     0 = ordinary token
C     1 = delimiter token
C     2 = string token
C LINLEN  (In)  length of string
C ERRCOD  (Out) error code
C     0 = no error
C     1 = no more item in line
C     2 = string not finished at end-of-line
C
C Character lexical set codes
C
      INTEGER    LXTOK, LXSEP, LXDEL, LXSTR, LXCOM
      PARAMETER (LXTOK = 0, LXSEP = 1, LXDEL = 2, LXSTR = 3, LXCOM = 4)
C
C Character set list
C
      INTEGER LXDSET (0:128)
C
      COMMON /LXDATA/ LXDSET
      SAVE   /LXDATA/
C
C Help variable
C
      INTEGER CURC
C
C Find start of next item, if any, in line
C
      CURC = IPOS (2)
  100 CURC = CURC + 1
C
C End-of-line reached, no item returned
C
      IF (CURC .GT. LINLEN) THEN
        ERRCOD = 1
        RETURN
      END IF
C
C Act according to character type; if separator then just skip back
C
      GOTO (100, 200, 300, 400) LXDSET (ICHAR (LINE (CURC:CURC)))
C
C Ordinary token
C
      ITYP = 0
      IPOS (1) = CURC
      ERRCOD = 0
C
50    CURC = CURC + 1
C
C End-of-line reached
C
        IF (CURC .GT. LINLEN) THEN
          IPOS (2) = LINLEN
          RETURN
C
C Ordinary token finished
C
        ELSE IF (LXDSET (ICHAR (LINE (CURC:CURC))) .NE. LXTOK) THEN
          IPOS (2) = CURC - 1
          RETURN
        END IF
C
      GOTO 50
C
C Delimiter character; return it
C
  200 ITYP = 1
      IPOS (1) = CURC
      IPOS (2) = CURC
      ERRCOD = 0
      RETURN
C
C String item
C
  300 ITYP = 2
      IPOS (1) = CURC
C
350   CURC = CURC + 1
C
C End-of-line reached before string finished
C
        IF (CURC .GT. LINLEN) THEN
          IPOS (2) = LINLEN
          ERRCOD = 2
          RETURN
C
C String item finished
C
        ELSE IF (LXDSET (ICHAR (LINE (CURC:CURC))) .EQ. LXSTR) THEN
          IPOS (2) = CURC
          ERRCOD = 0
          RETURN
        END IF
C
      GOTO 350
C
C Comment start character; by definition nothing more on line
C
  400 ERRCOD = 1
      RETURN
C
      END
C
C Expand string abbreviation if unique in string list.
C
C Note: leading and trailing blanks in input string are significant.
C
C Uses no separately compiled procedures.
C
C Copyright (C) 1988 Per Kraulis
C
C Per Kraulis, Dept Molecular Biology, Uppsala University Sweden.
C  16-Sep-1987  first attempts
C  29-Oct-1987  lexically ordered input list; binary search
C  24-Mar-1988  unordered list option removed
C   9-Jun-1988  renamed
C  11-Jul-1990  added test for string length in list
C
C-----------------------------------------------------------
      SUBROUTINE SEXPND (EXPAND, STR, SLIST, LSTLEN, ERRCOD)
C
      CHARACTER*(*) EXPAND, STR
      CHARACTER*(*) SLIST (*)
      INTEGER       LSTLEN, ERRCOD
C
C EXPAND  (Out) expanded string
C STR     (In)  string to expand
C SLIST   (In)  list of string to expand into
C LSTLEN  (In)  number of strings in list; array size
C ERRCOD  (Out) error code
C     0 = no error
C     1 = no corresponding string found in list
C     2 = ambiguous; more than one corresponding string found in list
C     3 = input string ord longer than expanded; expansion not possible
C
      INTEGER STRLEN, LO, HI, MIDDLE
C
C Check that expansion is possible and if anything in list
C
      IF (LSTLEN .LT. 1) THEN
        ERRCOD = 1
        RETURN
      END IF
C
      STRLEN = LEN (STR)
      IF (STRLEN .GT. LEN (EXPAND)  .OR.
     $    STRLEN .GT. LEN (SLIST (1))) THEN
        ERRCOD = 3
        RETURN
      END IF
C
C Binary search; compare until middle of interval matches, or 0 interval
C
      LO = 1
      HI = LSTLEN
C
C Simulated repeat-until loop
C
100   MIDDLE = (LO + HI) / 2
C
C A matching string has been found
C
        IF (STR .EQ. SLIST (MIDDLE) (1:STRLEN)) THEN
C
C Check if ambiguous; if so, such strings must be adjacent in list
C
          IF (MIDDLE .GT. 1) THEN
            IF (STR .EQ. SLIST (MIDDLE - 1) (:STRLEN)) THEN
              ERRCOD = 2
              RETURN
            END IF
          END IF
          IF (MIDDLE .LT. LSTLEN) THEN
            IF (STR .EQ. SLIST (MIDDLE + 1) (:STRLEN)) THEN
              ERRCOD = 2
              RETURN
            END IF
          END IF
C
C Non-ambiguous match; expand string and return
C
          EXPAND = SLIST (MIDDLE)
          ERRCOD = 0
          RETURN
C
C No match; set new interval limits, upwards or downwards in list
C
        ELSE IF (LGT (STR, SLIST (MIDDLE) (:STRLEN))) THEN
          LO = MIDDLE + 1
        ELSE
          HI = MIDDLE - 1
        END IF
C
C Return with failure if interval has shrunk to nothing
C
      IF (LO .GT. HI) THEN
        ERRCOD = 1
        RETURN
      END IF
C
C Loop back
C
      GOTO 100
C
      END
C Package SNCNVT
C
C Conversion of character string from/to integer/real/fix data.
C
C SITOCH s  write integer value into string
C SRTOCH s  write float real value into string
C SFTOCH s  write fix real value into string
C SCHTOI s  convert string into integer value
C SCHTOR s  convert string into real value
C
C Uses no separately compiled procedures.
C
C Copyright (C) 1987 Per Kraulis
C
C Per Kraulis, Dept Molecular Biology, Uppsala University Sweden.
C   9-Sep-1987  written, tested
C  19-Jun-1990  added P format code to SRTOCH
C
C--------------------------------------
      SUBROUTINE SITOCH (CH, I, ERRCOD)
C
      CHARACTER*(*) CH
      INTEGER       I, ERRCOD
C
C CH      (Out) string to write integer value in
C I       (In)  integer value to write
C ERRCOD  (Out) error code
C      0 = no error
C      1 = string equivalent of integer could not fit into given string
C
      CHARACTER*5 FMTLIN
      SAVE FMTLIN
      DATA FMTLIN /'(I  )'/
C
      WRITE (FMTLIN (3:4), 10) LEN (CH)
10    FORMAT (I2)
      WRITE (CH, FMTLIN, ERR=90) I
      ERRCOD = 0
      RETURN
90    ERRCOD = 1
      RETURN
      END
C
C-------------------------------------------
      SUBROUTINE SRTOCH (CH, R, DEC, ERRCOD)
C
      CHARACTER*(*) CH
      REAL          R
      INTEGER       DEC, ERRCOD
C
C CH      (Out) string to write floating real value in
C R       (In)  floating real value to write
C DEC     (In)  number of decimals to use
C ERRCOD  (Out) error code
C      0 = no error
C      1 = string equivalent of real could not fit into given string
C      2 = illegal number of decimals
C
      CHARACTER*10 FMTLIN
      SAVE FMTLIN
      DATA FMTLIN /'(1PE  .  )'/
C
      IF (DEC .LT. 1 .OR. DEC .GT. 99) THEN
        ERRCOD = 2
        RETURN
      END IF
C
      WRITE (FMTLIN (5:6), 10) LEN (CH)
10    FORMAT (I2)
      WRITE (FMTLIN (8:9), 10) DEC
      WRITE (CH, FMTLIN, ERR=90) R
      ERRCOD = 0
      RETURN
C
90    ERRCOD = 1
      RETURN
      END
C
C-------------------------------------------
      SUBROUTINE SFTOCH (CH, F, DEC, ERRCOD)
C
      CHARACTER*(*) CH
      REAL          F
      INTEGER       DEC, ERRCOD
C
C CH      (Out) string to write fix real value in
C F       (In)  fix real value to write
C DEC     (In)  number of decimals to use
C ERRCOD  (Out) error code
C      0 = no error
C      1 = string equivalent of real could not fit into given string
C      2 = illegal number of decimals
C
      CHARACTER*8 FMTLIN
      SAVE FMTLIN
      DATA FMTLIN /'(F  .  )'/
C
      IF (DEC .LT. 0 .OR. DEC .GT. 99) THEN
        ERRCOD = 2
        RETURN
      END IF
C
      WRITE (FMTLIN (3:4), 10) LEN (CH)
10    FORMAT (I2)
      WRITE (FMTLIN (6:7), 10) DEC
      WRITE (CH, FMTLIN, ERR=90) F
      ERRCOD = 0
      RETURN
C
90    ERRCOD = 1
      RETURN
      END
C
C--------------------------------------
      SUBROUTINE SCHTOI (I, CH, ERRCOD)
C
      INTEGER       I, ERRCOD
      CHARACTER*(*) CH
C
C I       (Out) integer value to read
C CH      (In)  string to read integer value from
C ERRCOD  (Out) error code
C      0 = no error
C      1 = could not read integer value string
C
      CHARACTER*10 FMTLIN
      SAVE FMTLIN
      DATA FMTLIN /'(BN,I    )'/
C
      WRITE (FMTLIN (6:9), 10) LEN (CH)
10    FORMAT (I4)
      READ (CH, FMTLIN, ERR=90) I
      ERRCOD = 0
      RETURN
C
90    ERRCOD = 1
      RETURN
      END
C
C--------------------------------------
      SUBROUTINE SCHTOR (R, CH, ERRCOD)
C
      REAL          R
      CHARACTER*(*) CH
      INTEGER       ERRCOD
C
C R       (Out) real value to read
C CH      (In)  string to read real value from
C ERRCOD  (Out) error code
C      0 = no error
C      1 = could not read real value from string
C
      CHARACTER*12 FMTLIN
      SAVE FMTLIN
      DATA FMTLIN /'(BN,F    .0)'/
C
      WRITE (FMTLIN (6:9), 10) LEN (CH)
10    FORMAT (I4)
      READ (CH, FMTLIN, ERR=90) R
      ERRCOD = 0
      RETURN
C
90    ERRCOD = 1
      RETURN
      END
C Package STRBLK
C
C Finding and eliminating blanks in string.
C
C SFNBLK if  position of first nonblank character
C SLNBLK if  position of last nonblank character
C SLEJYR s   left justify, remove blanks
C SRIJYR s   right justify, remove blanks
C SLESHF s   left shift, leave blanks after first nonblank untouched
C SRISHF s   right shift, leave blanks before last nonblank untouched
C SAPPND s   append second string to first, starting after last nonblank
C
C Uses no separately compiled procedures.
C
C Copyright (C) 1987 Per Kraulis
C
C Per Kraulis, Dept Molecular Biology, Uppsala University Sweden.
C    2-May-1987  main code written
C   16-Sep-1987  package created by splitting up STRUTL
C    3-Oct-1987  procedures SLESHF, SRISHF and SAPPND added
C   21-Oct-1990  fixed bug in SRISHF; slight internal modifications
C
C
C----------------------------------
      INTEGER FUNCTION SFNBLK (STR)
C
      CHARACTER*(*) STR
C
      INTEGER POS
C
      DO 100 POS = 1, LEN (STR)
        IF (STR (POS : POS) .NE. ' ') THEN
          SFNBLK = POS
          RETURN
        END IF
100   CONTINUE
C
      SFNBLK = 0
      RETURN
      END
C
C
C----------------------------------
      INTEGER FUNCTION SLNBLK (STR)
C
      CHARACTER*(*) STR
C
      INTEGER POS
C
      DO 100 POS = LEN (STR), 1, -1
        IF (STR (POS : POS) .NE. ' ') THEN
          SLNBLK = POS
          RETURN
        END IF
100   CONTINUE

      SLNBLK = 0
      RETURN
      END
C
C
C----------------------------
      SUBROUTINE SLEJYR (STR)
C
      CHARACTER*(*) STR
C
      INTEGER POS, CURR
C
      CURR = 0
      DO 100 POS = 1, LEN (STR)
        IF (STR (POS : POS) .NE. ' ') THEN
          CURR = CURR + 1
          STR (CURR : CURR) = STR (POS : POS)
          IF (POS .NE. CURR) STR (POS : POS) = ' '
        END IF
100   CONTINUE
C
      RETURN
      END
C
C
C----------------------------
      SUBROUTINE SRIJYR (STR)
C
      CHARACTER*(*) STR
C
      INTEGER POS, CURR
C
      CURR = LEN (STR) + 1
      DO 100 POS = LEN (STR), 1, -1
        IF (STR (POS : POS) .NE. ' ') THEN
          CURR = CURR - 1
          STR (CURR : CURR) = STR (POS : POS)
          IF (POS .NE. CURR) STR (POS : POS) = ' '
        END IF
100   CONTINUE
C
      RETURN
      END
C
C
C----------------------------
      SUBROUTINE SLESHF (STR)
C
      CHARACTER*(*) STR
C
      INTEGER SFNBLK
C
      INTEGER POS
C
      POS = SFNBLK (STR)
      IF (POS .LE. 1) RETURN
      STR = STR (POS :)
      RETURN
      END
C
C
C----------------------------
      SUBROUTINE SRISHF (STR)
C
      CHARACTER *(*) STR
C
      INTEGER SLNBLK
C
      INTEGER POS, STRLEN
C
      STRLEN = LEN (STR)
      POS = SLNBLK (STR)
      IF (POS .EQ. 0) RETURN
      IF (POS .EQ. STRLEN) RETURN
      STR (STRLEN - POS + 1 :) = STR (: POS)
      STR (1 : STRLEN - POS) = ' '
      RETURN
      END
C
C
C-----------------------------------
      SUBROUTINE SAPPND (STR1, STR2)
C
      CHARACTER*(*) STR1, STR2
C
      INTEGER SLNBLK
C
      INTEGER POS
C
      POS = SLNBLK (STR1)
      IF (POS .LT. LEN (STR1)) STR1 (POS + 1 :) = STR2
      RETURN
      END
C Package STRUTL
C
C Simple string utility procedures. Assumes ASCII character code.
C
C SUPCAS s   all characters into upcase
C SLOCAS s   all characters into lowcase
C SCHCHR s   change all occurrences of a specified character
C SMINDX if  position of first of specified set of characters, 0 if none
C SRINDX if  position of last occurrence of string
C SDIGIT lf  TRUE if character is digit, otherwise FALSE
C SALPHA lf  TRUE if character is alphabetical, otherwise FALSE
C
C Uses no separately compiled procedures.
C
C Copyright (C) 1988 Per Kraulis
C
C Per Kraulis, Dept Molecular Biology, Uppsala University Sweden.
C    2-May-1987  main code written
C    3-Jun-1987  SMINDX added
C   16-Sep-1987  package split up; see STRBLK, SNCNVT
C   15-Mar-1988  SDIGIT and SALPHA added
C   15-Jul-1989  SRINDX added
C
C
C----------------------------
      SUBROUTINE SUPCAS (STR)
C
      CHARACTER*(*) STR
C
      INTEGER I, ASCII
C
      DO 100 I = 1, LEN (STR)
        ASCII = ICHAR (STR (I:I))
        IF (ASCII.GE.97 .AND. ASCII.LE.122) STR (I:I) = CHAR (ASCII-32)
100   CONTINUE
      RETURN
      END
C
C
C----------------------------
      SUBROUTINE SLOCAS (STR)
C
      CHARACTER*(*) STR
C
      INTEGER I, ASCII
C
      DO 100 I = 1, LEN (STR)
        ASCII = ICHAR (STR (I:I))
        IF (ASCII.GE.65 .AND. ASCII.LE.90) STR (I:I) = CHAR (ASCII+32)
100   CONTINUE
      RETURN
      END
C
C
C----------------------------------------
      SUBROUTINE SCHCHR (STR, OLDC, NEWC)
C
      CHARACTER*(*) STR
      CHARACTER*1   OLDC, NEWC
C
      INTEGER I
C
      DO 100 I = 1, LEN (STR)
        IF (STR (I:I) .EQ. OLDC) STR (I:I) = NEWC
100   CONTINUE
      RETURN
      END
C
C
C--------------------------------------
      INTEGER FUNCTION SMINDX (STR, CH)
C
      CHARACTER*(*) STR, CH
C
      INTEGER I, POS, MINPOS
C
      SMINDX = 0
      MINPOS = LEN (STR) + 1
      DO 100 I = 1, LEN (CH)
        POS = INDEX (STR, CH(I:I))
        IF (POS .NE. 0) THEN
          IF (POS .LT. MINPOS) THEN
            SMINDX = POS
            MINPOS = POS
          END IF
        END IF
100   CONTINUE
      RETURN
      END
C
C
C--------------------------------------
      INTEGER FUNCTION SRINDX (STR, CH)
C
      CHARACTER*(*) STR, CH
C
      INTEGER I, LENSTR, LENCH
C
      LENSTR = LEN (STR)
      LENCH  = LEN (CH)
      DO 100 I = LENSTR - LENCH + 1, 1, -1
        IF (STR (I : I + LENCH - 1) .EQ. CH) THEN
          SRINDX = I
          RETURN
        END IF
100   CONTINUE
      SRINDX = 0
      RETURN
      END
C
C
C---------------------------------
      LOGICAL FUNCTION SDIGIT (CH)
C
      CHARACTER*1 CH
C
      INTEGER ASCII
C
      ASCII = ICHAR (CH)
      SDIGIT = ASCII.GE.48 .AND. ASCII.LE.57
      RETURN
      END
C
C
C---------------------------------
      LOGICAL FUNCTION SALPHA (CH)
C
      CHARACTER*1 CH
C
      INTEGER ASCII
C
      ASCII = ICHAR (CH)
      SALPHA = (ASCII.GE.65 .AND. ASCII.LE.90) .OR.
     $         (ASCII.GE.97 .AND. ASCII.LE.122)
      RETURN
      END
C Package TTYIOP
C
C Terminal input/output procedures
C
C Some of these procedures use the '$' format descriptor
C
C TTYGTL s  output prompt and get line
C TTYGTI s  output prompt and get integer value 
C TTYGTR s  output prompt and get real value
C TTYCR  s  output N empty lines
C TTYPTL s  output line
C TTYPTI s  output line and integer value
C TTYPTF s  output line and fix real value
C TTYPTR s  output line and floating real value
C
C Uses no separately compiled procedures.
C
C Copyright (C) 1987 Per Kraulis
C
C Per Kraulis, Dept Molecular Biology, Uppsala University Sweden
C  16-Sep-1987  written
C   7-Oct-1987  modified to handle empty input line properly
C
C
C---------------------------------------------
      SUBROUTINE TTYGTL (PROMPT, LINE, ERRCOD)
C
      CHARACTER*(*) PROMPT, LINE
      INTEGER       ERRCOD
C
C ERRCOD  (Out) error code
C     0 = no error
C     1 = could not read line
C     2 = end-of-file encountered
C
      WRITE (6, 10) PROMPT
10    FORMAT (' ',A,$)
      READ (5, 20, ERR=90, END=95) LINE
20    FORMAT (A)
      ERRCOD = 0
      RETURN
C
90    ERRCOD = 1
      RETURN
95    ERRCOD = 2
      RETURN
      END
C
C
C---------------------------------------------
      SUBROUTINE TTYGTI (PROMPT, IVAL, ERRCOD)
C
      CHARACTER*(*) PROMPT
      INTEGER       IVAL, ERRCOD
C
C ERRCOD  (Out) error code
C     0 = no error
C     1 = could not read integer value
C     2 = end-of-file encountered
C
      CHARACTER*80 INLINE
C
      WRITE (6, 10) PROMPT
10    FORMAT (' ',A,$)
      READ (5, 20, ERR=90, END=95) INLINE
20    FORMAT (A)
      IF (INLINE.EQ.' ') GOTO 90
      READ (INLINE, 30, ERR=90) IVAL
30    FORMAT (BN, I80)
      ERRCOD = 0
      RETURN
C
90    ERRCOD = 1
      RETURN
95    ERRCOD =2
      RETURN
      END
C
C
C---------------------------------------------
      SUBROUTINE TTYGTR (PROMPT, RVAL, ERRCOD)
C
      CHARACTER*(*) PROMPT
      REAL          RVAL
      INTEGER       ERRCOD
C
C ERRCOD  (Out) error code
C     0 = no error
C     1 = could not read real value
C     2 = end-of-file encountered
C
      CHARACTER*80 INLINE
C
      WRITE (6, 10) PROMPT
10    FORMAT (' ',A,$)
      READ (5, 20, ERR=90, END=95) INLINE
20    FORMAT (A)
      IF (INLINE.EQ.' ') GOTO 90
      READ (INLINE, 30, ERR=90) RVAL
30    FORMAT (BN, F80.0)
      ERRCOD = 0
      RETURN
C
90    ERRCOD = 1
      RETURN
95    ERRCOD = 2
      RETURN
      END
C
C
C-------------------------
      SUBROUTINE TTYCR (N)
C
      INTEGER N
C
      INTEGER I
C
      DO 10 I = 1, N
        WRITE (6, '(/)')
10    CONTINUE
      RETURN
      END
C
C
C-----------------------------
      SUBROUTINE TTYPTL (LINE)
C
      CHARACTER*(*) LINE
C
      WRITE (6, 10) LINE
10    FORMAT (' ',A)
      RETURN
      END
C
C
C------------------------------------------------
      SUBROUTINE TTYPTI (LINE, IVAL, POS, ERRCOD)
C
      CHARACTER*(*) LINE
      INTEGER       IVAL, POS, ERRCOD
C
C LINE    (In) line to output
C IVAL    (In) integer value to output
C POS     (In) number of positions to use for integer value
C ERRCOD  (Out) error code
C     0 = no error
C     1 = integer value position overflow
C
      CHARACTER*11 FMTLIN
      SAVE FMTLIN
      DATA FMTLIN /'('' '',A,I  )'/
C
      WRITE (FMTLIN (9:10), 10, ERR=90) POS
10    FORMAT (I2)
      WRITE (6, FMTLIN, ERR=90) LINE, IVAL
      ERRCOD = 0
      RETURN
C
90    ERRCOD = 1
      RETURN
      END
C
C
C-----------------------------------------------------
      SUBROUTINE TTYPTF (LINE, FVAL, POS, DEC, ERRCOD)
C
      CHARACTER*(*) LINE
      REAL          FVAL
      INTEGER       POS, DEC, ERRCOD
C
C LINE    (In) line to output
C FVAL    (In) fix real value to output
C POS     (In) number of positions to use for fix real value
C DEC     (In) number of decimals to output
C ERRCOD  (Out) error code
C     0 = no error
C     1 = fix real value position overflow
C
      CHARACTER*14 FMTLIN
      SAVE FMTLIN
      DATA FMTLIN /'('' '',A,F  .  )'/
C
      WRITE (FMTLIN (9:10), 10, ERR=90) POS
10    FORMAT (I2)
      WRITE (FMTLIN (12:13), 10, ERR=90) DEC
      WRITE (6, FMTLIN, ERR=90) LINE, FVAL
      ERRCOD = 0
      RETURN
C
90    ERRCOD = 1
      RETURN
      END
C
C
C-----------------------------------------------------
      SUBROUTINE TTYPTR (LINE, RVAL, POS, DEC, ERRCOD)
C
      CHARACTER*(*) LINE
      REAL          RVAL
      INTEGER       POS, DEC, ERRCOD
C
C LINE    (In) line to output
C RVAL    (In) floating real value to output
C POS     (In) number of positions to use for floating real value
C DEC     (In) number of decimals to output
C ERRCOD  (Out) error code
C     0 = no error
C     1 = floating real value position overflow
C
      CHARACTER*14 FMTLIN
      SAVE FMTLIN
      DATA FMTLIN /'('' '',A,E  .  )'/
C
      WRITE (FMTLIN (9:10), 10, ERR=90) POS
10    FORMAT (I2)
      WRITE (FMTLIN (12:13), 10, ERR=90) DEC
      WRITE (6, FMTLIN, ERR=90) LINE, RVAL
      ERRCOD = 0
      RETURN
C
90    ERRCOD = 1
      RETURN
      END
C*******************************************************************
C Package LQPLAN
C
C Compute a least-square plane from a set of 3D coordinates.
C
C LQPLAN s  calculate least-square plane from a set of points
C LPDIST rf  return distance from point to given plane
C LPRMSD rf  compute root-mean-square deviation of points to given plane
C
C Uses no separately compiled procedures.
C
C Copyright (C) 1988 Per Kraulis
C
C Per Kraulis, Dept Molecular Biology, Uppsala University Sweden.
C   12-May-1988  first attempts
C   13-May-1988  written
C   16-May-1988  debugged
C
C
C-----------------------------------------------------------
      SUBROUTINE LQPLAN (PLANE, CENTER, POINTS, NUM, ERRCOD)
C
      INTEGER NUM, ERRCOD
      REAL    PLANE (4), CENTER (3), POINTS (3, NUM)
C
C PLANE   (Out) plane parameters: xyz of normal, and dist to origin
C CENTER  (Out) center of gravity of points
C POINTS  (In)  coordinates of set of points
C NUM     (In)  total number of points
C ERRCOD  (Out) error code
C      0 = no error
C      1 = too few points
C      2 = ill conditioned; determinant of matrix close to zero
C
C MINDET   minimum absolute value of determinant
C
      REAL       MINDET
      PARAMETER (MINDET = 1.0E-6)
C
C Help variables
C
      INTEGER I, OFFSET
      INTEGER ORD (3)
      REAL    FACTOR
      REAL    XYZ (3), MTX (3, 3)
C
C Skip if too few points input
C
      IF (NUM .LT. 3) THEN
        ERRCOD = 1
        RETURN
      END IF
C
C No offset has been added to any direction
C
      OFFSET = 0
C
C Compute sum of coordinates
C
100   XYZ (1) = 0.0
      XYZ (2) = 0.0
      XYZ (3) = 0.0
      DO 200 I = 1, NUM
        XYZ (1) = XYZ (1) + POINTS (1, I)
        XYZ (2) = XYZ (2) + POINTS (2, I)
        XYZ (3) = XYZ (3) + POINTS (3, I)
200   CONTINUE
C
C Compute center of gravity, if not already done
C
      IF (OFFSET .EQ. 0) THEN
        CENTER (1) = XYZ (1) / REAL (NUM)
        CENTER (2) = XYZ (2) / REAL (NUM)
        CENTER (3) = XYZ (3) / REAL (NUM)
      END IF
C
C Compute matrix elements
C
      MTX (1, 1) = 0.0
      MTX (1, 2) = 0.0
      MTX (1, 3) = 0.0
      MTX (2, 2) = 0.0
      MTX (2, 3) = 0.0
      MTX (3, 3) = 0.0
      DO 300 I = 1, NUM
        MTX (1, 1) = MTX (1, 1) + POINTS (1, I) * POINTS (1, I)
        MTX (1, 2) = MTX (1, 2) + POINTS (1, I) * POINTS (2, I)
        MTX (1, 3) = MTX (1, 3) + POINTS (1, I) * POINTS (3, I)
        MTX (2, 2) = MTX (2, 2) + POINTS (2, I) * POINTS (2, I)
        MTX (2, 3) = MTX (2, 3) + POINTS (2, I) * POINTS (3, I)
        MTX (3, 3) = MTX (3, 3) + POINTS (3, I) * POINTS (3, I)
300   CONTINUE
      MTX (2, 1) = MTX (1, 2)
      MTX (3, 1) = MTX (1, 3)
      MTX (3, 2) = MTX (2, 3)
C
C Check if problem is ill conditioned
C
      IF (ABS (MTX (1,1) * MTX (2,2) * MTX (3,3) +
     $         MTX (1,2) * MTX (2,3) * MTX (3,1) +
     $         MTX (1,3) * MTX (2,1) * MTX (3,2) -
     $         MTX (1,1) * MTX (2,3) * MTX (3,2) -
     $         MTX (1,2) * MTX (2,1) * MTX (3,3) -
     $         MTX (1,3) * MTX (2,2) * MTX (3,1)) .LT. MINDET) THEN
C
C Try adding offset and recompute centre of gravity and matrix
C
cpjx         type * , ' Ill conditioned'
        IF (OFFSET .LT. 3) THEN
C
C Remove old offset, if any
C
          IF (OFFSET .GT. 0) THEN
            DO 400 I = 1, NUM
              POINTS (OFFSET, I) = POINTS (OFFSET, I) - 1.0
400         CONTINUE
          END IF
C
C Add offset, try first in x direction, then y, then z
C
          OFFSET = OFFSET + 1
          DO 450 I = 1, NUM
            POINTS (OFFSET, I) = POINTS (OFFSET, I) + 1.0
450       CONTINUE
          GOTO 100
C
C If neither x, y nor z offset works, then really ill conditioned
C
        ELSE
          ERRCOD = 2
          RETURN
        END IF
      END IF
C
C Use Gauss elimination and row pivoting (or whatever it's called)
C
      IF (ABS (MTX (1,1)) .GT. ABS (MTX (2,1))) THEN
        ORD (1) = 1
      ELSE
        ORD (1) = 2
      END IF
      IF (ABS (MTX (3,1)) .GT. ABS (MTX (ORD (1), 1))) ORD (1) = 3
      DO 500 I = 1, 3
        IF (I .NE. ORD (1)) THEN
          FACTOR = MTX (I, 1) / MTX (ORD (1), 1)
          MTX (I, 2) = MTX (I, 2) - FACTOR * MTX (ORD (1), 2)
          MTX (I, 3) = MTX (I, 3) - FACTOR * MTX (ORD (1), 3)
          XYZ (I)    = XYZ (I)    - FACTOR * XYZ (ORD (1))
        END IF
500   CONTINUE
C
      IF (ORD (1) .EQ. 1) THEN
        IF (ABS (MTX (2,2)) .GT. ABS (MTX (3,2))) THEN
          ORD (2) = 2
          ORD (3) = 3
        ELSE
          ORD (2) = 3
          ORD (3) = 2
        END IF
      ELSE IF (ORD (1) .EQ. 2) THEN
        IF (ABS (MTX (1,2)) .GT. ABS (MTX (3,2))) THEN
          ORD (2) = 1
          ORD (3) = 3
        ELSE
          ORD (2) = 3
          ORD (3) = 1
        END IF
      ELSE
        IF (ABS (MTX (1,2)) .GT. ABS (MTX (2,2))) THEN
          ORD (2) = 1
          ORD (3) = 2
        ELSE
          ORD (2) = 2
          ORD (3) = 1
        END IF
      END IF
cpjx           type * , ' denominator 1 = ',MTX (ORD (2), 2)
      FACTOR = MTX (ORD (3), 2) / MTX (ORD (2), 2)
      MTX (ORD (3), 3) = MTX (ORD (3), 3) - FACTOR * MTX (ORD (2), 3)
      XYZ (ORD (3))    = XYZ (ORD (3))    - FACTOR * XYZ (ORD (2))
C
C Solve for the plane coefficients; backwards substitution
C
cpjx           type * , ' denominator 2 = ',MTX (ORD (3), 3)
      PLANE (3) =  XYZ (ORD (3)) / MTX (ORD (3), 3)
cpjx           type * , ' denominator 3 = ',MTX (ORD (2), 2)
      PLANE (2) = (XYZ (ORD (2)) - MTX (ORD (2), 3) * PLANE (3)) /
     $             MTX (ORD (2), 2)
cpjx           type * , ' denominator 4 = ',MTX (ORD (1), 1)
      PLANE (1) = (XYZ (ORD (1)) - MTX (ORD (1), 2) * PLANE (2) -
     $             MTX (ORD (1), 3) * PLANE (3)) / MTX (ORD (1), 1)
      PLANE (4) = SQRT (PLANE(1)**2 + PLANE(2)**2 + PLANE(3)**2)
      IF ( PLANE (4) .NE. 0 ) THEN
         PLANE (4) = 1.0 / PLANE (4)
      ELSE
cpjx         type * , ' Can''t compute plane (4)'
         return
      END IF
      PLANE (1) = PLANE (4) * PLANE (1)
      PLANE (2) = PLANE (4) * PLANE (2)
      PLANE (3) = PLANE (4) * PLANE (3)
C
C If offset added, remove it from plane parameter and coordinates
C
      IF (OFFSET .NE. 0) THEN
        PLANE (4) = PLANE (4) - PLANE (OFFSET)
        DO 600 I = 1, NUM
          POINTS (OFFSET, I) = POINTS (OFFSET, I) - 1.0
600     CONTINUE
      END IF
C
      ERRCOD = 0
      RETURN
      END
C
C
C----------------------------------------
      REAL FUNCTION LPDIST (PLANE, POINT)
C
      REAL PLANE (4), POINT (3)
C
C PLANE  (In)  plane parameters
C POINT  (In)  coordinate of point
C
      LPDIST = PLANE (1) * POINT (1) +
     $         PLANE (2) * POINT (2) +
     $         PLANE (3) * POINT (3) - PLANE (4)
      RETURN
      END
C
C
C----------------------------------------------
      REAL FUNCTION LPRMSD (PLANE, POINTS, NUM)
C
      INTEGER NUM
      REAL    PLANE (4), POINTS (3, NUM)
C
C PLANE   (In)  plane parameters
C POINTS  (In)  coordinates of points
C NUM     (In)  number of points
C
      REAL    SUM
      INTEGER I
C
      SUM = 0.0
      DO 100 I = 1, NUM
        SUM = SUM + (PLANE (1) * POINTS (1, I) +
     $               PLANE (2) * POINTS (2, I) +
     $               PLANE (3) * POINTS (3, I) - PLANE (4)) **2
100   CONTINUE
C
      IF (NUM .GT. 1) SUM = SUM / REAL (NUM)
      LPRMSD = SQRT (SUM)
      RETURN
      END
