C
C
C     This code is distributed under the terms and conditions of the
C     CCP4 Program Suite Licence Agreement as a CCP4 Application.
C     A copy of the CCP4 licence can be obtained by writing to the
C     CCP4 Secretary, Daresbury Laboratory, Warrington WA4 4AD, UK.
C
************************************************************************
*
*
*       Package VECLIB
*       Miscellaneous routines for fundamental operations on vectors
*       and matrices. Also some other odds and ends.
*
*
*       Contains:
*       s/r matcmp     compare two n x m real matrices
*       s/r rmatcmp    = matcmp
*       s/r imatcmp    compare two n x m integer matrices
*
*       s/r matcop     copy n x m real matrix
*       s/r rmatcop    = matcop
*       s/r imatcop    copy n x m integer matrix
*       s/r drmatcop   copy n x m double precision matrix to real
*
*       s/r matclr     clear n x m real matrix
*       s/r rmatclr    = matclr
*       s/r imatclr    clear n x m integer matrix
*       s/r dmatclr    clear n x m double precision matrix
*       s/r cmatclr    clear n x m character matrix
*
*       s/r rmatset    set all elements in n x m real matrix
*       s/r imatset    set all elements in n x m integer matrix
*       s/r dmatset    set all elements in n x m double precision matrix
*       s/r cmatset    set all elements in n x m character matrix
*
*       s/r vecdif     subtract two real vectors
*       s/r vecadd     add two real vectors
*       s/r vecabs     abs of real vector elements
*       s/r rvecdif    = vecdif
*       s/r rvecadd    = vecadd
*       s/r rvecdif    = vecdif
*       s/r rvecmod    modulus of real vector
*       s/r ivecdif    subtract two integer vectors
*       s/r ivecadd    add two integer vectors
*       s/r ivecabs    abs of integer vector elements
*
*       s/r ivecf      convert integer vector to real
*       s/r fveci      convert real vector to integer
*
*       s/r mtxtps     transpose n x m real matrix
*       s/r mtxinv     invert n x n matrix
*       s/r mtxmlt     multiply n x m matrix by m x n matrix
*       s/r vmxmlt     mulitply n x 1 vector by m x n matrix
*
*       s/r transpose  transpose n x n real matrix
*       s/r matinv     invert 3 x 3 real matrix
*       s/r vmtply     multiply 3 x 1 vector by 3 x 3 matrix
*       s/r mmtply     multiply 3 x 3 matrix by 3 x 3 matrix
*       s/r cosabg     cosine of three angles
*       s/r ortho      get (de)orthogoanlization matrix from cell
*       s/r pnt2pln    get plane equation from three points
*       fn  plndst     distance between point and plane
*       s/r polar2mtx  convert polar angles to matrix
*       s/r nfold2mtx  convert direction cosines and rot angle to matrix
*       s/r msk        .and. two logical vectors, count and point
*       s/r mskcmp     compare two logical vectors
*
*       s/r isort      sort real array, co-sort integer vector
*       s/r fsort      sort real array, co-sort real vector
*       s/r qsortfi    quicksort real array, co-sort integer vector
*       s/r qsortff    quicksort real array, co-sort real vector
*       s/r qsortffi   quicksort real array, co-sort real and integer vector
*       s/r qsortpf    quicksort pointers
*       s/r nsorta     Sort N columns on values in key column, acending
*
*       fun det        determinant of 3 x 3 matrix
*       fun vlen       length (A) of fractional vector in ]-0.5,0.5[
*       fun vlen2      length (A) of fractional vector
*
*       fun imaxvc     maximum element in vector (integer)
*       fun iminvc     minimum element in vector (integer)
*       fun fmaxvc     maximum element in vector (real)
*       fun fminvc     minimum element in vector (real)
*       fun random     pseudo-random number generator (real)
*
************************************************************************
************************************************************************
*
*       Subroutine MATCMP
*
*       Function:
*       Compare two n x m matrices a and b and return number of matching
*       elements in match.
*
*       Arguments:
*       A         (i)      matrix 1
*       B         (i)      matrix 2
*       N         (i)      first dimension of A and B
*       M         (i)      second dimension of A and B
*       Match     (o)      number of matching elements
*       Tol       (i)      allowed discrepancy
*
*       History:
*       Written October 1987 S. Knight
************************************************************************
        subroutine matcmp ( a , b , n , m , match , tol )

        implicit none

*       Arguments
        integer         n , m , match
        real            a( n,m ) , b( n,m )
        real            tol

*       Local variables
        integer         i , j
*-----------------------------------------------------------------------
        match = 0
        do i = 1 , m
        do j = 1 , n
        if ( abs ( a( j,i ) - b( j,i ) ) .lt. tol ) match = match + 1
        end do
        end do
        return
        end
************************************************************************
************************************************************************
*
*       Subroutine RMATCMP
*
*       Function:
*       Compare two n x m matrices a and b and return number of matching
*       elements in match.
*
*       Arguments:
*       A         (i)      matrix 1
*       B         (i)      matrix 2
*       N         (i)      first dimension of A and B
*       M         (i)      second dimension of A and B
*       Match     (o)      number of matching elements
*       Tol       (i)      allowed discrepancy
*
*       History:
*       Written October 1987 S. Knight
************************************************************************
        subroutine rmatcmp ( a , b , n , m , match , tol )

        implicit none

*       Arguments
        integer         n , m , match
        real            a( n,m ) , b( n,m )
        real            tol

*       Local variables
        integer         i , j
*-----------------------------------------------------------------------
        match = 0
        do i = 1 , m
        do j = 1 , n
        if ( abs ( a( j,i ) - b( j,i ) ) .lt. tol ) match = match + 1
        end do
        end do
        return
        end
************************************************************************
************************************************************************
*
*       Subroutine IMATCMP
*
*       Function:
*       Compare two n x m integer matrices a and b and return number of 
*       matching elements in match.
*
*       Arguments:
*       A         (i)      matrix 1
*       B         (i)      matrix 2
*       N         (i)      first dimension of A and B
*       M         (i)      second dimension of A and B
*       Match     (o)      number of matching elements
*
*       History:
*       Written October 1987 S. Knight
************************************************************************
        subroutine imatcmp ( a , b , n , m , match )

        implicit none

*       Arguments
        integer         n , m , match
        integer         a( n,m ) , b( n,m )

*       Local variables
        integer         i , j
*-----------------------------------------------------------------------
        match = 0
        do i = 1 , m
        do j = 1 , n
        if ( a( j,i ) .eq. b( j,i ) ) match = match + 1
        end do
        end do
        return
        end
************************************************************************
************************************************************************
*
*       Subroutine MATCOP
*
*       Function:
*       Copy n x m real matrix a to b
*
*       Arguments:
*       A         (i)      matrix to copy
*       B         (o)      copy of A
*       N         (i)      first dimension of A and B
*       M         (i)      second dimension of A and B
*
*       History:
*       Written March 1988 S. Knight
************************************************************************

        subroutine matcop ( a , b , n , m )

        implicit none

*       Arguments
        integer         n , m
        real            a( n,m ) , b( n,m )

*       Local varibles
        integer         i , j
*-----------------------------------------------------------------------
        do i=1,m
        do j=1,n
        b(j,i)=a(j,i)
        end do
        end do
        return
        end
************************************************************************
************************************************************************
*
*       Subroutine RMATCOP
*
*       Function:
*       Copy n x m real matrix a to b
*
*       Arguments:
*       A         (i)      matrix to copy
*       B         (o)      copy of A
*       N         (i)      first dimension of A and B
*       M         (i)      second dimension of A and B
*
*       History:
*       Written March 1988 S. Knight
************************************************************************

        subroutine rmatcop ( a , b , n , m )

        implicit none

*       Arguments
        integer         n , m
        real            a( n,m ) , b( n,m )

*       Local varibles
        integer         i , j
*-----------------------------------------------------------------------
        do i=1,m
        do j=1,n
        b(j,i)=a(j,i)
        end do
        end do
        return
        end
************************************************************************
************************************************************************
*
*       Subroutine IMATCOP
*
*       Function:
*       Copy n x m integer matrix a to b
*
*       Arguments:
*       A         (i)      matrix to copy
*       B         (o)      copy of A
*       N         (i)      first dimension of A and B
*       M         (i)      second dimension of A and B
*
*       History:
*       Written March 1988 S. Knight
************************************************************************

        subroutine imatcop ( a , b , n , m )

        implicit none

*       Arguments
        integer         n , m
        integer         a(n,m) , b(n,m)

*       Local varibles
        integer         i , j
*-----------------------------------------------------------------------
        do i=1,m
        do j=1,n
        b(j,i)=a(j,i)
        end do
        end do
        return
        end
************************************************************************
************************************************************************
*
*       Subroutine DRMATCOP
*
*       Function:
*       Copy n x m double precision to real matrix 
*
*       Arguments:
*       A         (i)      matrix to copy
*       B         (o)      copy of A
*       N         (i)      first dimension of A and B
*       M         (i)      second dimension of A and B
*
*       History:
*       Written March 1988 S. Knight
************************************************************************

        subroutine drmatcop ( a , b , n , m )

        implicit none

*       Arguments
        integer          n , m
        double precision a( n,m )
        real             b( n,m )

*       Local varibles
        integer         i , j
*-----------------------------------------------------------------------
        do i=1,m
        do j=1,n
        b(j,i)=a(j,i)
        end do
        end do
        return
        end
************************************************************************
************************************************************************
*
*       Subroutine MATCLR
*
*       Function:
*       Clear n x m real matrix a
*
*       Arguments:
*       A         (i/o)    matrix to clear
*       N         (i)      first dimension of A
*       M         (i)      second dimension of A
*
*       History:
*       Written March 1988 S. Knight
************************************************************************

        subroutine matclr ( a , n , m )

        implicit none

*       Arguments
        integer         n , m
        real            a( n,m )

*       Local varibles
        integer         i , j
*-----------------------------------------------------------------------
        do i=1,m
        do j=1,n
        a(j,i)=0.0
        end do
        end do
        return
        end
************************************************************************
************************************************************************
*
*       Subroutine RMATCLR
*
*       Function:
*       Clear n x m real matrix a
*
*       Arguments:
*       A         (i/o)    matrix to clear
*       N         (i)      first dimension of A
*       M         (i)      second dimension of A
*
*       History:
*       Written March 1988 S. Knight
************************************************************************

        subroutine rmatclr ( a , n , m )

        implicit none

*       Arguments
        integer         n , m
        real            a( n,m )

*       Local varibles
        integer         i , j
*-----------------------------------------------------------------------
        do i=1,m
        do j=1,n
        a(j,i)=0.0
        end do
        end do
        return
        end
************************************************************************
************************************************************************
*
*       Subroutine IMATCLR
*
*       Function:
*       Clear n x m integer matrix a
*
*       Arguments:
*       A         (i/o)    matrix to clear
*       N         (i)      first dimension of A
*       M         (i)      second dimension of A
*
*       History:
*       Written March 1988 S. Knight
************************************************************************

        subroutine imatclr(a,n,m)

        implicit none

*       Arguments
        integer         n , m
        integer         a(n,m)

*       Local varibles
        integer         i , j
*-----------------------------------------------------------------------
        do i=1,m
        do j=1,n
        a(j,i)=0
        end do
        end do
        return
        end
************************************************************************
************************************************************************
*
*       Subroutine DMATCLR
*
*       Function:
*       Clear n x m double precision matrix a
*
*       Arguments:
*       A         (i/o)    matrix to clear
*       N         (i)      first dimension of A
*       M         (i)      second dimension of A
*
*       History:
*       Written March 1988 S. Knight
************************************************************************

        subroutine dmatclr(a,n,m)

        implicit none

*       Arguments
        integer          n , m
        double precision a(n,m)

*       Local varibles
        integer         i , j
*-----------------------------------------------------------------------
        do i=1,m
        do j=1,n
        a(j,i)=0
        end do
        end do
        return
        end
************************************************************************
************************************************************************
*
*       Subroutine CMATCLR
*
*       Function:
*       Clear n x m character matrix a
*
*       Arguments:
*       A         (i/o)    matrix to clear
*       N         (i)      first dimension of A
*       M         (i)      second dimension of A
*
*       History:
*       Written March 1988 S. Knight
************************************************************************

        subroutine cmatclr(a,n,m)

        implicit none

*       Arguments
        integer         n , m
        character*(*)   a(n,m)

*       Local varibles
        integer         i , j
*-----------------------------------------------------------------------
        do i=1,m
        do j=1,n
        a(j,i)=' '
        end do
        end do
        return
        end
************************************************************************
************************************************************************
*
*       Subroutine RMATSET
*
*       Function:
*       Set all elements in n x m real matrix a
*
*       Arguments:
*       A         (i/o)    matrix to set
*       N         (i)      first dimension of A
*       M         (i)      second dimension of A
*       Rval      (i)      real value to insert in matrix
*
*       History:
*       Written October 1992 S. Knight
************************************************************************

        subroutine rmatset ( a , n , m , rval )

        implicit none

*       Arguments
        integer         n , m
        real            a( n,m )
        real            rval

*       Local varibles
        integer         i , j
*-----------------------------------------------------------------------
        do 20 i = 1 , m
          do 10 j = 1 , n
            a( j,i ) = rval
 10       continue
 20     continue
        return
        end
************************************************************************
************************************************************************
*
*       Subroutine IMATSET
*
*       Function:
*       Set all elements in n x m integer matrix a
*
*       Arguments:
*       A         (i/o)    matrix to set
*       N         (i)      first dimension of A
*       M         (i)      second dimension of A
*       Ival      (i)      integer value to insert in matrix
*
*       History:
*       Written October 1992 S. Knight
************************************************************************

        subroutine imatset ( a , n , m , ival )

        implicit none

*       Arguments
        integer         n , m
        integer         a( n,m )
        integer         ival

*       Local varibles
        integer         i , j
*-----------------------------------------------------------------------
        do 20 i = 1 , m
          do 10 j = 1 , n
            a( j,i ) = ival
 10       continue
 20     continue
        return
        end
************************************************************************
************************************************************************
*
*       Subroutine DMATSET
*
*       Function:
*       Set all elements in n x m double precision matrix a
*
*       Arguments:
*       A         (i/o)    matrix to set
*       N         (i)      first dimension of A
*       M         (i)      second dimension of A
*       Dval      (i)      double precicision value to insert in matrix
*
*       History:
*       Written October 1992 S. Knight
************************************************************************

        subroutine dmatset ( a , n , m , dval )

        implicit none

*       Arguments
        integer          n , m
        double precision a( n,m )
        double precision dval

*       Local varibles
        integer         i , j
*-----------------------------------------------------------------------
        do 20 i = 1 , m
          do 10 j = 1 , n
            a( j,i ) = dval
 10       continue
 20     continue
        return
        end
************************************************************************
************************************************************************
*
*       Subroutine CMATSET
*
*       Function:
*       Set all elements in n x m character matrix a
*
*       Arguments:
*       A         (i/o)    matrix to set
*       N         (i)      first dimension of A
*       M         (i)      second dimension of A
*       Sval      (i)      String to insert in matrix
*
*       History:
*       Written October 1992 S. Knight
************************************************************************

        subroutine cmatset ( a , n , m , sval )

        implicit none

*       Arguments
        integer         n , m
        character*(*)   a(n,m)
        character*(*)   sval

*       Local varibles
        integer         i , j
*-----------------------------------------------------------------------
        do 20 i = 1 , m
          do 10 j = 1 , n
            a( j,i ) = sval
 10       continue
 20     continue
        return
        end
************************************************************************
************************************************************************
*
*       Subroutine VECDIF
*
*       Function:
*       Subtract real vector b from vector a, put result in c
*
*       Arguments:
*       A         (i)      first vector
*       B         (i)      second vector
*       C         (o)      vector difference A - B
*       N         (i)      dimension of vectors
*
*       History:
*       Written November 1991 S. Knight
************************************************************************

        subroutine vecdif ( a , b , c , n )

        implicit none

*       Arguments
        integer         n
        real            a( n ) , b( n ) , c( n )

*       Local varibles
        integer         i
*-----------------------------------------------------------------------
        do i = 1 , n
        c( i ) = a( i ) - b( i )
        end do
        return
        end
************************************************************************
************************************************************************
*
*       Subroutine VECADD
*
*       Function:
*       Add real vectors a and b, put result in c
*
*       Arguments:
*       A         (i)      first vector
*       B         (i)      second vector
*       C         (o)      vector sum A + B
*       N         (i)      dimension of vectors
*
*       History:
*       Written November 1991 S. Knight
************************************************************************

        subroutine vecadd ( a , b , c , n )

        implicit none

*       Arguments
        integer         n
        real            a( n ) , b( n ) , c( n )

*       Local varibles
        integer         i
*-----------------------------------------------------------------------
        do i = 1 , n
        c( i ) = a( i ) + b( i )
        end do
        return
        end
************************************************************************
************************************************************************
*
*       Subroutine VECABS
*
*       Function:
*       Absolute value of real vector elements, put result in b
*
*       Arguments:
*       A         (i)      argument vector
*       B         (o)      result vector
*       N         (i)      dimension of vectors
*
*       History:       
*       Written November 1991 S. Knight
************************************************************************

        subroutine vecabs ( a , b , n )

        implicit none

*       Arguments
        integer         n
        real            a( n ) , b( n )

*       Local varibles
        integer         i
*-----------------------------------------------------------------------
        do i = 1 , n
        b( i ) = abs ( a( i ) )
        end do
        return
        end
************************************************************************
************************************************************************
*
*       Subroutine RVECDIF
*
*       Function:
*       Subtract real vector b from vector a, put result in c
*
*       Arguments:
*       A         (i)      first vector
*       B         (i)      second vector
*       C         (o)      vector difference A - B
*       N         (i)      dimension of vectors
*
*       History:
*       Written November 1991 S. Knight
************************************************************************

        subroutine rvecdif ( a , b , c , n )

        implicit none

*       Arguments
        integer         n
        real            a( n ) , b( n ) , c( n )

*       Local varibles
        integer         i
*-----------------------------------------------------------------------
        do i = 1 , n
        c( i ) = a( i ) - b( i )
        end do
        return
        end
************************************************************************
************************************************************************
*
*       Subroutine RVECADD
*
*       Function:
*       Add real vectors a and b, put result in c
*
*       Arguments:
*       A         (i)      first vector
*       B         (i)      second vector
*       C         (o)      vector sum A + B
*       N         (i)      dimension of vectors
*
*       History:
*       Written November 1991 S. Knight
************************************************************************

        subroutine rvecadd ( a , b , c , n )

        implicit none

*       Arguments
        integer         n
        real            a( n ) , b( n ) , c( n )

*       Local varibles
        integer         i
*-----------------------------------------------------------------------
        do i = 1 , n
        c( i ) = a( i ) + b( i )
        end do
        return
        end
************************************************************************
************************************************************************
*
*       Subroutine RVECABS
*
*       Function:
*       Absolute value of real vector elements, put result in b
*
*       Arguments:
*       A         (i)      argument vector
*       B         (o)      result vector
*       N         (i)      dimension of vectors
*
*       History:       
*       Written November 1991 S. Knight
************************************************************************

        subroutine rvecabs ( a , b , n )

        implicit none

*       Arguments
        integer         n
        real            a( n ) , b( n )

*       Local varibles
        integer         i
*-----------------------------------------------------------------------
        do i = 1 , n
        b( i ) = abs ( a( i ) )
        end do
        return
        end
************************************************************************
************************************************************************
*
*       Subroutine RVECMOD
*
*       Function:
*       Modulus of vector elements, put result in b
*
*       Arguments:
*       A         (i)      argument vector
*       B         (o)      result vector
*       Modulus   (i)      modulus
*       N         (i)      dimension of vectors
*
*       History:       
*       Written November 1995 S. Knight
************************************************************************

        subroutine rvecmod ( a , b , modulus , n )

        implicit none

*       Arguments
        integer         n
        real            a( n ) , b( n ) , modulus

*       Local varibles
        integer         i
*-----------------------------------------------------------------------
        do i = 1 , n
        b( i ) = mod ( a( i ) , modulus )
        end do
        return
        end
************************************************************************
************************************************************************
*
*       Subroutine IVECDIF
*
*       Function
*       Subtract integer vector b from vector a, put result in c
*
*       Arguments:
*       A         (i)      first vector
*       B         (i)      second vector
*       C         (o)      vector difference A - B
*       N         (i)      dimension of vectors
*
*       History:
*       Written November 1991 S. Knight
************************************************************************

        subroutine ivecdif ( a , b , c , n )

        implicit none

*       Arguments
        integer         n
        integer         a( n ) , b( n ) , c( n )

*       Local varibles
        integer         i
*-----------------------------------------------------------------------
        do i = 1 , n
        c( i ) = a( i ) - b( i )
        end do
        return
        end
************************************************************************
************************************************************************
*
*       Subroutine IVECADD
*
*       Function:
*       Add integer vectors a and b, put result in c
*
*       Arguments:
*       A         (i)      first vector
*       B         (i)      second vector
*       C         (o)      vector sum A + B
*       N         (i)      dimension of vectors
*
*       History:
*       Written November 1991 S. Knight
************************************************************************

        subroutine ivecadd ( a , b , c , n )

        implicit none

*       Arguments
        integer         n
        integer         a( n ) , b( n ) , c( n )

*       Local varibles
        integer         i
*-----------------------------------------------------------------------
        do i = 1 , n
        c( i ) = a( i ) + b( i )
        end do
        return
        end
************************************************************************
************************************************************************
*
*       Subroutine IVECABS
*
*       Function:
*       Absolute value of integer vector elements, put result in b
*
*       Arguments:
*       A         (i)      argument vector
*       B         (o)      result vector
*       N         (i)      dimension of vectors
*
*       History:       
*       Written November 1991 S. Knight
************************************************************************

        subroutine ivecabs ( a , b , n )

        implicit none

*       Arguments
        integer         n
        integer         a( n ) , b( n )

*       Local varibles
        integer         i
*-----------------------------------------------------------------------
        do i = 1 , n
        b( i ) = abs ( a( i ) )
        end do
        return
        end
************************************************************************
************************************************************************
*
*       Subroutine FVECI
*
*       Function:
*       Convert real n x 1 vector to integer, result in b
*
*       Arguments:
*       A         (i)      real vector
*       B         (o)      integer vector
*       N         (i)      dimension of vectors
*
*       History:
*       Written November 1991 S. Knight
************************************************************************

        subroutine fveci ( a , b , n )

        implicit none

*       Arguments
        integer         n
        real            a( n )
        integer         b( n )

*       Local varibles
        integer         i
*-----------------------------------------------------------------------
        do i = 1 , n
        b( i ) = nint ( a( i ) )
        end do
        return
        end
************************************************************************
************************************************************************
*
*       Subroutine IVECF
*
*       Function:
*       Convert integer n x 1 vector to real, result in b
*
*       Arguments:
*       A         (i)      integer vector
*       B         (o)      real vector
*       N         (i)      dimension of vectors
*
*       History:
*       Written November 1991 S. Knight
************************************************************************

        subroutine ivecf ( a , b , n )

        implicit none

*       Arguments
        integer         n
        integer         a( n )
        real            b( n )

*       Local varibles
        integer         i
*-----------------------------------------------------------------------
        do i = 1 , n
        b( i ) = float ( a( i ) )
        end do
        return
        end
************************************************************************
************************************************************************
*
*       Subroutine TRANSPOSE
*
*       Function:
*       Transpose n x n matrix a, return result in at
*
*       Arguments:
*       A         (i)      matrix to be transposed
*       AT        (o)      transpose of A
*       N         (i)      dimension of matrix 
*
*       History:
*       Prehistoric
************************************************************************

        subroutine transpose ( a , at , n )

        implicit none

*       Arguments
        integer         n
        real            a(n,n) , at(n,n)

*       Local varibles
        integer         i , j
*-----------------------------------------------------------------------
        do i=1,n
        do j=1,n
        at(j,i)=a(i,j)
        end do
        end do
        return
        end
************************************************************************
************************************************************************
*
*       Subroutine MATINV 
*
*       Function:
*       Invert 3 x 3 real matrix 
*
*       Arguments:
*       R         (i)      3 x 3 matrix to be inverted
*       Ri        (o)      3 x 3 inverse matrix
*       Errcod    (o)      error code
*                          errcod = 0 - no error
*                          errcod = 1 - singular matrix
*
*       History:
*       Written 1989 S. Knight
************************************************************************
        subroutine matinv ( r , ri , errcod )

        implicit none


*       Arguments
        real                r( 3,3 ) , ri( 3,3 )
        integer             errcod

*       Local variables
        real                det
*=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
*       Initiate error code
        errcod = 0

*       Compute value of determinant
        det = r(1,1) * ( r(2,2) * r(3,3) - r(2,3) * r(3,2) ) +
     &        r(2,1) * ( r(3,2) * r(1,3) - r(1,2) * r(3,3) ) +
     &        r(3,1) * ( r(1,2) * r(2,3) - r(1,3) * r(2,2) )

*       Check matrix not singular
        if ( abs ( det ) .lt. 0.0000001 ) then
          errcod = 1
          return
        endif

*       Compute elements of inverse matrix
        ri(1,1) =   ( r(2,2) * r(3,3) - r(2,3) * r(3,2) ) / det
        ri(1,2) = - ( r(1,2) * r(3,3) - r(1,3) * r(3,2) ) / det
        ri(1,3) =   ( r(1,2) * r(2,3) - r(1,3) * r(2,2) ) / det
        ri(2,1) = - ( r(2,1) * r(3,3) - r(3,1) * r(2,3) ) / det
        ri(2,2) =   ( r(1,1) * r(3,3) - r(1,3) * r(3,1) ) / det
        ri(2,3) = - ( r(1,1) * r(2,3) - r(1,3) * r(2,1) ) / det
        ri(3,1) =   ( r(2,1) * r(3,2) - r(3,1) * r(2,2) ) / det
        ri(3,2) = - ( r(1,1) * r(3,2) - r(1,2) * r(3,1) ) / det
        ri(3,3) =   ( r(1,1) * r(2,2) - r(1,2) * r(2,1) ) / det

        return

*=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
        end
************************************************************************
************************************************************************
*
*       Subroutine VMTPLY
*
*       Function:
*       Premultiplication of 3 x 1 column vector by 3 x 3 matrix
*
*       Arguments:
*       A         (i)      matrix
*       V         (i)      vector
*       AV        (o)      resulting vector (A*V)
*
*       History:
*       Prehistoric
************************************************************************
        subroutine vmtply ( a , v , av )

        implicit none   
        real            a(3,3),v(3),av(3)

        av(1) = a(1,1)*v(1) + a(1,2)*v(2) + a(1,3)*v(3)
        av(2) = a(2,1)*v(1) + a(2,2)*v(2) + a(2,3)*v(3)
        av(3) = a(3,1)*v(1) + a(3,2)*v(2) + a(3,3)*v(3)
        return
        end
************************************************************************
************************************************************************
*
*       Subroutine MMTPLY
*
*       Function:
*       3 x 3 matrix multiplication
*
*       Arguments:
*       A         (i)      first matrix
*       B         (i)      second matrix
*       AB        (o)      resulting matrix (A*B)
*
*       History:
*       Prehistoric
************************************************************************
        subroutine mmtply ( a , b , ab )
        implicit none
        real            a(3,3) , b(3,3) , ab(3,3)


        ab(1,1) = a(1,1)*b(1,1) + a(1,2)*b(2,1) + a(1,3)*b(3,1)
        ab(1,2) = a(1,1)*b(1,2) + a(1,2)*b(2,2) + a(1,3)*b(3,2)
        ab(1,3) = a(1,1)*b(1,3) + a(1,2)*b(2,3) + a(1,3)*b(3,3)

        ab(2,1) = a(2,1)*b(1,1) + a(2,2)*b(2,1) + a(2,3)*b(3,1)
        ab(2,2) = a(2,1)*b(1,2) + a(2,2)*b(2,2) + a(2,3)*b(3,2)
        ab(2,3) = a(2,1)*b(1,3) + a(2,2)*b(2,3) + a(2,3)*b(3,3)

        ab(3,1) = a(3,1)*b(1,1) + a(3,2)*b(2,1) + a(3,3)*b(3,1)
        ab(3,2) = a(3,1)*b(1,2) + a(3,2)*b(2,2) + a(3,3)*b(3,2)
        ab(3,3) = a(3,1)*b(1,3) + a(3,2)*b(2,3) + a(3,3)*b(3,3)

        return
        end
************************************************************************
************************************************************************
*
*       Subroutine COSABG
*
*       Function:
*       Compute cosine of cell angles
*
*       Arguments:
*       A         (i)      cell angles
*       CA        (o)      cosine of cell angles
*
*       History:
*       Written September 1991 S. Knight
************************************************************************
        subroutine cosabg ( a , ca )
        implicit none
        real            a( 3 ) , ca( 3 )
        real            conv
        parameter       ( conv = 3.1415927/180. )
        integer         i

        do i = 1 , 3
          if ( abs ( a( i ) - 90. ) .lt. 0.01 ) then
            ca( i ) = 0.0
          else
            ca( i ) = cos ( conv * a( i ) )
          endif
        end do

        return
        end
************************************************************************
************************************************************************
*
*       Subroutine ORTHO
*
*       Function:
*       Set matrix for conversion between crystallographic and
*       orthogonal frame
*       
*       Comments:
*       Adapted from CCP4 routine RBFRO1
*       
*       Arguments:
*       Cell      (i)    unit cell parameters
*       Ortcod    (i)    orthogonalization code:
*                        ortcod = 1 >        a  // xo   c* // zo
*                               = 2 >        b  // xo   a* // zo
*                               = 3 >        c  // xo   b* // zo
*                               = 4 >  hex a+b  // xo   c* // zo
*                               = 5 >        a* // xo   c  // zo
*                               = 6 >        a  // xo   b* // yo
*       Way       (i)    direction of conversion:
*                        way = DEORT > orthogonal to crystallographic
*                        way = TOORT > crystallographic to orthogonal
*       Rort      (o)    (de)orthogonalization matrix
*       Errcod    (o)    error code
*                        errcod = 0 - no error
*                        errcod = 1 - illegal orthogonalization code 
*                        errcod = 2 - matrix inversion error
*                        errcod = 3 - illegal WAY string
*       
*       
*       History:
*       Written December 12 1991 (from CCP4 s/r RBFRO1) S. Knight
************************************************************************
        subroutine ortho ( cell , ortcod , way , rort , errcod )

        implicit none

*       Arguments
        real              cell( 6 )
        integer           ortcod
        character*(*)     way
        real              rort( 3,3 )
        integer           errcod

*       Local variables
        real              dtor
        real              r( 3,3 )
        real              a , b , c , alpha , beta , gamma
        real              sina , cosa , sinb , cosb , sing , cosg
        real              cosas , cosbs , cosgs , sinas , sinbs , sings
        parameter         ( dtor = 3.141592/180. )
*=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
*       Init error code
        errcod = 0

*       Initiate useful constants
        a      = cell(1)
        b      = cell(2)
        c      = cell(3)
        alpha  = cell(4) * dtor
        beta   = cell(5) * dtor
        gamma  = cell(6) * dtor
        sina   = sin ( alpha )
        cosa   = cos ( alpha )
        sinb   = sin ( beta )
        cosb   = cos ( beta )
        sing   = sin ( gamma )
        cosg   = cos ( gamma )
        cosas  = ( cosg * cosb - cosa ) / ( sinb * sing )
        cosbs  = ( cosa * cosg - cosb ) / ( sina * sing )
        cosgs  = ( cosa * cosb - cosg ) / ( sina * sinb )
        sinas  = sqrt ( 1.0 - cosas * cosas )
        sinbs  = sqrt ( 1.0 - cosbs * cosbs )
        sings  = sqrt ( 1.0 - cosgs * cosgs )

*       Init matrix
        call matclr ( r , 3 , 3 )

*
*       Calculate matrix (fractional crystallographic to Angstrom orthogonal)
*
        if ( ortcod .eq. 1 ) then

*       XO along a  Zo along c*
          r(1,1) =  a
          r(1,2) =  b * cosg
          r(1,3) =  c * cosb
          r(2,2) =  b * sing
          r(2,3) = -c * sinb * cosas
          r(3,3) =  c * sinb * sinas

        else if ( ortcod .eq. 2 ) then

*       XO along b  Zo along a*
          r(3,1) =  a * sing * sinbs
          r(1,1) =  a * cosg
          r(1,2) =  b
          r(1,3) =  c * cosa
          r(2,1) = -a * sing * cosbs
          r(2,3) =  c * sina

        else if ( ortcod .eq. 3 ) then

*       XO along c  Zo along b*
          r(2,1) =  a * sinb
          r(2,2) = -b * sina * cosgs
          r(3,2) =  b * sina * sings
          r(1,1) =  a * cosb
          r(1,2) =  b * cosa
          r(1,3) =  c

        else if ( ortcod .eq. 4 ) then

*       trigonal only - XO along a+b  YO along a-b  Zo along c*
          r(3,3) =  c
          r(1,1) =  a / 2.0
          r(1,2) =  a / 2.0
          r(2,1) = -a * sing
          r(2,2) =  a * sing

        else if ( ortcod .eq. 5 ) then

*       XO along a*   ZO along c
          r(1,1) =  a * sinb * sings
          r(3,1) =  a * cosb
          r(3,2) =  b * cosa
          r(3,3) =  c
          r(2,1) = -a * sinb * cosgs
          r(2,2) =  b * sina

        else if ( ortcod .eq. 6 ) then

*       XO along a  Yo along b*
          r(1,1) =  a
          r(1,2) =  b * cosg
          r(1,3) =  c * cosb
          r(2,2) =  b * sing * sinas
          r(3,2) = -b * sing * cosas
          r(3,3) =  c * sinb

        else

          errcod = 1

        endif


*       Copy matrix if crystallographic to orthogonal wanted
*       Invert matrix if orthogonal to crystallographic wanted
        if ( way .eq. 'TOORT' ) then
          call matcop ( r , rort , 3 , 3 )
        else if ( way .eq. 'DEORT' ) then
          call matinv ( r , rort , errcod )
          if ( errcod .ne. 0 ) errcod = 2
        else
          errcod = 3
        endif


        return

*=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
        end
*********************************************************************sdk
************************************************************************
*       Subroutine pnt2pln    
*
*       Function: get plane equation from three points
*
*       Arguments:
*       Points    (i)     Coordinates of three points
*       Plane     (o)     Plane equation
*       Errcod    (o)     Error code
*                         errcod = 0 - no error
*                         errcod = 1 - problem is ill conditioned
*
*
*       Result:
*       Return normal vector (PLANE(1-3)) and distance from origin
*       (PLANE(4))
*
*       History:
*       Written february 1996 sdk
************************************************************************
        subroutine pnt2pln ( points , plane , errcod )

        implicit none

*       Arguments
        integer           errcod
        real              points( 3,3 )
        real              plane( 4 )

*       Local
        integer           i
        real              r1( 3 ) , r2( 3 ) , d
        real              mind
        parameter         ( mind = 0.000001 )
*=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
*       Init error code
        errcod = 0

*       Form difference vectors
        do i = 1 , 3
          r1( i ) = points( i,2 ) - points( i,1 )
          r2( i ) = points( i,3 ) - points( i,1 )
        end do

*       Normal vector
        plane( 1 ) = r1( 2 ) * r2( 3 ) - r1( 3 ) * r2( 2 )
        plane( 2 ) = r1( 3 ) * r2( 1 ) - r1( 1 ) * r2( 3 )
        plane( 3 ) = r1( 1 ) * r2( 2 ) - r1( 2 ) * r2( 1 )
        d = sqrt ( plane( 1 )**2 + plane( 2 )**2 + plane( 3 )**2 )
*       Check that vectors aren't co-linear
        if ( d .lt. mind ) then
*       Ooops, problem is ill conditioned
          errcod = 1
          return
        endif
        plane( 1 ) = plane( 1 ) / d
        plane( 2 ) = plane( 2 ) / d
        plane( 3 ) = plane( 3 ) / d


*       Distance from origin
        plane( 4 ) = plane( 1 )*points( 1,1 ) + 
     &               plane( 2 )*points( 2,1 ) + 
     &               plane( 3 )*points( 3,1 )


        return
        end
*********************************************************************sdk
************************************************************************
*       Real function PlnDst
*
*       Function:
*       Return distance from point to plane
*
*       Arguments:
*       Point     (i)     coordinates of point
*       Plane     (i)     plane normal vector and dist from origin
*
*       History:
*       Written february 1996 sdk
************************************************************************
        real function plndst ( point , plane )
        implicit none
        real        point( 3 ) , plane( 4 )


        plndst = plane( 1 ) * point( 1 ) +
     &           plane( 2 ) * point( 2 ) +
     &           plane( 3 ) * point( 3 ) -
     &           plane( 4 )

        return
        end
*********************************************************************sdk
************************************************************************
*
*       Subroutine POLAR2MTX
*
*       Function:
*       Convert polar angles to rotation matrix 
*       
*       Comments:
*       Definition of angles:
*
*       Kappa is the clockwise rotation around the rotation axis looking
*       away from the origin.
*       Phi is the angle between the x axis and the projection of the rotation
*       axis on the xy plane. A positive rotation is a clockwise rotation when
*       looking down the z axis.
*       Omega is the angle between the z axis and the rotation axis.
*
*       
*       Arguments:
*       Omega     (i)     omega angle in degrees
*       Phi       (i)     phi angle in degrees
*       Kappa     (i)     kappa angle in degrees
*       Rot       (o)     3 x 3 rotation matrix
*       
*       
*       History:
*       Subroutine from standalone program February 13 1992 sdk
*       Written February 1990 S. Knight
************************************************************************
        subroutine polar2mtx ( omega , phi , kappa , rot )

        implicit none

*       Arguments
        real              omega , phi , kappa , rot(3,3)

*       Local variables
        integer           i , j
        double precision  domega , dphi , dkappa , r(3,3)
        double precision  atr , a1 , a2 , a3 , a4 , a5 , a6

        parameter         (atr = 3.141592653589793/180.)
*=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
*       Convert to radians and double precision
        domega = atr * omega
        dphi   = atr * phi
        dkappa = atr * kappa

*       a1,a2,a3 = direction cosines of rotation axis
        a1 = sin ( domega ) * cos( dphi )
        a2 = sin ( domega ) * sin( dphi )
        a3 = cos ( domega )
        a4 = sin ( dkappa )
        a5 = cos ( dkappa )
        a6 = 1 - a5

*       First row
        r(1,1) = a5 + a6 * a1**2
        r(1,2) = a1 * a2 * a6 - a3 * a4
        r(1,3) = a1 * a3 * a6 + a2 * a4

*       Second row
        r(2,1) = a1 * a2 * a6 + a3 * a4
        r(2,2) = a5 + a6 * a2**2
        r(2,3) = a2 * a3 * a6 - a1 * a4

*       Third row
        r(3,1) = a1 * a3 * a6 - a2 * a4
        r(3,2) = a2 * a3 * a6 + a1 * a4
        r(3,3) = a5 + a6 * a3**2

*       Copy result
        do 20 j = 1 , 3
          do 10 i = 1 , 3 
            rot( i,j ) = r( i,j )
 10       continue
 20     continue


        return
*=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
        end
*********************************************************************SDK
************************************************************************
*
*       Subroutine NFOLD2MTX
*
*       Function:
*       Convert direction cosines and rotation angle to rotation matrix 
*       
*       Comments:
*
*       
*       Arguments:
*       Dircos    (i)     direction cosines of rotation axis
*       Rotang    (i)     rotation angle
*       Rot       (o)     3 x 3 rotation matrix
*       
*       
*       History:
*       From Polar2Mtx March 3 1994 S. Knight
************************************************************************
        subroutine NFold2Mtx ( dircos , rotang , rot )

        implicit none

*       Arguments
        real              dircos( 3 ) , rotang , rot(3,3)

*       Local variables
        integer           i , j
        double precision  dkappa , r(3,3)
        double precision  atr , a1 , a2 , a3 , a4 , a5 , a6

        parameter         (atr = 3.141592653589793/180.)
*=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
*       a1,a2,a3 = direction cosines of rotation axis
        a1 = dircos( 1 )
        a2 = dircos( 2 )
        a3 = dircos( 3 )
        dkappa = atr * rotang
        a4 = sin ( dkappa )
        a5 = cos ( dkappa )
        a6 = 1 - a5

*       First row
        r(1,1) = a5 + a6 * a1**2
        r(1,2) = a1 * a2 * a6 - a3 * a4
        r(1,3) = a1 * a3 * a6 + a2 * a4

*       Second row
        r(2,1) = a1 * a2 * a6 + a3 * a4
        r(2,2) = a5 + a6 * a2**2
        r(2,3) = a2 * a3 * a6 - a1 * a4

*       Third row
        r(3,1) = a1 * a3 * a6 - a2 * a4
        r(3,2) = a2 * a3 * a6 + a1 * a4
        r(3,3) = a5 + a6 * a3**2

*       Copy result
        do 20 j = 1 , 3
          do 10 i = 1 , 3 
            rot( i,j ) = r( i,j )
 10       continue
 20     continue


        return
*=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
        end
*********************************************************************SDK
************************************************************************
*
*       Subroutine MSK
*
*       Function:
*       .and. two logical arrays. Count number of resulting true
*       elements, store pointers to true elements.
*
*       Arguments:
*       A         (i)      logical array 1
*       B         (i)      logical array 2
*       C         (o)      resulting logical array (A .and. B)
*       D         (o)      array of pointer to true elements in C
*       N         (i)      dimension of arrays
*       Neq       (o)      number of true elements in C
*
*       History:
*       Written 1988 S. Knight
************************************************************************
        subroutine msk ( a , b , c , d , n , neq )
        implicit none        

*       Arguments
        integer         n , neq
        integer         d(n)
        logical         a(n),b(n),c(n)

*       Local variables
        integer         i

        neq=0
        do i=1,n
        c(i)=(a(i).and.b(i))
        if(c(i))then
        neq=neq+1
        d(neq)=i
        endif
        end do
        return
        end
*********************************************************************sdk
************************************************************************
*
*       Subroutine MSKCMP
* 
*       Function:
*       Compare two logical arrays
*
*       Arguments:
*       A         (i)      logical array 1
*       B         (i)      logical array 2
*       N         (i)      dimension of arrays
*       Match     (o)      number of identical elements in A and B
*
*       History:
*       Written 1988 S. Knight
************************************************************************
        subroutine mskcmp ( a , b , n , match )
        implicit none
        integer         n , match
        logical         a(n),b(n)
        integer         i
        match=0
        do i=1,n
        if( ( a(i) .and. b(i) ) .or. 
     &  ( .not. a(i) .and. .not. b(i) ) ) match = match + 1
        end do
        return
        end
*********************************************************************sdk
************************************************************************
*
*       Subroutine ISORT
*
*       Function:
*       Sort value in descending order, co-sort peaks (integer)
*
*       Method:
*       Sorting by straight insertion
*
*       Arguments:
*       Np       (i)       number of elements to sort
*       Peaks    (i/o)     associated 3 x np integer array
*       Value    (i/o)     sort key real array
*
*       History:
*       October 2 1991 sdk
*       Modified to co-sort unpacked peak coordinates
*       March 14 1988 sdk
************************************************************************
        subroutine isort ( np , peaks , value )

        implicit none

*       Arguments
        integer         np
        real            value(np)
        integer         peaks(3,np)

*       Local variables
        integer         i , j
        real            valsav
        integer         peksav(3)

*-----------------------------------------------------------------------

        do i = 1 , np - 1
        valsav    = value(i+1)
        peksav(1) = peaks(1,i+1)
        peksav(2) = peaks(2,i+1)
        peksav(3) = peaks(3,i+1)
        do j = i , 1 , -1
          if ( valsav .le. value(j) ) then
            goto 20
          else
            value(j+1)   = value(j)
            peaks(1,j+1) = peaks(1,j)
            peaks(2,j+1) = peaks(2,j)
            peaks(3,j+1) = peaks(3,j)
          endif
        end do
 20     value(j+1)   = valsav
        peaks(1,j+1) = peksav(1)
        peaks(2,j+1) = peksav(2)
        peaks(3,j+1) = peksav(3)
        end do

        return
        end
*********************************************************************sdk
************************************************************************
*
*       Subroutine FSORT
*
*       Function:
*       Sort value in descending order, co-sort peaks (real)
*
*       Method:
*       Sorting by straight insertion
*
*       Arguments:
*       Np       (i)       number of elements to sort
*       Peaks    (i/o)     associated 3 x np real array
*       Value    (i/o)     sort key real array
*
*       History:
*       October 2 1991 sdk
*       Modified to co-sort unpacked peak coordinates
*       March 14 1988 sdk
************************************************************************
        subroutine fsort ( np , peaks , value )

        implicit none

*       Arguments
        integer         np
        real            value(np)
        real            peaks(3,np)

*       Local variables
        integer         i , j
        real            valsav
        real            peksav(3)

*-----------------------------------------------------------------------

        do i = 1 , np - 1
        valsav    = value(i+1)
        peksav(1) = peaks(1,i+1)
        peksav(2) = peaks(2,i+1)
        peksav(3) = peaks(3,i+1)
        do j = i , 1 , -1
          if ( valsav .le. value(j) ) then
            goto 20
          else
            value(j+1)   = value(j)
            peaks(1,j+1) = peaks(1,j)
            peaks(2,j+1) = peaks(2,j)
            peaks(3,j+1) = peaks(3,j)
          endif
        end do
 20     value(j+1)   = valsav
        peaks(1,j+1) = peksav(1)
        peaks(2,j+1) = peksav(2)
        peaks(3,j+1) = peksav(3)
        end do

        return
        end
*********************************************************************sdk
************************************************************************
*
*       Subroutine QSORTFF
*
*       Function:
*       Sort real value in descending order, co-sort peaks (real)
*
*       Method:
*       Quicksort (H.A.R. Hoare, "Quicksort", Comp. J. 5(1):10-15 
*       (1962).
*
*       Comment:
*       Code translated from Pascal version in
*       N. Wirth, Algorithms + Data Structures = Programs,
*       1976 Prentice-Hall Inc, Englewood Cliffs, New Jersey.
*
*       Arguments:
*       Np       (i)       number of elements to sort
*       Peaks    (i/o)     associated 3 x np real array
*       Value    (i/o)     sort key real array
*       Errcod   (o)       error code
*                          errcod = 0 - no error
*                          errcod = 1 - stack overflow
*
*       History:
*       Tested OK November 23 1991 sdk
*       First attempts November 23 1991 S. Knight
************************************************************************
        subroutine qsortff ( np , peaks , value , errcod )

        implicit none

*       Arguments
        integer         np
        real            value(np)
        real            peaks(3,np)
        integer         errcod

*       Local variables
        integer         i , j , l , r , s , x
        integer         m
        parameter       ( m = 12 )
        integer         stack( 2,m )

        real            valsav
        real            peksav(3)

*-----------------------------------------------------------------------
*       Initialize error code
        errcod = 0

*       Don't bother if nothing to sort
        if ( np .le. 1 ) return

*       Initialize stack
        s = 1
        stack( 1,1 ) = 1
        stack( 2,1 ) = np



*       Conditional loop 1
 100    continue
*       Take top request from stack
          l = stack( 1,s )
          r = stack( 2,s )
          s = s - 1

*       Conditional loop 2
 200      continue
*       Split interval l-r until nothing left, chose middle of 
*       interval as comparand
            i = l
            j = r
            x = ( i + j ) / 2

*       Conditional loop 3
 300        continue 
*       Partition
              if ( value( i ) .gt. value( x ) ) then
                i = i + 1
                goto 300
              endif
 310          if ( value( x ) .gt. value( j ) ) then
                j = j - 1
                goto 310
              endif

*       Swap and if necessary reset comparand pointer
              if ( i .le. j ) then
                valsav       = value( i )
                peksav( 1 )  = peaks( 1,i )
                peksav( 2 )  = peaks( 2,i )
                peksav( 3 )  = peaks( 3,i )

                value( i )   = value( j )
                peaks( 1,i ) = peaks( 1,j )
                peaks( 2,i ) = peaks( 2,j )
                peaks( 3,i ) = peaks( 3,j )

                value( j )   = valsav
                peaks( 1,j ) = peksav( 1 )
                peaks( 2,j ) = peksav( 2 )
                peaks( 3,j ) = peksav( 3 )

                if ( x .eq. i ) then
                  x = j
                else if ( x .eq. j ) then
                  x = i
                endif

                i = i + 1
                j = j - 1
              endif

*       End of conditional loop 3
            if ( i .le. j ) goto 300


*       Set up requests: short interval now, long interval to stack
            if ( j - l .lt. r - i ) then
              if ( i .lt. r ) then
*       Stack request for sorting right partition
                s = s + 1
                if ( s .gt. m ) goto 999
                stack( 1,s ) = i
                stack( 2,s ) = r
              endif
*       and continue sorting left partition 
              r = j
            else
              if ( l .lt. j ) then
*       Stack request for sorting left partition
                s = s + 1
                if ( s .gt. m ) goto 999
                stack( 1,s ) = l
                stack( 2,s ) = j
              endif
*       and continue sorting right partition 
              l = i
            endif

*       End of conditional loop 2
          if ( l .lt. r ) goto 200


*       End of conditional loop 3
        if ( s .gt. 0 ) goto 100

*       All done
        return

 999    errcod = 1
        return

        end
*********************************************************************sdk
************************************************************************
*
*       Subroutine QSORTFI
*
*       Function:
*       Sort real value in descending order, co-sort peaks (integer)
*
*       Method:
*       Quicksort (H.A.R. Hoare, "Quicksort", Comp. J. 5(1):10-15 
*       (1962).
*
*       Comment:
*       Code translated from Pascal version in
*       N. Wirth, Algorithms + Data Structures = Programs,
*       1976 Prentice-Hall Inc, Englewood Cliffs, New Jersey.
*
*       Arguments:
*       Np       (i)       number of elements to sort
*       Peaks    (i/o)     associated 3 x np integer array
*       Value    (i/o)     sort key real array
*       Errcod   (o)       error code
*                          errcod = 0 - no error
*                          errcod = 1 - stack overflow
*
*       History:
*       Tested OK November 23 1991 sdk
*       First attempts November 23 1991 S. Knight
************************************************************************
        subroutine qsortfi ( np , peaks , value , errcod )

        implicit none

*       Arguments
        integer         np
        real            value(np)
        integer         peaks(3,np)
        integer         errcod

*       Local variables
        integer         i , j , l , r , s , x
        integer         m
        parameter       ( m = 12 )
        integer         stack( 2,m )

        real            valsav
        integer         peksav(3)

*-----------------------------------------------------------------------
*       Initialize error code
        errcod = 0

*       Don't bother if nothing to sort
        if ( np .le. 1 ) return

*       Initialize stack
        s = 1
        stack( 1,1 ) = 1
        stack( 2,1 ) = np



*       Conditional loop 1
 100    continue
*       Take top request from stack
          l = stack( 1,s )
          r = stack( 2,s )
          s = s - 1

*       Conditional loop 2
 200      continue
*       Split interval l-r until nothing left, chose middle of 
*       interval as comparand
            i = l
            j = r
            x = ( i + j ) / 2

*       Conditional loop 3
 300        continue 
*       Partition
              if ( value( i ) .gt. value( x ) ) then
                i = i + 1
                goto 300
              endif
 310          if ( value( x ) .gt. value( j ) ) then
                j = j - 1
                goto 310
              endif

*       Swap and if necessary reset comparand pointer
              if ( i .le. j ) then
                valsav       = value( i )
                peksav( 1 )  = peaks( 1,i )
                peksav( 2 )  = peaks( 2,i )
                peksav( 3 )  = peaks( 3,i )

                value( i )   = value( j )
                peaks( 1,i ) = peaks( 1,j )
                peaks( 2,i ) = peaks( 2,j )
                peaks( 3,i ) = peaks( 3,j )

                value( j )   = valsav
                peaks( 1,j ) = peksav( 1 )
                peaks( 2,j ) = peksav( 2 )
                peaks( 3,j ) = peksav( 3 )

                if ( x .eq. i ) then
                  x = j
                else if ( x .eq. j ) then
                  x = i
                endif

                i = i + 1
                j = j - 1
              endif

*       End of conditional loop 3
            if ( i .le. j ) goto 300


*       Set up requests: short interval now, long interval to stack
            if ( j - l .lt. r - i ) then
              if ( i .lt. r ) then
*       Stack request for sorting right partition
                s = s + 1
                if ( s .gt. m ) goto 999
                stack( 1,s ) = i
                stack( 2,s ) = r
              endif
*       and continue sorting left partition 
              r = j
            else
              if ( l .lt. j ) then
*       Stack request for sorting left partition
                s = s + 1
                if ( s .gt. m ) goto 999
                stack( 1,s ) = l
                stack( 2,s ) = j
              endif
*       and continue sorting right partition 
              l = i
            endif

*       End of conditional loop 2
          if ( l .lt. r ) goto 200


*       End of conditional loop 3
        if ( s .gt. 0 ) goto 100

*       All done
        return

 999    errcod = 1
        return

        end
*********************************************************************sdk
************************************************************************
*
*       Subroutine QSORTFFI
*
*       Function:
*       Sort real value in descending order, co-sort real and integer
*       arrays
*
*       Method:
*       Quicksort (H.A.R. Hoare, "Quicksort", Comp. J. 5(1):10-15 
*       (1962).
*
*       Comment:
*       Code translated from Pascal version in
*       N. Wirth, Algorithms + Data Structures = Programs,
*       1976 Prentice-Hall Inc, Englewood Cliffs, New Jersey.
*
*       Arguments:
*       Np       (i)       number of elements to sort
*       Peaks    (i/o)     associated 3 x np real array
*       Value    (i/o)     sort key real array
*       Type     (i/o)     associated 1 x np integer array
*       Errcod   (o)       error code
*                          errcod = 0 - no error
*                          errcod = 1 - stack overflow
*
*       History:
*       Tested OK November 23 1991 sdk
*       First attempts November 23 1991 S. Knight
************************************************************************
        subroutine qsortffi ( np , peaks , value , type , errcod )

        implicit none

*       Arguments
        integer         np
        real            value(np)
        real            peaks(3,np)
        integer         type(np)
        integer         errcod

*       Local variables
        integer         i , j , l , r , s , x
        integer         m
        parameter       ( m = 12 )
        integer         stack( 2,m )

        real            valsav
        real            peksav(3)
        integer         typsav
*-----------------------------------------------------------------------
*       Initialize error code
        errcod = 0

*       Don't bother if nothing to sort
        if ( np .le. 1 ) return

*       Initialize stack
        s = 1
        stack( 1,1 ) = 1
        stack( 2,1 ) = np



*       Conditional loop 1
 100    continue
*       Take top request from stack
          l = stack( 1,s )
          r = stack( 2,s )
          s = s - 1

*       Conditional loop 2
 200      continue
*       Split interval l-r until nothing left, chose middle of 
*       interval as comparand
            i = l
            j = r
            x = ( i + j ) / 2

*       Conditional loop 3
 300        continue 
*       Partition
              if ( value( i ) .gt. value( x ) ) then
                i = i + 1
                goto 300
              endif
 310          if ( value( x ) .gt. value( j ) ) then
                j = j - 1
                goto 310
              endif

*       Swap and if necessary reset comparand pointer
              if ( i .le. j ) then
                valsav       = value( i )
                typsav       = type( i )
                peksav( 1 )  = peaks( 1,i )
                peksav( 2 )  = peaks( 2,i )
                peksav( 3 )  = peaks( 3,i )
                
                value( i )   = value( j )
                type( i )    = type( j )
                peaks( 1,i ) = peaks( 1,j )
                peaks( 2,i ) = peaks( 2,j )
                peaks( 3,i ) = peaks( 3,j )

                value( j )   = valsav
                type( j )    = typsav
                peaks( 1,j ) = peksav( 1 )
                peaks( 2,j ) = peksav( 2 )
                peaks( 3,j ) = peksav( 3 )

                if ( x .eq. i ) then
                  x = j
                else if ( x .eq. j ) then
                  x = i
                endif

                i = i + 1
                j = j - 1
              endif

*       End of conditional loop 3
            if ( i .le. j ) goto 300


*       Set up requests: short interval now, long interval to stack
            if ( j - l .lt. r - i ) then
              if ( i .lt. r ) then
*       Stack request for sorting right partition
                s = s + 1
                if ( s .gt. m ) goto 999
                stack( 1,s ) = i
                stack( 2,s ) = r
              endif
*       and continue sorting left partition 
              r = j
            else
              if ( l .lt. j ) then
*       Stack request for sorting left partition
                s = s + 1
                if ( s .gt. m ) goto 999
                stack( 1,s ) = l
                stack( 2,s ) = j
              endif
*       and continue sorting right partition 
              l = i
            endif

*       End of conditional loop 2
          if ( l .lt. r ) goto 200


*       End of conditional loop 3
        if ( s .gt. 0 ) goto 100

*       All done
        return

 999    errcod = 1
        return

        end
*********************************************************************sdk
************************************************************************
*
*       Subroutine QSORTPF
*
*       Function:
*       Sort pointers in pntarr on values in key (descending)
*
*       Method:
*       Quicksort (H.A.R. Hoare, "Quicksort", Comp. J. 5(1):10-15 
*       (1962).
*
*       Comment:
*       Code translated and adapted from Pascal version in
*       N. Wirth, Algorithms + Data Structures = Programs,
*       1976 Prentice-Hall Inc, Englewood Cliffs, New Jersey.
*
*       Arguments:
*       Np       (i)       number of elements to sort
*       Pntarr   (i/o)     1 x np integer array of pointers
*       Key      (i/o)     sort key real array (1 x np)
*       Errcod   (o)       error code
*                          errcod = 0 - no error
*                          errcod = 1 - stack overflow
*
*       History:
*       First attempts November 23 1991 S. Knight
************************************************************************
        subroutine qsortpf ( np , pntarr , key , errcod )

        implicit none

*       Arguments
        integer         np
        integer         pntarr( np )
        real            key( np )
        integer         errcod

*       Local variables
        integer         i , j , l , r , s , x
        integer         m
        parameter       ( m = 12 )
        integer         stack( 2,m )
        integer         pntsav
*-----------------------------------------------------------------------
*       Initialize error code
        errcod = 0

*       Don't bother if nothing to sort
        if ( np .le. 1 ) return

*       Initialize stack
        s = 1
        stack( 1,1 ) = 1
        stack( 2,1 ) = np



*       Conditional loop 1
 100    continue
*       Take top request from stack
          l = stack( 1,s )
          r = stack( 2,s )
          s = s - 1

*       Conditional loop 2
 200      continue
*       Split interval l-r until nothing left, chose middle of 
*       interval as comparand
            i = l
            j = r
            x = ( i + j ) / 2

*       Conditional loop 3
 300        continue 
*       Partition
              if ( key( pntarr(i) ) .gt. key( pntarr(x) ) ) then
                i = i + 1
                goto 300
              endif
 310          if ( key( pntarr(x) ) .gt. key( pntarr(j) ) ) then
                j = j - 1
                goto 310
              endif

*       Swap and if necessary reset comparand pointer
              if ( i .le. j ) then
                pntsav      = pntarr( i )
                pntarr( i ) = pntarr( j )
                pntarr( j ) = pntsav
                if ( x .eq. i ) then
                  x = j
                else if ( x .eq. j ) then
                  x = i
                endif

                i = i + 1
                j = j - 1
              endif

*       End of conditional loop 3
            if ( i .le. j ) goto 300


*       Set up requests: short interval now, long interval to stack
            if ( j - l .lt. r - i ) then
              if ( i .lt. r ) then
*       Stack request for sorting right partition
                s = s + 1
                if ( s .gt. m ) goto 999
                stack( 1,s ) = i
                stack( 2,s ) = r
              endif
*       and continue sorting left partition 
              r = j
            else
              if ( l .lt. j ) then
*       Stack request for sorting left partition
                s = s + 1
                if ( s .gt. m ) goto 999
                stack( 1,s ) = l
                stack( 2,s ) = j
              endif
*       and continue sorting right partition 
              l = i
            endif

*       End of conditional loop 2
          if ( l .lt. r ) goto 200


*       End of conditional loop 3
        if ( s .gt. 0 ) goto 100

*       All done
        return

 999    errcod = 1
        return

        end
*********************************************************************sdk
************************************************************************
*
*       Subroutine NSORTA
*
*       Function:
*       Sort N columns on values in given column, acending
*
*       Method:
*       Sorting by straight insertion
*
*       Arguments:
*       MaxC     (i)       fast dimension of sort matrix
*       MaxR     (i)       slow dimension of sort matrix
*       NC       (i)       number of columns
*       NR       (i)       number of rows
*       Key      (i)       pointer to sort-column
*       Srtmtx   (i/o)     nc x nr integer matrix to sort
*       Errcod   (o)       error code
*                          errcod = 0 - no error
*                          errcod = 1 - illegal key value
*                          errcod = 2 - too many columns
*
*       History:
*       Modified from ISORT october 6 - S. Knight
************************************************************************
        subroutine nsorta ( maxc , maxr , nc , nr , key , srtmtx , 
     &  errcod )

        implicit none

*       Arguments
        integer          maxc , maxr
        integer          nc , nr
        integer          key
        integer          srtmtx( maxc,maxr )
        integer          errcod

*       Local variables
        integer          i , j , k
        integer          maxcol
        parameter        ( maxcol = 40 )
        integer          isave( maxcol )

*-----------------------------------------------------------------------
*       Init
        errcod = 0

*       Check input
        if ( key .lt. 1 .or. key .gt. nc ) then
          errcod = 1
          return
        endif
        if ( nc .gt. maxcol ) then
          errcod = 2
          return
        endif
      
*       Sort acending
        do 70 i = 1 , nr - 1
          do 10 j = 1 , nc
            isave( j ) = srtmtx( j,i+1 )
 10       continue
          do 40 j = i , 1 , -1
            if ( isave( key ) .ge. srtmtx( key,j ) ) then
              goto 50
            else
              do 20 k = 1 , nc
                srtmtx( k,j+1 ) = srtmtx( k,j )
 20           continue
            endif
 40       continue
 50       do 60 k = 1 , nc
            srtmtx( k,j+1 ) = isave( k )
 60       continue
 70     continue

        return
        end
*********************************************************************sdk
************************************************************************
*
*       Real function DET
*
*       Function:
*       Return determinant of 3 x 3 matrix
*
*       Arguments:
*       A         (i)      3 x 3 real matrix
*
*       History:
*       Prehistoric
************************************************************************
        real function det ( a )
        implicit none
        real            a(3,3)


        det = a(1,1)*a(2,2)*a(3,3) +
     &        a(1,2)*a(2,3)*a(3,1) +

     &        a(1,3)*a(2,1)*a(3,2) -
     &        a(1,3)*a(2,2)*a(3,1) -
     &        a(1,2)*a(2,1)*a(3,3) -
     &        a(1,1)*a(2,3)*a(3,2)

        return
        end
************************************************************************
************************************************************************
*
*       Real function VLEN
*
*       Function:
*       Calculate length in Angstrom of vector u.
*
*       Comment:
*       The vector is reduced to the interval -0.5 - 0.5.
*
*       Arguments :
*       U       (i)     3 x 1 vector (fractional)
*       Cell    (i)     cell parameters in angstrom and degrees
*       Cosang  (i)     cosines of the cell angles
*
*
*       History:
*       November 24 1991 sdk
*       Changed from s/r to function
*       November 8 1991 sdk
*       Much simplified, now just straight distance from fractional
*       coordinates
*       November 1988 sdk
*       Flag argument added
*       Written March 1988 S. Knight
************************************************************************
        real function vlen ( u , cell , cosang )

        implicit none

*       Arguments
        real            u(3) , cell(6) , cosang(3)

*       Local variables
        integer         i
        real            v(3)
        real            d(3)
*-----------------------------------------------------------------------
        do i = 1 , 3
          v( i ) = u( i )
          if ( v( i ) .gt.  0.5 ) v( i ) = v( i ) - 1.0
          if ( v( i ) .lt. -0.5 ) v( i ) = v( i ) + 1.0
          d( i ) = v( i ) * cell( i )
        end do

        vlen = sqrt ( d(1)**2 + d(2)**2 + d(3)**2 +
     &                d(1) * d(2) * cosang(3)     +
     &                d(1) * d(3) * cosang(2)     +
     &                d(2) * d(3) * cosang(1)     )

        return
        end
*********************************************************************sdk
************************************************************************
*
*       Real function VLEN2
*
*       Function:
*       Calculate length in Angstrom of vector u.
*
*       Comment:
*       The vector is used as passed.
*
*       Arguments :
*       U       (i)     3 x 1 vector (fractional)
*       Cell    (i)     cell parameters in angstrom and degrees
*       Cosang  (i)     cosines of the cell angles
*
*
*       History:
*       From VLEN February 16 1994 S. Knight
************************************************************************
        real function vlen2 ( u , cell , cosang )

        implicit none

*       Arguments
        real            u(3) , cell(6) , cosang(3)

*       Local variables
        integer         i
        real            d(3)
*-----------------------------------------------------------------------
	do i = 1 , 3
          d( i ) = u( i ) * cell( i )
        end do

        vlen2 = sqrt ( d(1)**2 + d(2)**2 + d(3)**2 +
     &                 d(1) * d(2) * cosang(3)     +
     &                 d(1) * d(3) * cosang(2)     +
     &                 d(2) * d(3) * cosang(1)     )

        return
        end
*********************************************************************sdk
************************************************************************
*
*       Integer function IMAXVC
*
*       Function:
*       Return value of largest element in vector (integer)
*
*       Arguments :
*       V       (i)     1 x N integer vector
*       N       (i)     Number of elements in vector
*
*
*       History:
*       Written July 2 1992 S. Knight
************************************************************************
        integer function imaxvc ( v , n )

        implicit none

*       Arguments
        integer         n
        integer         v( n )

*       Local variables
        integer         i
*-----------------------------------------------------------------------
        imaxvc = -214783648 ! -1 * (2**31)
        do 10 i = 1 , n
         imaxvc = max ( imaxvc , v( i ) )
 10     continue
        return
        end
*********************************************************************sdk
************************************************************************
*
*       Integer function IMINVC
*
*       Function:
*       Return value of smallest element in vector (integer)
*
*       Arguments :
*       V       (i)     1 x N integer vector
*       N       (i)     Number of elements in vector
*
*
*       History:
*       Written July 2 1992 S. Knight
************************************************************************
        integer function iminvc ( v , n )

        implicit none

*       Arguments
        integer         n
        integer         v( n )

*       Local variables
        integer         i
*-----------------------------------------------------------------------
        iminvc = 2147483647 ! (2**31) - 1
        do 10 i = 1 , n
         iminvc = min ( iminvc , v( i ) )
 10     continue
        return
        end
*********************************************************************sdk
************************************************************************
*
*       Real function FMAXVC
*
*       Function:
*       Return value of largest element in vector (real)
*
*       Arguments :
*       V       (i)     1 x N real vector
*       N       (i)     Number of elements in vector
*
*
*       History:
*       Written July 2 1992 S. Knight
************************************************************************
        real function fmaxvc ( v , n )

        implicit none

*       Arguments
        integer         n
        real            v( n )

*       Local variables
        integer         i
*-----------------------------------------------------------------------
        fmaxvc = -1.e+37
        do 10 i = 1 , n
         fmaxvc = max ( fmaxvc , v( i ) )
 10     continue
        return
        end
*********************************************************************sdk
************************************************************************
*
*       Real function FMINVC
*
*       Function:
*       Return value of smallest element in vector (real)
*
*       Arguments :
*       V       (i)     1 x N real vector
*       N       (i)     Number of elements in vector
*
*
*       History:
*       Written July 2 1992 S. Knight
************************************************************************
        real function fminvc ( v , n )

        implicit none

*       Arguments
        integer         n
        real            v( n )

*       Local variables
        integer         i
*-----------------------------------------------------------------------
        fminvc = 1.e+38
        do 10 i = 1 , n
         fminvc = min ( fminvc , v( i ) )
 10     continue
        return
        end
************************************************************************
*
*       Real function RANDOM
*
*       Function:
*       Return random number in interval [0.0,1.0[
*
*       Comments:
*       The code is MACHINE SPECIFIC.
*
*       Arguments:
*       Seed     (i)       seed
*
*       History:
*       Written November 5 1992 S. Knight
************************************************************************
        real function random ( seed )

        implicit none

*       Arguments
        integer          seed

*       Local variables
cpjx        double precision x
        logical  iset
        integer  iseed,iseed2
        real     rvec(1)
        external ranmar

*       Data, save statements
        data     iset/.false./
        save     iset
*       Intrinsic function
cpjx        double precision rand
*-----------------------------------------------------------------------
*       VAX/VMS
*        random = ran ( seed )
* 10     if ( random .eq. 1.0 ) goto 10

*       SGI, ALPHA
cpjx 10     x = rand()
cpjx        if ( x .eq. 1.0 ) goto 10
cpjx        random = x
cpjx Code below uses random number generator in ccp4 library
cpjx which should be portable
C
        if (.not.iset) then
          iseed2 = seed + 12345
          iseed  = mod(seed,31329)
          iseed2 = mod(iseed2,30082)
          call rmarin(iseed,iseed2)
          iset = .true.
        end if
C
 10     call ranmar(rvec,1)
        if (rvec(1).eq.1.0) goto 10
        random = rvec(1)
        return
        end
*********************************************************************SDK
*********************************************************************sdk
************************************************************************
************************************************************************
*********************************************************************sdk
