/*

     This code is distributed under the terms and conditions of the
     CCP4 Program Suite Licence Agreement as a CCP4 Application.
     A copy of the CCP4 licence can be obtained by writing to the
     CCP4 Secretary, Daresbury Laboratory, Warrington WA4 4AD, UK.
*/

#include <stdlib.h>
#include <math.h>
#include <stdio.h>
#include <ccp4_sysdep.h>
#include "matrix.h"
#include "ccp4_general.h"

/******************************************************************************
*     Prototypes not defined in matrix.h                                      *
*******************************************************************************/
#if defined (PROTOTYPE)

  float *vector(int nl, int nh);

  void free_vector(float *v, int nl, int nh);

  int *ivector(int nl, int nh);

  void free_ivector(int *v, int nl, int nh);

  float **matrix(int nrl, int nrh, int ncl, int nch);

  void free_matrix(float **m, int nrl, int nrh, int ncl, int nch);

  void LUdecomp(float **a, int n, int *indx, float *d);

  void LUbacksub(float **a, int n, int *indx, float *b);

  float **invert_matrix(float **a, int n);

#endif   /* (PROTOTYPE)  */



/******************************************************************************/

#if defined (PROTOTYPE)
  float *vector(int nl, int nh)
#else
  float *vector(nl, nh)
  int nl, nh;
#endif
{ float *v;

  v=(float *)malloc((unsigned) (nh-nl+1)*sizeof(float));
  if (!v) 
  { fprintf(stderr, "allocation failure in vector()\n");
    ccperror(1,"*** Program Terminating ***");}
  return v-nl;}


/******************************************************************************/

#if defined (PROTOTYPE)
  void free_vector(float *v, int nl, int nh)
#else
  void free_vector(v, nl, nh)
  float *v;
  int nl, nh;
#endif
{ free((char*) (v+nl));}


/******************************************************************************/

#if defined (PROTOTYPE)
  int *ivector(int nl, int nh)
#else
  int *ivector(nl, nh)
  int nl, nh;
#endif
{ int *v;
  
  v=(int *)malloc((unsigned) (nh-nl+1)*sizeof(int));
  if (!v) 
  { fprintf(stderr, "allocation failure in ivector()\n");
    ccperror(1,"*** Program Terminating ***");}
  return v-nl;}


/******************************************************************************/

#if defined (PROTOTYPE)
  void free_ivector(int *v, int nl, int nh)
#else
  void free_ivector(v, nl, nh)
  float *v;
  int nl, nh;
#endif
{ free((char*) (v+nl));}


/******************************************************************************/

#if defined (PROTOTYPE)
  float **matrix(int nrl, int nrh, int ncl, int nch)
#else
  float **matrix(nrl, nrh, ncl, nch)
  int nrl, nrh, ncl, nch;
#endif
{ int i;
  float **m;
  
  m=(float **) malloc((unsigned) (nrh-nrl+1)*sizeof(float*));
  if (!m) 
  { fprintf(stderr, "allocation failure 1 in matrix()\n");
    ccperror(1,"*** Program Terminating ***");}
  m -= nrl;
  for(i=nrl;i<=nrh;i++)
  { m[i]=(float *) malloc((unsigned) (nch-ncl+1)*sizeof(float));
    if (!m[i]) 
    { fprintf(stderr, "allocation failure 2 in matrix()\n");
      ccperror(1,"*** Program Terminating ***");}
    m[i] -= ncl;}
  return m;}


/******************************************************************************/

#if defined (PROTOTYPE)
  void free_matrix(float **m, int nrl, int nrh, int ncl, int nch)
#else
  void free_matrix(m, nrl, nrh, ncl, nch)
  int nrl, nrh, ncl, nch;
  float **m;
#endif
{ int i;

  for(i=nrh;i>=nrl;i--)
    free((char*) (m[i]+ncl));
  free((char*) (m+nrl));}


/******************************************************************************/

#if defined (PROTOTYPE)
  void LUbacksub(float **a, int n, int *indx, float *b)
#else
  void  LUbacksub(a, n, indx, b)
  float **a, *b;
  int n, *indx;
#endif
{ int i, ii=0, ip, j;
  float sum;
  
  for (i = 1; i <= n; i++)
  { ip = indx[i];
    sum = b[ip];
    b[ip] = b[i];
    if (ii)
      for (j = ii;j <= i - 1; j++)
	sum -=  a[i][j] * b[j];
    else if (sum)
      ii = i;
    b[i] = sum;}
  for (i = n; i >= 1; i--)
  { sum = b[i];
    for (j = i + 1; j <= n; j++) 
      sum -=  a[i][j] * b[j];
    b[i] = sum / a[i][i];}}


/******************************************************************************/

#define TINY 1.0e-20;

#if defined (PROTOTYPE)
  void LUdecomp(float **a, int n, int *indx, float *d)
#else
  void LUdecomp(a, n, indx, d)
  float **a, *d;
  int n, *indx;
#endif
{ int i, imax, j, k;
  float big, dum, sum, temp;
  float *vv;
#if !defined (PROTOTYPE)
  float *vector();
#endif

  vv = vector(1, n);
  *d = 1.0;
  for (i = 1; i <= n; i++)
  { big = 0.0;
    for (j = 1; j <= n; j++)
      if ((temp = fabs(a[i][j])) > big)
	big = temp;
    if (big  ==  0.0)
    { fprintf(stderr, "Singular matrix in routine LUdecomp\n");
      ccperror(1,"*** Program Terminating ***");}
    vv[i] = 1.0 / big;}
  for (j = 1; j <= n; j++)
  { for (i = 1; i < j; i++)
      { sum = a[i][j];
	for (k = 1; k < i; k++)
	  sum -=  a[i][k] * a[k][j];
	a[i][j] = sum;}
    big = 0.0;
    for (i = j; i <= n; i++)
    { sum = a[i][j];
      for (k = 1; k < j; k++)
	sum -=  a[i][k] * a[k][j];
      a[i][j] = sum;
      if ((dum = vv[i] * fabs(sum)) >=  big)
      { big = dum;
	imax = i;}}
    if (j !=  imax)
    { for (k = 1; k <= n; k++)
      { dum = a[imax][k];
	a[imax][k] = a[j][k];
	a[j][k] = dum;}
      *d  =  -(*d);
      vv[imax] = vv[j];}
    indx[j] = imax;
    if (a[j][j]  ==  0.0)
      a[j][j] = TINY;
    if (j !=  n)
    { dum = 1.0 / a[j][j];
      for (i = j + 1; i <= n; i++)
	a[i][j] *= dum;}}
  free_vector(vv,1,n);}


#undef TINY


/******************************************************************************/

#if defined (PROTOTYPE)
  float **invert_matrix(float **a, int n)
#else
  float **invert_matrix(a, n)
  float **a;
  int n;
#endif
/* Returns a pointer to the inverse of matrix 'a[1..n][1..n]'. When the 
   inversed matrix is not used anymore, its location should be freed with
   free_matrix(inv, 1, n, 1, n). */

{ float **inv, *col, d;
  int i, j, *indx;
#if !defined (PROTOTYPE)
  float *vector();
  void LUdecomp();
  float **matrix();
  void LUbacksub();
  void free_vector();
  void free_ivector();
#endif

  inv = matrix(1, n, 1, n);
  col = vector(1, n);
  indx = ivector(1, n);
  LUdecomp(a, n, indx, &d);
  for (j = 1; j <= n; ++j)
  { for (i = 1; i <= n; ++i)
      col[i] = 0;
    col[j] = 1;
    LUbacksub(a, n, indx, col);
    for (i = 1; i <= n; ++i)
      inv[i][j] = col[i];}
  free_vector(col, 1, n);
  free_ivector(indx, 1, n);
  return(inv);}


/******************************************************************************/

