/*
     SOLOMON density modification program
     Copyright (C) 1995 Jan-Pieter Abrahams

     This code is distributed under the terms and conditions of the
     CCP4 Program Suite Licence Agreement as a CCP4 Application.
     A copy of the CCP4 licence can be obtained by writing to the
     CCP4 Secretary, Daresbury Laboratory, Warrington WA4 4AD, UK.
*/

#include <stdio.h>
#include <math.h>
#include <string.h>
#include <stdlib.h>
#include <ccp4_sysdep.h>                    /* CCP4 machine dep. include file */
#include "solomon.h"
#include "matrix.h"
#include "ccp4_unitcell.h"
#include "ccp4_general.h"

BOOLEAN verbose = FALSE;

#if defined (PROTOTYPE)
int main(int argc, char *argv[])
#else
int main(argc, argv)
  int argc;
  char *argv[];
#endif
{
#if defined (PROTOTYPE)
  void solmain(void);
#else
  void solmain();
#endif
  ccp4fyp(argc,argv);
  ccp4_banner();
  solmain();
  ccperror(0," *** Normal Termination ***");
}

void solmain()

{ MAP *map, *tst;
  MASK *mask[1024];
  BOOLEAN average, extrude, keepmean;
  int nmask, i;
  char newmapname[256], rmsmapname[256], mapmin[256], solmaskname[256];
  float sd, mean, radius, patchsd, mintrunc, maxtrunc;
  float solvdens, solvmul, mapminscale, protmul, sdred, validate[2];
  float ptos, sharp;
#if !defined (PROTOTYPE)
  double *NormalDistr();
  int ParseInput();
  void FreeMap();
  void TrimMasks();
  MAP *MapFill();
  void AverageMap();
  MAP *CombMaps();
  MAP *FlatMap();
  void WriteMap();
#endif

  (void) NormalDistr();
  nmask = ParseInput(mask, &map, &average, newmapname, rmsmapname, &radius, 
		     &mintrunc, &maxtrunc, &ptos, &solvdens, &solvmul,
		     &protmul, mapmin, &mapminscale, &sdred, solmaskname,
		     validate, &extrude, &keepmean, &sharp);
  if (average != TRUE)
  { printf("\n************************************************************\n\n");
    printf("   Solomon will trim masks using the cell & grid of %s\n", map->infile);
    printf("\n************************************************************\n\n");
    FreeMap(map);
    if (!VECINMASK(mask[0]->ori, mask[0]))
    { printf("Translating asymmetric unit making its origin all positive\n");
      for (i = 0; i < 3; ++i)
	mask[0]->ori[i] = PMOD(mask[0]->ori[i], mask[0]->grid[i]);}
    TrimMasks(mask, nmask);
    }
  else
  { printf("************************************************************\n\n");
    printf("   Solomon will modify %s\n", map->infile);
    if (newmapname[0] != '\0')
      printf("   Modified map written to MAPOUT. \n\n");
    printf("************************************************************\n\n");
    sd = 0;
    patchsd = 0;
    if (verbose)
      printf("Read %d masks\n", nmask - 1);
    for (i = 1; i < nmask; ++i)
      if (strcmp(mask[i]->title, "monitor") == 0)
      { tst = MapFill(mask[i], map, mask[i]->trans, mask[0]->trans,
		      mask[0]->ntrans);
	printf("Relative Sd of %s: %f\n", mask[i]->infile,
	       (patchsd = SigmaMap(tst, mask[i], &mean)) / map->rms);
	FreeMap(tst);}
    for (i = 1, sd = -1; (i < nmask) && (sd <= 0); ++i)
      if (strcmp(mask[i]->title, "weight") == 0)
      { tst = MapFill(mask[i], map, mask[i]->trans, mask[0]->trans,
		      mask[0]->ntrans);
	patchsd = sd = SigmaMap(tst, mask[i], &mean);
	printf("\nSd of %s: %f\n", mask[i]->infile, sd);
	printf("This will be used as an estimate of the noise of the map.\n");
	printf("    Relative flatness of solvent in the patch: %f\n",
	        sd / map->rms);
	FreeMap(tst);}
    if (sd < 0)
      printf("\nAveraging will not be weighted\n");
    else
      printf("\nAveraging will be weighted\n");
    for (i = 1; i < nmask; ++i)
      AverageMap(map, mask[i], mask[0]->trans, mask[0]->ntrans, sd);
    if ((newmapname[0] != '\0') || (rmsmapname[0] != '\0'))
    { map = CombMaps(map, mask, nmask);
      map = FlatMap(map, map->solv, radius, mintrunc, maxtrunc, mask, nmask,
		    patchsd, ptos, solvdens, solvmul, rmsmapname, protmul,
		    mapmin, mapminscale, sdred, solmaskname, validate, 
		    extrude, keepmean, sharp);
      WriteMap(map, newmapname);}
  }
}



/******************************************************************************/

#if defined (PROTOTYPE)
  void TrimMasks(MASK **mask, int nmask)
#else
  void TrimMasks(mask, nmask)
  MASK **mask;
  int nmask;
#endif
/* Trims overlap between (symmetry related) copies of all masks in 'mask[]'.
   Copies of all specified trimmed masks will be written to disk, together 
   with their symmetry operators. The array 'mask[]' contains 'nmask'
   entries. The first mask of the array 'mask[]' describes the asymmetric
   unit and its origin should be all positive. */
  
{ MASK *unitcell, *tstmsk[2];
  int totmask, ntrim, i, j, k, back, maxtrans;
  int name;
  char ncsname[256], *dot, temp[10];
#if !defined (PROTOTYPE)
  void ShutUp();
  MASK *ShrinkMask();
  MASK *TransMask();
  void FillAsUn();
  void FreeMask();
  MASK *MallocMask();
  void ClearMask();  
  void WriteMask();
  void WriteNcs();
#endif

  ShutUp(TRUE);

  /* Produce all symmetry related masks. */
  for (i = 1, totmask = nmask; i < nmask; ++i)
  { ShrinkMask(mask[i]);
    printf("Transforming %s\n", mask[i]->infile);
    for (j = 0; j < mask[i]->ntrans; ++j, ++totmask)
    { mask[totmask] = TransMask(mask[i], mask[i]->trans + j);
      if (mask[i]->trans[j].outfile[0] != '\0')
	printf("  transform %d produces %s\n", j / mask[0]->ntrans,
	       mask[i]->trans[j].outfile);}}

  /* Now fill the asymmetric unit. */
  mask[totmask] = NULL;
  FillAsUn(mask[0], mask + nmask);
  printf("Asymmetric unit filled!\n");

  /* Trim the masks so that no point in the asymmetric unit is hit twice. */
  for (i = 1, maxtrans = 0; i < nmask; ++i)
    maxtrans = MAX(maxtrans, mask[i]->ntrans);
  for (i = 1; i < nmask; ++i)
  { if (mask[i]->outfile[0] != '\0')
      printf("Mask %s has %d gridpoints set\n", mask[i]->outfile, 
	     mask[i]->nset);
    else
      printf("Mask %s symmetry mate has %d gridpoints set\n",
	     mask[i]->outfile, mask[i]->nset);}
      
  j = 0;
  do
  { ntrim = 0;
    do
      for (i = (j%2) ? 1 : (nmask-1); (j%2) ? (i<nmask) : (i>0);
	                                      i = (j%2)?(i+1):(i-1))
      { k = j % mask[i]->ntrans;
	if (strcmp(mask[i]->title, "solvent patch") != 0)
	  ntrim += TrimMask(mask[i]->trans[k].child, mask[0], 4);}
    while ((++j % maxtrans) != 0);}
  while (ntrim != 0);
  printf("Masks trimmed!!\n");

  for (i = 1; i < nmask; ++i)
  { if (mask[i]->outfile[0] != '\0')
      printf("Mask %s now has %d gridpoints set\n", mask[i]->outfile, 
	     mask[i]->nset);
    else
      printf("Mask %s symmetry mate now has %d gridpoints set\n",
	     mask[i]->outfile, mask[i]->nset);}
 
  /* Fold the masks back on the original ones */
  unitcell = MallocMask(mask[0]->grid, mask[0]->ori, mask[0]->grid, mask[0]->cell);
  for (i = 1, back = nmask; i < nmask; ++i)
  { ClearMask(unitcell);
    for (j = 0; j < mask[i]->ntrans; ++j, ++back)
    { tstmsk[0] = TransMask(mask[back], mask[back]->trans);
      tstmsk[1] = NULL;
      if (mask[back]->outfile[0] != '\0')
      { name = back;
        printf("Folding back %s\n", mask[back]->outfile);}
      else
	printf("Folding back crystallographically related copy of %s\n",
	       mask[name]->outfile);
      FillAsUn(unitcell, tstmsk);
      FreeMask(tstmsk[0]);}
    TestMask(mask[i], unitcell, mask[i]->ntrans);}

  for (i = 1; i < nmask; ++i)
  { if (mask[i]->outfile[0] != '\0')
      printf("Mask %s now has %d gridpoints set\n", mask[i]->outfile, 
	     mask[i]->nset);
    else
      printf("Mask %s symmetry mate now has %d gridpoints set\n",
	     mask[i]->outfile, mask[i]->nset);}

  /* Redistribute masks */
  for (i = nmask; i < totmask; ++i)
    FreeMask(mask[i]);
  ShutUp(FALSE);
  for (i = 1, totmask = nmask; i < nmask; ++i)
  { ShrinkMask(mask[i]);
    printf("Transforming %s\n", mask[i]->infile);
    for (j = 0; j < mask[i]->ntrans; ++j, ++totmask)
    { mask[totmask] = TransMask(mask[i], mask[i]->trans + j);
      if (mask[i]->trans[j].outfile[0] != '\0')
	printf("  transform %d produces %s\n", j / mask[0]->ntrans,
	       mask[i]->trans[j].outfile);}}
  mask[totmask] = NULL;
  ClearMask(mask[0]);
  FillAsUn(mask[0], mask + nmask);

  /* Write out the required masks */
  if (mask[0]->outfile[0] != '\0')
    WriteMask(mask[0]);
  for (i = 1; i < nmask; ++i)
  { strcpy(mask[i]->outfile, mask[i]->trans[0].outfile);
    WriteMask(mask[i]);}
  for (i = nmask; i < totmask; ++i)
  { if (mask[i]->outfile[0] != '\0')
    { WriteMask(mask[i]);
      for (j = 0; j < mask[i]->ntrans; j += mask[0]->ntrans)
      { strcpy(ncsname, mask[i]->outfile);
	if ((dot = strrchr(ncsname, '.')) != NULL)
	  dot[0] = '\0';
	strcat(ncsname, ".ncs");
	sprintf(temp, "%d", j / mask[0]->ntrans);
	strcat(ncsname, temp);
	WriteNcs(mask[i]->trans + j, ncsname);}}}}



/******************************************************************************/

#if defined (PROTOTYPE)
  void FillAsUn(MASK *as_mask, MASK **mask)
#else
  void  FillAsUn(as_mask, mask)
  MASK *as_mask;
  MASK **mask;
#endif
/* Loops through all masks in 'mask', and if it finds a point which is set,
   it will increment the corresponding gridpoint in the mask of the 
   asymmetric unit, defined by 'as_mask'. The array containing the masks should
   be NULL terminated. */

{ int *asymu, *ori, *grid, *ext, i, i0, i1, i2, asym0, asym1, asym2;
  Byte ***mskdata, ***as_mskdata;
  
  asymu = as_mask->ext;
  grid = as_mask->grid;
  as_mskdata = as_mask->msk;
  for (i = 0; mask[i] != NULL; ++i)
  { ori = mask[i]->ori;
    ext = mask[i]->ext;
    mskdata = mask[i]->msk;
    for (i0 = 0; i0 < ext[0]; ++i0)
    { asym0 = PMOD(ori[0] + i0, grid[0]);
      asym0 = (asym0 < 0) ? (asym0 + grid[0]) : asym0;
      if (asym0 < asymu[0])
      { for (i1 = 0; i1 < ext[1]; ++i1)
	{ asym1 = PMOD(ori[1] + i1, grid[1]);
	  asym1 = (asym1 < 0) ? (asym1 + grid[1]) : asym1;
	  if (asym1 < asymu[1])
	  { for (i2 = 0; i2 < ext[2]; ++i2)
	    { asym2 = PMOD(ori[2] + i2, grid[2]);
	      asym2 = (asym2 < 0) ? (asym2 + grid[2]) : asym2;
	      if (asym2 < asymu[2])
		if ((as_mskdata[asym0][asym1][asym2] += mskdata[i0][i1][i2]) == 1)
		  ++as_mask->nset;}}}}}}}



/******************************************************************************/

#if defined (PROTOTYPE)
  int TrimMask(MASK *mask, MASK *as_mask, int nset)
#else
  int TrimMask(mask, as_mask, nset)
  MASK *mask, *as_mask;
  int nset;
#endif
/* Checks all the set gridpoints in 'mask' which are on the border of the set
   area and resets them if corresponding gridpoints in the asymmetric unit
   ('as_mask') are set more than once. The corresponding gridpoint in the
   asymmetric unit is updated. A point is considered to be at the border of the
   set area of the mask if the number of unset neighbour points is equal or
   larger than 'nset'. Returns the number of points which have been reset.*/

{ int *asymu, *ori, *grid, ext[3], i0, i1, i2, asym0, asym1, asym2;
  int ntrim = 0, j0, j1, j2, nempty;
  Byte ***mskdata, ***as_mskdata;

  asymu = as_mask->ext;
  grid = as_mask->grid;
  as_mskdata = as_mask->msk;
  ori = mask->ori;
  mskdata = mask->msk;
  for (i0 = 0; i0 < 3; ++i0)
    ext[i0] = mask->ext[i0] - 1;
  for (i0 = 1; i0 < ext[0]; ++i0)
  { asym0 = PMOD(ori[0] + i0, grid[0]);
    asym0 = (asym0 < 0) ? (asym0 + grid[0]) : asym0;
    if (asym0 < asymu[0])
    { for (i1 = 1; i1 < ext[1]; ++i1)
      { asym1 = PMOD(ori[1] + i1, grid[1]);
	asym1 = (asym1 < 0) ? (asym1 + grid[1]) : asym1;
	if (asym1 < asymu[1])
        { for (i2 = 1; i2 < ext[2]; ++i2)
          { asym2 = PMOD(ori[2] + i2, grid[2]);
	    asym2 = (asym2 < 0) ? (asym2 + grid[2]) : asym2;
	    if (asym2 < asymu[2])
	      if ((mskdata[i0][i1][i2] == 1) && (as_mskdata[asym0][asym1][asym2] > 1))
	      { for (j0 = i0 - 1, nempty = 0; j0 < (i0 + 2); ++j0)
		  for (j1 = i1 - 1; j1 < (i1 + 2); ++j1)
		    for (j2 = i2 - 1; j2 < (i2 + 2); ++j2)
		      nempty += (mskdata[j0][j1][j2] == 0) ? 1 : 0;
		if (nempty >= nset)
		{ --as_mskdata[asym0][asym1][asym2];
		  --mskdata[i0][i1][i2];
		  ++ntrim;}}}}}}}
  mask->nset -= ntrim;
  return(ntrim);}
  


/******************************************************************************/

#if defined (PROTOTYPE)
  void AddTrans(MASK *mask, TRANSFORM *trans, char *outfile)
#else
  void  AddTrans(mask, trans, outfile)
  MASK *mask;
  TRANSFORM *trans;
  char *outfile;
#endif
/* Adds a new transformation to the specified mask. */

{ int i;

  if (mask->trans == NULL)
    mask->trans = (TRANSFORM *) malloc(sizeof(TRANSFORM));
  else
    mask->trans = (TRANSFORM *) realloc(mask->trans, 
					(mask->ntrans + 1) * sizeof(TRANSFORM));
  if (mask->trans == NULL)
  { printf("Not enough space to add transforms to masks!!");
    ccperror(1,"*** Program Terminating ***");}
  else
  { for (i = 0; i < 9; ++i)
      mask->trans[mask->ntrans].rotate[i] = trans->rotate[i];
    for (i = 0; i < 3; ++i)
      mask->trans[mask->ntrans].translate[i] = trans->translate[i];
    mask->trans[mask->ntrans].child = NULL;
    if (outfile != NULL)
      strcpy(mask->trans[mask->ntrans].outfile, outfile);
    else
      mask->trans[mask->ntrans].outfile[0] = '\0';
    ++mask->ntrans;}}
    

/******************************************************************************/

#if defined (PROTOTYPE)
  char *ReadLine(FILE *file, char *key, char *to, BOOLEAN begin)
#else
  char *ReadLine(file, key, to, begin)
  FILE *file;
  char *key, *to;
  BOOLEAN begin;
#endif
/* Returns the next line from `file` which starts with `key`, stripped of
   the first token. In identifying this line, no distinction is made between
   upper and lower case. The search stops if EOF is hit, or (if `to != NULL`)
   a line starting with any of the keys, separated by spaces, which are
   defined in `to` is reached. If the search stops without finding a line 
   starting with `key`, NULL is returned. The file is rewound before parsing if
   `begin == TRUE`. If `key == NULL` no subsequent parsing will be undertaken. 
   The file is read in on the first call with a non-NULL `file`
   and subsequent calls with `file == NULL` access the internal buffer.

   Since everything is stored internally, this function also works for stdin.
   The buffers are not of infinite size however, so do not parse files larger
   than 10 kByte, files which contain more than 1024 lines or files with lines
   containing more than 1024 characters. */

{ static char buffer[1024*10];
  static char *string[1024];
  static char line[1024];
  static int index = 0, nlines = 0;
  char *tokeys[100];
  int i, j;
  char charbuff[1024], *rv;

  /* Check if the file is read into the internal buffer: */
  if (file != NULL)
  { string[nlines = 0] = buffer;
    for ( ; fgets(string[nlines], 1024, file) != NULL; ++nlines)
    { string[nlines + 1] = strchr(string[nlines], 0) + 1;
      if ((nlines == 1022) || ((strchr(string[nlines], 0) - buffer) > 9216))
      { fprintf(stderr, "Internal buffer of ReadLine too small!!!\n");
        ccperror(1,"*** Program Terminating ***");}}
    index = 0;}

  if (begin == TRUE)
    index = 0;
  if (key == NULL)
    return(NULL);
  if (to != NULL)
  { strcpy(charbuff, to);
    tokeys[0] = strtok(charbuff, " ");
    for (i = 1; (i < 100) && ((tokeys[i] = strtok(NULL, " ")) != NULL); ++i);
    if (tokeys[i] != NULL)
    { fprintf(stderr, "Too many keys for ReadLine!!!\n");
      ccperror(1,"*** Program Terminating ***");}}
  else
    tokeys[0] = NULL;

  /* Look for the key and return the corresponding line: */
  for ( ;index < nlines; ++index)
  { if (strncasecmp(string[index], key, 4) == 0)
    { for (i = j = 0; i < 1023; ++i)
        if (!((string[index][i] == ' ') && (string[index][i + 1] == ' ')))
	  if ((line[j++] = string[index][i]) == '\0')
	    i = 1024;
      
      ++index;
      rv = strtok(line, " ");
      rv = strtok(NULL, "\n");
      if (rv == NULL)
      { printf("Data line--- %s\n", key);
	return("");}
      printf("Data line--- %s %s\n", key, rv);
      return(rv);}
    for (i = 0; tokeys[i] != NULL; ++i)
      if (strncasecmp(string[index], tokeys[i], 4) == 0)
        return(NULL);}
  printf(" *** %s   not found, defaults employed\n", key);
  return(NULL);}


/******************************************************************************/

#if defined (PROTOTYPE)
  int ParseInput(MASK **mask, MAP **map, BOOLEAN *average, char *newmapname,
	       char *rmsmapname, float *radius, float *mintrunc, float *maxtrunc,
	       float *ptos, float *solvdens, float *solvmul, float *protmul,
	       char *mapmin, float *mapminscale, float *sdred, char *solmaskname,
	       float *validate, BOOLEAN *extrude, BOOLEAN *keepmean, float *sharp)
#else
  int  ParseInput(mask, map, average, newmapname, rmsmapname, radius, mintrunc,
               maxtrunc, ptos, solvdens, solvmul, protmul, mapmin, mapminscale,
               sdred, solmaskname, validate, extrude, keepmean, sharp)
  MASK **mask;
  MAP **map;
  BOOLEAN *average, *extrude, *keepmean;
  char *newmapname, *rmsmapname;
  float *radius, *mintrunc, *maxtrunc, *ptos, *solvdens, *solvmul, *protmul;
  float *mapminscale, *sdred, *validate, *sharp;
  char *mapmin, *solmaskname;
#endif
/* Reads the masks as specified by the commands in 'stdin'. The format is:
   (not necessarily in this order)
   ...
   ...
   verbose if more output is desired
   slvfrc  (fraction of the map to be flattened)
   trunc   (followed by one or two numbers: minimum & maximum percentage of
            protein to be truncated by solvent flattening)
   ptos     (fractional of protein versus solvent density, default: 1.31)
   extrude  (If present, the mean protein to solvent ratio will be adjusted
             to "ptos" by adding a constant to the solvent fraction, assuming
	     that the minimum in the protein region represents zero density
	     after adding in the F000);
   slvdens  (mean electron density of the solvent, used as mean relative to
             which the local standard deviation is determined, default: 0)
   mulsolv (multiplier of solvent density, or "auto", followed by required
            modification of the rms of the solvent with respect to the rms 
            of the protein)
   sharp   (exponent by which the protein density should be sharpened. Default
            is 0).
   mulprot (multiplier of the protein)
   mapmin  (filename of map to be subtracted from protein area, the map will be 
           scaled by the factor following the number, default 1. If followed by
	   ")
   radius  (radius for determining solvent)
   mapout  (Averaged map will be saved as specified, assigned to MAPOUT)
   rmsmap  (Local rmsmap will be saved as specified, assign to RMSMAP)
   asmask  (filename of mask defining which part of the asymmetric unit 
            is covered by other masks).
   csym    (filename of O-style crystallographic symmetry operators)
   sdred   (value defining the reduction of the standard deviation of the density
            of the protein chain. If included, excursions into the solvent by 
	    density connected to high protein denisty will be romoved from the
	    solvent mask)
   slvmask  filename (Name of the mask file to which the solvent mask is to be
            written)
   ...
   ...
   patch   radius   (Spherical mask with given radius, centered on the
           gridpoint with the lowest local rms)
   mskout  filename (name of the patch)
   mskin   (filename of O-style mask, or "unit.msk" (mask of unit cell) if such
            a file does not exist)
   monitor (if present, calculates sd of mask before averaging)
   weight  (if present, uses sd within mask to determine sd of noise)
   insert  (if present, inserts corresponding map without averaging or flattening)
   solvent (if present, treats the named mask as solvent & flattens it)
   trans   (filename of O-style non-crystallographic symmetry ops)
   ...     (Repeat 'trans (filename)' as often as needed for this mask. Make
            sure that the identity operator is included. Every file defines
	    one symmetry operator.)
   mskout  (filename of altered mask)
   mskin   (other mask)
   trans   (with its own non-crystallographic symmetry operators)
   ...
   mskout  (filename of altered mask)
   ...
   EOF   (Repeat 'mskin (filename)\n trans (filename)\n trans (filename) ...'
          as often as required)

   Lines starting with unrecognized keywords are ignored.  The number of masks
   read in is returned. The first mask contains the asymmetric unit.
   If at least one "newmask" keyword was encountered, "average" will be set to
   'FALSE', otherwise it will be set to 'TRUE'. */

{ TRANSFORM trans;
  char buffer[1024], *line;
  double tot;
  int i, j, nmask;
  float patchrad;
  BOOLEAN lineread;
  FILE *file;
#if !defined (PROTOTYPE)
  char *ReadLine();
  MAP *ReadMap();
  MASK *MallocMask();
  void AddTrans();
  MASK *ReadMask();
  void ReadTrans();
  MASK *MinPatch();
  void WriteMask();
#endif

  *average = TRUE;
  mapmin[0] = '\0';

  /* Get the asymmetric unit. */
  line = ReadLine(stdin, "verbose", NULL, TRUE);
  if (line != NULL) 
    verbose = TRUE;
  /* Validate undocumented */
  validate[0] = validate[1] = 0;
  strcpy(buffer, "MAPIN");
  *map = ReadMap(buffer);
  mask[0] = MallocMask((*map)->grid, (*map)->ori, (*map)->ext, (*map)->cell);
  line = ReadLine(NULL, "asmask", NULL, TRUE);
  if (line != NULL)
    sscanf(line,"%s", mask[0]->outfile);
  line = ReadLine(NULL, "slvfrc", NULL, TRUE);
  map[0]->solv = -1;
  if (line != NULL)
    sscanf(line, "%f", &map[0]->solv);
  if (map[0]->solv < 0) 
  { printf ("Failed to provide solvent content!!\n");
    ccperror(1,"*** Program Terminating ***");}
  if (map[0]->solv > 1)
  { printf("Solvent fraction larger than 1, assuming %f percent solvent\n",
    map[0]->solv);
    map[0]->solv /= 100;}
  line = ReadLine(NULL, "radius", NULL, TRUE);
  if (line == NULL)
    *radius = 3.5;
  else if (sscanf(line, "%f", radius) == 0)
    *radius = 3.5;
  line = ReadLine(NULL, "sdred", NULL, TRUE);
  if (line == NULL)
    *sdred = -1.0;
  else if (sscanf(line, "%f", sdred) == 0)
    *sdred = 1.0;
  line = ReadLine(NULL, "ptos", NULL, TRUE);
  *ptos = 1.31;
  if (line != NULL)
    sscanf(line, "%f", ptos);
  line = ReadLine(NULL, "sharp", NULL, TRUE);
  *sharp = 0;
  if (line != NULL)
    sscanf(line, "%f", sharp);
  if (ReadLine(NULL, "extrude", NULL, TRUE) == NULL)
    *extrude = FALSE;
  else
    *extrude = TRUE;
  if (ReadLine(NULL, "keepmean", NULL, TRUE) == NULL)
    *keepmean = FALSE;
  else
    *keepmean = TRUE;
  line = ReadLine(NULL, "slvdens", NULL, TRUE);
  *solvdens = 0.0;
  if (line != NULL)
    sscanf(line, "%f", solvdens);
   line = ReadLine(NULL, "mulsolv", NULL, TRUE);
  *solvmul = -0.75;
  if (line != NULL)
  { if (strncasecmp(line, "auto", 4) == 0)
    { printf("Autoscaling solvent with respect to change in protein rms\n");
      *solvmul = -1.00;
      sscanf(line, "auto %f", solvmul);
      *solvmul += 1000;}
    else
    { *solvmul = -0.75;
       sscanf(line, "%f", solvmul);}}
  line = ReadLine(NULL, "trunc", NULL, TRUE);
  *mintrunc = 0.35;
  *maxtrunc = 1;
  if (line != NULL)
  { if (strncasecmp(line, "auto", 4) == 0)
    { if (verbose)
      { printf("Determining the F000 term automatically. \n");
        printf("  Feed back Fc and PHIc of missing data, include as much\n");
	printf("  low resolution data as possible!!\n");}
      sscanf(line, "auto %f", maxtrunc);
      *mintrunc = -2;}
    else
      sscanf(line, "%f %f", mintrunc, maxtrunc);}
  line = ReadLine(NULL, "mulprot", NULL, TRUE);
  *protmul = 1.0;
  if (line != NULL)
    sscanf(line, "%f", protmul);
  line = ReadLine(NULL, "mapmin", NULL, TRUE);
  *mapminscale = 1.0;
  if (line != NULL)
  { strcpy(mapmin, "MAPMIN");
    /*  Will autoscale when no scale given with mapmin keyword.
        Not sure if this is correct feature undocumented       */
    if (-1 == sscanf(line, "%f", mapminscale))
      *mapminscale = HUGE;}
  if (ReadLine(NULL, "mapout", NULL, TRUE) == NULL)
    newmapname[0] = '\0';
  else
    strcpy(newmapname, "MAPOUT");
  line = ReadLine(NULL, "slvmask", NULL, TRUE);
  if (line != NULL)
    sscanf(line, "%s", solmaskname);
  else
    solmaskname[0] = '\0';
  if (ReadLine(NULL, "rmsmap", NULL, TRUE) == NULL)
    rmsmapname[0] = '\0';
  else
    strcpy(rmsmapname, "RMSMAP");
  /* Get the crystallographic symmetry operators. */
  line = ReadLine(NULL, "csym", NULL, TRUE);
  if (line == NULL)
  { if (verbose)
    { printf("No crystallographic symmetry operators given!!! (keyword csym)\n");
      printf("Using the operators from %s\n\n", map[0]->infile);}
    { for (i = 0; i < (map[0]->nsyms * 16); i += 16)
      { trans.rotate[0] = map[0]->sym[i + 0];
	trans.rotate[1] = map[0]->sym[i + 4];
	trans.rotate[2] = map[0]->sym[i + 8];
	trans.rotate[3] = map[0]->sym[i + 1];
	trans.rotate[4] = map[0]->sym[i + 5];
	trans.rotate[5] = map[0]->sym[i + 9];
	trans.rotate[6] = map[0]->sym[i + 2];
	trans.rotate[7] = map[0]->sym[i + 6];
	trans.rotate[8] = map[0]->sym[i + 10];
	trans.translate[0] = map[0]->cell[0] * map[0]->sym[i + 3];
	trans.translate[1] = map[0]->cell[1] * map[0]->sym[i + 7];
	trans.translate[2] = map[0]->cell[2] * map[0]->sym[i + 11];
	AddTrans(mask[0], &trans, NULL);}}}
  else
  { sscanf(line, "%s", buffer);
    file = fopen(buffer, "r");
    if (file == NULL)
    { printf("Could not find %s\n", buffer);
      ccperror(1,"*** Program Terminating ***");}
    fgets(buffer, 1024, file);
    while (fgets(buffer, 1024, file) != NULL)
    { for (i = 0; i < 9; i += 3)
      { sscanf(buffer, " %lf %lf %lf", trans.rotate + i, trans.rotate + i + 1,
	                               trans.rotate + i + 2);
	fgets(buffer, 1024, file);}
      sscanf(buffer, " %lf %lf %lf", trans.translate + 0, trans.translate + 1, 
	                             trans.translate + 2);
      for (j = 0; j < 3; ++j)
	trans.translate[j] *= mask[0]->cell[j];
      AddTrans(mask[0], &trans, NULL);}
    fclose(file);}

  /* Get the masks. */
  (void) ReadLine(NULL, NULL, NULL, TRUE);
  for (nmask = 1; (line = ReadLine(NULL, "mskin", NULL, FALSE)) != NULL; ++nmask)
  { if (strncmp(line, "unit.msk", 8) == 0)
    { if ((file = fopen("unit.msk", "r")) == NULL)
      { mask[nmask] = MallocMask((*map)->grid, (*map)->ori, (*map)->ext, (*map)->cell);
	mask[nmask]->nset = map[0]->ext[0] * map[0]->ext[1] * map[0]->ext[2];
	for (i = 0; i < mask[nmask]->nset; ++i)
	  mask[nmask]->msk[0][0][i] = 1;}
      else
      { fclose(file);
	mask[nmask] = ReadMask(line);}}
    else
      mask[nmask] = ReadMask(line);

    /* Check for inconsistencies. */
    for (i = 0, tot = 0; i < 5; ++i)
      tot += ABS(mask[0]->cell[i] - mask[nmask]->cell[i]);
    if (tot > 0.6)
    { printf("Mask %s has an inconsistent unit cell!!\n", line);
      ccperror(1,"*** Program Terminating ***");}
    for (i = 0; i < 3; ++i)
      if (mask[nmask]->grid[i] != mask[0]->grid[i])
      { printf("Inconsistent gridsize, check %s!\n", mask[nmask]->infile);
	ccperror(1,"*** Program Terminating ***");}

  /* Get their transforms, check if they are to be monitored or used for
     weighting, are to be flattened or just inserted. */
    for (lineread = TRUE; lineread == TRUE; )
    { lineread = FALSE;
      while ((line = ReadLine(NULL, "trans",
		 "patch mskin mskout monitor weight solvent insert", FALSE)) != NULL)
      { lineread = TRUE;
	sscanf(line, "%s", buffer);
	line = ReadLine(NULL, "mskout",
			"patch mskin trans monitor weight solvent insert", FALSE);
	ReadTrans(buffer, mask[nmask], mask[0], line);
	if (line != NULL)
	  *average = FALSE;}
      while ((line = ReadLine(NULL, "monitor",
		  "patch mskin mskout trans weight solvent insert", FALSE)) != NULL)
      { lineread = TRUE;
	strcpy(mask[nmask]->title, "monitor");}
      while ((line = ReadLine(NULL, "solvent",
		  "patch mskin mskout trans weight monitor insert", FALSE)) != NULL)
      { lineread = TRUE;
	strcpy(mask[nmask]->title, "solvent");}
      while ((line = ReadLine(NULL, "insert",
		  "patch mskin mskout trans weight solvent monitor", FALSE)) != NULL)
      { lineread = TRUE;
	strcpy(mask[nmask]->title, "insert");}
      while ((line = ReadLine(NULL, "weight",
		  "patch mskin mskout trans monitor solvent insert", FALSE)) != NULL)
      { lineread = TRUE;
	strcpy(mask[nmask]->title, "weight");}}}

  /* Get the patches. */
  (void) ReadLine(NULL, NULL, NULL, TRUE);
  for ( ; (line = ReadLine(NULL, "patch", NULL, FALSE)) != NULL; ++nmask)
  { if (sscanf(line, "%f", &patchrad) == 0)
    { printf ("Failed to provide radius of patch!!\n");
      ccperror(1,"*** Program Terminating ***");}
    printf("Determining a spherical patch with lowest variation and radius %f\n",
	   patchrad);
    mask[nmask] = MinPatch(*map, mask[0]->trans, mask[0]->ntrans, patchrad);
    AddTrans(mask[nmask], mask[0]->trans, NULL);
    strcpy(mask[nmask]->title, "solvent patch");
    
  /* Get their new names, check if they are to be monitored or used for
     weighting. */
    for (lineread = TRUE; lineread == TRUE; )
    { lineread = FALSE;
      while ((line = ReadLine(NULL, "mskout",
			      "patch mskin monitor weight", FALSE)) != NULL)
      { *average = FALSE;
	sscanf(line, "%s", mask[nmask]->trans[0].outfile);
	sscanf(line, "%s", mask[nmask]->outfile);
	if (strncmp(mask[nmask]->title, "solvent patch", 13) == 0)
	  WriteMask(mask[nmask]);}
      while ((line = ReadLine(NULL, "monitor",
			      "patch mskin mskout trans weight", FALSE)) != NULL)
      { lineread = TRUE;
	strcpy(mask[nmask]->title, "monitor");}
      while ((line = ReadLine(NULL, "weight",
			      "patch mskin mskout trans monitor", FALSE)) != NULL)
      { lineread = TRUE;
	strcpy(mask[nmask]->title, "weight");}}}
	
  return(nmask);}



/******************************************************************************/

#if defined (PROTOTYPE)
  void TransAToGrid(TRANSFORM *ang, TRANSFORM *g, float *cell, int *grid)
#else
  void TransAToGrid(ang, g, cell, grid)
  TRANSFORM *ang, *g; 
  float *cell;
  int *grid;
#endif
/* Changes a translation defined in Angstroms ('ang') into one defined
   in gridpoints ('g'). */

{ int i;
  
  for (i = 0; i < 9; ++i)
    g->rotate[i] = ang->rotate[i];
  for (i = 0; i < 3; ++i)
    g->translate[i] = (ang->translate[i] / cell[i]) * grid[i];}



/******************************************************************************/

#if defined (PROTOTYPE)
  void ReadTrans(char *filename, MASK *mask, MASK *asymu, char *outfile)
#else
   void ReadTrans(filename, mask, asymu, outfile)
   char *filename, *outfile;
   MASK *mask, *asymu;
#endif
/* Reads non-crystallographic O-style transformation from 'filename', combines
   it with the crystallographic transformations, and includes the resulting 
   transformations in 'mask'. The mask 'asymu' contains the asymmetric unit
   and the crystallographic symmetry operations. Only the first newly generated 
   transformation will have its '.outfile' set to 'outfile'. */

{ TRANSFORM trans, comb;
  FILE *file;
  char buffer[1024];
  int i, j;
#if !defined (PROTOTYPE)
  void Mat3Mul();
  void Vec3Rot();
  void AddTrans();
#endif

  file = fopen(filename, "r");
  if (file == NULL)
  { printf("Cannot open file %s\n", filename);
    ccperror(1,"*** Program Terminating ***");}

/* If the user has deleted the first line from an "O" ncs file, this
   will ensure proper continuation: */
  if (sscanf(fgets(buffer, 1024, file), " %lf %lf %lf",
         trans.rotate + 0, trans.rotate + 1, trans.rotate + 2) != 3)
    sscanf(fgets(buffer, 1024, file), " %lf %lf %lf",
           trans.rotate + 0, trans.rotate + 1, trans.rotate + 2);
  else
    printf("%s is not an O-style ncs file starting with a format line\n",
        filename);

  i=0;
  j = sscanf(fgets(buffer, 1024, file), " %lf %lf %lf",
	 trans.rotate + 3, trans.rotate + 4, trans.rotate + 5);
  i = i+j;
  j = sscanf(fgets(buffer, 1024, file), " %lf %lf %lf",
	 trans.rotate + 6, trans.rotate + 7, trans.rotate + 8);
  i = i+j;
  j = sscanf(fgets(buffer, 1024, file), " %lf %lf %lf",
	 trans.translate + 0, trans.translate + 1, trans.translate + 2);
  i = i+j;
  if (i != 9) {
    printf("The transformation file %s has an incorrect format\n",filename);
    ccperror(1,"*** Program Terminating ***");
  }

  for (i = 0; i < asymu->ntrans; ++i)
  { Mat3Mul(comb.rotate, asymu->trans[i].rotate, trans.rotate);
    Vec3Rot(comb.translate, asymu->trans[i].rotate, trans.translate);
    for (j = 0; j < 3; ++j)
      comb.translate[j] += asymu->trans[i].translate[j];
    if (i == 0)
      AddTrans(mask, &comb, outfile);
    else
      AddTrans(mask, &comb, "");}
  fclose(file);}



/******************************************************************************/

#if defined (PROTOTYPE)
  void Mat3Mul(double *r, double *m1, double *m2)
#else
  void Mat3Mul(r, m1, m2)
  double *r, *m1, *m2;
#endif
/* Computes the product 'm1 x m2' of 3x3 matrices 'm1' and 'm2'. The result is 
   returned in 'r'. */

/*{ r[0] = m1[0] * m2[0] + m1[1] * m2[3] + m1[2] * m2[6];
  r[1] = m1[0] * m2[1] + m1[1] * m2[4] + m1[2] * m2[7];
  r[2] = m1[0] * m2[2] + m1[1] * m2[5] + m1[2] * m2[8];
  r[3] = m1[3] * m2[0] + m1[4] * m2[3] + m1[5] * m2[6];
  r[4] = m1[3] * m2[1] + m1[4] * m2[4] + m1[5] * m2[7];
  r[5] = m1[3] * m2[2] + m1[4] * m2[5] + m1[5] * m2[8];
  r[6] = m1[6] * m2[0] + m1[7] * m2[3] + m1[8] * m2[6];
  r[7] = m1[6] * m2[1] + m1[7] * m2[4] + m1[8] * m2[7];
  r[8] = m1[6] * m2[2] + m1[7] * m2[5] + m1[8] * m2[8];}*/
{ r[0] = m1[0] * m2[0] + m1[3] * m2[1] + m1[6] * m2[2];
  r[1] = m1[1] * m2[0] + m1[4] * m2[1] + m1[7] * m2[2];
  r[2] = m1[2] * m2[0] + m1[5] * m2[1] + m1[8] * m2[2];
  r[3] = m1[0] * m2[3] + m1[3] * m2[4] + m1[6] * m2[5];
  r[4] = m1[1] * m2[3] + m1[4] * m2[4] + m1[7] * m2[5];
  r[5] = m1[2] * m2[3] + m1[5] * m2[4] + m1[8] * m2[5];
  r[6] = m1[0] * m2[6] + m1[3] * m2[7] + m1[6] * m2[8];
  r[7] = m1[1] * m2[6] + m1[4] * m2[7] + m1[7] * m2[8];
  r[8] = m1[2] * m2[6] + m1[5] * m2[7] + m1[8] * m2[8];}



/******************************************************************************/

#if defined (PROTOTYPE)
  void Vec3Rot(double *r, double *m, double *v)
#else
  void Vec3Rot(r, m, v)
  double *r, *m, *v;
#endif
/* Multiplies vector 'v' by matrix 'm'. The result is returned in 'r'. */

{ r[0] = m[0] * v[0] + m[3] * v[1] + m[6] * v[2];
  r[1] = m[1] * v[0] + m[4] * v[1] + m[7] * v[2];
  r[2] = m[2] * v[0] + m[5] * v[1] + m[8] * v[2];}
/*{ r[0] = m[0] * v[0] + m[1] * v[1] + m[2] * v[2];
  r[1] = m[3] * v[0] + m[4] * v[1] + m[5] * v[2];
  r[2] = m[6] * v[0] + m[7] * v[1] + m[8] * v[2];}*/



/******************************************************************************/

#if defined (PROTOTYPE)
  void WriteNcs(TRANSFORM *trans, char *filename)
#else
  void WriteNcs(trans, filename)
  TRANSFORM *trans;
  char *filename;
#endif
/* Writes transformation operation as 'filename' */

{ FILE *ncs;
  int j;

  ncs = fopen(filename, "w");
  fprintf(ncs, ".LSQ_RT_%s R 12 (3f15.8)\n", filename);
  for (j = 0; j < 9; j += 3)
    fprintf(ncs, "%15.8f%15.8f%15.8f\n", trans->rotate[j], 
	     trans->rotate[j + 1], trans->rotate[j + 2]); 
  fprintf(ncs, "%15.8f%15.8f%15.8f\n", trans->translate[0], 
	   trans->translate[1], trans->translate[2]);
  fclose(ncs);}



 
/******************************************************************************/

#if defined (PROTOTYPE)
  MASK *TransMask(MASK *mask, TRANSFORM *trans)
#else
  MASK  *TransMask(mask, trans)
  MASK *mask;
  TRANSFORM *trans;
#endif
/* Generates a new mask by transforming a specified mask and returns a pointer
   to it. Any transformations defined for the old mask are copied into the new
   mask, after applying the inverse of the transform to these transformations
   too. This means that application of any of the original transforms to the 
   original mask is equivalent to applying any of the new corresponding
   transforms to the new mask!! */

{ int i, i0, i1, i2, j0, j1, j2, ori[3], maxv[3], ext[3], grid[3];
  double v[3], ncheck, dd, dist;
  MASK *newmask;
  TRANSFORM backtrans, trf;
#if !defined (PROTOTYPE)
  void TransVector();
  MASK *MallocMask();
  void InvertTrans();
  MASK *ShrinkMask();
  void Mat3Mul();
  void Vec3Rot();
  void AddTrans();
#endif

  /* Make some space for the new mask... */
  for (i = 0; i < 3; ++i)
    maxv[i] = -(ori[i] = 99999);
  for (i0 = 0; i0 < 2; ++i0)
    for(i1 = 0; i1 < 2; ++i1)
      for (i2 = 0; i2 < 2; ++i2)
      { v[0] = i0 * mask->ext[0] + mask->ori[0];
	v[1] = i1 * mask->ext[1] + mask->ori[1];
	v[2] = i2 * mask->ext[2] + mask->ori[2];
	TransVector(v, trans, mask->cell, mask->grid);
	for (i = 0; i < 3; ++i)
	{ maxv[i] = MAX(maxv[i], (int) ceil(v[i]));
	  ori[i] = MIN(ori[i], (int) floor(v[i]));}}
  for (i = 0; i < 3; ++i)
  { --ori[i];
    ++maxv[i];
    ext[i] = maxv[i] - ori[i];}
  newmask = MallocMask(mask->grid, ori, ext, mask->cell);
  strcpy(newmask->outfile, trans->outfile);

  /* Set those gridpoints in the new mask which correspond with set gridpoints
     of the old mask. This is done by transforming them back to the old mask using
     the inverse of the specified transform, and checking if the majority of
     gridpoints within a sphere of 0.87 from the back-transform of the new gridpoint
     are set. */
  InvertTrans(trans, &backtrans);
  for (i0 = 0; i0 < ext[0]; ++i0)
    for (i1 = 0; i1 < ext[1]; ++i1)
      for (i2 = 0; i2 < ext[2]; ++i2)
      { v[0] = i0 + ori[0];
	v[1] = i1 + ori[1];
	v[2] = i2 + ori[2];
	TransVector(v, &backtrans, mask->cell, mask->grid);
	if (VECINMASK(v, mask))
	{ dd = ncheck = 0;
	  for (grid[0] = SMINT(v[0]); (grid[0]-1) < v[0]; ++grid[0])
	  { j0 = grid[0] - mask->ori[0];
	    for (grid[1] = SMINT(v[1]); (grid[1]-1) < v[1]; ++grid[1])
	    { j1 = grid[1] - mask->ori[1];
	      for (grid[2] = SMINT(v[2]); (grid[2]-1) < v[2]; ++grid[2])
	      { j2 = grid[2] - mask->ori[2];
	        for (i = 0, dist = 0; i < 3; ++i)
		  dist += SQR(v[i] - grid[i]);
		dist = sqrt(dist);
		if (dist < 0.5)
		  { if (mask->msk[j0][j1][j2] >= 1)
	            { newmask->msk[i0][i1][i2] = 1;
		      ++ newmask->nset;}
		  grid[0] = grid[1] = grid[2] = 99999;
                  ncheck = 99999.0;}
		else if (dist < 0.87)
		{ ncheck += 1;
		  dd += mask->msk[j0][j1][j2];}}}}
	  if (ncheck  < (dd * 2))
	  { newmask->msk[i0][i1][i2] = 1;
	    ++newmask->nset;}}}

  /* Now trim the new mask to save some space and add the new transforms. */
  ShrinkMask(newmask);
  for (i = 0; i < mask->ntrans; ++i)
  { Mat3Mul(trf.rotate, mask->trans[i].rotate, backtrans.rotate);
    Vec3Rot(trf.translate, mask->trans[i].rotate, backtrans.translate);
    trf.translate[0] += mask->trans[i].translate[0];
    trf.translate[1] += mask->trans[i].translate[1];
    trf.translate[2] += mask->trans[i].translate[2];
    AddTrans(newmask, &trf, NULL);}
  strcpy(newmask->outfile, trans->outfile); 
  newmask->parent = mask;
  trans->child = newmask;
  return(newmask);}



/******************************************************************************/

#if defined (PROTOTYPE)
  void TransVector(double *v, TRANSFORM *trans, float *cell, int *grid)
#else
  void TransVector(v, trans, cell, grid)
  double *v;
  TRANSFORM *trans;
  float *cell;
  int *grid;
#endif
/* Transforms 'v' (expressed in grid coordinates) using the transform defined
   by 'trans' (which is is defined in orthogonal Angstroms). Because both are 
   differently expressed, the cell and the grid have got to be known by this
   function. */

{ double r[3];
  double *t;
  static float strcell[6] = {0, 0, 0, 0, 0, 0};
  float volume;
  static TRANSFORM om, sm;
  int i,j,k;
#if !defined (PROTOTYPE)
  void Vec3Rot();
  void InvertTrans();
#endif  

  if ((cell[3] == 90.0) && (cell[4] == 90.0) && (cell[5] == 90.0))
  { /* From grid to Angstrom */ 
    v[0] *= cell[0] / grid[0];
    v[1] *= cell[1] / grid[1];
    v[2] *= cell[2] / grid[2];
    
    /* Rotate... */
    Vec3Rot(r, trans->rotate, v);
    
    /* Translate... */ 
    t = trans->translate;
    r[0] += t[0];
    r[1] += t[1];
    r[2] += t[2];

    /* From Angstroms to grid... */  
    v[0] = (r[0] / cell[0]) * grid[0];
    v[1] = (r[1] / cell[1]) * grid[1];
    v[2] = (r[2] / cell[2]) * grid[2];}
  else 
  { /* To fractional cell coordinates... */
    v[0] /= grid[0];
    v[1] /= grid[1];
    v[2] /= grid[2];

    /* Orthogonalise... */
    if ((cell[0] != strcell[0]) || (cell[1] != strcell[1]) ||
	(cell[2] != strcell[2]) || (cell[3] != strcell[3]) ||
	(cell[4] != strcell[4]) || (cell[5] != strcell[5]))
    { double dcell[6];
      double rf[3][3], ro[3][3];
      dcell[0] = strcell[0] = cell[0];
      dcell[1] = strcell[1] = cell[1];
      dcell[2] = strcell[2] = cell[2];
      dcell[3] = strcell[3] = cell[3];
      dcell[4] = strcell[4] = cell[4];
      dcell[5] = strcell[5] = cell[5];
      volume = ccp4uc_frac_orth_mat(dcell,1,ro,rf);
      for (i = 0, k = 0; i < 3; ++i)
        for ( j = 0 ; j < 3 ; ++j, ++k)
          om.rotate[k] = ro[i][j];
      InvertTrans(&om, &sm);}
    Vec3Rot(r, om.rotate, v);

    /* Rotate... */
    Vec3Rot(v, trans->rotate, r);

    /* Translate... */
    t = trans->translate;
    v[0] += t[0];
    v[1] += t[1];
    v[2] += t[2];

    /* Skew back... */
    Vec3Rot(r, sm.rotate, v);

    /* Back to grid... */
    v[0] = r[0] * grid[0];
    v[1] = r[1] * grid[1];
    v[2] = r[2] * grid[2];}}
    

  
  
/******************************************************************************/

#if defined (PROTOTYPE)
  void InvertTrans(TRANSFORM *trans, TRANSFORM *backtrans)
#else
  void InvertTrans(trans, backtrans)
  TRANSFORM *trans, *backtrans;
#endif
/* Inverts the transformation defined by 'trans' and returns the result in
   'backtrans' */

{ float **src, **rot;
  double t[3];
  int i;
#if !defined (PROTOTYPE)
  void Vec3Rot();
  void free_matrix();
  float **invert_matrix();
#endif

  /* First make a matrix which is accepted by the Numerical Recipe... */
  src = matrix(1, 3, 1, 3);
  src[1][1] = (float) trans->rotate[0];
  src[1][2] = (float) trans->rotate[1];
  src[1][3] = (float) trans->rotate[2];
  src[2][1] = (float) trans->rotate[3];
  src[2][2] = (float) trans->rotate[4];
  src[2][3] = (float) trans->rotate[5];
  src[3][1] = (float) trans->rotate[6];
  src[3][2] = (float) trans->rotate[7];
  src[3][3] = (float) trans->rotate[8];

  /* Invert this matrix and put it back in 'trans'... */
  rot = invert_matrix(src, 3);
  backtrans->rotate[0] = rot[1][1];
  backtrans->rotate[1] = rot[1][2];
  backtrans->rotate[2] = rot[1][3];
  backtrans->rotate[3] = rot[2][1];
  backtrans->rotate[4] = rot[2][2];
  backtrans->rotate[5] = rot[2][3];
  backtrans->rotate[6] = rot[3][1];
  backtrans->rotate[7] = rot[3][2];
  backtrans->rotate[8] = rot[3][3];
  free_matrix(src, 1, 3, 1, 3);
  free_matrix(rot, 1, 3, 1, 3);

  /* Now apply this inverse rotation to the translation and negate. */
  Vec3Rot(t, backtrans->rotate, trans->translate);
  for (i = 0; i < 3; ++i)
    backtrans->translate[i] = -t[i];}
