/*

     This code is distributed under the terms and conditions of the
     CCP4 Program Suite Licence Agreement as a CCP4 Application.
     A copy of the CCP4 licence can be obtained by writing to the
     CCP4 Secretary, Daresbury Laboratory, Warrington WA4 4AD, UK.
*/
#include <ccp4_utils.h>     
#include <ccp4_types.h>     

typedef char BOOLEAN;

typedef char Byte;


typedef double VECTOR[3];


typedef struct { double rotate[9];    /* Matrix defining the rotation. */
		 double translate[3]; /* Vector defining subsequent 
					 translation. */
		 char outfile[256];   /* File to which transformed mask will be 
					 written. */
		 struct mask *child;  /* Mask produced by the transform, or NULL
					 if no mask has yet been generated. */
	       } TRANSFORM;


typedef struct mask 
               { char outfile[256];   /* File to which mask will be written. */
		 char infile[256];    /* File from which mask will be read. */
                 int grid[3];         /* Grid of the mask. */
                 int ori[3];          /* Smallest coordinates of mask. */
		 int ext[3];          /* Extent of the mask. */
		 float cell[6];       /* Unit cell. */
		 int ntrans;          /* Number of transformationss to be
					 applied to mask. */
		 struct mask *parent; /* Mask from which this mask was derived. */
		 TRANSFORM *trans;    /* The transformations. */
		 Byte ***msk;         /* The actual mask. */
		 char title[81];      /* Title of mask */
		 int nset;            /* Number of set gridpoints */
	       } MASK;

typedef struct { char outfile[256];   /* File to which map will be written. */
		 char infile[256];    /* File from which map was read. */
		 char title[81];      /* Map title. */
		 int order[3];        /* Order of axes as read from the file */
		 int mode;            /* At the momen only mode 2 is supported. */
		 int spcgrp;          /* Spacegroup. */
                 int grid[3];         /* Grid of the map. */
                 int ori[3];          /* Smallest coordinates of map. */
                 int ext[3];          /* Extent of the map. */
                 float cell[6];       /* Unit cell. */
		 float solv;          /* Solvent fraction if known, otherwise -1 */
		 float min;           /* Minimum value in map. */
		 float max;           /* Maximum value in map. */
		 float mean;          /* Mean value of map. */
		 float rms;           /* Root mean square of map. */
                 int nsyms;           /* Number of symmetry operators to be
                                         applied to map. */
                 float sym[4*4*192];  /* The symmetry operators. */
                 float ***map;        /* The actual map. */
               } MAP;

typedef struct { int nclas;           /* Number of classes in histogram. */
		 int tot;             /* Total number of observations. */
		 int *histo;          /* Actual histogram */
		 float min;           /* Minimum value. */
		 float max;           /* Maximum value. */
		 float mean;          /* Mean value of all contributing obs. */
		 float rms;           /* Rms value of all contributing obs. */
		 float csiz;          /* range of each class. */
	       } HISTOGRAM;

#define SQR(x) ((x)*(x))

#define CUB(x) ((x)*(x)*(x))
/* Already defined in library.h
#define DEGREE(x) ((((x < 0)?(x)+2*3.14159265358979323846:(x))*360)/(2*3.14159265358979323846))

#define RADIAN(x) ((((x<0)?(x)+360:(x))*2*3.14159265358979323846)/360)
*/
#define MAX(x, y) (((x)>(y))?(x):(y))

#define MIN(x, y) (((x)<(y))?(x):(y))

#define ABS(x) (((x)<0)?-(x):(x))

#define SIGN(x) (((x)<0)?-1:1)

#define SMINT(x) (((x)<0)?((int)((x)-1)):(int)(x))
/* Returns integer smaller than, or equal to (x) */

#define NINT(x) (((x)<0)?((int)((x)-0.5)):(int)((x)+0.5))
/* Returns integer closest to (x) */

#define PMOD(x,y) (((x%y)<0)?((y)+((x)%(y))):((x)%(y)))

#define VECINMASK(v, m)	( ( (v)[0] >= (m)->ori[0]) && ( (v)[1] >= (m)->ori[1]) && ( (v)[2] >= (m)->ori[2]) && ( ((v)[0]+1) <= ((m)->ori[0]+(m)->ext[0])) && ( ((v)[1]+1) <= ((m)->ori[1]+(m)->ext[1])) && ( ((v)[2]+1) <= ((m)->ori[2]+(m)->ext[2])))
/* Returns TRUE if the vector "v" falls inside the area defined by MASK 'm' */

#define VECINMAP(v, m) VECINMASK(v, m)
/* Returns TRUE if the vector "v" falls inside the area defined by MAP 'm' */


#define SPLINE(x, a, b, c, d) ((b)+(x)*((c)-((2*(a)-(d)-3*(b))/6))+SQR(x)*((((c)+(a))/2)-b)+CUB(x)*(((d)-(a)+3*((c)-(b)))/6))
/* Cubic spline interpolation: 0 <= (splinx) <= 1, the values a, b, c & d are the 
   measured values at -1, 0, 1 & 2 respectively. */
 
#define ucase(c) ((c>='a' && c <= 'z') ? c - 'a' + 'A' : c)
/* Change char c to uppercase */ 

#if !defined (VMS) && !defined (_MSV_VER)
#define malloc ccp4_utils_malloc
/* ccp4malloc is a wrapper in library.c for malloc to trap for failures */

#define realloc ccp4_utils_realloc
/* ccp4realloc is a wrapper in library.c for realloc to trap for failures */

#define calloc ccp4_utils_calloc
/* ccp4calloc is a wrapper in library.c for calloc to trap for failures */
#endif

#ifndef TRUE
#define TRUE 1
#endif
#ifndef FALSE
#define FALSE 0
#endif

#undef HUGE
#define HUGE 1.0e30		/* HUGE_VAL is double; this should be
				   OK evrywhere. */


#if defined (PROTOTYPE)

int strncasecmp(const char *, const char *, size_t);

MAP *MallocMap(int grid[], int ori[], int ext[], float cell[]);
/* Returns a pointer to an empty map with the requested origin & extent. */

MAP *SubMap(MAP *map, int ori[], int ext[]);
/* Returns a pointer to a new, empty map which inherits most of its 
   characteristics from the input map, but with the specified origin & extent. */

MAP *ReadMap(char *mapname);
/* Read a CCP4 map. */

char *StringFrtToC(char *c_string, int *frt_string);
/* Translates a string encoded by integers into a character string. Stops
   if 0 is reached and returns a pointer to c_string. */

int *StringCToFrt(int *frt_string, char *c_string, int strlen);
/* Translates character string into a string encoded by integers. */

void PrintMap(MAP *map);
/* Prints relevant information of the structure `map`. */

MAP *MapFill(MASK *mask, MAP *map, TRANSFORM *trans, TRANSFORM *csym, int nsym);
/* Fills a 'submap' with density taken from 'map' using the transformation
   defined by 'trans'. The crystallographic transformations are passed in
   'csym', the number of crystallographic transformations is passed in 
   'nsym'. The original map should contain all the density required. The
   transfromation defines which density the submask would contain if it
   were to be moved into the original map. The submap has the same origin &
   extent as the mask, and only gridpoints set in the mask are filled with
   density. */

float MapGrid(MAP *map, int *pnt, TRANSFORM *csym, int nsym);
/* Returns the electron density of 'map' at the gridpoint defined by 'pnt' or
   any of its symmetry mates defined by 'csym' (the crystallographic symmetry
   operators, nsym is the number of crystallographic symmetry operators). */

#endif /* (PROTOTYPE) */


#define VABSQR(v) (SQR((v)[0])+SQR((v)[1])+SQR((v)[2]))
/* VECTOR v;
   Returns the square of the abolute value of vector v. */

#define vabs(v) (sqrt(VABSQR(v)))
/* VECTOR v;
   Returns the absolute value of vector v. */


#if defined (PROTOTYPE)

void TrimMasks(MASK **mask, int nmask);
/* Trims overlap between (symmetry related) copies of all masks in 'mask[]'.
   Copies of all specified trimmed masks will be written to disk, together 
   with their symmetry operators. The array 'mask[]' contains 'nmask'
   entries. */

MASK *MallocMask(int grid[], int ori[], int ext[], float cell[]);
/* Returns a pointer to an empty MASK with the requested origin and extent. */

void FreeMask(MASK *mask);
/* Frees memory occupied by 'mask'. */

void AddTrans(MASK *mask, TRANSFORM *trans, char *outfile);
/* Adds a new transformation to the specified mask. */

char *ReadLine(FILE *file, char *key, char *to, BOOLEAN begin);
/* Returns the next line from `file` which starts with `key`, stripped of
   the first token. In identifying this line, no distinction is made between
   upper and lower case. The search stops if EOF is hit, or (if `to != NULL`)
   a line starting with any of the keys, separated by spaces, which are
   defined in `to` is reached. If the search stops without finding a line 
   starting with `key`, NULL is returned. The file is rewound before parsing if
   `begin == TRUE`. If `key == NULL` no subsequent parsing will be undertaken. 
   The file is read in on the first call with a non-NULL `file`
   and subsequent calls with `file == NULL` access the internal buffer.

   Since everything is stored internally, this function also works for stdin.
   The buffers are not of infinite size however, so do not parse files larger
   than 10 kbyte, files which contain more than 1024 lines or files with lines
   containing more than 1024 characters. */

int ParseInput(MASK **mask, MAP **map, BOOLEAN *average, char *newmapname,
	       char *rmsmapname, float *radius, float *mintrunc, float *maxtrunc,
	       float *ptos, float *solvdens, float *solvmul, float *protmul,
	       char *mapmin, float *mapminscale, float *sdred, char *solmaskname,
	       float *validate, BOOLEAN *extrude, BOOLEAN *keepmean, float *sharp);

/* Reads the masks as specified by the commands in 'stdin'. The format is:
   (not necessarily in this order)
   ...
   ...
   verbose if more output is desired
   slvfrc  (fraction of the map to be flattened)
   trunc   (followed by one or two numbers: minimum & maximum percentage of
            protein to be truncated by solvent flattening)
   ptos     (fractional of protein versus solvent density, default: 1.31)
   extrude  (If present, the mean protein to solvent ratio will be adjusted
             to "ptos" by adding a constant to the solvent fraction, assuming
	     that the minimum in the protein region represents zero density
	     after adding in the F000);
   slvdens (mean electron density of the solvent, used as mean relative to
             which the local standard deviation is determined, default: 0)
   slvmul  (multiplier of solvent density, or "auto", followed by required
            modification of the rms of the solvent with respect to the rms 
            of the protein)
   sharp   (exponent by which the protein density should be sharpened. Default
            is 1).
   mulprot (multiplier of the protein)
   mapmin  (Map to be subtracted from protein area (associated with logical 
            MAPMIN, the map will be scaled by the factor following the number, 
            default 1. If followed ")
   radius  (radius for determining solvent)
   mapout  (Averaged map will be saved as specified, associated with logical 
            MAPOUT)
   rmsmap  (Local rmsmap will be saved as specified, associated with logical
            RMSMAP)
   asmask  (filename of mask defining which part of the asymmetric unit 
            is covered by other masks).
   csym    (filename of O-style crystallographic symmetry operators)
   sdred   (value defining the reduction of the standard deviation of the density
            of the protein chain. If included, excursions into the solvent by 
	    density connected to high protein denisty will be romoved from the
	    solvent mask)
   slvmask  filename (Name of the mask file to which the solvent mask is to be
            written)
   ...
   ...
   patch   radius   (Spherical mask with given radius, centered on the
           gridpoint with the lowest local rms)
   mskout  filename (name of the patch)
   mskin   (filename of O-style mask, or "unit.msk" (mask of unit cell) if such
            a file does not exist)
   monitor (if present, calculates sd of mask before averaging)
   weight  (if present, uses sd within mask to determine sd of noise)
   insert  (if present, inserts corresponding map without averaging or flattening)
   solvent (if present, treats the named mask as solvent & flattens it)
   trans   (filename of O-style non-crystallographic symmetry ops)
   ...     (Repeat 'trans (filename)' as often as needed for this mask. Make
            sure that the identity operator is included. Every file defines
	    one symmetry operator.)
   mskout  (filename of altered mask)
   mskin   (other mask)
   trans   (with its own non-crystallographic symmetry operators)
   ...
   mskout  (filename of altered mask)
   ...
   EOF    (Repeat 'mskin (filename)\n trans (filename)\n trans (filename) ...'
          as often as required)

   Lines starting with unrecognized keywords are ignored.  The number of masks
   read in is returned. The first mask contains the asymmetric unit.
   If at least one "newmask" keyword was encountered, "average" will be set to
   'FALSE', otherwise it will be set to 'TRUE'. */

void ReadTrans(char *filename, MASK *mask, MASK *asymu, char *outfile);
/* Reads non-crystallographic O-style transformation from 'filename', combines
   it with the crystallographic transformations, and includes the resulting 
   transformations in 'mask'. The mask 'asymu' contains the asymmetric unit
   and the crystallographic symmetry operations. All newly generated 
   transformations will have the their '.outfile' set to 'outfile'. */

void PrintMask(MASK *mask);
/* Prints some information on the specified mask. */

MASK *TransMask(MASK *mask, TRANSFORM *trans);
/* Generates a new mask by transforming a specified mask and returns a pointer
   to it. Any transformations defined for the old mask are copied into the new
   mask, after applying the inverse of the transform to these transformations
   too. This means that application of any of the original transforms to the 
   original mask is equivalent to applying any of the new corresponding
   transforms to the new mask!! */

void TransVector(double *v, TRANSFORM *trans, float *cell, int *grid);
/* Transforms 'v' (expressed in grid coordinates) using the transform defined
   by 'trans' (which is is defined in orthogonal Angstroms). Because both are 
   differently expressed, the cell and the grid have got to be known by this
   function. */

void InvertTrans(TRANSFORM *trans, TRANSFORM *backtrans);
/* Inverts the transformation defined by 'trans' and returns the result in
   'backtrans' */

MASK *ShrinkMask(MASK *mask);
/* Shrinks the extent of a mask and updates its origin in order to save some 
   space. */

void Mat3Mul(double *r, double *m1, double *m2);
/* Multiplies 3x3 matrices 'm1' and 'm2'. The result is returned in 'r'. */

void Vec3Rot(double *r, double *m, double *v);
/* Multiplies vector 'v' by matrix 'm'. The result is returned in 'r'. */

void DumpMask(MASK *mask);
/* Dumps the contents of the mask to stdout. */

void FillAsUn(MASK *as_mask, MASK **mask);
/* Loops through all masks in 'mask', and if it finds a point which is set,
   it will increment the corresponding gridpoint in the mask of the 
   asymmetric unit, defined by 'as_mask'. The array containing the masks should
   be NULL terminated. */

int TrimMask(MASK *mask, MASK *as_mask, int nset);
/* Checks all the set gridpoints in 'mask' which are on the border of the set
   area and resets them if corresponding gridpoints in the asymmetric unit
   ('as_mask') are set more than once. The corresponding gridpoint in the
   asymmetric unit is updated. A point is considered to be at the border of the
   set area of the mask if the number of unset neighbour points is equal or
   larger than 'nset'. Returns the number of points which have been reset.*/

void TestMask(MASK *mask, MASK *un_mask, int nset);
/* Checks all the set gridpoints in 'mask' and resets them unless the value
   of the corresponding gridpoint in 'un_mask' is equal or higher than 'nset'. */

void ClearMask(MASK *mask);
/* Resets all gridpoints in 'mask' */

void WriteNcs(TRANSFORM *trans, char *filename);
/* Writes transformation operation as 'filename' */

void AverageMap(MAP *map, MASK *mask, TRANSFORM *csym, int nsym, float sd);
/* Averages the area within the map defined by 'mask' and the transformations
   associated with 'mask'. The averaged submap is written to disk as mask->infile,
   where the extension is replaced by '.map'. The crystallographic symmetry
   operators are defined by 'csym', and there are 'nsym' crystallographic
   symmetry operators. If (sd >= 0), 'sd' is an indication of the standard
   deviation of the noise of the map, and therefore the averaging can (and will)
   be weighted according to a normal probability distribution with an s.d. of
   sqrt(2)*sd (because the two observations which are compared have a s.d.). This
   probability distribution is normalised so that if both observations are
   identical, the weight for the observation to be added in, is 1. */

void FreeMap(MAP *map);
/* Frees memory allocated to 'map'. */

int AddGrid(MAP *map, int j0, int j1, int j2, MAP *submap, MASK *mask,
	    TRANSFORM *trans, int ntrans);
/* Checks if the gridpoint [j0+map->ori[0]] [j1+map->ori[1]] [j2+map->ori[2]]
   or one of its symmetry mates as defined by the symmetries in the array
   'trans[]', is present in 'submap', and is set in 'mask'. If it is, it sets
   the electron density of this gridpoint in 'map' to the value found in
   'submap', and returns a '1'. Otherwise it returns a '0'. */

MAP *CombMaps(MAP *map, MASK **mask, int nmask);
/* Combines maps whose names will be derived from the names of the masks in the 
   array 'mask[]'The density of 'map' is changed. The number of masks is passed
   in 'nmask'. The first mask of the array 'mask[]' defines the crystallographic
   symmetry operators. Returns a pointer to 'map'.*/

void ClearMap(MAP *map);
/* Resets all gridpoints in 'map' */

void ShutUp(BOOLEAN gag);
/* Impedes output on stdout if `gag == TRUE`, or allows output on stdout
   if `gag == FALSE`. */

void GridFillMap(MAP *frm, MAP *to, TRANSFORM *trans, int ntrans);
/* The map 'to' is filled with density values from map 'frm', using the
   transformations defined in 'trans[]' for mapping 'frm' on 'to'. The
   routine is much faster than 'MapFill', but is much less flexible: no
   interpolation is undertaken. Therefore this procedure is only suitable
   if 'trans[]' only defines crystallographic symmetry operations. */

void TransAToGrid(TRANSFORM *ang, TRANSFORM *g, float *cell, int *grid);
/* Changes a transform defined in orthogonal Angstroms ('ang') into one defined
   in gridpoints ('g'). */

void ExtractMap(MAP *frm, MAP *to, TRANSFORM *trans, int ntrans);
/* The map 'to' is filled with density values from map 'frm', using the
   transformations defined in 'trans[]' for mapping 'frm' on 'to'. The
   routine is much faster than 'MapFill', but is much less flexible: no
   interpolation is undertaken. Therefore this procedure is only suitable
   if 'trans[]' only defines crystallographic symmetry operations. It is 
   assumed that every gridpoint of map 'to' can be retrieved from map 'frm'.
   If this is not the case, use the routine MapGridFill. */

void MapGridFill(MAP *frm, MAP *to, MASK *cumm, MASK *mask);
/* Density values from map 'frm' are added to those of map 'to', using the
   transformations defined in 'cumm->trans' for mapping 'frm' on 'to'. Only
   those points which are set in 'mask' will be added. Use this routine only
   in combination with crystallographic symmetry operators: it is assumed 
   that the symmetry operators exactly map gridpoints on other gridpoints.
   Use this routine if 'to' is larger than 'frm' and not all the gridpoints
   of 'to' can be retrieved from 'frm'. Both 'cumm' and 'frm' should have
   the same ori & ext, and any points which are added to 'to', will be
   marked in 'cumm' by adding 1 to the corresponding mask gridpoint. */

float SplInt(float ***map, double *v);
/* 3D Spline interpolation of an electron density map at the position defined
   by 'v' */

float LinInt(float ***map, double *v);
/* Linear interpolation of an electron density map at the position defined
   by 'v' */

float MeanMap(MAP *map, MASK *mask);
/* Returns the mean of the fraction of the map defined by 'mask'. If 
   ('mask' == NULL), the mean of the whole map is returned. */

float SigmaMap(MAP *map, MASK *mask, float *mean);
/* Returns the standard deviation of the fraction of the map defined by 'mask'.
   If ('mask' == NULL), the s.d. of the whole map is returned. The mean of the
   map is returned in 'mean'. */

float SigmaPosMap(MAP *map, MASK *mask, float *mean);
/* Returns the standard deviation of the fraction of the positive density of
   the map within 'mask'.
   If ('mask' == NULL), the s.d. of the whole map is returned. The mean of the
   map is returned in 'mean'. */

float SkewnessMap(MAP *map, MASK *mask, float mean, float sd);
/* Returns the skewness of the fraction of the map defined by 'mask'.
   If ('mask' == NULL), the skew of the whole map is returned. The mean and 
   standard deviation should be passed in 'mean' and 'sd'. */

float GridDist(int *p, int *q, MAP *map);
/* Returns the distance in Angstroms between gridpoints p[] & q[]. */

HISTOGRAM *HistMap(MAP *map, MASK *mask, int nclasses);
/* Histograms map into 'nclasses', if 'mask' != NULL, only the gridpoints set in
   'mask' are histogrammed. */

HISTOGRAM *DumpHist(HISTOGRAM *hst);
/* Dumps info on 'hst' & returns 'hst'. */

float SafeSolv(MAP *map, float safe, float solvcont, int ncl, int n);
/* Returns the maximum local rms which can safely be attributed to solvent,
   which, when used for constructing a mask, will not contain more than a
   'safe' fraction of protein. However, not more that 'solvcont' fraction
   of the map will be flattened. */

MASK *ContourMap(MAP *map, float val);
/* Creates a mask in which all the gridpoints corresponding to values in 'map'
   which are higher than 'val', are set. */

float SolvSkew(MAP *map);
/* Returns the skewness of the histogram assuming its centre is the mean, rather
   than the actual mean. This is like determining the skewness of the solvent,
   assuming the absence of protein. */

MASK *SphericalMask(MAP *map, float radius, int *centre);
/* Returns a spherical mask with the same grid & cell as the map, centered around
   'centre'. */

void FreeHist(HISTOGRAM *hst);
/* Frees space allocated to 'hst' */

void WriteMap(MAP *map, char *mapname);
/* Writes 'map' as 'mapname'. */

float MinMap(MAP *map, MASK *mask, float fraction);
/* Returns the maximum density-cutoff defining the 'fraction' of gridpoints. If
   'mask != NULL', only the points within 'mask' are considered. */

MASK *MinPatch(MAP *map, TRANSFORM *csym, int nsym, float radius);
/* Returns a spherical mask with the specified radius, which is centered on 
   the gridpoint with the lowest local rms (using the same radius). */

MAP *VarMap(MAP *map, float radius, TRANSFORM *csym, int nsym, float solvmean);
/* Returns a map which contains the local variation of 'map' within a spherical
   volume defined by 'radius', relative to solvmean. */

MAP *RmsMap(MAP *map, float radius, TRANSFORM *csym, int nsym);
/* Returns a map which contains the local rms of 'map' within a spherical
   volume defined by 'radius' */

MAP *RmsMap2(MAP *map, float radius, TRANSFORM *csym, int nsym);
/* Returns a map which contains the local rms of 'map' within a spherical
   volume defined by 'radius' */

MASK *CarveMask(MASK *frm, MASK *to, TRANSFORM *trans, int ntrans);
/* Gridpoints set in both masks 'frm' and 'to', will be removed from
   'to'. Returns 'to'. The transfromations should define crystallographic
   symmetries, ie. transform gridpoints onto gridpoints. */

double *NormalDistr();
/* Returns a pointer to an array which defines a normal probability distribution 
   with a standard deviation of 1 and a mean of 0, with 30000 elements, each
   element specifying an increment of 0.001 in sigma. */

MAP *FlatMap(MAP *map, float solv, float radius, float mintrunc, float maxtrunc,
	     MASK **mask, int nmask, float sd, float ptos, float solvmean,
	     float solvmul, char *rmsmapname, float protmul, char *mapmin,
	     float mapminscale, float sdred, char *solmaskname, float *validate,
	     BOOLEAN extrude, BOOLEAN keepmean, float sharp);
/* Flattens 'map', uses a solvent content as defined by 'solv', an integration
   radius for determining the local rms of 'radius' (if 'radius' is positive),
   and truncates a fraction 'mintrunc' of the lowest density values in the 
   protein, and also density values which are higher than a 'maxtrunc' fraction. 
   The array 'mask[]' can define patches which are not to be flattened (masks with
   the title 'weight' or 'monitor') or which are to be flattened (masks with the
   title 'solvent'). The solvent is set to the proper level (according to 
   'ptos', the ratio between the mean protein density and the mean solvent
   density), and attenuated by the factor 'solvmul' (which can be negative!!)
   If there exists a mapfile called `mapmin`, the protein density will be
   multiplied by `protmul`, and at each gridpoint within the protein mask, the
   corresponding gridpoint from mapmin, after scaling by `mapminscale`, will be
   subtracted. If the fraction 'sdred' > 0, the deviation from the mean of the
   protein chain will be reduced by this fraction and the solvent mask will 
   include any excursions into the solvent connected to high density within
   the protein. If 'solvmul' is larger than 900, 1000 will be subtracted, and the
   rms of the solvent will be scaled relative to the rms of the protein.
   The local rms map will be written to disk as 'rmsmapname' if a valid name
   was specified. */

MASK* AddMask(MASK *frm, MASK *to, TRANSFORM *trans, int ntrans);
/* Gridpoints set in either masks 'frm' and 'to', will be set in 
   'to'. Returns 'to'. The transfromations should define crystallographic
   symmetries, ie. transform gridpoints onto gridpoints. */

void FillGridp(MAP *map, MASK *mask, int *p, TRANSFORM *trans, int ntrans,
	       float thresh, int depth, MASK *tstmsk, Byte border);
/* Recursively fills 'mask': if the gridpoint defined by p[] is lower than 
   'thresh', it will be excluded in the mask 'mask', likewise for its neighbours.
   If the boundaries of the map are overstepped, the neighbours are determined
   by using the crystallographic transformations defined by 'trans'. If 'tstmsk'
   != NULL, points are only set if they are within 'tstmsk'. Gridpoints bordering
   the resulting mask will be set to the value 'border'. */

MASK *GetSolvMask(MAP *map, MASK **mask, int nmask, float radius,
		  char *rmsmapname, float solvmean);
/* Returns a solvent mask, containing the lowest fraction of the local rms-map
   as defined by the solvent content, using the integration radius defined by
   'radius', to this mask are added any masks with the title 'solvent', and
   masks bearing the title 'weight' or 'monitor' are excluded. The local rms
   map will be written to disk as 'rmsmapname' if a valid name was specified.*/

MAP *ReadMap(char *mapname);
/* Read a CCP4 map. */

MASK* SetMask(MASK *frm, MASK *to, int set, TRANSFORM *trans, int ntrans);
/* Gridpoints set in masks 'frm', will be set to the value `set` in mask
   'to'. Returns 'to'. The transfromations should define crystallographic
   symmetries, ie. transform gridpoints onto gridpoints. */

MASK *NotMask(MASK *mask);
/* Negates 'mask', turning 1's into 0's and v.v. Other values are kept. */

MASK *ProtSausage(MAP *map, MASK **mask, int nmask, MASK *solmask, float sdred);
/* Determines connected regions within the protein as defined by `solmask`, and
   reduces the standard deviation of these regions from their mean by a factor
   `sdred`. The standard deviation from the mean of the remaining regions within 
   the solvent mask is similarely reduced. An updated solvent mask is returned 
   (it could be that connected density makes an excursion into the solvent). */

void ValidateMap(MAP *map, MASK **mask, MASK *solmask, float *validate);
/* Determines the top 'validate[0]' fraction of the protein part of the map, 
   and expands this fraction to the cutoff value defined by the protein fraction 
   'validate[0]'. then it prints some statistics about this area. */

MASK *GrowMask(MAP *map, float high, float low, MASK **mask, MASK *solmsk);
/* First sets all gridpoints in a mask corresponding to `map` which fall 
   above the `high` fraction of the intensities. Then it adds gridpoints to the
   mask, provided they are above the `low` fraction and that their neighbour
   gridpoints which have already been set. */

void MapHistoMax(MAP *map, MASK *mask, TRANSFORM *trans, int ntrans);
/* Dumps a histogram of the maxima of the map which fall within "mask". If
   "mask" == NULL, the histogram desribes the entire map. */

MASK *MapMaxima(MAP *map, MASK *prot, TRANSFORM *trans, int ntrans);
/* Returns a mask where every point within 'prot' which is a local maximum is set
   to 1, the rest of the points are set to 0. If 'prot' == NULL, the entire 
   map is used. */

MASK *AsymMask(MAP *map, TRANSFORM *trans, int ntrans);
/* Returns a mask, the size of 'map', in which all the unique gridpoints have
   been set. Any gridpoints repeated by crystallographic symmetry as defined by
   'trans' and 'ntrans', are marked only once. */

MASK *ReadMask(char *filename);
/* Reads an O or CCP4 mask-file and returns a pointer to the
   corresponding MASK. (KDC) */

void WriteMask(MASK *mask);
/* Writes a mask to disk in O or CCP4 format */

double *NormalDistr(void);
/* Returns a pointer to an array which defines a normal probability
   distribution with a standard deviation of 1 and a mean of 0. */

#endif /* (PROTOTYPE) */
