C
C
C     This code is distributed under the terms and conditions of the
C     CCP4 Program Suite Licence Agreement as a CCP4 Application.
C     A copy of the CCP4 licence can be obtained by writing to the
C     CCP4 Secretary, Daresbury Laboratory, Warrington WA4 4AD, UK.
C
c    deckname: lineq1.f
      subroutine lineq1 (a,b,x,nd,n,nr,s)
c    purpose
c       solves the real matrix equation ax=b with nr right-hand sides.
c    usage
c       call lineq1(a,b,x,nd,n,nr,s)
c    description of parameters
c          a      - (n x n) real coefficient matrix.
c          b      - (n x nr) real right-hand side array.
c          x      - (n x nr) real array for return of solution vectors.
c          nd     - the number of rows for the arrays a, b, and x
c                   in the dimension statement in the user program.
c          n      - the number of equations to be solved.
c          nr     - the number of right-hand sides to be solved.
c          s      - integer variable returned non-zero only if
c                   matrix a is singular to machine accuracy.
c    remarks
c       arrays a and b are not destroyed.
c    method
c       the matrix a is factored into lower and upper triangular
c       matrices l and u and then the equations lz=b and ux=z are
c       solved in turn. iterative refinement is used so solutions
c       are very accurate whenever s is returned equal to zero.
c
c       forsythe and moler   computer solution of linear algebraic
c       systems   prentice-hall 1967.
c
c    written by david s. dodson,  06/01/70
c               computer sciences department purdue university
c
c    revised for cdc205 by b.f. putnam, purdue university, 11/1/83
c
c    revised to use linpack subroutines sgeco and sgefa
c    dave seaman, purdue university, 05/08/91
c     ..................................................................
      integer nd,n,nr,s
      real a(nd,n),b(nd,nr),x(nd,nr)
c
      integer job
      parameter (job=0)
C set length of wk large enough
c     real wk(100)
      real wk(1000)
      integer iwk (1000)
      equivalence (wk,iwk)
      integer i, j, ia, ipvt, iz
C
c     if( nd .gt. 3 .or. n .gt. 3 .or. nr .gt. 3 ) then
c       write(1000,*)
c1000   format(/' ERROR, call of lineq1 with system larger',
c    a    ' than 3-by-3')
c       stop
c     end if
c
C set ipvt to sufficiently large value
c     ipvt = 10
      ipvt = n+1
C
      iz = ipvt + n
      ia = iz + n
      do 20 j=1,n
        do 10 i=1,n
          wk(ia+(j-1)*n+i) = a(i,j)
   10   continue
   20 continue
      call sgeco(wk(ia+1),n,n,iwk(ipvt+1),rcond,wk(iz+1))
      if (1.0 + rcond .eq. 1.0) then
        s = 1
      else
        do 40 j=1,nr
          do 30 i=1,n
            x(i,j) = b(i,j)
   30     continue
          call sgesl(wk(ia+1),n,n,iwk(ipvt+1),x(1,j),job)
   40   continue
        s = 0
      endif
      end
      subroutine sgeco(a,lda,n,ipvt,rcond,z)                            
      integer lda,n,ipvt(n)                                             
      real a(lda,n),z(n)                                                
      real rcond                                                        
c                                                                       
c     sgeco factors a real matrix by gaussian elimination               
c     and estimates the condition of the matrix.                        
c                                                                       
c     if  rcond  is not needed, sgefa is slightly faster.               
c     to solve  a*x = b , follow sgeco by sgesl.                        
c     to compute  inverse(a)*c , follow sgeco by sgesl.                 
c     to compute  determinant(a) , follow sgeco by sgedi.               
c     to compute  inverse(a) , follow sgeco by sgedi.                   
c                                                                       
c     on entry                                                          
c                                                                       
c        a       real(lda, n)                                           
c                the matrix to be factored.                             
c                                                                       
c        lda     integer                                                
c                the leading dimension of the array  a .                
c                                                                       
c        n       integer                                                
c                the order of the matrix  a .                           
c                                                                       
c     on return                                                         
c                                                                       
c        a       an upper triangular matrix and the multipliers         
c                which were used to obtain it.                          
c                the factorization can be written  a = l*u  where       
c                l  is a product of permutation and unit lower          
c                triangular matrices and  u  is upper triangular.       
c                                                                       
c        ipvt    integer(n)                                             
c                an integer vector of pivot indices.                    
c                                                                       
c        rcond   real                                                   
c                an estimate of the reciprocal condition of  a .        
c                for the system  a*x = b , relative perturbations       
c                in  a  and  b  of size  epsilon  may cause             
c                relative perturbations in  x  of size  epsilon/rcond . 
c                if  rcond  is so small that the logical expression     
c                           1.0 + rcond .eq. 1.0                        
c                is true, then  a  may be singular to working           
c                precision.  in particular,  rcond  is zero  if         
c                exact singularity is detected or the estimate          
c                underflows.                                            
c                                                                       
c        z       real(n)                                                
c                a work vector whose contents are usually unimportant.  
c                if  a  is close to a singular matrix, then  z  is      
c                an approximate null vector in the sense that           
c                norm(a*z) = rcond*norm(a)*norm(z) .                    
c                                                                       
c     linpack. this version dated 08/14/78 .                            
c     cleve moler, university of new mexico, argonne national lab.      
c                                                                       
c     subroutines and functions                                         
c                                                                       
c     linpack sgefa                                                     
c     blas saxpy,sdot,sscal,sasum                                       
c     fortran abs,amax1,sign                                            
c                                                                       
c     internal variables                                                
c                                                                       
      real sdot,ek,t,wk,wkm                                             
      real anorm,s,sasum,sm,ynorm                                       
      integer info,j,k,kb,kp1,l                                         
c                                                                       
c                                                                       
c     compute 1-norm of a                                               
c                                                                       
      anorm = 0.0e0                                                     
      do 10 j = 1, n                                                    
         anorm = amax1(anorm,sasum(n,a(1,j),1))                         
   10 continue                                                          
c                                                                       
c     factor                                                            
c                                                                       
      call sgefa(a,lda,n,ipvt,info)                                     
c                                                                       
c     rcond = 1/(norm(a)*(estimate of norm(inverse(a)))) .              
c     estimate = norm(z)/norm(y) where  a*z = y  and  trans(a)*y = e .  
c     trans(a)  is the transpose of a .  the components of  e  are      
c     chosen to cause maximum local growth in the elements of w  where  
c     trans(u)*w = e .  the vectors are frequently rescaled to avoid    
c     overflow.                                                         
c                                                                       
c     solve trans(u)*w = e                                              
c                                                                       
      ek = 1.0e0                                                        
      do 20 j = 1, n                                                    
         z(j) = 0.0e0                                                   
   20 continue                                                          
      do 100 k = 1, n                                                   
         if (z(k) .ne. 0.0e0) ek = sign(ek,-z(k))                       
         if (abs(ek-z(k)) .le. abs(a(k,k))) go to 30                    
            s = abs(a(k,k))/abs(ek-z(k))                                
            call sscal(n,s,z,1)                                         
            ek = s*ek                                                   
   30    continue                                                       
         wk = ek - z(k)                                                 
         wkm = -ek - z(k)                                               
         s = abs(wk)                                                    
         sm = abs(wkm)                                                  
         if (a(k,k) .eq. 0.0e0) go to 40                                
            wk = wk/a(k,k)                                              
            wkm = wkm/a(k,k)                                            
         go to 50                                                       
   40    continue                                                       
            wk = 1.0e0                                                  
            wkm = 1.0e0                                                 
   50    continue                                                       
         kp1 = k + 1                                                    
         if (kp1 .gt. n) go to 90                                       
            do 60 j = kp1, n                                            
               sm = sm + abs(z(j)+wkm*a(k,j))                           
               z(j) = z(j) + wk*a(k,j)                                  
               s = s + abs(z(j))                                        
   60       continue                                                    
            if (s .ge. sm) go to 80                                     
               t = wkm - wk                                             
               wk = wkm                                                 
               do 70 j = kp1, n                                         
                  z(j) = z(j) + t*a(k,j)                                
   70          continue                                                 
   80       continue                                                    
   90    continue                                                       
         z(k) = wk                                                      
  100 continue                                                          
      s = 1.0e0/sasum(n,z,1)                                            
      call sscal(n,s,z,1)                                               
c                                                                       
c     solve trans(l)*y = w                                              
c                                                                       
      do 120 kb = 1, n                                                  
         k = n + 1 - kb                                                 
         if (k .lt. n) z(k) = z(k) + sdot(n-k,a(k+1,k),1,z(k+1),1)      
         if (abs(z(k)) .le. 1.0e0) go to 110                            
            s = 1.0e0/abs(z(k))                                         
            call sscal(n,s,z,1)                                         
  110    continue                                                       
         l = ipvt(k)                                                    
         t = z(l)                                                       
         z(l) = z(k)                                                    
         z(k) = t                                                       
  120 continue                                                          
      s = 1.0e0/sasum(n,z,1)                                            
      call sscal(n,s,z,1)                                               
c                                                                       
      ynorm = 1.0e0                                                     
c                                                                       
c     solve l*v = y                                                     
c                                                                       
      do 140 k = 1, n                                                   
         l = ipvt(k)                                                    
         t = z(l)                                                       
         z(l) = z(k)                                                    
         z(k) = t                                                       
         if (k .lt. n) call saxpy(n-k,t,a(k+1,k),1,z(k+1),1)            
         if (abs(z(k)) .le. 1.0e0) go to 130                            
            s = 1.0e0/abs(z(k))                                         
            call sscal(n,s,z,1)                                         
            ynorm = s*ynorm                                             
  130    continue                                                       
  140 continue                                                          
      s = 1.0e0/sasum(n,z,1)                                            
      call sscal(n,s,z,1)                                               
      ynorm = s*ynorm                                                   
c                                                                       
c     solve  u*z = v                                                    
c                                                                       
      do 160 kb = 1, n                                                  
         k = n + 1 - kb                                                 
         if (abs(z(k)) .le. abs(a(k,k))) go to 150                      
            s = abs(a(k,k))/abs(z(k))                                   
            call sscal(n,s,z,1)                                         
            ynorm = s*ynorm                                             
  150    continue                                                       
         if (a(k,k) .ne. 0.0e0) z(k) = z(k)/a(k,k)                      
         if (a(k,k) .eq. 0.0e0) z(k) = 1.0e0                            
         t = -z(k)                                                      
         call saxpy(k-1,t,a(1,k),1,z(1),1)                              
  160 continue                                                          
c     make znorm = 1.0                                                  
      s = 1.0e0/sasum(n,z,1)                                            
      call sscal(n,s,z,1)                                               
      ynorm = s*ynorm                                                   
c                                                                       
      if (anorm .ne. 0.0e0) rcond = ynorm/anorm                         
      if (anorm .eq. 0.0e0) rcond = 0.0e0                               
      return                                                            
      end                                                               
      subroutine sgesl(a,lda,n,ipvt,b,job)                              
      integer lda,n,ipvt(n),job                                         
      real a(lda,n),b(n)                                                
c                                                                       
c     sgesl solves the real system                                      
c     a * x = b  or  trans(a) * x = b                                   
c     using the factors computed by sgeco or sgefa.                     
c                                                                       
c     on entry                                                          
c                                                                       
c        a       real(lda, n)                                           
c                the output from sgeco or sgefa.                        
c                                                                       
c        lda     integer                                                
c                the leading dimension of the array  a .                
c                                                                       
c        n       integer                                                
c                the order of the matrix  a .                           
c                                                                       
c        ipvt    integer(n)                                             
c                the pivot vector from sgeco or sgefa.                  
c                                                                       
c        b       real(n)                                                
c                the right hand side vector.                            
c                                                                       
c        job     integer                                                
c                = 0         to solve  a*x = b ,                        
c                = nonzero   to solve  trans(a)*x = b  where            
c                            trans(a)  is the transpose.                
c                                                                       
c     on return                                                         
c                                                                       
c        b       the solution vector  x .                               
c                                                                       
c     error condition                                                   
c                                                                       
c        a division by zero will occur if the input factor contains a   
c        zero on the diagonal.  technically this indicates singularity  
c        but it is often caused by improper arguments or improper       
c        setting of lda .  it will not occur if the subroutines are     
c        called correctly and if sgeco has set rcond .gt. 0.0           
c        or sgefa has set info .eq. 0 .                                 
c                                                                       
c     to compute  inverse(a) * c  where  c  is a matrix                 
c     with  p  columns                                                  
c           call sgeco(a,lda,n,ipvt,rcond,z)                            
c           if (rcond is too small) go to ...                           
c           do 10 j = 1, p                                              
c              call sgesl(a,lda,n,ipvt,c(1,j),0)                        
c        10 continue                                                    
c                                                                       
c     linpack. this version dated 08/14/78 .                            
c     cleve moler, university of new mexico, argonne national lab.      
c                                                                       
c     subroutines and functions                                         
c                                                                       
c     blas saxpy,sdot                                                   
c                                                                       
c     internal variables                                                
c                                                                       
      real sdot,t                                                       
      integer k,kb,l,nm1                                                
c                                                                       
      nm1 = n - 1                                                       
      if (job .ne. 0) go to 50                                          
c                                                                       
c        job = 0 , solve  a * x = b                                     
c        first solve  l*y = b                                           
c                                                                       
         if (nm1 .lt. 1) go to 30                                       
         do 20 k = 1, nm1                                               
            l = ipvt(k)                                                 
            t = b(l)                                                    
            if (l .eq. k) go to 10                                      
               b(l) = b(k)                                              
               b(k) = t                                                 
   10       continue                                                    
            call saxpy(n-k,t,a(k+1,k),1,b(k+1),1)                       
   20    continue                                                       
   30    continue                                                       
c                                                                       
c        now solve  u*x = y                                             
c                                                                       
         do 40 kb = 1, n                                                
            k = n + 1 - kb                                              
            b(k) = b(k)/a(k,k)                                          
            t = -b(k)                                                   
            call saxpy(k-1,t,a(1,k),1,b(1),1)                           
   40    continue                                                       
      go to 100                                                         
   50 continue                                                          
c                                                                       
c        job = nonzero, solve  trans(a) * x = b                         
c        first solve  trans(u)*y = b                                    
c                                                                       
         do 60 k = 1, n                                                 
            t = sdot(k-1,a(1,k),1,b(1),1)                               
            b(k) = (b(k) - t)/a(k,k)                                    
   60    continue                                                       
c                                                                       
c        now solve trans(l)*x = y                                       
c                                                                       
         if (nm1 .lt. 1) go to 90                                       
         do 80 kb = 1, nm1                                              
            k = n - kb                                                  
            b(k) = b(k) + sdot(n-k,a(k+1,k),1,b(k+1),1)                 
            l = ipvt(k)                                                 
            if (l .eq. k) go to 70                                      
               t = b(l)                                                 
               b(l) = b(k)                                              
               b(k) = t                                                 
   70       continue                                                    
   80    continue                                                       
   90    continue                                                       
  100 continue                                                          
      return                                                            
      end                                                               
      real function sasum (n,sx,incx)                                   
      integer            n,incx                                         
      real               sx(n)                                          
      integer            i,m,mp1,ns                                     
      sasum = 0.0e0                                                     
      if (n.le.0) return                                                
      if (incx.eq.1) go to 10                                           
      ns = n*incx                                                       
      do 5 i=1,ns,incx                                                  
         sasum = sasum+abs(sx(i))                                       
    5 continue                                                          
      return                                                            
   10 m = n-(n/6)*6                                                     
      if (m.eq.0) go to 20                                              
      do 15 i=1,m                                                       
         sasum = sasum+abs(sx(i))                                       
   15 continue                                                          
      if (n.lt.6) return                                                
   20 mp1 = m+1                                                         
      do 25 i=mp1,n,6                                                   
         sasum = sasum+abs(sx(i))+abs(sx(i+1))+abs(sx(i+2))+abs(sx(i    
     1   +3))+abs(sx(i+4))+abs(sx(i+5))                                 
   25 continue                                                          
      return                                                            
      end                                                               
      subroutine sgefa(a,lda,n,ipvt,info)                               
      integer lda,n,ipvt(n),info                                        
      real a(lda,n)                                                     
c                                                                       
c     sgefa factors a real matrix by gaussian elimination.              
c                                                                       
c     sgefa is usually called by sgeco, but it can be called            
c     directly with a saving in time if  rcond  is not needed.          
c     (time for sgeco) = (1 + 9/n)*(time for sgefa) .                   
c                                                                       
c     on entry                                                          
c                                                                       
c        a       real(lda, n)                                           
c                the matrix to be factored.                             
c                                                                       
c        lda     integer                                                
c                the leading dimension of the array  a .                
c                                                                       
c        n       integer                                                
c                the order of the matrix  a .                           
c                                                                       
c     on return                                                         
c                                                                       
c        a       an upper triangular matrix and the multipliers         
c                which were used to obtain it.                          
c                the factorization can be written  a = l*u  where       
c                l  is a product of permutation and unit lower          
c                triangular matrices and  u  is upper triangular.       
c                                                                       
c        ipvt    integer(n)                                             
c                an integer vector of pivot indices.                    
c                                                                       
c        info    integer                                                
c                = 0  normal value.                                     
c                = k  if  u(k,k) .eq. 0.0 .  this is not an error       
c                     condition for this subroutine, but it does        
c                     indicate that sgesl or sgedi will divide by zero  
c                     if called.  use  rcond  in sgeco for a reliable   
c                     indication of singularity.                        
c                                                                       
c     linpack. this version dated 08/14/78 .                            
c     cleve moler, university of new mexico, argonne national lab.      
c                                                                       
c     subroutines and functions                                         
c                                                                       
c     blas saxpy,sscal,isamax                                           
c                                                                       
c     internal variables                                                
c                                                                       
      real t                                                            
      integer isamax,j,k,kp1,l,nm1                                      
c                                                                       
c                                                                       
c     gaussian elimination with partial pivoting                        
c                                                                       
      info = 0                                                          
      nm1 = n - 1                                                       
      if (nm1 .lt. 1) go to 70                                          
      do 60 k = 1, nm1                                                  
         kp1 = k + 1                                                    
c                                                                       
c        find l = pivot index                                           
c                                                                       
         l = isamax(n-k+1,a(k,k),1) + k - 1                             
         ipvt(k) = l                                                    
c                                                                       
c        zero pivot implies this column already triangularized          
c                                                                       
         if (a(l,k) .eq. 0.0e0) go to 40                                
c                                                                       
c           interchange if necessary                                    
c                                                                       
            if (l .eq. k) go to 10                                      
               t = a(l,k)                                               
               a(l,k) = a(k,k)                                          
               a(k,k) = t                                               
   10       continue                                                    
c                                                                       
c           compute multipliers                                         
c                                                                       
            t = -1.0e0/a(k,k)                                           
            call sscal(n-k,t,a(k+1,k),1)                                
c                                                                       
c           row elimination with column indexing                        
c                                                                       
            do 30 j = kp1, n                                            
               t = a(l,j)                                               
               if (l .eq. k) go to 20                                   
                  a(l,j) = a(k,j)                                       
                  a(k,j) = t                                            
   20          continue                                                 
               call saxpy(n-k,t,a(k+1,k),1,a(k+1,j),1)                  
   30       continue                                                    
         go to 50                                                       
   40    continue                                                       
            info = k                                                    
   50    continue                                                       
   60 continue                                                          
   70 continue                                                          
      ipvt(n) = n                                                       
      if (a(n,n) .eq. 0.0e0) info = n                                   
      return                                                            
      end                                                               
      subroutine sscal  (n,sa,sx,incx)                                  
      integer            incx,n                                         
      real               sa,sx(n)                                       
      integer            i,m,mp1,ns                                     
      if (n.le.0) return                                                
      if (incx.eq.1) go to 10                                           
      ns = n*incx                                                       
      do 5 i=1,ns,incx                                                  
         sx(i) = sa*sx(i)                                               
    5 continue                                                          
      return                                                            
   10 m = n-(n/5)*5                                                     
      if (m.eq.0) go to 20                                              
      do 15 i=1,m                                                       
         sx(i) = sa*sx(i)                                               
   15 continue                                                          
      if (n.lt.5) return                                                
   20 mp1 = m+1                                                         
      do 25 i=mp1,n,5                                                   
         sx(i) = sa*sx(i)                                               
         sx(i+1) = sa*sx(i+1)                                           
         sx(i+2) = sa*sx(i+2)                                           
         sx(i+3) = sa*sx(i+3)                                           
         sx(i+4) = sa*sx(i+4)                                           
   25 continue                                                          
      return                                                            
      end                                                               
      real function sdot (n,sx,incx,sy,incy)                            
      integer            n,incx,incy                                    
      real               sx(n),sy(n)                                    
      integer            i,m,mp1,ns,ix,iy                               
      sdot = 0.0e0                                                      
      if (n.le.0) return                                                
      if (incx.eq.incy) if (incx-1) 5,15,35                             
    5 continue                                                          
      ix = 1                                                            
      iy = 1                                                            
      if (incx.lt.0) ix = (-n+1)*incx+1                                 
      if (incy.lt.0) iy = (-n+1)*incy+1                                 
      do 10 i=1,n                                                       
         sdot = sdot+sx(ix)*sy(iy)                                      
         ix = ix+incx                                                   
         iy = iy+incy                                                   
   10 continue                                                          
      return                                                            
   15 m = n-(n/5)*5                                                     
      if (m.eq.0) go to 25                                              
      do 20 i=1,m                                                       
         sdot = sdot+sx(i)*sy(i)                                        
   20 continue                                                          
      if (n.lt.5) return                                                
   25 mp1 = m+1                                                         
      do 30 i=mp1,n,5                                                   
         sdot = sdot+sx(i)*sy(i)+sx(i+1)*sy(i+1)+sx(i+2)*sy(i+2)+sx(i   
     1   +3)*sy(i+3)+sx(i+4)*sy(i+4)                                    
   30 continue                                                          
      return                                                            
   35 continue                                                          
      ns = n*incx                                                       
      do 40 i=1,ns,incx                                                 
         sdot = sdot+sx(i)*sy(i)                                        
   40 continue                                                          
      return                                                            
      end                                                               
      subroutine saxpy  (n,sa,sx,incx,sy,incy)                          
      integer            n,incx,incy                                    
      real               sx(n),sy(n),sa                                 
      integer            i,ix,iy,m,mp1,ns                               
      if (n.le.0.or.sa.eq.0.e0) return                                  
      if (incx.eq.incy) if (incx-1) 5,15,35                             
    5 continue                                                          
      ix = 1                                                            
      iy = 1                                                            
      if (incx.lt.0) ix = (-n+1)*incx+1                                 
      if (incy.lt.0) iy = (-n+1)*incy+1                                 
      do 10 i=1,n                                                       
         sy(iy) = sy(iy)+sa*sx(ix)                                      
         ix = ix+incx                                                   
         iy = iy+incy                                                   
   10 continue                                                          
      return                                                            
   15 m = n-(n/4)*4                                                     
      if (m.eq.0) go to 25                                              
      do 20 i=1,m                                                       
         sy(i) = sy(i)+sa*sx(i)                                         
   20 continue                                                          
      if (n.lt.4) return                                                
   25 mp1 = m+1                                                         
      do 30 i=mp1,n,4                                                   
         sy(i) = sy(i)+sa*sx(i)                                         
         sy(i+1) = sy(i+1)+sa*sx(i+1)                                   
         sy(i+2) = sy(i+2)+sa*sx(i+2)                                   
         sy(i+3) = sy(i+3)+sa*sx(i+3)                                   
   30 continue                                                          
      return                                                            
   35 continue                                                          
      ns = n*incx                                                       
      do 40 i=1,ns,incx                                                 
         sy(i) = sa*sx(i)+sy(i)                                         
   40 continue                                                          
      return                                                            
      end                                                               
      integer function isamax (n,sx,incx)                               
      integer            n,incx                                         
      real               sx(n)        
      integer            i,ii,ns                                        
      real               smax,xmag                                      
      isamax = 0                                                        
      if (n.le.0) return                                                
      isamax = 1                                                        
      if (n.le.1) return                                                
      if (incx.eq.1) go to 15                                           
      smax = abs(sx(1))                                                 
      ns = n*incx                                                       
      ii = 1                                                            
      do 10 i=1,ns,incx                                                 
         xmag = abs(sx(i))                                              
         if (xmag.le.smax) go to 5                                      
         isamax = ii                                                    
         smax = xmag                                                    
    5    ii = ii+1                                                      
   10 continue                                                          
      return                                                            
   15 smax = abs(sx(1))                                                 
      do 20 i=2,n                                                       
         xmag = abs(sx(i))                                              
         if (xmag.le.smax) go to 20                                     
         isamax = i                                                     
         smax = xmag                                                    
   20 continue                                                          
      return                                                            
      end                                                               

