c     chef - reflection accumulation subroutine
c     
c     Copyright (C) 2008 STFC, Graeme Winter
c     
c     This code is distributed under the BSD license, a copy of which
c     is included in the root directory if this package.
c     
c     16th January, 2008
c
c     Since many subroutines will want to accumulate all symmetry related
c     reflections then perform some calculation based on them, this 
c     subroutine will gather the reflections then call a passed in function
c     (which will have a NULL result - other than success or failure)
c     to perform the required calculations
c

      subroutine accumulate(calcfunc, status, message)

      implicit none

      integer calcfunc, status
      character message*(*)

c     common blocks containing MTZ information & standard reflection 
c     storage stuff

      include 'parameters.fh'
      include 'arrays.fh'
      include 'options.fh'
      include 'mtz.fh'

c     external subroutines which may be used here

      external lrreff, lrrewd, centr, lenstr

c     function types

      integer lenstr

c     calculation function - defined externally

      external calcfunc

c     local variables - hc, kc, lc - "current" values

      integer i, m, h, k, l, hc, kc, lc, misym, ihkl(3)
      real ival, sigival, base
      real recin(mcol), s
      logical mtzeof, cent

c     local variables to accumulate the completeness as a function 
c     of "base" - p => plus (or native) m => minus, cnref => (unique) counts, 
c     jmax => largest observed batch number, minbatch => first observations
c     of this reflection ...

      integer cnrefp(maxbatch), cnrefm(maxbatch)
      integer pminbatch, mminbatch, jmaxbatch, qminbatch
      
c     then b => both e => either (again for anomalous pairs...)

      integer cnrefb(maxbatch), cnrefe(maxbatch)

c     overall statistics
      
      integer nrefp, nrefm, nrefall
      logical opls, omns

c     start calculation

      do m = 1, nmtz
         
c     count the number of reflections -> nrefall - also initialises
c     centric reflection subroutine so we can compute those reflections
c     right... need to get dmin, dmax from somewhere!

         call nref(m, o_dmin, o_dmax, nrefall)

 999     format('Expected number of reflections in ', 
     +        2f6.2, i7)

         if (o_pcomp) then
            write(*, 999) o_dmin, o_dmax, nrefall
         end if

c     initialise arrays
         
         do i = 1, maxbatch
            cnrefp(i) = 0
            cnrefm(i) = 0
            cnrefb(i) = 0
            cnrefe(i) = 0
         end do
         
         pminbatch = maxbatch
         mminbatch = maxbatch
         jmaxbatch = 0
         
         nrefp = 0
         nrefm = 0
         opls = .false.
         omns = .false.

         cent = .false.

         call lrrewd(hklin(m))

         hc = 0
         kc = 0
         lc = 0
      
         b_npls = 0
         b_nmns = 0

 100     continue

         call lrreff(hklin(m), s, recin, mtzeof)

         if (mtzeof) goto 101

c     check it is within current resolution boundaries

         if ((s .lt. o_smin) .or. (s .gt. o_smax)) goto 100

         h = nint(recin(1))
         k = nint(recin(2))
         l = nint(recin(3))

         misym = nint(recin(4))

         ival = recin(5)
         sigival = recin(6)
         base = recin(7)

c     is this reflection within scope? (1) in terms of base value

         if (base .lt. o_bmin) goto 100
         if ((o_bmax .gt. o_bmin) .and. 
     +        (base .gt. o_bmax)) goto 100

c     (2) in terms of spacing storage

         if (nint((base - o_bmin) / o_bwidth) .gt. maxbatch) then
            status = 2
            write(message, 1) nint((base - o_bmin) / o_bwidth)
            return
         end if

c     record limits now...

         if (nint((base - o_bmin) / o_bwidth) .gt. jmaxbatch) then
            jmaxbatch = nint((base - o_bmin) / o_bwidth)
         end if
         
c     anomalous pair separation - will be useful for rmerge calculations
c     this will move them all to I+

         if (.not. anomalous) then
            if (mod(misym, 2) .eq. 0) misym = misym - 1
         end if


         if ((h .eq. hc) .and. 
     +        (k .eq. kc) .and.
     +        (l .eq. lc)) then

c     then this is another observation of a known reflection
            
            if (mod(misym, 2) .eq. 1) then
               
               b_npls = b_npls + 1

               if (b_npls .ge. maxrefl) then
                  status = 2
                  write(message, 2)
                  return
               end if

               b_ipls(b_npls) = ival
               b_spls(b_npls) = sigival
               b_dpls(b_npls) = 1.0 / sqrt(s)
               b_bpls(b_npls) = base
               b_jpls(b_npls) = nint((base - o_bmin) / o_bwidth)
               b_opls(b_npls) = misym

               opls = .true.
               if (nint((base - o_bmin) / o_bwidth) .lt. pminbatch) 
     +              pminbatch = nint((base - o_bmin) / o_bwidth)

            end if

c     if it is centric, count it towards the opposition as well ;o)

            if ((mod(misym, 2) .eq. 0) .or.
     +           (cent .and. anomalous)) then

               if (b_nmns .ge. maxrefl) then
                  status = 2
                  write(message, 2)
                  return
               end if

               b_nmns = b_nmns + 1
               b_imns(b_nmns) = ival
               b_smns(b_nmns) = sigival
               b_dmns(b_nmns) = 1.0 / sqrt(s)
               b_bmns(b_nmns) = base
               b_jmns(b_nmns) = nint((base - o_bmin) / o_bwidth)
               b_omns(b_nmns) = misym

               omns = .true.               
               if (nint((base - o_bmin) / o_bwidth) .lt. mminbatch) 
     +              mminbatch = nint((base - o_bmin) / o_bwidth)
               
            end if

         else
               
c     this is a new reflection - so first deal with the old
c     buffered observations

            if (opls) then
               do i = pminbatch, maxbatch
                  cnrefp(i) = cnrefp(i) + 1
               end do
               nrefp = nrefp + 1
            end if

            if (omns) then
               do i = mminbatch, maxbatch
                  cnrefm(i) = cnrefm(i) + 1
               end do
               nrefm = nrefm + 1
            end if

            if (opls .or. omns) then
               qminbatch = min(pminbatch, mminbatch)
               do i = qminbatch, maxbatch
                  cnrefe(i) = cnrefe(i) + 1
               end do
            end if

            if (opls .and. omns) then
               qminbatch = max(pminbatch, mminbatch)
               do i = qminbatch, maxbatch
                  cnrefb(i) = cnrefb(i) + 1
               end do
            end if


c     reset these flags

            pminbatch = 1.0e6
            mminbatch = 1.0e6
            
            nrefp = 0
            nrefm = 0
            opls = .false.
            omns = .false.

c     call calculation function - this will do the actual work (external)

            status = calcfunc()

c     then reset the buffers
            
            b_npls = 0
            b_nmns = 0

c     check to see if it is centric...

            ihkl(1) = h
            ihkl(2) = k
            ihkl(3) = l
            call centr(ihkl, cent)
            
            hc = h
            kc = k
            lc = l

c     and store this reflection

            if (mod(misym, 2) .eq. 1) then
               
               b_npls = b_npls + 1
               b_ipls(b_npls) = ival
               b_spls(b_npls) = sigival
               b_dpls(b_npls) = 1.0 / sqrt(s)
               b_bpls(b_npls) = base
               b_jpls(b_npls) = nint((base - o_bmin) / o_bwidth)
               b_opls(b_npls) = misym

               opls = .true.
               if (nint((base - o_bmin) / o_bwidth) .lt. pminbatch) 
     +              pminbatch = nint((base - o_bmin) / o_bwidth)

            end if

c     if it is centric, count it towards the opposition as well ;o)

            if ((mod(misym, 2) .eq. 0) .or. 
     +           (cent .and. anomalous)) then

               b_nmns = b_nmns + 1
               b_imns(b_nmns) = ival
               b_smns(b_nmns) = sigival
               b_dmns(b_nmns) = 1.0 / sqrt(s)
               b_bmns(b_nmns) = base
               b_jmns(b_nmns) = nint((base - o_bmin) / o_bwidth)
               b_omns(b_nmns) = misym

               omns = .true.
               if (nint((base - o_bmin) / o_bwidth) .lt. mminbatch) 
     +              mminbatch = nint((base - o_bmin) / o_bwidth)

            end if


         end if

c     read next reflection

         goto 100

c     end of file on the reflections

 101     continue

c     perform calculations on cached reflections
         
         if (opls) then
            do i = pminbatch, maxbatch
               cnrefp(i) = cnrefp(i) + 1
            end do
            nrefp = nrefp + 1
         end if
         
         if (omns) then
            do i = mminbatch, maxbatch
               cnrefm(i) = cnrefm(i) + 1
            end do
            nrefm = nrefm + 1
         end if

         if (opls .or. omns) then
            do i = mminbatch, maxbatch
               cnrefe(i) = cnrefe(i) + 1
            end do
         end if
         
         if (opls .and. omns) then
            do i = mminbatch, maxbatch
               cnrefb(i) = cnrefb(i) + 1
            end do
         end if

         status = calcfunc()

c     print out completeness statistics if desired - requires calculation
c     of nref within the given resolution limits...

         if (o_pcomp) then

            write(*, 5) pname(m)(1:lenstr(pname(m))),
     +           xname(m)(1:lenstr(xname(m))),
     +           dname(m)(1:lenstr(dname(m)))

            if (anomalous) then
               write(*, 6)
               do i = 1, jmaxbatch
                  write(*, 3) real(i) * o_bwidth + o_bmin,
     +                 real(cnrefp(i)) / real(nrefall), 
     +                 real(cnrefm(i)) / real(nrefall),
     +                 real(cnrefe(i)) / real(nrefall),
     +                 real(cnrefb(i)) / real(nrefall)
               end do
            else
               do i = 1, jmaxbatch
                  write(*, 4) real(i) * o_bwidth + o_bmin,
     +                 real(cnrefp(i)) / real(nrefall)
               end do
            end if
         end if

c     move to next reflection file

      end do

      o_pcomp = .false.
         
 1    format('insufficient storage for rb: ', i8)
 2    format('insufficient reflection storage')
 3    format(f8.1, 4f5.2)
 4    format(f8.1, f5.2)
 5    format('Completeness vs. BASELINE: ', a, '/', a, '/', a)
 6    format('   BASE   I+   I-   I   dI')

      end
