#!/usr/bin/env tclsh

#
# Generate an .mtz file with FMEAN and DANO columns for all datasets
# in the file inputxml that only contain F+ and F-.  Then, merge this
# file and the original file into one file with CAD.
#

#
# sness - This program is one of the place where things could break:
# If the input MTZ labels are very strange the [string trimright]
# construction I'm using might have problems.  Is there a better way
# to do this?
#

# Set CRANK path and load utility functions
set crankbin "$env(CRANK)/bin"
source [file join $crankbin crankutils.tcl]
source [file join $crankbin crankprint.tcl]

if { [lindex $argv 3] == "" } {
	puts "Usage:"
    puts " "
    puts "add-fmean-to-mtz.tcl inputxml inputmtz outputxml output.mtz"
	puts "Where:"
	puts "  inputxml -> Name of input .xml file"
	puts "  outputxml -> Name of .xml file to output"
	puts "  output.mtz -> Name of output .mtz file"
	puts ""
	puts "Copyright (c) Steven Ness 2004"
    exit
}

set inputxml [lindex $argv 0]
if { [file exists $inputxml] == 0 } {
	crank_error "mtz2drear::inputxml file does not exist"
}
XMLParsefile $inputxml

set inputmtz [lindex $argv 1]
set outputxml [lindex $argv 2]
set outputmtz [lindex $argv 3]

set n_datasets $XMLParse(soap__run__mtz_in__n_datasets)

set j 0
set k 1
set l 0
for { set i 0 } { $i < $n_datasets } { incr i } {
	if { $XMLParse(soap__run__mtz_in__data__anomalous__$i) == 1 } {
		set fp_plus $XMLParse(soap__run__mtz_in__data__f_plus__$i)
		set sigfp_plus $XMLParse(soap__run__mtz_in__data__sigf_plus__$i)
		set fp_minus $XMLParse(soap__run__mtz_in__data__f_minus__$i)
		set sigfp_minus $XMLParse(soap__run__mtz_in__data__sigf_minus__$i)

		set f_id [string trimright $fp_plus "(+-)"]
		set sf_id [string trimright $sigfp_plus "(+-)"]

		set d_id "D$f_id"
		set sd_id "D$sf_id"

		set f_output($i) $f_id
		set sf_output($i) $sf_id

		set d_output($i) $d_id
		set sd_output($i) $sd_id

		set labin_lines($j) "F[subst $k](+)=$fp_plus SIGF[subst $k](+)=$sigfp_plus F[subst $k](-)=$fp_minus SIGF[subst $k](-)=$sigfp_minus"
		set labout_lines($j) "F[subst $k]=$f_id SIGF[subst $k]=$sf_id D[subst $k]=$d_id SIGD[subst $k]=$sd_id"
		incr j
		incr k
		incr l
	} else {
		set f_output($i) $XMLParse(soap__run__mtz_in__data__f__$i)
		set sf_output($i) $XMLParse(soap__run__mtz_in__data__sigf__$i)
		set d_output($i) ""
		set sd_output($i) ""
		incr l
	}
}

#
# Create mtzMADmod command
#
set mtz_command "HKLIN $inputmtz HKLOUT $outputmtz"

set mtz_script "TITLE Automatically generated with add-fmean-to-mtz.tcl"

# LABIN section
set mtz_script "$mtz_script\nLABIN"
for { set i 0 } { $i < $j } { incr i } {
	set mtz_script "$mtz_script -\n   $labin_lines($i)"
}

# LABOUT section
set mtz_script "$mtz_script\nLABOUT"
for { set i 0 } { $i < $j } { incr i } {
	set mtz_script "$mtz_script -\n   $labout_lines($i)"
}

set mtz_script "$mtz_script\nEND"

puts "add-fmean-to-mtz.tcl::mtz_command=($mtz_command)"
puts "add-fmean-to-mtz.tcl::mtz_script($mtz_script)"

set command "mtzMADmod $mtz_command << \"$mtz_script\""
set status [catch {eval exec $command } output ]
puts $output


#
# Now, join these two files together with cad
#

set outputxmlfile [open $outputxml w]

puts $outputxmlfile "<soap>"
puts $outputxmlfile "\t<run>"
puts $outputxmlfile "\t\t<mtz_in>"

for { set i 0 } { $i < $n_datasets } { incr i } {
	if { $XMLParse(soap__run__mtz_in__data__anomalous__$i) == 1 } {
		set fp $f_output($i)
		set sigfp $sf_output($i)
		set d $d_output($i)
		set sigd $sd_output($i)
		puts $outputxmlfile "\t\t\t<data id=\"$i\">"
		puts $outputxmlfile "\t\t\t\t<f>$fp</f>"
		puts $outputxmlfile "\t\t\t\t<sigf>$sigfp</sigf>"
		if { $d != "" } {
			puts $outputxmlfile "\t\t\t\t<d>$d</d>"
			puts $outputxmlfile "\t\t\t\t<sigd>$sigd</sigd>"
		}
		puts $outputxmlfile "\t\t\t</data>"
	}
}

puts $outputxmlfile "\t\t</mtz_in>"
puts $outputxmlfile "\t</run>"
puts $outputxmlfile "</soap>"

close $outputxmlfile
