#!/bin/sh
#
# blessing.sh
#
# Script to run Blessing's stuff to get E's for Crunch
#

#
# blessing.sh reads in Scalepack .sca files and calculates difference
# E values by running the various programs in the DREAR suite:
# (http://www.hwi.buffalo.edu/SnB/DREARhelp/DREAR.htm)
#
# Usage:
#  blessing.sh code runtype
#
#  Where:
#	code    -> Compound code (.sca file must start with this prefix)
#	runtype -> SAD - For SAD data
#
#  In the current directory, you should have:
#   contents1 -> Contents of this file
#   code1.sca -> Native or SAD .sca file
#   code2.sca -> Derivative .sca file (if running with SIR)
#

#
# The file contents1 should have the following format:
#       8
#SYNCHROTRON
#       2
#C      750
#Fe       8
#       1
#Fe    0.35   1.25
#
# Where:
# line1 -> Multiplicity of the general position
# line2 -> Radiation type (also supported CU, MO, AG)
# line3 -> The number of atomic types
# line4 -> Number of carbons
# line5 -> Atom type of anomalous scatterers
# line6 -> Number of anomalous scatterers
# line7 -> The f' and f'' of the anomalous scatterers
#

#
# blessing suite written by : Dr. Robert Blessing
#
# wrappers written by : RAG de Graaff
# wrappers editted by : Steven Ness
#

#
# Usage:
#
if [ "$CRANK" = "" ]; then
  echo "Error in running blessing.sh"
  echo ""
  echo "   The CRANK environment variable is not set"
  echo "   Please set it with"
  echo ""
  echo "   setenv CRANK /location/of/crank/top/directory"
  echo "     if using tcsh"
  echo "     --or--"
  echo "   export CRANK=/location/of/crank/top/directory"
  echo "      if using bash"
fi

if [ "$2" = "" ]; then
	echo "Usage:"
	echo "blessing.sh code runtype"
	echo ""
	echo "Where:"
	echo "  code    -> Compound code (.sca file must start with this prefix)"
	echo "  runtype -> SAD - For SAD data"
	echo ""
	echo "(For descriptions of contents1 and other details"
	echo "check out the source code for this script: blessing.sh)"
	exit
fi


wdir=`pwd`
dble=$CRANK/packages/blessing/drivers
dpro=$CRANK/packages/blessing/programs

do_init()
{
compoundcode=$1
runtype=$2
if [ "$runtype" = clear ]; then
do_clear
exit
fi
if [ ! -f $wdir/slimits ]; then
cp $dble/slimits .
fi
if [ ! -f $wdir/contents1 ]; then
echo "Celcontents etc. have not been specified, we need to do this!"
echo "Please create or copy the file \"contents1\" to this directory"
exit
fi
if [ ! -f $wdir/diff.user ]; then
cp $dble/diff.user .
fi
if [ ! -f $wdir/scale.user ]; then
cp $dble/scale.user .
fi
if [ ! -f $wdir/filetje ]; then
cp $dble/filetje .
fi
if [ "$runtype" = SAD ]; then
runtype="SAS"
fi
if [ "$runtype" = SAS ]; then
if [ -f $wdir/$compoundcode.sca ] || [ ! -f $wdir/$compoundcode"1.sca" ]; then
cp $wdir/$compoundcode.sca  $wdir/$compoundcode"1.sca"
fi
else
if [ ! -f $wdir/contents2 ]; then
echo "Celcontents etc. have not been specified for the derivative!"
exit
fi
fi
#
# Input obtained from the file wdir/newinput
#
do_pgrp
read center < $wdir/pointgroupfile
tail -1 $wdir/pointgroupfile > temp
read pointgroup < temp
rm -f temp
head -2 newinput > cell
do_makesortavinp
}

do_clear()
{
rm -f bayes* *bayes edata* eval* locscl* levy* diffe* slimits
rm -f *sortav *locscl *file cell *hkl *sort sort* *diffe
rm -f smax fort* filetje
}

do_makesortavinp()
#
# This routine transforms a .sca file with I+ and I- for the reflections
# into a file containing Friedel pairs for input to sortav and bayes
# SIR data are just copied, it will do two files in this case
# Forbidden reflections are taken out
#
{
	echo "Make sortav input"
	nfile=1
	ln -s $wdir/$compoundcode$nfile.sca fort.10
	ln -s $wdir/sortav$nfile.inp fort.11
	ln -s $wdir/newinput fort.12
ln -s $wdir/slimits fort.20
# 
# slimits contain dmax and dmin limits supplied by the user
#
	$dpro/friedel
	if [ "$runtype" = SIR ]; then
		nfile=2
	rm -f fort.*
		ln -s $wdir/$compoundcode$nfile.sca fort.10
		ln -s $wdir/sortav$nfile.inp fort.11
		ln -s $wdir/newinput fort.12
		ln -s $wdir/slimits fort.20
# 
# slimits contain dmax and dmin limits supplied by the user
#
		$dpro/friedel
	fi
	mv fort.21 slimits
		rm -f fort.*
}

do_pgrp()
#
# gets the pointgroup from the spacegroup symbol and 
# crystal class given in newinput
#
{
	head -5 newinput|tail -1 > $wdir/spacegroupfile
	read spg< $wdir/spacegroupfile
	tail -1 newinput > spacegroupfile
	echo "          "$spg >> spacegroupfile
	ln -s $wdir/spacegroupfile fort.35
	ln -s $wdir/pointgroupfile fort.36
	$dpro/pgrp
	rm -f fort*
}

do_sortav()
#
# Runs the program sortav
# Forbidden reflections must have been removed
#
{
	echo "Do sortav"
if [ -f $wdir/sortav.lp ]; then
mv $wdir/sortav.lp $wdir/sortav1.lp
mv $wdir/sortav.dat $wdir/sortav1.dat
fi
	echo " Sortav run for "$compoundcode, "runtype" $runtype > $wdir/sortav.dat
	echo "  1" >>  $wdir/sortav.dat
	#
	# Only one reflection file at the time
	#
	echo $wdir"/sortav"$nfile".inp" >> $wdir/sortav.dat
	echo $wdir"/"$compoundcode$nfile".sortav" >> $wdir/sortav.dat
	echo "    0" >> sortav.dat
	echo $pointgroup >> $wdir/sortav.dat
	tail -1 cell >>  $wdir/sortav.dat
if [ -f $wdir/missing.hkl ]; then
mv  $wdir/missing.hkl  $wdir/missing1.hkl
fi
rm -f  $wdir/*.sortav $wdir/reject.dat $wdir/twobad.dat
cat $wdir/filetje >> sortav.dat
# 
# Filetje contains the default input provided by the java script
# which is used for running SnB/DREAR These values cannot be changed by 
# the user of SnB
#
cat slimits >> sortav.dat
$dpro/sortav
}

do_bayes()
#
# Runs bayes.f, calcualtes F and E from I
#
{
	echo "Do bayes"
if [ -f $wdir/bayes.lp ]; then
mv $wdir/bayes.lp $wdir/bayes1.lp
mv $wdir/bayes.dat $wdir/bayes1.dat
fi
	echo " Bayes run for "$compoundcode > $wdir/bayes.dat
	echo "    0" >>  $wdir/bayes.dat
	echo "$wdir/$compoundcode$nfile.sortav" >> $wdir/bayes.dat
	echo "$wdir/$compoundcode$nfile.bayes" >> $wdir/bayes.dat
	echo $center >> $wdir/bayes.dat
	echo $pointgroup >> $wdir/bayes.dat
	tail -1 cell >>  $wdir/bayes.dat 
	echo "       100" >> $wdir/bayes.dat
	echo "       100" >> $wdir/bayes.dat
	echo "    0" >> bayes.dat
	$dpro/bayes
}

do_levy()
#
# Calculates scale and overall B
#
{
	echo "Do levy"
if [ -f $wdir/levy.lp ]; then
mv $wdir/levy.lp $wdir/levy1.lp
mv $wdir/levy.dat $wdir/levy1.dat
fi
rm -f  $wdir/edata.levy $wdir/$compoundcode$nfile.levy
echo " Levy run for "$compoundcode > $wdir/levy.dat
echo "    0" >>  $wdir/levy.dat
echo "$wdir/$compoundcode$nfile.bayes" >> $wdir/levy.dat
echo $center >> $wdir/levy.dat
echo $pointgroup >> $wdir/levy.dat
tail -1 cell >>  $wdir/levy.dat 
cat contents$nfile >> levy.dat
cat slimits >> levy.dat
$dpro/levy
rm -f fort*
}

do_eval()
#
# Calculates E1 and E2, from F+ and F- or Fnative and F derivative
#
{
echo "Do eval"
if [ -f $wdir/eval.lp ]; then
mv $wdir/eval.lp $wdir/eval1.lp 
fi
$dpro/eval
mv $wdir/edata.eee $wdir/edata$nfile.eee
mv $wdir/edata.hkl $wdir/edata$nfile.hkl
mv $wdir/eval.dat $wdir/eval$nfile.dat
}

do_locscl()
#
# Does local scaling if specified by the user (YES in the file scale.user)
#
{
	echo "Do locscl"
	echo "Locscl run for "$compoundcode > $wdir/locscl.dat
	if [ "$runtype" = SAS ]; then
	echo "SAS" >> $wdir/locscl.dat
	echo "$wdir/edata1.eee" >> $wdir/locscl.dat
echo "    0" >> $wdir/locscl.dat
	else
	echo $runtype >> locscl.dat
	echo "$wdir/edata1.eee" >> $wdir/locscl.dat
	echo "$wdir/edata2.eee" >> $wdir/locscl.dat
	echo "    0" >> $wdir/locscl.dat
echo "    0" >> $wdir/locscl.dat
	fi
	#
	# Do we do local scaling?
	#
	read scale < scale.user
	if [ "$scale" = YES ]; then
	head -2 scale.user > temp
	tail -1 temp > tmp
	read cutoff < tmp
	rm temp tmp
	mh="         0"
	mk="         0"
	ml="         0"
	zlimit=6.0
	else
	cutoff="    3.000"
	mh="        -1"
	mk="        -1"
	ml="        -1"
	zlimit=0.0
	fi
	echo $cutoff >> $wdir/locscl.dat
	echo $pointgroup >> $wdir/locscl.dat
	tail -1 cell >> $wdir/locscl.dat
	echo $mh$mk$ml >> locscl.dat
	rm -f temp
	echo "         0" >> locscl.dat
	echo "     "$zlimit >> locscl.dat
	$dpro/locscl
	rm -f fort*
}

do_diffe()
#
# Calculates difference E's from E+ and E- or En and Ed
# Eliminates unreasonable values
# Parameters are specified in diff.user, the template contains defaults
#
{
	echo "Do diffe"
	echo "Diffe for"$compoundcode > $wdir/diffe.dat
	echo $wdir"/data.locscl" >> $wdir/diffe.dat
	echo $runtype >> $wdir/diffe.dat 
	echo $wdir"/eval1.dat" >>  $wdir/diffe.dat
if [ "$runtype" = "SIR" ]; then
echo $wdir"/eval2.dat" >>  $wdir/diffe.dat
fi
	cat diff.user >> $wdir/diffe.dat
	cat slimits >> $wdir/diffe.dat
	$dpro/diffe
	rm -f fort*
}

do_drear()
{
	ln -s $wdir/sort.diffe fort.10
	ln -s $wdir/$compoundcode.drear fort.11
	$dpro/drear
	rm -f fort*
}

do_sad()
{
	nfile=1
	if [ ! -f $wdir/$compoundcode$nfile.sortav ]; then
		do_sortav
	fi
	if [ ! -f $wdir/$compoundcode$nfile.bayes ]; then
		do_bayes
	fi
	if [ ! -f $wdir/eval.dat ]; then
		do_levy 
	fi
	rm -f $wdir/edata$nfile.eee $wdir/locscl.lp
	rm -f $wdir/data.neg $wdir/data.pos $wdir/*.locscl $wdir/sort.diffe
	rm -f $wdir/diffe.lp
	do_eval
	do_locscl
	do_diffe
	do_drear
}

do_sir()
{
nfile=1
do_sortav
mv $wdir/sortav.lp $wdir/sortav1.lp
do_bayes
mv  $wdir/edata.bayes $wdir/edata1.bayes
do_levy 
rm -f $wdir/edata$nfile.eee  $wdir/eval.lp $wdir/locscl.lp
rm -f $wdir/data.neg $wdir/data.pos $wdir/*.locscl $wdir/sort.diffe
rm -f $wdir/diffe.lp
do_eval
nfile=2
if [ ! -f $wdir/$compoundcode$nfile.sortav ]; then
do_sortav
fi
if [ ! -f $wdir/$compoundcode$nfile.bayes ]; then
do_bayes
mv  $wdir/edata.bayes $wdir/edata2.bayes
fi
if [ ! -f $wdir/eval2.dat ]; then
do_levy 
fi
rm -f $wdir/edata$nfile.eee $wdir/locscl.lp
rm -f $wdir/data.neg $wdir/data.pos $wdir/*.locscl $wdir/sort.diffe
rm -f $wdir/diffe.lp
do_eval
do_locscl
do_diffe
do_drear
}
#
# End of function definitions
#

do_init $@

if [ "$runtype" = SAS ]; then
	echo "Calculate SAD E's for "$compoundcode
	do_sad
# 	if [ -f $compoundcode.drear ]; then
# 		do_clear
# 	fi
	exit 
	else
	if [ "$runtype" = SIR ]; then
		echo "Calculate SIR E's for "$compoundcode
		do_sir
#		if [ -f $compoundcode.drear ]; then
#			do_clear
#		fi
	exit
	fi
fi
echo " Sorry, we only do SAD or SIR now"
exit
