#!/usr/bin/env tclsh

#
# Take a heavy atom file in fractional coordinates (FRAC format) and
# convert it to orthogonal coordinates (PDB format).
#

#
# Convert the input Crank XML Coords file (with fractional
# coordinates) and convert it into a PDB file (with orthogonal
# coordinates)
#
# Example coordconv script:
#
# coordconv XYZOUT $2 XYZIN $1 << END-convrnxp > /dev/null
# INPUT PDB
# OUTPUT FRAC
# END
# END-convrnxp
#
# Pseudocode:
# 1) Output XML coordinates in FRAC file format
# 2) Run the CCP4 program COORDCONV to convert this to PDB format
#
proc convert_coordinates { outcoordspdb } {
	global XMLParse

	#
	# Output XML coordinates in FRAC file format
	#
	set fracfileid [open "tmp.frac" w]
	set i 1
	while { [info exists XMLParse(coordinate_data__atom__atom_id__$i)] } {
		set output [format "%5i%10.5f%10.5f%10.5f%10.5f%5.2f%10i     %-4sXXX A1" \
						$i \
						$XMLParse(coordinate_data__atom__x__$i) \
						$XMLParse(coordinate_data__atom__y__$i) \
						$XMLParse(coordinate_data__atom__z__$i) \
						$XMLParse(coordinate_data__atom__bfactor__$i) \
						$XMLParse(coordinate_data__atom__occupancy__$i) \
						$i \
						$XMLParse(coordinate_data__atom__element_name__$i)]
		puts $fracfileid $output
		incr i
	}
	close $fracfileid

	#
	# Run the CCP4 program COORDCONV to convert this to PDB format
	#
	set mtz_command "XYZIN tmp.frac XYZOUT $outcoordspdb"

	set cell_a $XMLParse(dataset_info__cell__cell_a)
	set cell_b $XMLParse(dataset_info__cell__cell_b)
	set cell_c $XMLParse(dataset_info__cell__cell_c)
	set cell_alpha $XMLParse(dataset_info__cell__cell_alpha)
	set cell_beta $XMLParse(dataset_info__cell__cell_beta)
	set cell_gamma $XMLParse(dataset_info__cell__cell_gamma)

	set mtz_script "CELL $cell_a $cell_b $cell_c $cell_alpha $cell_beta $cell_gamma"
	set mtz_script "$mtz_script\nINPUT FRAC\nOUTPUT PDB\nEND\n"

	set command "coordconv $mtz_command << \"$mtz_script\""
	set status [catch {eval exec $command } output ]
	
	puts $output

	file delete "tmp.frac"

}

#
# Main
#
if { [lindex $argv 2] == "" } {
	puts "Usage:"
	puts "  coordconv_frac2pdb.tcl in.coords.xml mtzxml out.coords.pdb"
	puts ""
	puts "Where:"
	puts "  coords.xml -> Input coordinates in Crank coords XML format"
	puts "  mtzxml -> Crank MTZ xml file for this compound"
	puts ""
	puts " coords.xml file must contain only one trial"
	puts " "
	puts " Copyright/GPL Feb 2005 Steven Ness"
	exit
}

set crankbin "$env(CRANK)/bin"
source [file join $crankbin crankutils.tcl]

set coordsxmlfile [lindex $argv 0]
set mtzxmlfile [lindex $argv 1]
set outcoordspdb [lindex $argv 2]

XMLParsefile $coordsxmlfile
XMLParsefile $mtzxmlfile

convert_coordinates $outcoordspdb


