#!/usr/bin/env tclsh

#
# From the file inputmtz, extract all the columns specified in the
# Crank XML input file inputxml into the file outputmtz 
#
#
# We need to do this because the user could have arbitrarily named
# columns in their .mtz file that could conflict with the names of
# columns we add in Crank.
#

# Set CRANK path and load utility functions
set crankbin "$env(CRANK)/bin"
source [file join $crankbin crankutils.tcl]
source [file join $crankbin crankprint.tcl]

if { [lindex $argv 2] == "" } {
	puts "Usage:"
    puts " "
    puts "extract-xml-columns.tcl inputxml inputmtz crankxml crankmtz output.mtz"
	puts "Where:"
	puts "  inputxml -> Name of inputxml file"
	puts "  output.mtz -> Name of output .mtz file"
	puts ""
	puts "Copyright (c) Steven Ness 2004"
    exit
}

set inputxml [lindex $argv 0]
if { [file exists $inputxml] == 0 } {
	crank_error "extract-xml-columns.tcl::input xml file does not exist::$inputxml"
}
XMLParsefile $inputxml

set inputmtz [lindex $argv 1]

#
# We can run in two modes: 
#  1) Normal mode, with two input .xml files and two input .mtz files
#  2) SIR mode, with one input .xml file and one input .mtz file
#
if { [lindex $argv 3] == "" } {
	set two_files 0

	set outputmtz [lindex $argv 2]
} else {
	set two_files 1

	set crankxml [lindex $argv 2]
	if { [file exists $crankxml] == 0 } {
		crank_error "extract-xml-columns.tcl::crank xml file does not exist::$crankxml"
	}
	XMLParsefile $crankxml

	set crankmtz [lindex $argv 3]
	set outputmtz [lindex $argv 4]
}

puts "two_files=($two_files)"

set n_datasets $XMLParse(soap__run__mtz_in__n_datasets)

set e 1
set e_added 1
set j 0
for { set i 0 } { $i < $n_datasets } { incr i } {
	set type $XMLParse(soap__run__mtz_in__data__type__$i)

	if { $XMLParse(soap__run__mtz_in__data__anomalous__$i) == 1 } {
		# If we have anomalous data, but also have F and SIGF columns, we
		# have already been through mtzMADmod, and the new columns are actually
		# in the file $crankmtz.  Deal with that properly here.
		if { [info exists XMLParse(soap__run__mtz_in__data__f__$i) ] } {
			# orig.mtz
			set fp_plus $XMLParse(soap__run__mtz_in__data__f_plus__$i)
			set sigfp_plus $XMLParse(soap__run__mtz_in__data__sigf_plus__$i)
			set fp_minus $XMLParse(soap__run__mtz_in__data__f_minus__$i)
			set sigfp_minus $XMLParse(soap__run__mtz_in__data__sigf_minus__$i)

			# added.mtz
			set fp $XMLParse(soap__run__mtz_in__data__f__$i)
			set sigfp $XMLParse(soap__run__mtz_in__data__sigf__$i)
			set d $XMLParse(soap__run__mtz_in__data__d__$i)
			set sigd $XMLParse(soap__run__mtz_in__data__sigd__$i)
			if { [info exists XMLParse(soap__run__mtz_in__data__isym__$i)] } {
				set isym $XMLParse(soap__run__mtz_in__data__isym__$i)
			}

			if {  [info exists XMLParse(soap__run__mtz_in__data__isym__$i)] } {
				set labin_lines($i) "E[expr $e]=$fp_plus E[expr $e+1]=$sigfp_plus E[expr $e+2]=$fp_minus E[expr $e+3]=$sigfp_minus E[expr $e+4]=$isym"
				set labout_lines($i) "E[expr $e]=$fp_plus E[expr $e+1]=$sigfp_plus E[expr $e+2]=$fp_minus E[expr $e+3]=$sigfp_minus E[expr $e+4]=$isym"
				set xname($i) "E$e TO E[expr $e + 4]=$type"
			} else {
				set labin_lines($i) "E[expr $e]=$fp_plus E[expr $e+1]=$sigfp_plus E[expr $e+2]=$fp_minus E[expr $e+3]=$sigfp_minus"
				set labout_lines($i) "E[expr $e]=$fp_plus E[expr $e+1]=$sigfp_plus E[expr $e+2]=$fp_minus E[expr $e+3]=$sigfp_minus"
				set xname($i) "E$e TO E[expr $e + 3]=$type"
			}

			set labin_lines_added($j) "E[expr $e_added]=$fp E[expr $e_added+1]=$sigfp E[expr $e_added+2]=$d E[expr $e_added+3]=$sigd"
			set labout_lines_added($j) "E[expr $e_added]=$fp E[expr $e_added+1]=$sigfp E[expr $e_added+2]=$d E[expr $e_added+3]=$sigd"
			set xname_added($j) "E$e_added TO E[expr $e_added + 3]=$type"
			incr j

			if {  [info exists XMLParse(soap__run__mtz_in__data__isym__$i)] } {
				set e [expr $e + 5]
			} else {
				set e [expr $e + 4]
			}
			set e_added [expr $e_added + 4]

		} else {
			set fp_plus $XMLParse(soap__run__mtz_in__data__f_plus__$i)
			set sigfp_plus $XMLParse(soap__run__mtz_in__data__sigf_plus__$i)
			set fp_minus $XMLParse(soap__run__mtz_in__data__f_minus__$i)
			set sigfp_minus $XMLParse(soap__run__mtz_in__data__sigf_minus__$i)

			set labin_lines($i) "E[expr $e]=$fp_plus E[expr $e+1]=$sigfp_plus E[expr $e+2]=$fp_minus E[expr $e+3]=$sigfp_minus"
			set labout_lines($i) "E[expr $e]=$fp_plus E[expr $e+1]=$sigfp_plus E[expr $e+2]=$fp_minus E[expr $e+3]=$sigfp_minus"
			set xname($i) "E$e TO E[expr $e + 3]=$type"
			set e [expr $e + 4]
		}
	} else {
		set fp $XMLParse(soap__run__mtz_in__data__f__$i)
		set sigfp $XMLParse(soap__run__mtz_in__data__sigf__$i)

		set labin_lines($i) "E[expr $e]=$fp E[expr $e+1]=$sigfp"
		set labout_lines($i) "E[expr $e]=$fp E[expr $e+1]=$sigfp"
		set xname($i) "E$e TO E[expr $e + 1]=$type"
		set e [expr $e + 2]
	}
}

set total $i
set anom_total $j

#
# Create cad command
#
if { $two_files == 1 } {
	set mtz_command "cad HKLIN1 $inputmtz HKLIN2 $crankmtz HKLOUT $outputmtz"
} else {
	set mtz_command "cad HKLIN1 $inputmtz HKLOUT $outputmtz"
}

set mtz_script "TITLE Automatically extracted with extract-xml-columns.tcl"

#
# inputmtz
#

# LABIN section
set mtz_script "$mtz_script\nLABIN FILE 1"
for { set i 0 } { $i < $total } { incr i } {
	set mtz_script "$mtz_script -\n   $labin_lines($i)"
}

# LABOUT section
set mtz_script "$mtz_script\nLABOUT FILE 1"
for { set i 0 } { $i < $total } { incr i } {
	set mtz_script "$mtz_script -\n   $labout_lines($i)"
}

set mtz_script "$mtz_script\nXNAME FILE_NUMBER 1 "
for { set i 0 } { $i < $total } { incr i } {
	set mtz_script "$mtz_script $xname($i)"
}

#
# crankmtz
#

if { $two_files == 1 } {
	# LABIN section
	set mtz_script "$mtz_script\nLABIN FILE 2"
	for { set i 0 } { $i < $anom_total } { incr i } {
		set mtz_script "$mtz_script -\n   $labin_lines_added($i)"
	}

	# LABOUT section
	set mtz_script "$mtz_script\nLABOUT FILE 2"
	for { set i 0 } { $i < $anom_total } { incr i } {
		set mtz_script "$mtz_script -\n   $labout_lines_added($i)"
	}

	set mtz_script "$mtz_script\nXNAME FILE_NUMBER 2 "
	for { set i 0 } { $i < $anom_total } { incr i } {
		set mtz_script "$mtz_script $xname_added($i)"
	}
}

set mtz_script "$mtz_script\nEND"

puts $mtz_command
puts $mtz_script

set command "$mtz_command << \"$mtz_script\""
set status [catch {eval exec $command } output ]

puts $output