#!/usr/bin/env tclsh

#
# crank - Automated macromolecular structure solution
#
# Copyright (C) 2004 Steven Ness
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#
# getwilsonvalues.tcl
#
# Usage:
#   getwilsonvalues.tcl input.mtz nresidues f sigf
#
# Where:
#   input.mtz -> input .mtz file
#   nresidues -> number of residues in the protein
#   f         -> F label of dataset to calculate
#   sigf      -> SIGF label of dataset to calculate
#
# This program outputs b-factor and solvent content as determined from
# the CCP4 program WILSON.  
#

set debug 0

#
# runcommand - Execute an outside program
#
# Run a command using an "exec" wrapped with "catch".  If the external
# command fails, catch the error and return a string describing the
# error message, specifically:
#
# SUCCESS     - Command succeeded
# CHILDKILLED - Child process was killed
# CHILDSTATUS - Child exited with a non-zero error status
# CHILDSUSP   - Child was suspended
# POSIX       - One of the kernel calls to launch the command failed
#                (e.g. file could not be found)
#
# To output more debugging information, such as the output of the 
# command, set debug 1
#
# This clever code is adapted from (http://wiki.tcl.tk/1039)
#
proc runcommand { command } {

	global debug

	set status [catch {eval exec $command } output]

	if { $debug == 1 } {
		puts "-------crankutils::runcommand-------"
		puts "command:($command)"
		puts "status:($status)"
		puts "errorCode=($::errorCode)"
		puts "output:($output)"
		puts "------------------------------------"
	}

    if { $status == 0 } {
        # The command succeeded, and wrote nothing to stderr.
        # $output contains what it wrote to stdout, unless you
        # redirected it.
		set returnval "SUCCESS"

    } elseif { [string equal $::errorCode NONE] } {
        # The command exited with a normal status, but wrote something
        # to stderr, which is included in $output.
		set returnval "SUCCESS"

    } else {

		puts "Error encountered when running:"
		puts "command:($command)"
		puts "crank::runcommand::ERROR CODE($::errorCode)"
		puts "output:($output)"

        switch -exact -- [lindex $::errorCode 0] {

            CHILDKILLED {
                # A child process, whose process ID was $pid,
                # died on a signal named $sigName.  A human-
                # readable message appears in $msg.
                foreach { - pid sigName msg } $::errorCode break
				puts "pid=($pid)"
				puts "sigName=($sigName)"
				puts "msg=($msg)"
				puts "crank::runcommand::STATUS:CHILDKILLED - child killed"
				set returnval "CHILDKILLED"

            }

            CHILDSTATUS {
                # A child process, whose process ID was $pid,
                # exited with a non-zero exit status, $code.
                foreach { - pid code } $::errorCode break
				puts "crank::runcommand::STATUS:CHILDSTATUS - child exited with non-zero exit status"
				set returnval "CHILDSTATUS"

            }

            CHILDSUSP {
                # A child process, whose process ID was $pid,
                # has been suspended because of a signal named
                # $sigName.  A human-readable description of the
                # signal appears in $msg.
                foreach { - pid sigName msg } $::errorCode break
				puts "crank::runcommand::STATUS:CHILDSUSP - child was suspended"
				set returnval "CHILDSUSP"

            }


            POSIX {
				# One of the kernel calls to launch the command
				# failed.  The error code is in $errName, and a
				# human-readable message is in $msg.
                foreach { - errName msg } $::errorCode break
				puts "crank::runcommand::STATUS:POSIX - kernel call failed"
				set returnval "POSIX"
            }
        }
    }

	return $returnval
}

if { [lindex $argv 4] == "" } {
    puts "Usage:"
    puts "getwilsonvalues.tcl input.mtz nresidues f sigf <xml>\n\n"
	puts " Where:"
	puts "   input.mtz -> input .mtz file"
	puts "   nresidues -> number of residues in the protein"
	puts "   f         -> F label of dataset to calculate"
	puts "   sigf      -> SIGF label of dataset to calculate"
	puts "   <xml>     -> One of \"xml\" or \"noxml\", determines the type of output"
	puts "\nSteven Ness - Aug 2004"
    exit
}

set input_mtz [lindex $argv 0]
set nresidues [lindex $argv 1]
set f [lindex $argv 2]
set sigf [lindex $argv 3]

if { [string compare [lindex $argv 4] "xml"] == 0 } {
	set xml 1
} else {
	set xml 0
}

# Check for existance of WILSON program

# Run the command

set output [exec /home/sness/ccp4-5.0/bin/wilson HKLIN $input_mtz << "TITLE RUN FROM GETWILSONVALUES.SH\nWILSON\nNRESIDUES $nresidues\nLABIN FP=$f SIGFP=$sigf"]
#puts $output

if { ![regexp {.*occupied by atoms *= *([-0-9.]*).*} $output full atoms] } {
 	puts "Search string not found"
}

if { ![regexp {.*line gives: *B *= *([-0-9.]*).*} $output full bfactor] } {
 	puts "Search string not found"
}

if { ![info exists atoms] } {
	puts "Could not find \"occupied by atoms\" line.  Exitting\n"
	exit
}

if { ![info exists bfactor] } {
	puts "Could not find \"occupied by atoms\" line.  Exitting\n"
	exit
}

set solvent_content [expr 1.0 - $atoms]

if { $xml } {
puts "<dataset_info>"
puts "\t<solvent_content>$solvent_content</solvent_content>"
puts "\t<b_factor>$bfactor</b_factor>"
puts "</dataset_info>"
} else {
	puts "solvent=$solvent_content\tbfactor=$bfactor"
}
