#!/usr/bin/env tclsh

#
# crank - Automated macromolecular structure solution
#
# Copyright (C) 2004 Steven Ness
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#

#
# mtz2drear.tcl
#
# Converts .mtz files to DREAR file format using the blessing.sh
# script which runs the DREAR suite by Dr. Robert Blessing
#

#
# Usage:
#
# mtz2drear.tcl input.xml
#

#
# Nota bene:
#
#
# This script is the new, better version of mtz2drear.sh, a script
# that uses the CCP4 program ECALC to convert .mtz files to DREAR file
# format.  This script, using the DREAR suite, produces much better
# E-values for CRUNCH2, and should be the one you use.
#

#
# This script calls blessing.sh, a script that runs the various
# programs in the DREAR suite.  This blessing.sh script requires the
# prescence of specifically named files with formatted contents.
#
# The files:
# 1) newinput    -> spacegroup description
# 2) contents1   -> atomic contents of the asymmetric unit
# 3) diff.user   -> parameters for the diffe program (xmin, ymin, etc.)
# 4) scale.user  -> perform local scaling
#
#
# File #1 - newinput
# ------------------
# c.acidurci Ferredoxin (caufd) caufd
# 33.95000  33.95000  74.82000  90.00000  90.00000  90.00000
#    8
#    15
#     P 43 21 2
#     8
#   1  0  0  0.0000000  0  1  0  0.0000000  0  0  1  0.0000000
#   0 -1  0  0.5000000  1  0  0  0.5000000  0  0  1  0.7500000
#  -1  0  0  0.0000000  0 -1  0  0.0000000  0  0  1  0.5000000
#   0  1  0  0.5000000 -1  0  0  0.5000000  0  0  1  0.2500000
#   1  0  0  0.5000000  0 -1  0  0.5000000  0  0 -1  0.2500000
#   0  1  0  0.0000000  1  0  0  0.0000000  0  0 -1  0.0000000
#  -1  0  0  0.5000000  0  1  0  0.5000000  0  0 -1  0.7500000
#   0 -1  0  0.0000000 -1  0  0  0.0000000  0  0 -1  0.5000000
#     4
# 
# Where:
# line 1     -> Title and description
# line 2     -> Number of scatterers
# line 3     -> Scattering power (15 is fine)
# line 4     -> Spacegroup
# line 5     -> Number of symmetry operators
# line 6-end -> Symmetry operators
# line end   -> Crystal class
#               1 - Triclinic
#               2 - Monoclinic
#               3 - Orthorhombic
#               4 - Tetragonal 
#               5 - Rhombohedral
#               6 - Trigonal
#               7 - Hexagonal 
#               8 - Cubic
#
# File #2 - contents1
# -------------------
#       8
#SYNCHROTRON
#       2
#C      750
#Fe       8
#       1
#Fe    0.35   1.25
# 
# Where:
# line 1 -> Multiplicity of general position (number of symmetry
#           operators)
# line 2 -> Type of radiation (SYNCHROTRON, CU, MO or AG)
# line 3 -> Number of atomic types
# line 4 -> Total number of first kind of atom
# line 5 -> Total number of first anomalously scattering atom
# line 6 -> The number of anomalous scatterers
# line 7 -> f' and f'' for the anomalous scatterer
#
# 
# File #3 - diff.user
# -------------------
#     6.000
#     3.000
#     1.000
#     0.000
#     0.000
#
# Where:
# line 1 -> Tmax
# line 2 -> Xmin
# line 3 -> Ymin
# line 4 -> Zmin
# line 5 -> Zmax
#
#
# File #4 - scale.user
# --------------------
# NO
#
# Where:
# line 1 -> Perform local scaling

#
# Pseudocode
#
# 1) Read in input XML file
# 2) Read in input mtz XML file
# 3) Generate newinput
# 4) Generate contents1
# 5) Generate contents2 (if needed)
# 5) Generate diff.user
# 6) Generate scale.user
# 7) Run blessing.sh
#

proc generate_newinput { lattice } {
	global XMLParse

	set outid [open "newinput" w]

	#  Output description + code
	puts $outid "$XMLParse(dataset_info__description) $XMLParse(dataset_info__code)"

	#  Output cell parameters
	set output [format "%-10.5f%-10.5f%-10.5f%-10.5f%-10.5f%-8.5f" \
					$XMLParse(dataset_info__cell__cell_a) \
					$XMLParse(dataset_info__cell__cell_b) \
					$XMLParse(dataset_info__cell__cell_c) \
					$XMLParse(dataset_info__cell__cell_alpha) \
					$XMLParse(dataset_info__cell__cell_beta) \
					$XMLParse(dataset_info__cell__cell_gamma) ]
	puts $outid $output

	#  Output # of atoms to look for
	if { $XMLParse(dataset_info__natoms) != "" } {
		set output [format "%5i" $XMLParse(dataset_info__natoms)]
		puts $outid $output
	} else {
		puts $outid "Could not find dataset_info__natoms - Required data item"
		puts $outid "Exitting"
		exit
	}

	#  Output scattering power
	if { $XMLParse(soap__crunch2__scattering_power) != "" } {
		#set val $XMLParse(soap__crunch2__scattering_power)
		#set rounded [expr round($val)] 
		set output [format "%5i" [expr round($XMLParse(soap__crunch2__scattering_power))]]
		puts $outid $output
	} else {
		set output [format "%5i" 15]
		puts $outid $output
	}

	#  Output spacegroup
	puts -nonewline $outid "    $XMLParse(dataset_info__spacegroup__lattice)"
	set i 0
	while { [info exists XMLParse(dataset_info__spacegroup__operator__$i)] } {
		puts -nonewline $outid " $XMLParse(dataset_info__spacegroup__operator__$i)"
		incr i
	}
	puts $outid ""

	#  Output # of symmetry operators
	puts $outid "    $XMLParse(dataset_info__n_symops)"

	#  Output symmetry operator cards
	set i 0
	while { [info exists XMLParse(dataset_info__symmetry_operator__aa__$i)] } {
		set output [format "%3.0f%3.0f%3.0f %10.7f" \
						$XMLParse(dataset_info__symmetry_operator__aa__$i) \
						$XMLParse(dataset_info__symmetry_operator__ab__$i) \
						$XMLParse(dataset_info__symmetry_operator__ac__$i) \
						$XMLParse(dataset_info__symmetry_operator__atrans__$i)]
		puts -nonewline $outid $output

		set output [format "%3.0f%3.0f%3.0f %10.7f" \
						$XMLParse(dataset_info__symmetry_operator__ba__$i) \
						$XMLParse(dataset_info__symmetry_operator__bb__$i) \
						$XMLParse(dataset_info__symmetry_operator__bc__$i) \
						$XMLParse(dataset_info__symmetry_operator__btrans__$i)]
		puts -nonewline $outid $output

		set output [format "%3.0f%3.0f%3.0f %10.7f" \
						$XMLParse(dataset_info__symmetry_operator__ca__$i) \
						$XMLParse(dataset_info__symmetry_operator__cb__$i) \
						$XMLParse(dataset_info__symmetry_operator__cc__$i) \
						$XMLParse(dataset_info__symmetry_operator__ctrans__$i)]
		puts $outid $output

		incr i
	}
	puts $outid "    $lattice"

	close $outid
}

proc generate_contents { } {
	global XMLParse
	# Change this in the future to do all the common light
	# atoms in proteins and DNA, C, N, O, P, S.

	set n_symops $XMLParse(dataset_info__n_symops)
	set xray_source $XMLParse(soap__crunch2__drear__xray_source)
	set num_c $XMLParse(dataset_info__all_atoms__c)
	set num_n $XMLParse(dataset_info__all_atoms__n)
	set num_o $XMLParse(dataset_info__all_atoms__o)
	set num_s $XMLParse(dataset_info__all_atoms__s)
	set atom_name $XMLParse(dataset_info__atom_name) 
	set natoms $XMLParse(dataset_info__natoms)

	#
	# Set fp and fpp.  
	#
	# This could be done better in the case of missing fp and fpp
	# values.  Ideally, we should estimate this information if it
	# is missing.
	# 
	if { $XMLParse(soap__run__experiment) == "SAD" } {
		if { [ info exists XMLParse(soap__run__mtz_in__data__fprime__0) ] } {
			set fp $XMLParse(soap__run__mtz_in__data__fprime__0)
		} else {
			set fp 0.0
		}
		if { [ info exists XMLParse(soap__run__mtz_in__data__fprimeprime__0) ] } {
			set fpp $XMLParse(soap__run__mtz_in__data__fprimeprime__0)
		} else { 
			set fpp 0.0
		}
	} else {
		if { [ info exists XMLParse(soap__run__mtz_in__data__fprime__1) ] } {
			set fp $XMLParse(soap__run__mtz_in__data__fprime__1)
		} else {
			set fp 0.0
		}
		if { [ info exists XMLParse(soap__run__mtz_in__data__fprimeprime__1) ] } {
			set fpp $XMLParse(soap__run__mtz_in__data__fprimeprime__1)
		} else { 
			set fpp 0.0
		}
	}

	set outid [open "contents1" w]
 
 	puts $outid "        $n_symops"
 	puts $outid "$xray_source"

	if { $XMLParse(soap__run__experiment) == "SAD" } {
		puts $outid "        5"
	} else {
		puts $outid "        4"
	}

	set output [format "%-2s%10i" "C" $num_c]
	puts $outid $output
	set output [format "%-2s%10i" "N" $num_n]
	puts $outid $output
	set output [format "%-2s%10i" "O" $num_o]
	puts $outid $output
	set output [format "%-2s%10i" "S" $num_s]
	puts $outid $output

	if { $XMLParse(soap__run__experiment) == "SAD" } {
		puts $outid "$atom_name       $natoms"
		puts $outid "    1"
		set output [format "%-2s%10.7f%10.7f" $atom_name $fp $fpp]
		puts $outid $output
	} else {
		puts $outid "     0"
	}

	close $outid

	if { $XMLParse(soap__run__experiment) == "SIR" } {		
		set outid [open "contents2" w]
		
		puts $outid "        $n_symops"
		puts $outid "$xray_source"
		puts $outid "        5"
		set output [format "%-2s%10i" "C" $num_c]
		puts $outid $output
		set output [format "%-2s%10i" "N" $num_n]
		puts $outid $output
		set output [format "%-2s%10i" "O" $num_o]
		puts $outid $output
		set output [format "%-2s%10i" "S" $num_s]
		puts $outid $output
		puts $outid "$atom_name       $natoms"
		puts $outid "    1"
		set output [format "%-2s%10.7f%10.7f" $atom_name $fp $fpp]
		puts $outid $output
		close $outid
	}
	
}

proc generate_diff_user { } {
	global XMLParse

	set outid [open "diff.user" w]

	set tmax $XMLParse(soap__crunch2__drear__tmax)
	set xmin $XMLParse(soap__crunch2__drear__xmin)
	set ymin $XMLParse(soap__crunch2__drear__ymin)
	set zmin $XMLParse(soap__crunch2__drear__zmin)
	set zmax $XMLParse(soap__crunch2__drear__zmax)

	puts $outid "     $tmax"
	puts $outid "     $xmin"
	puts $outid "     $ymin"
	puts $outid "     $zmin"
	puts $outid "     $zmax"

	close $outid
}

proc generate_scale_user { } {
	global XMLParse

	set outid [open "scale.user" w]

	set local_scaling $XMLParse(soap__crunch2__drear__local_scaling)
	set scaling_cutoff $XMLParse(soap__crunch2__drear__scaling_cutoff)

	if { $local_scaling == 1 } {
		puts $outid "YES"
		puts $outid "$scaling_cutoff"
		
	} else {
		puts $outid "NO"
	}

	close $outid

}

proc generate_slimits { } {
	global XMLParse

	set outid [open "slimits" w]

#	set low_res $XMLParse(dataset_info__low_res_limit)
#	set high_res $XMLParse(dataset_info__high_res_limit)

	set low_res 999.0
	set high_res [expr $XMLParse(soap__crunch2__max_resolution)]
#	set high_res [expr $XMLParse(soap__crunch2__max_resolution) / 2]

	puts $outid "$low_res $high_res"

	close $outid
	
	file copy "slimits" "slimits.copy"
}
					   


#
# Main Tcl code
#

# Set CRANK path and load utility functions
set crankbin "$env(CRANK)/bin"
source [file join $crankbin crankutils.tcl]

if { [lindex $argv 1] == "" } {
	puts "Usage:"
	puts "mtz2drear.tcl input.xml mtz.xml"
	puts " "
	puts " Where:"
	puts "   input.xml - Input .xml file containing commands and data"
	puts "   mtz.xml - Input .xml file containing crystal information"
	exit
}

# Load input XML file
set inputxml [lindex $argv 0]
if { [file exists $inputxml] == 0 } {
	crank_error "mtz2drear::inputxml file does not exist"
}
XMLParsefile $inputxml

# Load input XML mtz file
set mtzxml [lindex $argv 1]
if { [file exists $mtzxml] == 0 } {
	crank_error "mtz2drear::inputmtzxml file does not exist"
}
XMLParsefile $mtzxml

#
# Output all datastructures, for debugging purposes
# 
if { $debug >= 1 } {
	set allvars [array get XMLParse]
	for { set i 0 } { $i < [llength $allvars] } { incr i } {
		puts -nonewline "[lindex $allvars $i]\t"
		incr i
		puts "[lindex $allvars $i] "
	}
	exit
}

#
# Read symop.lib and determine spacegroup number
#
ReadSymops [FileJoin [GetEnvPath CLIBD] symop.lib] 1

set number $XMLParse(dataset_info__spacegroup__number)
  if { [set i [lsearch $crystal_data(SYM_NUMBER) $number]] >= 0 } {
	  set lattice [expr [lindex $crystal_data(SYM_LATTICE) $i ] + 1]
  }

#
# Generate the various files for blessing.sh 
#
generate_newinput $lattice
generate_contents
generate_diff_user
generate_scale_user
generate_slimits

#
# Run blessing.sh
#
#
if { $XMLParse(soap__run__experiment) == "SAD" } {
	set status [runcommand [concat $crankbin/blessing.sh crank SAD]]
} else {
	set status [runcommand [concat $crankbin/blessing.sh crank SIR]]
}