#!/bin/sh

#
# run-rantan.sh
#

# Generate substructures from F(+) and F(-) or SIR data using the
# CCP4 program RANTAN.
#
# Use the CCP4 program ECALC to convert F(+) and F(-) to EA values.
# With the use of different flags, one can change resolution
# (RESOLUTION), number of reflections in each shell (SHELL).  
# EXCLUDE CENTRIC is tured on by default, as is necessary for use with
# CRUNCH2
#
# Rag's emailed info:
# EXCLUDE CENTRIC, SHELL 500 and RESOLUTION x where x is the resolution of the
#  data in Angstrom + 0.5.
#

#
# Usage:
#
#  mtz2drear.sh input.mtz output.drear logfile f(+) sigf(+) f(-) sigf(-) shell resolution
#
# Where:
# $1 - Input .mtz file
# $2 - f(+) column name
# $3 - sigf(+) column name
# $4 - f(-) column name
# $5 - sigf(-) column name
# $6 - RESOLUTION - resolution cutoff
#

# Check for correct usage
if [ "$9" = "" ];
    then
    echo "Usage:"
    echo "run-rantan.sh input.mtz f(+) sigf(+) f(-) sigf(-) highres lowres threshold numpeaks"
	echo "Where:"
	echo "input.mtz    - Input .mtz file"
	echo "f(+)         - f(+) column name"
	echo "sigf(+)      - sigf(+) column name"
	echo "f(-)         - f(-) column name"
	echo "sigf(-)      - sigf(-) column name"
	echo "hires        - high resolution cutoff"
	echo "lowres       - low resolution cutoff"
	echo "threshold    - peak picking threshold"
	echo "numpeaks     - number of peaks to pick"
	echo ""
	echo "Steven Ness - (c) 2004"
    exit
fi

inputmtz="$1"
f_plus="$2"
sigf_plus="$3"
f_minus="$4"
sigf_minus="$5"
high_res="$6"
low_res="$7"
threshold="$8"
numpeaks="$9"

#
# Run ecalc to generate E and SIG
#
ecalc hklin $inputmtz hklout tmp1.mtz << END
TITLE calculate evalue from f+ f-
LABIN FP=$f_plus SIGFP=$sigf_plus FPH=$f_minus SIGFPH=$sigf_minus
EXCLUDE CENTRIC
#RESOLUTION $high_res 
END

#
# First, run RANTAN to find a set of phases that fits the tangent formula
#
rantan HKLIN tmp1.mtz HKLOUT tmp2.mtz <<EOF
TITLE RANTAN _ JIA _ TRY1
LABIN  EVAL=E
LABOUT  PHI1=PHASE1 WT1=WT1
NOUT 1
END
EOF

#
# Then, use FFT to make a map of that density
#
fft HKLIN tmp2.mtz MAPOUT map1.tmp <<EOF
TITLE RANTAN _ JIA _ TRY1
LABIN -
   F1=E SIG1=SIGE PHI=PHASE1 W=WT1
END
EOF

#
# Next, use MAPMASK to make a map of only the asymmetric unit
#
mapmask MAPIN map1.tmp MAPOUT map2.map <<EOF
SYMMETRY 'P 21 21 21'
XYZLIM ASU
EOF

#
# Run PEAKMAX to find the peaks in this map
#
peakmax MAPIN map2.map XYZOUT rantan.pdb XYZFRC rantan.ha<<EOF
THRESHHOLD RMS $threshold
NUMPEAKS $numpeaks
OUTPUT BROOKHAVEN FRAC
EOF


#
# Run COORDCONV to convert those coordinates into something we can compare
#
coordconv XYZOUT rantan.frac XYZIN rantan.ha <<EOF
INPUT HA
OUTPUT FRAC
EOF
