#!/bin/csh -f
#
##########################################################
##########################################################
#
# AVERAGING PROCEDURE USING SOLOMON
#
# Script to run several cycles of density modification.
# The intial .MTZ file must contain the Hendrickson Lattman
# coefficients, the native structure factors, best phases 
# and figures of merit.
#
# Note that all the intermediate files are deleted.
#
##########################################################
# 
# BEWARE! This script uses a calculation procedure based on
# bc. You should have a script in your path that looks like:
# 
# >> #! /bin/csh -f
# >> bc << end_calc
# >> $1
# >> end_calc
#
# 
##########################################################
#

# If it exists, source the .par file that has been previously
# generated by Crank

if ($1 != "") then
source $1
else
echo "Usage:"
echo ""
echo "solomon_crank.com input.par"
echo ""
echo "The .par file should be created by xml2solomonpar.tcl"
echo "and will contain lines to set the various variables"
echo "required by this script."
exit
endif

#
# Normally, you don't have to edit below this line
###################################################################################
#
set TotCyc=`calc "$PrepCycle + $ExtCycle + $EndCycle"`
if ($ExtCycle > 0) then
  set resinc=`calc "($endres - $strtres) / ($ExtCycle)"`
else
  set resinc=0
endif
set CurrentCycle=$LastCycle
set res=$strtres
set slvdens=0
#
#
if ($CurrentCycle >= $PrepCycle) then
  if ($CurrentCycle <= `calc "$PrepCycle + $ExtCycle"`) then
    set res=`calc "$strtres + (($CurrentCycle - $PrepCycle) * $resinc)"`
  else
    set res=$endres
  endif
endif
#
#
if (-w ${label}_Solomon.log) /bin/rm ${label}_Solomon.log
if (-w ${label}_stats.log) /bin/rm ${label}_stats.log
#
if (-w ${data}P1.mtz) then
  echo "using P1 data from previous run"
  goto FFT
endif

#
# sness - First, check to make sure that the input file is present
#
if ( ! -e ${data}.mtz ) then
	echo "Crank::run_solomon.com::Error the file ${data}.mtz does not exist.  Exitting"
	exit
endif

#
#
#
##########################################################
#  A) Use CAD to resort the data to P1
##########################################################
#
#

cad HKLIN1 ${data}.mtz HKLOUT ${data}P1.mtz << end-cad > /dev/null
OUTLIM SPACEGROUP P1
TITLE  ${data} extended to P1 cell
LABIN FILE 1 E1=${F} E2=${SIGF} E3=${PHIB} E4=${FOM} E5=${HLA} E6=${HLB} E7=${HLC} E8=${HLD}
CTYPE FILE 1 E1=F E2=Q E3=P E4=W E5=A E6=A E7=A E8=A
#LABOUT ALLIN
end
end-cad
#
##########################################################
#  B) Calculate map with best phases and FOM weighted 
#  amplitudes.
##########################################################
#
#
FFT:
if (-w ${label}_cycle${LastCycle}.map) then
  echo "using initial map from previous run"
  goto Begin
endif
#
fft HKLIN ${data}P1.mtz \
    MAPOUT ${label}_cycle${LastCycle}.map \
    << end-fft > /dev/null
LABIN F1=${F} SIG1=${SIGF} PHI=${PHIB} W=${FOM}
FFTSPACEGROUP 1
RESO 9999 $strtres
end
end-fft
#
Begin:
#
#
@ CurrentCycle=$LastCycle + 1
echo "Current Cycle = " ${CurrentCycle} 
#
#
##########################################################
#  1) Modify the density with Solomon
##########################################################
#
#
Solomon:
#
if ($CurrentCycle == $EndCycle) set mulsolv=0
#
solomon MAPIN ${label}_cycle${LastCycle}.map MAPOUT ${label}_cycle${CurrentCycle}av.map << end-solomon >> ${label}_Solomon.log
verbose
slvfrc $slvfrc
mulsolv auto $mulsolv
slvdens $slvdens
radius $radius
trunc $trunlo $trunhi
extrude
mapout
end

!Put any symmetry constraints here...

end-solomon
#
set slvdens=`grep "solvdens:" ${label}_Solomon.log|tail -1|awk '{print $3}'`
set ratio=`grep "ratio:" ${label}_Solomon.log|tail -1|awk '{print $2}'`
echo "solvent density =" $slvdens >> ${label}_stats.log
echo "ratio protein/solvent =" $ratio >> ${label}_stats.log
#
#
##########################################################
#  2) generate sfs and phases from this modified map
##########################################################
#
#
if ($CurrentCycle >= $PrepCycle) then
  if ($CurrentCycle <= `calc "$PrepCycle + $ExtCycle"`) then
    set res=`calc "$strtres + (($CurrentCycle - $PrepCycle) * $resinc)"`
  endif
endif
#
echo "Running sfall HKLIN ${data}P1.mtz MAPIN ${label}_cycle${CurrentCycle}av.map HKLOUT ${label}_mod_cycle${CurrentCycle}.mtz"
# sfall HKLIN ${data}P1.mtz \
#        MAPIN ${label}_cycle${CurrentCycle}av.map\
#        HKLOUT ${label}_mod_cycle${CurrentCycle}.mtz \
#        << end-sfrkall > ${label}_sfall.log
sfall HKLIN ${data}P1.mtz \
        MAPIN ${label}_cycle${CurrentCycle}av.map\
        HKLOUT ${label}_mod_cycle${CurrentCycle}.mtz \
        << end-sfrkall
TITLE   Sfs from density modification
MODE SFCALC MAPIN HKLIN
LABIN FP=${F}  SIGFP=${SIGF} F0=${PHIB} F1=${FOM} F2=${HLA} F3=${HLB} -
      F4=${HLC} F5=${HLD}
RESO 9999 $res
RSCB 6 $strtres
BINS 40
#SFSG 1
LABO   FC=FCmod PHIC=PHICmod
end
end-sfrkall
exit
#
/bin/rm ${label}_cycle${CurrentCycle}av.map
echo Cycle $CurrentCycle >> ${label}_stats.log
grep "Overall Reliability index is" ${label}_sfall.log \
						 >> ${label}_stats.log
#
#
##########################################################
#  3) combine phases
##########################################################
#
sigmaa HKLIN ${label}_mod_cycle${CurrentCycle}.mtz \
       HKLOUT ${label}_cycle${CurrentCycle}.mtz \
       << end-sigmaa > /dev/null
TITLE   Phase combination ${label}
LABI FP=${F}  SIGFP=${SIGF}                  -
     HLA=${HLA} HLB=${HLB} HLC=${HLC} HLD=${HLD} PHIBP=${PHIB}  WP=${FOM}     -
     FC=FCmod PHIC=PHICmod
LABO PHCMB=PHCMB1 WCMB=FOMCMB1 FWT=${FC} PHFWT=${PHIC}
RANGES 10 1000
RESO  9999 $res
ERROR
COMBINE PAR 1 
END
end-sigmaa
#
##########################################################
#  5) recalculate map.
##########################################################
#
fft HKLIN ${label}_cycle${CurrentCycle}.mtz \
    MAPOUT ${label}_cycle${CurrentCycle}.map \
    << end-fft > /dev/null
TITLE F1 15-$res A, after ${CurrentCycle} averaging cycles
LABI F1=${FC}  SIG1=${FC} PHI=${PHIC}
RESO 9999 $res
FFTSPACEGROUP 1
end
end-fft
#
#
#
#
#
if ($CurrentCycle > 1) then
  set oldc=`calc "$CurrentCycle - 1"`
  if (-w ${label}_cycle${CurrentCycle}.map) then 
  /bin/rm ${label}_cycle${oldc}.map
  endif
  if (-w ${label}_cycle${oldc}.mtz) then
  /bin/rm ${label}_cycle${oldc}.mtz
  /bin/rm ${label}_mod_cycle${oldc}.mtz
  endif
endif
@ LastCycle=$LastCycle + 1
#
#
if ( $CurrentCycle<$TotCyc) goto Begin
#

/bin/cp ${label}_cycle${CurrentCycle}.mtz solomon.out.mtz
