#!/usr/bin/env tclsh

#
# crank - Automated macromolecular structure solution
#
# Copyright (C) 2004 Steven Ness
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#

#
# runshelxc
#

#
# Run SHELXC to generate FA values from input XML
#

#
# Usage:
#
#   runshelxcxml.tcl code.in.user.xml code.in.mtz.xml
#

#
# Necessary input:
#   1) CRANK XML user datasetinfo file
#   2) CRANK XML mtz datasetinfo file
#

# Set CRANK path
set crankbin "$env(CRANK)/bin"
source [file join $crankbin crankutils.tcl]

#
# Main
#

# Check for usage information
if { [lindex $argv 2] == "" } {
	puts "Usage:"
	puts "   runshelxcxml.tcl code code.in.user.xml code.in.mtz.xml"
	puts " "
	puts " Necessary input:"
	puts "   1) CRANK XML user datasetinfo file"
	puts "   2) CRANK XML mtz datasetinfo file"
	puts ""
	puts " Copyright / GPL - Feb 2004 - Steven Ness"
	puts " "
	exit
}

#
# Setup variables from command line arguments
#
set code [lindex $argv 0]
set userxml [lindex $argv 1]
set mtzxml [lindex $argv 2]
set soapxml [lindex $argv 3]

#
# Check to make sure files exist
#
if { [file exists $userxml] == 0 } {
	crank_error "runshelxcxml.tcl::userxml file does not exist"
}

if { [file exists $mtzxml] == 0 } {
	crank_error "runshelxcxml.tcl::mtzxml file does not exist"
}

if { [file exists $soapxml] == 0 } {
	crank_error "runshelxcxml.tcl::soapxml file does not exist"
}

#
# Load all XML files
#
XMLParsefile $userxml
XMLParsefile $mtzxml
XMLParsefile $soapxml

#
# Generate SHELXC run files
#

puts "Starting runshelxcxml.tcl"
puts "--------------------------"

#
# Get current system time
#
set start_time [exec date]
puts "Start time = ($start_time)."

set status [runcommand [concat $crankbin/xml2shelxc.tcl $userxml $mtzxml $soapxml > shelxc-command.com]]
set status [runcommand [concat chmod ugo+x shelxc-command.com]]
set status [runcommand [concat shelxc-command.com > logfile]]
exit

#
# Run shelxc command file keeping output
#
puts "Running SHELXC."
set status [runcommand [concat shelxc-ccp4i $code try >& shelxc.run.output]]
if { $status != "SUCCESS" } {
	crank_error "runshelxcxml::newshelxc-ccp4 failed"
}

#
# Get current system time
#
set end_time [exec date]
puts "End time = ($end_time)."

#
# Now, process the output files and put everything
# into one big XML report file
#
puts "Processing output files and generating output."

# Determine cpu time
set shelxc_fileid [open "shelxc.run.output" r]
while { [gets $shelxc_fileid line] >= 0} {
	if { [regexp { *([0-9.]*)user *([0-9.]*)system ([0-9.:]*)} $line full user_time sys_time wall_time] } {
		break
	}
}

# Determine computer name
set computer_name [exec hostname]

#
# Output run data as XML
#
set runxml_fileid [open "$code.out.run.xml" w]
puts $runxml_fileid "<run>"
puts $runxml_fileid "\t<program>shelxc</program>"
puts $runxml_fileid "\t<start_time>$start_time</start_time>"
puts $runxml_fileid "\t<end_time>$end_time</end_time>"
puts $runxml_fileid "\t<cpu_time>"

if { [info exists user_time] } {
	puts $runxml_fileid "\t\t<user>$user_time</user>"
}

if { [info exists sys_time] } {
	puts $runxml_fileid "\t\t<sys>$sys_time</sys>"
}

if { [info exists wall_time] } {
	puts $runxml_fileid "\t\t<wall>$wall_time</wall>"
}

puts $runxml_fileid "\t</cpu_time>"
puts $runxml_fileid "\t<computer_name>$computer_name</computer_name>"
puts $runxml_fileid "</run>"
close $runxml_fileid

#
# Concatenate all output XML into a big XML report file
#
set output_fileid [open $code.out.crank.xml w] 
puts $output_fileid "<?xml version=\"1.0\" standalone=\"yes\"?>"
puts $output_fileid "<!-- CRANK XML --->\n"
close $output_fileid

set output [exec echo -e "\\074!-- CRANK input ---\\076" >> $code.out.crank.xml]
set output [exec cat $userxml | grep -v "xml version" >> $code.out.crank.xml]
set output [exec echo " " >> $code.out.crank.xml]

set output [exec cat $mtzxml | grep -v "xml version" >> $code.out.crank.xml]
set output [exec echo " " >> $code.out.crank.xml]

set output [exec echo -e "\\074!-- CRANK output ---\076\n" >> $code.out.crank.xml]
set output [exec cat $code.out.run.xml | grep -v "xml version" >> $code.out.crank.xml]
set output [exec echo " " >> $code.out.crank.xml]

set output [exec cat $code.out.coords.xml | grep -v "xml version" >> $code.out.crank.xml]
set output [exec echo " " >> $code.out.crank.xml]





