#!/usr/bin/env tclsh

#
# crank - Automated macromolecular structure solution
#
# Copyright (C) 2004 Steven Ness
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#

#
# shelxclog3graph.tcl
#
# Convert a SHELXC log file to "CRANK XML stats file"
#
#  This program collates the following information from
#  a SHELXC log file
#   1) FOM - All reflections
#   2) SDLU
#
# Look through the file until the end, where the string
#   "Script for further refinement" is found.
#   Then output the data to stdout in CRANK XML coordinate format
# However, a few lines before this "Script for further refinement"
#  line, there is a line that contains the figure of merit.  Save
#  this line as well.
#
#
# Usage: shelxclog2graph.tcl code shelxc.log
#

if { [lindex $argv 1] == "" } {
	puts "Usage:"
	puts "shelxclog2graph.tcl code shelxc.log"
	puts " "
	puts "Copyright / GPL - Feb 2004 - Steven Ness "
	puts " "
	exit
}

proc get_data_section { shelxclog_fileid } {

	# Process number of reflections line
	gets $shelxclog_fileid line

	if { ! [regexp { *([0-9.]*) Reflections read from ([A-Z0-9]*).*} $line full num_refl natname] } {
		puts "shelxclog2graph.tcl::Error in parsing reflections line"
		exit 1
	}

	# Process number of unique reflections line
	gets $shelxclog_fileid line

	if { ! [regexp { *([0-9.]*) Unique reflections.*} $line full num_unique ] } {
		puts "shelxclog2graph.tcl::Error in parsing unique reflections line"
		exit 1
	}

	# Process resolution bins line
	gets $shelxclog_fileid line
	if { ! [regexp { Resl.   Inf - ([0-9.]*) - ([0-9.]*) - ([0-9.]*) - ([0-9.]*) - ([0-9.]*) - ([0-9.]*) - ([0-9.]*) - ([0-9.]*) - ([0-9.]*) - ([0-9.]*) - ([0-9.]*) *} \
				$line full r1 r2 r3 r4 r5 r6 r7 r8 r9 r10 r11] } {
		puts "shelxclog2graph.tcl::Error in parsing resolution bins line"
		puts "line=$line"
		puts "full=$full"
		puts "vars=$r1 $r2 $r3 $r4 $r5 $r6 $r7 $r8 $r9 $r10"
		exit 1
	}

	# Process N(data) line
	gets $shelxclog_fileid line
	if { ! [regexp { N\(data\) *([0-9]*) *([0-9]*) *([0-9]*) *([0-9]*) *([0-9]*) *([0-9]*) *([0-9]*) *([0-9]*) *([0-9]*) *([0-9]*) *([0-9]*) *} \
				$line full n1 n2 n3 n4 n5 n6 n7 n8 n9 n10 n11] } {
		puts "shelxclog2graph.tcl::Error in parsing ndata line"
		puts "line=$line"
		puts "full=$full"
		puts "vars=$n1 $n2 $n3 $n4 $n5 $n6 $n7 $n8 $n9 $n10"
		exit 1
	}

	# Process <I/sig> line
	gets $shelxclog_fileid line
	if { ! [regexp { <I/sig> *([0-9.]*) *([0-9.]*) *([0-9.]*) *([0-9.]*) *([0-9.]*) *([0-9.]*) *([0-9.]*) *([0-9.]*) *([0-9.]*) *([0-9.]*) *([0-9.]*) *} \
				$line full i1 i2 i3 i4 i5 i6 i7 i8 i9 i10 i11] } {
		puts "shelxclog2graph.tcl::Error in parsing I/sig bins line"
		puts "line=$line"
		puts "full=$full"
		puts "vars=$i1 $i2 $i3 $i4 $i5 $i6 $i7 $i8 $i9 $i10 $i11"
		exit 1
	}

	# Process %Complete line
	gets $shelxclog_fileid line
	if { ! [regexp { %Complete *([0-9.]*) *([0-9.]*) *([0-9.]*) *([0-9.]*) *([0-9.]*) *([0-9.]*) *([0-9.]*) *([0-9.]*) *([0-9.]*) *([0-9.]*) *([0-9.]*) *} \
				$line full c1 c2 c3 c4 c5 c6 c7 c8 c9 c10 c11] } {
		puts "shelxclog2graph.tcl::Error in parsing complete bins line"
		puts "line=$line"
		puts "full=$full"
		puts "vars=$c1 $c2 $c3 $c4 $c5 $c6 $c7 $c8 $c9 $c10"
		exit 1
	}

	# Process <d"/sig> line
	gets $shelxclog_fileid line
	if { ! [regexp { <d\"/sig> *([0-9.]*) *([0-9.]*) *([0-9.]*) *([0-9.]*) *([0-9.]*) *([0-9.]*) *([0-9.]*) *([0-9.]*) *([0-9.]*) *([0-9.]*) *([0-9.]*) *} \
				$line full d1 d2 d3 d4 d5 d6 d7 d8 d9 d10 d11] } {
# 		puts "shelxclog2graph.tcl::Error in parsing d/sig bins line"
# 		puts "line=$line"
# 		puts "full=$full"
# 		puts "vars=$d1 $d2 $d3 $d4 $d5 $d6 $d7 $d8 $d9 $d10"
# 		exit 1
		set d1 0.0
		set d2 0.0
		set d3 0.0
		set d4 0.0
		set d5 0.0
		set d6 0.0
		set d7 0.0
		set d8 0.0
		set d9 0.0
		set d10 0.0
		set d11 0.0
	}

	if { [string compare $natname "NAT"] == 0 } {
		set native 1
	} else {
		set native 0
	}

	puts "\$TABLE: Shelxc log file information $natname :"
	puts "\$GRAPHS "
	puts ": N(data) :A:1,2:"
	puts ": <I/sig> :N:1,3:"
	puts ": %Complete :N:1,4:"
	if { ! $native } {
		puts ": <d\"/sig> :N:1,5:"
	}
	puts "\$\$"
	if { $native } {
		puts " resolution N(data) <I/sig> %Complete"
	} else {
		puts " resolution N(data) <I/sig> %Complete <d\"/sig>"
	}
	puts "\$\$"
	if { $native } {
		puts " resolution n i c"
	} {
		puts " resolution n i c d"
	}
	puts "\$\$"
	if { $native } {
		puts "$r1 $n1 $i1 $c1"
		puts "$r2 $n2 $i2 $c2"
		puts "$r3 $n3 $i3 $c3"
		puts "$r4 $n4 $i4 $c4"
		puts "$r5 $n5 $i5 $c5"
		puts "$r6 $n6 $i6 $c6"
		puts "$r7 $n7 $i7 $c7"
		puts "$r8 $n8 $i8 $c8"
		puts "$r9 $n9 $i9 $c9"
		puts "$r10 $n10 $i10 $c10"
		puts "$r11 $n11 $i11 $c11"
	} else {
		puts "$r1 $n1 $i1 $c1 $d1"
		puts "$r2 $n2 $i2 $c2 $d2"
		puts "$r3 $n3 $i3 $c3 $d3"
		puts "$r4 $n4 $i4 $c4 $d4"
		puts "$r5 $n5 $i5 $c5 $d5"
		puts "$r6 $n6 $i6 $c6 $d6"
		puts "$r7 $n7 $i7 $c7 $d7"
		puts "$r8 $n8 $i8 $c8 $d8"
		puts "$r9 $n9 $i9 $c9 $d9"
		puts "$r10 $n10 $i10 $c10 $d10"
		puts "$r11 $n11 $i11 $c11 $d11"
	}

	
	if { $native } {
		gets $shelxclog_fileid
		gets $shelxclog_fileid
	} else {
		gets $shelxclog_fileid
	}


}

# Process command line arguments
set trial_code [lindex $argv 0]
set shelxclog_filename [lindex $argv 1]

# # Open the shelxc logfile and output it all to stdout
# set shelxclog_fileid [open $shelxclog_filename r]
# while { [gets $shelxclog_fileid line] >= 0 } {
# 	puts $line
# }
# close $shelxclog_fileid

# Open the shelxc logfile and output it all to stdout
set shelxclog_fileid [open $shelxclog_filename r]

# Skip the first 7 lines
for { set i 0 } { $i < 7 } { incr i } {
	gets $shelxclog_fileid line
}

get_data_section $shelxclog_fileid
get_data_section $shelxclog_fileid
get_data_section $shelxclog_fileid
get_data_section $shelxclog_fileid



close $shelxclog_fileid

