#!/usr/bin/env tclsh

#
# crank - Automated macromolecular structure solution
#
# Copyright (C) 2004 Steven Ness
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#

#
# shelxclog3xmlstats.tcl
#
# Convert a SHELXC log file to "CRANK XML stats file"
#
#  This program collates the following information from
#  a SHELXC log file
#   1) FOM - All reflections
#   2) SDLU
#
# Look through the file until the end, where the string
#   "Script for further refinement" is found.
#   Then output the data to stdout in CRANK XML coordinate format
# However, a few lines before this "Script for further refinement"
#  line, there is a line that contains the figure of merit.  Save
#  this line as well.
#
#
# Usage: shelxclog2xmlstats.tcl code shelxc.log
#

#
# The output XML is in the following format:
#
# <shelxc_log>
#   <data_info>
#   <n_reflections></n_reflections>
#   <unique_reflections></unique_reflections>
#   <reflections_shelxd></reflections_shelxd>
#   <reflections_shelxe></reflections_shelxe>
#   <high_res></high_res>
#     <data>
#       <resolution>
#         <high></high>
#         <low></low>
#       </resolution>
#       <ndata></ndata>
#       <i_sigi></i_sigi>
#       <percent_complete></percent_complete>
#       <deltaf_sig></deltaf_sig>
#     </data>
#   </data_info>
#   <shelx_params>
#     <titl></titl>
#     <cell>
#       <wavelength></wavelength>
#       <a></a>
#       <b></b>
#       <c></c>
#       <alpha></alpha>
#       <beta></beta>
#       <gamma></gamma>
#     </cell>
#     <latt></latt>
#     <symm id=""></symm>
#     <sfac></sfac>
#     <unit></unit>
#     <shel>
#       <max></max>
#       <min></min>
#     </shel>
#     <pats></pats>
#     <find></find>
#     <mind></mind>
#     <dsul></dsul>
#     <ntry></ntry>
#     <seed></seed>
#     <hklf></hklf>
#   </shelx_params>
# </shelxc_log>
#

if { [lindex $argv 1] == "" } {
	puts "Usage:"
	puts "shelxclog2xmlstats.tcl code shelxc.log"
	puts " "
	puts "Copyright / GPL - Feb 2004 - Steven Ness "
	puts " "
	exit
}

# Process command line arguments
set trial_code [lindex $argv 0]
set shelxclog_filename [lindex $argv 1]

# Open the shelxc logfile
set shelxc_log_fileid [open $shelxclog_filename r]

# Skip the first 7 lines
for { set i 0 } { $i < 7 } { incr i } {
	gets $shelxc_log_fileid line
}

# Process number of reflections line
gets $shelxc_log_fileid line

if { ! [regexp { *([0-9.]*) Reflections read from.*} $line full num_refl ] } {
	puts "shelxclog2xmlstats.tcl::Error in parsing reflections line"
	exit 1
}

# Process number of unique reflections line
gets $shelxc_log_fileid line

if { ! [regexp { *([0-9.]*) Unique reflections, highest resolution *([0-9.]*).*} $line full num_unique high_res] } {
	puts "shelxclog2xmlstats.tcl::Error in parsing unique reflections line"
	exit 1
}

# Process resolution bins line
gets $shelxc_log_fileid line
if { ! [regexp { Resl.   Inf - ([0-9.]*) - ([0-9.]*) - ([0-9.]*) - ([0-9.]*) - ([0-9.]*) - ([0-9.]*) - ([0-9.]*) - ([0-9.]*) - ([0-9.]*) - ([0-9.]*) - ([0-9.]*) *} \
			$line full r1 r2 r3 r4 r5 r6 r7 r8 r9 r10 r11] } {
	puts "shelxclog2xmlstats.tcl::Error in parsing resolution bins line"
	puts "line=$line"
	puts "full=$full"
	puts "vars=$r1 $r2 $r3 $r4 $r5 $r6 $r7 $r8 $r9 $r10"
	exit 1
}

lappend r_list $r1
lappend r_list $r2
lappend r_list $r3
lappend r_list $r4
lappend r_list $r5
lappend r_list $r6
lappend r_list $r7
lappend r_list $r8
lappend r_list $r9
lappend r_list $r10
lappend r_list $r11

# Process N(data) line
gets $shelxc_log_fileid line
if { ! [regexp { N\(data\) *([0-9]*) *([0-9]*) *([0-9]*) *([0-9]*) *([0-9]*) *([0-9]*) *([0-9]*) *([0-9]*) *([0-9]*) *([0-9]*) *([0-9]*) *} \
			$line full n1 n2 n3 n4 n5 n6 n7 n8 n9 n10 n11] } {
	puts "shelxclog2xmlstats.tcl::Error in parsing ndata line"
	puts "line=$line"
	puts "full=$full"
	puts "vars=$n1 $n2 $n3 $n4 $n5 $n6 $n7 $n8 $n9 $n10"
	exit 1
}

lappend n_list $n1
lappend n_list $n2
lappend n_list $n3
lappend n_list $n4
lappend n_list $n5
lappend n_list $n6
lappend n_list $n7
lappend n_list $n8
lappend n_list $n9
lappend n_list $n10
lappend n_list $n11


# Process <I/sig> line
gets $shelxc_log_fileid line
if { ! [regexp { <I/sig> *([0-9.]*) *([0-9.]*) *([0-9.]*) *([0-9.]*) *([0-9.]*) *([0-9.]*) *([0-9.]*) *([0-9.]*) *([0-9.]*) *([0-9.]*) *([0-9.]*) *} \
			$line full i1 i2 i3 i4 i5 i6 i7 i8 i9 i10 i11] } {
	puts "shelxclog2xmlstats.tcl::Error in parsing I/sig bins line"
	puts "line=$line"
	puts "full=$full"
	puts "vars=$i1 $i2 $i3 $i4 $i5 $i6 $i7 $i8 $i9 $i10 $i11"
	exit 1
}

lappend i_list $i1
lappend i_list $i2
lappend i_list $i3
lappend i_list $i4
lappend i_list $i5
lappend i_list $i6
lappend i_list $i7
lappend i_list $i8
lappend i_list $i9
lappend i_list $i10
lappend i_list $i11


# Process %Complete line
gets $shelxc_log_fileid line
if { ! [regexp { %Complete *([0-9.]*) *([0-9.]*) *([0-9.]*) *([0-9.]*) *([0-9.]*) *([0-9.]*) *([0-9.]*) *([0-9.]*) *([0-9.]*) *([0-9.]*) *([0-9.]*) *} \
			$line full c1 c2 c3 c4 c5 c6 c7 c8 c9 c10 c11] } {
	puts "shelxclog2xmlstats.tcl::Error in parsing complete bins line"
	puts "line=$line"
	puts "full=$full"
	puts "vars=$c1 $c2 $c3 $c4 $c5 $c6 $c7 $c8 $c9 $c10"
	exit 1
}

lappend c_list $c1
lappend c_list $c2
lappend c_list $c3
lappend c_list $c4
lappend c_list $c5
lappend c_list $c6
lappend c_list $c7
lappend c_list $c8
lappend c_list $c9
lappend c_list $c10
lappend c_list $c11


# Process <d"/sig> line
gets $shelxc_log_fileid line
if { ! [regexp { <d\"/sig> *([0-9.]*) *([0-9.]*) *([0-9.]*) *([0-9.]*) *([0-9.]*) *([0-9.]*) *([0-9.]*) *([0-9.]*) *([0-9.]*) *([0-9.]*) *([0-9.]*) *} \
			$line full d1 d2 d3 d4 d5 d6 d7 d8 d9 d10 d11] } {
	puts "shelxclog2xmlstats.tcl::Error in parsing d/sig bins line"
	puts "line=$line"
	puts "full=$full"
	puts "vars=$d1 $d2 $d3 $d4 $d5 $d6 $d7 $d8 $d9 $d10"
	exit 1
}

lappend d_list $d1
lappend d_list $d2
lappend d_list $d3
lappend d_list $d4
lappend d_list $d5
lappend d_list $d6
lappend d_list $d7
lappend d_list $d8
lappend d_list $d9
lappend d_list $d10
lappend d_list $d11

# Process number of reflections line
gets $shelxc_log_fileid line
gets $shelxc_log_fileid line

if { ! [regexp { *([0-9.]*) Reflections.*} $line full num_out_fa_refl ] } {
	puts "shelxclog2xmlstats.tcl::Error in parsing SHELXD/E reflections line"
	exit 1
}

# Process number of reflections line
gets $shelxc_log_fileid line
gets $shelxc_log_fileid line

if { ! [regexp { *([0-9.]*) Reflections.*} $line full num_out_refl ] } {
	puts "shelxclog2xmlstats.tcl::Error in parsing SHELXE reflections line"
	exit 1
}

# Read three lines, until the start of the SHELX script
gets $shelxc_log_fileid line
gets $shelxc_log_fileid line
gets $shelxc_log_fileid line

# Read TITL line
gets $shelxc_log_fileid line
if { ! [regexp { *TITL(.*) *} $line full shelx_titl ] } {
	puts "shelxclog2xmlstats.tcl::Error in parsing TITL line"
	exit 1
}

#puts "TITL($shelx_titl)"

# Read CELL line
gets $shelxc_log_fileid line
if { ! [regexp { *CELL *([0-9.]*) *([0-9.]*) *([0-9.]*) *([0-9.]*) *([0-9.]*) *([0-9.]*) *([0-9.]*) *} $line full shelx_cell_wave shelx_cell_a shelx_cell_b shelx_cell_c shelx_cell_alpha shelx_cell_beta shelx_cell_gamma ] } {
	puts "shelxclog2xmlstats.tcl::Error in parsing CELL line"
	exit 1
}

#puts "CELL($shelx_cell)"

# Read LATT line
gets $shelxc_log_fileid line
if { ! [regexp { *LATT *(.*)} $line full shelx_latt ] } {
	puts "shelxclog2xmlstats.tcl::Error in parsing LATT line"
	exit 1
}

#puts "LATT($shelx_latt)"

# Read SYMM cards
set nsym_cards 0
set done 0

while { !$done } {
	gets $shelxc_log_fileid line
	if { ! [regexp { *SYMM(.*)} $line full card ] } {
		set done 1
		break
	}
	lappend symm_cards $card
	incr nsym_cards
}

# for { set i 0 } { $i < $nsym_cards } { incr i } {
# 	puts "SYMM([lindex $symm_cards $i])"
# }

# Read SFAC line
if { ! [regexp { *SFAC *(.*)} $line full shelx_sfac ] } {
	puts "shelxclog2xmlstats.tcl::Error in parsing SFAC line"
	exit 1
}

#puts "SFAC($shelx_sfac)"

# Read UNIT line
gets $shelxc_log_fileid line
if { ! [regexp { *UNIT *(.*)} $line full shelx_unit ] } {
	puts "shelxclog2xmlstats.tcl::Error in parsing UNIT line"
	exit 1
}

#puts "UNIT($shelx_unit)"

# Read SHEL line
gets $shelxc_log_fileid line
if { ! [regexp { *SHEL *([0-9.]*) *([0-9.]*) *} $line full shelx_shel_low shelx_shel_high ] } {
	puts "shelxclog2xmlstats.tcl::Error in parsing SHEL line"
	exit 1
}

#puts "SHEL($shelx_shel)"

# Read PATS line
gets $shelxc_log_fileid line
if { ! [regexp { *PATS *(.*)} $line full shelx_pats ] } {
	puts "shelxclog2xmlstats.tcl::Error in parsing PATS line"
	exit 1
}

#puts "PATS($shelx_pats)"

# Read FIND line
gets $shelxc_log_fileid line
if { ! [regexp { *FIND *(.*)} $line full shelx_find ] } {
	puts "shelxclog2xmlstats.tcl::Error in parsing FIND line"
	exit 1
}

#puts "FIND($shelx_find)"

# Read MIND line
gets $shelxc_log_fileid line
if { ! [regexp { *MIND *(.*)} $line full shelx_mind ] } {
	puts "shelxclog2xmlstats.tcl::Error in parsing MIND line"
	exit 1
}

#puts "MIND($shelx_mind)"

# Read NTRY line
gets $shelxc_log_fileid line
if { ! [regexp { *NTRY *(.*)} $line full shelx_ntry ] } {
	puts "shelxclog2xmlstats.tcl::Error in parsing NTRY line"
	exit 1
}

#puts "NTRY($shelx_ntry)"

# Read SEED line
gets $shelxc_log_fileid line
if { ! [regexp { *SEED *(.*)} $line full shelx_seed ] } {
	puts "shelxclog2xmlstats.tcl::Error in parsing SEED line"
	exit 1
}

#puts "SEED($shelx_seed)"

# Read HKLF line
gets $shelxc_log_fileid line
if { ! [regexp { *HKLF *(.*)} $line full shelx_hklf ] } {
	puts "shelxclog2xmlstats.tcl::Error in parsing HKLF line"
	exit 1
}

#puts "HKLF($shelx_hklf)"

# Read END line
gets $shelxc_log_fileid line
if { ! [regexp { *END *(.*)} $line full shelx_end ] } {
	puts "shelxclog2xmlstats.tcl::Error in parsing END line"
	exit 1
}

#puts "END($shelx_end)"


puts "<shelxc_log>"
puts "\t<data_info>"
puts "\t<n_reflections>$num_refl</n_reflections>"
puts "\t<unique_reflections>$num_unique</unique_reflections>"
puts "\t<reflections_shelxd>$num_out_fa_refl</reflections_shelxd>"
puts "\t<reflections_shelxe>$num_out_refl</reflections_shelxe>"
puts "\t<high_res>$high_res</high_res>"

for { set i 0 } { $i < 11 } { incr i } {
	puts "\t<data id=\"$i\">"
	puts "\t\t<resolution>"
	puts "\t\t\t<high>[lindex $r_list $i]</high>"
	puts "\t\t\t<low>[lindex $r_list $i]</low>"
	puts "\t\t</resolution>"
	puts "\t\t<ndata>[lindex $n_list $i]</ndata>"
	puts "\t\t<i_sigi>[lindex $i_list $i]</i_sigi>"
	puts "\t\t<percent_complete>[lindex $c_list $i]</percent_complete>"
	puts "\t\t<deltaf_sig>[lindex $d_list $i]</deltaf_sig>"
	puts "\t</data>"
}
puts "\t</data_info>"

puts "\t<shelx_run_params>"
puts "\t\t<titl>$shelx_titl</titl>"
puts "\t\t<cell>"
puts "\t\t\t<wavelength>$shelx_cell_wave</wavelength>"
puts "\t\t\t<a>$shelx_cell_a</a>"
puts "\t\t\t<b>$shelx_cell_b</b>"
puts "\t\t\t<c>$shelx_cell_c</c>"
puts "\t\t\t<alpha>$shelx_cell_alpha</alpha>"
puts "\t\t\t<beta>$shelx_cell_beta</beta>"
puts "\t\t\t<gamma>$shelx_cell_gamma</gamma>"
puts "\t\t</cell>"
puts "\t\t<latt>$shelx_latt</latt>"

for {set i 0} { $i < $nsym_cards } { incr i }  {
	puts "\t\t<symm id=\"$i\">[lindex $symm_cards $i]</symm>"
}

puts "\t\t<sfac>$shelx_sfac</sfac>"
puts "\t\t<unit>$shelx_unit</unit>"
puts "\t\t<shel>"
puts "\t\t\t<low>$shelx_shel_low</low>"
puts "\t\t\t<high>$shelx_shel_high</high>"
puts "\t\t</shel>"
puts "\t\t<pats>$shelx_pats</pats>"
puts "\t\t<find>$shelx_find</find>"
puts "\t\t<mind>$shelx_mind</mind>"
puts "\t\t<ntry>$shelx_ntry</ntry>"
puts "\t\t<seed>$shelx_seed</seed>"
puts "\t\t<hklf>$shelx_hklf</hklf>"
puts "\t</shelx_run_params>"
puts "</shelxc_log>"
