#!/usr/bin/env tclsh

#
# crank - Automated macromolecular structure solution
#
# Copyright (C) 2004 Steven Ness
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#

#
# xml2bp3com.tcl
#
# Tcl program to take the *.user.xml and *.mtz.xml files,
# the "CRANK XML dataset info" files, and one 
# trial.coords.xml file and create a BP3 com file
#
#
# Usage:
#  xml2bp3com.tcl user.xml mtz.xml trial.coords.xml
#

# Set CRANK path
set crankbin "$env(CRANK)/bin"
source [file join $crankbin crankutils.tcl]

#
# Output a CRUNCH2 newinput file corresponding to the *.user.xml,
# *.mtz.xml and *.coords.xml XML files read in.  
#
proc OutputBP3comfile { } {
	global XMLParse

	# Find maximum occupancy, we will scale all the other occupancies by it
	# to normalize them.
	set i 0
	set max_occu 0.0
	while { [info exists XMLParse(coordinate_data__atom__x__$i)] } {
		if { $XMLParse(coordinate_data__atom__occupancy__$i) > $max_occu } {
			set max_occu $XMLParse(coordinate_data__atom__occupancy__$i)
		}
		incr i
	}


	puts "\#!/bin/sh"
	puts " "

	set bp3_path "$XMLParse(soap__bp3__program)"

	set mtz_in "../mtz/input.mtz"
	set mtz_out "bp3.out.mtz"

	puts "$bp3_path HKLIN $mtz_in HKLOUT $mtz_out << END"
	puts " "

	# 
	# If running anything but SAD output Native crystal first
	# 
	if { [string tolower $XMLParse(soap__run__experiment)] == "sir" } {
		puts "xtal NATIVE"
		puts "dname native"
		set f_nat $XMLParse(soap__run__mtz_in__data__f__0)
		set sigf_nat $XMLParse(soap__run__mtz_in__data__sigf__0)
		puts "column F=$f_nat SF=$sigf_nat"
		puts ""
	}

	if { [string tolower $XMLParse(soap__run__experiment)] == "siras" } {
		puts "xtal NATIVE"
		puts "dname native"
		set f_nat $XMLParse(soap__run__mtz_in__data__f__0)
		set sigf_nat $XMLParse(soap__run__mtz_in__data__sigf__0)
		puts "column F=$f_nat SF=$sigf_nat"
		puts ""
	}

	if { [string tolower $XMLParse(soap__run__experiment)] == "mad" } {
		set native_column 0
		for { set i 0 } { $i < $XMLParse(soap__run__mtz_in__n_datasets) } { incr i } {
			if { [string compare [string tolower $XMLParse(soap__run__mtz_in__data__type__$i)] "native"] == 0 } {
				set native_column $i
			}
		}

		if { $native_column != 0 } {
			puts "xtal NATIVE"
			puts "dname NATIVE"
			set f_nat $XMLParse(soap__run__mtz_in__data__f__$native_column)
			set sigf_nat $XMLParse(soap__run__mtz_in__data__sigf__$native_column)
			puts "column F=$f_nat SF=$sigf_nat"
			puts ""
		}
	}

	if { [string tolower $XMLParse(soap__run__experiment)] == "sad" } {
		puts "xtal CRYSTAL1"
	} else {
		puts "xtal DERIVATIVE"
	}

	
	# Output all atoms
	set i 0
	while { [info exists XMLParse(coordinate_data__atom__x__$i)] } {
		# Atom name
		puts "atom $XMLParse(coordinate_data__atom__element_name__$i)"

		# Coordinates
		puts -nonewline "xyz "
		puts -nonewline "$XMLParse(coordinate_data__atom__x__$i) "
		puts -nonewline "$XMLParse(coordinate_data__atom__y__$i) "
		puts -nonewline "$XMLParse(coordinate_data__atom__z__$i)"

		if { [info exists XMLParse(soap__bp3__mode)] } {
			if { $XMLParse(soap__bp3__mode) == "refine_sdlu_occu" } {
				puts " NOREF X Y Z"
			} else {
				puts ""
			}
		} else {
			puts ""
		}

		# Occupancy
		set occupancy_modifier 1.0

		if { [string compare [string tolower $XMLParse(soap__run__experiment)] "sad"] == 0 } {
			set occupancy_modifier $XMLParse(soap__bp3__max_sad_occupancy)
		}

		if { [string compare [string tolower $XMLParse(soap__run__experiment)] "sir"] == 0 } {
			set occupancy_modifier $XMLParse(soap__bp3__max_sir_occupancy)
		}

		if { [string compare [string tolower $XMLParse(soap__run__experiment)] "mad"] == 0 } {
			set occupancy_modifier $XMLParse(soap__bp3__max_mad_occupancy)
		}

		if { [string compare [string tolower $XMLParse(soap__run__experiment)] "mir"] == 0 } {
			set occupancy_modifier $XMLParse(soap__bp3__max_mir_occupancy)
		}

		set occupancy [expr $XMLParse(coordinate_data__atom__occupancy__$i) / $max_occu]
		set occupancy [expr $occupancy * $occupancy_modifier]

		set output [format "occu %10.3f" $occupancy]
		puts $output

		# B-factor
		puts -nonewline "biso $XMLParse(dataset_info__bfactor)"

		puts ""

		incr i
	}

	if { [string tolower $XMLParse(soap__run__experiment)] == "sad" } {
 		puts "dname sad"
		set i 0
	} 

	if { [string tolower $XMLParse(soap__run__experiment)] == "sir" } {
		puts "dname derivative"
		set i 1
	}

	if { [string tolower $XMLParse(soap__run__experiment)] == "siras" } {
		puts "dname derivative"
		set i 1
	}

	if { [string compare [string tolower $XMLParse(soap__run__experiment)] "mad" ] != 0 } {

		if { $XMLParse(soap__run__mtz_in__data__anomalous__$i) == 0 } {
			# Columns
			set f $XMLParse(soap__run__mtz_in__data__f__$i)
			set sigf $XMLParse(soap__run__mtz_in__data__sigf__$i)
			puts " column F=$f SF=$sigf"
			
		} else {
			# Columns
			set fplus $XMLParse(soap__run__mtz_in__data__f_plus__$i)
			set sigfplus $XMLParse(soap__run__mtz_in__data__sigf_plus__$i)
			set fminus $XMLParse(soap__run__mtz_in__data__f_minus__$i)
			set sigfminus $XMLParse(soap__run__mtz_in__data__sigf_minus__$i)
			puts " column F+=$fplus SF+=$sigfplus F-=$fminus SF-=$sigfminus"
			# Atom form factors
			if { [info exists XMLParse(soap__run__mtz_in__data__fprime__$i)] } {
				set atomname $XMLParse(dataset_info__atom_name)
				set fprime $XMLParse(soap__run__mtz_in__data__fprime__$i)
				set fprimeprime $XMLParse(soap__run__mtz_in__data__fprimeprime__$i)
			}
		}

		if { [info exists XMLParse(soap__run__mtz_in__data__fprime__$i)] } {
			set atomname $XMLParse(dataset_info__atom_name)
			set fprime $XMLParse(soap__run__mtz_in__data__fprime__$i)
			set fprimeprime $XMLParse(soap__run__mtz_in__data__fprimeprime__$i)
			puts " form $atomname FP=$fprime FPP=$fprimeprime"
		}
	} else {
		#
		# Doing MAD
		#
		for { set i 0 } { $i < $XMLParse(soap__run__mtz_in__n_datasets) } { incr i } {
			# If it's a native dataset, we've already handled it
			if { [string compare [string tolower $XMLParse(soap__run__mtz_in__data__type__$i)] "native"] != 0 } {
				# Dataset name
				puts "dname $XMLParse(soap__run__mtz_in__data__type__$i)_$i"

				if { $XMLParse(soap__run__mtz_in__data__anomalous__$i) == 0 } {
					# Columns
					set f $XMLParse(soap__run__mtz_in__data__f__$i)
					set sigf $XMLParse(soap__run__mtz_in__data__sigf__$i)
					puts " column F=$f SF=$sigf"
					
				} else {
					# Columns
					set fplus $XMLParse(soap__run__mtz_in__data__f_plus__$i)
					set sigfplus $XMLParse(soap__run__mtz_in__data__sigf_plus__$i)
					set fminus $XMLParse(soap__run__mtz_in__data__f_minus__$i)
					set sigfminus $XMLParse(soap__run__mtz_in__data__sigf_minus__$i)
					puts " column F+=$fplus SF+=$sigfplus F-=$fminus SF-=$sigfminus"
					# Atom form factors
					if { [info exists XMLParse(soap__run__mtz_in__data__fprime__$i)] } {
						set atomname $XMLParse(dataset_info__atom_name)
						set fprime $XMLParse(soap__run__mtz_in__data__fprime__$i)
						set fprimeprime $XMLParse(soap__run__mtz_in__data__fprimeprime__$i)
					}
				}
				if { [info exists XMLParse(soap__run__mtz_in__data__fprime__$i)] } {
					set atomname $XMLParse(dataset_info__atom_name)
					set fprime $XMLParse(soap__run__mtz_in__data__fprime__$i)
					set fprimeprime $XMLParse(soap__run__mtz_in__data__fprimeprime__$i)
					puts " form $atomname FP=$fprime FPP=$fprimeprime"
				}
				
			}
		}
	}

# 	# Put dataset info
# 	puts "dname data1"
	
# 	# Put column info
# 	set fplus "$XMLParse(soap__bp3__mtz_in__fplus)"
# 	set sigfplus "$XMLParse(soap__bp3__mtz_in__sigfplus)"
# 	set fminus "$XMLParse(soap__bp3__mtz_in__fminus)"
# 	set sigfminus "$XMLParse(soap__bp3__mtz_in__sigfminus)"

# 	puts "column F+=$fplus SF+=$sigfplus F-=$fminus SF-=$sigfminus"

# 	# Output atomic form factors
# 	set atomtype "$XMLParse(coordinate_data__atom__element_name__0)"
# 	set fprime "$XMLParse(dataset_info__f_prime)"
# 	set fdoubleprime "$XMLParse(dataset_info__f_double_prime)"

# 	puts "form $atomtype FP=$fprime FPP=$fdoubleprime"

	
	#
	# BP3 Options
	#
	if { [info exists XMLParse(soap__bp3__refall)] } {
		if { $XMLParse(soap__bp3__refall) == "yes" } {
			puts "refa"
		}
	}

	if { [info exists XMLParse(soap__bp3__shel)] } {
		if { $XMLParse(soap__bp3__shel) == "yes" } {
			puts "shel"
		}
	}

	if { [info exists XMLParse(soap__bp3__bins)] } {
		if { $XMLParse(soap__bp3__bins) != "0" } {
			puts "bins $XMLParse(soap__bp3__bins)"
		}
	}

	if { [info exists XMLParse(soap__bp3__ncycles)] } {
		if { $XMLParse(soap__bp3__ncycles) != "0" } {
			puts "cycle $XMLParse(soap__bp3__ncycles)"
		}
	}

	if { [info exists XMLParse(soap__bp3__phase)] } {
		if { $XMLParse(soap__bp3__phase) != "0" } {
			puts "phase"
		}
	}

	if { [info exists XMLParse(soap__bp3__output_pdb)] } {
		if { $XMLParse(soap__bp3__output_pdb) != "" } {
			puts "pdbout"
		}
	}

	puts "END"

}


#
# Main
#
if { [lindex $argv 2] == "" } {
	puts "Usage:"
	puts "  xml2bp3com.tcl input.xml mtz.xml coords.xml"
	puts " "
	puts " Copyright/GPL Feb 2004 Steven Ness"
	exit
}

#
# Setup variable names
#
set inputxml [lindex $argv 0]
set mtzxml [lindex $argv 1]
set coordsxml [lindex $argv 2]

#
# Check to make sure files exist
#
if { [file exists $inputxml] == 0 } {
	crank_error "xml2bp3com.tcl::inputxml file does not exist"
}
if { [file exists $mtzxml] == 0 } {
	crank_error "xml2bp3com.tcl::mtzxml file does not exist"
}
if { [file exists $coordsxml] == 0 } {
	crank_error "xml2bp3com.tcl::coordsxml file does not exist"
}

#
# Read in XML files
#
XMLParsefile $inputxml
XMLParsefile $mtzxml
XMLParsefile $coordsxml

#
# Generate .com file
#
OutputBP3comfile
