#!/usr/bin/env tclsh

#
# crank - Automated macromolecular structure solution
#
# Copyright (C) 2004 Steven Ness
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#

#
# xml2crunch2newinput.tcl
#
# Tcl program to take the *.user.xml and *.mtz.xml files,
# the "CRANK XML dataset info" files, and create a CRUNCH2
# newinput file
#
# Usage:
#  xml2crunch2newinput.tcl user.xml mtz.xml
#

# Set CRANK path
set crankbin "$env(CRANK)/bin"
source [file join $crankbin crankutils.tcl]

#
# Output a CRUNCH2 newinput file corresponding to the 
#  *.user.xml and *.mtz.xml XML files read in.
#
# The business end of this program, the rest of the code 
# mainly just reads in and processes the XML.
#
# Pseudocode:
#  Output description + code
#  Output cell parameters
#  Output # of atoms to look for
#  Output scattering power
#  Output spacegroup
#  Output # of symmetry operators
#  Output symmetry operator cards
#
# Example newinput file:
#
# Test newcrunch2 - caufd
# 33.95000  33.95000  74.82000  90.00000  90.00000  90.00000
#    8
#    15
#     P 43 21 2
#     8
#   1  0  0  0.0000000  0  1  0  0.0000000  0  0  1  0.0000000
#   0 -1  0  0.5000000  1  0  0  0.5000000  0  0  1  0.7500000
#  -1  0  0  0.0000000  0 -1  0  0.0000000  0  0  1  0.5000000
#   0  1  0  0.5000000 -1  0  0  0.5000000  0  0  1  0.2500000
#  -1  0  0  0.5000000  0  1  0  0.5000000  0  0 -1  0.7500000
#   0  1  0  0.0000000  1  0  0  0.0000000  0  0 -1  0.0000000
#   1  0  0  0.5000000  0 -1  0  0.5000000  0  0 -1  0.2500000
#   0 -1  0  0.0000000 -1  0  0  0.0000000  0  0 -1  0.5000000
#
proc OutputCRUNCH2newinput { } {
	global XMLParse

	#  Output description + code
	puts "$XMLParse(dataset_info__description) $XMLParse(dataset_info__code)"

	#  Output cell parameters
	set output [format "%-10.5f%-10.5f%-10.5f%-10.5f%-10.5f%-8.5f" \
					$XMLParse(dataset_info__cell__cell_a) \
					$XMLParse(dataset_info__cell__cell_b) \
					$XMLParse(dataset_info__cell__cell_c) \
					$XMLParse(dataset_info__cell__cell_alpha) \
					$XMLParse(dataset_info__cell__cell_beta) \
					$XMLParse(dataset_info__cell__cell_gamma) ]
	puts $output

	#  Output # of atoms to look for
	if { $XMLParse(dataset_info__natoms) != "" } {
		set output [format "%5i" $XMLParse(dataset_info__natoms)]
		puts $output
	} else {
		puts "Could not find dataset_info__natoms - Required data item"
		puts "Exitting"
		exit
	}

	#  Output scattering power
	if { $XMLParse(soap__crunch2__scattering_power) != "" } {
		#set val $XMLParse(soap__crunch2__scattering_power)
		#set rounded [expr round($val)] 
		set output [format "%5i" [expr round($XMLParse(soap__crunch2__scattering_power))]]
		puts $output
	} else {
		set output [format "%5i" 15]
		puts $output
	}

	#  Output spacegroup
	puts -nonewline "    $XMLParse(dataset_info__spacegroup__lattice)"
	set i 0
	while { [info exists XMLParse(dataset_info__spacegroup__operator__$i)] } {
		puts -nonewline " $XMLParse(dataset_info__spacegroup__operator__$i)"
		incr i
	}
	puts ""

	#  Output # of symmetry operators
	set output [format "%5i" $XMLParse(dataset_info__n_symops)]
	puts $output

	#  Output symmetry operator cards
	set i 0
	while { [info exists XMLParse(dataset_info__symmetry_operator__aa__$i)] } {
		set output [format "%3.0f%3.0f%3.0f %10.7f" \
						$XMLParse(dataset_info__symmetry_operator__aa__$i) \
						$XMLParse(dataset_info__symmetry_operator__ab__$i) \
						$XMLParse(dataset_info__symmetry_operator__ac__$i) \
						$XMLParse(dataset_info__symmetry_operator__atrans__$i)]
		puts -nonewline $output

		set output [format "%3.0f%3.0f%3.0f %10.7f" \
						$XMLParse(dataset_info__symmetry_operator__ba__$i) \
						$XMLParse(dataset_info__symmetry_operator__bb__$i) \
						$XMLParse(dataset_info__symmetry_operator__bc__$i) \
						$XMLParse(dataset_info__symmetry_operator__btrans__$i)]
		puts -nonewline $output

		set output [format "%3.0f%3.0f%3.0f %10.7f" \
						$XMLParse(dataset_info__symmetry_operator__ca__$i) \
						$XMLParse(dataset_info__symmetry_operator__cb__$i) \
						$XMLParse(dataset_info__symmetry_operator__cc__$i) \
						$XMLParse(dataset_info__symmetry_operator__ctrans__$i)]
		puts $output

		incr i
	}
	puts ""

}


#
# Main
#
if { [lindex $argv 1] == "" } {
	puts "Usage:"
	puts "  xml2crunch2newinput.tcl user.xml mtz.xml"
	puts " "
	puts " Copyright/GPL Feb 2004 Steven Ness"
}

#
# Setup variable names
#
set userxml [lindex $argv 0]
set mtzxml [lindex $argv 1]

#
# Check to make sure files exist
#
if { [file exists $userxml] == 0 } {
	crank_error "xml2crunch2newinput.tcl::userxml file does not exist"
}

if { [file exists $mtzxml] == 0 } {
	crank_error "xml2crunch2newinput.tcl::mtzxml file does not exist"
}

XMLParsefile $userxml
XMLParsefile $mtzxml

OutputCRUNCH2newinput

