#!/usr/bin/env tclsh

#
# crank - Automated macromolecular structure solution
#
# Copyright (C) 2004 Steven Ness
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#

#
# Tcl program to take the *.user.xml and *.mtz.xml files,
# the "CRANK XML dataset info" files, and create a PMF
# crunch.use file
#

#
# Usage
#  xml2pmf.use.tcl code.soap.xml
#

# Set CRANK path
set crankbin "$env(CRANK)/bin"
source [file join $crankbin crankutils.tcl]

#
# Output a PMF pmf.use file corresponding to the 
#   pmf.soap.in.xml XML file
#
# The business end of this program, the rest of the code 
# mainly just reads in and processes the XML.
#
# Example pmf.use
#  (Maximum resolution 4.0A, 10 trials)
#
# ISTLM  15
# NCOOR  10
# NPATT  1
#
#
#     This file contains the defaults for the parameters governing a
#     run of PMF. The user may prepare a file named pmf.use
#     containing alternative values for the variables listed below.
#     Only the variables that need to be changed need be given.
#     Lines starting with a '#' are treated as comment.
#
#NBEGIN   1                   Default trials 1 to 20 are done
# NEIND  20
# IELIM   0                   Default all E's are used
#  NOTF   1                   NOTF high: zero reflections are important
#ISTLM   40                   ISTLM is 10 times the resolution limit used
#NCOOR    3                   NCOOR/10 is the fraction of the number of
#                            atoms which is present in the starting set
#NPATT    1                   Defines type of patterson: 1 gives F^2, 2
#                            F*E and 3 E^2


proc OutputPMF.use { } {
	global XMLParse

	# Output PMF maximum resolution limits
	if { [info exists XMLParse(soap__crunch2__pmf__max_resolution)] &&
		 $XMLParse(soap__crunch2__pmf__max_resolution) != ""  } {
		puts "ISTLM    [expr $XMLParse(soap__crunch2__pmf__max_resolution) * 10.0]"
	} 

	# Output CRUNCH2 start trial
	if { [info exists XMLParse(soap__crunch2__ntrials)] &&
		 $XMLParse(soap__crunch2__ntrials) != "" } {
		puts "NBEGIN    1"
		puts "NEIND     $XMLParse(soap__crunch2__ntrials)"
	} else {
		puts "NBEGIN    1"
		puts "NEIND     10"
	}

	#
	# Output PMF fraction of coordinates to look for. NCOOR/10 is the
	# fraction of the number of atoms in the substructure. 
	#
	#
	# If the value for soap__crunch2__pmf__ncoor is specified, we use
	# it, with the assumption that it was set for a certain reason.
	# 
	# If it wasn't set in the HTML, then, based on the number of atoms
	# we are looking for, output a reasonable value for NCOOR.  The
	# simple algorithm we use is:
	#  
	# <= 4 atoms -> NCOOR 10
	# 5-10 atoms -> NCOOR 5
	# 10-20 atoms -> NCOOR 3
	# 20+ atoms -> NCOOR 2
	#
	if { [info exists XMLParse(soap__crunch2__pmf__ncoor)] } {
		if { $XMLParse(soap__crunch2__pmf__ncoor) != "" } {
			set ncoor $XMLParse(soap__crunch2__pmf__ncoor)
		}
	}

	if { ![info exists ncoor] } {
		set natoms $XMLParse(dataset_info__natoms)
		if { $natoms <= 4 } { 
			set ncoor 10
		} elseif { $natoms <= 10 } {
			set ncoor 5
		} elseif { $natoms <= 20 } {
			set ncoor 3
		} else {
			set ncoor 2
		}
	}

	puts "NCOOR     $ncoor"

	# Output PMF patterson sharpening
	if { [info exists XMLParse(soap__crunch2__pmf__npatt)] } {
		if { $XMLParse(soap__crunch2__pmf__npatt) != ""  } {
			set npatt $XMLParse(soap__crunch2__pmf__npatt)
		} 
	}

	if { ![info exists npatt] } {
		set npatt 1
	}

	puts "NPATT     $npatt"
}


#
# Main
#
if { [lindex $argv 0] == "" } {
	puts "Usage:"
	puts "  xml2pmfcrunch.use.tcl code.soap.xml"
	puts " "
	puts " Copyright/GPL Feb 2004 Steven Ness"
}

set xmlfile_soap [lindex $argv 0]

XMLParsefile $xmlfile_soap

OutputPMF.use




