#!/usr/bin/env tclsh

#
# crank - Automated macromolecular structure solution
#
# Copyright (C) 2004 Steven Ness
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#

#
# xml2shelx.tcl
#
# Tcl program to create the input file for SHELX.
#
# Example SHELX input for SAD data is:
#
#CELL 58.036 58.036 151.29 90 90 90
#SAD ../thau-nat.hkl
#SPAG P41212
#FIND 9
#DSUL 8
#MIND -3.5
#NTRY 100
#
# Usage:
#  xml2shelx.tcl user.xml mtz.xml
#

# Set CRANK path
set crankbin "$env(CRANK)/bin"
source [file join $crankbin crankutils.tcl]

#
# Output a CRUNCH2 newinput file corresponding to the 
#  *.user.xml and *.mtz.xml XML files read in.
#
# The business end of this program, the rest of the code 
# mainly just reads in and processes the XML.
#
# Pseudocode:
# 1) Output CELL information (CELL)
# 2) Output SAD datafile location (SAD)
# 3) Output Spacegroup, no spaces between operators (SPAG)
# 3) Output Number of atoms to find (FIND)
# 4) Output Number of atoms participating in disulphide bonds (DSUL)
# 5) Output Minimum distance between atoms (MIND)
#
# CRANK SHELX XML
#
#<dataset_info>
#   <code></code>
#   <atom_name></atom_name>
#   <natoms></natoms>
#   <high_res_limit></high_res_limit>
#   <scalepack_file></scalepack_file>
#</dataset_info>
#<shelx>
#   <dsul></dsul>
#   <mind></mind>
#   <ntry></ntry>
#</shelx>
#
#CELL 58.036 58.036 151.29 90 90 90
#SAD thau-sad.sca
#SPAG P41212
#FIND 9
#DSUL 8
#MIND -3.5
#NTRY 100
#
proc OutputShelx { } {
	global XMLParse

	puts "shelxc $XMLParse(dataset_info__code) <<EOF"

	#  Output cell parameters
	set output [format "CELL %-10.5f%-10.5f%-10.5f%-10.5f%-10.5f%-8.5f" \
					$XMLParse(dataset_info__cell__cell_a) \
					$XMLParse(dataset_info__cell__cell_b) \
					$XMLParse(dataset_info__cell__cell_c) \
					$XMLParse(dataset_info__cell__cell_alpha) \
					$XMLParse(dataset_info__cell__cell_beta) \
					$XMLParse(dataset_info__cell__cell_gamma) ]
	puts $output

	if { [string compare [string tolower $XMLParse(soap__run__experiment)] sad] == 0 } {
		 puts "SAD ../sca/$XMLParse(dataset_info__code)_$XMLParse(soap__run__mtz_in__data__type__0)_0.sca"
	}

	if { [string compare [string tolower $XMLParse(soap__run__experiment)] sir] == 0 } {
		 puts "NAT ../sca/$XMLParse(dataset_info__code)_$XMLParse(soap__run__mtz_in__data__type__0)_0.sca"
		 puts "SIR ../sca/$XMLParse(dataset_info__code)_$XMLParse(soap__run__mtz_in__data__type__1)_1.sca"
	}

	if { [string compare [string tolower $XMLParse(soap__run__experiment)] siras] == 0 } {
		 puts "NAT ../sca/$XMLParse(dataset_info__code)_$XMLParse(soap__run__mtz_in__data__type__0)_0.sca"
		 puts "SIRA ../sca/$XMLParse(dataset_info__code)_$XMLParse(soap__run__mtz_in__data__type__1)_1.sca"
	}

	if { [string compare [string tolower $XMLParse(soap__run__experiment)] mad] == 0 } {
		for { set i 0 } { $i < $XMLParse(soap__run__mtz_in__n_datasets) } { incr i } {
			set type $XMLParse(soap__run__mtz_in__data__type__$i)
			set filename "../sca/$XMLParse(dataset_info__code)_$XMLParse(soap__run__mtz_in__data__type__$i)_$i.sca"

			if { [string compare $type NATIVE] == 0 } {
				set type "NAT"
			}

			puts "$type $filename"
		}
	}

	#  Output spacegroup
	if { [info exists XMLParse(dataset_info__spacegroup__operator__0)] } {
		set operator_0 $XMLParse(dataset_info__spacegroup__operator__0)
	} else {
		set operator_0 ""
	}

	if { [info exists XMLParse(dataset_info__spacegroup__operator__1)] } {
		set operator_1 $XMLParse(dataset_info__spacegroup__operator__1)
	} else {
		set operator_1 ""
	}

	if { [info exists XMLParse(dataset_info__spacegroup__operator__2)] } {
		set operator_2 $XMLParse(dataset_info__spacegroup__operator__2)
	} else {
		set operator_2 ""
	}
	set spacegroup "$XMLParse(dataset_info__spacegroup__lattice)${operator_0}${operator_1}${operator_2}"
	
	puts "SPAG $spacegroup"

	puts "FIND $XMLParse(dataset_info__natoms)"
	puts "MIND -[expr abs($XMLParse(soap__shelx__mind))]"
	puts "NTRY $XMLParse(soap__shelx__ntry)"
	puts "SHEL $XMLParse(soap__shelx__shel__max) $XMLParse(soap__shelx__shel__min)"
	puts "SFAC $XMLParse(dataset_info__atom_name)"

	if { [info exists XMLParse(soap__shelx__dsul)] } {
		if { $XMLParse(soap__shelx__dsul) > 0 } {
			puts "DSUL $XMLParse(soap__shelx__dsul)"
		}
	}

	if { [info exists XMLParse(soap__shelx__pats)] } {
		set tmpstr [string toupper $XMLParse(soap__shelx__pats)]
		if { [string compare $tmpstr "ON" ] == 0 } {
			puts "PATS"
		}
	}

	if { [info exists XMLParse(soap__shelx__weed)] } {
		if { $XMLParse(soap__shelx__dsul) > 0 } {
			puts "WEED $XMLParse(soap__shelx__weed)"
		}
	}
#	puts "HKLF 3"
	puts "EOF"

	if { [info exists XMLParse(soap__shelx__run_shelxd)] } {
		if { [string compare $XMLParse(soap__shelx__run_shelxd) "1"] == 0 } {
			puts "shelxd $XMLParse(dataset_info__code)_fa"
		}
	}

	if { [info exists XMLParse(soap__shelx__run_shelxe)] } {
		if { [string compare $XMLParse(soap__shelx__run_shelxe) "1"] == 0 } {
			set code $XMLParse(dataset_info__code)
			set solvent $XMLParse(dataset_info__solvent_content)
			puts "shelxe $code ${code}_fa -s$solvent -m20"
			puts "shelxe $code ${code}_fa -s$solvent -m20 -i"
		}
	}


}


#
# Main
#
if { [lindex $argv 1] == "" } {
	puts "Usage:"
	puts "  xml2bp3com.tcl input.xml mtz.xml"
	puts " "
	puts " Copyright/GPL Feb 2004 Steven Ness"
	exit
}

#
# Setup variable names
#
set inputxml [lindex $argv 0]
set mtzxml [lindex $argv 1]

#
# Check to make sure files exist
#
if { [file exists $inputxml] == 0 } {
	crank_error "xml2bp3com.tcl::inputxml file does not exist"
}
if { [file exists $mtzxml] == 0 } {
	crank_error "xml2bp3com.tcl::mtzxml file does not exist"
}

#
# Read in XML files
#
XMLParsefile $inputxml
XMLParsefile $mtzxml

#
# Generate .com file
#
OutputShelx

