#!/usr/bin/env tclsh

#
# crank - Automated macromolecular structure solution
#
# Copyright (C) 2004 Steven Ness
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#

#
# xml2solomonpar.tcl
#
# Generate a .par file for the new Solomon script solomon_crank.com
#

#
# Usage:
#  xml2solomonpar.tcl user.xml mtz.xml
#


# Set CRANK path
set crankbin "$env(CRANK)/bin"
source [file join $crankbin crankutils.tcl]

#
# Output a CRUNCH2 newinput file corresponding to the *.user.xml,
# *.mtz.xml and *.coords.xml XML files read in.  
#
proc OutputSOLOMONcomfile { hand } {
	global XMLParse

	#
	# General parameters
	#

	# Prefix of all files produced by this script
	puts "set label='SOL'"

	# File containing MIR data (without .mtz!)
	puts "set data='$XMLParse(dataset_info__code).in'"

	# xyzlim for P1
	puts "set xyzlim='0 1.0 0 1.0 0 1.0'"

	# Resolution to start with
	# If <soap><solomon><start_res> is present, use it,
	# otherwise, use sensible defaults
	if { [info exists XMLParse(soap__solomon__start_res)] } {
		puts "set strtres=$XMLParse(soap__solomon__start_res)"
	} else {
		puts "set strtres=2.5"
	}

	# Resolution to extend to
	if { [info exists XMLParse(soap__solomon__end_res)] } {
		puts "set endres=$XMLParse(soap__solomon__end_res)"
	} else {
		puts "set endres=2.5"
	}

	#
	# LABIN:
	#

	# MTZ-label of Fobs
	if { [info exists XMLParse(soap__solomon__mtz_in__fp)] } {
		puts "set F='$XMLParse(soap__solomon__mtz_in__fp)'"
	} else {
		crank_error "crank::xml2solomonpar.tcl::mtz_in__fp not present"
	}

	# MTZ-label of the standard deviation of Fobs
	if { [info exists XMLParse(soap__solomon__mtz_in__sigfp)] } {
		puts "set SIGF='$XMLParse(soap__solomon__mtz_in__sigfp)'"
	} else {
		crank_error "crank::xml2solomonpar.tcl::mtz_in__sigfp not present"
	}

	# MTZ-label of best phi
	if { [info exists XMLParse(soap__solomon__mtz_in__phib)] } {
		puts "set PHIB='$XMLParse(soap__solomon__mtz_in__phib)'"
	} else {
		crank_error "crank::xml2solomonpar.tcl::mtz_in__phib not present"
	}

	# MTZ-label of figure of merit
	if { [info exists XMLParse(soap__solomon__mtz_in__fomo)] } {
		puts "set FOM='$XMLParse(soap__solomon__mtz_in__fomo)'"
	} else {
		crank_error "crank::xml2solomonpar.tcl::mtz_in__fomo not present"
	}

	# MTZ-labels of Hendrickson Lattman coeffs
	if { [info exists XMLParse(soap__solomon__mtz_in__hla)] } {
		puts "set HLA='$XMLParse(soap__solomon__mtz_in__hla)'"
	} else {
		crank_error "crank::xml2solomonpar.tcl::mtz_in__hla not present"
	}

	if { [info exists XMLParse(soap__solomon__mtz_in__hlb)] } {
		puts "set HLB='$XMLParse(soap__solomon__mtz_in__hlb)'"
	} else {
		crank_error "crank::xml2solomonpar.tcl::mtz_in__hlb not present"
	}

	if { [info exists XMLParse(soap__solomon__mtz_in__hlc)] } {
		puts "set HLC='$XMLParse(soap__solomon__mtz_in__hlc)'"
	} else {
		crank_error "crank::xml2solomonpar.tcl::mtz_in__hlc not present"
	}

	if { [info exists XMLParse(soap__solomon__mtz_in__hld)] } {
		puts "set HLD='$XMLParse(soap__solomon__mtz_in__hld)'"
	} else {
		crank_error "crank::xml2solomonpar.tcl::mtz_in__hld not present"
	}

	#
	# LABOUT:
	#

	# MTZ-label of final FOM-weighted structure factors
	if { [info exists XMLParse(soap__solomon__mtz_out__fdm)] } {
		puts "set FC='$XMLParse(soap__solomon__mtz_out__fdm)'"
	} else {
		crank_error "crank::xml2solomonpar.tcl::soap__solomon_mtz_out__fdm not present"
	}

	# MTZ-label of improved phases
	if { [info exists XMLParse(soap__solomon__mtz_out__phidm)] } {
		puts "set PHIC='$XMLParse(soap__solomon__mtz_out__phidm)'"
	} else {
		crank_error "crank::xml2solomonpar.tcl::soap__solomon_mtz_out__phidm not present"
	}

	#
	# CYCLE calculation:
	#

	# Cycle to start with 
	puts "set LastCycle=0"

	# Number of cycles before extending phases	
	if { [info exists XMLParse(soap__solomon__prep_cycles)] } {
		puts "set PrepCycle=$XMLParse(soap__solomon__prep_cycles)"
	} else {
		puts "set PrepCycle=0"
	}

	# Number of extension cycles
	if { [info exists XMLParse(soap__solomon__ext_cycles)] } {
		puts "set ExtCycle=$XMLParse(soap__solomon__ext_cycles)"
	} else {
		puts "set ExtCycle=0"
	}

	# Number of cycles after extension
	if { [info exists XMLParse(soap__solomon__end_cycles)] } {
		puts "set EndCycle=$XMLParse(soap__solomon__end_cycles)"
	} else {
		puts "set EndCycle=10"
	}

	#
	# Solomon options: Read documentation for optimal values
	#

	# Solvent fraction of the crystal
	if { [info exists XMLParse(dataset_info__solvent_content)] } {
		puts "set slvfrc='$XMLParse(dataset_info__solvent_content)'"
	} else {
		puts "set slvfrc=0.50"
	}

	# The flipping factor
	if { [info exists XMLParse(soap__solomon__flipping_factor)] } {
		puts "set mulsolv='$XMLParse(soap__solomon__flipping_factor)'"
	} else {
		puts "set mulsolv=-1"
	}

	# radius used for determining solvent region
	if { [info exists XMLParse(soap__solomon__radius)] } {
		puts "set radius='$XMLParse(soap__solomon__radius)'"
	} else {
		puts "set radius=3.5"
	}

	# truncation levels
	if { [info exists XMLParse(soap__solomon__trunc_low)] } {
		puts "set trunlo='$XMLParse(soap__solomon__trunc_low)'"
	} else {
		puts "set trunlo=0.2"
	}

	if { [info exists XMLParse(soap__solomon__trunc_high)] } {
		puts "set trunhi='$XMLParse(soap__solomon__trunc_high)'"
	} else {
		puts "set trunhi=1"
	}
}


#
# Main
#
if { [lindex $argv 1] == "" } {
	puts "Usage:"
	puts "  xml2bp3com.tcl input.xml <hand> "
	puts " "
	puts "Where:"
	puts "  input.xml -> input CRANK XML file with dataset info"
	puts "  <hand>    -> 0 for normal hand, 1 for other hand" 
	puts " Copyright/GPL Feb 2004 Steven Ness"
	exit
}

#
# Setup variable names
#
set inputxml [lindex $argv 0]
set hand [lindex $argv 1]

#
# Check to make sure files exist
#
if { [file exists $inputxml] == 0 } {
	crank_error "xml2bp3com.tcl::inputxml file does not exist"
}

#
# Read in XML files
#
XMLParsefile $inputxml

#
# Generate .com file
#
OutputSOLOMONcomfile $hand



