#!/usr/bin/env tclsh

#
# crank - Automated macromolecular structure solution
#
# Copyright (C) 2004 Steven Ness
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#

#
# Tcl program to output the best (or the nth best) trial
# from a *.coords.xml file.
#


# Set some paths and source some important files
# so that we can run CCP4i utility commands outside of the CCP4i
# interface
set system(OPSYS) [string toupper $tcl_platform(platform)]
source [file join $env(CCP4I_TOP) bin $system(OPSYS) startup.tcl]
source [file join $env(CCP4I_TOP) src system.tcl]
source [file join $env(CCP4I_TOP) src utils.tcl]
source [SearchPath TOP utils xml_utils.tcl]


#
# ***************************************************
# XMLParse - coords
#
# Please note, this code has been subtly changed from the
#  regular XMLParse case to handle coordinate files, which
#  have multiple trials and atoms, and are therefore tricky.
#
# *** This code shouldn't be used as a basis for further
# *** development, rather, use the XMLParse constructions
# *** in the other files in this directory.
#
#
# Parse the XML file into the global XMLParse 
# data structure in memory.
#
# Each XML character data tag will be converted into
# an element in this XMLParse associative array, with
# the name being the XML parse tree with "__" between 
# elements, and the value will be the Tcl string trimmed
# value of the character data.
#
# For example, the following XML code:
#
# <start>
#   <tag1>
#     <tag1_1>value1</tag1_1>
#   <tag1>
#   <tag2>value2</tag2>
# <start>
#
# Will be stored in memory as:
#
# XMLParse(start__tag1__tag1_1) == value1
# XMLParse(start__tag2) == value2
#
# ***************************************************
#

#---------------------------------------------------------------------
proc XMLElementStartHandler { name attlist } {
	global xmldata
	global XMLParse
	global XMLWork

	# Each arm of the switch statement deals with a different
	# element in the XML file
	#puts "XML-start=($name)"

	# Add this name to the current "full" name
	set XMLWork(section_name) $name
	lappend XMLWork(section_full_name) $name

	if { $XMLWork(section_name) == "trial" } {
		set section_trial_id [GetXmlAttributeValue $attlist "id"]
		lappend XMLWork(section_full_name) $section_trial_id
		set XMLWork(trial_id) $section_trial_id
		incr XMLWork(n_trials)
# 		if { $section_trial_id > $XMLWork(n_trials) } {
# 			set XMLWork(n_trials) $section_trial_id
# 		}
	}

	set XMLWork(section_cat_name) [join $XMLWork(section_full_name) "__"]
	incr XMLWork(section_depth)

# 	if { $XMLWork(section_name) == "trial" } {
# 		set section_trial_id [GetXmlAttributeValue $attlist "id"]
# 		set XMLWork(section_trial_id) $section_trial_id
# 	}

 	if { $XMLWork(section_name) == "atom" } {
 		set section_atom_id [GetXmlAttributeValue $attlist "id"]
 		set XMLWork(section_atom_id) $section_atom_id
		set XMLWork(natoms_${XMLWork(trial_id)}) $section_atom_id
 	}


}


#---------------------------------------------------------------------
proc XMLElementEndHandler { name } {
	global xmldata
	global XMLParse
	global XMLWork

	# Each arm of the switch statement deals with a different
	# element in the XML file
	#puts "XML-end1=($name) $XMLWork(section_depth) $XMLWork(section_trial_id_depth)"

	if { $name == "trial" } {
		set XMLWork(section_full_name) [lreplace $XMLWork(section_full_name) \
											[expr [llength $XMLWork(section_full_name)] - 2] \
											[expr [llength $XMLWork(section_full_name)] - 1]]
	} else {
	set XMLWork(section_full_name) [lreplace $XMLWork(section_full_name) \
										  [expr [llength $XMLWork(section_full_name)] - 1] \
										  [expr [llength $XMLWork(section_full_name)] - 1]]
	}

	incr XMLWork(section_depth) -1


	#puts "XML-end2=($name) $XMLWork(section_depth) $XMLWork(section_trial_id_depth)"

}

#---------------------------------------------------------------------
proc XMLCharacterHandler { data args } {
	#---------------------------------------------------------------------
	#d_sum Internal parsing procedure for processing CRANK XML files
	#d_desc
	#d_arg name Name of the XML element
	#d_arg attlist Attribute list for that element
	global xmldata
	global XMLParse
	global XMLWork

	# Each arm of the switch statement deals with a different
	# element in the XML file
	set data [string trim $data]

	if { $data != "" } {

		if { $XMLWork(section_name) == "score" } {
			set fullvarname "$XMLWork(section_cat_name)"
		} elseif { $XMLWork(section_name) == "scoretype" } {
			set fullvarname "$XMLWork(section_cat_name)"
		} else {
			set fullvarname "$XMLWork(section_cat_name)__$XMLWork(section_atom_id)"
		}

		set XMLParse($fullvarname) $data

		#puts "section $fullvarname = $data"

	}
}


#---------------------------------------------------------------------
proc XMLParsefile { xmlfile } {
	# Global used by the XML parser
 	global xmldata
	
	# Data structure to hold actual data
	global XMLParse 
	
	# Working data structure to hold temporary work data
	global XMLWork

	set XMLWork(section_depth) 0
	set XMLWork(section_atom_id) ""
	set XMLWork(section_atom_id_depth) 0

	set XMLWork(n_trials) 0

 	# Initialise and configure the XML parser
 	set parser 0
 	set parser [ xml::parser ]
 	$parser configure -elementstartcommand XMLElementStartHandler
 	$parser configure -elementendcommand XMLElementEndHandler
 	$parser configure -characterdatacommand XMLCharacterHandler
 	$parser configure -ignorewhitespace true

 	# Read and parse the specified file
 	if { [ReadFile $xmlfile xmltext] } {
 		$parser parse $xmltext
 		set xmldata(XML_FILE) $xmlfile
 		set xmldata(XML_LOADED) 1
 	} else {
 		# Couldn't load the data
 		set xmldata(XML_LOADED) 0
 	}
}

#
# Output a CRUNCH2 newinput file corresponding to the 
#  *.user.xml and *.mtz.xml XML files read in.
#
# The business end of this program, the rest of the code 
# mainly just reads in and processes the XML.
#
proc Outputbesttrial { nth_best } {
	global XMLParse
	global XMLWork
	
	#
	# Find the nth best trial
	#

	# First, make a list of all the ids and scores
	set curr_trial 0
  	while { $curr_trial < $XMLWork(n_trials) } {
		set tmp [list $curr_trial $XMLParse(coordinate_data__trial__${curr_trial}__score)]
		lappend scores $tmp
		incr curr_trial
	}

# 	puts $scores
	set sorted [lsort -decreasing -index 1 $scores]
# 	puts $sorted
# 	puts [lindex $sorted 0]

	if { $nth_best > $XMLWork(n_trials) } {
		set nth_best $XMLWork(n_trials)
	}
	set nth_best_trial [lindex [lindex $sorted [expr $nth_best - 1]] 0]
# 	puts "Nth Best ($nth_best) trial = $nth_best_trial"
# 	exit

	# Output the best trial
	puts "<coordinate_data>"
	puts "\t<code>$XMLParse(coordinate_data__code__)</code>"
	
	# Output all atoms
	
	for { set j 0 } { $j <= $XMLWork(natoms_${nth_best_trial}) } { incr j } {
		puts "\t<atom id=\"$j\">"
		puts "\t\t<atom_id>$XMLParse(coordinate_data__trial__${nth_best_trial}__atom__atom_id__${j})</atom_id>"
		puts "\t\t<element_name>$XMLParse(coordinate_data__trial__${nth_best_trial}__atom__element_name__${j})</element_name>"
		puts "\t\t<x>$XMLParse(coordinate_data__trial__${nth_best_trial}__atom__x__${j})</x>"
		puts "\t\t<y>$XMLParse(coordinate_data__trial__${nth_best_trial}__atom__y__${j})</y>"
		puts "\t\t<z>$XMLParse(coordinate_data__trial__${nth_best_trial}__atom__z__${j})</z>"
		puts "\t\t<occupancy>$XMLParse(coordinate_data__trial__${nth_best_trial}__atom__occupancy__${j})</occupancy>"
		puts "\t</atom>"
	}

	# Output footer
	puts "\t<score>$XMLParse(coordinate_data__trial__${nth_best_trial}__score)</score>"
	
	if { $XMLParse(coordinate_data__trial__${nth_best_trial}__scoretype) == "bp3_fom"} { 
		set scoretype "bp3_fom"
	} elseif { $XMLParse(coordinate_data__trial__${nth_best_trial}__scoretype) == "crunch2_fom" } {
		set scoretype "crunch2_fom"
	} else {
		puts "Error: <scoretype> not one of bp3_fom or crunch2_fom"
		puts "Exitting"
		exit
	}

	puts "\t<scoretype>$scoretype</scoretype>"
	puts "</coordinate_data>"

}


# 	# Find the nth best trial
# 	set curr_trial 0
# 	set best_score 0.0
# 	set best_trial 0
#  	while { $curr_trial <= $XMLWork(n_trials) } {
# 		if { $XMLParse(coordinate_data__trial__${curr_trial}__score) > $best_score } {
# 			set best_score $XMLParse(coordinate_data__trial__${curr_trial}__score)
# 			set best_trial $curr_trial
# 		}
# 		incr curr_trial
# 	}



#
# Main
#
if { [lindex $argv 0] == "" } {
	puts "Usage:"
	puts "  xmloutputbesttrial.tcl nth_best coords.xml"
	puts " "
	puts " Copyright/GPL Feb 2004 Steven Ness"
	exit
}

set nth_best [lindex $argv 0]
set xmlfile [lindex $argv 1]

# Output the last set of coordinates to a new file
set output [exec xmloutputlastcoordinatedata.tcl $xmlfile > tmp.coords.xml]

XMLParsefile tmp.coords.xml

# set allvars [array get XMLParse]
# for { set i 0 } { $i < [llength $allvars] } { incr i } {
# 	puts -nonewline "[lindex $allvars $i]\t"
# 	incr i
# 	puts "[lindex $allvars $i] "
# }
# exit

Outputbesttrial $nth_best

file delete tmp.coords.xml
